import torch
import torch.nn as nn
import numpy as np
from graph import Graph, MentionGraph
from transformers import XLMRobertaModel
import torch.nn.functional as F
from copy import deepcopy
from collections import defaultdict
from gcn_utils import InstanceGCN

WINDOW_SIZE = 50

NUM_DISTANCES = int(np.sum([1,  # padding token
                            1,  # out of left context
                            1,  # out of right context
                            WINDOW_SIZE,  # left context
                            1,  # trigger word
                            WINDOW_SIZE  # right context
                            ]))


def get_dist_embeds(triggers, batch_size, seq_len, embed_layer):
    padding_id = 0
    out_of_left_id = 1
    out_of_right_id = 2
    with torch.no_grad():
        positions = torch.from_numpy(np.array(range(seq_len)))
        batch_position = positions.expand(batch_size, seq_len)
        batch_distance = batch_position - triggers.unsqueeze(
            1).long().cpu() + WINDOW_SIZE + 3  # 3 = 1 for padding, 1 for out of leftcontext, 1 for out of rightcontext

        out_of_left_mask = (batch_distance < 3).long()
        out_of_right_mask = (batch_distance >= NUM_DISTANCES).long()

        batch_outofleft = torch.ones_like(batch_distance).long() * out_of_left_mask * out_of_left_id
        batch_outofright = torch.ones_like(batch_distance).long() * out_of_right_mask * out_of_right_id

        out_of_context_mask = out_of_left_mask + out_of_right_mask
        out_of_context_mask = (out_of_context_mask != 1).long()
        batch_distance = batch_distance * out_of_context_mask + batch_outofleft + batch_outofright
        batch_distance = batch_distance.cuda()

    dist_reps = embed_layer(batch_distance)  # [batch size, seq len, dist dim]
    return dist_reps


def get_trigger_reps(batch_reps, triggers):
    '''
    :param batch_reps: [batch size, sequence length, rep dim]
    :param triggers: [batch size, ]
    :return: reps of anchor words: [batch size, rep dim]
    '''
    batch_reps = batch_reps.clone()
    ids = triggers.view(-1, 1).long()
    ids = ids.expand(ids.size(0), batch_reps.size(2)).unsqueeze(1)

    trigger_reps = torch.gather(batch_reps, 1, ids)
    trigger_reps = trigger_reps.squeeze(1)
    return trigger_reps


class Tree(object):
    """
    Reused tree object from stanfordnlp/treelstm.
    """

    def __init__(self):
        self.parent = None
        self.num_children = 0
        self.children = list()

    def add_child(self, child):
        child.parent = self
        self.num_children += 1
        self.children.append(child)

    def size(self):
        if getattr(self, '_size'):
            return self._size
        count = 1
        for i in range(self.num_children):
            count += self.children[i].size()
        self._size = count
        return self._size

    def depth(self):
        if getattr(self, '_depth'):
            return self._depth
        count = 0
        if self.num_children > 0:
            for i in range(self.num_children):
                child_depth = self.children[i].depth()
                if child_depth > count:
                    count = child_depth
            count += 1
        self._depth = count
        return self._depth

    def __iter__(self):
        yield self
        for c in self.children:
            for x in c:
                yield x


def get_head_token(head_id):
    return head_id - 1


def get_full_adj(head_ids, self_loop=True):
    batch_size, seq_len = head_ids.shape
    head_ids = head_ids.data.cpu().numpy()
    adj_list = []
    for b_id in range(batch_size):
        adj = np.zeros((seq_len, seq_len))
        for i in range(seq_len):
            adj[i][i] = 1 if self_loop else 0
            head = int(head_ids[b_id][i] - 1)
            if head >= 0:
                adj[i][head] = 1
                adj[head][i] = 1
        adj_list.append(adj)
    adj_list = np.array(adj_list)
    return torch.from_numpy(adj_list).float().cuda()


def get_pruned_adj(batch_head_ids, batch_token_nums, batch_subj_positions, batch_obj_positions, prune):
    def head_to_tree(head, actual_len, prune, subj_pos, obj_pos):
        """
        Convert a sequence of head indexes into a tree object.
        """
        head = head[:actual_len].tolist()
        root = None

        if prune < 0:
            nodes = [Tree() for _ in head]

            for i in range(len(nodes)):
                h = head[i]
                nodes[i].idx = i
                nodes[i].dist = -1  # just a filler
                if h == 0:
                    root = nodes[i]
                else:
                    nodes[get_head_token(head_id=h)].add_child(nodes[i])
        else:
            # find dependency path
            subj_pos = [i for i in range(actual_len) if subj_pos[i] == 0]
            obj_pos = [i for i in range(actual_len) if obj_pos[i] == 0]

            common_ancestors = None

            subj_ancestors = set(subj_pos)
            for s in subj_pos:
                h = head[s]
                tmp = [s]
                while h > 0:
                    head_token = get_head_token(head_id=h)
                    tmp += [head_token]
                    subj_ancestors.add(head_token)
                    h = head[head_token]  # get parent node

                if common_ancestors is None:
                    common_ancestors = set(tmp)
                else:
                    common_ancestors.intersection_update(tmp)  # apply join operation between two sets

            obj_ancestors = set(obj_pos)
            for o in obj_pos:
                h = head[o]
                tmp = [o]
                while h > 0:
                    head_token = get_head_token(head_id=h)
                    tmp += [head_token]
                    obj_ancestors.add(head_token)
                    h = head[head_token]
                common_ancestors.intersection_update(tmp)

            # find lowest common ancestor
            if len(common_ancestors) == 1:
                lca = list(common_ancestors)[0]
            else:
                child_count = {k: 0 for k in common_ancestors}
                for ca in common_ancestors:
                    head_of_ca = head[ca]
                    head_token_of_ca = get_head_token(head_id=head_of_ca)
                    if head_of_ca > 0 and head_token_of_ca in common_ancestors:
                        child_count[head_token_of_ca] += 1

                # the LCA has no child in the CA set
                for ca in common_ancestors:
                    if child_count[ca] == 0:
                        lca = ca
                        break

            path_nodes = subj_ancestors.union(obj_ancestors).difference(common_ancestors)
            path_nodes.add(lca)

            # compute distance to path_nodes
            dist = [-1 if i not in path_nodes else 0 for i in range(actual_len)]

            for i in range(actual_len):
                if dist[i] < 0:
                    stack = [i]
                    while stack[-1] >= 0 and stack[-1] not in path_nodes:
                        stack.append(get_head_token(head[stack[-1]]))

                    if stack[-1] in path_nodes:
                        for d, j in enumerate(reversed(stack)):
                            dist[j] = d
                    else:
                        for j in stack:
                            if j >= 0 and dist[j] < 0:
                                dist[j] = int(1e4)  # aka infinity

            highest_node = lca
            nodes = [Tree() if dist[i] <= prune else None for i in range(actual_len)]

            for i in range(len(nodes)):
                if nodes[i] is None:
                    continue
                h = head[i]
                nodes[i].idx = i
                nodes[i].dist = dist[i]
                if h > 0 and i != highest_node:
                    head_token = get_head_token(head_id=h)
                    assert nodes[head_token] is not None
                    nodes[head_token].add_child(nodes[i])

            root = nodes[highest_node]

        assert root is not None
        return root

    def tree_to_adj(sent_len, tree, directed=True, self_loop=False):
        """
        Convert a tree object to an (numpy) adjacency matrix.
        """
        ret = np.zeros((sent_len, sent_len), dtype=np.float32)

        queue = [tree]
        idx = []
        while len(queue) > 0:
            t, queue = queue[0], queue[1:]

            idx += [t.idx]

            for c in t.children:
                ret[t.idx, c.idx] = 1
            queue += t.children

        if not directed:
            ret = ret + ret.T

        if self_loop:
            for i in idx:
                ret[i, i] = 1

        return ret

    maxlen = max(batch_token_nums)
    batch_head_ids, batch_subj_positions, batch_obj_positions = batch_head_ids.cpu().numpy(), batch_subj_positions.cpu().numpy(), batch_obj_positions.cpu().numpy()
    trees = [head_to_tree(batch_head_ids[i], batch_token_nums[i], prune, batch_subj_positions[i],
                          batch_obj_positions[i]) for i in range(len(batch_token_nums))]
    adj = [tree_to_adj(maxlen, tree, directed=False, self_loop=False).reshape(1, maxlen, maxlen) for tree in
           trees]
    adj = np.concatenate(adj, axis=0)
    adj = torch.cuda.FloatTensor(adj)  # [batch size, max len, max len]
    return adj


class GCN(nn.Module):

    def __init__(self, in_dim, hidden_dim, num_layers):
        super(GCN, self).__init__()
        self.num_layers = num_layers
        self.hidden_dim = hidden_dim
        self.in_dim = in_dim
        self.gcn_drop = nn.Dropout(0.5)

        # gcn layer
        self.W = nn.ModuleList()

        for layer in range(self.num_layers):
            input_dim = self.in_dim if layer == 0 else self.hidden_dim
            self.W.append(nn.Linear(input_dim, self.hidden_dim))

    def forward(self, gcn_inputs, head_ids):
        adj = get_full_adj(head_ids)
        # gcn layer
        denom = adj.sum(2).unsqueeze(2) + 1
        pool_mask = (adj.sum(2) + adj.sum(1)).eq(0)

        for l in range(self.num_layers):
            Ax = adj.bmm(gcn_inputs)
            AxW = self.W[l](Ax)
            AxW = AxW + self.W[l](gcn_inputs)  # self loop
            AxW = AxW / denom

            gAxW = F.relu(AxW)
            gcn_inputs = self.gcn_drop(gAxW) if l < self.num_layers - 1 else gAxW

        return gcn_inputs, pool_mask


def compute_binary_reprs(obj1_reprs, obj2_reprs):  # note that, (obj1, obj2) != (obj2, obj1)
    batch_size, _, rep_dim = obj1_reprs.shape
    num_obj1 = obj1_reprs.shape[1]
    num_obj2 = obj2_reprs.shape[1]

    cloned_obj1s = obj1_reprs.repeat(1, 1, num_obj2).view(batch_size, -1, rep_dim)
    cloned_obj2s = obj2_reprs.repeat(1, num_obj1, 1).view(batch_size, -1, rep_dim)
    return cloned_obj1s, cloned_obj2s


def compute_span_reprs(word_reprs, span_idxs):
    '''
    word_reprs.shape: [batch size, num words, word dim]
    span_idxs.shape: [batch size, num spans, 2]
    '''
    batch_span_reprs = []
    batch_size, _, _ = word_reprs.shape
    _, num_spans, _ = span_idxs.shape
    for bid in range(batch_size):
        span_reprs = []
        for sid in range(num_spans):
            start, end = span_idxs[bid][sid]
            words = word_reprs[bid][start: end]  # [span size, word dim]
            span_reprs.append(torch.mean(words, dim=0))
        span_reprs = torch.stack(span_reprs, dim=0)  # [num spans, word dim]
        batch_span_reprs.append(span_reprs)
    batch_span_reprs = torch.stack(batch_span_reprs, dim=0)  # [batch size, num spans, word dim]
    return batch_span_reprs


def compute_word_reprs_first(piece_reprs, first_idxs):
    '''
    piece_reprs.shape: [batch size, max length, rep dim]
    first_idxs.shape: [batch size, num words]
    '''
    batch_word_reprs = []
    batch_size, _, _ = piece_reprs.shape
    _, num_words = first_idxs.shape
    for bid in range(batch_size):
        word_reprs = []
        for wid in range(num_words):
            word_reprs.append(piece_reprs[bid][first_idxs[bid][wid]])
        word_reprs = torch.stack(word_reprs, dim=0)  # [ num words, rep dim]
        batch_word_reprs.append(word_reprs)
    batch_word_reprs = torch.stack(batch_word_reprs, dim=0)  # [batch size, num words, rep dim]
    return batch_word_reprs


def compute_word_reps_avg(piece_reprs, component_idxs):
    '''
        piece_reprs.shape: [batch size, max length, rep dim]
        component_idxs.shape: [batch size, num words, 2]
    '''
    batch_word_reprs = []
    batch_size, _, _ = piece_reprs.shape
    _, num_words, _ = component_idxs.shape
    for bid in range(batch_size):
        word_reprs = []
        for wid in range(num_words):
            wrep = torch.mean(piece_reprs[bid][component_idxs[bid][wid][0]: component_idxs[bid][wid][1]], dim=0)
            word_reprs.append(wrep)
        word_reprs = torch.stack(word_reprs, dim=0)  # [num words, rep dim]
        batch_word_reprs.append(word_reprs)
    batch_word_reprs = torch.stack(batch_word_reprs, dim=0)  # [batch size, num words, rep dim]
    return batch_word_reprs


def log_sum_exp(tensor, dim=0, keepdim: bool = False):
    """LogSumExp operation used by CRF."""
    m, _ = tensor.max(dim, keepdim=keepdim)
    if keepdim:
        stable_vec = tensor - m
    else:
        stable_vec = tensor - m.unsqueeze(dim)
    return m + (stable_vec.exp().sum(dim, keepdim=keepdim)).log()


def sequence_mask(lens, max_len=None):
    """Generate a sequence mask tensor from sequence lengths, used by CRF."""
    batch_size = lens.size(0)
    if max_len is None:
        max_len = lens.max().item()
    ranges = torch.arange(0, max_len, device=lens.device).long()
    ranges = ranges.unsqueeze(0).expand(batch_size, max_len)
    lens_exp = lens.unsqueeze(1).expand_as(ranges)
    mask = ranges < lens_exp
    return mask


def token_lens_to_offsets(token_lens):
    """Map token lengths to first word piece indices, used by the sentence
    encoder.
    :param token_lens (list): token lengths (word piece numbers)
    :return (list): first word piece indices (offsets)
    """
    max_token_num = max([len(x) for x in token_lens])
    offsets = []
    for seq_token_lens in token_lens:
        seq_offsets = [0]
        for l in seq_token_lens[:-1]:
            seq_offsets.append(seq_offsets[-1] + l)
        offsets.append(seq_offsets + [-1] * (max_token_num - len(seq_offsets)))
    return offsets


def token_lens_to_idxs(token_lens, start_tokens, end_tokens):
    batch_size = len(start_tokens)
    max_token_num = max([end_tokens[i] - start_tokens[i] for i in range(batch_size)])
    max_token_len = max(
        [max([tlen for tid, tlen in enumerate(token_lens[i]) if tid >= start_tokens[i] and tid < end_tokens[i]]) for i
         in range(batch_size)])
    idxs = []
    for sid, seq_token_lens in enumerate(token_lens):
        seq_idxs = []
        offset = 0
        for tid, token_len in enumerate(seq_token_lens):
            if tid >= start_tokens[sid] and tid < end_tokens[sid]:
                seq_idxs.append([offset, offset + token_len])
            offset += token_len
        seq_idxs.extend([[-1, 0]] * (max_token_num - len(seq_idxs)))
        idxs.append(seq_idxs)
    return idxs, max_token_num, max_token_len


def graphs_to_node_idxs(batch, graphs):
    """
    :param graphs (list): A list of Graph objects.
    :return: mention/trigger index matrix, mask tensor, max number, and max length
    """
    mention_idxs = []
    trigger_idxs = []

    max_mention_num = max(max(graph.mention_num for graph in graphs), 1)
    max_trigger_num = max(max(graph.trigger_num for graph in graphs), 1)
    max_mention_len = max(max([e[1] - e[0] for e in graph.mentions] + [1])
                          for graph in graphs)
    max_trigger_len = max(max([t[1] - t[0] for t in graph.triggers] + [1])
                          for graph in graphs)
    for bid, graph in enumerate(graphs):
        seq_mention_idxs = []
        seq_trigger_idxs = []

        for mention in graph.mentions:
            mention_len = mention[1] - mention[0]
            seq_mention_idxs.append([mention[0], mention[1]])
        seq_mention_idxs.extend([[0, 1]] * (max_mention_num - graph.mention_num))
        mention_idxs.append(seq_mention_idxs)

        for trigger in graph.triggers:
            seq_trigger_idxs.append([trigger[0], trigger[1]])

        seq_trigger_idxs.extend([[0, 1]] * (max_trigger_num - graph.trigger_num))

        trigger_idxs.append(seq_trigger_idxs)

    return (
        mention_idxs, max_mention_num, max_mention_len,
        trigger_idxs, max_trigger_num, max_trigger_len
    )


def graphs_to_node_idxs_outer_argument(batch, graphs):
    """
    :param graphs (list): A list of Graph objects.
    :return: mention/trigger index matrix, mask tensor, max number, and max length
    """
    mention_idxs = []
    argument_idxs = []
    trigger_idxs = []

    max_mention_num = max(max(graph.mention_num for graph in graphs), 1)
    max_argument_num = max(max(graph.argument_num for graph in graphs), 1)
    max_trigger_num = max(max(graph.trigger_num for graph in graphs), 1)
    max_mention_len = max(max([e[1] - e[0] for e in graph.mentions] + [1])
                          for graph in graphs)
    max_argument_len = max(max([a[1] - a[0] for a in graph.arguments] + [1]) for graph in graphs)
    max_trigger_len = max(max([t[1] - t[0] for t in graph.triggers] + [1])
                          for graph in graphs)
    for bid, graph in enumerate(graphs):
        seq_mention_idxs = []
        seq_argument_idxs = []
        seq_trigger_idxs = []

        for mention in graph.mentions:
            mention_len = mention[1] - mention[0]
            seq_mention_idxs.append([mention[0], mention[1]])
        seq_mention_idxs.extend([[0, 1]] * (max_mention_num - graph.mention_num))
        mention_idxs.append(seq_mention_idxs)

        for argument in graph.arguments:
            argument_len = argument[1] - argument[0]
            seq_argument_idxs.append([argument[0], argument[1]])
        seq_argument_idxs.extend([[0, 1]] * (max_argument_num - graph.argument_num))
        argument_idxs.append(seq_argument_idxs)

        for trigger in graph.triggers:
            seq_trigger_idxs.append([trigger[0], trigger[1]])

        seq_trigger_idxs.extend([[0, 1]] * (max_trigger_num - graph.trigger_num))

        trigger_idxs.append(seq_trigger_idxs)

    return (
        mention_idxs, max_mention_num, max_mention_len,
        argument_idxs, max_argument_num, max_argument_len,
        trigger_idxs, max_trigger_num, max_trigger_len
    )


def graphs_to_node_idxs_doc(batch, graphs):
    """
    :param graphs (list): A list of Graph objects.
    :return: mention/trigger index matrix, mask tensor, max number, and max length
    """
    mention_idxs = []

    max_mention_num = max(max(graph.mention_num for graph in graphs), 1)
    max_mention_len = max(max([e[1] - e[0] for e in graph.mentions] + [1])
                          for graph in graphs)

    for bid, graph in enumerate(graphs):
        seq_mention_idxs = []

        for mention in graph.mentions:
            seq_mention_idxs.append([mention[0], mention[1]])
        seq_mention_idxs.extend([[0, 1]] * (max_mention_num - graph.mention_num))
        mention_idxs.append(seq_mention_idxs)

    return (
        mention_idxs, max_mention_num, max_mention_len
    )


def generate_pairwise_idxs(num1, num2):
    idxs = []
    distance_idxs = []
    for i in range(num1):
        for j in range(num2):
            idxs.append(i)
            idxs.append(j + num1)
            distance_idxs.append(np.abs(i - j))
    return idxs, distance_idxs


def trigger_tag_paths_to_spans(paths, token_nums, vocab):
    """Convert predicted tag paths to a list of spans (entity mentions or event
    triggers).
    :param paths: predicted tag paths.
    :return (list): a list (batch) of lists (sequence) of spans.
    """
    batch_mentions = []
    itos = {i: s for s, i in vocab.items()}
    for i, path in enumerate(paths):
        mentions = []
        cur_mention = None
        path = path.tolist()[:token_nums[i].item()]
        for j, tag in enumerate(path):
            tag = itos[tag]
            if tag == 'O':
                prefix = tag = 'O'
            else:
                prefix, tag = tag.split('-', 1)

            if prefix == 'B':
                if cur_mention:
                    mentions.append(cur_mention)
                cur_mention = [j, j + 1, tag]
            elif prefix == 'I':
                if cur_mention is None:
                    # treat it as B-*
                    cur_mention = [j, j + 1, tag]
                elif cur_mention[-1] == tag:
                    cur_mention[1] = j + 1
                else:
                    # treat it as B-*
                    mentions.append(cur_mention)
                    cur_mention = [j, j + 1, tag]
            else:
                if cur_mention:
                    mentions.append(cur_mention)
                cur_mention = None
        if cur_mention:
            mentions.append(cur_mention)
        mentions = [(i, j, tag) for i, j, tag in mentions]
        batch_mentions.append(mentions)
    return batch_mentions


def mention_tag_paths_to_spans(paths, token_nums, vocab):
    """Convert predicted tag paths to a list of spans (entity mentions or event
    triggers).
    :param paths: predicted tag paths.
    :return (list): a list (batch) of lists (sequence) of spans.
    """
    batch_mentions = []
    itos = {i: s for s, i in vocab.items()}
    for i, path in enumerate(paths):
        mentions = []
        cur_mention = None
        path = path.tolist()[:token_nums[i].item()]
        for j, tag in enumerate(path):
            tag = itos[tag]
            if tag == 'O':
                prefix = tag = 'O'
            else:
                prefix, tag = tag.split('-', 1)

            if prefix == 'B':
                if cur_mention:
                    mentions.append(cur_mention)
                cur_mention = [j, j + 1, tag]
            elif prefix == 'I':
                if cur_mention is None:
                    # treat it as B-*
                    cur_mention = [j, j + 1, tag]
                elif cur_mention[-1] == tag:
                    cur_mention[1] = j + 1
                else:
                    # treat it as B-*
                    mentions.append(cur_mention)
                    cur_mention = [j, j + 1, tag]
            else:
                if cur_mention:
                    mentions.append(cur_mention)
                cur_mention = None
        if cur_mention:
            mentions.append(cur_mention)
        mentions = [(i, j, tag) for i, j, tag in mentions]
        batch_mentions.append(mentions)
    return batch_mentions


def get_sentence_distances(trigger_sent_ids, graphs, max_sentence_distance):
    max_trigger_num = len(trigger_sent_ids[0])
    dist_ids = []
    for bid, graph in enumerate(graphs):
        distances = []
        for i in range(graph.trigger_num):
            for j in range(graph.trigger_num):
                abs_dist = np.abs(trigger_sent_ids[bid][i] - trigger_sent_ids[bid][j])
                abs_dist = min(max_sentence_distance, abs_dist)
                distances.append(abs_dist)
            distances.extend([max_sentence_distance] * (max_trigger_num - graph.trigger_num))
        distances.extend([max_sentence_distance] * max_trigger_num * (max_trigger_num - graph.trigger_num))
        dist_ids.append(distances)
    return dist_ids


class Linears(nn.Module):
    """Multiple linear layers with Dropout."""

    def __init__(self, dimensions, activation='relu', dropout_prob=0.0, bias=True):
        super().__init__()
        assert len(dimensions) > 1
        self.layers = nn.ModuleList([nn.Linear(dimensions[i], dimensions[i + 1], bias=bias)
                                     for i in range(len(dimensions) - 1)])
        self.activation = getattr(torch, activation)
        self.dropout = nn.Dropout(dropout_prob)

    def forward(self, inputs):
        for i, layer in enumerate(self.layers):
            if i > 0:
                inputs = self.activation(inputs)
                inputs = self.dropout(inputs)
            inputs = layer(inputs)
        return inputs


class CRF(nn.Module):
    def __init__(self, label_vocab, bioes=False):
        super(CRF, self).__init__()

        self.label_vocab = label_vocab
        self.label_size = len(label_vocab) + 2
        # self.same_type = self.map_same_types()
        self.bioes = bioes

        self.start = self.label_size - 2
        self.end = self.label_size - 1
        transition = torch.randn(self.label_size, self.label_size)
        self.transition = nn.Parameter(transition)
        self.initialize()

    def initialize(self):
        self.transition.data[:, self.end] = -100.0
        self.transition.data[self.start, :] = -100.0

        for label, label_idx in self.label_vocab.items():
            if label.startswith('I-') or label.startswith('E-'):
                self.transition.data[label_idx, self.start] = -100.0
            if label.startswith('B-') or label.startswith('I-'):
                self.transition.data[self.end, label_idx] = -100.0

        for label_from, label_from_idx in self.label_vocab.items():
            if label_from == 'O':
                label_from_prefix, label_from_type = 'O', 'O'
            else:
                label_from_prefix, label_from_type = label_from.split('-', 1)

            for label_to, label_to_idx in self.label_vocab.items():
                if label_to == 'O':
                    label_to_prefix, label_to_type = 'O', 'O'
                else:
                    label_to_prefix, label_to_type = label_to.split('-', 1)

                if self.bioes:
                    is_allowed = any(
                        [
                            label_from_prefix in ['O', 'E', 'S']
                            and label_to_prefix in ['O', 'B', 'S'],

                            label_from_prefix in ['B', 'I']
                            and label_to_prefix in ['I', 'E']
                            and label_from_type == label_to_type
                        ]
                    )
                else:
                    is_allowed = any(
                        [
                            label_to_prefix in ['B', 'O'],

                            label_from_prefix in ['B', 'I']
                            and label_to_prefix == 'I'
                            and label_from_type == label_to_type
                        ]
                    )
                if not is_allowed:
                    self.transition.data[
                        label_to_idx, label_from_idx] = -100.0

    def pad_logits(self, logits):
        """Pad the linear layer output with <SOS> and <EOS> scores.
        :param logits: Linear layer output (no non-linear function).
        """
        batch_size, seq_len, _ = logits.size()
        pads = logits.new_full((batch_size, seq_len, 2), -100.0,
                               requires_grad=False)
        logits = torch.cat([logits, pads], dim=2)
        return logits

    def calc_binary_score(self, labels, lens):
        batch_size, seq_len = labels.size()

        # A tensor of size batch_size * (seq_len + 2)
        labels_ext = labels.new_empty((batch_size, seq_len + 2))
        labels_ext[:, 0] = self.start
        labels_ext[:, 1:-1] = labels
        mask = sequence_mask(lens + 1, max_len=(seq_len + 2)).long()
        pad_stop = labels.new_full((1,), self.end, requires_grad=False)
        pad_stop = pad_stop.unsqueeze(-1).expand(batch_size, seq_len + 2)
        labels_ext = (1 - mask) * pad_stop + mask * labels_ext
        labels = labels_ext

        trn = self.transition
        trn_exp = trn.unsqueeze(0).expand(batch_size, self.label_size,
                                          self.label_size)
        lbl_r = labels[:, 1:]
        lbl_rexp = lbl_r.unsqueeze(-1).expand(*lbl_r.size(), self.label_size)
        # score of jumping to a tag
        trn_row = torch.gather(trn_exp, 1, lbl_rexp)

        lbl_lexp = labels[:, :-1].unsqueeze(-1)
        trn_scr = torch.gather(trn_row, 2, lbl_lexp)
        trn_scr = trn_scr.squeeze(-1)

        mask = sequence_mask(lens + 1).float()
        trn_scr = trn_scr * mask
        score = trn_scr

        return score

    def calc_unary_score(self, logits, labels, lens):
        """Checked"""
        labels_exp = labels.unsqueeze(-1)
        scores = torch.gather(logits, 2, labels_exp).squeeze(-1)
        mask = sequence_mask(lens).float()
        scores = scores * mask
        return scores

    def calc_gold_score(self, logits, labels, lens):
        """Checked"""
        unary_score = self.calc_unary_score(logits, labels, lens).sum(
            1).squeeze(-1)
        binary_score = self.calc_binary_score(labels, lens).sum(1).squeeze(-1)
        return unary_score + binary_score

    def calc_norm_score(self, logits, lens):
        batch_size, _, _ = logits.size()
        alpha = logits.new_full((batch_size, self.label_size), -100.0)
        alpha[:, self.start] = 0
        lens_ = lens.clone()

        logits_t = logits.transpose(1, 0)
        for logit in logits_t:
            logit_exp = logit.unsqueeze(-1).expand(batch_size,
                                                   self.label_size,
                                                   self.label_size)
            alpha_exp = alpha.unsqueeze(1).expand(batch_size,
                                                  self.label_size,
                                                  self.label_size)
            trans_exp = self.transition.unsqueeze(0).expand_as(alpha_exp)
            mat = logit_exp + alpha_exp + trans_exp
            alpha_nxt = log_sum_exp(mat, 2).squeeze(-1)

            mask = (lens_ > 0).float().unsqueeze(-1).expand_as(alpha)
            alpha = mask * alpha_nxt + (1 - mask) * alpha
            lens_ = lens_ - 1

        alpha = alpha + self.transition[self.end].unsqueeze(0).expand_as(alpha)
        norm = log_sum_exp(alpha, 1).squeeze(-1)

        return norm

    def loglik(self, logits, labels, lens):
        norm_score = self.calc_norm_score(logits, lens)
        gold_score = self.calc_gold_score(logits, labels, lens)
        return gold_score - norm_score

    def viterbi_decode(self, logits, lens):
        """Borrowed from pytorch tutorial
        Arguments:
            logits: [batch_size, seq_len, n_labels] FloatTensor
            lens: [batch_size] LongTensor
        """
        batch_size, _, n_labels = logits.size()
        vit = logits.new_full((batch_size, self.label_size), -100.0)
        vit[:, self.start] = 0
        c_lens = lens.clone()

        logits_t = logits.transpose(1, 0)
        pointers = []
        for logit in logits_t:
            vit_exp = vit.unsqueeze(1).expand(batch_size, n_labels, n_labels)
            trn_exp = self.transition.unsqueeze(0).expand_as(vit_exp)
            vit_trn_sum = vit_exp + trn_exp
            vt_max, vt_argmax = vit_trn_sum.max(2)

            vt_max = vt_max.squeeze(-1)
            vit_nxt = vt_max + logit
            pointers.append(vt_argmax.squeeze(-1).unsqueeze(0))

            mask = (c_lens > 0).float().unsqueeze(-1).expand_as(vit_nxt)
            vit = mask * vit_nxt + (1 - mask) * vit

            mask = (c_lens == 1).float().unsqueeze(-1).expand_as(vit_nxt)
            vit += mask * self.transition[self.end].unsqueeze(
                0).expand_as(vit_nxt)

            c_lens = c_lens - 1

        pointers = torch.cat(pointers)
        scores, idx = vit.max(1)
        paths = [idx.unsqueeze(1)]
        for argmax in reversed(pointers):
            idx_exp = idx.unsqueeze(-1)
            idx = torch.gather(argmax, 1, idx_exp)
            idx = idx.squeeze(-1)

            paths.insert(0, idx.unsqueeze(1))

        paths = torch.cat(paths[1:], 1)
        scores = scores.squeeze(-1)

        return scores, paths

    def calc_conf_score_(self, logits, labels):
        batch_size, _, _ = logits.size()

        logits_t = logits.transpose(1, 0)
        scores = [[] for _ in range(batch_size)]
        pre_labels = [self.start] * batch_size
        for i, logit in enumerate(logits_t):
            logit_exp = logit.unsqueeze(-1).expand(batch_size,
                                                   self.label_size,
                                                   self.label_size)
            trans_exp = self.transition.unsqueeze(0).expand(batch_size,
                                                            self.label_size,
                                                            self.label_size)
            score = logit_exp + trans_exp
            score = score.view(-1, self.label_size * self.label_size) \
                .softmax(1)
            for j in range(batch_size):
                cur_label = labels[j][i]
                cur_score = score[j][cur_label * self.label_size + pre_labels[j]]
                scores[j].append(cur_score)
                pre_labels[j] = cur_label
        return scores


class TriggerModel(nn.Module):
    def __init__(self,
                 config,
                 vocabs):
        super().__init__()
        self.config = config
        # vocabularies
        self.vocabs = vocabs

        # XLMR encoder
        self.xlmr_dim = 1024 if 'large' in config.xlmr_model_name else 768
        self.xlmr = XLMRobertaModel.from_pretrained(config.xlmr_model_name,
                                                    cache_dir=config.cache_dir,
                                                    output_hidden_states=True)
        self.xlmr_dropout = nn.Dropout(p=config.xlmr_dropout)

        # ###################### basic events #######################

        self.trigger_label_stoi = vocabs['trigger_label']  # BIO tags for ['Communicate-Event', 'Cull-Livestock', ...]
        self.trigger_label_itos = {i: s for s, i in self.trigger_label_stoi.items()}
        self.trigger_label_num = len(self.trigger_label_stoi)

        self.trigger_label_ffn = Linears([self.xlmr_dim, config.event_hidden_num,
                                          self.trigger_label_num],
                                         dropout_prob=config.linear_dropout,
                                         bias=config.linear_bias,
                                         activation=config.linear_activation)

        self.trigger_crf = CRF(self.trigger_label_stoi, bioes=False)

    def encode(self, piece_idxs, attention_masks, token_lens, start_tokens, end_tokens):
        batch_size, _ = piece_idxs.size()
        all_xlmr_outputs = self.xlmr(piece_idxs, attention_mask=attention_masks)
        xlmr_outputs = all_xlmr_outputs[0]

        idxs, token_num, token_len = token_lens_to_idxs(token_lens, start_tokens, end_tokens)
        idxs = piece_idxs.new(idxs) + 1
        xlmr_outputs = compute_word_reps_avg(xlmr_outputs, idxs)

        xlmr_outputs = self.xlmr_dropout(xlmr_outputs)
        return xlmr_outputs

    def forward(self, batch):
        # encoding
        word_reprs = self.encode(batch.piece_idxs,
                                 batch.attention_masks,
                                 batch.token_lens, batch.start_tokens, batch.end_tokens)
        batch_size, _, _ = word_reprs.size()

        # basic trigger detection
        trigger_label_scores = self.trigger_label_ffn(word_reprs)
        trigger_label_scores = self.trigger_crf.pad_logits(trigger_label_scores)
        trigger_label_loglik = self.trigger_crf.loglik(trigger_label_scores,
                                                       batch.trigger_label_idxs,
                                                       batch.token_nums).unsqueeze(-1) * batch.weights.unsqueeze(-1)

        loss = - trigger_label_loglik.mean()

        return loss

    def predict(self, batch):
        word_reprs = self.encode(batch.piece_idxs,
                                 batch.attention_masks,
                                 batch.token_lens, batch.start_tokens, batch.end_tokens)
        batch_size, _, _ = word_reprs.size()

        # basic trigger detection
        trigger_label_scores = self.trigger_label_ffn(word_reprs) # [batch size, max words, num bio labels]
        word_probs = torch.softmax(trigger_label_scores, dim=-1)
        trigger_label_scores = self.trigger_crf.pad_logits(trigger_label_scores)
        _, trigger_label_preds = self.trigger_crf.viterbi_decode(trigger_label_scores, batch.token_nums)
        batch_triggers = trigger_tag_paths_to_spans(trigger_label_preds,
                                                    batch.token_nums,
                                                    self.trigger_label_stoi) # [batch size, num triggers, 3]

        # ----------------------
        # add confidence scores
        word_probs = word_probs.data.cpu().numpy().tolist()
        new_batch_triggers = []
        for bid in range(len(batch_triggers)):
            num_triggers = len(batch_triggers[bid])
            new_triggers = []
            for i in range(num_triggers):
                start_token, end_token, event_type = batch_triggers[bid][i]
                score = word_probs[bid][start_token][self.trigger_label_stoi['B-{}'.format(event_type)]]
                new_triggers.append([start_token, end_token, event_type, score])

            new_batch_triggers.append(new_triggers)

        batch_triggers = new_batch_triggers
        # ----------------------

        batch_graphs = [Graph(did, eid, sid, doc_txt, toks, [], tlist, [], self.vocabs, start, end)
                        for did, eid, sid, doc_txt, toks, tlist, start, end in
                        zip(batch.doc_ids, batch.entry_ids, batch.sent_ids, batch.doc_texts, batch.tokens,
                            batch_triggers, batch.start, batch.end)]
        return batch_graphs


class ArgumentModel(nn.Module):
    def __init__(self,
                 config,
                 vocabs):
        super().__init__()
        self.config = config
        # vocabularies
        self.vocabs = vocabs

        self.trigger_label_stoi = vocabs['trigger_label']  # BIO tags for ['Communicate-Event', 'Cull-Livestock', ...]
        self.trigger_label_itos = {i: s for s, i in self.trigger_label_stoi.items()}
        self.trigger_label_num = len(self.trigger_label_stoi)

        self.event_type_stoi = vocabs['event_type']  # ['Communicate-Event', 'Cull-Livestock', ..., 'Other']
        self.event_type_itos = {i: s for s, i in self.event_type_stoi.items()}
        self.event_type_num = len(self.event_type_stoi)

        self.mention_label_stoi = vocabs['mention_label']  # name, nominal, pronoun, anchor, time
        self.mention_label_itos = {i: s for s, i in self.mention_label_stoi.items()}
        self.mention_label_num = len(self.mention_label_stoi)

        self.role_type_stoi = vocabs['role_type']  # {'O', 'agents', 'patients', 'ref-events', 'money', 'time-attachments'}
        self.role_type_itos = {i: s for s, i in self.role_type_stoi.items()}
        self.role_type_num = len(self.role_type_stoi)

        # XLMR encoder
        self.xlmr_dim = 1024 if 'large' in config.xlmr_model_name else 768
        self.xlmr = XLMRobertaModel.from_pretrained(config.xlmr_model_name,
                                                    cache_dir=config.cache_dir,
                                                    output_hidden_states=True)
        self.xlmr_dropout = nn.Dropout(p=config.xlmr_dropout)

        self.upos_embedding = nn.Embedding(
            num_embeddings=len(vocabs['upos']),
            embedding_dim=self.config.feat_dim,
            padding_idx=0
        )

        self.dist_embedding = nn.Embedding(
            num_embeddings=NUM_DISTANCES,
            embedding_dim=self.config.feat_dim,
            padding_idx=0
        )

        self.deprel_embedding = nn.Embedding(
            num_embeddings=len(vocabs['deprel']),
            embedding_dim=self.config.feat_dim,
            padding_idx=0
        )

        self.gcn_layer = GCN(
            in_dim=self.config.feat_dim,
            hidden_dim=self.config.gcn_dim,
            num_layers=2
        )

        self.trans_rep = nn.Linear(self.xlmr_dim + self.config.feat_dim * 2 + self.config.gcn_dim * 4, self.xlmr_dim)

        # local classifiers
        # span detection
        self.mention_label_ffn = Linears([self.xlmr_dim, config.role_hidden_num,
                                          self.mention_label_num],
                                         dropout_prob=config.linear_dropout,
                                         bias=config.linear_bias,
                                         activation=config.linear_activation)

        self.trigger_label_ffn = Linears([self.xlmr_dim, config.event_hidden_num,
                                          self.trigger_label_num],
                                         dropout_prob=config.linear_dropout,
                                         bias=config.linear_bias,
                                         activation=config.linear_activation)

        # event classification
        self.event_type_ffn = Linears([self.xlmr_dim, config.event_hidden_num,
                                       self.event_type_num],
                                      dropout_prob=config.linear_dropout,
                                      bias=config.linear_bias,
                                      activation=config.linear_activation)

        self.role_type_ffn = Linears([self.xlmr_dim, config.role_hidden_num,
                                      self.role_type_num],
                                     dropout_prob=config.linear_dropout,
                                     bias=config.linear_bias,
                                     activation=config.linear_activation)

        self.event_criteria = torch.nn.CrossEntropyLoss()
        self.role_criteria = torch.nn.CrossEntropyLoss(reduction='none')

        # others
        self.trigger_crf = CRF(self.trigger_label_stoi, bioes=False)
        self.mention_crf = CRF(self.mention_label_stoi, bioes=False)
        self.pad_vector = nn.Parameter(torch.randn(1, 1, self.xlmr_dim))

    def encode(self, piece_idxs, attention_masks, token_lens, start_tokens, end_tokens):
        batch_size, _ = piece_idxs.size()
        all_xlmr_outputs = self.xlmr(piece_idxs, attention_mask=attention_masks)
        xlmr_outputs = all_xlmr_outputs[0]

        idxs, token_num, token_len = token_lens_to_idxs(token_lens, start_tokens, end_tokens)
        idxs = piece_idxs.new(idxs) + 1
        xlmr_outputs = compute_word_reps_avg(xlmr_outputs, idxs)

        xlmr_outputs = self.xlmr_dropout(xlmr_outputs)
        return xlmr_outputs

    def scores(self, xlmr_outputs, batch, graphs):
        (
            mention_idxs, mention_num, mention_len,
            trigger_idxs, trigger_num, trigger_len
        ) = graphs_to_node_idxs(batch, graphs)

        batch_size, _, xlmr_dim = xlmr_outputs.size()

        trigger_idxs = xlmr_outputs.new_tensor(trigger_idxs, dtype=torch.long)
        mention_idxs = xlmr_outputs.new_tensor(mention_idxs, dtype=torch.long)

        # mention type scores
        mention_reprs = compute_span_reprs(xlmr_outputs, mention_idxs)
        role_type_scores = self.role_type_ffn(mention_reprs)

        # event type scores
        trigger_reprs = compute_span_reprs(xlmr_outputs, trigger_idxs)
        event_type_scores = self.event_type_ffn(trigger_reprs)

        return role_type_scores, event_type_scores

    def forward(self, batch):
        # encoding
        xlmr_reprs = self.encode(batch.piece_idxs,
                                 batch.attention_masks,
                                 batch.token_lens, batch.start_tokens, batch.end_tokens)
        batch_size, seq_len, _ = xlmr_reprs.size()
        upos_reps = self.upos_embedding(batch.upos_ids)
        dist_reps = get_dist_embeds(batch.triggers, batch_size, seq_len, self.dist_embedding)
        word_deprel_embs = self.deprel_embedding(batch.deprel_ids)
        word_deprel_reps, _ = self.gcn_layer(word_deprel_embs, batch.head_ids)
        trigger_deprel_reps = get_trigger_reps(word_deprel_reps, batch.triggers).unsqueeze(1).repeat(1, seq_len,
                                                                                                     1)  # [batch size, sep len, xlmr dim]

        word_feats = torch.cat(
            [xlmr_reprs, upos_reps, dist_reps, word_deprel_reps, trigger_deprel_reps,
             torch.abs(trigger_deprel_reps - word_deprel_reps), trigger_deprel_reps * word_deprel_reps],
            dim=2
        )
        word_reprs = self.trans_rep(word_feats)

        mention_label_scores = self.mention_label_ffn(word_reprs)

        mention_label_scores = self.mention_crf.pad_logits(mention_label_scores)
        mention_label_loglik = self.mention_crf.loglik(mention_label_scores,
                                                       batch.mention_label_idxs,
                                                       batch.token_nums).unsqueeze(-1) * batch.weights.unsqueeze(-1)

        role_type_scores, event_type_scores = self.scores(word_reprs, batch, batch.graphs)

        # role classification
        role_type_scores = role_type_scores.view(-1, self.role_type_num)

        classification_loss = self.role_criteria(role_type_scores, batch.role_type_idxs).view(batch_size,
                                                                                              -1) * batch.weights.unsqueeze(
            -1)

        loss = classification_loss.mean() - mention_label_loglik.mean()

        return loss

    def predict(self, batch):
        xlmr_reprs = self.encode(batch.piece_idxs,
                                 batch.attention_masks,
                                 batch.token_lens, batch.start_tokens, batch.end_tokens)
        batch_size, seq_len, _ = xlmr_reprs.size()
        upos_reps = self.upos_embedding(batch.upos_ids)
        dist_reps = get_dist_embeds(batch.triggers, batch_size, seq_len, self.dist_embedding)
        word_deprel_embs = self.deprel_embedding(batch.deprel_ids)
        word_deprel_reps, _ = self.gcn_layer(word_deprel_embs, batch.head_ids)
        trigger_deprel_reps = get_trigger_reps(word_deprel_reps, batch.triggers).unsqueeze(1).repeat(1, seq_len,
                                                                                                     1)  # [batch size, sep len, xlmr dim]

        word_feats = torch.cat(
            [xlmr_reprs, upos_reps, dist_reps, word_deprel_reps, trigger_deprel_reps,
             torch.abs(trigger_deprel_reps - word_deprel_reps), trigger_deprel_reps * word_deprel_reps],
            dim=2
        )
        word_reprs = self.trans_rep(word_feats)

        mention_label_scores = self.mention_label_ffn(word_reprs)
        mention_label_scores = self.mention_crf.pad_logits(mention_label_scores)
        _, mention_label_preds = self.mention_crf.viterbi_decode(mention_label_scores,
                                                                 batch.token_nums)

        batch_mentions = mention_tag_paths_to_spans(mention_label_preds,
                                                    batch.token_nums,
                                                    self.mention_label_stoi)

        pred_graphs = [MentionGraph(did, eid, sid, doc_txt, toks, [], elist, self.vocabs, start, end)
                       for did, eid, sid, doc_txt, toks, elist, start, end in
                       zip(batch.doc_ids, batch.entry_ids, batch.sent_ids, batch.doc_texts, batch.tokens,
                           batch_mentions, batch.start, batch.end)]

        role_type_scores = self.scores(word_reprs, batch, pred_graphs)[0]
        role_type_probs = torch.softmax(role_type_scores,dim=-1).data.cpu().numpy().tolist() # [batch size, num triggers * num mentions, num roles]
        role_type_scores = role_type_scores.data.cpu().numpy().tolist()

        pred_graphs = []
        # Decode each document in the batch
        for bid in range(batch_size):
            seq_mentions = batch_mentions[bid]
            mention_num = len(seq_mentions)

            graph = deepcopy(batch.graphs[bid])
            if mention_num == 0:
                # skip decoding
                pred_graphs.append(graph)
                continue

            for mid, mention in enumerate(seq_mentions):
                start, end, mention_type = mention
                role_type, _ = max([(l, s) for l, s in enumerate(role_type_scores[bid][mid])], key=lambda x: x[1])
                if role_type > 0:
                    graph.add_role(start, end, mention_type, self.role_type_itos[role_type], role_type_probs[bid][mid][role_type] * 0.99)

            pred_graphs.append(graph)
        return pred_graphs


class TemplateModel(nn.Module):
    def __init__(self,
                 config,
                 vocabs):
        super().__init__()
        self.config = config
        # vocabularies
        self.vocabs = vocabs

        # XLMR encoder
        self.xlmr_dim = 1024 if 'large' in config.xlmr_model_name else 768
        self.xlmr = XLMRobertaModel.from_pretrained(config.xlmr_model_name,
                                                    cache_dir=config.cache_dir,
                                                    output_hidden_states=True)
        self.xlmr_dropout = nn.Dropout(p=config.xlmr_dropout)

        self.trans_rep = nn.Linear(self.xlmr_dim * 2, self.xlmr_dim)

        self.type_slot_itos = {i: s for s, i in self.vocabs['type_slot'].items()}
        self.completion_slot_itos = {i: s for s, i in self.vocabs['completion_slot'].items()}
        self.overtime_slot_itos = {i: s for s, i in self.vocabs['overtime_slot'].items()}
        self.coordinated_slot_itos = {i: s for s, i in self.vocabs['coordinated_slot'].items()}

        hidden_ffns = {}
        template_classifiers = {}
        type_classifiers = {}
        completion_classifiers = {}
        overtime_classifiers = {}
        coordinated_classifiers = {}

        for template_type, template_id in self.vocabs['template-type'].items():
            if template_type != 'O':
                self.hidden_dim = 768
                hidden_ffns[template_type] = Linears([self.xlmr_dim, self.hidden_dim],
                                                     dropout_prob=config.linear_dropout,
                                                     bias=config.linear_bias,
                                                     activation=config.linear_activation)

                template_classifiers[template_type] = Linears([self.hidden_dim, config.event_hidden_num, 2],
                                                              dropout_prob=config.linear_dropout,
                                                              bias=config.linear_bias,
                                                              activation=config.linear_activation)

                # categorical slots
                type_classifiers[template_type] = Linears([self.hidden_dim, config.event_hidden_num,
                                                           len(vocabs['type_slot'])],
                                                          dropout_prob=config.linear_dropout,
                                                          bias=config.linear_bias,
                                                          activation=config.linear_activation)

                completion_classifiers[template_type] = Linears([self.hidden_dim, config.event_hidden_num,
                                                                 len(vocabs['completion_slot'])],
                                                                dropout_prob=config.linear_dropout,
                                                                bias=config.linear_bias,
                                                                activation=config.linear_activation)

                overtime_classifiers[template_type] = Linears([self.hidden_dim, config.event_hidden_num,
                                                               len(vocabs['overtime_slot'])],
                                                              dropout_prob=config.linear_dropout,
                                                              bias=config.linear_bias,
                                                              activation=config.linear_activation)

                coordinated_classifiers[template_type] = Linears([self.hidden_dim, config.event_hidden_num,
                                                                  len(vocabs['coordinated_slot'])],
                                                                 dropout_prob=config.linear_dropout,
                                                                 bias=config.linear_bias,
                                                                 activation=config.linear_activation)

        self.hidden_ffns = nn.ModuleDict(hidden_ffns)
        self.template_classifiers = nn.ModuleDict(template_classifiers)
        self.type_classifiers = nn.ModuleDict(type_classifiers)
        self.completion_classifiers = nn.ModuleDict(completion_classifiers)
        self.overtime_classifiers = nn.ModuleDict(overtime_classifiers)
        self.coordinated_classifiers = nn.ModuleDict(coordinated_classifiers)

        self.cls_criteria = torch.nn.CrossEntropyLoss(reduction='none')

    def encode(self, piece_idxs, attention_masks, token_lens, start_tokens, end_tokens):
        batch_size, _ = piece_idxs.size()
        all_xlmr_outputs = self.xlmr(piece_idxs, attention_mask=attention_masks)
        xlmr_outputs = all_xlmr_outputs[0]
        cls_reprs = xlmr_outputs[:, 0, :]  # [batch size, xlmr dim]

        idxs, token_num, token_len = token_lens_to_idxs(token_lens, start_tokens, end_tokens)
        idxs = piece_idxs.new(idxs) + 1
        xlmr_outputs = compute_word_reps_avg(xlmr_outputs, idxs)

        xlmr_outputs = self.xlmr_dropout(xlmr_outputs)
        return xlmr_outputs, cls_reprs

    def forward(self, batch):
        # encoding
        # initially, the input is the first 512 tokens of the document.
        # later on, you might want to try excluding sentences without basic events to include more content
        # in the input of the template model
        word_reprs, cls_reprs = self.encode(batch.piece_idxs,
                                            batch.attention_masks,
                                            batch.token_lens, batch.start_tokens, batch.end_tokens)
        trigger_mask = (batch.type_ids > 0).float()
        trigger_pooled = torch.sum(word_reprs * trigger_mask.unsqueeze(-1), dim=1) / (1e-12 + torch.sum(trigger_mask,
                                                                                               dim=1).unsqueeze(-1))  # [batch size, rep dim]
        batch_size, _ = cls_reprs.size()

        cls_reprs = self.trans_rep(torch.cat([cls_reprs, trigger_pooled], dim=1))

        cls_loss = 0
        for template_type, template_id in self.vocabs['template-type'].items():
            if template_type == 'O':
                continue

            # template anchor classification
            anchor_reprs = self.hidden_ffns[template_type](cls_reprs)

            template_scores = self.template_classifiers[template_type](anchor_reprs)
            type_slot_scores = self.type_classifiers[template_type](anchor_reprs)
            completion_slot_scores = self.completion_classifiers[template_type](anchor_reprs)
            overtime_slot_scores = self.overtime_classifiers[template_type](anchor_reprs)
            coordinated_slot_scores = self.coordinated_classifiers[template_type](anchor_reprs)

            cls_loss += self.cls_criteria(template_scores,
                                          batch.template_type_idxs[template_type]) * batch.weights.unsqueeze(-1)
            cls_loss += self.cls_criteria(type_slot_scores,
                                          batch.type_slot_idxs[template_type]) * batch.weights.unsqueeze(-1)
            cls_loss += self.cls_criteria(completion_slot_scores,
                                          batch.completion_slot_idxs[template_type]) * batch.weights.unsqueeze(-1)
            cls_loss += self.cls_criteria(overtime_slot_scores,
                                          batch.overtime_slot_idxs[template_type]) * batch.weights.unsqueeze(-1)
            cls_loss += self.cls_criteria(coordinated_slot_scores,
                                          batch.coordinated_slot_idxs[template_type]) * batch.weights.unsqueeze(-1)
        return cls_loss.mean()

    def predict(self, batch):
        # encoding
        word_reprs, cls_reprs = self.encode(batch.piece_idxs,
                                            batch.attention_masks,
                                            batch.token_lens, batch.start_tokens, batch.end_tokens)
        trigger_mask = (batch.type_ids > 0).float()
        trigger_pooled = torch.sum(word_reprs * trigger_mask.unsqueeze(-1), dim=1) / (1e-12 + torch.sum(trigger_mask,
                                                                                               dim=1).unsqueeze(-1))  # [batch size, rep dim]
        batch_size, _ = cls_reprs.size()

        cls_reprs = self.trans_rep(torch.cat([cls_reprs, trigger_pooled], dim=1))

        predicted_templates = []

        for template_type, template_id in self.vocabs['template-type'].items():
            if template_type == 'O':
                continue

            # template anchor classification
            anchor_reprs = self.hidden_ffns[template_type](cls_reprs)

            template_scores = self.template_classifiers[template_type](anchor_reprs)
            template_probs = torch.softmax(template_scores, dim=-1).data.cpu().numpy().tolist()

            type_slot_scores = self.type_classifiers[template_type](anchor_reprs)
            completion_slot_scores = self.completion_classifiers[template_type](anchor_reprs)
            overtime_slot_scores = self.overtime_classifiers[template_type](anchor_reprs)
            coordinated_slot_scores = self.coordinated_classifiers[template_type](anchor_reprs)

            # template anchor classification
            template_type_preds = torch.argmax(template_scores, dim=1).data.cpu().numpy().tolist()
            type_slot_preds = torch.argmax(type_slot_scores, dim=1).data.cpu().numpy().tolist()
            completion_slot_preds = torch.argmax(completion_slot_scores, dim=1).data.cpu().numpy().tolist()
            overtime_slot_preds = torch.argmax(overtime_slot_scores, dim=1).data.cpu().numpy().tolist()
            coordinated_slot_preds = torch.argmax(coordinated_slot_scores, dim=1).data.cpu().numpy().tolist()

            for bid in range(batch_size):
                if template_type_preds[bid] > 0:
                    new_template = {
                        'entry-id': batch.entry_ids[bid],
                        'template-type': template_type,
                        'confidence': template_probs[bid][template_type_preds[bid]]
                    }
                    if type_slot_preds[bid] > 0:
                        new_template['type'] = self.type_slot_itos[type_slot_preds[bid]]
                    if completion_slot_preds[bid] > 0:
                        new_template['completion'] = self.completion_slot_itos[completion_slot_preds[bid]]
                    if overtime_slot_preds[bid] > 0:
                        new_template['over-time'] = True
                    if coordinated_slot_preds[bid] > 0:
                        new_template['coordinated'] = True

                    predicted_templates.append(new_template)
        return predicted_templates


class SlotModel(nn.Module):
    def __init__(self,
                 config,
                 vocabs):
        super().__init__()
        self.config = config
        # vocabularies
        self.vocabs = vocabs
        self.irrealis_type_stoi = vocabs['irrealis_type']
        self.irrealis_type_itos = {i: s for s, i in self.irrealis_type_stoi.items()}
        # XLMR encoder
        self.xlmr_dim = 1024 if 'large' in config.xlmr_model_name else 768
        self.xlmr = XLMRobertaModel.from_pretrained(config.xlmr_model_name,
                                                    cache_dir=config.cache_dir,
                                                    output_hidden_states=True)
        self.xlmr_dropout = nn.Dropout(p=config.xlmr_dropout)

        # basic-span GCN
        self.basic_gcn = InstanceGCN(trigger_dim=self.xlmr_dim, entity_dim=self.xlmr_dim, num_layers=3)
        # slot classifiers
        tmp = {}
        for template_type in self.vocabs['template-type2slots']:
            for slot_name in self.vocabs['template-type2slots'][template_type]:
                if slot_name + '-ffn' not in tmp:
                    tmp[slot_name + '-ffn'] = Linears([self.xlmr_dim, config.role_hidden_num, 2],
                                                      dropout_prob=config.linear_dropout,
                                                      bias=config.linear_bias,
                                                      activation=config.linear_activation)

        self.name2classifier = nn.ModuleDict(tmp)

        # irrealis classifier
        self.irrealis_classifier = Linears(
            [self.xlmr_dim + len(vocabs['slot_type']), config.role_hidden_num,
             len(vocabs['irrealis_type'])],
            dropout_prob=config.linear_dropout,
            bias=config.linear_bias,
            activation=config.linear_activation)
        # loss functions
        self.criteria = torch.nn.CrossEntropyLoss()

    def encode(self, piece_idxs, attention_masks, token_lens, start_tokens, end_tokens):
        batch_size, _ = piece_idxs.size()
        all_xlmr_outputs = self.xlmr(piece_idxs, attention_mask=attention_masks)
        xlmr_outputs = all_xlmr_outputs[0]

        idxs, token_num, token_len = token_lens_to_idxs(token_lens, start_tokens, end_tokens)
        idxs = piece_idxs.new(idxs) + 1
        xlmr_outputs = compute_word_reps_avg(xlmr_outputs, idxs)

        xlmr_outputs = self.xlmr_dropout(xlmr_outputs)
        return xlmr_outputs

    def scores(self, xlmr_outputs, batch, graphs):
        (
            mention_idxs, mention_num, mention_len,
            trigger_idxs, trigger_num, trigger_len
        ) = graphs_to_node_idxs(batch, graphs)

        batch_size, _, xlmr_dim = xlmr_outputs.size()

        trigger_idxs = xlmr_outputs.new_tensor(trigger_idxs, dtype=torch.long)
        mention_idxs = xlmr_outputs.new_tensor(mention_idxs, dtype=torch.long)

        # event reprs
        trigger_reprs = compute_span_reprs(xlmr_outputs, trigger_idxs)
        # mention reprs
        mention_reprs = compute_span_reprs(xlmr_outputs, mention_idxs)

        # encoding trigger-trigger, trigger-entity relationships from predicted basic events and arguments
        trigger_reprs, mention_reprs = self.basic_gcn(
            adjacency_matrix=batch.basic_graphs,
            T=trigger_reprs,
            E=mention_reprs,
            t_n=trigger_num,
            e_n=mention_num
        )

        # candidate reprs
        candidate_reprs = torch.cat(
            [trigger_reprs, mention_reprs], dim=1
        )  # [batch size, max trigger num + max mention num, xlmr dim]

        return candidate_reprs, trigger_num, mention_num

    def forward(self, batch, abstract_data=False):
        # encoding
        word_reprs = self.encode(batch.piece_idxs,
                                 batch.attention_masks,
                                 batch.token_lens, batch.start_tokens, batch.end_tokens)
        batch_size, _, _ = word_reprs.size()
        # basic triggers and mentions' representations
        candidate_reprs, trigger_num, mention_num = self.scores(word_reprs, batch, batch.graphs)

        slot_loss = 0
        irrealis_loss = 0
        for bid in range(batch_size):
            weight = batch.weights[bid]
            template_type = batch.template_types[bid]
            for slot_id, slot_name in enumerate(self.vocabs['template-type2slots'][template_type]):
                classifier = self.name2classifier[slot_name + '-ffn']
                slot_reprs = classifier(candidate_reprs[bid])
                slot_loss += weight * self.criteria(slot_reprs, batch.slot_type_idxs[bid][slot_id])

                # irrealis classification
                if slot_name in self.vocabs['slot2irrealis']:
                    for cand_id in range(slot_reprs.shape[0]):
                        if batch.slot_type_idxs[bid][slot_id][cand_id].item() == 1:
                            slot_indicator = torch.zeros(size=(1, len(self.vocabs['slot_type'])),
                                                         requires_grad=False).cuda()
                            slot_indicator[:, self.vocabs['slot_type'][slot_name]] = 1.
                            irrealis_reprs = torch.cat(
                                [candidate_reprs[bid][cand_id].unsqueeze(0), slot_indicator],
                                dim=1
                            )  # [1, xlmr dim + num slot]
                            irrealis_scores = self.irrealis_classifier(irrealis_reprs)
                            irrealis_loss += weight * self.criteria(irrealis_scores,
                                                                    batch.irrealis_idxs[bid][slot_id][
                                                                        cand_id].unsqueeze(0))
        loss = slot_loss + irrealis_loss
        return loss

    def predict(self, batch):
        # encoding
        word_reprs = self.encode(batch.piece_idxs,
                                 batch.attention_masks,
                                 batch.token_lens, batch.start_tokens, batch.end_tokens)
        batch_size, _, _ = word_reprs.size()
        # basic triggers and mentions' representations
        candidate_reprs, trigger_num, mention_num = self.scores(word_reprs, batch, batch.graphs)

        batch_predictions = []
        for bid in range(batch_size):
            predictions = []
            template_type = batch.template_types[bid]
            for slot_id, slot_name in enumerate(self.vocabs['template-type2slots'][template_type]):
                classifier = self.name2classifier[slot_name + '-ffn']
                slot_reprs = classifier(candidate_reprs[bid])  # [num candidates, 2]
                slot_probs = torch.softmax(slot_reprs, dim=-1).data.cpu().numpy().tolist() # [num candidates, 2]

                preds = torch.argmax(slot_reprs, dim=1).data.cpu().numpy().tolist()  # [num candidates]

                # filter illegal predictions
                if slot_name in self.vocabs['slot2filler_types'] and 'ssid' in self.vocabs['slot2filler_types'][
                    slot_name] and 'event-id' not in \
                        self.vocabs['slot2filler_types'][slot_name]:
                    preds = [p if k >= trigger_num else 0 for k, p in
                             enumerate(preds)]
                elif slot_name in self.vocabs['slot2filler_types'] and 'event-id' in self.vocabs['slot2filler_types'][
                    slot_name] and 'ssid' not in \
                        self.vocabs['slot2filler_types'][slot_name]:
                    preds = [p if k < trigger_num else 0 for k, p in
                             enumerate(preds)]

                probs = [slot_probs[i][p] for i, p in enumerate(preds)]
                entry_ids = [batch.entry_ids[bid] for _ in range(len(preds))]
                template_ids = [batch.template_ids[bid] for _ in range(len(preds))]
                names = [slot_name for _ in range(len(preds))]
                slot_ids = batch.slot_ids[bid][slot_id]

                # irrealis

                if slot_name in self.vocabs['slot2irrealis']:
                    preds_irrealis = []
                    for cand_id in range(len(preds)):
                        if preds[cand_id] == 1:
                            slot_indicator = torch.zeros(size=(1, len(self.vocabs['slot_type'])),
                                                         requires_grad=False).cuda()
                            slot_indicator[:, self.vocabs['slot_type'][slot_name]] = 1.
                            irrealis_reprs = torch.cat(
                                [candidate_reprs[bid][cand_id].unsqueeze(0), slot_indicator],
                                dim=1
                            )  # [1, xlmr dim + num slot]
                            irrealis_scores = self.irrealis_classifier(irrealis_reprs)
                            preds_irrealis.append(torch.argmax(irrealis_scores, dim=1).data.cpu().numpy().tolist()[0])
                        else:
                            preds_irrealis.append(-100)
                else:
                    preds_irrealis = [-100] * len(preds)

                for a, b, c, d, e, f, ff in zip(entry_ids, template_ids, names, slot_ids, preds, preds_irrealis, probs):
                    if d == 'skip' or e != 1:
                        continue
                    pred = {
                        'entry-id': a, 'template-id': b, 'filler-id': d,
                        'slot': slot_name.lower(), 'confidence': ff
                    }
                    irrealis = self.irrealis_type_itos.get(f, 'O')
                    if irrealis != 'O' and slot_name in self.vocabs['slot2irrealis'] and irrealis in \
                            self.vocabs['slot2irrealis'][slot_name]:
                        pred['irrealis'] = irrealis

                    batch_predictions.append(pred)

        return batch_predictions


class EventCorefModel(nn.Module):
    def __init__(self,
                 config,
                 vocabs):
        super().__init__()
        self.config = config
        # vocabularies
        self.vocabs = vocabs

        self.coref_type_stoi = {
            'O': 0, 'RefEvent': 1, 'CorefEvent': 2
        }
        self.coref_type_itos = {i: s for s, i in self.coref_type_stoi.items()}
        self.coref_type_num = len(self.coref_type_stoi)

        # XLMR encoder
        self.xlmr_dim = 1024 if 'large' in config.xlmr_model_name else 768
        self.xlmr = XLMRobertaModel.from_pretrained(config.xlmr_model_name,
                                                    cache_dir=config.cache_dir,
                                                    output_hidden_states=True)
        self.xlmr_dropout = nn.Dropout(p=config.xlmr_dropout)

        self.type_embedding = nn.Embedding(
            num_embeddings=len(self.vocabs['type']),
            embedding_dim=100,
            padding_idx=0
        )

        self.trans_rep = nn.Linear(self.xlmr_dim + 100, self.xlmr_dim)

        self.coref_type_ffn = nn.Linear(self.xlmr_dim * 4, self.coref_type_num)

        self.coref_criteria = torch.nn.CrossEntropyLoss(reduction='none')

    def encode(self, piece_idxs, attention_masks, token_lens, start_tokens, end_tokens):
        batch_size, _ = piece_idxs.size()
        all_xlmr_outputs = self.xlmr(piece_idxs, attention_mask=attention_masks)
        xlmr_outputs = all_xlmr_outputs[0]

        idxs, token_num, token_len = token_lens_to_idxs(token_lens, start_tokens, end_tokens)
        idxs = piece_idxs.new(idxs) + 1
        xlmr_outputs = compute_word_reps_avg(xlmr_outputs, idxs)

        xlmr_outputs = self.xlmr_dropout(xlmr_outputs)
        return xlmr_outputs

    def scores(self, word_reprs, batch, graphs):
        (
            mention_idxs, mention_num, mention_len,
            trigger_idxs, trigger_num, trigger_len
        ) = graphs_to_node_idxs(batch, graphs)

        batch_size, _, xlmr_dim = word_reprs.size()

        trigger_idxs = word_reprs.new_tensor(trigger_idxs, dtype=torch.long)
        trigger_reprs = compute_span_reprs(word_reprs, trigger_idxs)

        # coref type score
        trg1_reprs, trg2_reprs = compute_binary_reprs(trigger_reprs, trigger_reprs)

        coref_reprs = torch.cat(
            [trg1_reprs, trg2_reprs, trg1_reprs * trg2_reprs, trg1_reprs - trg2_reprs],
            dim=2
        )
        coref_type_scores = self.coref_type_ffn(coref_reprs)
        return coref_type_scores, trigger_num

    def forward(self, batch):
        # encoding
        xlmr_reprs = self.encode(batch.piece_idxs,
                                 batch.attention_masks,
                                 batch.token_lens, batch.start_tokens, batch.end_tokens)
        batch_size, seq_len, _ = xlmr_reprs.size()
        type_reprs = self.type_embedding(batch.type_ids)

        word_reprs = self.trans_rep(torch.cat(
            [xlmr_reprs, type_reprs], dim=2
        ))

        # classification
        coref_type_scores, _ = self.scores(word_reprs, batch, batch.graphs)  # [batch size, trg num ** 2, coref num]

        coref_type_scores = coref_type_scores.view(-1, self.coref_type_num)
        classification_loss = self.coref_criteria(coref_type_scores, batch.role_type_idxs).view(batch_size,
                                                                                                -1) * batch.weights.unsqueeze(
            -1)
        return classification_loss.mean()

    def predict(self, batch):
        # encoding
        xlmr_reprs = self.encode(batch.piece_idxs,
                                 batch.attention_masks,
                                 batch.token_lens, batch.start_tokens, batch.end_tokens)
        batch_size, seq_len, _ = xlmr_reprs.size()
        type_reprs = self.type_embedding(batch.type_ids)

        word_reprs = self.trans_rep(torch.cat(
            [xlmr_reprs, type_reprs], dim=2
        ))

        # classification
        coref_type_scores, max_trigger_num = self.scores(word_reprs, batch,
                                                         batch.graphs)  # [batch size, trg num ** 2, coref num]

        pred_corefs = torch.argmax(coref_type_scores, dim=2).data.cpu().tolist()  # [BS, trg num ** 2]
        batch_scores = torch.softmax(coref_type_scores, dim=2).data.cpu().tolist()  # [BS, trg num ** 2, coref num]

        batch_predictions = []
        for bid in range(batch_size):
            sent_predictions = {
                'entry-id': batch.entry_ids[bid],
                'event_list': batch.eventids[bid],
                'predictions': []
            }
            actual_trigger_num = batch.graphs[bid].trigger_num
            for i in range(actual_trigger_num):
                for j in range(actual_trigger_num):
                    if i == j:
                        continue
                    if pred_corefs[bid][i * max_trigger_num + j] > 0:
                        if pred_corefs[bid][i * max_trigger_num + j] == self.coref_type_stoi['RefEvent']:
                            if pred_corefs[bid][j * max_trigger_num + i] == self.coref_type_stoi['O']:
                                sent_predictions['predictions'].append({
                                    'entry-id': batch.entry_ids[bid],
                                    'head-eventid': batch.eventids[bid][i],
                                    'tail-eventid': batch.eventids[bid][j],
                                    'coref-type': 'RefEvent',
                                    'confidence': batch_scores[bid][i * max_trigger_num + j][
                                        pred_corefs[bid][i * max_trigger_num + j]]
                                })
                        else:
                            # pred_corefs[bid][i][j] == self.coref_type_stoi['CorefEvent']
                            if pred_corefs[bid][j * max_trigger_num + i] == self.coref_type_stoi['CorefEvent']:
                                sent_predictions['predictions'].append({
                                    'entry-id': batch.entry_ids[bid],
                                    'head-eventid': batch.eventids[bid][i],
                                    'tail-eventid': batch.eventids[bid][j],
                                    'coref-type': 'CorefEvent',
                                    'confidence': batch_scores[bid][i * max_trigger_num + j][
                                        pred_corefs[bid][i * max_trigger_num + j]]
                                })
            batch_predictions.append(sent_predictions)
        return batch_predictions


def max_pooling(vecs, mask):
    h = vecs.clone()
    h = h.masked_fill(mask, -1e12)
    return torch.max(h, 1)[0]


def avg_pooling(vecs, mask):
    h = vecs.clone()
    h = h.masked_fill(mask, 0)
    return h.sum(1) / (mask.size(1) - mask.float().sum(1))
