# Copyright (C) 2007 - 2021 The MITRE Corporation. See the toplevel
# file LICENSE.txt for license terms.

from __future__ import print_function, absolute_import
from .PyCompat import Py3

# MATConfig is a dict, but it's oh, so much more.
# You can override settings with a config file, or
# with the environment. This is a read-only dictionary,
# as far as the user is concerned.

# Override order: env -> MAT_SETTINGS_FILE in environ -> MAT_settings.config
# This needs to be enforced as time goes by.

import os, sys

if Py3:
    import configparser
    from collections import UserDict as IterableUserDict
else:
    import ConfigParser as configparser
    from UserDict import IterableUserDict

class ConfigError(Exception):
    pass

# By default, now, the configuration is empty. The
# config info for the default full distribution is added in the
# import for the MAT package itself.

class ConfigDict(IterableUserDict):

    def __init__(self, useEnv = True, verboseConfig = False):
        IterableUserDict.__init__(self)
        self.verboseConfig = verboseConfig
        self.sourceAndVerboseStatus = {}
        thisDir = os.path.dirname(__file__)
        self.useEnv = useEnv
        self.fixedKeys = set()

    #
    # Public API
    #

    def configureFromSetting(self, k, v, fixed = False):
        if k in self.fixedKeys:
            pass
        elif fixed:
            self._setFixedKey(k, v)
        else:
            self._setKey(k, v)

    def configureFromDictionary(self, d, fixed = False):
        for k, v in d.items():
            self.configureFromSetting(k, v, fixed = fixed)

    def configureFromFile(self, f, augment = False):
        f = f.strip()
        if not f:
            # Empty string.
            return
        if not os.path.exists(f):
            raise ConfigError("config file doesn't exist")
        self._readConfigFile(os.path.abspath(f), augment = augment)

    #
    # everything else
    #

    def _setFixedKey(self, k, v):
        # This sets a value, but will not allow it
        # to be reset.
        self.data[k] = v
        self.fixedKeys.add(k)
        self.sourceAndVerboseStatus[k] = ["<fixed by installation>", False]

    # In general, the user shouldn't be modifying keys, and the
    # normal dictionary method is disabled. But sometimes it has
    # to be done.
    
    def _setKey(self, k, v):
        self._setKeyAndSource(k, v, "<set by installation>")

    # We're going to update the paths automatically when they're set.
    def _setKeyAndSource(self, k, v, source):
        oldVal = self.data.get(k)
        self.data[k] = v
        self.sourceAndVerboseStatus[k] = [source, False]
        # Special handling.
        if k.endswith("_PYTHONLIB"):
            import sys
            if (oldVal is not None) and (oldVal in sys.path):
                sys.path.remove(oldVal)
            sys.path.insert(0, v)

    def _readConfigFile(self, f, augment = False):
        p = configparser.RawConfigParser()
        p.optionxform = str
        p.read([f])
        for sect in p.sections():
            if sect == "_GLOBALS":
                prefix = ""
            else:
                prefix = sect + ":"
            for opt in p.options(sect):
                k = prefix + opt
                # Can't override these via augmentSettings.
                if augment and (k in self.fixedKeys):
                    continue
                self._setKeyAndSource(k, p.get(sect, opt), f)

    def __getitem__(self, key):
        useEnv = self.useEnv
        if key in self.fixedKeys:
            useEnv = False
        if useEnv and (key in os.environ):
            v = os.environ[key]
            if (key not in self.sourceAndVerboseStatus) or \
               (self.sourceAndVerboseStatus[key][0] != "<shell environment>"):
                self.sourceAndVerboseStatus[key] = ["<shell environment>", False]
        else:
            v = IterableUserDict.__getitem__(self, key)
        # If we've gotten this far, getitem hasn't thrown KeyError, which means
        # that the key HAS to have an entry in sourceAndVerboseStatus.
        # But let's check.
        if key not in self.sourceAndVerboseStatus:
            self.sourceAndVerboseStatus[key] = ["<unknown source>", False]
        if self.verboseConfig and (not self.sourceAndVerboseStatus[key][1]):
            self.sourceAndVerboseStatus[key][1] = True
            print("[Read value of config var %s from %s]" % (key, self.sourceAndVerboseStatus[key][0]), file=sys.stderr)
        return v            

    def __setitem__(self, key, val):
        raise KeyError("Read-only dictionary")

    def __contains__(self, key):
        useEnv = self.useEnv
        if key in self.fixedKeys:
            useEnv = False        
        return (useEnv and (key in os.environ)) or \
               (IterableUserDict.__contains__(self, key))

    def setVerboseConfig(self, boolVal):
        self.verboseConfig = boolVal

    def copy(self):
        newConf = ConfigDict(self.useEnv, self.verboseConfig)
        newConf.data = self.data.copy()
        newConf.sourceAndVerboseStatus = dict([(k, v[:]) for (k, v) in self.sourceandVerboseStatus.items()])
        return newConf

MATConfig = ConfigDict()
