# Copyright (C) 2021 The MITRE Corporation. See the toplevel
# file LICENSE.txt for license terms.

from __future__ import (absolute_import, print_function)
from .PyCompat import Py3, _strtypes, _numtypes

# Readers and writers for documents.

import sys, codecs, re, json, os

class NotImplementedError(Exception):
    pass

from .Document import (AnnotatedDoc, LoadError, AnnotationReporter, DocumentError)
from .Annotation import (AnnotationCore, StringAttributeType, 
                         AttributeValueList, AttributeValueSet, AttributeValueSequence, 
                         AnnotationAttributeType, AnnotationError)
from .Operation import (OptionBearer, OptionTemplate, OpArgument)
from .XMLNode import (XMLNodeDescFromFile, XMLNodeFromFile, XMLNodeFromString)
from .Config import MATConfig
from . import ExecutionContext
from .PyCompat import _jsondumps

# Toplevel getter.

_IO_LIB = {}

def declareDocumentIO(name, cls, input, output):
    global _IO_LIB
    _IO_LIB[name] = cls, input, output

def getDocumentIO(name, **kw):
    return _IO_LIB[name][0](**kw)

def getDocumentIOClass(name):
    return _IO_LIB[name][0]

def getInputDocumentIOClass(name):
    v = _IO_LIB[name]
    if not v[1]:
        raise KeyError
    return v[0]

def getOutputDocumentIOClass(name):
    v = _IO_LIB[name]
    if not v[2]:
        raise KeyError
    return v[0]

def documentIODeclared(name):
    return name in _IO_LIB

def inputDocumentIODeclared(name):
    return (name in _IO_LIB) and _IO_LIB[name][1]

def outputDocumentIODeclared(name):
    return (name in _IO_LIB) and _IO_LIB[name][2]

def allDocumentIO(exclusions = None):
    return [k for k in _IO_LIB.keys() if ((exclusions is None) or (k not in exclusions))]

def allInputDocumentIO(exclusions = None):
    return [name for (name, (cls, input, output)) in _IO_LIB.items()
            if input and ((exclusions is None) or (name not in exclusions))]

def allOutputDocumentIO(exclusions = None):
    return [name for (name, (cls, input, output)) in _IO_LIB.items()
            if output and ((exclusions is None) or (name not in exclusions))]

class SaveError(Exception):
    pass

#
# Base class.
#

class DocumentIO(OptionBearer):

    @classmethod
    def addInputOptions(cls, aggregator, **kw):
        super(DocumentIO, cls).addOptions(aggregator, subtype = "input", **kw)

    @classmethod
    def addOutputOptions(cls, aggregator, **kw):
        super(DocumentIO, cls).addOptions(aggregator, subtype = "output", **kw)

    @classmethod
    def findDynamicOptionMapping(cls, d = None, filters = None):
        global _IO_LIB
        if d is None: d = {}
        if "input" in filters:
            d.update(dict([(name, cls) for name, (cls, input, output) in _IO_LIB.items()
                           if input]))
        elif "output" in filters:
            d.update(dict([(name, cls) for name, (cls, input, output) in _IO_LIB.items()
                           if input]))
        else:
            d.update(dict([(name, cls) for name, (cls, input, output) in _IO_LIB.items()]))
        if "excluderaw" in filters:
            try:
                del d["raw"]
            except KeyError:
                pass
        return d
    
#
# All file-based IO modules.
#

# SAM 1/12/11: I'm going to add a capability to transform the document
# before it's written, or after it's read. This is to do things like
# promote annotations from ENAMEX type=X to X or other conversions.
# I'm going to do it on a document basis because I need to convert the
# annotation types, etc. This means that (a) the deserialize method
# must get a tag table, not the task, because the tag table might be
# different; and (b) the reader seed has to be handled carefully, because
# it shouldn't be polluted with the "incorrect" annotation structure, and
# (c) side effect changes in writers won't make any difference.

# The way you use the convertor is to make an instance of it when
# the file IO is created. This means that you have to register a separate
# file IO which is the combination of the convertor and the IO. This doesn't
# bother me; I don't have time or resources to promote the convertor to
# the command line, and that degree of flexibility isn't necessary anyway,
# because we'll be working with stable sources of data.

# Because we now need to copy the input to the output for inputConvert()
# we can no longer have a default null operator - it has to know that
# it's a null operator. 

class DocumentConvertor(object):

    def __init__(self):
        self.annotationTypeRepository = self.createAnnotationTypeRepository()

    def createAnnotationTypeRepository(self):
        return None

    def getInputTagTable(self, taskTagTable):
        return taskTagTable

    def deserializeAndConvert(self, docIO, source, seedDocument):
        if hasattr(self, "inputConvert"):
            d = AnnotatedDoc(globalTypeRepository = self.annotationTypeRepository)
            docIO.deserialize(source, d)
            self.inputConvert(d, seedDocument)
            # Make sure that the phasesDone are preserved. Other metadata
            # we probably don't care about.
            if "phasesDone" in d.metadata:
                # Is this right? Should it be a set union? I'm going
                # to say no, for the moment.
                seedDocument.metadata["phasesDone"] = d.metadata["phasesDone"][:]
        else:
            docIO.deserialize(source, seedDocument)

    # No longer called for side effect. A seed document is now
    # passed. The targetDoc will have the global repository
    # if available.

    # Not defined unless it's needed. Can't think of
    # another way of indicating that it's a no-op.
    
    # def inputConvert(self, sourceDoc, targetDoc)

    # Changes by side effect.

    def perhapsConvert(self, annotDoc):
        if hasattr(self, "outputConvert"):
            d = AnnotatedDoc(globalTypeRepository = self.annotationTypeRepository)
            # Make sure that the phasesDone are preserved. Other metadata
            # we probably don't care about.
            self.outputConvert(annotDoc, d)
            if "phasesDone" in annotDoc.metadata:
                d.metadata["phasesDone"] = annotDoc.metadata["phasesDone"][:]
            return d
        else:
            return annotDoc

    # def outputConvert(self, annotDoc, targetDoc)

#
# You should be able to register a convertor for a given task.
#

# This can only live on children of DocumentSourceIO.

class RegistrationError(Exception):
    pass

_FILEIO_CONVERTORS = {}

def registerConvertorInstructionSet(ioName, taskName, cls = None, file = None, xml = None):
    global _FILEIO_CONVERTORS
    iocls = getDocumentIOClass(ioName)
    if not issubclass(iocls, DocumentSourceIO):
        raise RegistrationError("convertors can be registered only on DocumentSourceIO classes")
    if not (cls or file or xml):
        raise RegistrationError("none of cls or file or xml are provided")
    if ((cls and file) or (cls and xml) or (xml and file)):
        raise RegistrationError("only one of cls or file or xml may be provided")
    # What goes in here is a function which returns an instruction set engine.
    if cls:
        if not (issubclass(cls, DocumentInstructionSetEngine) or (issubclass(cls, DocumentConvertor))):
            raise RegistrationError("cls must be a subclass of DocumentInstructionSetEngine or DocumentConvertor")
        _FILEIO_CONVERTORS[(iocls, taskName)] = cls
    elif file:
        if not os.path.isabs(file):
            raise RegistrationError("file must be an absolute pathname")
        _FILEIO_CONVERTORS[(iocls, taskName)] = lambda: DocumentInstructionSetEngine(instructionSetFile = file)
    else:
        _FILEIO_CONVERTORS[(iocls, taskName)] = lambda: DocumentInstructionSetEngine(instructionSetXML = xml)

# And now, the convertible class itself.

class DocumentSourceIO(DocumentIO):

    def __init__(self, task = None, convertor = None, **kw):
        self.truncateOnUpdate = True
        self.task = task
        self.convertor = convertor
        if (convertor is None) and task:
            instructionSetEngineFactory = _FILEIO_CONVERTORS.get((self.__class__, task.name))
            if instructionSetEngineFactory is not None:
                isEngine = instructionSetEngineFactory()
                if isinstance(isEngine, DocumentConvertor):
                    self.convertor = isEngine
                else:
                    self.convertor = DocumentInstructionSetEngineConvertor(isEngine)

    #
    # Reading
    #

    # This needs to be implemented in the child.

    def readFromSource(self, source, sourceName = None, **kw):
        raise NotImplementedError

    # This "frame" is independent of the source type, but does the appropriate
    # job of wrapping up the deserialization in the appropriate seed document.
    
    # It turns out that there are some formats which we can't
    # just read bytes and then take apart. 

    def deserializeFromSource(self, source, seedDocument = None, taskSeed = None,
                              update = False):
        taskSeed = taskSeed or self.task
        if seedDocument and self.convertor and not (update and self.truncateOnUpdate):
            raise LoadError("Can't have a convertor and a seed document if you're not truncating")
        if taskSeed:
            globalRepository = taskSeed.getAnnotationTypeRepository()
        else:
            globalRepository = None
        if seedDocument is not None:
            if update and self.truncateOnUpdate:
                seedDocument.truncate()
            if taskSeed:
                if not seedDocument.atypeRepository.globalTypeRepository:
                    seedDocument.atypeRepository.globalTypeRepository = globalRepository
                elif seedDocument.atypeRepository.globalTypeRepository != globalRepository:
                    raise LoadError("global type repository for seed document doesn't match task")
        else:
            seedDocument = AnnotatedDoc(globalTypeRepository = globalRepository)
        # If there's a convertor, we REALLY don't want to use the global
        # type repository - we want to find whatever we can. But then
        # inputConvert maybe should migrate from one document to another?
        # Yes. So we create an empty document, without a repository,
        # and deserialize IT.
        if self.convertor:
            self.convertor.deserializeAndConvert(self, source, seedDocument)
        else:
            self.deserialize(source, seedDocument)
            
        # In MAT 3.0, phasesDone is going to be phased out, so to speak.
        # For the moment, though, we need to infer it, and the fact is that
        # the phase names in 2.0 were steps, while they're annotation sets
        # in 3.0. So I think what we need to do is take a look at the taskSeed,
        # and see if the phasesDone are a subset of the annotation sets.
        # If they're not, kill the phasesDone and redo.

        if taskSeed and ("phasesDone" in seedDocument.metadata) and \
           not (set(seedDocument.metadata["phasesDone"]) < taskSeed.getAnnotationTypeRepository().getAnnotationSetNames()):
            # Wipe it and start over.
            del seedDocument.metadata["phasesDone"]

        # More 2.0 retrofitting: the SEGMENT annotations in 2.0 were
        # global; here, they're for a given managed set. So I need to
        # look for segment annotations which have no step value, and
        # update them if there's a single managed set in the task.

        if taskSeed:
            segments = seedDocument.getAnnotations(["SEGMENT"])
            if segments:
                allSets = set([segment.get("set") for segment in segments])
                # If it's just None, then we update. If it's
                # something else, but one of them is None, then it's a bug.
                if None in allSets:
                    if len(allSets) == 1:
                        # Get the managed steps. If there's more than one, we can't
                        # assign. But there should only be one in the 2.0 tasks.
                        managedSets = taskSeed.getManagedSets()
                        if len(managedSets) != 1:
                            # REMOVE THE SETS. Seems like the best bad answer.
                            seedDocument.removeAnnotations(["SEGMENT"])
                        else:
                            for s in segments:
                                s["set"] = list(managedSets)[0]
                    else:
                        raise LoadError("found document with some SEGMENTs with sets, some without")
        
        # If we've got a task, we should try to infer the phases
        # done, if not available. The same step type may be checked
        # more than once, but because the effects of each step with the
        # same name should be the same, there should be no harm done,
        # because the steps are a set.
        
        # And, now, the steps are truly global. So we don't need to look
        # at the workflows.
        
        if taskSeed is not None and ("phasesDone" not in seedDocument.metadata):
            taskSeed._inferPhasesDone(seedDocument)

        return seedDocument

    # This must be implemented by the children. 
    # annotDoc is an annotated document.
    
    def deserialize(self, source, annotDoc):
        raise NotImplementedError

    #
    # Writing
    #

    def writeToTarget(self, annotDoc, target):
        raise NotImplementedError

# This is the class that expects, ultimately, to be able to read
# a Unicode string from a stream or file.
    
class DocumentFileIO(DocumentSourceIO):

    NON_BMP_WARN = "warn"
    NON_BMP_SCRUB_OR_WARN = "scrub_or_warn"
    NON_BMP_FAIL = "fail"
    NON_BMP_IGNORE = "ignore"

    def __init__(self, encoding = None, handle_non_bmp = None, **kw):
        # The command line tool may provide an explicit None
        # as an argument, which is almost never a good thing
        if encoding is None:
            encoding = 'utf-8'
        self.encoding = encoding
        self.nonBMPHandling = handle_non_bmp or MATConfig.get("HANDLE_NON_BMP") or "warn"
        DocumentSourceIO.__init__(self, **kw)

    # We can read from a byte sequence, or from a Unicode string,
    # or from a file.
    
    def readFromSource(self, source, sourceName = None, **kw):
        closeIt = False        
        if hasattr(source, "readline"):
            # Already a file pointer. If it's been
            # open with codecs.open, or it's sys.stdin, it'll have
            # an encoding set. But sys.stdin only reads byte sequences,
            # not Unicode strings, unlike codecs.open.
            fp = source
            sourceName = sourceName or "<unknown>"
        elif source == "-":
            fp = sys.stdin
            sourceName = sourceName or "<stdin>"
        else:
            fp = codecs.open(source, "r", self.encoding)
            sourceName = sourceName or source
            closeIt = True
        try:
            if fp is sys.stdin:
                # read from a byte stream, but it's got an encoding.
                annotDoc = self.readFromByteSequence(fp.read(), encoding = fp.encoding, sourceName = sourceName, **kw)
            elif fp.encoding is not None:
                annotDoc = self.readFromUnicodeString(fp.read(), sourceName = sourceName, **kw)
            else:
                annotDoc = self.readFromByteSequence(fp.read(), sourceName = sourceName, **kw)
            if closeIt:
                fp.close()
        except Exception as e:
            if ExecutionContext._DEBUG:
                raise
            else:
                if isinstance(e, UnicodeDecodeError):
                    eStr = self._unicodeErrorString(e, fp.encoding or self.encoding)
                else:
                    eStr = str(e)
                raise LoadError("Error loading from source " + str(source) + ": " + eStr)
        return annotDoc

    # Internal function.
    
    def _unicodeErrorString(self, e, curEncoding):
        cands = set(["ascii", "latin1", "utf-8", "windows-1252"])
        if curEncoding:
            cands.discard(curEncoding)
        return "The document doesn't appear to be in the expected encoding (%s). The offending character sequence starts at character %d and ends at character %d. Try a different encoding (for English, other likely candidates are %s)." % (curEncoding, e.start, e.end, ", ".join(cands))

    # Note that in Python 3, sys.stdin produces a Unicode string, not
    # a byte sequence. So the code above in readFromSource() is wrong, but
    # it's properly compensated for here. Writing is another matter.
    
    def readFromByteSequence(self, s, encoding = None, **kw):
        if type(s) is type(u''):
            return self.readFromUnicodeString(s, **kw)
        else:
            try:
                return self.readFromUnicodeString(s.decode(encoding or self.encoding), **kw)
            except UnicodeDecodeError as e:                
                # Catch this here as well as in readFromSource(), just in case.
                raise LoadError(self._unicodeErrorString(e, encoding or self.encoding))

    # In Python 3, there's an increased chance that having non-BMP characters
    # in the input string will cause problems. I should warn about this, by default.
    # The raw IO will do something different. The problem is that we don't know
    # whether we're in Python 2 or Python 3. In Python 3, or in a Python 2
    # compilation where UCS-4 is enabled, all we need to do is replace all characters
    # above a certain code point with the replacement character. In Python 2 without
    # UCS-4, we have to look for surrogate pairs.

    # First, let's robustly test in Python whether we have surrogate pairs or not.
    SURROGATES_USED = (len(u'\U00010000') == 2)
    
    # Here, s is a Unicode string.
    def _checkNonBMP(self, doc, sourceName):
        s = doc.signal
        if not self.SURROGATES_USED:
            # I want to collect these as a list of strings, because
            # that's how I can do the substitution.
            nonBMPFragments = [(c, i) for (i, c) in enumerate(s) if c > u'\uffff']
            nonBMPCodePoints = [ord(c[0]) for c in nonBMPFragments]
        else:
            # See if there are any high surrogates.
            highSurrogates = [c for c in s if c >= u'\ud800' and c <= u'\udbff']
            nonBMPCodePoints = []
            nonBMPFragments = []
            if highSurrogates:
                # I need to put the code points together
                i = 0
                for c in s:
                    if c in highSurrogates:
                        # Formula for surrogate pair -> code point.
                        lowSurrogate = ord(s[i+1]) - 0xdc00
                        highSurrogate = ord(s[i]) - 0xd800
                        codePoint = 0x10000 + lowSurrogate + (2**10 * highSurrogate)
                        nonBMPCodePoints.append(codePoint)
                        nonBMPFragments.append((s[i:i+2], i))
                    i += 1
        if nonBMPCodePoints:
            self.handleNonBMP(doc, sourceName, nonBMPCodePoints, nonBMPFragments)
    
    def handleNonBMP(self, doc, sourceName, codePoints, fragmentPairs):
        msg = "Warning: signal of document %s contains characters outside the Unicode Basic Multilingual Plane, at approximate offsets %s. See the 'Unicode issues' section of the MAT documentation for details." % (sourceName, ", ".join(["%d" % p[1] for p in fragmentPairs]))
        if self.nonBMPHandling == self.NON_BMP_FAIL:
            raise LoadError(msg)
        elif self.nonBMPHandling != self.NON_BMP_IGNORE:
            # warn or scrub
            print(msg, file = sys.stderr)
    
    # As a consequence of the new convertors, it will not be permitted
    # to have a convertor AND a seedDocument if the document isn't truncated.
    
    def readFromUnicodeString(self, s, seedDocument = None, taskSeed = None,
                              update = False, sourceName = "<unknown>"):
        if type(s) is not type(u''):
            raise LoadError("for %s, input to Unicode deserialization is not Unicode" % sourceName)
        doc = self.deserializeFromSource(s, seedDocument = seedDocument,
                                         taskSeed = taskSeed, update = update)
        if self.nonBMPHandling != self.NON_BMP_IGNORE:
            self._checkNonBMP(doc, sourceName)
        return doc
    
    #
    # And now, the writing
    #
    
    def writeToTarget(self, annotDoc, target):
        closeIt = False
        if hasattr(target, "writelines"):
            fp = target
        elif target == "-":
            fp = sys.stdout
        else:
            # If the file doesn't exist, and self.encoding is not known,
            # the file will be created anyway, which is a Bad Thing.
            # So let's look up the encoding first, and if it's not there,
            # it'll throw the error BEFORE I try to open the file.
            codecs.lookup(self.encoding)
            fp = codecs.open(target, "w", self.encoding)
            closeIt = True
        try:
            if self.convertor:
                annotDoc = self.convertor.perhapsConvert(annotDoc)
            if fp is sys.stdout:
                # write to a byte stream, but it's got an encoding.
                # Should I always treat stdout as a byte stream? That
                # means that if you cat it to another file,
                # the contents of the file will be appropriately
                # encoded, which is good.
                if Py3:
                    # In Python 3, stdout is a Unicode output stream.
                    fp.write(self.writeToUnicodeString(annotDoc))
                else:
                    fp.write(self.writeToByteSequence(annotDoc))
            elif fp.encoding is not None:
                fp.write(self.writeToUnicodeString(annotDoc))
            else:
                fp.write(self.writeToByteSequence(annotDoc))
            fp.flush()
        except UnicodeEncodeError as e:
            if ExecutionContext._DEBUG:
                raise
            else:
                raise SaveError(str(e))
        if closeIt:
            fp.close()

    def writeToByteSequence(self, annotDoc, encoding = None):
        return self.writeToUnicodeString(annotDoc).encode(encoding or self.encoding)

    # Returns a Unicode string.
    
    def writeToUnicodeString(self, annotDoc):
        raise NotImplementedError

#
# JSON reading/writing
#

class JSONDocumentIO(DocumentFileIO):

    def __init__(self, encoding = None, legacyWriter = False, **kw):
        # ALWAYS utf-8.
        self.legacyWriter = legacyWriter
        DocumentFileIO.__init__(self, 'utf-8', **kw)

    def readFromByteSequence(self, s, encoding = None, **kw):
        # Never allow the encoding to be changed.
        return DocumentFileIO.readFromByteSequence(self, s, **kw)

    # On Windows, if you have an emergency and the server has failed
    # and you save the offered MAT-JSON as UTF-8, like you should, you're likely to get a Unicode
    # byte order mark at the beginning, because Windows tools are really, really dumb.
    # See https://docs.microsoft.com/en-us/windows/win32/intl/using-byte-order-marks
    
    def deserialize(self, s, annotDoc):
        try:
            d = json.loads(s)
        except ValueError:
            if s and s[0] == u'\ufeff':
                # Try it again.
                try:
                    d = json.loads(s[1:])
                except ValueError:
                    raise LoadError("input doesn't appear to be JSON")
            else:
                raise LoadError("input doesn't appear to be JSON")
        self._deserializeFromJSON(d, annotDoc)

    # This is seldom used as a separate function, but the
    # WebClient needs it.

    # One problem is that JSON, itself, doesn't have a distinction
    # between float and int, but I do. In the browser, if you type
    # 6.0, it gets evaluated as 6. So every value associated with
    # a float slot needs to be coerced to a float, if it's an int.
    # This should happen on decode.
    
    def _deserializeFromJSON(self, d, annotDoc):
        # Check for the version. If it's greater than 1, barf.
        version = d.get("version", 1)
        if version > 2:
            raise LoadError("MAT-JSON version is later than version 2")
        try:
            annotDoc.signal = d["signal"]
        except KeyError:
            pass
        try:
            annotDoc.metadata = d["metadata"]
        except KeyError:
            pass

        # Deal with the float thing.
        def make_float(x):
            if (x is not None) and (type(x) is int):
                return float(x)
            else:
                return x
            
        if "asets" in d:
            annotMap = {}
            aPairs = []
            for adir in d["asets"]:
                hasSpan = adir.get("hasSpan", True)
                hasID = adir.get("hasID", False)
                t = annotDoc.findAnnotationType(adir["type"], hasSpan = hasSpan)

                # Retrofitting for MAT 2.0: SEGMENTs no longer
                # have a few attributes that they had in
                # 2.0, and they're still locked for attribute adding. So
                # discard if we have any attributes other than the
                # ones present in 3.0, and the ones required
                # for retrofitting. If the doc wasn't created with
                # a global annotation table, however, don't bother with this.

                if (adir["type"] == "SEGMENT") and annotDoc.atypeRepository.globalTypeRepository:
                    # Map the "attrs" and "annots" lists.
                    attrMask = []
                    newAttrs = []
                    foundSet = False
                    i = 0
                    indices = (hasID and [0, 1, 2]) or [0, 1]
                    for k in adir["attrs"]:
                        if type(k) not in _strtypes:
                            nm = k["name"]
                        else:
                            nm = k
                        if nm == "set":
                            # If we find "set", the chances are this
                            # has already been updated, because that's a 3.0-specific
                            # feature.
                            foundSet = True
                        if nm in t.attr_table:
                            newAttrs.append(k)
                            attrMask.append(i + len(indices))
                        i += 1
                    if not foundSet:
                        # Probably not a 3.0 segment entry; replace the
                        # attrs and mask the annots.
                        adir["attrs"] = newAttrs
                        attrMask = indices + attrMask
                        adir["annots"] = [[v[i] for i in attrMask] for v in adir["annots"]]                            
                
                # TEMPORARY SHIM.
                attrIndices = None
                attrTypes = None
                annotIndices = None
                digesters = None
                annotDigesters = None
                maybeAnnotVals = False
                if adir["attrs"]:
                    attrIndices = []
                    annotIndices = []
                    digesters = []
                    annotDigesters = []
                    for attr in adir["attrs"]:
                        if type(attr) in _strtypes:
                            tName = attr
                            tType = tAggr = None
                        else:
                            tName = attr["name"]
                            tType = attr.get("type")
                            tAggr = attr.get("aggregation")
                        # When we decode, we have to make sure that
                        # the attributes are ordered appropriately. They may
                        # have been defined otherwise for other documents.
                        # And this will be a threading problem, of course...
                        # This should no longer be a problem, since the
                        # annotations are now local to a document.
                        attrIndex = t.ensureAttribute(tName, aType = tType, aggregation = tAggr)
                        attrIndices.append(attrIndex)
                        # We need to deal with the float/int issue here.
                        if tType == "annotation":
                            # The IDs are all strings. We need to look
                            # them up in the annotMap. Eventually.
                            maybeAnnotVals = True
                            annotIndices.append(attrIndex)
                            if tAggr == "list":
                                digesters.append(lambda x: ((x is not None) and AttributeValueList([annotMap[v] for v in x])) or None)
                            elif tAggr == "set":
                                digesters.append(lambda x: ((x is not None) and AttributeValueSet([annotMap[v] for v in x])) or None)
                            else:
                                digesters.append(lambda x: ((x is not None) and annotMap[x]) or None)
                            annotDigesters.append(digesters[-1])
                        elif tType == "float":
                            if tAggr == "list":
                                digesters.append(lambda x: ((x is not None) and AttributeValueList([make_float(v) for v in x])) or None)
                            elif tAggr == "set":
                                digesters.append(lambda x: ((x is not None) and AttributeValueSet([make_float(v) for v in x])) or None)
                            else:
                                digesters.append(make_float)
                        elif tAggr == "set":
                            digesters.append(lambda x: ((x is not None) and AttributeValueSet(x)) or None)
                        elif tAggr == "list":
                            digesters.append(lambda x: ((x is not None) and AttributeValueList(x)) or None)
                        else:
                            digesters.append(lambda x: x)
                # Do postponement of creating attributes, just in case
                # there are annotation-valued attributes. And this has
                # to be done all the way to the end of the asets loop.
                for a in adir["annots"]:
                    aI = 0
                    if hasSpan:
                        newAnnot = annotDoc.createAnnotation(a[0], a[1], t, blockAdd = True)
                        aI = 2
                    else:
                        newAnnot = annotDoc.createSpanlessAnnotation(t, blockAdd = True)
                    if hasID:
                        if a[aI] is not None:
                            newAnnot.setID(a[aI])
                            annotMap[a[aI]] = newAnnot
                        aI += 1
                    # If you have some attributes, loop through
                    # all the attr values, and add it at the appropriate
                    # index. Postpone if there may be annotation-valued attributes.
                    # Actually, postpone ONLY the annotation-valued attributes themselves;
                    # otherwise, if you have an annotation with an effective label
                    # AND annotation-valued attributes, if it's a value of
                    # some other annotation-valued attribute which expects the
                    # effective label to be set, you'll be hosed.
                    # Also, we have to be careful here because if the
                    # valu is null, we shouldn't set it. The defaults have
                    # already applied when the annotation is created, and we don't
                    # want to override them. This works differently in Javascript, where
                    # the annotation is created with the features, and the defaults
                    # are already applied after the features are added.
                    if attrIndices is not None:
                        if len(a) > (aI + len(adir["attrs"])):
                            raise AnnotationError("MAT-JSON annotation '%s' has more attribute values than the label has attributes" % newAnnot.describe())
                        annotVals = []
                        i = 0
                        for val in a[aI:]:
                            thisIdx = attrIndices[i]
                            if thisIdx in annotIndices:
                                annotVals.append(val)
                            elif val is not None:
                                newAnnot[attrIndices[i]] = digesters[i](val)
                            i += 1
                        if annotVals:
                            aPairs.append((newAnnot, annotVals, annotIndices, annotDigesters))
                    annotDoc._addAnnotation(newAnnot)
            # And now, this is for the annot values. Note that we
            # can have Nones in this list.
            for newAnnot, attrInput, attrIndices, digesters in aPairs:
                i = 0
                for val in attrInput:
                    if val is not None:
                        try:
                            dv = digesters[i](val)
                        except Exception as e:
                            raise AnnotationError("for annotation '%s', attribute %s, and candidate value %s: %s" % (newAnnot.describe(), newAnnot.atype.attr_list[attrIndices[i]].name, val, str(e)))
                        newAnnot[attrIndices[i]] = dv
                    i += 1

    # I need this because of the special role it plays in the Web services.

    # In order to serialize annotations which have IDs and spans, we need
    # to modify this. Then, we'll need to convert annotations in annotation-valued
    # attribute slots into IDs.

    # I've added a legacyWriter flag for my own convenience. All spanless
    # annotations will be dropped, and all annotation-valued attributes will be
    # nulled out,  and all attrs will be strings. Anyone who uses this writer
    # will have to use the fancy operations in brokerAnnotations to absorb the
    # results, of course...

    def _renderAnnotationSingleValue(self, v):
        if v is None:
            return None
        else:
            return v.id

    def _renderSequence(self, v):
        if v is None:
            return v
        else:
            return list(v)

    def _renderAnnotationSequence(self, v):
        if v is None:
            return v
        else:
            return [a.id for a in v]
    
    def renderJSONObj(self, annotDoc):
        d = {"signal": annotDoc.signal,
             "metadata": annotDoc.metadata,
             "asets": []}
        if not self.legacyWriter:
            d["version"] = 2
        asets = d["asets"]
        for aType, annots in annotDoc.atypeDict.items():
            hasSpan = aType.hasSpan
            # Don't use any() here - the list comprehension evaluates
            # everything.
            hasID = False
            for a in annots:
                if a.id is not None:
                    hasID = True
                    break
            aD = {"type": aType.lab}
            if self.legacyWriter:
                if not hasSpan:
                    continue
                aD["attrs"] = [a.name for a in aType.attr_list]
                # Anything that's an aggregation or a non-string value should be None
                meths = []
                for attr in aType.attr_list:
                    if (attr.aggregation is None) and isinstance(attr, StringAttributeType):
                        meths.append(lambda x: x)
                    else:
                        meths.append(lambda x: None)
                aD["annots"] = [[annot.start, annot.end] +
                                [meth(v) for (meth, v) in zip(meths, annot.attrs)]
                                for annot in annots]
            else:
                # Just store the type and aggregation. All the WFC stuff has already
                # been checked.
                attrList = aType.attr_list
                aD.update({"hasID": hasID,
                           "hasSpan": hasSpan,
                           "attrs": [{"name": t.name, "type": t._typename_, "aggregation": t.aggregation}
                                     for t in attrList]})
                # Assume the annotation is well-formed.
                meths = []
                for attr in attrList:
                    if isinstance(attr, AnnotationAttributeType):
                        if (attr.aggregation is None):
                            meths.append(self._renderAnnotationSingleValue)
                        else:
                            meths.append(self._renderAnnotationSequence)
                    elif attr.aggregation is not None:
                        meths.append(self._renderSequence)
                    else:
                        meths.append(lambda x: x)
                if hasID and hasSpan:                
                    aD["annots"] = [[annot.start, annot.end, annot.id] + 
                                    [meth(v) for (meth, v) in zip(meths, annot.attrs)]
                                    for annot in annots]
                elif hasID:
                    aD["annots"] = [[annot.id] + [meth(v) for (meth, v) in zip(meths, annot.attrs)]
                                    for annot in annots]
                elif hasSpan:
                    aD["annots"] = [[annot.start, annot.end] + [meth(v) for (meth, v) in zip(meths, annot.attrs)]
                                    for annot in annots]
                else:
                    aD["annots"] = [[meth(v) for (meth, v) in zip(meths, annot.attrs)] for annot in annots]
            asets.append(aD)
        return d

    def writeToUnicodeString(self, annotDoc):
        # Let's put the actual characters in there. Makes it
        # easier to examine the docs, and they're shorter.
        return _jsondumps(self.renderJSONObj(annotDoc), ensure_ascii = False)

# Make the legacy writer visible to the command line.

class LegacyJSONDocumentIO(JSONDocumentIO):

    def __init__(self, legacyWriter = True, **kw):
        JSONDocumentIO.__init__(self, legacyWriter = True, **kw)

declareDocumentIO("mat-json", JSONDocumentIO, True, True)
declareDocumentIO("mat-json-v1", LegacyJSONDocumentIO, False, True)

#
# Raw reading/writing
#

class RawDocumentIO(DocumentFileIO):

    def handleNonBMP(self, doc, sourceName, codePoints, fragmentPairs):
        # The codePoints are integers, and the fragments are bits of
        # strings. To replace them, you can use the .translate() method.
        # But that only works if we have UCS-4 support; otherwise, we
        # have to use re.
        if self.nonBMPHandling == self.NON_BMP_SCRUB_OR_WARN:
            if self.SURROGATES_USED:
                import re
                fragments = [p[0] for p in fragmentPairs]
                pat = re.compile("|".join(set(fragments)))
                cmap = {c: u'\u25a1' for c in set(fragments)}
                doc.signal = pat.sub(lambda m: cmap[m.group(0)], doc.signal)
            else:
                doc.signal = doc.signal.translate({c: u'\u25a1' for c in codePoints})
        else:
            DocumentFileIO.handleNonBMP(self, doc, sourceName, codePoints, fragmentPairs)
    
    def deserialize(self, s, annotDoc):
        annotDoc.signal = s

    def writeToUnicodeString(self, annotDoc):
        return annotDoc.signal

declareDocumentIO("raw", RawDocumentIO, True, True)

#
# A manager for file pairing, currently used by MATEngine and MATTransducer.
# Eventually, this will be superseded by the corpus manager.
#

# The task is NOT required, but it's passed to all the readers and writers
# if it IS provided.

class ManagerError(Exception):

    pass

class DocumentIOManager(object):

    def __init__(self, task = None):

        self.task = task

    def configure(self, input_file = None, input_dir = None, input_file_re = None,
                  input_encoding = None, input_file_type = None, steps = None, undo_through = None,
                  output_file = None, output_file_type = None, output_dir = None,
                  output_fsuff = None, output_encoding = None,
                  inputFileList = None, inputFileType = None, outputFileType = None,
                  **params):
        
        # Do some sanity checking.
        if inputFileType is None:
            if not inputDocumentIODeclared(input_file_type):
                raise ManagerError("input_file_type must be one of " + ", ".join(["'"+x+"'" for x in allInputDocumentIO()]))
        if (output_file_type is not None) and \
           (not outputDocumentIODeclared(output_file_type)):
            raise ManagerError("output_file_type must be one of " + ", ".join(["'"+x+"'" for x in allOutputDocumentIO()]))
        if ((output_file is not None) or (output_dir is not None)) and (output_file_type is None) and (outputFileType is None):
            raise ManagerError("output_file or output_dir specified without output_file_type")
        if output_file is not None and (not input_file):
            raise ManagerError("Can't specify output_file without input_file")

        if output_file and self._outputIsInWorkspace(os.path.dirname(output_file)):
            raise ManagerError("document manager can't write to a workspace")

        if output_dir and self._outputIsInWorkspace(output_dir):
            raise ManagerError("document manager can't write to a workspace")

        if inputFileList is None:
            if input_file and input_dir:
                raise ManagerError("can't specify both input_file and input_dir.")
            if not (input_file or input_dir):
                raise ManagerError("one of input_file or input_dir must be specified")
        elif (output_file is not None) or (output_dir is not None):
            raise ManagerError("can't specify output file or dir with inputFileList")
        
        self.input_file = input_file
        self.input_dir = input_dir
        self.input_file_re = input_file_re
        self.input_encoding = input_encoding
        self.input_file_type = input_file_type
        self.output_file = output_file
        self.output_file_type = output_file_type
        self.output_dir = output_dir
        self.output_fsuff = output_fsuff
        self.output_encoding = output_encoding
        self.inputFileList = inputFileList
        self.inputFileType = inputFileType
        self.outputFileType = outputFileType
        self.writeable = (input_file and output_file) or (input_dir and output_dir)
        
        if self.inputFileType is None:
            self.inputFileType = getDocumentIO(self.input_file_type, encoding = self.input_encoding,
                                               task = self.task, **params)
        # Load the writer, while I'm at it. A bit perverse, but this is where I get the
        # params.

        if self.writeable:
            if self.outputFileType is None:
                self.outputFileType = getDocumentIO(self.output_file_type, encoding = self.output_encoding,
                                                    task = self.task, **params)
                
    # Utility.
    def _outputIsInWorkspace(self, p):
        # p may be a folder. That means that its dirname would
        # be "folders", and the parent of "folders" would have
        # a properties.txt file.
        return (os.path.basename(os.path.dirname(p)) == "folders") and \
               os.path.isfile(os.path.join(os.path.dirname(os.path.dirname(p)), "ws_db.db"))
    
    def isWriteable(self):
        return self.writeable
   
    # The loading should probably be a generator, but let's not bother with that right now.
    # This yields a list of pairs of <fullpath>, <matdocument>.

    def loadPairs(self, keepGoing = False):

        inputFileList = self._processInputFileList()

        docPairs = []
        skipPairs = []

        if ExecutionContext._DEBUG:
            keepGoing = False
        
        for f in inputFileList:
            if keepGoing:
                try:
                    docPairs.append((f, self._loadFile(f)))
                except Exception as e:
                    skipPairs.append((f, str(e)))
            else:
                docPairs.append((f, self._loadFile(f)))

        return docPairs, skipPairs

    # And here's the long awaited generator. This yields a 3-tuple (<fullpath>, <matdoc>, <error>).
    # error is set if matdoc is None, which is only possible if keepGoing is True.

    def loadPairsIncrementally(self, keepGoing = False):

        inputFileList = self._processInputFileList()

        docPairs = []
        skipPairs = []

        if ExecutionContext._DEBUG:
            keepGoing = False
        
        for f in inputFileList:
            if keepGoing:                
                try:
                    yield (f, self._loadFile(f), None)
                except Exception as e:
                    yield (f, None, str(e))
            else:
                yield (f, self._loadFile(f), None)
        
    def _processInputFileList(self):
        if self.inputFileList is not None:
            return self.inputFileList
        elif self.input_file is not None:
            return [self.input_file]
        else:
            input_file_re = self.input_file_re
            if input_file_re is not None:
                input_file_re = re.compile("^"+input_file_re+"$")
            files = os.listdir(self.input_dir)
            inputFileList = []
            for f in files:
                fullP = os.path.join(self.input_dir, f)
                if (not os.path.isdir(fullP)) and ((input_file_re is None) or (input_file_re.match(f))):
                    inputFileList.append(fullP)
            return inputFileList

    def _loadFile(self, p):
        try:
            return self.inputFileType.readFromSource(p, taskSeed = self.task)
        except IOError:
            if ExecutionContext._DEBUG:
                raise
            else:
                raise ManagerError("Error opening file %s for reading." % p)
        except LoadError as e:
            if ExecutionContext._DEBUG:
                raise
            else:
                raise ManagerError("Load error: " + str(e))

    def outPath(self, fName):
        if self.output_file:
            oFile = self.output_file
        else:
            oFile = os.path.join(self.output_dir, os.path.basename(fName))
            if self.output_fsuff:
                oFile += self.output_fsuff
        return oFile        

    def writeDocument(self, fName, doc):
        if self.output_dir and not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir)
        oFile = self.outPath(fName)
        try:
            self.outputFileType.writeToTarget(doc, oFile)
        except SaveError as e:
            if ExecutionContext._DEBUG:
                raise
            else:
                raise ManagerError("Error saving file %s: %s" % (oFile, str(e)))

#
# DocumentMappings
#

# These are used by document convertors, or by themselves
# (e.g., in MATTransducer). This is an XML-configurable
# object.

# The idea is that the document instruction set has a number of
# methods which filter on labels, or on attributes, or on attribute values.
# The resulting sequences are themselves classes which implement
# various operations and return themselves. So we can chain
# these, and they're always referring back to the original
# instruction set and modifying it.

DIS_DESC = XMLNodeDescFromFile(os.path.join(os.path.dirname(__file__), "doc_convertor_template.xml"))

class ConversionReporter(AnnotationReporter):

    def __init__(self):
        AnnotationReporter.__init__(self, partitionByLabel = False)
        self.addPosition(headerPrefix = "source", concordanceContext = True)
        self.addPosition(headerPrefix = "target", showText = False)
        self.reasons = []
        
    def addRow(self, row):
        if len(row) != 3:
            raise DocumentError("row is different length than positions")
        AnnotationReporter.addRow(self, row[:2])
        self.reasons.append(row[2])

    def getHeadersAndRows(self):
        headers, rows = AnnotationReporter.getHeadersAndRows(self)
        return headers + ["removal reason"], [a + [b] for (a, b) in zip(rows, self.reasons)]

class DocumentInstructionSetEngine(object):

    def __init__(self, instructionSetFile = None, instructionSetXML = None,
                 recordConversions = False):
        self.instructionSetFile = instructionSetFile
        self.instructionSetXML = instructionSetXML
        self.copyMetadata = False
        if instructionSetFile and instructionSetXML:
            raise LoadError("Can't have both instructionSetFile and instructionSetXML")
        if instructionSetFile:
            self._digestInstructionDom(XMLNodeFromFile(instructionSetFile, DIS_DESC))
        elif instructionSetXML:
            self._digestInstructionDom(XMLNodeFromString(instructionSetXML, DIS_DESC))
        self.recordConversions = False
        self.conversionRecorder = None
        self.conversionList = []
        if recordConversions:
            self.enableConversionRecording()

    def enableConversionRecording(self):
        self.recordConversions = True
        self.conversionRecorder = ConversionReporter()

    def _digestInstructionDom(self, domResult):
        # So the game is that the instructions are ALMOST exactly what we
        # call, but it has to be massaged just a bit in places.
        self.instructions = []
        for instr in domResult.orderedChildren:
            instrName = instr.label
            if instrName == "labels":
                self.instructions.append(self._digestLabelInstructions(instr))
            elif instrName == "copy_metadata":
                if instr.attrs.get("keys"):
                    self.copyMetadata = [s.strip() for s in instr.attrs["keys"].split(",")]
                else:
                    self.copyMetadata = True
            else:
                self.instructions.append((instrName, instr.attrs, None))

    def _digestLabelInstructions(self, labelInstructionNode):
        childInstructions = []
        kv = labelInstructionNode.attrs.copy()
        for instr in labelInstructionNode.orderedChildren:
            instrName = instr.label
            if instrName == "with_attrs":
                try:
                    kv["with_attrs"].append(instr.wildcardAttrs)
                except KeyError:
                    kv["with_attrs"] = [instr.wildcardAttrs]
            elif instrName == "of_attr":
                newV = {"label": instr.attrs["label"],
                        "label_re": instr.attrs["label_re"],
                        "attr": instr.attrs["attr"],
                        "attr_re": instr.attrs["attr_re"]}
                try:
                    kv["of_attr"].append(newV)
                except KeyError:
                    kv["of_attr"] = [newV]                                          
            elif instrName == "apply":
                # Make sure the function exists.
                instrAttrs = instr.attrs.copy()
                try:
                    instrAttrs["fn"] = eval(instrAttrs["fn"])
                except:
                    raise LoadError("Can't find function named %s" % instrAttrs["fn"])
                childInstructions.append(("apply", instrAttrs, None))
            elif instrName in ("discard_attrs", "discard_if_null"):
                instrAttrs = instr.attrs.copy()
                if instrAttrs.get("attrs"):
                    instrAttrs["attrs"] = [s.strip() for s in instrAttrs["attrs"].split(",")]
                childInstructions.append((instrName, instrAttrs, None))
            elif instrName == "split_attr":
                instrAttrs = instr.attrs.copy()
                instrAttrs["target_attrs"] = [s.strip() for s in instrAttrs["target_attrs"].split(",")]
                childInstructions.append((instrName, instrAttrs, None))
            elif instrName == "join_attrs":
                instrAttrs = instr.attrs.copy()
                instrAttrs["source_attrs"] = [s.strip() for s in instrAttrs["source_attrs"].split(",")]
                childInstructions.append((instrName, instrAttrs, None))
            elif instrName == "attrs":
                childInstructions.append(self._digestAttrInstructions(instr))
            else:
                childInstructions.append((instrName, instr.attrs, None))
        return ("labels", kv, childInstructions or None)

    def _digestAttrInstructions(self, attrInstructionNode):
        childInstructions = []
        for instr in attrInstructionNode.orderedChildren:
            if instr.label == "values":
                childInstructions.append(self._digestValueInstructions(instr))
            elif instr.label == "split":
                instrAttrs = instr.attrs.copy()
                instrAttrs["target_attrs"] = [s.strip() for s in instrAttrs["target_attrs"].split(",")]
                childInstructions.append((instr.label, instrAttrs, None))                
            else:
                childInstructions.append((instr.label, instr.attrs, None))
        return ("attrs", attrInstructionNode.attrs, childInstructions or None)

    def _digestValueInstructions(self, valueInstructionNode):
        return ("values", valueInstructionNode.attrs,
                [(instr.label, instr.attrs, None) for instr in valueInstructionNode.orderedChildren] or None)

    def _execute(self, sourceDoc, targetDoc):
        instSet = DocumentMappingPair(sourceDoc, targetDoc, self)        
        self.execute(instSet)
        instSet.instantiateAnnotations()
        if self.copyMetadata:
            self._copyMetadata(sourceDoc, targetDoc)

    def _copyMetadata(self, sourceDoc, targetDoc):
        if self.copyMetadata is True:
            # Copy all the keys, except for phasesDone. See MATTransducer.
            keys = set(sourceDoc.metadata.keys()) - set(["phasesDone"])
        else:
            keys = self.copyMetadata
        for k in keys:
            if k in sourceDoc.metadata:
                targetDoc.metadata[k] = self._recursiveCopyMVal(sourceDoc.metadata[k])

    def _recursiveCopyMVal(self, v):
        if type(v) is tuple:
            return tuple([self._recursiveCopyMVal(elt) for elt in v])
        elif type(v) is list:
            return [self._recursiveCopyMVal(elt) for elt in v]
        elif type(v) is dict:
            return dict([(k, self._recursiveCopyMVal(elt)) for (k, elt) in v.items()])
        else:
            return v
    
    # Override execute() if you're writing your instructions by hand.
    
    def execute(self, mappingPair):
        self._executeInstructions(mappingPair, self.instructions)
        
    def recordConversion(self, row):
        self.conversionRecorder.addRow(row)
        self.conversionList.append(row)

    def _executeInstructions(self, bundle, instructions):
        for name, kv, childInstrs in instructions:
            r = getattr(bundle, name)(**kv)
            if childInstrs:
                self._executeInstructions(r, childInstrs)

# Bleah. There are three problems with my original implementation:
# First, it's hard to understand and code against; two, it
# requires a bunch of mostly redundant code; and three, it 
# makes it really, really expensive to figure out the current
# state of an annotation which is a value of an annotation attribute
# (because it doesn't point back to the new label).
# Its main virtue is efficiency. 

# So I need to fix this, in a way that will permit me to 
# do the right thing now, and recode later. 

# OK. The solution is that nobody, but NOBODY, should update
# allEntries except via _incorporateEntry (except the DocumentMappingPair
# initializer, of course). So far, no one does, and I don't
# think I can enforce this, but. I won't have solved the complex
# coding or code duplication problem, but I'm time-limited
# at the moment and there's a LOT of code here. You can only
# do what you can do.

# Actually, I can do one more thing: I can convert a bunch of
# these things into structures, and give them __setattr__ and
# __getattr__ so the current implementation will still work.
# This will give me the opportunity to ensure that no one
# can use the original implementation to update
# the annotation type annots list without
# doing the appropriate updates, for instance.

class DMAnnotationType(object):

    def __init__(self, label, hasSpan, dmAnnots):
        self.label = label
        self.hasSpan = hasSpan
        for annot in dmAnnots:
            annot.atype = self
        self.annots = set(dmAnnots)

    def maybeCopy(self, lab):
        if lab == self.label:
            return self
        else:
            return DMAnnotationType(lab, self.hasSpan, self.annots)

    def incorporateEntry(self, entry):
        if self.hasSpan != entry.hasSpan:
            raise LoadError("hasSpan values collide for label %s" % self.label)
        else:
            for annot in entry.annots:
                annot.atype = self
            self.annots |= entry.annots

    def removeAnnots(self, annots):
        self.annots -= set(annots)
    
    # These are for backward compatibility, until I have
    # a chance to rewrite the code.

    def __getitem__(self, key):
        if key in ("label", "hasSpan", "annots"):
            return getattr(self, key)
        else:
            raise KeyError(key)

    def __setitem__(self, key, val):
        if key in ("label", "hasSpan"):
            setattr(self, key, val)
        else:
            # Can't set the annots this way.
            raise KeyError(key + " not settable")

class DMAnnotation(object):

    def __init__(self, instr, a):
        self.annot = a
        # Note that these values are still source annotation values.
        # Especially sets, lists and annotations.
        self.attrs = dict([(attr.name, val) for (attr, val) in zip(a.atype.attr_list, a.attrs)
                           if val is not None])
        self.touched = False
        self.forceID = False
        self.atype = None
        self.removalReason = None
        instr._dmAnnotationMap[a] = self

    # These are for backward compatibility, until I have
    # a chance to rewrite the code.
    
    def __getitem__(self, key):
        if key in ("annot", "attrs", "touched"):
            return getattr(self, key)
        else:
            raise KeyError(key)

    def __setitem__(self, key, val):
        if key in ("attrs", "touched"):
            setattr(self, key, val)
        else:
            # You can't set the atype this way, or the annot.
            raise KeyError(key + " not settable")

class DocumentMappingPair(object):

    def __init__(self, sourceDoc, targetDoc, engine):
        self.allEntries = {}
        self.engine = engine
        self.signal = sourceDoc.signal
        self.sourceDoc = sourceDoc
        self.targetDoc = targetDoc
        self.discardFailedDuringConversion = None
        self._dmAnnotationMap = {}
        for lab, atype in sourceDoc.atypeRepository.items():
            # Unpack the annotations, so I can operate on them without worrying
            # about wellformedness conditions.
            self.allEntries[lab] = DMAnnotationType(lab, atype.hasSpan,
                                                    [DMAnnotation(self, annot)
                                                     for annot in sourceDoc.getAnnotations(atypes = [lab])])

    #
    # Public filter method
    #

    # with_attrs is a list of dictionaries. The values in the dictionary
    # are values to match, with special handling for annotation labels.
    # of_attr is a list of dictionaries with the keys label, label_re,
    # attr, attr_re which describe the location(s) in which the existing
    # annotations must appear (the CURRENT label and attr, mind you,
    # so, not so efficient yet, because I have to keep mapping back
    # from the true to the DM).

    def labels(self, source = None, source_re = None, excluding = None, excluding_re = None,
               with_attrs = None, of_attr = None):
        if source is not None:
            if source in self.allEntries:
                labs = [source]
            else:
                labs = []
        else:
            if source_re is not None:
                if type(source_re) in _strtypes:
                    # We have to make sure that this is a non-capturing
                    # pattern, because map() may do a replace.
                    source_re = re.compile("^(?:" + source_re + ")$")
                labs = [lab for lab in self.allEntries.keys() if source_re.match(lab)]
            else:
                # Use list() to freeze dictionary keys in Python 3.
                labs = list(self.allEntries.keys())
            # Don't look at excluding if we started with source - makes no sense.
            if excluding is not None:
                try:
                    labs.remove(excluding)
                except ValueError:
                    pass
            elif excluding_re is not None:
                if type(excluding_re) in _strtypes:
                    excluding_re = re.compile("^(" + excluding_re + ")$")
                labs = [lab for lab in labs if not excluding_re.match(lab)]


        if (with_attrs) is None and (of_attr is None):
            # Return a sequence of labels.
            # SAM 1/17/13: I originally thought that I could
            # do this by labels alone, but the problem is that
            # the label sequence uses the GLOBAL cache every
            # time it executes an operation; so if it, e.g.,
            # maps to a label that was already mapped to,
            # the next operations in that block would
            # access all the annotations mapped to that label,
            # not just the ones mapped in that block. So
            # doing this by labels is just not going to
            # work. And I have to make sure to copy the
            # annotation set, so it isn't the same as
            # the one in the reference.
            return _DISLabPairSequence(self, [(lab, list(self.allEntries[lab]["annots"])) for lab in labs], source_re)
        else:
            # Return a sequence of (lab, annotations).
            # As a special case, I'm going to allow
            # the values to be compared to the current labels of
            # annotation-valued attributes. And only the
            # label - no specifying attr-val pairs for
            # the value.
            # For each attrDict, the annotations I test
            # have to match everything.
            # But first, collect the candidate annotations.
            candAttrs = None
            if of_attr:
                # First, if there's an of_attr, collect all
                # the annotations this could possibly be.
                # These are ACTUAL annotations.
                candAttrs = set()                
                for ofAttrEntry in of_attr:
                    self._augmentCandAttrValues(candAttrs, **ofAttrEntry)
            pairs = []            
            for lab in labs:
                annots = []
                for annot in self.allEntries[lab]["annots"]:
                    # If we have some context established, if
                    # the annotation isn't in it, bail. And
                    # if there are no attrs to check, just
                    # accept it and continue.
                    if candAttrs is not None:
                        if annot["annot"] not in candAttrs:
                            continue
                        elif with_attrs is None:
                            annots.append(annot)
                            continue
                    # for each annot, let's cache the converted
                    # values, just in case. The conversion is
                    # the same no matter what.
                    attrCache = {}
                    # For each attrDict, all the values have to match.
                    for attrDict in with_attrs:
                        ok = True
                        for k, v in attrDict.items():
                            try:
                                av, conv = attrCache[k]
                            except KeyError:
                                # Convert them to strings. I imagine
                                # bad things will happen with floats,
                                # but I'm not about to worry about it
                                # at the moment. Lists and sets will never match.
                                av = annot["attrs"].get(k)                                
                                conv = None
                                if (av is None) or isinstance(av, AttributeValueSequence):
                                    ok = False
                                    break
                                else:
                                    if isinstance(av, AnnotationCore):
                                        # Find its current label value.
                                        av = self._dmAnnotationMap[av].atype.label
                                    elif type(av) in (int, float, str):
                                        av = str(av)
                                    elif type(av) is bool:
                                        if av is True:
                                            av = "yes"
                                        else:
                                            av = "no"
                                        conv = lambda x: x.lower()
                                    attrCache[k] = av, conv
                            if conv: v = conv(v)
                            if v != av:
                                ok = False
                                break
                        if ok:
                            # All the key/val pairs work.
                            annots.append(annot)
                            # Break out of the list of with_attrs.
                            break
                if annots:
                    pairs.append((lab, annots))
            return _DISLabPairSequence(self, pairs, source_re)

    def _augmentCandAttrValues(self, candSet, label = None, label_re = None,
                               attr = None, attr_re = None):
        if label is not None:
            if label in self.allEntries:
                entries = [self.allEntries[label]]
            else:
                entries = []
        else:
            if label_re is not None:
                if type(label_re) in _strtypes:
                    label_re = re.compile("^(" + label_re + ")$")
                entries = [e for (lab, e) in self.allEntries.items() if label_re.match(lab)]
            else:
                entries = self.allEntries.values()
        # Now, the attrs.
        for entry in entries:
            okAttrs = set()

            if attr is not None:
                okAttrs.add(attr)
            else:
                if (attr_re is not None) and (type(attr_re) in _strtypes):
                    attr_re = re.compile("^(" + attr_re + ")$")
            for annot in entry["annots"]:
                for k, v in annot["attrs"].items():
                    if not v:
                        continue
                    doIt = False
                    if k in okAttrs:
                        doIt = True
                    elif (attr is None) and ((attr_re is None) or attr_re.match(k)):
                        okAttrs.add(k)
                        doIt = True
                    if doIt:
                        if isinstance(v, AnnotationCore):
                            candSet.add(v)
                        elif isinstance(v, AttributeValueSequence):
                            v = list(v)
                            if isinstance(v[0], AnnotationCore):
                                candSet.update(v)                        

    def discard_untouched(self):
        # Use list() to freeze dictionary keys in Python 3.
        for lab, instr in list(self.allEntries.items()):
            annotsToRemove = [a for a in instr["annots"] if not a["touched"]]
            for a in annotsToRemove:
                a.removalReason = "discard_untouched rule"
            instr.removeAnnots(annotsToRemove)
            if not instr["annots"]:
                del self.allEntries[lab]

    def instantiateAnnotations(self):
        targetDoc = self.targetDoc
        targetDoc.signal = self.signal
        discardFailed = self.discardFailedDuringConversion
        instructionSet = self.allEntries
        annotAttrValuesToAdd = []
        annotMap = {}
        for lab, instr in instructionSet.items():
            lab = instr["label"]
            hasSpan = instr["hasSpan"]
            t = targetDoc.findAnnotationType(lab, hasSpan = hasSpan)
            # Just in case we already have attributes defined.
            nonAnnotAttrs = set()
            annotAttrs = set()
            for attrObj in t.attr_list:
                if attrObj._typename_ != "annotation":
                    nonAnnotAttrs.add(attrObj.name)
            for aBundle in instr["annots"]:
                annot = aBundle["annot"]
                attrSeed = aBundle["attrs"]
                # So how do we copy the annotations? What if
                # there are references? To do the references,
                # I probably need to put off those particular updates.
                # In addition, in order to preserve the annotation IDs
                # from the input, I need to do the nonannotation attributes first.
                # Use list() to freeze dictionary items in Python 3.
                for attr, val in list(attrSeed.items()):
                    if isinstance(val, AnnotationCore):
                        if attr in nonAnnotAttrs:
                            raise LoadError("found attribute type clash during conversion")
                        # Sequence or not, we postpone it.
                        annotAttrValuesToAdd.append((aBundle, annot, attr, val))
                        del attrSeed[attr]                        
                    elif isinstance(val, AttributeValueSequence):
                        if (len(val) > 0) and isinstance(list(val)[0], AnnotationCore):
                            if attr in nonAnnotAttrs:
                                raise LoadError("found attribute type clash during conversion")
                            annotAttrValuesToAdd.append((aBundle, annot, attr, val))
                            del attrSeed[attr]
                        else:         
                            attrSeed[attr] = val.copy()
                            if len(val) > 0:
                                nonAnnotAttrs.add(attr)
                    else:
                        nonAnnotAttrs.add(attr)
                if hasSpan:
                    annotMap[annot] = targetDoc.createAnnotation(annot.start, annot.end, t, attrSeed)
                else:
                    annotMap[annot] = targetDoc.createSpanlessAnnotation(t, attrSeed)
                if annot.id is not None:
                    annotMap[annot].setID(annot.id)
                elif aBundle.forceID:
                    annotMap[annot].getID()

        # We want to have the option of discarding anything that
        # wasn't completely mapped, either because of some
        # selectional restriction or because of an unmapped filler.

        failedSourceAnnots = set()
        
        for dmAnnot, annot, attrName, sourceVal in annotAttrValuesToAdd:
            targetAnnot = annotMap[annot]
            failAnnot = False
            if isinstance(sourceVal, AttributeValueSequence):
                members = list(sourceVal)
                newMembers = []
                for m in members:
                    newMember = annotMap.get(m)
                    if newMember:
                        newMembers.append(newMember)
                    else:                        
                        failAnnot = True
                        dmAnnot.removalReason = "marked for failure because of missing element of sequence value for %s attribute" % attrName
                        break
                if not failAnnot:
                    try:
                        targetAnnot[attrName] = sourceVal.__class__(newMembers)
                    except AnnotationError as e:
                        dmAnnot.removalReason = "marked for failure because of error in setting %s attribute: %s" % (attrName, str(e))
                        failAnnot = True
            else:
                newValue = annotMap.get(sourceVal)
                if newValue:
                    try:
                        targetAnnot[attrName] = newValue
                    except AnnotationError as e:
                        dmAnnot.removalReason = "marked for failure because of error in setting %s attribute: %s" % (attrName, str(e))
                        failAnnot = True
                else:
                    dmAnnot.removalReason = "marked for failure because of missing value for %s attribute" % attrName
                    failAnnot = True
            if failAnnot and discardFailed and (targetAnnot.atype.lab in discardFailed):
                targetDoc.removeAnnotation(targetAnnot)
                failedSourceAnnots.add(annot)

        if self.engine.recordConversions:
            # Report what happened to the annotations.
            for a in self.sourceDoc.getAnnotations():
                if annotMap.get(a) and (a not in failedSourceAnnots):
                    self.engine.recordConversion([a, annotMap[a], None])
                else:
                    # Look up the source annot in the dmAnnotation map,
                    # and see if there's a removal reason.
                    removalReason = None
                    dmA = self._dmAnnotationMap.get(a)
                    if dmA and dmA.removalReason:
                        removalReason = dmA.removalReason
                    self.engine.recordConversion([a, None, removalReason])
                    
    #
    # Utility functions
    #
    
    def _newEntry(self, lab, hasSpan, annots):
        return DMAnnotationType(lab, hasSpan, annots)

    # entry is a DMAnnotationType, as is curEntry.
    def _incorporateEntry(self, lab, entry):
        curEntry = self.allEntries.get(lab)
        if curEntry is None:
            self.allEntries[lab] = entry.maybeCopy(lab)
        else:
            curEntry.incorporateEntry(entry)

    def _convertValueInternal(self, val, targetType):
        if targetType == "int":
            return int(val)
        elif targetType == "float":
            return float(val)
        elif targetType == "boolean":
            # Note that the source can be either a string, in which
            # case we do the string conversion, or a boolean, in
            # which case we keep it.
            if type(val) in _strtypes:
                return ((val == "yes") and True) or False
            elif val:
                # Anything that's not False, e.g. 1
                return True
            else:
                return False
        else:
            return str(val)

    def _convertValue(self, val, targetType, targetAggregation):
        # The incoming values will likely be strings, but you never know.
        # If the incoming value is a list or a set or an AttributeValueList or an AttributeValueSet,
        # we have to convert all the values in the list. THEN do the aggregation conversion.
        if targetAggregation is None:
            tval = type(val)
            if tval is list:
                return AttributeValueList([self._convertValueInternal(v, targetType) for v in val])
            elif tval is set:
                return AttributeValueSet([self._convertValueInternal(v, targetType) for v in val])
            elif tval is AttributeValueList:
                return AttributeValueList([self._convertValueInternal(v, targetType) for v in val])
            elif tval is AttributeValueSet:
                return AttributeValueSet([self._convertValueInternal(v, targetType) for v in val])
            else:
                return self._convertValueInternal(val, targetType)
        elif targetType is None:
            if targetAggregation == "set":
                if type(val) in (list, AttributeValueList, set):
                    return AttributeValueSet(val)
                elif type(val) is not AttributeValueSet:
                    return AttributeValueSet([val])
                else:
                    return val
            elif targetAggregation == "list":
                if type(val) in (list, AttributeValueSet, set):
                    return AttributeValueList(val)
                elif type(val) is not AttributeValueList:
                    return AttributeValueList([val])
                else:
                    return val
            elif targetAggregation == "singleton":
                if type(val) in (list, AttributeValueList):
                    return val[0]
                elif type(val) in (set, AttributeValueSet):
                    return val.pop()
                else:
                    return val
        else:
            if type(val) in (list, set, AttributeValueList, AttributeValueSet):
                val = [self._convertValueInternal(v, targetType) for v in val]                
            if targetAggregation == "set":
                if type(val) is list:
                    return AttributeValueSet(val)
                else:
                    return AttributeValueSet([val])
            elif targetAggregation == "list":
                if type(val) is list:
                    return AttributeValueList(val)
                else:
                    return AttributeValueList([val])
            elif targetAggregation == "singleton":
                if type(val) is list:
                    return val[0]
                else:
                    return val
                
    # Any of target_value, target_type, target, target_aggregation may be null.
    # If there's an attrRE and a target and backreferences in the target, then we need
    # to substitute; similarly, if there's a valueRE and a target_value and backreferences
    # in the target_value, we need to substitute.
    
    def _mapAttrs(self, annotPairs, target, target_value, target_type, target_aggregation, attrRE, valueRE):
        if (target_type is not None) and (target_type not in ('int', 'float', 'boolean', 'string')):
            raise LoadError("unknown target_type '%s' (must be int, float, boolean, string)" % target_type)
        if (target_aggregation is not None) and (target_aggregation not in ("set", "list", "singleton")):
            raise LoadError("unknown target_aggregation '%s' (must be set, list, singleton)" % target_type)
        if (target is None) and (target_type is None) and (target_value is None) and (target_aggregation is None):
            raise LoadError("one of target, target_type, target_value, target_aggregation must be provided")

        # First, map the value if you need to.
        if (target_value is not None) and ((target_type is not None) or (target_aggregation is not None)):
            target_value = self._convertValue(target_value, target_type, target_aggregation)

        replaceAttr = target and attrRE and (re.search(r"\\\d", target) is not None)
        replaceValue = target_value and valueRE and (re.search(r"\\\d", target_value) is not None)

        # So we want to be able to convert existing values from type to type,
        # but what comes in is a different format than what's already set. So
        # we have to be careful.
        
        for annot, keys in annotPairs:
            for key in keys:
                if key in annot["attrs"]:
                    if replaceValue:
                        # We must have a value, and it must match the RE.
                        v = valueRE.sub(target_value, annot["attrs"][key])
                    elif target_value is not None:
                        v = target_value
                    else:
                        v = annot["attrs"][key]
                    if (target_type is not None) or (target_aggregation is not None):
                        v = self._convertValue(v, target_type, target_aggregation)
                    if target is not None:
                        del annot["attrs"][key]
                    if replaceAttr:
                        # We know that they key must match the attrRE.
                        annot["attrs"][attrRE.sub(target, key)] = v
                    else:
                        annot["attrs"][target or key] = v
                    annot["touched"] = True

    def _removeAnnots(self, lab, annots):
        instr = self.allEntries[lab]
        instr.removeAnnots(annots)
        hasSpan = instr.hasSpan
        if not instr.annots:
            del self.allEntries[lab]
        # This may be the only bit of info I need from the old entry.
        return hasSpan

class _DISLabPairSequence(object):

    def __init__(self, dis, labAnnotPairs, sourceRe):
        self.docPair = dis
        self.pairs = labAnnotPairs
        # If the labels were selected by regular expression,
        # this will be a compiled RE.
        self.sourceRe = sourceRe

    # Operations supported: apply, demote, discard, discard_failed, touch, untouch, map,
    # map_attr, map_attrs, promote_attr, discard_attr, split_attr, join_attrs, set_attr,
    # discard_attrs, attrs.

    # For compatibility with DISLabPairSequence, fn accepts an entry
    # and must return a list of entries. The fn takes responsibility for
    # touching, etc.
    
    def apply(self, fn = None):
        if fn is None:
            raise LoadError("fn is required")
        pairs = self.pairs
        self.pairs = []
        pairDict = {}
        entryUpdates = []
        for lab, annots in pairs:
            if annots:
                hasSpan = self.docPair._removeAnnots(lab, annots)
                newEntries = fn(self.docPair._newEntry(lab, hasSpan, annots))
                for newInstr in newEntries:
                    entryUpdates.append(newInstr)
                    try:
                        pairDict[newInstr["label"]] += list(newInstr["annots"])
                    except KeyError:
                        pairDict[newInstr["label"]] = list(newInstr["annots"])
        for newInstr in entryUpdates:
            self.docPair._incorporateEntry(newInstr["label"], newInstr)
        # Use list() to freeze dictionary items in Python 3.
        self.pairs = list(pairDict.items())
        return self

    def demote(self, target_label = None, target_attr = None):
        if target_label is None:
            raise LoadError("target_label is required")
        if target_attr is None:
            raise LoadError("target_attr is required")
        doReplacement = (self.sourceRe is not None) and (re.search(r"\\\d", target_label) is not None)
        pairs = self.pairs
        self.pairs = []
        outAnnots = []
        # Sort them by target candidate.
        if doReplacement:
            d = {}
            for lab, annots in pairs:
                thisTarget = self.sourceRe.sub(target_label, lab)
                try:
                    d[thisTarget].append((lab, annots))
                except KeyError:
                    d[thisTarget] = [(lab, annots)]
            targetPairs = d.items()
        else:            
            targetPairs = [(target_label, pairs)]
        entryUpdates = dict([(t, []) for (t, ignore) in targetPairs])
        for thisTarget, pairs in targetPairs:
            targetOutAnnots = []
            for lab, annots in pairs:
                if annots:
                    targetOutAnnots += annots
                    for a in annots:
                        a["touched"] = True
                        a["attrs"][target_attr] = lab
                    if lab != thisTarget:                    
                        hasSpan = self.docPair._removeAnnots(lab, annots)
                        entryUpdates[thisTarget].append(self.docPair._newEntry(thisTarget, hasSpan, annots))
            if targetOutAnnots:
                outAnnots.append((thisTarget, targetOutAnnots))
        if outAnnots:
            self.pairs += outAnnots
        for thisTarget, instrs in entryUpdates.items():
            for instr in entryUpdates:
                self.docPair._incorporateEntry(thisTarget, instr)
        return self

    def promote_attr(self, source = None):
        if source is None:
            raise LoadError("source is required")
        pairs = self.pairs
        self.pairs = []
        entryUpdates = []
        pairDict = {}
        for lab, annots in pairs:
            # For each annot, if it has the attribute, remove it
            # and promote it.
            annotsToRemove = [annot for annot in annots if (source in annot["attrs"])]
            if annotsToRemove:
                hasSpan = self.docPair._removeAnnots(lab, annotsToRemove)
                for annot in annotsToRemove:
                    targetLabel = annot["attrs"][source]
                    del annot["attrs"][source]
                    annot["touched"] = True
                    entryUpdates.append(self.docPair._newEntry(targetLabel, hasSpan, [annot]))
                    try:
                        pairDict[targetLabel].append(annot)
                    except KeyError:
                        pairDict[targetLabel] = [annot]
            if len(annotsToRemove) < len(annots):
                # There are still some left.
                try:
                    pairDict[lab] += list(set(annots) - set(annotsToRemove))
                except KeyError:
                    pairDict[lab] = list(set(annots) - set(annotsToRemove))
        # Use list() to freeze dictionary items in Python 3.
        self.pairs = list(pairDict.items())
        for instr in entryUpdates:
            self.docPair._incorporateEntry(instr["label"], instr)
        return self

    def discard(self):
        pairs = self.pairs
        self.pairs = []
        for lab, annots in pairs:
            if annots:
                for a in annots:
                    a.removalReason = "discard rule"
                self.docPair._removeAnnots(lab, annots)
        return self

    def discard_if_null(self, attrs = None):
        if attrs is None:
            raise LoadError("attrs is None")
        pairs = self.pairs
        self.pairs = []
        for lab, annots in pairs:
            toRemove = []
            toKeep = []
            for a in annots:
                removed = False
                for attr in attrs:
                    if a["attrs"].get(attr) is None:
                        toRemove.append(a)
                        a.removalReason = "discard_if_null rule"
                        removed = True
                        break
                if not removed:
                    toKeep.append(a)
            if toRemove:
                self.docPair._removeAnnots(lab, toRemove)
            if toKeep:
                self.pairs.append((lab, toKeep))
        return self

    def discard_failed(self):
        labs = [p[0] for p in self.pairs]
        if self.docPair.discardFailedDuringConversion is None:
            self.docPair.discardFailedDuringConversion = set(labs)
        else:
            self.docPair.discardFailedDuringConversion.update(labs)
        return self

    def touch(self):
        for lab, annots in self.pairs:
            for a in annots:
                a["touched"] = True
        return self

    def untouch(self):
        for lab, annots in self.pairs:
            for a in annots:
                a["touched"] = False
        return self

    def force_id(self):
        for lab, annots in self.pairs:
            for a in annots:
                a.forceID = True
        return self

    # If I have a sourceRe, and there are \d patterns in
    # the target, it's a replacement.

    def map(self, target = None):
        if target is None:            
            raise LoadError("target is required")
        doReplacement = (self.sourceRe is not None) and (re.search(r"\\\d", target) is not None)            
        pairs = self.pairs
        self.pairs = []
        outAnnots = []
        # Sort them by target candidate.
        if doReplacement:
            d = {}
            for lab, annots in pairs:
                thisTarget = self.sourceRe.sub(target, lab)
                try:
                    d[thisTarget].append((lab, annots))
                except KeyError:
                    d[thisTarget] = [(lab, annots)]
            targetPairs = d.items()
        else:            
            targetPairs = [(target, pairs)]
        entryUpdates = dict([(t, []) for (t, ignore) in targetPairs])
        for thisTarget, pairs in targetPairs:
            targetOutAnnots = []
            for lab, annots in pairs:
                if annots:
                    for a in annots:
                        a["touched"] = True
                    targetOutAnnots += annots
                    if lab != thisTarget:
                        hasSpan = self.docPair._removeAnnots(lab, annots)                    
                        entryUpdates[thisTarget].append(self.docPair._newEntry(thisTarget, hasSpan, annots))
            if targetOutAnnots:
                outAnnots.append((thisTarget, targetOutAnnots))
        if outAnnots:
            self.pairs += outAnnots
        for thisTarget, instrs in entryUpdates.items():
            for instr in instrs:
                self.docPair._incorporateEntry(thisTarget, instr)
        return self

    def make_spanless(self, demoted_label = None, demoted_attr = None):
        if (demoted_label or demoted_attr) and not (demoted_label and demoted_attr):
            raise LoadError("if demoted_label or demoted_attr are specified, both must be specified")
        doReplacement = (self.sourceRe is not None) and \
                        demoted_label and (re.search(r"\\\d", demoted_label) is not None)
        # We have to be careful here. If any of the entries
        # are subsets of the full list of annotations for that label,
        # we can't do it.
        for lab, annots in self.pairs:
            instr = self.docPair.allEntries.get(lab)
            if (instr is not None) and (set(annots) != instr["annots"]):
                raise LoadError("can't make a subset of a label spanless")
        # Now, do it again.
        newAnnots = {}
        for lab, annots in self.pairs:
            instr = self.docPair.allEntries.get(lab)
            if instr is not None:
                instr["hasSpan"] = False
                for a in instr["annots"]:
                    if (demoted_label is not None) and (a["annot"].start is not None):
                        if doReplacement:
                            thisDemotedLabel = self.sourceRe.sub(demoted_label, lab)
                        else:
                            thisDemotedLabel = demoted_label
                        # Add an annotation. A REAL one in the argument position,
                        # because that's what we currently have, and a
                        # local one for the reconstitution. This will get
                        # copied, so it's kind of wasteful, but that's the
                        # way the code currently works. So use the target doc's
                        # atype (creating if necessary - after all, we WILL
                        # be creating these eventually, unless some later
                        # operation changes the label, in which case there will
                        # be a stray, empty annotation type in the target doc,
                        # which I can live with), but don't add the annotation
                        # to the target doc.
                        trueAnnot = self.docPair.targetDoc.createAnnotation(a["annot"].start, a["annot"].end, thisDemotedLabel, blockAdd = True)
                        dmAnnot = DMAnnotation(self.docPair, trueAnnot)
                        # Touch it!
                        dmAnnot["touched"] = True
                        try:
                            newAnnots[thisDemotedLabel].append(dmAnnot)
                        except KeyError:
                            newAnnots[thisDemotedLabel] = [dmAnnot]
                        a["attrs"][demoted_attr] = trueAnnot
                    a["touched"] = True
        if newAnnots:
            for thisDemotedLabel, theseAnnots in newAnnots.items():
                self.docPair._incorporateEntry(thisDemotedLabel, self.docPair._newEntry(thisDemotedLabel, True, theseAnnots))
        return self

    def map_attr(self, source = None, source_re = None, target = None, target_type = None, target_aggregation = None):
        if (source is None) and (source_re is None):
            raise LoadError("source or source_re is required")
        elif source is not None:
            # We're not going to use this.
            source_re = None
        elif source_re is not None:
            source_re = re.compile("^(?:" + source_re + ")$")
        annotPairs = []
        if source is not None:
            for lab, annots in self.pairs:
                annotPairs += [(a, [source]) for a in annots]
        else:
            okAttrs = set()
            for lab, annots in self.pairs:
                # Note that the same annotation can conceivably be tickled by more than one
                # of these attributes. But I don't want to push the annotation more than once;
                # if I operate on it, I want to operate on it once.
                if annots:
                    tuples = []
                    for a in annots:
                        resAttrs = []
                        for k in a["attrs"].keys():
                            if k in okAttrs:
                                resAttrs.append(k)
                            elif source_re.match(k):
                                # If source was not None, we would have been approved
                                # by okAttrs already. So if source_re is ALSO None,
                                # there's nothing to block approval, so we approve.
                                okAttrs.add(k)
                                resAttrs.append(k)
                        if resAttrs:
                            annotPairs.append((a, resAttrs))
        
        self.docPair._mapAttrs(annotPairs, target, None, target_type, target_aggregation, source_re, None)
        return self

    # This is not available from the XML. It doesn't do anything with regular expressions.
    
    def map_attrs(self, map = None):
        if map is None:
            raise LoadError("map is required")
        allAnnots = []
        for lab, annots in self.pairs:
            allAnnots += annots
        for source, target in map.items():
            self.docPair._mapAttrs([(a, [source]) for a in allAnnots], target, None, None, None, None, None)
        return self

    # Not available from the XML.
            
    def discard_attr(self, source = None):
        if source is None:
            raise LoadError("source is required")
        for lab, annots in self.pairs:
            for annot in annots:
                try:
                    del annot["attrs"][source]
                    annot["touched"] = True
                except KeyError:
                    pass
        return self
    
    def set_attr(self, attr = None, value = None, value_type = None,
                 value_aggregation = None):
        if attr is None:
            raise LoadError("attr is required")
        if value is None:
            raise LoadError("value is required")
        
        # If the aggregation is not None, we really have to create it
        # each time.

        convertedV = None
        if value_aggregation is None:
            convertedV = self.docPair._convertValue(value, value_type, None)

        for lab, annots in self.pairs:
            for annot in annots:
                if value_aggregation is None:
                    annot["attrs"][attr] = convertedV
                else:
                    annot["attrs"][attr] = self.docPair._convertValue(value, value_type, value_aggregation)
                annot["touched"] = True
        
        return self

    def discard_attrs(self, attrs = None, attr_re = None):
        if (attrs is None) and (attr_re is None):
            raise LoadError("one of attrs or attr_re is required")
        elif attrs is not None:
            attr_re = None
        else:
            attr_re = re.compile("^(?:" + attr_re + ")$")
        for lab, annots in self.pairs:
            for annot in annots:
                if attrs is not None:
                    for source in attrs:
                        try:
                            del annot["attrs"][source]
                            annot["touched"] = True
                        except KeyError:
                            pass
                else:
                    # Use list() to freeze dictionary keys in Python 3.
                    for source in list(annot["attrs"].keys()):
                        if attr_re.match(source):
                            del annot["attrs"][source]
                            annot["touched"] = True
        return self

    def split_attr(self, attr = None, target_attrs = None):
        if attr is None:
            raise LoadError("attr is required")
        if target_attrs is None:
            raise LoadError("target_attrs is required")
        for lab, annots in self.pairs:
            for annot in annots:
                v = annot["attrs"].get(attr)
                if v is not None:
                    # If it's a set or list, map each
                    # value to the target attributes. If
                    # it's a singleton, map it to the first one.
                    # Whatever you run out of first is where you stop.
                    del annot["attrs"][attr]
                    annot["touched"] = True
                    if isinstance(v, AttributeValueSequence):
                        v = list(v)
                        for tv, ta in zip(v, target_attrs):
                            annot["attrs"][ta] = tv
                    elif target_attrs:
                        annot["attrs"][target_attrs[0]] = v                            
        return self

    def join_attrs(self, attr = None, source_attrs = None, target_aggregation = None):
        if attr is None:
            raise LoadError("attr is required")
        if source_attrs is None:
            raise LoadError("source_attrs is required")
        if target_aggregation not in ("list", "set"):
            raise LoadError("target_aggregation must be one of list, set")
        for lab, annots in self.pairs:
            for annot in annots:
                newV = []
                found = False
                for sa in source_attrs:
                    v = annot["attrs"].get(sa)
                    if v is not None:
                        found = True
                        if isinstance(v, AttributeValueSequence):
                            newV += list(v)
                        else:
                            newV.append(v)
                        del annot["attrs"][sa]
                if found:
                    if target_aggregation == "set":
                        newV = AttributeValueSet(newV)
                    else:
                        newV = AttributeValueList(newV)
                    annot["attrs"][attr] = newV
                    annot["touched"] = True
        return self
    
    def attrs(self, source = None, source_re = None, excluding = None, excluding_re = None):

        okAttrs = set()
        forbiddenAttrs = set()

        if source is not None:
            okAttrs.add(source)
            # We're never going to use these.
            source_re = excluding = excluding_re = None
        else:
            if excluding is not None:
                forbiddenAttrs.add(excluding)
                excluding_re = None
            if (source_re is not None) and (type(source_re) in _strtypes):
                # Again, because we support backreferences now, this parenthesis has
                # to be guaranteed to be non-grouping.
                source_re = re.compile("^(?:" + source_re + ")$")
            if (excluding is None) and (excluding_re is not None) and (type(excluding_re) in _strtypes):
                excluding_re = re.compile("^(" + excluding_re + ")$")

        labPairs = []
        for lab, annots in self.pairs:
            # Note that the same annotation can conceivably be tickled by more than one
            # of these attributes. But I don't want to push the annotation more than once;
            # if I operate on it, I want to operate on it once.
            if annots:
                tuples = []
                for a in annots:
                    resAttrs = []
                    for k in a["attrs"].keys():
                        if k in okAttrs:
                            resAttrs.append(k)
                        elif k in forbiddenAttrs:
                            continue
                        elif (excluding_re is not None) and excluding_re.match(k):
                            # Here, we have to deal with the case where there's no
                            # source, as well as the excluding case.
                            forbiddenAttrs.add(k)
                            continue
                        elif (source is None) and ((source_re is None) or source_re.match(k)):
                            # If source was not None, we would have been approved
                            # by okAttrs already. So if source_re is ALSO None,
                            # there's nothing to block approval, so we approve.
                            okAttrs.add(k)
                            resAttrs.append(k)
                    if resAttrs:
                        tuples.append((a, resAttrs))
                if tuples:
                    labPairs.append((lab, tuples))
        return _DISAttrSequence(self.docPair, labPairs, source_re, None)

class _DISAttrSequence(object):

    # pairs are (lab, [(annot, keys), ...])
    
    def __init__(self, dis, labAnnotAttrTuples, attrRE, valueRE):
        self.docPair = dis
        self.pairs = labAnnotAttrTuples
        self.attrRE = attrRE
        self.valueRE = valueRE        

    # Operations supported: promote, discard, split, map, values.

    def promote(self):
        # For promote, only use the first attr found. AND you return an
        # empty list, since you can't really chain after this.
        pairs = self.pairs
        self.pairs = []
        entryUpdates = []
        for lab, annotPairs in pairs:
            if annotPairs:
                hasSpan = self.docPair._removeAnnots(lab, [annot for (annot, keys) in annotPairs])
                for annot, keys in annotPairs:
                    k = keys[0]
                    targetLabel = annot["attrs"][k]
                    del annot["attrs"][k]
                    annot["touched"] = True
                    # This deals with the hasSpan stuff transparently, although
                    # it's probably a bit verbose.
                    entryUpdates.append(self.docPair._newEntry(targetLabel, hasSpan, [annot]))
        for instr in entryUpdates:
            self.docPair._incorporateEntry(instr["label"], instr)
        return self                    

    # This is discarding the ATTRIBUTE.
    def discard(self):        
        for lab, annotPairs in self.pairs:
            for annot, keys in annotPairs:
                for k in keys:
                    del annot["attrs"][k]
                annot["touched"] = True
        return self

    def split(self, target_attrs = None):
        if target_attrs is None:
            raise LoadError("target_attrs is required")
        for lab, annotPairs in self.pairs:
            for annot, keys in annotPairs:
                for attr in keys:
                    v = annot["attrs"].get(attr)
                    if v is not None:
                        # If it's a set or list, map each
                        # value to the target attributes. If
                        # it's a singleton, map it to the first one.
                        # Whatever you run out of first is where you stop.
                        del annot["attrs"][attr]
                        annot["touched"] = True
                        if isinstance(v, AttributeValueSequence):
                            v = list(v)
                            for tv, ta in zip(v, target_attrs):
                                annot["attrs"][ta] = tv
                        elif target_attrs:
                            annot["attrs"][target_attrs[0]] = v                            
        return self

    def discard_annot(self):
        for lab, annotPairs in self.pairs:
            toRemove = [annot for (annot, keys) in annotPairs]
            for a in toRemove:
                a.removalReason = "discard_annot rule"
            self.docPair._removeAnnots(lab, toRemove)
        self.pairs = []
        return self

    def discard_annot_if_null(self):
        pairs = self.pairs
        self.pairs = []
        for lab, annotPairs in self.pairs:
            toKeep = []
            toRemove = []
            for annot, keys in annotPairs:
                removed = False
                for key in keys:
                    if annot["attrs"].get(key) is None:
                        toRemove.append(annot)
                        annot.removalReason = "discard_annot_if_null rule"
                        removed = True
                        break
                if not removed:
                    toKeep.append((annot, keys))
            if toRemove:
                self.docPair._removeAnnots(lab, toRemove)
            if toKeep:
                self.pairs.append((lab, toKeep))
        return self

    def values(self, source = None, source_re = None, excluding = None, excluding_re = None):
        
        okVals = set()
        forbiddenVals = set()

        if source is not None:
            okVals.add(source)
            # We're never going to use these.
            source_re = excluding = excluding_re = None
        else:
            if excluding is not None:
                forbiddenVals.add(excluding)
                excluding_re = None
            if (source_re is not None) and (type(source_re) in _strtypes):
                source_re = re.compile("^(?:" + source_re + ")$")
            if (excluding is None) and (excluding_re is not None) and (type(excluding_re) in _strtypes):
                excluding_re = re.compile("^(" + excluding_re + ")$")

        # Now, what do we do? We basically filter this list and return a new one of these.
        # You'll be able to map as a perverse case of attrs(), but that's OK.

        labPairs = []
        for lab, annotPairs in self.pairs:
            tuples = []
            for annot, keys in annotPairs:
                resKeys = []
                for k in keys:
                    # It's possible that this attr no longer exists,
                    # if you're mapping multiple values to some other
                    # attribute and this value has already been mapped.
                    v = annot["attrs"].get(k)
                    # We need to use a very small subset of
                    # values: just int, float, str, unicode, boolean.
                    # Just the ones we can convert to a string.
                    if v is None:
                        continue
                    if (type(v) in _numtypes) or (type(v) is bool) or (type(v) in _strtypes):
                        if v is False:
                            v = "no"
                        elif v is True:
                            v = "yes"
                        elif type(v) in _numtypes:
                            v = str(v)
                    elif isinstance(v, AnnotationCore):
                        v = self.docPair._dmAnnotationMap[v].atype.label
                    elif isinstance(v, AttributeValueSequence):
                        if not v:
                            continue
                        else:
                            v = list(v)
                            if isinstance(v[0], AnnotationCore):
                                v = [self.docPair._dmAnnotationMap[sv].atype.label for sv in v]
                            elif type(v[0]) is bool:
                                v = [(sv and "yes") or "no" for sv in v]
                            v = "|"+",".join([str(sv) for sv in v])+"|"
                    else:
                        continue
                    # Now, evaluate it.
                    if v in okVals:
                        resKeys.append(k)
                    elif v in forbiddenVals:
                        continue
                    elif (excluding_re is not None) and excluding_re.match(v):
                        # Here, we have to deal with the case where there's no
                        # source, as well as the excluding case.
                        forbiddenVals.add(v)                                
                        continue
                    elif (source is None) and ((source_re is None) or source_re.match(v)):
                        # If source was not None, we would have been approved
                        # by okAttrs already. So if source_re is ALSO None,
                        # there's nothing to block approval, so we approve.
                        okVals.add(v)
                        resKeys.append(k)                
                if resKeys:                    
                    tuples.append((annot, resKeys))
            if tuples:
                labPairs.append((lab, tuples))

        return _DISAttrSequence(self.docPair, labPairs, self.attrRE, source_re)

    def map(self, target = None, target_value=None, target_type = None, target_aggregation = None):
        # In this case, we know that the annots have the keys, because they've already
        # been filtered. We don't always know that.
        if self.pairs:
            allAnnotKeyPairs = []
            for lab, annotPairs in self.pairs:
                allAnnotKeyPairs += annotPairs
            self.docPair._mapAttrs(allAnnotKeyPairs, target, target_value, target_type, target_aggregation, self.attrRE, self.valueRE)
        return self

class DocumentInstructionSetEngineConvertor(DocumentConvertor):

    def __init__(self, engine):
        self.engine = engine
        DocumentConvertor.__init__(self)

    def inputConvert(self, sourceDoc, targetDoc):
        self.engine._execute(sourceDoc, targetDoc)
