# Copyright (C) 2007 - 2021 The MITRE Corporation. See the toplevel
# file LICENSE.txt for license terms.

from __future__ import absolute_import
from .PyCompat import _strtypes

import os, sys, traceback, codecs

# MAT imports

from .Operation import (XMLOpArgumentAggregator, OpArgument, DynamicChoiceOption, OptionTemplate)
from . import Error
from . import Document

from . import PluginMgr
from . import DocumentIO
from . import ExecutionContext

class ConfigurationError(Exception):
    pass

class ShortUsageConfigurationError(ConfigurationError):
    pass

class NoUsageConfigurationError(ConfigurationError):
    pass

# No error checking yet.

# Steps can be things like "zone", "tag".

# Input for zone is the signal; output for render is the signal.
# All other inputs and outputs are the annotation set.

# Moving all the configuration digestion into here: finding
# named tasks, loading files, etc.

# New configuration for MATEngine. task is now taskObj, and it's now
# a keyword. The names of the keys have to match the command line.

# I want to be able to default to a single known task, and a single
# known workflow in that task. I think I'll handle that in the engine
# itself. Well, we can't default to a single workflow, because in
# order to default, you need to have computed the operational task,
# which we don't do until runtime. Actually, if the known task
# only has one workflow of its own, it should default to it, but
# there's no guarantee that the child will implement it. So we can default.

class MATEngine(object):

    def __init__(self, taskObj = None, workflow = None, task = None, pluginDir = None,
                 synchronizePhasesDone = True):
        self.taskObj = taskObj
        # By default, I'm going to synchronize the phases done,
        # because it's important. This won't happen in the workspaces,
        # since woe be to whoever meddles there anyway...
        self.synchronizePhasesDone = synchronizePhasesDone
        if self.taskObj is None:
            # We need to find one.
            pluginDir = pluginDir or PluginMgr.LoadPlugins()
            if task is None:
                allTasks = pluginDir.getAllTasks()
                if len(allTasks) == 1:
                    self.taskObj = allTasks[0]
                else:
                    # Nada.
                    raise ShortUsageConfigurationError(self, "task not specified")
            else:
                self.taskObj = pluginDir.getTask(task)
                if self.taskObj is None:
                    raise ShortUsageConfigurationError(self, "unknown task '%s'" % task)
        # At this point, there's a task obj. Later, we'll get the operational task.
        # When we run.
        self.operationalTask = None
        self.workFlow = workflow
        if self.workFlow is None:
            # We'll default to a single workflow, but even that workflow
            # won't be guaranteed to work.
            wfList = self.taskObj.getWorkflows().keys()
            if len(wfList) == 1:
                # In Python 3, dictionary keys are a view and can't be indexed.
                self.workFlow = list(wfList)[0]
            else:
                raise ConfigurationError(self, "workflow must be specified")

    def _ensureOperationalTask(self, steps, language, **params):
        if self.operationalTask is None:
            try:                
                self.operationalTask = self.taskObj.getTaskImplementation(self.workFlow,
                                                                          steps or [], language = language,
                                                                          **params)
                if self.operationalTask is None:
                    raise NoUsageConfigurationError(self, "couldn't find task implementation for the given parameters")
            except KeyError:
                raise NoUsageConfigurationError(self, "couldn't find task implementation for the given parameters")
            except PluginMgr.PluginError as e:
                raise ConfigurationError(self, "Plugin error: %s" % e)

    # Here's a version of the Run() method below which can be called with
    # an aggregator. The aggregator is going to have all sorts of
    # arguments. We need: input_file, input_dir, input_file_re,
    # input_encoding, input_file_type, output_file, output_dir,
    # output_fsuff, output_file_type, output_encoding, workflow, steps,
    # print_steps, undo_through. I suppose I should be adding these
    # by making the MATEngine an option bearer, but that's just not
    # in the cards at the moment.

    INTERNAL_ARGS = OptionTemplate([OpArgument("input_file", hasArg = True),
                                    OpArgument("input_dir", hasArg = True),
                                    OpArgument("input_file_re", hasArg = True),
                                    OpArgument("input_encoding", hasArg = True),
                                    DynamicChoiceOption("--input_file_type", DocumentIO.DocumentIO,
                                                        subtype = "input", storeLiteral = True),
                                    OpArgument("output_file", hasArg = True),
                                    OpArgument("output_dir", hasArg = True),
                                    OpArgument("output_fsuff", hasArg = True),
                                    DynamicChoiceOption("--output_file_type", DocumentIO.DocumentIO,
                                                        subtype = "output", storeLiteral = True),
                                    OpArgument("output_encoding", hasArg = True),
                                    OpArgument("language", hasArg = True),
                                    OpArgument("workflow", hasArg = True),
                                    OpArgument("steps", hasArg = True),
                                    OpArgument("print_steps", hasArg = True),
                                    OpArgument("undo_through", hasArg = True),
                                    OpArgument("mark_gold", hasArg = True),
                                    OpArgument("mark_reconciled", hasArg = True),
                                    OpArgument("fresh_task")
                                    ])
    
    def aggregatorExtract(self, aggregator, failOnFileTypes = False, **params):
        self.INTERNAL_ARGS.addOptions(aggregator)
        # First, start with the task.
        self.taskObj.addOptions(aggregator)
        # Now, we extract.
        return aggregator.extract(defaults = params)

    # The problem here is that there are OTHER arguments to aggregatorRun
    # which aren't part of the internal args, and never come out of
    # a command line parse. This wasn't a problem before, but
    # now it is.
    
    def aggregatorRun(self, aggregator,
                      # These three parameters might be passed in when this is called
                      # programmatically, but not from the command line.
                      inputFileList = None, inputFileType = None, outputFileType = None,
                      **params):
        return self.Run(inputFileList = inputFileList, inputFileType = inputFileType, outputFileType = outputFileType,
                        **self.aggregatorExtract(aggregator, failOnFileTypes = True, **params))

    # The Run method is the most convenient public face of the
    # engine. It allows you to configure all sorts of things about
    # where the files are read from and where they go.

    # The keyword arguments here must match the option argument names
    # on the command line. Some of them are intended only to be used
    # internally, such inputFileList. 

    def Run(self, input_file = None, input_dir = None, input_file_re = None,
            input_encoding = None, input_file_type = None, steps = None, undo_through = None,
            output_file = None, output_file_type = None, output_dir = None,
            output_fsuff = None, output_encoding = None,
            # These three parameters might be passed in when this is called
            # programmatically, but not from the command line.
            inputFileList = None, inputFileType = None, outputFileType = None,
            language = None,
            **params):
        
        # First, preprocess some of the arguments.
        steps = self._maybeDigestList(steps)

        # Make sure we have a task.
        self._ensureOperationalTask(steps, language, **params)

        dm = DocumentIO.DocumentIOManager(task = self.operationalTask)
        try:
            dm.configure(input_file = input_file, input_dir = input_dir,
                         input_file_re = input_file_re,
                         input_encoding = input_encoding, input_file_type = input_file_type, 
                         output_file = output_file, output_file_type = output_file_type,
                         output_dir = output_dir,
                         output_fsuff = output_fsuff, output_encoding = output_encoding,
                         inputFileList = inputFileList,
                         inputFileType = inputFileType,
                         outputFileType = outputFileType,
                         **params)
        except DocumentIO.ManagerError as e:
            if ExecutionContext._DEBUG:
                raise
            else:
                raise ConfigurationError(self, str(e))

        try:
            # Now, create the idata pairs. We're not going to keep going,
            # so we ignore the skipPairs.
            inputPairs, ignore = dm.loadPairs()
        except Exception as e:
            if ExecutionContext._DEBUG:
                raise
            else:
                raise NoUsageConfigurationError(self, str(e))

        # Central call. This is where the work gets done.
        
        iDataPairs = self.RunDataPairs(inputPairs, steps, undoThrough = undo_through, language = language, **params)
        resultTable = {}
        for fname, iData in iDataPairs:
            resultTable[fname] = iData

        # And then the end.

        if dm.isWriteable():
            # The complexity here is that the output may not be in the
            # appropriate format to save. That is, if it's raw, then
            # we have to turn the signal into a document. If the
            # output file type is mat-json, utf-8 is automatically enforced.
            for fname, idata in inputPairs:
                Output = resultTable[fname]
                if type(Output) is type(""):
                    Output = Document.AnnotatedDoc(signal = Output.decode('ascii'))
                elif type(Output) is type(u''):
                    Output = Document.AnnotatedDoc(signal = Output)
                elif not isinstance(Output, Document.AnnotatedDoc):
                    raise NoUsageConfigurationError(self, "Output is neither text nor a document")
                try:
                    dm.writeDocument(fname, Output)
                except Exception as e:
                    if ExecutionContext._DEBUG:                        
                        raise
                    else:
                        raise NoUsageConfigurationError(self, "Error opening file %s for writing: %s" % (dm.outPath(fname), str(e)))

        return iDataPairs

    # OK, time to introduce batch processing. Let's make iData 
    # a little more complicated than before; it should be a sequence
    # of (<fname>, <docobj>) pairs. The idea is that each step
    # can have a tagStep or tagBatchStep method. If it has the
    # latter, I pass in the iDataPairs; otherwise, I loop through
    # the pairs myself.

    # More on the relation between the steps and the steps already
    # in the document: the steps don't all need to be in the workflow,
    # since the things that can be done in a given workflow are a
    # subset of what can be done to a document. The only restriction
    # is that if there's a step to be done which hasn't been
    # done yet, but there are later steps that HAVE been done,
    # you should barf. Well, I can't know this. It may be that
    # there are steps done which logically precede the steps
    # that the engine should do which have been skipped, and
    # we really SHOULD be at the end, but we don't know it.
    # I think we should treat the document phases done as
    # an unordered set, and rely on the system to ensure the
    # operational order.

    def _maybeDigestList(self, v):
        if (v is not None) and (type(v) in _strtypes):
            v = v.split(",")
            if v == ['']:
                v = []
        return v

    def RunDataPairs(self, iDataPairs, steps = None, 
                     undoThrough = None, 
                     mark_gold = None, mark_reconciled = None, fresh_task = False,
                     language = None,
                     **params):

        self._ensureOperationalTask(steps, language, **params)

        if self.workFlow not in self.operationalTask.getWorkflows():
            raise Error.MATError(
                "[init]", "workflow %s not found (choices are %s)" % (self.workFlow, ", ".join(self.operationalTask.getWorkflows().keys())))

        workflow = self.operationalTask.getWorkflows()[self.workFlow]

        # Preprocess some of the arguments.
        mark_gold = self._maybeDigestList(mark_gold)
        mark_reconciled = self._maybeDigestList(mark_reconciled)

        synchronizePhasesDone = self.synchronizePhasesDone

        if fresh_task:
            # Do it now.
            for fname, iData in iDataPairs:
                try:
                    del iData.metadata["phasesDone"]
                except KeyError:
                    pass
                # And also the segments (this doesn't happen for just synchronization).
                iData.removeAnnotations(["SEGMENT"])
                self.operationalTask._inferPhasesDone(iData)
            # Don't do it later.
            synchronizePhasesDone = False

        if undoThrough is not None:
            if not workflow.params["undoable"]:
                raise Error.MATError(
                    "[init]", "workflow %s is not undoable" % self.workFlow)
            try:
                successors = workflow.getStepAndSuccessors(undoThrough)
            except PluginMgr.PluginError:
                raise Error.MATError(
                    "[init]", "no step %s to undo through" % undoThrough)
            successors = successors[:]
            successors.reverse()
            # For step, for each document, if the document has done the
            # step, undo it. We reverse the successors, so we undo them in order.
            # If the current workflow doesn't have the step, we have to
            # use the default step from the step implementation.
            for successorStep in successors:
                try:
                    for fname, iData in iDataPairs:
                        if successorStep.believedAtLeastPartiallyDone(iData):
                            successorStep.undo(iData)
                            successorStep.stepUndone(iData)
                    self.ReportBatchUndoStepResult(successorStep, iDataPairs)
                except Exception as e:
                    if ExecutionContext._DEBUG:
                        raise
                    else:
                        raise Error.MATError(successorStep.stepName, str(e), show_tb = True)

        # Previously, if there were no steps, I was just returning the pairs.
        # But I can't do that now, because I need to deal with mark_gold and mark_reconciled.

        # Also, I'm going to ensure that the phasesDone list is synchronized at this point.
        # Just in case somebody hand-modified a file. It's not cheap, but I think it's
        # important. This happens after undo, because undo is a dynamic probe and we don't
        # need it for that.

        if synchronizePhasesDone:
            for fname, iData in iDataPairs:
                try:
                    del iData.metadata["phasesDone"]
                except KeyError:
                    pass
                self.operationalTask._inferPhasesDone(iData)
        
        if not steps:
            steps = []
        else:
            # Copy it, because we're going to surgically alter it.
            steps = steps[:]
        
        for stepObj in workflow.stepList:
            
            stepName = stepObj.stepName
            prettyName = stepObj.prettyName

            # At this point, the step names are sometimes registered under the step prefix.
            
            if steps and ((steps[0] == stepName) or (prettyName and (steps[0] == prettyName))):

                impl = stepObj.stepImplementation
                localParams = params
                stepParams = params.get(stepName.replace(" ", "_") + "_", None)
                if stepParams:
                    localParams = localParams.copy()
                    del localParams[stepName.replace(" ", "_") + "_"]
                    localParams.update(stepParams)
                
                if impl and impl.runSettings:
                    # So we have to be very, very careful not to have
                    # extra values in either the stepObj params or in the
                    # cmdline params. At one point, the cmdline params
                    # were reporting defaults too aggressively,  and it wasn't
                    # possible to distinguish between an explicitly specified
                    # value and a default value. Now,  it should be cleanly
                    # managed. I've also fixed the same thing for
                    # the step values.
                    runParams = impl.enhanceAndExtract(XMLOpArgumentAggregator(impl.runSettings))
                    # NOW we can override with the command line.
                    for key, val in localParams.items():
                        if val is not None:
                            runParams[key] = val
                    localParams = runParams
                                    
                # Filter the ones which need to be done. 
                pairsToDo = [(fname, iData) for fname, iData in iDataPairs
                             if stepObj.stepCanBeDone(iData)]

                if pairsToDo:
                    fOrder = [fname for fname, iData in iDataPairs]
                    d = dict(iDataPairs)
                    try:
                        pairsDone = stepObj.doBatch(pairsToDo, language = language, **localParams)
                    except Exception as e:
                        if ExecutionContext._DEBUG:
                            raise
                        else:
                            raise Error.MATError(stepName, str(e), show_tb = True)

                    # Update the dictionary and reconstitute iDataPairs.
                    # Only record the step done if the output object is the same as
                    # the input object.
                    for fname, iData in pairsDone:
                        oldData = d[fname]
                        if (iData is oldData) and \
                           isinstance(iData, Document.AnnotatedDoc):
                            stepObj.recordStep(iData)
                        d[fname] = iData

                    # Do this AFTER the step is recorded.
                    self.ReportBatchStepResult(stepObj, pairsDone)
                    iDataPairs = [(fname, d[fname]) for fname in fOrder]
                
                steps[0:1] = []

            # Whether or not there was something to do, I should try to
            # mark them gold or reconciled.

            if (mark_reconciled is not None) and \
               ((stepName in mark_reconciled) or (prettyName in mark_reconciled)):
                if not (stepObj.managedSets <= stepObj.finalSetMentions):
                    raise Error.MATError(stepName, "step can't be marked reconciled because some of its managed sets are modified later in the workflow")
                for (fname, d) in iDataPairs:
                    stepObj.markReconciled(d)
            elif (mark_gold is not None) and \
                 ((stepName in mark_gold) or (prettyName in mark_gold)):
                if not (stepObj.managedSets <= stepObj.finalSetMentions):
                    raise Error.MATError(stepName, "step can't be marked gold because some of its managed sets are modified later in the workflow")
                for (fname, d) in iDataPairs:
                    stepObj.markGold(d)

        if steps:
            raise Error.MATError("[engine]", "Unknown step " + steps[0])

        return iDataPairs

    def ReportBatchStepResult(self, stepObj, iDataPairs):
        # This can be overridden, if desired.
        for fname, iData in iDataPairs:
            self.ReportStepResult(stepObj, fname, iData)

    def ReportBatchUndoStepResult(self, stepObj, iDataPairs):
        # This can be overridden, if desired.
        for fname, iData in iDataPairs:
            self.ReportUndoStepResult(stepObj, fname, iData)

    def ReportStepResult(self, stepObj, fname, iData):
        pass

    def ReportUndoStepResult(self, stepObj, fname, iData):
        pass

