# Copyright (C) 2014 - 2019 The MITRE Corporation. See the toplevel
# file LICENSE.txt for license terms.

from __future__ import print_function

# 7/15/19: Here, we define an annotation tree. This is targeted toward
# applications which need to manipulate the hierarchy of annotations.
# Initial application is for feature engineering for classifiers.
# There's probably a lot of overkill here, since it was developed
# originally for cases where the hierarchy was "sloppy", and
# things could dominate other things if they overlap but
# don't completely enclose, and I'm not going to try to sort
# that out right now.

# Lots of smart properties:

# parent, previous, next
# multiple levels, which you can define
# levels can have overlap; previous/next order is first by smallest start, then by biggest last.
# tokens are always on the bottom.
# document is always at the top.
# you can have elements besides annotations.

# Your parent isn't necessarily the next level up - it's the next level
# that completely encloses you. Every element besides the root (whose
# element is a MAT document) has a parent.

# Oops, we need multiple parents if levels can have overlap.

class TreeError(Exception):
    pass

import sys

ETYPE_ROOT, ETYPE_TOKEN, ETYPE_ANNOT = (0, 1, 2)

from .Annotation import Annotation

class TreeAnnotationCore(object):

    def __init__(self, annot, root, level, etype):
        self.annot = annot
        self.root = root
        self.root.annotsToTrees[annot] = self
        self.etype = etype
        # This is an INTEGER
        self.level = level

class TreeAnnotation(TreeAnnotationCore):

    def __init__(self, annot, root, level, etype):
        TreeAnnotationCore.__init__(self, annot, root, level, etype)
        # These two are set up before the index loop in buildTree.
        self.tokenStart = None
        self.tokenEnd = None
        # These four are set up in the index loop in buildTree.
        self.parents = None
        self.children = None
        self.previousInLevel = None
        self.nextInLevel = None

    def getString(self):
        return self.annot.doc.signal[self.annot.start:self.annot.end]

    def addChild(self, child):
        if self.children is None:
            self.children = [child]
        else:
            self.children.append(child)
        if child.parents is None:
            child.parents = set([self])
        else:
            child.parents.add(self)

    def getTokens(self):
        if (self.tokenStart is not None) and (self.tokenEnd is not None):
            return self.root.getSequence("token")[self.tokenStart:self.tokenEnd]
        else:
            return None

    # If allowWithinLevelParents is set, then we have to climb up to the level - 1,
    # not just the level.

    def getParentsOfLabelType(self, label):
        level = self.root.builder.levelHash[label]
        parents = self._getParentsOfLevel(level, None)
        return set([p for p in parents if p.annot.atype.lab == label])

    def getParentsOfType(self, lName):
        # First, find the type from the root.
        level = self.root.levelsToIndices[lName]
        # There are multiple parents. On each branch,
        # keep going until you reach the root.
        return self._getParentsOfLevel(level, None)

    def getUniqueParentOfType(self, lType):
        pSet = self.getParentsOfType(lType)
        if (pSet is None) or (len(pSet) != 1):
            raise TreeError("not a unique parent")
        return list(pSet)[0]

    def getUniqueParentOfLabelType(self, label):
        pSet = self.getParentsOfLabelType(label)
        if (pSet is None) or (len(pSet) != 1):
            raise TreeError("not a unique parent")
        return list(pSet)[0]

    def getChildrenOfParentOfType(self, childType, parentType):
        p = self.getUniqueParentOfType(parentType)
        return p.getChildrenOfType(childType)

    def _getParentsOfLevel(self, level, visited):
        # This first clause should never happen, unless
        # we ask the root for parents by accident.
        if self.parents is None:
            return None
        parentsAreCorrectLevel = False
        for p in self.parents:
            if p.etype == ETYPE_ROOT:
                # Don't even go to the root.
                return None
            if p.level < level:
                # We've gone too far.
                return None
            if p.level == level:
                # We don't need to go any farther.
                # All parents are of the same type.
                # We loop because it's the easiest way
                # to get into the set.
                parentsAreCorrectLevel = True
            break
        if parentsAreCorrectLevel:
            # If the parents are the correct type, if we don't allow
            # within level parents, we can return them.
            if not self.root.allowWithinLevelParents:
                return self.parents
            # Otherwise, we start off the set with these parents.
            r = set(self.parents)
        else:
            r = None
        if visited is None:
            visited = set([self])
        else:
            visited.add(self)
        for p in self.parents:
            if p not in visited:
                pTypes = p._getParentsOfLevel(level, visited)
                if pTypes is not None:
                    if r is None:
                        # Copy it.
                        r = set(pTypes)
                    else:
                        r.update(pTypes)
        return r

    def childOf(self, maybeParent, visited = None):
        if not self.parents:
            return False
        else:
            for p in self.parents:
                # The easiest way to get into the set.
                if p.etype < maybeParent.etype:
                    return False
                break
        if maybeParent in self.parents:
            return True
        if visited is None:
            visited = set([self])
        else:
            visited.add(self)
        for p in self.parents:
            if p not in visited:
                if p.childOf(maybeParent):
                    return True
        return False

    def getChildrenOfType(self, pType):
        if self is self.root:
            return self.root.levelSequenceMap[pType]
        else:
            allChildrenOfType = self.root.levelSequenceMap[pType]
            return [a for a in allChildrenOfType if a.annot.start >= self.annot.start and a.annot.end <= self.annot.end]

    def getContext(self, childType, contextType, parentType = None, tokenContextSize = None):
        if (parentType is None) and (tokenContextSize is None):
            raise TreeError("for context, either parent type or token context size required")
        else:
            if parentType is None:
                nodes = self.root.levelSequenceMap[childType]
            else:
                nodes = self.getChildrenOfParentOfType(childType, parentType)
            if contextType == "right":
                if tokenContextSize is None:
                    return [node for node in nodes if node.annot.start >= self.annot.end]
                else:
                    return [node for node in nodes
                            if (node.tokenStart >= self.tokenEnd) and ((node.tokenEnd - self.tokenEnd) <= tokenContextSize)]
            elif contextType == "left":
                if tokenContextSize is None:
                    return [node for node in nodes if node.annot.end <= self.annot.start]
                else:
                    return [node for node in nodes
                            if (node.tokenEnd <= self.tokenStart) and ((self.tokenStart - node.tokenStart) <= tokenContextSize)]
            elif contextType == None:
                if tokenContextSize is None:
                    return nodes
                else:
                    return [node for node in nodes
                            if (node.tokenStart >= (self.tokenStart - tokenContextSize)) and \
                               (node.tokenEnd <= (self.tokenEnd + tokenContextSize))]

    def summarize(self):
        if self.tokenStart and self.tokenEnd:
            return "%s[%d, %d-%d, tok %d-%d]" % \
                   (self.annot.atype.lab, self.etype, self.annot.start, self.annot.end, self.tokenStart, self.tokenEnd)
        else:
            return "%s[%d, %d-%d]" % (self.annot.atype.lab, self.etype, self.annot.start, self.annot.end)

# At this particular point, I'm not exactly sure why the spanless annotations
# ever have a tree annotation associated with them, but I'm almost
# certain to discover the need for it if I leave it out...
        
class SpanlessTreeAnnotation(TreeAnnotationCore):

    def summarize(self):
        return "%s[%d, %d-%d]" % (self.annot.atype.lab, self.etype, self.annot.start, self.annot.end)
                            
class Root(TreeAnnotation):

    def __init__(self, annot, fileName, builder, allowWithinLevelParents = False):
        self.annotsToTrees = {}
        TreeAnnotation.__init__(self, annot, self, -1, ETYPE_ROOT)
        self.builder = builder
        self.allowWithinLevelParents = allowWithinLevelParents
        self.levelSequenceMap = {}
        self.levelsToIndices = {}
        self.fileName = fileName        

    def getSequence(self, cls):
        return self.levelSequenceMap.get(cls)

    def describe(self):
        print("LEVELS")
        for k, values in self.levelSequenceMap.items():
            print(k, " ".join([v.summarize() for v in values]))
            for val in values:
                if val.parents:
                    parents = [v for v in val.parents if v is not self]
                else:
                    parents = None
                if val.children or parents:
                    print(val.summarize())
                    if val.children:
                        print("  CHILDREN", " ".join([v.summarize() for v in val.children]))
                    if parents:
                        print("  PARENTS", " ".join([v.summarize() for v in parents]))
            

# How to define a level:
# (name, type, labels, tiebreakOrder (top to bottom)).
# With no definition, it'll default to tokens at the
# bottom, all non-admin span annotations in the middle, and then
# the document as the root.
# Levels are defined in order. If you have custom elements,
# you have to have levels defined, otherwise it won't work.
# The labels, for annotations, are true labels.

class TreeBuilder(object):

    def __init__(self, task, levels = None, additionalContentLabels = None):
        self.task = task
        self.tokLabels = task.getAnnotationTypesByCategory("token")
        self.otherAdminLabels = set(task.getAnnotationTypesByCategory("zone")) | set(task.getAnnotationTypesByCategory("admin"))
        self.contentLabels = set(task.getAnnotationLabelsForSetNames(task.getNonAdminAnnotationSetNames()))
        if additionalContentLabels is not None:
            self.contentLabels.update(additionalContentLabels)
        self.levels = None
        self.levelHash = {}
        if levels:
            for level in levels:
                self.contentLabels.update(level[2])
            self.setLevels(levels[:])

    def getContentLevel(self):
        return ("content", ETYPE_ANNOT, list(self.contentLabels))

    def setLevels(self, levels = None):
        if self.levels is not None:
            print("Warning: redefining levels", file = sys.stderr)
        self.levels = levels
        # The default levels are "content" > "token".
        if self.levels is None:
            self.levels = [self.getContentLevel()]
        self.levels.append(("token", ETYPE_TOKEN, self.tokLabels))
        self.levelHash = {}
        i = 0
        for l in self.levels:
            for lab in l[2]:
                self.levelHash[lab] = i
            i += 1
        # Now, generate global hierarchy. If there's no hierarchy pair,
        # there's no preference.
        self.hierarchyPairs = set()
        self._populateHierarchyPairs(self.levels)

    # Once the hierarchy pair set is populated, the relative order of
    # any pair is enforced by whether one order or the other is in the pair set.

    def _populateHierarchyPairs(self, levels):
        thisLevel = levels[0]
        if len(thisLevel) == 4:
            for i in range(len(thisLevel[3]) - 1):
                self.hierarchyPairs.update([(thisLevel[3][i], b) for b in thisLevel[3][i + 1:]])
        if len(levels) > 1:
            accumulatedElements = self._populateHierarchyPairs(levels[1:])
            for lab in thisLevel[2]:
                self.hierarchyPairs.update([(lab, e) for e in accumulatedElements])
            return set(thisLevel[2]) | accumulatedElements
        else:
            return set(thisLevel[2])            

    # The default.
    
    def buildTrees(self, docPairList):
        return [self.buildTree(fileName, doc) for (fileName, doc) in docPairList]
    
    def buildTree(self, fileName, doc, allowWithinLevelParents = False):
        if self.levels is None:
            self.setLevels()

        # I want to create an annotation which is the root, but is not
        # stored as an annotation in the document.

        rootAnnot = Annotation(doc, 0, len(doc.signal), "_ROOT_")
        r = Root(rootAnnot, fileName, self, allowWithinLevelParents = allowWithinLevelParents)
                
        # Instantiate the sequence map.
        for i, l in enumerate(self.levels):
            r.levelSequenceMap[l[0]] = []
            r.levelsToIndices[l[0]] = i

        # First, everything becomes TreeNode elements. Start with the tokens.        
        tokLevel = len(self.levels) - 1
        toks = doc.orderAnnotations(self.tokLabels)

        # To deal with token start/end, I think it's best
        # to set it up in advance.
        tokStart = {}
        tokEnd = {}

        foundTokens = len(toks) > 0

        elts = []
        lastTokTree = None
        for i, tok in enumerate(toks):
            tokTree = TreeAnnotation(tok, r, tokLevel, ETYPE_TOKEN)
            elts.append(tokTree)
            tokStart[tok.start] = tokTree.tokenStart = i
            tokEnd[tok.end] = tokTree.tokenEnd = i + 1
            tokTree.previousInLevel = lastTokTree
            if lastTokTree:
                lastTokTree.nextInLevel = tokTree
            lastTokTree = tokTree

        # So the idea with setTokenEnd() and setTokenStart() is that
        # the token boundaries shouldn't have to match exactly. But
        # what if the span we're working on is SMALLER than any token?
        # What do we do? Nope, this is a pathological case which should
        # be addressed in the tokenizer.

        def setTokenEnd(elt):
            if foundTokens:
                charEnd = origCharEnd = elt.annot.end
                while charEnd > elt.annot.start:
                    try:
                        i = tokEnd[charEnd]
                        if charEnd != origCharEnd:
                            tokEnd[origCharEnd] = i
                        elt.tokenEnd = i
                        break
                    except KeyError:
                        charEnd -= 1

        def setTokenStart(elt):
            if foundTokens:
                origCharStart = charStart = elt.annot.start
                while charStart < elt.annot.end:
                    try:
                        i = tokStart[charStart]
                        if charStart != origCharStart:
                            tokStart[origCharStart] = i
                        elt.tokenStart = i
                        break
                    except KeyError:
                        charStart += 1

        setTokenStart(r)
        setTokenEnd(r)

        # And make damn sure that the order stays the same.
        r.levelSequenceMap["token"] = elts[:]
        # And set up token start and end for the root.
        # What if the boundaries aren't on token boundaries?
        # I think we need to ask its token children.

        # Next, the content annotations.
        levelHash = self.levelHash
        levels = self.levels

        # We want to be absolutely sure we can deal with spanless annotations.
        
        for e in doc.getAnnotations(self.contentLabels, spannedOnly = True):
            l = levelHash.get(e.atype.lab)
            if l is not None:
                t = TreeAnnotation(e, r, l, ETYPE_ANNOT)
                setTokenStart(t)
                setTokenEnd(t)
                elts.append(t)

        for e in doc.getAnnotations(self.contentLabels, spanlessOnly = True):
            l = levelHash.get(e.atype.lab)
            if l is not None:
                t = SpanlessTreeAnnotation(e, r, l, ETYPE_ANNOT)
                level = levels[t.level]
                r.levelSequenceMap[level[0]].append(t)
        
        # So there are a couple philosophies about the ordering strategy.
        # If you allow within-level ordering, you have to have some
        # sort of tiebreaker. But then you're not guaranteed of looping through
        # all the elements if you provide the sequence - UNLESS the sequence
        # isn't the concatenation of the children lists.
        
        # Originally, I was going to sort the annotations smallest start,
        # greatest end and then create the tree from that list, but the list
        # would require a good deal of complicated end-node computation.
        # I think, instead, I'm going to do my standard "collect all the indices"
        # algorithm and create the tree from THAT. There, the tree order
        # ends up enforcing the order, mostly. I'll still have cases
        # in the case of within-level overlap where I have to reorder
        # the child lists. Maybe.

        indexHash = {}
        for e in elts:
            try:
                indexHash[e.annot.start][0].append(e)
            except KeyError:
                indexHash[e.annot.start] = [[e], []]
            try:
                indexHash[e.annot.end][1].append(e)
            except KeyError:
                indexHash[e.annot.end] = [[], [e]]
        allIndices = list(indexHash.keys())
        allIndices.sort()

        levelOpenCache = [set() for i in range(len(self.levels))]

        # This is set up to work with Python 3 comparison,
        # which is optimized by keys. The idea is that an
        # annotation is greater than another if its end is
        # greater than another, or, if the ends are equal, if
        # the level is less.
        def elementKey(e):
            return (e.annot.end, -e.level)

        for idx in allIndices:
            # OK, what do we do?
            [starting, ending] = indexHash[idx]
            
            # First, end everything that's ending.
            for e in ending:
                levelOpenCache[e.level].remove(e)
            
            # Sort the starting list by greatest end. But it's not just
            # the greatest end; if the ends are the same, we want to do the
            # dominating levels first. So we need to reverse the list.
            starting.sort(key = elementKey, reverse = True)

            # What we're setting up here is previous, next, children, parents.
            for s in starting:
                level = levels[s.level]
                if s.etype != ETYPE_TOKEN:
                    # Tokens have already been added.
                    if r.levelSequenceMap[level[0]]:
                        r.levelSequenceMap[level[0]][-1].nextInLevel = s
                        s.previousInLevel = r.levelSequenceMap[level[0]][-1]
                    r.levelSequenceMap[level[0]].append(s)                    
                # Find the first entry in the open cache
                # which has an entry whose end is <= to the end of this one.
                if allowWithinLevelParents:
                    startLevel = s.level
                else:
                    startLevel = s.level - 1

                # The higher the level number, the lower in
                # the tree the element is. Tokens have the highest
                # levelnumber.

                # If startLevel reaches -1 (or if it starts there,
                # for that matter), attach to the root.
                
                while startLevel > -1:
                    cacheEntry = levelOpenCache[startLevel]
                    found = False
                    for c in cacheEntry:
                        if c.annot.end >= s.annot.end:
                            # We're good.
                            found = True
                            c.addChild(s)
                    if found:
                        break                    
                    startLevel -= 1
                if startLevel == -1:
                    r.addChild(s)
                levelOpenCache[s.level].add(s)

        return r
