# Copyright (C) 2007 - 2018 The MITRE Corporation. See the toplevel
# file LICENSE.txt for license terms.

from __future__ import print_function, absolute_import

# This file contains all the basic pieces to do bootstrapping:
# build a corpus, build a model, run a test. The experiment engine
# is an extension of this now.

import sys, os, shutil, random, glob, time, codecs
# I don't think this is necessary, but it doesn't hurt.
random.seed()

# We actually need to create a better abstraction of document
# sets - we may end up using them in multiple places, but the first
# use is going to be for the experiment engine and cross-validation.

# These document sets should support a bunch of things:
# - randomly split into multiple sets
# - consolidate to a single directory, with disambiguated file names
# - instantiate from a number of different types of sources, such as
# a list of files, a file which contains a list of files, or pattern + prefix

class DocumentSetError(Exception):
    pass

from .Operation import DictOpArgumentAggregator

from . import PluginMgr
from . import ExecutionContext
from . import DocumentIO
from . import ToolChain

#
# The document set object.
#

# filePats: a list of glob expressions which picks out
# a candidate set of training docs.
# fileFile: a file which contains filenames, one per line.
# use either this or filePats/randomSubset.

# Two ways of providing a list of files: either filePats and a
# size, or a file of filenames, or a list of the files directly.

class DocumentSet(object):
    
    def __init__(self, corpusName, partitionFile = None, filePats = None,
                 fileFile = None, fileList = None,
                 prefix = None,
                 partitions = None, partitionIsFixed = False, crossvalidationFolds = None,
                 flushedStdout = sys.stdout):

        self.corpusName = corpusName
        # Can't use sets, because I really need the order to be maintained.
        self._initVars()
        if partitionFile is not None:
            self.loadPartition(partitionFile)
        else:
            if partitions and (crossvalidationFolds is not None):
                raise DocumentSetError("can't specify both crossvalidation and partitions for corpus '%s'" % corpusName)
            self._populate(filePats, fileFile, fileList, prefix, flushedStdout)
            # I need to know this when I try to figure out what to do with the
            # corpus. We're going to name the folds something distinctive, like "fold_<n>",
            # and if any of the partitions have that pattern and crossvalidationFolds is None,
            # we'll raise an error.
            self.crossvalidationFolds = crossvalidationFolds
            self._partition(partitions or None, partitionIsFixed, crossvalidationFolds)

    def _initVars(self):
        
        self.files = []
        self._filesSeen = {}
        self.partitionDict = None
        self.deconflictions = None

    def _populate(self, filePats, fileFile, fileList, prefix, flushedStdout):

        # In the general case, who cares how many of these you use.

        if (fileList is None) and (filePats is None) and (fileFile is None):
            raise DocumentSetError("neither fileFile nor filePats nor fileList is specified")
        if prefix is not None:
            if not os.path.isabs(prefix):
                raise DocumentSetError("prefix must be absolute")
        if fileList:
            for f in fileList:
                if os.path.isabs(f):
                    self._addFile(f)
                elif prefix is None:
                    raise DocumentSetError("fileList contains relative pathnames, but there's no prefix")
                else:
                    self._addFile(os.path.join(prefix, f))
        elif fileFile:
            fp = codecs.open(fileFile, "r", "utf-8")
            for f in [line.strip() for line in fp.readlines()]:
                if os.path.isabs(f):
                    self._addFile(f)
                else:
                    # The justification for this is that a file which contains
                    # pathnames will have been archived by the save() method.
                    raise DocumentSetError("fileFile must contain absolute pathnames")
            fp.close()
        else:
            files = []
            if type(filePats) is type(""):
                filePats = [filePats]
            # Remove duplicates
            for p in filePats:
                i = 0
                if not os.path.isabs(p):
                    if prefix is None:
                        raise DocumentSetError("filePats contains relative pathnames, but there's no prefix")
                    p = os.path.join(prefix, p)
                for f in glob.glob(p):
                    if not os.path.isfile(f):
                        raise DocumentSetError("file %s specified by pattern %s is not a regular file" % (p, f))
                    self._addFile(f)
                    i += 1
                if i == 0:
                    print("Warning: pattern %s did not pick out any files" % p, file=flushedStdout)

        if not self.files:
            raise DocumentSetError("document set is empty")

    def _addFile(self, f):
        if f not in self._filesSeen:
            self.files.append(f)
            self._filesSeen[f] = True

    # What does a partition look like? It's a list of pairs, where
    # the first element is the name of the partition, and the second is
    # the share. The share is a number. All the numbers are normalized to
    # a share of 1, and then we randomize the list of files, and
    # select slice of the randomized list.

    # If partitionIsFixed is true, the number is an actual number of
    # files, rather than a share. -1 means "everything else".

    FIXED_PARTITION_REMAINDER = -1
    
    def _partition(self, partitions, partitionIsFixed, crossvalidationFolds):
        # crossvalidationFolds is implemented as multiple partitions
        # in the corpus. The model build and run have to check to see if
        # a model has crossvalidation folds. We'll use the name of the
        # partitions to indicate crossvalidation when we load our partition.
        if partitions is not None:
            for pName, total in partitions:
                if pName.startswith("fold_"):
                    raise DocumentSetError("non-crossvalidation partition name '%s' cannot start with 'fold_'" % pName)
        if crossvalidationFolds is not None:
            if crossvalidationFolds < 2:
                raise DocumentSetError("crossvalidation folds must be > 1")
            # If crossvalidation is specified, the partition information is ignored.
            partitionIsFixed = False
            partitions = [("fold_%d" % i, 1) for i in range(crossvalidationFolds)]
        # Save the partition list in NORMALIZED form.
        if partitions is not None:
            self.partitionDict = {}
            files = self.files[:]
            random.shuffle(files)
            trueP = []
            if partitionIsFixed:
                remainder = None
                start = 0
                for pName, total in partitions:
                    if total == self.FIXED_PARTITION_REMAINDER:
                        if remainder is not None:
                            raise DocumentSetError("multiple remainder pairs specified: %s, %s" % (pName, remainder))
                        remainder = pName
                    elif start + total > len(files):
                        raise DocumentSetError("number of requested files for fixed partition '%s' exceed number of available files" % pName)
                    else:
                        trueP.append([pName, start, start + total])
                        start += total
                # Not all the files might be used.
                if remainder is not None:
                    trueP.append([pName, start, len(files)])
            else:
                total = float(sum([x[1] for x in partitions]))
                # Now, we convert the numbers in the partitions
                # into shares of the size of the number of files.
                # The last one has to grab the remainder. I don't
                # want to make this too complicated.
                i = 0.0
                for pName, share in partitions:
                    j = i + (len(files) * (share / total))
                    trueP.append([pName, int(round(i)), int(round(j))])
                    i = j
                trueP[-1][-1] = len(files)
            for pName, start, end in trueP:
                self.partitionDict[pName] = files[start:end]
        else:
            # Make sure to randomize the order.
            self.files = self.files[:]
            random.shuffle(self.files)

    def getFiles(self, partition = None):
        if partition is not None:
            if self.partitionDict is None:
                raise KeyError(partition)
            else:
                return self.partitionDict[partition]
        else:
            return self.files

    def savePaths(self, outPath):
        fp = codecs.open(outPath, "w", "utf-8")
        for f in self.files:
            fp.write(f + "\n")
        fp.close()

    def savePartition(self, outPath):
        fp = codecs.open(outPath, "w", "utf-8")
        if self.partitionDict is not None:
            for name, files in self.partitionDict.items():
                for f in files:
                    fp.write("%s,%s\n" % (name, f))
        fp.close()

    def isCrossvalidationCorpus(self):
        return self.crossvalidationFolds is not None

    def loadPartition(self, path):
        self.partitionDict = {}
        fp = codecs.open(path, "r", "utf-8")
        for line in fp.readlines():
            [name, f] = line.strip().split(",", 1)
            try:
                self.partitionDict[name].append(f)
            except KeyError:
                self.partitionDict[name] = [f]
            self._addFile(f)
        keys = self.partitionDict.keys()
        if keys:
            howManyFolds = len([k for k in keys if k.startswith("fold_")])
            if howManyFolds > 0:
                # You can't have a partition that starts with this name without
                # its being a crossvalidation corpus.
                if howManyFolds < len(keys):
                    raise DocumentSetError("not all partitions in crossvalidation corpus start with fold_")
                self.crossvalidationFolds = howManyFolds
        fp.close()

    def _deconflict(self):
        if self.deconflictions is not None:
            return
        self.deconflictions = {}
        duplicateBasenameDict = {}
        basenamesToMap = []
        for f in self.files:
            # This will be overwritten in some cases later when we figure
            # out what needs to be deconflicted.
            b = os.path.basename(f)
            self.deconflictions[f] = b
            if b in duplicateBasenameDict:
                if len(duplicateBasenameDict[b]) == 1:
                    basenamesToMap.append(b)
                duplicateBasenameDict[b].append(f)
            else:
                duplicateBasenameDict[b] = [f]
        if basenamesToMap:
            for b in basenamesToMap:
                # There will be no duplicates in the list, because
                # document sets have duplicates removed.
                dirPrefix = os.path.dirname(os.path.commonprefix(duplicateBasenameDict[b]))
                for f in duplicateBasenameDict[b]:
                    suff = f[len(dirPrefix):]
                    # The new filename is this suffix with all the dir
                    # components replaced.
                    stack = []
                    split = os.path.split(suff)
                    while split[0] != suff:
                        stack[0:0] = [split[1]]
                        suff = split[0]
                        split = os.path.split(suff)
                    newBasename = "_".join(stack)
                    # And replace it in the list of files in the document set.
                    self.deconflictions[f] = newBasename

    # Consolidate a set of documents into a directory.
    # Use deconfliction.
    
    def consolidate(self, dir, fn):
        self._deconflict()
        for f in self.files:
            fn(f, os.path.join(dir, self.deconflictions[f]))

    # In the default case, these two are no-ops. They're required for the
    # interaction with a potential parent bootstrapping or experiment engine.
    
    def setContext(self, engine):
        pass

    def prepare(self, *args, **kw):
        pass

#
# And now, something to wrap around the ModelBuilder to be used in
# bootstrapping systems. This is simplified from the original version in
# CarafeTrain.py, which does all sorts of frightening cacheing of settings, etc.
#

class TrainingRunError(Exception):
    pass

class TrainingRunInstance(object):

    def __init__(self, engine, template, modelSubdir, trainingCorpus,
                 engineSettings = None, corpusSettings = None):
        self.modelName = template.modelName
        self.modelSubdir = modelSubdir
        self.configName = template.configName
        self.trainableStep = template.trainableStep
        
        # I think these will be actual training corpora.
        self.trainingSet = trainingCorpus
        self.engineSettings = engineSettings
        self.corpusSettings = corpusSettings
        self.flushedStdout = template.flushedStdout
        self.template = template

        self.setContext(engine)        
        
    # In this case, there really does need to be a context set,
    # and the engine needs to have a task, etc.

    # This method is only called from the __init__. So it doesn't need to be
    # implemented for the crossvalidation run.

    def setContext(self, engine):

        self._updateFromEngine(engine)

        if not os.path.exists(self.modelDir):
            print("Creating model directory", self.modelDir, "...", file=self.flushedStdout)
            os.makedirs(self.modelDir)

        self._configureBuilder()
        
    # The trainer requires all the files in a single directory, so we
    # take the files and put them somewhere random.
    # The format of the model output dir is a file containing
    # the file list at the toplevel, and a subdir for each increment.
    # I'm going to use the modelOutputDir also as the temp location,
    # because that's the safest thing to do - that location will have
    # to be appropriately protected for sensitive data.
    
    def train(self, interleave = False, collectCorpusStatistics = False):

        with ExecutionContext.Tmpdir(specifiedParent = self.modelDir) as tmpDir:
            mName = self.modelName

            print("Building model ...", file=self.flushedStdout)
            t = time.time()
            self.builder.run(os.path.join(self.modelDir, "model"),
                             self.trainingSet.getFiles(), 
                             tmpDir = tmpDir,
                             collectCorpusStatistics = collectCorpusStatistics,
                             oStream = self.flushedStdout)
            print("Build completed in %.2f seconds." % (time.time() - t,), file=self.flushedStdout)
            # Interleave will only be True if there's a single modelName.
            if interleave:
                for rTemplate in self.engine.runList:
                    if not rTemplate.modelNames:
                        continue
                    if rTemplate.modelNames[0] == mName:
                        for r in rTemplate.yieldInstances(self.engine):
                            r.setContext(self.engine, [self])
                            r.prepareTest()
                            rName = r.runName
                            print("\n### Performing run", rName, "in directory", r.runSubdir, file=self.flushedStdout)
                            r.test()
                            print("### Done.", file=self.flushedStdout)
                            r.finishTest()

    # The builder returns {origPath -> (totalItemsByTag, totalToks, totalToksByTag)}
    # This needs to yield something different:
    #
    # {"totalDocuments": 0,
    #                "totalItems": 0,
    #                "totalTokens": 0,
    #                "totalItemTokens": 0,
    #             "totalItemsByTag": {},
    #                "totalItemTokensByTag": {}}
    
    def reportCorpusStatistics(self):
        if not hasattr(self, "localCorpusStatistics"):
            rawStats = self.builder.corpusStatistics
            seed = {"totalDocuments": 0,
                    "totalItems": 0,
                    "totalTokens": 0,
                    "totalItemTokens": 0,
                    "totalItemsByTag": {},
                    "totalItemTokensByTag": {}}

            for origPath, stats in rawStats.items():
                self._augmentCorpusStatistics(seed, stats)
            self.localCorpusStatistics = seed
        return self.localCorpusStatistics

    @staticmethod
    def _augmentCorpusStatistics(seed, rawStats):
        totalItemsByTag, totalToks, totalToksByTag = rawStats
        seed["totalDocuments"] += 1
        seed["totalTokens"] += totalToks
        for lab, i in totalItemsByTag.items():
            try:
                seed["totalItemsByTag"][lab] += i
            except KeyError:
                seed["totalItemsByTag"][lab] = i
            seed["totalItems"] += i
        for lab, i in totalToksByTag.items():
            try:
                seed["totalItemTokensByTag"][lab] += i
            except KeyError:
                seed["totalItemTokensByTag"][lab] = i
            seed["totalItemTokens"] += i
               
    #
    # Internal methods which can be consumed.
    #
    
    def _updateFromEngine(self, engine):

        self.engine = engine
        self.modelDir = engine.getModelDir(self)
    
    def _configureBuilder(self):
        task = self.engine.task
        # Use a local class. If there are no annotation sets in this
        # case, the model info might barf.
        buildInfo = _computeBuildInfo(task, self.configName, self.trainableStep)
        if self.trainableStep is None:
            self.trainableStep = buildInfo.step.stepName
        mBClass = buildInfo.getModelClass()
        if mBClass is None:
            raise TrainingRunError("model %s has no engine for the task" % self.modelName)
        builderSettings = self.engineSettings or {}
        self.builder = buildInfo.buildModelBuilder(**builderSettings)

# If you do it this way, it will either break because there are multiple
# trainable steps, or it will return a model info object with the step.

def _computeBuildInfo(task, configName, stepName):
    try:
        return task.getModelInfo(configName = configName, step = stepName)
    except PluginMgr.PluginError as e:
        raise TrainingRunError("can't compute model build info: %s" % str(e))

# The challenge here is keeping enough information about the
# corpora to be able to ensure that (a) corpus increments are guaranteed to
# be successively larger slices of the SAME ORDER, and (b) corpus balances
# can be preserved. And then there's the issue about restart - but that's
# only in CarafeTrain, and I think I have to address it there.
# So I think I'll collect them as pairs of dsetname/partition and file,
# THEN shuffle them. Then that will be what I'll use when I select from it.

# So we relly need to have a number of ways of figuring out what the
# model info should be. If there's a trainableEngine, we can try to
# find it (and, in that case, we need to specify annotation sets).
# Otherwise, if there's a trainableStep, we can ask it. Otherwise,
# we run the risk of there being multiple trainable steps.

# For crossvalidation, the TrainingRun can't yield all the individual
# fold-excluding models, because when the runs get hold of the models,
# it generates the cross-product of all the instances yielded by the
# template sequence, and that's just the wrong thing. I need the TrainingRun
# to yield a CrossvalidationTrainingRunInstance.

# And it has to look just like a TrainingRunInstance.

class CrossvalidationTrainingRunInstance(object):

    def __init__(self, foldInstances, modelSubdir, trainingSet):
        self.foldInstances = foldInstances

        template = self.foldInstances[0].template

        self.modelName = template.modelName
        # I don't think this is ever used.
        self.modelSubdir = modelSubdir
        self.configName = template.configName
        self.trainableStep = self.foldInstances[0].trainableStep
        
        # Just in case.
        self.trainingSet = trainingSet
        self.engineSettings = self.foldInstances[0].engineSettings
        self.corpusSettings = self.foldInstances[0].corpusSettings
        self.flushedStdout = template.flushedStdout
        self.template = template

        self.engine = self.foldInstances[0].engine
        self.modelDir = self.engine.getModelDir(self)

    def train(self, interleave = False, collectCorpusStatistics = False, **kw):
        for m in self.foldInstances:
            # Don't interleave the individual folds.
            m.train(interleave = False, collectCorpusStatistics = collectCorpusStatistics, **kw)
        # Interleave will only be True if there's a single modelName.
        if interleave:
            for rTemplate in self.engine.runList:
                if not rTemplate.modelNames:
                    continue
                if rTemplate.modelNames[0] == self.modelName:
                    for r in rTemplate.yieldInstances(self.engine):
                        r.setContext(self.engine, [self])
                        r.prepareTest()
                        rName = r.runName
                        print("\n### Performing run", rName, "in directory", r.runSubdir, file=self.flushedStdout)
                        r.test()
                        print("### Done.", file=self.flushedStdout)
                        r.finishTest()

    # The builder returns {origPath -> (totalItemsByTag, totalToks, totalToksByTag)}
    # This needs to yield something different:
    #
    # {"totalDocuments": 0,
    #                "totalItems": 0,
    #                "totalTokens": 0,
    #                "totalItemTokens": 0,
    #             "totalItemsByTag": {},
    #                "totalItemTokensByTag": {}}

    # And the aggregation has to be careful to consider each file
    # only once.
    
    def reportCorpusStatistics(self):
        if not hasattr(self, "localCorpusStatistics"):
        
            seed = {"totalDocuments": 0,
                    "totalItems": 0,
                    "totalTokens": 0,
                    "totalItemTokens": 0,
                    "totalItemsByTag": {},
                    "totalItemTokensByTag": {}}
            seen = set()
            for m in self.foldInstances:
                rawStats = m.builder.corpusStatistics
                for origPath, stats in rawStats.items():
                    if origPath not in seen:
                        m._augmentCorpusStatistics(seed, stats)
                        seen.add(origPath)
            self.localCorpusStatistics = seed
        return self.localCorpusStatistics

class TrainingRun(object):

    def __init__(self, mName, trainingCorpora = None,
                 engineSettings = None, corpusSettings = None,
                 configName = None,
                 trainableStep = None,
                 iterators = None,
                 flushedStdout = sys.stdout,
                 instanceClass = TrainingRunInstance,
                 crossvalidationInstanceClass = CrossvalidationTrainingRunInstance):

        self.modelName = mName        
        self.iterators = iterators
        if iterators:
            for iterator in iterators[:-1]:
                if iterator.mustBeInnermost:
                    raise TrainingRunError("Iterator %s must be innermost, but is not last" % iterator)
        self.configName = configName
        self.trainableStep = trainableStep        
        # If one of the training corpora is a crossvalidation corpus,
        # there can be only one of them. However, we can't check that until
        # we actually have an engine.
        self.trainingCorpora = trainingCorpora        
        self.engineSettings = engineSettings
        self.corpusSettings = corpusSettings
        self.flushedStdout = flushedStdout
        self.instanceClass = instanceClass
        self.crossvalidationInstanceClass = crossvalidationInstanceClass
        # This is updated in _collectTrainingSet, which is a little perverse.
        # But that's where we look at the corpora.
        self.doCrossvalidation = False
        # We collect the instances. If the run
        # asks for them and you can use the cache,
        # then use the cache.
        self._cached = False
        self.allInstances = []

    # Note that because _createTrainingCorpus is called repeatedly,
    # if there's iteration, on the output of _collectTrainingSet, there's
    # no need to guarantee an order in _collectTrainingSet.

    # Here, we have to check to see if any of the corpora are crossvalidation
    # corpora. If they are, there can only be one training corpus. Later,
    # I'll follow up in createInstance.

    def _collectTrainingSet(self, engine):
        elts = []        
        for dSetName, partition in self.trainingCorpora:
            c = engine.getCorpus(dSetName)
            if c.isCrossvalidationCorpus():
                if len(self.trainingCorpora) > 1:
                    raise TrainingRunError("training run '%s' has more than one training corpus, but one of them is a crossvalidation corpus" % self.modelName)
                if partition is not None:
                    raise TrainingRunError("training run '%s' may not specify a partition of a crossvalidation corpus" % self.modelName)
                self.doCrossvalidation = True
                for p in c.partitionDict.keys():
                    elts += [((dSetName, p), f) for f in c.getFiles(p)]
            else:
                elts += [((dSetName, partition), f) for f in c.getFiles(partition)]
        random.shuffle(elts)
        return elts

    def _createTrainingCorpus(self, tSetElements, size = None):
        # VERY IMPORTANT to trim BEFORE DocumentSet is created,
        # since DocumentSet does a random shuffle.
        if size is not None:
            if self.doCrossvalidation:
                raise TrainingRunError("training run '%s' specifies the corpus size directly, but the training corpus is a crossvalidation corpus" % self.modelName)
            tSetElements = tSetElements[:size]
        tFiles = [p[1] for p in tSetElements]
        return DocumentSet(None, fileList = tFiles, flushedStdout = self.flushedStdout)

    def _satisfyIterators(self, engine, iterators, subdirName, tSetElements = None, **kw):
        if not iterators:
            # We've reached the end.
            if self.doCrossvalidation:
                # So here, we need to divide the tSetElements into appropriate pots, and then
                # generate an instance for each. The fold names will be fold_<n>.
                # And remember, there's only one corpus.
                mInstances = []
                for partition in set([t[0] for t in tSetElements]):
                    fold = partition[1]
                    mInstance = self._createInstance(engine, subdirName + "_except_for_" + fold,
                                                     tSetElements = [(pPair, f) for (pPair, f) in tSetElements if pPair != partition],
                                                    **kw)
                    mInstances.append(mInstance)
                yield self.crossvalidationInstanceClass(mInstances, subdirName,
                                                        self._createTrainingCorpus(tSetElements, **kw.get("corpusSettings", {})))
            else:
                yield self._createInstance(engine, subdirName, tSetElements = tSetElements, **kw)
        else:
            for newSubdirName, newKw in iterators[0](subdirName, tSetElements = tSetElements, **kw):
                for inst in self._satisfyIterators(engine, iterators[1:], newSubdirName, **newKw):
                    yield inst

    # both tSetElements and engineSettings are keywords. The former is a list
    # of pairs ((dSetName, partition), f), and the second is guaranteed to have a dictionary value.
    
    def _createInstance(self, engine, modelSubdir, tSetElements = None, engineSettings = None,
                        corpusSettings = None, 
                        # Soak up the extra context that was passed in to the iterators
                        **kw):
        return self.instanceClass(engine, self, modelSubdir,
                                  self._createTrainingCorpus(tSetElements, **corpusSettings),
                                  engineSettings = engineSettings)
    #
    # The sole public method
    #

    # We collect the objects and record the cache as done, so if we ever
    # loop over the runs instead of the models, it'll work fine.
    
    def yieldInstances(self, engine, **kw):
        if self._cached:
            for i in self.allInstances:
                yield i
        else:
            tSetElements = self._collectTrainingSet(engine)
            for mInst in self._satisfyIterators(engine, self.iterators or [], self.modelName,
                                                tSetElements = tSetElements,
                                                engineSettings = self.engineSettings or {},
                                                corpusSettings = self.corpusSettings or {},
                                                **kw):
                self.allInstances.append(mInst)
                yield mInst
            self._cached = True

#
# Next, the TestRun object.
#

_jsonIO = DocumentIO.getDocumentIO('mat-json')
_rawUTF8 = DocumentIO.getDocumentIO('raw', encoding = 'utf-8')

class TestRunError(Exception):
    pass

class TestRunInstance(object):

    def __init__(self, engine, template, runSubdir, engineOptions = None):
        self.runName = template.runName
        self.runSubdir = runSubdir
        self.engineOptions = engineOptions
        self.enginePrepOptions = template.enginePrepOptions
        self.flushedStdout = template.flushedStdout
        self.template = template
        self.testSet = template.testSet
        self.doCrossvalidation = False
        
    # I was hoping that this could be called from __init__, but
    # it has to wait until the model is created, which is sometimes inside
    # the run creation loop.

    # mInstances might be the empty list.

    # It's here that we should deal with checking the testSet. If this
    # is a crossvalidation test, the testSet must be None; otherwise, the
    # test can't be a crossvalidation test. And it's an error if there's
    # no test set and it's not crossvalidation. We can't deal with
    # crossvalidation until here because we need the model instances
    # to do it. setContext() is always called before prepareTest().
    
    def setContext(self, engine, mInstances):
        self.models = mInstances
        self.engine = engine
        # if models is empty...
        if not self.models:
            self.runDir = os.path.join(self.engine.getRunDir(self), "nomodel")
        else:
            self.runDir = os.path.join(self.engine.getRunDir(self), "_".join([m.modelSubdir for m in self.models]))
        if self.testSet is None:
            # If the test set is None, all the mInstances had better be crossvalidation training run instances.
            if not mInstances:
                raise TestRunError("run '%s' has no models and no test sets" % self.runName)
            for mInstance in mInstances:
                if not isinstance(mInstance, CrossvalidationTrainingRunInstance):
                    raise TestRunError("run '%s' has no test sets, but not all its models are crossvalidation models" % self.runName)
            # OK, we know that all the models have the same training corpus, and that
            # the models are all crossvalidation models. So we can create a test set
            # from the training corpus for the first model. This will be used in _prepareTestCorpus().
            self.testSet = DocumentSet(None, fileList = mInstances[0].trainingSet.getFiles(), flushedStdout = self.flushedStdout)
            self.doCrossvalidation = True
        elif mInstances:
            # If test sets are specified, none of the models can be crossvalidation models.
            for mInstance in mInstances:
                if isinstance(mInstance, CrossvalidationTrainingRunInstance):
                    raise TestRunError("run '%s' has test sets, but at least one of its models is a crossvalidation model" % self.runName)
    
    def prepareTest(self, **kw):
        
        print("\n### Preparing run", self.runName, file=self.flushedStdout)

        runInputTestDir = os.path.join(self.runDir, "run_input")

        if os.path.isdir(runInputTestDir):
            shutil.rmtree(runInputTestDir)

        self._prepareTestCorpus()

    def test(self):

        hypDir = os.path.join(self.runDir, "hyp")
        os.makedirs(hypDir)
        msg = "Creating hypothesis..."
        
        matEngineOptions = self.engineOptions or {}
        runInputTestDir = os.path.join(self.runDir, "run_input")
        task = self.engine.task

        # It used to be that there was a single tag step,
        # so a single tagger_local and tagger_model key.
        # No longer.

        forbiddenKeys = set(["input_file_type", "output_encoding",
                             "output_file_type", "input_encoding", "input_dir", "input_file",
                             "output_dir", "output_file", "output_fsuff", "input_file_re"])
        
        for key in matEngineOptions.keys():
            if (key in forbiddenKeys) or \
               (self.models and
                ((key.endswith("_model")) or \
                 (key.endswith("_local")))):
                # Either it's a literal forbidden key, or it's a key
                # corresponding to the parent keys in CmdlineTagStep.
                raise TestRunError("%s not permitted in run settings" % key)
        
        # Now, pull over the input file information from the prep options,
        # if applicable.
        
        inputEncoding = "utf-8"
        inputFileType = "raw"
        if self.enginePrepOptions is not None:
            if "output_file_type" in self.enginePrepOptions:
                inputFileType = self.enginePrepOptions["output_file_type"]
            if "output_encoding" in self.enginePrepOptions:
                inputEncoding = self.enginePrepOptions["output_encoding"]

        print(msg, file=self.flushedStdout)

        print("Invoking MATEngine:", " ".join(['%s: "%s"' % pair for pair in matEngineOptions.items()]), file=self.flushedStdout)
        # The workflow had better be there. This will raise an error if it
        # isn't, most likely.
        
        # We're going to find the step in the models, and then we're going to
        # add the appropriate model options. Note that we have to preprocess
        # the step arguments here, because the iterators may have contributed
        # something, and because any parameters which have been passed in in
        # the dictionaries haven't been processed yet.

        import re

        if self.doCrossvalidation:
            
            # And if this is a crossvalidation step, we need to do a run for each
            # fold in the training corpus. But the target is always the same output dir.

            trueTrainingCorpus = self.engine.getCorpus(self.models[0].template.trainingCorpora[0][0])

            # The deconflictions in the test set (because the test set was deconflicted in the test set
            # preparation) map from full files to new basenames. The items in the partitionDict
            # are full files.
            for fold, files in trueTrainingCorpus.partitionDict.items():
                localBasenames = [self.testSet.deconflictions[f] for f in files]
                # Create a temporary directory, pull the files from the runInputTestDir
                # into that directory, and then do a run.                
                with ExecutionContext.Tmpdir(specifiedParent = self.runDir) as tmpDir:
                    for f in files:
                        basename = self.testSet.deconflictions[f]
                        shutil.copy(os.path.join(runInputTestDir, basename + ".prepped"), os.path.join(tmpDir, basename + ".prepped"))

                    self._doRun(matEngineOptions, task, inputFileType, inputEncoding,
                                tmpDir, hypDir, "_except_for_" + fold)
        else:

            self._doRun(matEngineOptions, task, inputFileType, inputEncoding,
                        runInputTestDir, hypDir)

    def _doRun(self, matEngineOptions, task, inputFileType, inputEncoding,
               inputDir, hypDir, modelFileSuffix = None):

        params = {}

        for m in self.models:
            modelDir = os.path.join(m.modelDir)
            if modelFileSuffix:
                modelDir += modelFileSuffix
            modelFile = os.path.join(modelDir, "model")

            stepName = re.sub("\W+", "_", m.trainableStep)
            params.update({stepName + "_local": True, stepName + "_model": modelFile})

        params.update(matEngineOptions)

        e = ToolChain.MATEngine(taskObj = task, workflow = matEngineOptions.get("workflow"))

        e.Run(input_file_type = inputFileType, input_encoding = inputEncoding,
              output_file_type = "mat-json", input_dir = inputDir,
              output_dir = hypDir, output_fsuff = ".tag.json", language = self.engine.language,
              **e.aggregatorExtract(DictOpArgumentAggregator(params)))

    def finishTest(self):
        pass

    #
    # Internal functions available for reuse.
    #

    def _prepareTestCorpus(self):

        expEngine = self.engine
        task = expEngine.task
        testSet = self.testSet

        # Create preprocessed versions of the test files.
        # If there's no enginePrepOptions, just make them raw.
        
        print("Preparing test files...", file=self.flushedStdout)
        runInputTestDir = os.path.join(self.runDir, "run_input")
        os.makedirs(runInputTestDir)

        if self.enginePrepOptions is None:
            # Let's use UTF-8, since we know that it works.
            testSet.consolidate(runInputTestDir,
                                lambda inf, outf: _rawUTF8.writeToTarget(_jsonIO.readFromSource(inf, taskSeed = task), outf + ".prepped"))
        else:
            testSet.consolidate(runInputTestDir,
                                lambda inf, outf: shutil.copy(inf, outf + ".prepinput"))
            # Invoke the MATEngine. output_file_type is required. That's pretty much it.            
            for key in ["tagger_local", "tagger_model", "input_file_type",
                        "input_encoding", "input_dir", "input_file",
                        "output_dir", "output_file", "output_fsuff", "input_file_re"]:
                if key in self.enginePrepOptions:
                    raise TestRunError("%s not permitted in run prep settings" % key)
            if "output_file_type" not in self.enginePrepOptions:
                raise TestRunError("output_file_type attribute required in run prep settings")
            if self.enginePrepOptions["output_file_type"] not in ["raw", "mat-json"]:
                raise TestRunError("output_file_type attribute in run prep settings must be either raw or mat-json")
            
            print("Invoking MATEngine:", " ".join(['%s: "%s"' % pair for pair in self.enginePrepOptions.items()]), file=self.flushedStdout)
            # The workflow had better be there. This will raise an error if it
            # isn't, most likely.
            e = ToolChain.MATEngine(taskObj = task, workflow = self.enginePrepOptions.get("workflow"))
            # Use the directory temporarily.
            e.Run(input_file_type = "mat-json",
                  input_dir = runInputTestDir,
                  output_dir = runInputTestDir,
                  input_file_re = "^.*\.prepinput$",
                  output_fsuff = ".postprocessed",
                  language = self.engine.language,
                  **e.aggregatorExtract(DictOpArgumentAggregator(self.enginePrepOptions)))
            # Done. Remove all files which end with .prepinput, strip
            # .postprocessed.
            for b in testSet.deconflictions.values():
                os.remove(os.path.join(runInputTestDir, b + ".prepinput"))
                os.rename(os.path.join(runInputTestDir, b + ".prepinput.postprocessed"),
                          os.path.join(runInputTestDir, b + ".prepped"))

# Iterators can't apply to the test set. So that should be prepared
# when the test run is asked to yield instances.

# Let's see whether we can do multiple models. So model
# here can be a list, or a comma-delimited string. And
# the result will be a list of modelNameS.

class TestRun(object):

    def __init__(self, rName, model = None, testCorpora = None,
                 engineOptions = None,
                 iterators = None,
                 enginePrepOptions = None, flushedStdout = sys.stdout,
                 instanceClass = TestRunInstance):
        self.runName = rName
        if type(model) in (set, tuple, list):
            self.modelNames = model
        elif model is not None:
            self.modelNames = [s.strip() for s in model.split(",")]
        else:
            # raise TrainingRunError, ("no model name for run %s" % rName)
            # No longer an error not to have any models.
            self.modelNames = []
        # Ensure a canonical order.
        self.modelNames.sort()
        self.testCorpora = testCorpora
        self.engineOptions = engineOptions
        self.enginePrepOptions = enginePrepOptions
        self.testSet = None
        self.flushedStdout = flushedStdout
        self.iterators = iterators
        if iterators:
            for iterator in iterators[:-1]:
                if iterator.mustBeInnermost:
                    raise TrainingRunError("Iterator %s must be innermost, but is not last" % iterator)
        self.allInstances = []
        self.instanceClass = instanceClass

    def _satisfyIterators(self, engine, iterators, subdirName, **kw):
        if not iterators:
            # We've reached the end.
            yield self._createInstance(engine, subdirName, **kw)
        else:
            for newSubdirName, newKw in iterators[0](subdirName, **kw):
                for inst in self._satisfyIterators(engine, iterators[1:], newSubdirName, **newKw):
                    yield inst

    # engineOptions is a keyword, guaranteed to have a dictionary value.
    
    def _createInstance(self, engine, runSubdir, engineOptions = None,
                        # Soak up the extra context that was passed in to the iterators
                        **kw):
        return self.instanceClass(engine, self, runSubdir, 
                                  engineOptions = engineOptions)

    def _configureTestCorpus(self, engine):

        tFiles = []
        # This can be empty in the case of crossvalidation. I don't have
        # the models here yet, and even if I did, I'm not sure this
        # is what I'd want to do with them. I don't do anything with the testSet
        # until the run instance anyway.
        if self.testCorpora:
            for corpusName, partition in self.testCorpora:
                corpus = engine.getCorpus(corpusName)
                tFiles += corpus.getFiles(partition = partition)
            self.testSet = DocumentSet(None, fileList = tFiles, flushedStdout = self.flushedStdout)

    #
    # The sole public method
    #
    
    def yieldInstances(self, engine, **kw):
        # We should prepare the test set here, but I'm not
        # sure where to put it yet. Let's just suppress that until the
        # instance is created. That will be a bunch of extra work
        # in the iterative cases, but no more than what's currently
        # being done. I don't think we should allow the option
        # of iterating on the test prep. For now, we just prepare the
        # test set.
        self._configureTestCorpus(engine)
        for rInst in self._satisfyIterators(engine, self.iterators or [], self.runName,
                                            engineOptions = self.engineOptions,
                                            **kw):
            self.allInstances.append(rInst)
            yield rInst

#
# Some basic iterators
#

# Each iterator has a __call__() method, which should be a
# generator.

import re

class BootstrapIterator(object):

    mustBeInnermost = False
    
    def _newDirName(self, curDirName, attr, val):
        return "%s_%s_%s" % (curDirName, re.sub("\W", "_", str(attr)),
                             re.sub("\W", "_", str(val)))        

class IncrementIterator(BootstrapIterator):

    def __init__(self, optionKey, attribute, startVal, endVal, increment, forceLast = False):
        self.optionKey = optionKey
        self.attribute = attribute
        self.startVal = startVal
        self.endVal = endVal
        self.increment = increment
        self.forceLast = forceLast

    def __call__(self, curSubdirName, **kw):
        v = self.startVal
        if self.optionKey not in kw:
            return
        else:
            while v <= self.endVal:
                # Copy the arguments, and the value of the option key.
                newKw = kw.copy()
                newKw[self.optionKey] = newKw[self.optionKey].copy()
                newKw[self.optionKey][self.attribute] = v
                yield self._newDirName(curSubdirName, self.attribute, v), newKw
                if self.forceLast and (v < self.endVal) and ((v + self.increment) > self.endVal):
                    v = self.endVal
                else:
                    v += self.increment

# ONLY used with the corpus.

class CorpusSizeIterator(IncrementIterator):

    def __init__(self, increment, startVal = None, endVal = None, forceLast = False):

        IncrementIterator.__init__(self, "corpusSettings", "size", startVal or increment, endVal, increment, forceLast = forceLast)

    def __call__(self, curSubdirName, tSetElements = None, **kw):
        if self.endVal is None:
            self.endVal = len(tSetElements)
        for d, newKw in IncrementIterator.__call__(self, curSubdirName, tSetElements = tSetElements, **kw):
            yield d, newKw        

class ValueIterator(BootstrapIterator):

    def __init__(self, optionKey, attribute, valueList):
        self.optionKey = optionKey
        self.attribute = attribute
        self.valueList = valueList

    def __call__(self, curSubdirName, **kw):
        if self.optionKey not in kw:
            return
        else:
            for v in self.valueList:
                # Copy the arguments, and the value of the option key.
                newKw = kw.copy()
                newKw[self.optionKey] = newKw[self.optionKey].copy()
                newKw[self.optionKey][self.attribute] = v
                yield self._newDirName(curSubdirName, self.attribute, v), newKw

#
# And finally,  the bootstrapper itself.
#

class BootstrapError(Exception):
    pass

class Bootstrapper(object):

    def __init__(self, dir = None, task = None, language = None,
                 corpora = None, models = None, runs = None,
                 corpusDirs = None, modelDirs = None, runDirs = None,
                 flushedStdout = sys.stdout):

        self.flushedStdout = flushedStdout
        self.dir = dir
        if dir is None:
            raise BootstrapError("no dir specified")
        self.task = task
        if task is None:
            raise BootstrapError("no task specified")
        # This can be None. 
        if language:
            try:
                self.task.getLanguagesAndCodes()[language]
            except KeyError:
                raise BootstrapError("language %s not supported by task" % language)
        self.language = language
        # A table of the mapping from names to
        # corpora.
        self.corporaTable = {}
        # A list of corpora.
        self.corporaList = []
        # A table of the current mapping from names
        # to model set templates.
        self.modelSetTable = {}
        # A list of model set templates.
        self.modelSetList = []
        # A table of the current mapping from names
        # to training run templates.
        self.runTable = {}
        # A list of run templates.
        self.runList = []

        self.corpusDirs = corpusDirs or {}
        self.modelDirs = modelDirs or {}
        self.runDirs = runDirs or {}

        # Postpone setting the context until the run() method.
        
        if corpora is not None:
            for corpus in corpora:
                cName = corpus.corpusName
                if cName in self.corporaTable:
                    raise BootstrapError("duplicate corpus name '%s'" % cName)
                self.corporaTable[cName] = corpus
                self.corporaList.append(corpus)

        if models is not None:
            for model in models:
                mName = model.modelName
                if mName in self.modelSetTable:
                    raise BootstrapError("duplicate model set name '%s'" % mName)
                for cName, mPart in model.trainingCorpora:
                    if cName not in self.corporaTable:
                        raise BootstrapError("model '%s' requires unknown corpus '%s'" % (mName, cName))
                self.modelSetTable[mName] = model
                self.modelSetList.append(model)

        if runs is not None:
            for run in runs:
                rName = run.runName
                if rName in self.runTable:
                    raise BootstrapError("duplicate run name '%s'" % rName)
                # Should be OK not to have models, now.
                if run.modelNames is None:
                    # raise BootstrapError, ("run %s has no model set" % rName)
                    run.modelNames = []
                # The corpora have to be the same across the models.
                commonCorpus = None
                for mName in run.modelNames:
                    if mName not in self.modelSetTable:
                        raise BootstrapError("run %s requires unknown model set %s" % (rName, mName))
                    mTemplate = self.modelSetTable[mName]
                    if commonCorpus is None:
                        commonCorpus = set(mTemplate.trainingCorpora)
                    elif commonCorpus != set(mTemplate.trainingCorpora):
                        raise BootstrapError("run %s has multiple models, but they're trained on different corpora" % rName)
                for rCorpus, rPart in run.testCorpora:
                    if rCorpus not in self.corporaTable:
                        raise BootstrapError("run %s requires unknown corpus %s" % (rName, rCorpus))
                self.runTable[rName] = run
                self.runList.append(run)

    # c is a DocumentSet.
    
    def getCorpusDir(self, c):
        cName = c.corpusName
        cSubdir = cName
        # Impose directory defaults. If dir is absent, use
        # "corpora". If dir isn't absolute, prepend the experiment engine dir.
        try:
            cDir = self.corpusDirs[cName]
        except KeyError:
            cDir = "corpora"
        if not os.path.isabs(cDir):
            cDir = os.path.join(self.dir, cDir)
        return os.path.join(cDir, cSubdir)

    # m is a TrainingRunInstance.
    
    def getModelDir(self, m):
        mName = m.modelName
        mSubdir = m.modelSubdir
        try:
            mDir = self.modelDirs[mName]
        except KeyError:
            mDir = "model_sets"
        if not os.path.isabs(mDir):
            mDir = os.path.join(self.dir, mDir)
        return os.path.join(mDir, mSubdir)

    # mTemplate is a TrainingRun.
    
    def getModelDirPrefix(self, mTemplate):
        mName = mTemplate.modelName
        try:
            mDir = self.modelDirs[mName]
        except KeyError:
            mDir = "model_sets"
        if not os.path.isabs(mDir):
            mDir = os.path.join(self.dir, mDir)
        return os.path.join(mDir, mName)

    # r is a TestRunInstance.
    
    def getRunDir(self, r):
        rName = r.runName
        rSubdir = r.runSubdir
        try:
            rDir = self.runDirs[rName]
        except KeyError:
            rDir = "runs"
        if not os.path.isabs(rDir):
            rDir = os.path.join(self.dir, rDir)
        return os.path.join(rDir, rSubdir)

    # rTemplate is a TestRun.
    
    def getRunDirPrefix(self, rTemplate):
        rName = rTemplate.runName
        try:
            rDir = self.runDirs[rName]
        except KeyError:
            rDir = "runs"
        if not os.path.isabs(rDir):
            rDir = os.path.join(self.dir, rDir)
        return os.path.join(rDir, rName)
    
    def getCorpus(self, cName):
        return self.corporaTable[cName]

    def getModel(self, mName):
        return self.modelSetTable[mName]

    # So the new workflow is this. 
    
    def run(self, interleave = True):
        
        # To run it, we need to prepare each corpus, build each model,
        # execute each run.
        
        origT = time.time()
        
        print("Run began at", time.ctime(origT), file=self.flushedStdout)

        hitError = False

        try:
            try:

                # So the idea is that we start with the runs,
                # iterating through the corpora and then models.
                
                # Prepare each corpus.
                for c in self.corporaList:
                    c.setContext(self)
                    cDir = self.getCorpusDir(c)
                    print("\n### Preparing corpus", c.corpusName, file=self.flushedStdout)
                    c.prepare()
                    print("### Done.", file=self.flushedStdout)

                # We can't do interleave if some of the runs require more than
                # one model, or if they have no models.
                
                for rTemplate in self.runList:
                    if len(rTemplate.modelNames) != 1:
                        interleave = False
                        break

                # We can't do interleaved scoring runs if some of the models are
                # already done. So I need to check first.

                for mTemplate in self.modelSetList:
                    for m in mTemplate.yieldInstances(self):
                        print("\n### Building model set", m.modelName, "in directory", m.modelSubdir, file=self.flushedStdout)
                        # If interleave is True, we'll do the runs.
                        m.train(interleave = interleave)
                        print("### Done.", file=self.flushedStdout)

                def runRecurse(r, mTemplates, mList = None):
                    if not mTemplates:
                        # We're at the bottom, or there were no mTemplates
                        # in the first place.
                        if mList is None: mList = []
                        r.setContext(self, mList)
                        r.prepareTest()
                        print("\n### Performing run", r.runName, "in directory", r.runSubdir, file=self.flushedStdout)
                        r.test()
                        print("### Done.", file=self.flushedStdout)
                        r.finishTest()
                    else:
                        if mList is None:
                            mList = []
                        for m in mTemplates[0].yieldInstances(self):
                            runRecurse(r, mTemplates[1:], mList + [m])
                
                if not interleave:
                    # Perform each run.
                    for rTemplate in self.runList:
                        mTemplates = [self.getModel(mName) for mName in rTemplate.modelNames]
                        # I need to do all possible pairs, which means I need to recurse.
                        for r in rTemplate.yieldInstances(self):
                            runRecurse(r, mTemplates)

            except Exception as e:

                if ExecutionContext._DEBUG:
                    hitError = True
                raise
                                    
        finally:

            if not (ExecutionContext._DEBUG and hitError):

                self.finish(origT)

    def finish(self, origT):
        pass
