# Copyright (C) 2007 - 2018 The MITRE Corporation. See the toplevel
# file LICENSE.txt for license terms.

from __future__ import print_function, absolute_import
from .PyCompat import _unicodestring, Py3

# The contents of this module have been moved here to avoid
# various import circularities that have been lurking in MAT for
# a long time. They are typically elements that rely on
# properties of the task, but don't need to refer to PluginMgr
# by name.

from .Error import PluginError

#
# Region filtering and span inference
#

# The utility which computes regions for model building and
# scoring is used in a couple places, and was originally
# defined as class methods. This will make it easy to retrofit.

class RegionFilterer(object):

    # This is used for the spanless annotations.
    @classmethod
    def _computeImpliedSpan(cls, a, impliedSpans):
        if a.atype.hasSpan:
            return (a.start, a.end)
        else:
            try:
                return impliedSpans[a]
            except KeyError:
                start = end = None
                for attrObj, val in zip(a.atype.attr_list, a.attrs):
                    if (attrObj._typename_ == "annotation") and (val is not None):
                        if not attrObj.aggregation:
                            thisStart, thisEnd = cls._computeImpliedSpan(val, impliedSpans)
                            if (start is None) or (thisStart < start):
                                start = thisStart
                            if (end is None) or (thisEnd > end):
                                end = thisEnd
                        else:
                            for subval in val:
                                thisStart, thisEnd = cls._computeImpliedSpan(subval, impliedSpans)
                                if (start is None) or (thisStart < start):
                                    start = thisStart
                                if (end is None) or (thisEnd > end):
                                    end = thisEnd
                # Might be (None, None)
                impliedSpans[a] = (start, end)
                return (start, end)    

    # In order to deal with spanless annotations via their implied spans,
    # I'm going to do some fancy dancing.

    @classmethod
    def filterByRegions(cls, orderedRegionList, spannedLists = None, spanlessLists = None, maxRegionHash = None):

        # Ensure the annot lists are ordered, and collect the things that
        # we can get implied spans for. If we can't compute a span,
        # use the max region extents for the document.

        spannedTuples = []
        
        if spannedLists:
            for l in spannedLists:
                l.sort(key = lambda x: x.start)
            spannedTuples = [[(a, a.start, a.end) for a in spannedList] for spannedList in spannedLists]

        spanlessTuples = []
        
        if spanlessLists:
            impliedSpans = {}
            for spanlessList in spanlessLists:
                tupleList = []
                for a in spanlessList:
                    start, end = cls._computeImpliedSpan(a, impliedSpans)
                    if start is not None:
                        tupleList.append((a, start, end))
                    else:
                        try:
                            start, end = maxRegionHash[a.doc]
                            tupleList.append((a, start, end))
                        except KeyError:
                            pass
                tupleList.sort(key = lambda x: x[1])
                spanlessTuples.append(tupleList)

        annotLists = spannedTuples + spanlessTuples

        annotListCount = len(annotLists)
        idxList = [0] * annotListCount
        finalLists = [[] for l in annotLists]
        
        rI = 0
        
        while rI < len(orderedRegionList):
            curRegion = orderedRegionList[rI]
            rI += 1
            j = 0
            while j < annotListCount:
                annotTupleList = annotLists[j]
                finalList = finalLists[j]
                while idxList[j] < len(annotTupleList):
                    curA, start, end = annotTupleList[idxList[j]]
                    if start < curRegion[0]:
                        # discard.
                        idxList[j] += 1
                    elif end <= curRegion[1]:
                        # keep.
                        finalList.append(curA)
                        idxList[j] += 1
                    else:
                        # wait.
                        break
                j += 1
        # So we need to return the spanned tuples and the spanless tuples
        # separately.
        return tuple(finalLists[:len(spannedTuples)]), tuple(finalLists[len(spannedTuples):])

#
# WriteableTable
#

# This was originally in Score.py, but we use it in MATTransducer and MATReport
# and I may want those in contexts where I don't even have Score.py yet.

# Note that in Python 3, csv does Unicode, and the files must be opened in text mode.
# In Python 2, it does NOT do unicode, and the files must be opened in binary
# mode. The excellent "Supporting Python 3" describes a wrapper class which
# handles this cross-platform; conveniently, I've already written my own.
# Everything that writes to CSV in MAT appears to go through this class,
# conveniently enough.

class WriteableTable(object):

    def writeCSV(self, path, colNames, rows):

        import csv
        # CSV rows.

        # The contents of the documents being scored are all Unicode strings,
        # so if those strings happen to be in there (and they might be),
        # we need to be able to deal with that. Note the documentation
        # on the CSV module; it doesn't actually do Unicode. So what you need
        # to do is encode them as UTF-8.

        if Py3:
            fp = open(path, "wt", encoding = "utf-8", newline = "")
            w = csv.writer(fp)
            w.writerow(colNames)
            w.writerows(rows)
            fp.close()
        else:
            def utf8writer(w, row):
                w.writerow([((type(a) is _unicodestring) and a.encode('utf-8')) or a for a in row])
            fp = open(path, "wb")
            w = csv.writer(fp)
            utf8writer(w, colNames)
            for row in rows:
                utf8writer(w, row)
            fp.close()

    def format(self, colNames, fmtRows):

        # Be careful; if fmtRows is empty, then max(n, *[]) will fail.
        if len(fmtRows) == 0:
            fmtString = " ".join("%" + str(len(colNames[i])) + "s" for i in range(len(colNames)))
        else:
            fmtString = " ".join(["%" + str(max(len(colNames[i]), *[len(row[i]) for row in fmtRows])) + "s" for i in range(len(colNames))])

        sList = [fmtString % tuple(colNames),
                 fmtString % (tuple(["-----"]) * len(colNames))]

        for row in fmtRows:
            sList.append(fmtString % tuple(row))
        
        return "\n".join(sList)
    
#
# Task-level general annotation information
#

# The annotation type sorters are used in various places throughout
# the system. These originally were in PluginMgr.py, but this
# ultimately led to import circularities once I really cleaned
# up the imports.

# The concept of content vs. token annotations comes up in a number of places.
# I've written some awful code to manage this, and it needs to be cleaned up.

class AnnotationTypeSorter(object):

    def __init__(self):
        self.cache = {}

    def getLabelType(self, atype):
        label = atype.lab
        if label not in self.cache:
            self.cache[label] = [atype.hasSpan, self.computeLabelType(label)]
        elif self.cache[label][0] != atype.hasSpan:
            raise PluginError("label '%s' is both spanned and spanless in the label cache" % label)
        return self.cache[label][1]

    def isContentType(self, atype):
        return self.getLabelType(atype) == self.CONTENT

    def isTokenType(self, atype):
        return self.getLabelType(atype) == self.TOKEN

    [CONTENT, TOKEN, OTHER] = [0, 1, 2]

    # Implemented by the children.
    def computeLabelType(self, label):
        raise PluginError("not implemented")            

class TaskAnnotationTypeSorter(AnnotationTypeSorter):

    def __init__(self, task, contentAnnotationSets = None):
        AnnotationTypeSorter.__init__(self)
        self.task = task
        if contentAnnotationSets is not None:
            self.contentAnnotationSets = contentAnnotationSets
        else:
            self.contentAnnotationSets = self.task.getNonAdminAnnotationSetNames()
        self.allContentLabels = self.task.getAnnotationTypeRepository().getLabelsForAnnotationSetNames(self.contentAnnotationSets, includeAttributeOnly = True)[0]

    def computeLabelType(self, label):
        # There are only three categories: content, token, other.
        if self.task.getCategoryForLabel(label) == "token":
            return self.TOKEN
        elif label in self.allContentLabels:
            # Note that I HAVE to get the whole label set, since getting the
            # set for the label will mess up if the label is one which hosts
            # an attribute in the content annotation sets, but isn't actually
            # in the sets.
            return self.CONTENT
        else:
            return self.OTHER

class NonTaskAnnotationTypeSorter(AnnotationTypeSorter):

    def __init__(self, contentAnnotations, tokenAnnotations, contentAnnotationByDefault = False):
        AnnotationTypeSorter.__init__(self)
        self.contentAnnotations = contentAnnotations
        self.tokenAnnotations = tokenAnnotations
        self.contentAnnotationByDefault = contentAnnotationByDefault
    def computeLabelType(self, label):
        # There are only three categories: content, token, other.
        if (self.contentAnnotations and (label in self.contentAnnotations)) or \
           (self.contentAnnotationByDefault and ((self.tokenAnnotations is None) or (label not in self.tokenAnnotations))):
            return self.CONTENT
        elif self.tokenAnnotations and label in self.tokenAnnotations:
            return self.TOKEN
        else:
            return self.OTHER
