# Copyright (C) 2018 The MITRE Corporation. See the toplevel
# file LICENSE.txt for license terms.

from __future__ import print_function, absolute_import
from .PyCompat import _strtypes

# XML reader/writer.

from .Operation import OpArgument, OptionTemplate
from .Document import LoadError
from .Annotation import (AnnotationAttributeType, StringAttributeType, FloatAttributeType,
                         BooleanAttributeType, IntAttributeType, AttributeValueList,
                         AttributeValueSet, AnnotationError,
                         AnnotationCore, AttributeValueSequence)

from .DocumentIO import (declareDocumentIO, DocumentFileIO, SaveError)
from .PyCompat import _jsondumps, Py3

import xml.parsers.expat, xml.sax.saxutils, re, base64, sys, json

# We need to know what the annotations are in the task,
# if we're supposed to filter out the "junk" tags and only translate
# the tags for the task.

# There are going to be a couple special things in here
# now. First, there will be the <_mat:atypes> element, which
# optionally declares all the available annotation types and
# their attributes and types. This is important for recording
# spanless annotations and for recording annotations which have
# annotation-valued attributes (and the attribute for that will be
# _mat:id). So <_mat:atypes> looks like this:
# <_mat:atypes><atype name="..." hasSpan="yes|no"><attr name="..." type="string|attribute"/>...</atype></atypes>
# It will have no whitespace at all, so it doesn't count toward
# the offsets, and it should appear, if it appears at all,
# before the first annotation. If I can, when I render, I'll
# put it at the beginning of the document, but if it's
# an overlay, it'll go immediately before the first annotation.

class _ParserState(object):

    def __init__(self, annotDoc, inputIsOverlay, translateAll):
        self.inputIsOverlay = inputIsOverlay
        self.xmlTranslateAll = translateAll
        if self.xmlTranslateAll:
            # The annotDoc will likely come in with an annotation repository.
            # It's not enough to cut it off from the global repository;
            # we have to convince it that the annotation types it ALREADY
            # has are not closed.
            annotDoc.unlockAtypeRepository()
        self.annotDoc = annotDoc
        # The parser expects BYTES, not unicode characters.
        # The documentation does not make this clear at all.
        self.parser = xml.parsers.expat.ParserCreate("UTF-8")
        self.parser.StartElementHandler = self._startElement
        self.parser.EndElementHandler = self._endElement
        self.parser.CharacterDataHandler = self._handleCdata
        self.parser.CommentHandler = self._handleComment
        self.parser.DefaultHandler = self._handleDefault
        self.stack = []
        self.pos = 0
        self.signalChunks = []
        self._digestingAtypes = False
        self._digestedAtypes = False
        self._curAtype = None
        # The annotation attributes must be postponed, because some of
        # them may be annotation-valued.
        self._annotPairs = []
        self._idMap = {}

    def _addSignalChunk(self, chunk):
        self.signalChunks.append(chunk)
        self.pos += len(chunk)

    # VERY subtle bug. Because inputIsOverlay counts in BYTES,
    # and because I have to pass a byte string to the parser, I
    # can't store the input as a Unicode string; I have to encode
    # it into UTF-8 first.
    
    def parse(self, s):
        # Boy, this is hideous. XML doesn't permit CRLF - each
        # XML processor is obligated to convert them to LF. The
        # only way to block that is to replace \r with the
        # appropriate entity - but then the parser barfs on any
        # &#xD outside the toplevel tag, because it's not actual whitespace.
        # Poo.
        # So the answer is to do the replacement ONLY inside the
        # toplevel tag. That's easy enough.
        if s.find("\r") > -1:
            # Find the first toplevel tag.
            # Exclude ! (comment and doctype) and ? (xmldesc and processing instruction).
            # This will barf if there's nothing in the toplevel tag.
            mStart = re.search("<([^!?][^\s>]*)", s)
            if mStart is not None:
                # Gotta be something in the toplevel tag in order for
                # me to care.
                topStart = mStart.start()
                topEnd = re.search("</" + mStart.group(1) + ">", s).end()
                s = s[:topStart] + s[topStart:topEnd].replace("\r", "&#xD;") + s[topEnd:]
        # In Python 3, Parse() can accept either a string or bytes.
        # So it's harmless to specifically encode it.
        self.input = s.encode('utf-8')
        self.parser.Parse(self.input, True)
        # Now that we're done parsing, we have to update all the
        # attributes for those which have annotation-valued attributes.
        # Well, it's a LOT more complicated than that. We should be only
        # doing this with annotation-valued attributes. The reason is that
        # for ENAMEX-style attributes, you have to know what the value of
        # some of the attributes is in order to recognize that the annotation
        # is a legal value.
        # But I also have to interpret the value types correctly. So I need
        # to set aside the annotation attributes and add them in a second pass.
        atypeAttrHash = {}
        annotValuedPairs = []
        for newAnnot, attrDict in self._annotPairs:
            atype = newAnnot.atype
            try:
                attrHash, annotValuedAttrNames = atypeAttrHash[atype]
            except KeyError:
                attrHash = {}
                annotValuedAttrNames = set()
                for t in atype.attr_list:
                    # Barf.
                    if isinstance(t, AnnotationAttributeType):
                        # The IDs are all strings. We need to look
                        # them up in the _idMap.
                        if t.aggregation == "list":
                            attrHash[t.name] = lambda x: AttributeValueList([self._idMap[v] for v in x.split(",")])
                        elif t.aggregation == "set":
                            attrHash[t.name] = lambda x: AttributeValueSet([self._idMap[v] for v in x.split(",")])
                        else:
                            attrHash[t.name] = lambda x: self._idMap[x]
                        annotValuedAttrNames.add(t.name)
                    elif isinstance(t, StringAttributeType):
                        if t.aggregation == "set":
                            attrHash[t.name] = lambda x: AttributeValueSet(x.split(","))
                        elif t.aggregation == "list":
                            attrHash[t.name] = lambda x: AttributeValueList(x.split(","))
                        else:
                            attrHash[t.name] = lambda x: x
                    else:
                        if isinstance(t, FloatAttributeType):
                            mapper = float
                        elif isinstance(t, IntAttributeType):
                            mapper = int
                        elif isinstance(t, BooleanAttributeType):
                            mapper = lambda x: x == "yes"
                        if t.aggregation == "set":
                            attrHash[t.name] = lambda x: AttributeValueSet([mapper(v) for v in x.split(",")])
                        elif t.aggregation == "list":
                            attrHash[t.name] = lambda x: AttributeValueList([mapper(v) for v in x.split(",")])
                        else:
                            attrHash[t.name] = mapper
                atypeAttrHash[atype] = attrHash, annotValuedAttrNames
            
            theseAnnotValued = {}
            for k, v in attrDict.items():
                if k in annotValuedAttrNames:
                    theseAnnotValued[k] = v
                else:
                    newAnnot[k] = attrHash[k](v)
            if theseAnnotValued:
                annotValuedPairs.append((newAnnot, theseAnnotValued))
        # Now, finally, add the annotation-valued attributes.
        for newAnnot, theseAnnotValued in annotValuedPairs:
            attrHash, ignore = atypeAttrHash[newAnnot.atype]
            for k, v in theseAnnotValued.items():
                newAnnot[k] = attrHash[k](v)
        newSignal = "".join(self.signalChunks)
        if self.annotDoc.signal and (self.annotDoc.signal != newSignal):
            raise LoadError("signal from XML file doesn't match original signal")
        self.annotDoc.signal = newSignal
        if self.inputIsOverlay:
            self.annotDoc.metadata["signal_type"] = "xml"            

    # In Python 3, the input and the collected strings
    # are of different, unmixable types. So we have to manage that.
    # Fortunately, in Python 2, the b"" notation is identical to "".

    if Py3:
        def _findTagForOverlay(self):
            return self.input[self.parser.CurrentByteIndex:self.input.find(b">", self.parser.CurrentByteIndex) + 1].decode("utf-8")
    else:
        def _findTagForOverlay(self):
            return self.input[self.parser.CurrentByteIndex:self.input.find(">", self.parser.CurrentByteIndex) + 1]        

    # SAM 6/2/14: If we want to be able to interpret zero-length spans as spanless
    # annotations, we're going to have to postpone creating the annotations (and their types) until
    # endElement if types weren't digested.
    
    def _startElement(self, eltName, attrDict):
        if eltName == "_mat:atype":
            if not self._digestingAtypes:
                raise LoadError("found _mat:atype outside _mat:atypes")
            self._digestAtype(attrDict)
            return
        elif eltName == "_mat:attr":
            if not self._digestingAtypes:
                raise LoadError("found _mat:attr outside _mat:atypes")
            self._digestAttr(attrDict)
            return
        elif eltName == "_mat:atypes":
            self._digestingAtypes = True
            return
        if (not self.xmlTranslateAll) and (not self.annotDoc.findAnnotationType(eltName, create = False)):
            if self.inputIsOverlay:
                # There doesn't appear to be any way of finding the
                # end of the element in the input from the parser bindings, grrr.
                # So we have to look for ">" in the signal. 
                self._addSignalChunk(self._findTagForOverlay())
            return
        self.stack[0:0] = [(eltName, attrDict, self.pos)]
        
    def _digestAtype(self, attrDict):
        self._curAtype = self.annotDoc.findAnnotationType(attrDict["name"], hasSpan  = attrDict["hasSpan"] == "yes")

    def _digestAttr(self, attrDict):
        self._curAtype.ensureAttribute(attrDict["name"], aType = attrDict.get("type"),
                                       aggregation = attrDict.get("aggregation"))

    def _endElement(self, eltName):
        if eltName in ("_mat:attr", "_mat:atype"):
            return
        elif eltName == "_mat:atypes":
            self._digestingAtypes = False
            self._digestedAtypes = True
            return
        # Ugh. You have to be careful when adding the input back in - this method is called
        # on self-closing elements, too. For normal elements, eltName is the same
        # as </ + eltName at the current byte index; for self-closing elements, that's not
        # true; in fact, it points at something past the />. So searching for the
        # next > is a bad idea; in fact, nothing should be done, because that search in the
        # _startElement method already gobbled up the />.
        if (not self.xmlTranslateAll) and (not self.annotDoc.findAnnotationType(eltName, create = False)):
            if self.inputIsOverlay:
                tag = self._findTagForOverlay()
                if tag == "</" + eltName + ">":
                    self._addSignalChunk(tag)
            return
        startEltName, attrDict, startPos = self.stack[0]
        if startEltName == eltName:
            # special case: startPos == pos and types weren't digested.
            # In this case, the annotation type MAY be spanless. So just see if
            # you find one. But if the type doesn't exist, it may not
            # raise an error.
            if (self.pos == startPos) and (not self._digestedAtypes):
                # Type type can be spanned or spanless. Check for the spanned one. If you get an
                # error, check for the spanless one. Otherwise, create a spanned one.
                try:
                    t = self.annotDoc.findAnnotationType(eltName, create = False)
                except AnnotationError:
                    t = self.annotDoc.findAnnotationType(eltName, hasSpan = False, create = False)
                if t is None:
                    t = self.annotDoc.findAnnotationType(eltName)
            else:
                # If the atypes were declared, you don't want to check the attributes.
                t = self.annotDoc.findAnnotationType(eltName, create = not self._digestedAtypes)
            if not self._digestedAtypes:
                for key in attrDict.keys():
                    if key != "_mat:id":
                        t.ensureAttribute(key)
            elif t is None:
                # There's one case where we don't want to fail here: if we're reading a __top
                # label and it's the outermost annotation, and xmlTranslateAll is True (which happens
                # if it's explicitly requested, or no task is provided). If xmlTranslateAll is false
                # and there's no type, the second "if" in this method above claims the case
                # and the end tag is discarded (which is good, because there will be no
                # corresponding start tag on the top of the stack). Here, we assume that __top was inserted
                # as a "safe" XML toplevel wrapper (see addTop in writeToUnicodeString in XMLDocumentIO).
                # Could this ever get me in trouble? I don't think so. If the atypes were
                # present in the document, and this isn't part of an XML overlay, that's what we
                # go with, or if we have a task, that's what we go with.
                if (eltName == "__top") and (len(self.stack) == 1):
                    # Just pop it.
                    self.stack = []
                    return
                else:
                    raise LoadError("annotation types were recorded in the XML, but the label '%s' is unknown" % eltName)
            if t.hasSpan:
                newAnnot = self.annotDoc.createAnnotation(startPos, self.pos, t, blockAdd = True)
            else:
                newAnnot = self.annotDoc.createSpanlessAnnotation(t, blockAdd = True)
            id = attrDict.get("_mat:id")
            if id is not None:
                newAnnot.setID(id)
                self._idMap[id] = newAnnot
                del attrDict["_mat:id"]
            self.annotDoc._addAnnotation(newAnnot)
            self._annotPairs.append((newAnnot, attrDict))
            self.stack[0:1] = []
        else:
            raise LoadError("annotation close doesn't match annotation open")

    # I'm going to do something special with comments: I'm going to
    # dump the document metadata in a comment when I encounter it. I believe
    # the comment can appear after the final tag, so I'll put it at the very
    # end of the file.

    METADATA_PAT = re.compile("^ _mat_metadata_ (.*) $")
    
    def _handleComment(self, data):
        m = self.METADATA_PAT.match(data)
        if m is None:
            if self.inputIsOverlay:
                # Pass through the comment.
                self._addSignalChunk("<!--")
                self._addSignalChunk(data)
                self._addSignalChunk("-->")
        else:
            # We've got the metadata.
            # In Python 3, we have to do something special, because
            # strings and bytes are different types now. So we have to
            # encode the base64 chunk as ascii, to get bytes (or a str in Python 2),
            # and then b64decode, and then ascii-decode the result
            # to pass to json.loads().
            jsonMetadata = base64.b64decode(m.group(1).encode("ascii")).decode("ascii")
            self.annotDoc.metadata = json.loads(jsonMetadata)

    def _handleCdata(self, data):
        if self.inputIsOverlay:
            data = xml.sax.saxutils.escape(data)
        self._addSignalChunk(data)

    # Pass through anything you find if it's a default.
    def _handleDefault(self, data):
        if self.inputIsOverlay:
            self._addSignalChunk(data)

# Utilizes the algorithm for cmp_to_key in Python 3,
# because there are no longer dedicated comparison functions
# in Python 3. I COULD add these to the annotation object,
# but I'mnot about to do that.

class _AnnotComparisonKey(object):
    def __init__(self, obj, *args):
        self.obj = obj
    def __lt__(self, other):
        return self._cmpAnnot(self.obj, other.obj) < 0
    def __gt__(self, other):
        return self._cmpAnnot(self.obj, other.obj) > 0
    def __eq__(self, other):
        return self._cmpAnnot(self.obj, other.obj) == 0
    def __le__(self, other):
        return self._cmpAnnot(self.obj, other.obj) <= 0
    def __ge__(self, other):
        return self._cmpAnnot(self.obj, other.obj) >= 0
    def __ne__(self, other):
        return self._cmpAnnot(self.obj, other.obj) != 0
    def _cmpAnnot(self, ann1, ann2):
        if ann1.start > ann2.start:
            return 1
        elif ann1.start < ann2.start:
            return -1
        elif ann1.end > ann2.end:
            return -1
        elif ann1.end < ann2.end:
            return 1
        else:
            return 0

class XMLDocumentIO(DocumentFileIO):

    def __init__(self, xml_input_is_overlay = False, xml_translate_all = False, signal_is_xml = False,
                 xml_output_tag_exclusions = None, xml_output_exclude_metadata = False,
                 encoding = None, **kw):
        # Changing the default encoding.
        if encoding is None:
            encoding = "utf-8"
        DocumentFileIO.__init__(self, encoding = encoding, **kw)
        if xml_input_is_overlay:
            signal_is_xml = True
        self.excludeMetadata = xml_output_exclude_metadata
        self.xmlInputIsOverlay = xml_input_is_overlay
        self.tagExclusions = None
        if xml_output_tag_exclusions is not None:
            if type(xml_output_tag_exclusions) in _strtypes:
                xml_output_tag_exclusions = xml_output_tag_exclusions.split(",")
                if xml_output_tag_exclusions == [""]:
                    xml_output_tag_exclusions = []
            self.tagExclusions = dict([(t, True) for t in xml_output_tag_exclusions])
        self.signalIsXML = signal_is_xml
        self.xmlTranslateAll = xml_translate_all

    # This must be implemented by the children. s is a Unicode string.
    # annotDoc is an annotated document.
    
    def deserialize(self, s, annotDoc):
        # If there's no global annotation type repository, we want xmlTranslateAll to be True.
        state = _ParserState(annotDoc, self.xmlInputIsOverlay, self.xmlTranslateAll or (not annotDoc.atypeRepository.globalTypeRepository))
        state.parse(s)        
    
    def writeToUnicodeString(self, annotDoc):
        signalIsXML = self.signalIsXML or \
                      (("signal_type" in annotDoc.metadata) and
                       annotDoc.metadata["signal_type"] == "xml")
        # Get all the annotations. Let's not care about overlap right now,
        # since overlap will happen if I'm writing everything out, because
        # it'll be nested. So just get the annotations and then
        # sort them, and if we ever get a crossing dependency, we'll
        # have to check otherwise.
        # Split the atypes into spanned and spanless.
        spanned = []
        spanless = []
        # I used to remove the tag exclusions when I collect the
        # indices, but I need to do it earlier in order to figure out
        # if I need a top element or not.        
        annots = []
        spanlessAnnots = []
        for atype in annotDoc.atypeDict.keys():
            if self.tagExclusions and (atype.lab in self.tagExclusions):
                continue
            if atype.hasSpan:
                spanned.append(atype.lab)
            else:
                spanless.append(atype.lab)
        if spanned:
            annots = annotDoc.getAnnotations(atypes = spanned)
        if spanless:
            spanlessAnnots = annotDoc.getAnnotations(atypes = spanless)
        # We now know they can nest. So let's sort them.        
        # Sort them first by earliest start, latest end.
        # Changed to the key from cmp for Python 3 compatibility.
        annots.sort(key = _AnnotComparisonKey)
        # Now, I can loop through the annots, and keep a stack,
        # and we know when to pop the stack because of
        # how the indexes work.
        indices = {}
        lastAnnot = None
        for annot in annots:
            if lastAnnot and \
               (((lastAnnot.start < annot.start) and (lastAnnot.end < annot.end)) or \
                ((lastAnnot.start > annot.start) and (lastAnnot.end > annot.end))):
                raise SaveError("crossing dependencies")
            try:
                indices[annot.start][0].append(annot)
            except KeyError:
                indices[annot.start] = [[annot], []]
            try:
                indices[annot.end][1].append(annot)
            except KeyError:
                indices[annot.end] = [[], [annot]]
        indexList = sorted(indices.keys())
        segs = []
        # So we need to add a toplevel XML tag if we don't already have one, and if
        # we're not adding our own info.
        # The signal is not XML, and
        # There are no spanless annotations, or
        # The maximal annotation starts after the beginning, or
        # the maximal annotation ends before the end, or
        # there are spanless annots, which will be inserted as zero-length
        # annots before the first spanned annotation, or
        # we add metadata, which will insert the annotation types
        # in a similar position.
        addTop = (not signalIsXML) and \
                 ((not annots) or \
                  (annots[0].start > 0) or \
                  (annots[0].end < len(annotDoc.signal)) or \
                  spanlessAnnots or \
                  (not self.excludeMetadata))
        if addTop:
            segs.append("<__top>")
        # Here's a special, sick case: let's say you have some zero-length
        # annotations which were read as such because they were spanless in an
        # XML export, and they're not declared for the document that was read,
        # and you have a __top annotation already. You must ensure that the
        # annotations which start and end at 0 appear AFTER the start of the
        # encompassing toplevel annotation.
        pos = 0
        atypesInserted = False
        # Special case: if the annotation has length 0, then
        # make sure you don't end it first.
        for i in indexList:
            if pos < i:
                seg = annotDoc.signal[pos:i]
                if not signalIsXML:
                    seg = xml.sax.saxutils.escape(seg)
                segs.append(seg)
                pos = i
            [starts, ends] = indices[i]
            # Reverse the ends.
            ends.reverse()
            both = set(starts) & set(ends)
            for endAnnot in ends:
                if endAnnot not in both:
                    segs.append("</" + endAnnot.atype.lab + ">")
            zeroLengthPostponed = None
            if both:
                if (i == 0) and (not addTop):
                    zeroLengthPostponed = both
                else:
                    for bAnnot in both:
                        # Self-terminating.
                        segs.append(self._formatAnnot(bAnnot, spanless = True))
            for startAnnot in starts:
                if startAnnot in both:
                    continue
                if not atypesInserted:
                    if not self.excludeMetadata:
                        segs.append(self._formatAtypes(annotDoc))
                    atypesInserted = True
                    if spanlessAnnots:
                        for sAnnot in spanlessAnnots:
                            segs.append(self._formatAnnot(sAnnot, spanless = True))
                segs.append(self._formatAnnot(startAnnot))
                if zeroLengthPostponed:
                    for bAnnot in zeroLengthPostponed:
                        segs.append(self._formatAnnot(bAnnot, spanless = True))
                    zeroLengthPostponed = None
        if pos < len(annotDoc.signal):
            seg = annotDoc.signal[pos:]
            if not signalIsXML:
                seg = xml.sax.saxutils.escape(seg)
            segs.append(seg)
        if addTop:
            segs.append("</__top>")
        if not self.excludeMetadata:
            # In Python 3, b64encode explicitly requires bytes.
            # In Python 2, a Unicode string is fine as long as there are
            # no ascii characters. So I dump the metadata to a Unicode string,
            # ensuring ASCII (the default), and encode. And then, to
            # concatenate it, I have to decode the result as ascii.
            segs.append("<!-- _mat_metadata_ "+ base64.b64encode(_jsondumps(annotDoc.metadata).encode("ascii")).decode("ascii") + " -->")
        return "".join(segs)

    def _formatAtypes(self, annotDoc):
        segs = ["<_mat:atypes>"]
        for atype in annotDoc.atypeDict.keys():
            segs.append("<_mat:atype name='%s' hasSpan='%s'>" %
                        (atype.lab, (atype.hasSpan and "yes") or "no"))
            for attr in atype.attr_list:
                segs.append("<_mat:attr name='%s' type='%s' aggregation='%s'/>" % (attr.name, attr._typename_, attr.aggregation or "none"))
            segs.append("</_mat:atype>")
        segs.append("</_mat:atypes>")
        return "".join(segs)

    def _formatAnnot(self, annot, spanless = False):
        elts = ["<", annot.atype.lab]
        if annot.attrs:
            for attr, val in zip(annot.atype.attr_list, annot.attrs):
                if val is not None:
                    # Handle annotations specially.
                    if attr._typename_ == "annotation":
                        if attr.aggregation:
                            v = ",".join([str(a.id) for a in val])
                        else:
                            v = str(val.id)
                    else:
                        v = attr.toStringNonNull(val)
                    elts.append(" " + attr.name + "="+ xml.sax.saxutils.quoteattr(v))
        if annot.id is not None:
            elts.append(" _mat:id=" + xml.sax.saxutils.quoteattr(annot.id))
        if spanless:
            elts.append("/>")
        else:
            elts.append(">")
        return "".join(elts)

    # Best to use the OpArgument infrastructure, so we can extract
    # arguments cleanly from CGI and cmdline.

    args = OptionTemplate(input = [OpArgument("xml_input_is_overlay",
                                              help = "If specified, the input XML will be treated as a mix of task-relevant annotations and underlying XML, and the extracted signal will be a well-formed XML file"),
                                   OpArgument("xml_translate_all",
                                              help = "If specified, all tags will be converted, whether or not they're found in whatever task is specified")],
                          input_heading = "Options for XML input",
                          output = [OpArgument("signal_is_xml",
                                               help = "If specified, the underlying signal will be treated as a well-formed XML file when the output file is rendered. If the input file type is also 'xml-inline', use the --xml_input_is_overlay flag to control this setting instead."),
                                    OpArgument("xml_output_tag_exclusions",
                                               help = "A comma-delimited list of annotation labels to exclude from the XML output",
                                               hasArg = True),
                                    OpArgument("xml_output_exclude_metadata",
                                               help = "Normally, the XML writer saves the document metadata inside an XML comment, so it can be read back in by the XML reader. This flag causes the metadata not to be written.")],
                          output_heading = "Options for XML output")

declareDocumentIO("xml-inline", XMLDocumentIO, True, True)

#
# And now, the "fake" version, for when somebody just randomly inserted SGML-ish tags.
#

# ONLY a reader.

class FakeXMLIO(DocumentFileIO):

    def writeToUnicodeString(self, annotDoc):
        raise NotImplementedError

    # In fake xml IO, it's just SGML-ish tags.
    
    TAG_LOCATOR = re.compile("\<([^>]+)\>", re.S)
    ATTR_NAME_LOCATOR = re.compile("^\s*([^\s=]+)\s*=\s*")
    # The escape for the double quote isn't strictly necessary,
    # but it works better with the Python Emacs mode.
    ATTR_VALUE_LOCATOR = re.compile("""^([\"'])(.*?)\\1\s*""", re.S)

    # We can't quite do this the way we did the fake XML IO reader, because
    # we have (a) null annotations and (b) nested annotations. So we need
    # to look for the tags, and keep gathering input.

    # Because this is pseudo-XML, there may be <> in the signal. So if
    # we find a tag which we can't parse, assume it's signal.

    def _parseTag(self, startTag):
        labToks = startTag.split(None, 1)
        attrs = {}
        if len(labToks) > 1:
            # There are attributes. Grab a token (no whitespace, no =), then
            # look for an =, then find the following value.
            attrStr = labToks[1]
            while attrStr:
                m = self.ATTR_NAME_LOCATOR.match(attrStr)
                if m is None:
                    # Attribute string is ill-formed.
                    return None, None
                attr = m.group(1)
                attrStr = attrStr[m.end():]
                # Now, we need a value.
                m = self.ATTR_VALUE_LOCATOR.match(attrStr)
                if m is None:
                    # Attribute string is ill-formed.
                    return None, None
                val = m.group(2)
                # Since this sucks up trailing whitespace, it should
                # terminate the string when we reach the end of it.
                attrStr = attrStr[m.end():]
                attrs[attr] = val
        return labToks[0], attrs

    # Toplevel.
    def deserialize(self, s, annotDoc):

        annotations = []
        tagStack = []

        signalStrings = []
        startI = 0
        contentLen = 0

        NULL, OPEN, CLOSE = 0, 1, 2

        while True:
            m = self.TAG_LOCATOR.search(s, startI)

            if m is None:
                signalStrings.append(s[startI:])
                break

            tagContent = m.group(1)

            # If it starts with a /, it's an end tag.
            # if it ends with a /, it's a null tag.
            if tagContent[-1] == "/":
                tagLiteral = tagContent[:-1]
                tagStatus = NULL
            elif  tagContent[0] == "/":
                tagLiteral = tagContent[1:]
                tagStatus = CLOSE
            else:
                tagLiteral = tagContent
                tagStatus = OPEN

            if tagStatus in (NULL, OPEN):
                label, attrs = self._parseTag(tagLiteral)
                if label is None:
                    # Probably actually signal instead. Skip it. Well, don't
                    # skip the WHOLE thing; the pattern may have captured
                    # a real tag too.
                    print("Found bogus tag match '%s'; treating first character as text and trying again." % s[m.start():m.end()].encode('ascii', 'replace'), file=sys.stderr)
                    signalStrings.append(s[startI:m.start() + 1])
                    contentLen += ((m.start() + 1) - startI)
                    startI = m.start() + 1
                    continue

            # So now, we know it's a non-bogus tag.

            prefix = s[startI:m.start()]
            contentLen += len(prefix)
            signalStrings.append(prefix)
            annotLoc = contentLen
            startI = m.end()

            if tagStatus is NULL:
                annotations.append([label, attrs, annotLoc, annotLoc])
            elif tagStatus is CLOSE:
                # Gotta match what's on the top of the stack.
                tagContent = tagContent[1:]
                if not tagStack:
                    raise IOError("no start tag for end tag")
                elif tagStack[-1][0] != tagContent:
                    raise IOError("mismatch start and end tags (%s vs %s)" % (tagStack[-1][0], tagContent))
                else:
                    [label, attrs, startIdx] = tagStack[-1]
                    tagStack[-1:] = []
                    annotations.append([label, attrs, startIdx, annotLoc])
            else:
                label, attrs = self._parseTag(tagContent)
                tagStack.append([label, attrs, annotLoc])

        # OK, now we add the annotations.
        for t, attrs, annotStart, annotEnd in annotations:
            annotDoc.createAnnotation(annotStart, annotEnd, t, attrs)
        annotDoc.signal = "".join(signalStrings)

declareDocumentIO('fake-xml-inline', FakeXMLIO, True, False)
