# Copyright (C) 2007 - 2018 The MITRE Corporation. See the toplevel
# file LICENSE.txt for license terms.

from __future__ import print_function, absolute_import
from .PyCompat import _strtypes, _unicodestring

import os, sys

# Descriptor, like a DTD.
from xml.dom import Node

# Convert the DOM into something that's actually useful, and
# do WFC checking while I'm at it. Sort of like DTDs, only useful.

# Pass in a structure that looks like this:

# {"nType": Node.ELEMENT_NODE,
#  "label": "task",
#  "obligAttrs": ["name"],
#  "optAttrs": ["instantiable", "parent", "class"],
#  "optSingleChildren": [ ... ]
#  "optMultipleChildren": [ ... ]
#  "collectTextChildren": False,
#  "wildcardAttrs": False
# }

# recursively. The resulting object has three important
# fields: text (None unless collectTextChildren is true, in
# which case it's a string), attrs (a hash of attributes,
# with all known but non-occurring attributes set to tNone),
# and children (a hash of children, either another XMLNode
# if single, a list of such if multiple, with all known
# but non-occurring attributes set to None if single,
# an empty list if multiple).

# New feature: importable definitions.

class XMLError(Exception):
    pass

# When importTmpl is not None, and import_from is present, go get the
# contents of the file using the search path and continue processing.

class XMLNode(object):

    def __init__(self, pContext = None, label = None, node = None, initialAttrs = None,
                 initialChildren = None, initialText = None,
                 **kw):
        self.pContext = pContext or _NULL_PARSE_CONTEXT
        self.precedingComments = []
        self.followingComments = []
        # initialize everything.
        
        self.label = label
        self.attrs = {}
        self.children = {}
        self.orderedChildren = []
        self.text = initialText
        self.wildcardAttrs = None

        if initialAttrs is not None:
            self.attrs.update(initialAttrs)
        if initialChildren is not None:
            # Pairs of node, isList
            self.orderedChildren = [p[0] for p in initialChildren]
            for n, isList in initialChildren:
                if isList:
                    try:
                        self.children[n.label].append(n)
                    except KeyError:
                        self.children[n.label] = [n]
                else:
                    self.children[n.label] = n           
        if node:
            self._updateFromNode(label = label, node = node, **kw)

    def _updateFromNode(self, node = None, nType = None, label = None, obligAttrs = None,
                        optAttrs = None, 
                        obligSingleChildren = None, obligMultipleChildren = None,
                        optSingleChildren = None, optMultipleChildren = None,
                        collectTextChildren = False, wildcardAttrs = False,
                        importTmpl = None, previousTmpl = None,
                        availableLater = None, parentAvailableLater = None):

        availableLater = availableLater or parentAvailableLater or {}
        previousTmplOverridePairs = None
        if previousTmpl is not None:
            # The node we're matching against had _xmlnode_earlier or _xmlnode_synonym set.
            # We need to fill things out. This will be available if previousTmpl is set -
            # the template digester guarantees it.
            previousTmpl, previousTmplOverridePairs = previousTmpl
            previousTmpl = availableLater[previousTmpl]

        # Initial error check
        if nType is not None and \
           node.nodeType is not nType:
            self.pContext.raiseError("node %s is not of type %d" % (node.nodeName, nType))
        if label is not None and \
           node.nodeName != label:
            self.pContext.raiseError("node %s is not named %s" % (node.nodeName, label))

        if wildcardAttrs:
            self.wildcardAttrs = {}
        
        if previousTmpl:
            prevObligAttrs = previousTmpl["obligAttrs"]
            prevOptAttrs = previousTmpl["optAttrs"]
            if obligAttrs or prevObligAttrs:
                obligAttrs = list(set(obligAttrs) | set(prevObligAttrs))
            # Anything that's obligatory in either set should be
            # obligatory here.
            if optAttrs or prevOptAttrs:
                optAttrs = list(set(optAttrs) | set(prevOptAttrs))
                if obligAttrs:
                    optAttrs = list(set(optAttrs) - set(obligAttrs))

        for attrList in [obligAttrs, optAttrs]:
            if attrList is not None:
                for attr in attrList:
                    self.attrs[attr] = None

        if previousTmpl and previousTmplOverridePairs:
            # Don't update these yet. We have to wait until
            # the attributes are processed. Make sure the keys
            # can't be used, though.
            prevAttrs = set(previousTmplOverridePairs.keys())
            if optAttrs:
                optAttrs = list(set(optAttrs) - set(prevAttrs))
            if obligAttrs:
                obligAttrs = list(set(obligAttrs) - set(prevAttrs))
                    
        importedNode = self._processAttributes(node, importTmpl, 
                                               wildcardAttrs, obligAttrs, optAttrs)

        # Now, if we have an imported node, we want to use it from here on down.
        if importTmpl:
            if importedNode:
                node = importedNode
            obligSingleChildren = importTmpl["obligSingleChildren"]
            obligMultipleChildren = importTmpl["obligMultipleChildren"]
            optSingleChildren = importTmpl["optSingleChildren"]
            optMultipleChildren = importTmpl["optMultipleChildren"]
            collectTextChildren = importTmpl["collectTextChildren"]
        elif previousTmpl:
            if previousTmplOverridePairs:
                self.attrs.update(previousTmplOverridePairs)
            # Remember, we won't have both. And I've guaranteed that the target
            # is real, because the values for _xmlnode_previous can't be themselves
            # pointers.
            obligSingleChildren = previousTmpl["obligSingleChildren"]
            obligMultipleChildren = previousTmpl["obligMultipleChildren"]
            optSingleChildren = previousTmpl["optSingleChildren"]
            optMultipleChildren = previousTmpl["optMultipleChildren"]
            collectTextChildren = previousTmpl["collectTextChildren"]

        for childList, isUnique in [(obligSingleChildren, True), (optSingleChildren, True),
                                    (obligMultipleChildren, False), (optMultipleChildren, False)]:
            if childList is not None:
                for child in childList:
                    if isUnique:
                        self.children[child["label"]] = None
                    else:
                        self.children[child["label"]] = []
            
        textChildren = []
        childNodes = {}
        
        for childNode in node.childNodes:
            if childNode.nodeType is Node.COMMENT_NODE:
                continue
            if childNode.nodeType in [Node.TEXT_NODE, Node.CDATA_SECTION_NODE]:
                if collectTextChildren:
                    v = childNode.nodeValue
                    textChildren.append(v)
                continue
            
            if childNode.nodeName in childNodes:
                childNodes[childNode.nodeName].append(childNode)
            else:
                childNodes[childNode.nodeName] = [childNode]        
        if collectTextChildren:
            self.text = "".join(textChildren)
        childNodeMapping = {}
        self._doChildren(node, obligSingleChildren, childNodes, True, True, childNodeMapping, availableLater)
        self._doChildren(node, obligMultipleChildren, childNodes, True, False, childNodeMapping, availableLater)
        self._doChildren(node, optSingleChildren, childNodes, False, True, childNodeMapping, availableLater)
        self._doChildren(node, optMultipleChildren, childNodes, False, False, childNodeMapping, availableLater)
        if childNodes:
            self.pContext.raiseError("node %s has extra child nodes: %s" % (node.nodeName, str(childNodes)))
        # Updated the ordered children. Insert comments
        # before the node.
        comments = []
        lastOrderedChild = None
        for childNode in node.childNodes:
            # I collect the comment nodes in case I have to regenerate this.
            if childNode.nodeType is Node.COMMENT_NODE:
                comments.append(childNode.data)
            else:
                try:
                    lastOrderedChild = childNodeMapping[childNode]
                    lastOrderedChild.precedingComments = comments
                    comments = []
                    self.orderedChildren.append(lastOrderedChild)
                except KeyError:
                    pass
        if comments and lastOrderedChild:
            lastOrderedChild.followingComments = comments

    def _processAttrSpecs(self, node, attrKeys, obligAttrs, optAttrs):
        if obligAttrs is not None:
            for attr in obligAttrs:
                if attr not in attrKeys:
                    self.pContext.raiseError("node %s does not have attribute %s" % (node.nodeName, attr))
                # Remove the attr from the attrKeys.
                attrKeys.remove(attr)
        if optAttrs is not None:
            for attr in optAttrs:
                if attr in attrKeys:
                    attrKeys.remove(attr)

    def _prepareAttributes(self, node, importTmpl):
        
        attributes = node.attributes                     
        foundAttrDict = self.attrs
        importedNode = None
        
        if attributes is not None:
            for i in range(attributes.length):
                attr = attributes.item(i)
                attrName = attr.name
                attrVal = attr.value
                # Interlude: if there's an importTmpl and import_from is present,
                if (importTmpl is not None) and (attrName == "import_from"):
                    importedNode = self._importNode(attrVal, importTmpl)
                    if node.childNodes:
                        # It can't have any child nodes.
                        self.pContext.raiseError("node %s with import_from has child nodes" % node.nodeName)
                else:
                    foundAttrDict[attrName] = attrVal
                    
        return [key for key, val in foundAttrDict.items() if val is not None], importedNode

    def _importNode(self, fName, importTmpl):
        if self.pContext.searchPath:
            for p in self.pContext.searchPath:
                fullP = os.path.join(p, fName)
                if os.path.isfile(fullP):
                    return XMLNodeFromFile(fullP, importTmpl, searchPath = self.pContext.searchPath)
        self.pContext.raiseError("can't find subordinate XML file %s in search path" % fName)        

    def _processAttributes(self, node, importTmpl, 
                           wildcardAttrs, obligAttrs, optAttrs):

        attrKeys, importedNode = self._prepareAttributes(node, importTmpl)
        
        self._processAttrSpecs(node, attrKeys, obligAttrs, optAttrs)

        # In the case where there's an importTmpl but no importedNode, we want to
        # ALSO process the importTmpl against this node. Then we check the attribute
        # keys here. Then, if there's both an importTmpl and an importedNode, we want
        # to also check the child node attributes. THEN, we use the child selections for
        # either this template (if there's no importTmpl) or the child.

        if importTmpl and (not importedNode):
            self._processAttrSpecs(node, attrKeys, importTmpl["obligAttrs"], importTmpl["optAttrs"])

        # Check the remainder.
        if attrKeys:
            if wildcardAttrs:
                for key in attrKeys:
                    self.wildcardAttrs[key] = self.attrs[key]
            else:
                self.pContext.raiseError("node %s has extra attributes %s" % (node.nodeName, ",".join(attrKeys)))

        # Now, we may want to ALSO check the child, and then merge the results.
        if importTmpl and importedNode:
            savedAttrs = self.attrs
            self.attrs = {}
            self._processAttributes(importedNode, None, importTmpl["wildcardAttrs"],
                                    importTmpl["obligAttrs"], importTmpl["optAttrs"])
            # I want the parent to take precedence.
            self.attrs.update(savedAttrs)
        return importedNode

    def _doChildren(self, node, childDictList, childNodes, isOblig, isUnique, childNodeMapping, availableLater):
        processedNodes = self.children
        if childDictList is not None:
            for d in childDictList:
                childName = d['label']
                if childName not in childNodes:
                    if isOblig:
                        self.pContext.raiseError("node %s is missing the %s child" % (node.nodeName, childName))
                    elif isUnique:
                        # Optional unique children, not present.
                        processedNodes[childName] = None
                else:
                    nodes = childNodes[childName]
                    if isUnique:
                        if len(nodes) > 1:
                            self.pContext.raiseError("node %s has too many %s children" % (node.nodeName, childName))
                        child = XMLNode(node = nodes[0], pContext = self.pContext, parentAvailableLater = availableLater, **d)
                        processedNodes[childName] = child
                        childNodeMapping[nodes[0]] = child
                    else:
                        childList = []
                        for n in nodes:
                            child = XMLNode(pContext = self.pContext, parentAvailableLater = availableLater,
                                            node = n, **d)
                            childList.append(child)
                            childNodeMapping[n] = child
                        processedNodes[childName] = childList
                    del childNodes[childName]

    def _print(self, indent = 0, fp = sys.stdout, colwidth = 80):
        import xml.sax.saxutils
        prefix = " " * indent
        self._printComments(self.precedingComments or [], prefix, fp)
        # Let's make this pretty.        
        openPrefix = prefix
        # First, format the header line.
        openTag = prefix + "<" + self.label
        openEntity = openTag
        for key, val in self.attrs.items():
            if val is None: continue
            # Encode the XML value.
            entry = " %s='%s'" % (key, xml.sax.saxutils.escape(val))
            if (len(openEntity) + len(entry) > colwidth) and (openEntity != openTag):
                # Wrap, but only if it's not the first element.
                print(openEntity, file=fp)
                openPrefix = " " * len(openTag)
                openEntity = openPrefix + entry
            else:
                openEntity += entry
        if self.text is not None:
            print("%s>%s</%s>" % (openEntity, xml.sax.saxutils.escape(self.text), self.label), file=fp)
        elif self.orderedChildren:
            print(openEntity + ">", file=fp)
            for c in self.orderedChildren:
                c._print(indent = indent + 2, fp = fp, colwidth = colwidth)
            print("%s</%s>" % (prefix, self.label), file=fp)
        else:
            print(openEntity + "/>", file=fp)
        self._printComments(self.followingComments or [], prefix, fp)        

    def _printComments(self, comments, prefix, fp):
        for comment in comments:
            # Split into newlines, strip, pad.
            lines = [line.strip() for line in comment.split("\n")]
            lines[0] = prefix + "<!-- " + lines[0]
            lines = [lines[0]] + [prefix + "     " + l for l in lines[1:]]
            lines[-1] += " -->"
            for line in lines:
                print(line, file=fp)

    def copy(self):
        isList = dict([(k, type(v) is list) for (k, v) in self.children.items()])
        n = XMLNode(pContext = self.pContext,
                    label = self.label,
                    initialText = self.text,
                    initialAttrs = self.attrs,
                    initialChildren = [(c.copy(), isList[c.label]) for c in self.orderedChildren])
        if self.wildcardAttrs is not None:
            n.wildcardAttrs = self.wildcardAttrs.copy()
        return n

import xml
import xml.dom.minidom
from xml.dom import Node

# 8/13/15: New constructs: _xml_later="yes" makes the node available for
# later copying or extending. _xml_earlier="<label>" refers. _xml_synonym="<label>"
# copies children, sets various attributes. All these can recurse, so
# we need to point to the earlier one, not copy it.

def _recursivelyConstructTemplate(tmplDir, domNode, availableLater, entryPoints, parent = None):
    d = {"nType": Node.ELEMENT_NODE,
         "label": domNode.nodeName,
         "obligAttrs": [],
         "optAttrs": [],
         "optSingleChildren": [],
         "optMultipleChildren": [],
         "obligSingleChildren": [],
         "obligMultipleChildren": [],
         "collectTextChildren": False,
         "wildcardAttrs": False,
         "importTmpl": None,
         "previousTmpl": None}
    attributes = domNode.attributes
    if attributes is not None:
        foundCount = False
        for i in range(attributes.length):
            attr = attributes.item(i)
            name = attr.name
            value = attr.value
            if name == "_xmlnode_count":
                if parent:
                    if value == "?":
                        parent["optSingleChildren"].append(d)
                    elif value == "*":
                        parent["optMultipleChildren"].append(d)
                    elif value == "+":
                        parent["obligMultipleChildren"].append(d)
                    else:
                        raise XMLError("don't recognize attr/value %s=%s" % (name, value))
                foundCount = True
            elif name == "_xmlnode_collecttextchildren":
                if value == "yes":
                    d["collectTextChildren"] = True
                else:
                    raise XMLError("don't recognize attr/value %s=%s" % (name, value))
            elif name == "_xmlnode_import":
                # Import the grammar from the specified file. The options are either to
                # have an import attribute which refers to the content, or have it in-line.
                # If import is present, nothing else can be; if import is absent, well,
                # you get the idea.
                if tmplDir is None:
                    raise XMLError("can't import from subordinate template because no template directory was found")
                importD = XMLNodeDescFromFile(os.path.join(tmplDir, value))
                if d["label"] not in importD:
                    raise XMLError("import template doesn't have the same label as its host %s" % d["label"])
                d["importTmpl"] = importD[d["label"]]
            elif name == "_xmlnode_wildcardattrs":
                if value == "yes":
                    d["wildcardAttrs"] = True
                else:
                    raise XMLError("don't recognize attr/value %s=%s" % (name, value))
            elif name == "_xmlnode_later":
                if value == "yes":
                    if d["label"] in availableLater:
                        raise XMLError("node '%s' is declared multiple times with _xmlnode_later" % d["label"])
                    else:
                        availableLater[d["label"]] = d
                else:
                    raise XMLError("don't recognize attr/value %s=%s" % (name, value))
            elif name == "_xmlnode_earlier":
                if value == "yes":
                    if d["label"] in availableLater:
                        if d["previousTmpl"] is not None:
                            raise XMLError("node '%s' cannot bear both _xmlnode_earlier and _xmlnode_synonym" % d["label"])                        
                        elif availableLater[d["label"]]["previousTmpl"]:
                            raise XMLError("node '%s' cannot be an _xmlnode_earlier value because it also points to a previous node" % d["label"])
                        else:
                            d["previousTmpl"] = (d["label"], {})
                    else:
                        raise XMLError("node '%s' is referenced by _xmlnode_earlier before being declared via _xmlnode_later" % d["label"])
                else:
                    raise XMLError("don't recognize attr/value %s=%s" % (name, value))
            elif name == "_xmlnode_synonym":
                if value in availableLater:
                    if d["previousTmpl"] is not None:
                        raise XMLError("node '%s' cannot bear both _xmlnode_earlier and _xmlnode_synonym" % d["label"])
                    elif availableLater[value]["previousTmpl"]:
                        raise XMLError("node '%s' cannot be an _xmlnode_synonym value because it also points to a previous node" % value)
                    else:
                        d["previousTmpl"] = (value, {})
                else:
                    raise XMLError("node '%s' is referenced by _xmlnode_synonym before being declared via _xmlnode_later" % value)
            elif name == "_xmlnode_entrypoint":
                if value == "yes":
                    entryPoints[d["label"]] = d
                else:
                    raise XMLError("don't recognize attr/value %s=%s" % (name, value))
            # New functionality: complex tokens here.
            else:
                toks = set([s.strip() for s in value.split(",")])
                # Must have one of these.
                if "obligatory" in toks:
                    d["obligAttrs"].append(name)
                elif "optional" in toks:
                    d["optAttrs"].append(name)
                elif d["previousTmpl"]:
                    d["previousTmpl"][1][name] = value
                else:
                    raise XMLError("don't recognize attr/value %s=%s" % (name, value))
        if parent and (not foundCount):
            parent["obligSingleChildren"].append(d)
    for child in domNode.childNodes:
        if child.nodeType == Node.ELEMENT_NODE:
            _recursivelyConstructTemplate(tmplDir, child, availableLater, entryPoints, d)
    if (d["importTmpl"] is not None) or (d["previousTmpl"] is not None):
        if (d["importTmpl"] is not None) and (d["previousTmpl"] is not None):
            raise XMLError("_xmlnode_import cannot cooccur with _xmlnode_synonym or _xmlnode_earlier")
        if (d["importTmpl"] is not None) and (("import_from" in d["obligAttrs"]) or ("import_from" in d["optAttrs"])):
            raise XMLError("import_from is a reserved attribute for elements with _xmlnode_import")
        if d["optSingleChildren"] or d["optMultipleChildren"] or d["obligSingleChildren"] or d["obligMultipleChildren"]:
            raise XMLError("elements with _xmlnode_import, _xmlnode_earlier, _xmlnode_synonym are not permitted to have local children")
        if d["collectTextChildren"]:
            raise XMLError("can't collect text children for elements with _xmlnode_import")
        if d["wildcardAttrs"]:
            raise XMLError("can't have wildcard attrs for elements with _xmlnode_import")
    return d

#
# General XML node manipulation.
#

def _moveAttribute(attr, oldNode, newNode):
    if oldNode.attrs[attr]:
        newNode.attrs[attr] = oldNode.attrs[attr]
        oldNode.attrs[attr] = None
        return True
    else:
        return False

def _moveChild(childName, oldParent, newParent, before = True):
    if oldParent.children[childName]:
        child = oldParent.children[childName]
        oldParent.children[childName] = None
        # And, of course, remove it from the orderedChildren.
        if type(child) in (tuple, list):
            idx = oldParent.orderedChildren.index(child[0])
            oldParent.orderedChildren = [ch for ch in oldParent.orderedChildren
                                         if ch not in child]
            for c in child:
                _insert(newParent, None, c, as_list = True, before = before)
        else:
            idx = oldParent.orderedChildren.index(child)
            oldParent.orderedChildren.remove(child)
            _insert(newParent, None, child, before = before)
        return idx
    else:
        return -1

def _removeNode(oldNodeParent, oldNode):
    oldNodeParent.orderedChildren.remove(oldNode)
    if oldNodeParent.children[oldNode.label] == oldNode:
        oldNodeParent.children[oldNode.label] = None
    else:
        oldNodeParent.children[oldNode.label].remove(oldNode)

def _replaceNode(oldNodeParent, oldNode, newNode, inherit_comments = False):
    pos = oldNodeParent.orderedChildren.index(oldNode)
    if newNode:
        oldNodeParent.orderedChildren[pos] = newNode
        oldNodeParent.children[newNode.label] = newNode
    oldNodeChildren = oldNodeParent.children[oldNode.label]
    if oldNodeChildren == oldNode:
        # It's a singleton.
        del oldNodeParent.children[oldNode.label]
    else:
        oldNodeChildren.remove(oldNode)
    if inherit_comments and newNode:
        newNode.precedingComments = oldNode.precedingComments
        oldNode.precedingComments = []
        newNode.followingComments = oldNode.followingComments
        oldNode.followingComments = []

def _insert(parentNode, refChild, newChild, as_list = False, before = True):
    if refChild is None:
        if before:
            parentNode.orderedChildren.append(newChild)
        else:            
            parentNode.orderedChildren[0:0] = [newChild]
    else:
        pos = parentNode.orderedChildren.index(refChild)
        if not before:
            pos += 1
        parentNode.orderedChildren[pos:pos] = [newChild]
    if as_list:
        if newChild.label in parentNode.children:
            if refChild and (refChild in parentNode.children[newChild.label]):
                pos = parentNode.children[newChild.label].index(refChild)
                if not before:
                    pos += 1
                parentNode.children[newChild.label][pos:pos] = [newChild]
            elif before:
                parentNode.children[newChild.label].append(newChild)
            else:
                parentNode.children[newChild.label][0:0] = [newChild]
        else:
            parentNode.children[newChild.label] = [newChild]
    else:
        parentNode.children[newChild.label] = newChild

# fName can be a file or a file name.

def XMLNodeDescFromFile(fName):
    if type(fName) in _strtypes:
        tmplDir = os.path.dirname(os.path.abspath(fName))
    elif hasattr(fName, "name"):
        tmplDir = os.path.dirname(os.path.abspath(fName.name))
    else:
        tmplDir = None
    dom = xml.dom.minidom.parse(fName)
    d = None
    availableLater = {}
    entryPoints = {}
    # Grab the first element child.
    for n in dom.childNodes:
        if n.nodeType == Node.ELEMENT_NODE:
            d = _recursivelyConstructTemplate(tmplDir, n, availableLater, entryPoints)
            break
    dom.unlink()
    if d["importTmpl"] is not None:
        raise XMLError("_xmlnode_import is not permitted on the toplevel node in a template file")
    d["availableLater"] = availableLater
    for v in entryPoints.values():
        v["availableLater"] = availableLater
    entryPoints[d["label"]] = d
    return entryPoints

# Reverse engineering for testing.

def _reformatDesc(d, indent = 0, childCount = None):
    sList = ["<" + d["label"]]
    importTmpl = d["importTmpl"]
    for k in set(d["obligAttrs"] + ((importTmpl and importTmpl["obligAttrs"]) or [])):
        sList.append(" %s='obligatory'" % k)
    for k in set(d["optAttrs"] + ((importTmpl and importTmpl["optAttrs"]) or [])):
        sList.append(" %s='optional'" % k)
    if childCount is not None:
        sList.append(" _xmlnode_count='%s'" % childCount)
    if d["collectTextChildren"] or (importTmpl and importTmpl["collectTextChildren"]):
        sList.append(" _xmlnode_collecttextchildren='yes'")
    if d["wildcardAttrs"] or (importTmpl and importTmpl["wildcardAttrs"]):
        sList.append(" _xmlnode_wildcardattrs='yes'")
    children = []
    for k, ct in [("obligSingleChildren", None), ("optSingleChildren", "?"),
                  ("obligMultipleChildren", "+"), ("optMultipleChildren", "*")]:
        for subD in d[k] + ((importTmpl and importTmpl[k]) or []):
            children.append(_reformatDesc(subD, indent + 2, ct))
    if not children:
        sList.append("/>")
        return (" " * indent) + "".join(sList)
    else:
        sList.append(">")
        return "\n".join([(" " * indent) + "".join(sList)] + children + [(" " * indent) + "</" + d["label"] + ">"])

class XMLParseContext(object):

    def __init__(self, searchPath = None, xmlString = None, xmlFile = None):
        self.searchPath = searchPath
        self.xmlString = xmlString
        self.xmlFile = xmlFile

    def raiseError(self, msg):
        if self.xmlString:
            msg = ("In XML string '%s': " % self.xmlString) + msg
        elif self.xmlFile:
            msg = ("In XML file %s: " % self.xmlFile) + msg
        raise XMLError(msg)

_NULL_PARSE_CONTEXT = XMLParseContext()

# fName can be a file or a file name. Sometimes the file is a
# UTF-8 file, and I want the values as Unicode.

def XMLNodeFromString(s, descTable, searchPath = None):
    if type(s) is _unicodestring:
        # s really has to be a byte sequence in UTF-8.
        s = s.encode("utf8")
    return _XMLNodeFromDom(xml.dom.minidom.parseString(s), descTable,
                           XMLParseContext(searchPath = searchPath,
                                           xmlString = s))

def XMLNodeFromFile(fName, descTable, searchPath = None):
    if type(fName) in _strtypes:
        xmlFile = os.path.abspath(fName)
        pList = [os.path.dirname(xmlFile)]        
    elif hasattr(fName, "name"):
        xmlFile = os.path.abspath(fName.name)
        pList = [os.path.dirname(xmlFile)]
    else:
        pList = xmlFile = None
    if searchPath is not None:
        if pList is None:
            pList = searchPath
        else:
            pList += searchPath
    return _XMLNodeFromDom(xml.dom.minidom.parse(fName), descTable,
                           XMLParseContext(searchPath = pList,
                                           xmlFile = xmlFile))

def _XMLNodeFromDom(dom, descTable, pContext = None):
    # dom is a document node. There should be
    # exactly one child which is an element.
    # There may be comments, for instance.
    node = None
    precedingComments = []
    for n in dom.childNodes:
        # Take the first element child.
        if n.nodeType == Node.COMMENT_NODE:
            if node is None:
                precedingComments.append(n.data)
            else:
                node.followingComments.append(n.data)
        if n.nodeType == Node.ELEMENT_NODE:
            if n.nodeName in descTable:
                node = XMLNode(pContext = pContext or _NULL_PARSE_CONTEXT,
                               node = n, **descTable[n.nodeName])
                node.precedingComments = precedingComments
                node.followingComments = []
            break
    dom.unlink()
    return node
