####################################################################################################
### -----------------------------------------  NOTICE  ----------------------------------------- ###
###                                                                                              ###
### This software (or technical data) was produced for the U.S. Government under contract        ###
### 2015-14120200002-002, and is subject to the Rights in Data-General Clause 52.227-14 , ALT    ###
### IV (MAY 2014) or (DEC 2007).                                                                 ###
###                                                                                              ###
### © 2019 The MITRE Corporation. All Rights Reserved.                                           ###
###                                                                                              ###
### -------------------------------------------------------------------------------------------- ###
####################################################################################################

"""
Test functions for the ``score.py`` CLI. To use, call `pyest` from the root of the project.

Every function must start with ``test_``, and every class must start with ``Test``. For
convenience, testing classes will test the correspinding class in the BP library.

Examples:

    def test_ans():
        assert 3 == 3

    class TestClass(object):

        def test_one(self):
            x = "this"
            assert "h" in x

        def test_two(self):
            x = "foo bar"
            with pytest.raises(AssertionError):
                assert hasattr(x, "check")
"""

__author__  = "Leland E. Vakarian <lsperl@mitre.org>"
__date__    = "30 July 2020"

import pytest

from lib import bp
import score


class TestScoreAbstract(object):


    ## correct scoring output


    def test_identical(self):
        """An annotated reference file and an identical system file should produce a score of 1.0.
        """

        ref_doc = bp.BPDocument.from_json(
            "tests/fixtures/tester.abstract_scoring.annotated_entry.bp.json"
        )
        sys_doc = bp.BPDocument.from_json(
            "tests/fixtures/tester.abstract_scoring.annotated_entry.bp.json"
        )

        scores, entry_pairs = score.score_abstract(sys_doc, ref_doc)

        assert scores.combined_score == 1.0, "Scoring results don't match from identical files!"


    def test_empty_entry_both(self):
        """A reference file with an unannotated entry and a system file with an unannotated entry
        should produce a score of 1.0, since the system file is an exact reflection of the contents
        of the reference file.
        """

        ref_doc = bp.BPDocument.from_json(
            "tests/fixtures/tester.abstract_scoring.unannotated_entry.bp.json"
        )
        sys_doc = bp.BPDocument.from_json(
            "tests/fixtures/tester.abstract_scoring.unannotated_entry.bp.json"
        )

        scores, entry_pairs = score.score_abstract(sys_doc, ref_doc)

        assert scores.combined_score == 1.0


    def test_empty_entry_ref(self):
        """A reference file with an unannotated entry and an annotated system file should produce a
        score of 0.0, since there are only false positives in the system doc.
        """

        ref_doc = bp.BPDocument.from_json(
            "tests/fixtures/tester.abstract_scoring.unannotated_entry.bp.json"
        )
        sys_doc = bp.BPDocument.from_json(
            "tests/fixtures/tester.abstract_scoring.annotated_entry.bp.json"
        )

        scores, entry_pairs = score.score_abstract(sys_doc, ref_doc)

        assert scores.combined_score == 0.0


    def test_empty_entry_sys(self):
        """An annotated reference file and a system file with an unannotated entry should produce a
        score of 0.0, since there are only false negatives (misses) in the system doc.
        """

        ref_doc = bp.BPDocument.from_json(
            "tests/fixtures/tester.abstract_scoring.annotated_entry.bp.json"
        )
        sys_doc = bp.BPDocument.from_json(
            "tests/fixtures/tester.abstract_scoring.unannotated_entry.bp.json"
        )

        scores, entry_pairs = score.score_abstract(sys_doc, ref_doc)

        assert scores.combined_score == 0.0


    def test_blank_ref_unannotated_sys(self):
        """An empty reference file (no entries) and an unannotated system file should get 100%
        credit since there are no false alarms or misses, and the reference set is empty.
        """

        ref_doc = bp.BPDocument.from_json(
            "tests/fixtures/tester.abstract_scoring.no_entries.bp.json"
        )
        sys_doc = bp.BPDocument.from_json(
            "tests/fixtures/tester.abstract_scoring.unannotated_entry.bp.json"
        )

        scores, entry_pairs = score.score_abstract(sys_doc, ref_doc)

        assert scores.combined_score == 1.0


    def test_blank_ref_annotated_sys(self):
        """An empty reference file (no entries) and an annotated system file should get 0 credit
        since it is only false alarms.
        """
        ref_doc = bp.BPDocument.from_json(
            "tests/fixtures/tester.abstract_scoring.no_entries.bp.json"
        )
        sys_doc = bp.BPDocument.from_json(
            "tests/fixtures/tester.abstract_scoring.annotated_entry.bp.json"
        )

        scores, entry_pairs = score.score_abstract(sys_doc, ref_doc)

        assert scores.combined_score == 0.0


class TestScoreBasic(object):


    def test_identical(self):
        """An annotated reference file and an identical system file should produce a score of 1.0.
        """

        ref_doc = bp.BPDocument.from_json(
            "tests/fixtures/basic.en.test-1.bp.json"
        )
        sys_doc = bp.BPDocument.from_json(
            "tests/fixtures/basic.en.test-1.bp.json"
        )

        scores, entry_pairs = score.score_basic(sys_doc, ref_doc)

        assert scores.combined_score == 1.0, "Scoring results don't match from identical files!"
        assert scores.combined_measures.f1 == 1.0, \
            "Scoring results don't match from identical files!"


    def test_stats(self):

        ref_doc = bp.BPDocument.from_json(
            "tests/fixtures/basic.eng.one-entry.ref.bp.json"
        )
        sys_doc = bp.BPDocument.from_json(
            "tests/fixtures/basic.eng.one-entry.sys-3.bp.json"
        )

        scores, entry_pairs = score.score_basic(sys_doc, ref_doc)

        assert scores.event_measures.miss == 1
        assert scores.event_measures.false_alarm == 1
        assert scores.event_measures.hit == 20
        assert scores.event_measures.match == 19
        assert scores.event_measures.value == 18

        assert scores.argmatch_measures.miss == 2
        assert scores.argmatch_measures.false_alarm == 1
        assert scores.argmatch_measures.hit == 38
        assert scores.argmatch_measures.match == 35
        assert scores.argmatch_measures.value == 34

        assert scores.combined_measures.miss == 3
        assert scores.combined_measures.false_alarm == 2
        assert scores.combined_measures.hit == 58
        assert scores.combined_measures.match == 54
        assert scores.combined_measures.value == 52


    def test_instability_same_order(self):

        ref_doc = bp.BPDocument.from_json(
            "tests/fixtures/instability/single-doc-basic.json"
        )
        sorted_sys_doc = bp.BPDocument.from_json(
            "tests/fixtures/instability/single-document.output.sorted.json", True
        )
        unsorted_sys_doc = bp.BPDocument.from_json(
            "tests/fixtures/instability/single-document.output.unsorted.json", True
        )

        sorted_scores, _ = score.score_basic(sorted_sys_doc, ref_doc, no_validation=True)
        unsorted_scores, _ = score.score_basic(unsorted_sys_doc, ref_doc, no_validation=True)

        assert sorted_scores.combined_score == unsorted_scores.combined_score


class TestScoreGranular(object):

    def test_issue_52_handler(self):
        """
        See https://gitlab.mitre.org/better_program_eval/bp_lib/-/issues/52.

        ```bash
        # should produce 1.0 score if Issue#52 handler is working
        ./score.py -v Granular \
            /mnt/c/src/bp_lib/tests/fixtures/issue-52-tester.outcome-x.bp.json \
            /mnt/c/src/bp_lib/tests/fixtures/issue-52-tester.outcome.bp.json
        ```
        """

        ref_doc = bp.BPDocument.from_json(
                "tests/fixtures/issue-52-tester.outcome.bp.json"
        )
        sys_doc = bp.BPDocument.from_json(
            "tests/fixtures/issue-52-tester.outcome-x.bp.json"
        )

        scores, entry_pairs = score.score_granular(sys_doc, ref_doc)

        assert scores.combined_score == 1.0, \
            "Old outcome slots weren't converted correctly, so score isn't identical files!"