#!/usr/bin/env python3

####################################################################################################
### -----------------------------------------  NOTICE  ----------------------------------------- ###
###                                                                                              ###
### This software (or technical data) was produced for the U.S. Government under contract        ###
### 2015-14120200002-002, and is subject to the Rights in Data-General Clause 52.227-14 , ALT    ###
### IV (MAY 2014) or (DEC 2007).                                                                 ###
###                                                                                              ###
### © 2019 The MITRE Corporation. All Rights Reserved.                                           ###
###                                                                                              ###
### -------------------------------------------------------------------------------------------- ###
####################################################################################################

"""
Validates various JSON document types.
"""

__author__  = "Leland E. Vakarian <lsperl@mitre.org>"
__date__    = "10 July 2020"

import click

from lib import bp
from lib.utils import CONTEXT_SETTINGS
from lib.utils import LOGGER


"""bool: Global flags handled by cli()."""
DEBUG = False
VERBOSE = False


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option('-v', '--verbose', count=True, help="Show verbose output. Pass again for debug.")
@click.option('-f', '--force', is_flag=True, help="Skip asking to overwrite output.")
@click.option(
    '-l',
    '--save-logs',
    'save_logs',
    type=click.Path(),
    help="Capture logging output (at whatever verbosity level) to the specified path."
)
def cli(verbose, force, save_logs):
    """Validates various types of JSON documents.

    \f

    This function serves as the entrypoint for the program when called on the CLI. It also handles
    global flag ``verbose``.

    Args:
        verbose (int): if 1, pretty-print output as the command runs. If 2, print debugging output
            in addition to level 1 output. If 0, prints nothing as the command runs. Defaults to 0.
        force (bool): if True, skip asking for permission to overwrite a file with output. Defaults
            to False.
        save_logs (str): a path to write captured logging output to.
    """
    # do this before anything else
    from lib.utils import capture_logs
    if save_logs:
        capture_logs(save_logs, verbose, force)

    global DEBUG
    global VERBOSE

    if verbose:
        LOGGER.info("Verbose output is on!")
        if verbose >= 1:
            VERBOSE = True
        if verbose > 1:
            DEBUG = True

    VERBOSE = verbose


@cli.command('bp-json', context_settings=CONTEXT_SETTINGS)
@click.argument('path', type=click.Path(exists=True))
def bp_json_cli(path):
    """Validates a BP JSON document against the current schema.

    \f

    Args:
        path (str): a path to a BP JSON document to validate.
    """
    global LOGGER
    global VERBOSE

    from lib.bp import BPDocument
    from lib.utils import open_json_blob

    doc = open_json_blob(path)

    if VERBOSE:
        LOGGER.info("Starting validation...")

    if BPDocument.dict_is_schema_valid(doc, VERBOSE):
        if VERBOSE:
            LOGGER.info("Document passes schema validation.")

    doc = BPDocument.from_dict(doc)

    if VERBOSE:
        LOGGER.info("<bp.BPDocument> object successfully constructed from document.")
    else:
        click.echo("OK")


@cli.command('dist-json', context_settings=CONTEXT_SETTINGS)
@click.argument('path', type=click.Path(exists=True))
def dist_json_cli(path):
    """Validates a ``dist.json`` document against the current schema.

    \f

    Args:
        path (str): a path to a ``dist.json`` document to validate.
    """
    global VERBOSE

    from lib.utils import open_json_blob
    from lib.utils import validate_dist

    doc = open_json_blob(path)

    if validate_dist(doc, VERBOSE):
        if VERBOSE:
            LOGGER.info("Document passes schema validation.")
        else:
            click.echo("OK")


@cli.command('tasks-json', context_settings=CONTEXT_SETTINGS)
@click.argument('path', type=click.Path(exists=True))
def tasks_json_cli(path):
    """Validates a ``tasks.json`` document against the current schema.

    \f

    Args:
        path (str): a path to a ``tasks.json`` document to validate.
    """
    global VERBOSE

    from lib.utils import open_json_blob
    from lib.utils import validate_tasks

    doc = open_json_blob(path)

    if validate_tasks(doc, VERBOSE):
        if VERBOSE:
            LOGGER.info("Document passes schema validation.")
        else:
            click.echo("OK")


if __name__ == '__main__':

    ## program proper

    cli()
