import os
import json
import time
import logging
import sys
from argparse import ArgumentParser
import random
import numpy as np
import tqdm
import torch
from torch.utils.data import DataLoader
from transformers import (XLMRobertaTokenizer, AdamW,
                          get_linear_schedule_with_warmup)
from model import *
from graph import Graph
from iterators import *
from util import *
from seqeval.metrics import f1_score
from datetime import datetime
from shutil import copyfile
from constants import *

command = 'python ' + ' '.join([x for x in sys.argv])

# configuration
parser = ArgumentParser()
# model hyper-parameters
parser.add_argument('--xlmr_model_name', default='xlm-roberta-large', type=str, choices=[
    'xlm-roberta-base', 'xlm-roberta-large'
])
parser.add_argument('--cache_dir', default='resource/xlmr', type=str)
parser.add_argument('--xlmr_dropout', default=0.3, type=float)
parser.add_argument('--global_warmup', default=0, type=int)
parser.add_argument('--linear_dropout', default=0.4, type=float)
parser.add_argument('--linear_bias', default=1, type=int)
parser.add_argument('--linear_activation', default='relu', type=str)
parser.add_argument('--event_hidden_num', default=600, type=int)
parser.add_argument('--mention_hidden_num', default=150, type=int)
parser.add_argument('--role_hidden_num', default=600, type=int)
parser.add_argument('--gcn_dim', default=450, type=int)
parser.add_argument('--feat_dim', default=50, type=int)
# training hyper-parameters
parser.add_argument('--train_file', default='datasets/basic-events/granular.train.{}.jsonl', type=str)
parser.add_argument('--output_file', default='/app/results.json', type=str)  # system prediction file
parser.add_argument('--dev_file', default='datasets/basic-events/granular.dev.{}.jsonl', type=str)
parser.add_argument('--test_file', default='datasets/basic-events/granular.test.{}.jsonl', type=str)
parser.add_argument('--chinese_file', default='datasets/basic-events/chinese.granular.test.{}.jsonl', type=str)
parser.add_argument('--korean_file', default='datasets/basic-events/korean.granular.test.{}.jsonl', type=str)
parser.add_argument('--russian_file', default='datasets/basic-events/russian.granular.test.{}.jsonl', type=str)
parser.add_argument('--log', default='granular-training-log.txt', type=str)
parser.add_argument('--p1_data_weight', default=1.0, type=float)
parser.add_argument('--accumulate_step', default=2, type=int)
parser.add_argument('--batch_size', default=16, type=int)
parser.add_argument('--eval_batch_size', default=2, type=int)
parser.add_argument('--max_epoch', default=100, type=int)
parser.add_argument('--learning_rate', default=1e-3, type=float)
parser.add_argument('--xlmr_learning_rate', default=2e-5, type=float)
parser.add_argument('--weight_decay', default=1e-3, type=float)
parser.add_argument('--xlmr_weight_decay', default=2e-5, type=float)
parser.add_argument('--grad_clipping', default=5.0, type=float)
parser.add_argument('--prune', default=1, type=int)
parser.add_argument('--model_name', default='trg', type=str)
# data processing
parser.add_argument('--max_sent_length', default=200, type=int)
parser.add_argument('--lang', default='english', type=str)
# function mode
parser.add_argument('--eval_trg_epoch', default=79, type=int)
parser.add_argument('--eval_arg_epoch', default=13, type=int)
parser.add_argument('--seed', default=3456, type=int)
parser.add_argument('--debug', action='store_true')
parser.add_argument('--docker_run', action='store_true')
parser.add_argument('--use_pretrained', action='store_true')
parser.add_argument('--reffile', default='', type=str)
parser.add_argument('--trigger_epoch', default='', type=str)
parser.add_argument('--argument_epoch', default='', type=str)
parser.add_argument('--template_epoch', default='', type=str)
parser.add_argument('--slot_epoch', default='', type=str)
parser.add_argument('--coref_epoch', default='', type=str)
config = parser.parse_args()

# set random seed
# Seed the behavior of the environment variable
os.environ['PYTHONHASHSEED'] = str(config.seed)
random.seed(config.seed)
np.random.seed(config.seed)
torch.manual_seed(config.seed)
torch.cuda.manual_seed(config.seed)
torch.cuda.manual_seed_all(config.seed)
torch.backends.cudnn.deterministic = True
torch.backends.cudnn.benchmark = False

log_dir = os.path.join(WORKING_DIR, 'logs/{}'.format(config.model_name))
config.cache_dir = os.path.join(WORKING_DIR, config.cache_dir)

config.train_file = config.train_file.format(config.model_name)
config.dev_file = config.dev_file.format(config.model_name)
config.test_file = config.test_file.format(config.model_name)
config.chinese_file = config.chinese_file.format(config.model_name)
config.korean_file = config.korean_file.format(config.model_name)
config.russian_file = config.russian_file.format(config.model_name)

config.train_file = os.path.join(WORKING_DIR, config.train_file)
config.dev_file = os.path.join(WORKING_DIR, config.dev_file)
config.test_file = os.path.join(WORKING_DIR, config.test_file)
config.chinese_file = os.path.join(WORKING_DIR, config.chinese_file)
config.korean_file = os.path.join(WORKING_DIR, config.korean_file)
config.russian_file = os.path.join(WORKING_DIR, config.russian_file)

# set GPU device
torch.cuda.set_device(0)

# logging
ensure_dir('logs/')
ensure_dir(log_dir)

for name in logging.root.manager.loggerDict:
    if 'transformers' in name:
        logging.getLogger(name).setLevel(logging.CRITICAL)

logging.basicConfig(format='%(message)s', level=logging.INFO,
                    filename=os.path.join(log_dir, config.log),
                    filemode='w')
logger = logging.getLogger(__name__)

trigger_model_save_fpath = os.path.join(log_dir, 'trigger-model.mdl')
argument_model_save_fpath = os.path.join(log_dir, 'argument-model.mdl')
template_model_save_fpath = os.path.join(log_dir, 'template-model.mdl')
slot_model_save_fpath = os.path.join(log_dir, 'slot-model.mdl')
coref_model_save_fpath = os.path.join(log_dir, 'coref-model.mdl')

# datasets
tokenizer = XLMRobertaTokenizer.from_pretrained(config.xlmr_model_name,
                                                cache_dir=config.cache_dir,
                                                do_lower_case=False)

logger.info('Command that produces this log: {}'.format(command))

if config.model_name.startswith('trg'):
    train_set = TriggerDataset(config.train_file, config)
    dev_set = TriggerDataset(config.dev_file, config)
    test_set = TriggerDataset(config.test_file, config)
    chinese_set = TriggerDataset(config.chinese_file, config)
    korean_set = TriggerDataset(config.korean_file, config)
    russian_set = TriggerDataset(config.russian_file, config)

    train_set.numberize(tokenizer, vocabs)
    dev_set.numberize(tokenizer, vocabs)
    test_set.numberize(tokenizer, vocabs)
    chinese_set.numberize(tokenizer, vocabs)
    korean_set.numberize(tokenizer, vocabs)
    russian_set.numberize(tokenizer, vocabs)

    batch_num = len(train_set) // config.batch_size
    dev_batch_num = len(dev_set) // config.eval_batch_size + \
                    (len(dev_set) % config.eval_batch_size != 0)
    test_batch_num = len(test_set) // config.eval_batch_size + \
                     (len(test_set) % config.eval_batch_size != 0)
    chinese_batch_num = len(chinese_set) // config.eval_batch_size + \
                        (len(chinese_set) % config.eval_batch_size != 0)
    korean_batch_num = len(korean_set) // config.eval_batch_size + \
                       (len(korean_set) % config.eval_batch_size != 0)
    russian_batch_num = len(russian_set) // config.eval_batch_size + \
                        (len(russian_set) % config.eval_batch_size != 0)

    # initialize the model
    model = TriggerModel(config, vocabs)
    model.cuda()

    init_trigger_model_save_fpath = os.path.join('logs/trg/trigger-model.mdl') + '.lang-chinese'
    if config.docker_run and os.path.exists(init_trigger_model_save_fpath):  # trick to reduce run 2's retraining time
        init_trigger_ckpt = torch.load(init_trigger_model_save_fpath)
        model.load_state_dict(init_trigger_ckpt['model'])
        print_and_log('Initialized Trigger model with checkpoint at {}'.format(init_trigger_model_save_fpath), logger)
        del init_trigger_ckpt
        torch.cuda.empty_cache()

    # print trainable weights
    n_trainable_params, n_nontrainable_params = 0, 0
    for p in model.parameters():
        n_params = torch.prod(torch.tensor(p.shape)).item()
        if p.requires_grad:
            n_trainable_params += n_params
        else:
            n_nontrainable_params += n_params
    logger.info('-' * 100)
    logger.info('> trainable params:')
    for name, param in model.named_parameters():
        if param.requires_grad:
            logger.info('>>> {0}: {1}'.format(name, param.shape))
    logger.info(
        'n_trainable_params: {0}, n_nontrainable_params: {1}'.format(n_trainable_params, n_nontrainable_params))
    logger.info('-' * 100)

    # optimizer
    param_groups = [
        {
            'params': [p for n, p in model.named_parameters() if n.startswith('xlmr')],
            'lr': config.xlmr_learning_rate, 'weight_decay': config.xlmr_weight_decay
        },
        {
            'params': [p for n, p in model.named_parameters() if not n.startswith('xlmr')
                       and 'crf' not in n],
            'lr': config.learning_rate, 'weight_decay': config.weight_decay
        },
        {
            'params': [p for n, p in model.named_parameters() if not n.startswith('xlmr')
                       and ('crf' in n)],
            'lr': config.learning_rate, 'weight_decay': 0
        }
    ]
    optimizer = AdamW(params=param_groups)
    schedule = get_linear_schedule_with_warmup(optimizer,
                                               num_warmup_steps=batch_num * 5,
                                               num_training_steps=batch_num * 100)

    # model state
    state = dict(model=model.state_dict(),
                 event_hidden_num=config.event_hidden_num,
                 role_hidden_num=config.role_hidden_num)

    global_step = 0

    best_english_dev_chinese = {
        'epoch': 0,
        'event': {'f1': 0},
        'argument': {'f1': 0},
        'combined': 0
    }
    best_english_test_chinese = {
        'epoch': 0,
        'event': {'f1': 0},
        'argument': {'f1': 0},
        'combined': 0
    }
    best_english_dev_korean = {
        'epoch': 0,
        'event': {'f1': 0},
        'argument': {'f1': 0},
        'combined': 0
    }
    best_english_test_korean = {
        'epoch': 0,
        'event': {'f1': 0},
        'argument': {'f1': 0},
        'combined': 0
    }
    best_english_dev_russian = {
        'epoch': 0,
        'event': {'f1': 0},
        'argument': {'f1': 0},
        'combined': 0
    }
    best_english_test_russian = {
        'epoch': 0,
        'event': {'f1': 0},
        'argument': {'f1': 0},
        'combined': 0
    }
    best_sample_chinese = {
        'epoch': 0,
        'event': {'f1': 0},
        'argument': {'f1': 0},
        'combined': 0
    }
    best_sample_korean = {
        'epoch': 0,
        'event': {'f1': 0},
        'argument': {'f1': 0},
        'combined': 0
    }
    best_sample_russian = {
        'epoch': 0,
        'event': {'f1': 0},
        'argument': {'f1': 0},
        'combined': 0
    }

    dev_infile = GRANULAR_ANALYSIS_EMPTY_P3
    test_infile = GRANULAR_DEVTEST_EMPTY_P3
    chinese_infile = SAMPLE_GRANULAR_CHINESE_P3
    korean_infile = SAMPLE_GRANULAR_KOREAN_P3
    russian_infile = SAMPLE_GRANULAR_RUSSIAN_P3

    dev_sysfile = os.path.join(log_dir, 'results.trigger.analysis.bp.json')
    test_sysfile = os.path.join(log_dir, 'results.trigger.devtest.bp.json')
    chinese_sysfile = os.path.join(log_dir, 'results.trigger.chinese-sample.bp.json')
    korean_sysfile = os.path.join(log_dir, 'results.trigger.korean-sample.bp.json')
    russian_sysfile = os.path.join(log_dir, 'results.trigger.russian-sample.bp.json')

    dev_reffile = GRANULAR_ANALYSIS_REF_P3
    test_reffile = GRANULAR_DEVTEST_REF_P3
    chinese_reffile = SAMPLE_GRANULAR_CHINESE_P3
    korean_reffile = SAMPLE_GRANULAR_KOREAN_P3
    russian_reffile = SAMPLE_GRANULAR_RUSSIAN_P3

    for epoch in range(config.max_epoch):
        torch.cuda.empty_cache()
        print_and_log('*' * 30, logger)
        print_and_log('Epoch: {}'.format(epoch), logger)
        print_and_log('command: {}'.format(command), logger)
        # training set
        progress = tqdm.tqdm(total=batch_num, ncols=75,
                             desc='Train {}'.format(epoch))
        model.train()
        optimizer.zero_grad()
        losses = []
        for batch_idx, batch in enumerate(DataLoader(
                train_set, batch_size=config.batch_size // config.accumulate_step,
                shuffle=True, collate_fn=train_set.collate_fn)):

            loss = model(batch)
            loss = loss * (1 / config.accumulate_step)

            loss.backward()

            if (batch_idx + 1) % config.accumulate_step == 0:
                progress.update(1)
                global_step += 1
                torch.nn.utils.clip_grad_norm_(
                    model.parameters(), config.grad_clipping)
                optimizer.step()
                schedule.step()
                optimizer.zero_grad()
                print_and_log('{}: step: {}/{}, loss: {}'.format(datetime.now(), batch_idx + 1, batch_num, loss.item()),
                              logger,
                              printout=False)
                losses.append(loss.item())
        progress.close()

        model.eval()
        # dev set
        progress = tqdm.tqdm(total=dev_batch_num, ncols=75,
                             desc='Dev {}'.format(epoch))

        dev_predictions = []
        for batch in DataLoader(dev_set, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=dev_set.collate_fn):
            progress.update(1)
            graphs = model.predict(batch)
            dev_predictions.extend(graphs)
        progress.close()
        write_trigger_predictions(
            dev_predictions,
            output_fpath=dev_sysfile + '.epoch-{}'.format(epoch),
            infile=dev_infile
        )
        dev_score = compute_basic_score(
            reffile=dev_reffile,
            sysfile=dev_sysfile + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'analysis.trigger.report.epoch-{}.csv'.format(epoch))
        )
        # test set
        progress = tqdm.tqdm(total=test_batch_num, ncols=75,
                             desc='Test {}'.format(epoch))

        test_predictions = []
        for batch in DataLoader(test_set, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=test_set.collate_fn):
            progress.update(1)
            graphs = model.predict(batch)
            test_predictions.extend(graphs)
        progress.close()

        write_trigger_predictions(
            test_predictions,
            output_fpath=test_sysfile + '.epoch-{}'.format(epoch),
            infile=test_infile
        )
        test_score = compute_basic_score(
            reffile=test_reffile,
            sysfile=test_sysfile + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'devtest.trigger.report.epoch-{}.csv'.format(epoch))
        )

        # Chinese test set
        progress = tqdm.tqdm(total=chinese_batch_num, ncols=75,
                             desc='Chinese {}'.format(epoch))

        chinese_trigger_predictions = []
        for batch in DataLoader(chinese_set, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=chinese_set.collate_fn):
            progress.update(1)
            graphs = model.predict(batch)
            chinese_trigger_predictions.extend(graphs)
        progress.close()

        write_trigger_predictions(
            chinese_trigger_predictions,
            output_fpath=chinese_sysfile + '.epoch-{}'.format(epoch),
            infile=chinese_infile
        )
        chinese_score = compute_basic_score(
            reffile=chinese_reffile,
            sysfile=chinese_sysfile + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'chinese.trigger.report.epoch-{}.csv'.format(epoch))
        )
        # korean test set
        progress = tqdm.tqdm(total=korean_batch_num, ncols=75,
                             desc='Korean {}'.format(epoch))

        korean_trigger_predictions = []
        for batch in DataLoader(korean_set, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=korean_set.collate_fn):
            progress.update(1)
            graphs = model.predict(batch)
            korean_trigger_predictions.extend(graphs)
        progress.close()

        write_trigger_predictions(
            korean_trigger_predictions,
            output_fpath=korean_sysfile + '.epoch-{}'.format(epoch),
            infile=korean_infile
        )
        korean_score = compute_basic_score(
            reffile=korean_reffile,
            sysfile=korean_sysfile + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'korean.trigger.report.epoch-{}.csv'.format(epoch))
        )
        # russian test set
        progress = tqdm.tqdm(total=russian_batch_num, ncols=75,
                             desc='Russian {}'.format(epoch))

        russian_trigger_predictions = []
        for batch in DataLoader(russian_set, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=russian_set.collate_fn):
            progress.update(1)
            graphs = model.predict(batch)
            russian_trigger_predictions.extend(graphs)
        progress.close()

        write_trigger_predictions(
            russian_trigger_predictions,
            output_fpath=russian_sysfile + '.epoch-{}'.format(epoch),
            infile=russian_infile
        )
        russian_score = compute_basic_score(
            reffile=russian_reffile,
            sysfile=russian_sysfile + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'russian.trigger.report.epoch-{}.csv'.format(epoch))
        )
        dev_score['epoch'] = epoch
        test_score['epoch'] = epoch
        chinese_score['epoch'] = epoch
        korean_score['epoch'] = epoch
        russian_score['epoch'] = epoch

        print_and_log('=' * 50, logger)
        print_and_log('Loss: {:.3f}'.format(sum(losses) / len(losses)), logger)
        print_and_log('-' * 20, logger)
        print_and_log('Dev: {}'.format(dev_score), logger, printout=not config.docker_run)
        print_and_log('Test: {}'.format(test_score), logger, printout=not config.docker_run)
        print_and_log('Chinese: {}'.format(chinese_score), logger, printout=not config.docker_run)
        print_and_log('Korean: {}'.format(korean_score), logger, printout=not config.docker_run)
        print_and_log('Russian: {}'.format(russian_score), logger, printout=not config.docker_run)

        if epoch == 0 or (dev_score['event']['f1'] + chinese_score['event']['f1']) > (
                best_english_dev_chinese['event']['f1'] + best_sample_chinese['event']['f1']):
            best_english_dev_chinese = dev_score
            best_english_test_chinese = test_score
            best_sample_chinese = chinese_score
            print_and_log('New best chinese model...', logger)

            torch.save(state, trigger_model_save_fpath + '.lang-chinese')

            # save best prediction files
            copyfile(dev_sysfile + '.epoch-{}'.format(epoch), dev_sysfile + '.lang-chinese.best')
            copyfile(test_sysfile + '.epoch-{}'.format(epoch), test_sysfile + '.lang-chinese.best')
            copyfile(chinese_sysfile + '.epoch-{}'.format(epoch), chinese_sysfile + '.best')

        if epoch == 0 or (dev_score['event']['f1'] + korean_score['event']['f1']) > (
                best_english_dev_korean['event']['f1'] + best_sample_korean['event']['f1']):
            best_english_dev_korean = dev_score
            best_english_test_korean = test_score
            best_sample_korean = korean_score
            print_and_log('New best korean model...', logger)

            torch.save(state, trigger_model_save_fpath + '.lang-korean')

            # save best prediction files
            copyfile(dev_sysfile + '.epoch-{}'.format(epoch), dev_sysfile + '.lang-korean.best')
            copyfile(test_sysfile + '.epoch-{}'.format(epoch), test_sysfile + '.lang-korean.best')
            copyfile(korean_sysfile + '.epoch-{}'.format(epoch), korean_sysfile + '.best')

        if epoch == 0 or (dev_score['event']['f1'] + russian_score['event']['f1']) > (
                best_english_dev_russian['event']['f1'] + best_sample_russian['event']['f1']):
            best_english_dev_russian = dev_score
            best_english_test_russian = test_score
            best_sample_russian = russian_score
            print_and_log('New best russian model...', logger)

            torch.save(state, trigger_model_save_fpath + '.lang-russian')

            # save best prediction files
            copyfile(dev_sysfile + '.epoch-{}'.format(epoch), dev_sysfile + '.lang-russian.best')
            copyfile(test_sysfile + '.epoch-{}'.format(epoch), test_sysfile + '.lang-russian.best')
            copyfile(russian_sysfile + '.epoch-{}'.format(epoch), russian_sysfile + '.best')

        print_and_log('=' * 50, logger, printout=not config.docker_run)
        print_and_log('Current best result:', logger, printout=not config.docker_run)
        print_and_log('-' * 20, logger, printout=not config.docker_run)
        print_and_log('Eng Dev for Chinese: {}'.format(best_english_dev_chinese), logger,
                      printout=not config.docker_run)
        print_and_log('Eng Test for Chinese: {}'.format(best_english_test_chinese), logger,
                      printout=not config.docker_run)
        print_and_log('Sample Chinese: {}'.format(best_sample_chinese), logger, printout=not config.docker_run)
        print_and_log('-' * 20, logger, printout=not config.docker_run)
        print_and_log('Eng Dev for Korean: {}'.format(best_english_dev_korean), logger, printout=not config.docker_run)
        print_and_log('Eng Test for Korean: {}'.format(best_english_test_korean), logger,
                      printout=not config.docker_run)
        print_and_log('Sample Korean: {}'.format(best_sample_korean), logger, printout=not config.docker_run)
        print_and_log('-' * 20, logger, printout=not config.docker_run)
        print_and_log('Eng Dev for Russian: {}'.format(best_english_dev_russian), logger,
                      printout=not config.docker_run)
        print_and_log('Eng Test for Russian: {}'.format(best_english_test_russian), logger,
                      printout=not config.docker_run)
        print_and_log('Sample Russian: {}'.format(best_sample_russian), logger, printout=not config.docker_run)

elif config.model_name.startswith('arg'):
    train_set = ArgumentDataset(config.train_file, config)

    dev_set_chinese = ArgumentDataset(config.dev_file + '.lang-chinese', config, eval=True)
    test_set_chinese = ArgumentDataset(config.test_file + '.lang-chinese', config, eval=True)

    dev_set_korean = ArgumentDataset(config.dev_file + '.lang-korean', config, eval=True)
    test_set_korean = ArgumentDataset(config.test_file + '.lang-korean', config, eval=True)

    dev_set_russian = ArgumentDataset(config.dev_file + '.lang-russian', config, eval=True)
    test_set_russian = ArgumentDataset(config.test_file + '.lang-russian', config, eval=True)

    chinese_set = ArgumentDataset(config.chinese_file, config, eval=True)
    korean_set = ArgumentDataset(config.korean_file, config, eval=True)
    russian_set = ArgumentDataset(config.russian_file, config, eval=True)

    train_set.numberize(tokenizer, vocabs)

    dev_set_chinese.numberize(tokenizer, vocabs)
    test_set_chinese.numberize(tokenizer, vocabs)

    dev_set_korean.numberize(tokenizer, vocabs)
    test_set_korean.numberize(tokenizer, vocabs)

    dev_set_russian.numberize(tokenizer, vocabs)
    test_set_russian.numberize(tokenizer, vocabs)

    chinese_set.numberize(tokenizer, vocabs)
    korean_set.numberize(tokenizer, vocabs)
    russian_set.numberize(tokenizer, vocabs)

    batch_num = len(train_set) // config.batch_size

    dev_batch_num_chinese = len(dev_set_chinese) // config.eval_batch_size + \
                            (len(dev_set_chinese) % config.eval_batch_size != 0)
    test_batch_num_chinese = len(test_set_chinese) // config.eval_batch_size + \
                             (len(test_set_chinese) % config.eval_batch_size != 0)

    dev_batch_num_korean = len(dev_set_korean) // config.eval_batch_size + \
                           (len(dev_set_korean) % config.eval_batch_size != 0)
    test_batch_num_korean = len(test_set_korean) // config.eval_batch_size + \
                            (len(test_set_korean) % config.eval_batch_size != 0)

    dev_batch_num_russian = len(dev_set_russian) // config.eval_batch_size + \
                            (len(dev_set_russian) % config.eval_batch_size != 0)
    test_batch_num_russian = len(test_set_russian) // config.eval_batch_size + \
                             (len(test_set_russian) % config.eval_batch_size != 0)

    chinese_batch_num = len(chinese_set) // config.eval_batch_size + \
                        (len(chinese_set) % config.eval_batch_size != 0)
    korean_batch_num = len(korean_set) // config.eval_batch_size + \
                       (len(korean_set) % config.eval_batch_size != 0)
    russian_batch_num = len(russian_set) // config.eval_batch_size + \
                        (len(russian_set) % config.eval_batch_size != 0)

    # initialize the model
    model = ArgumentModel(config, vocabs)
    model.cuda()

    init_argument_model_save_fpath = os.path.join('logs/arg/argument-model.mdl') + '.lang-chinese'
    if os.path.exists(init_argument_model_save_fpath):
        init_argument_ckpt = torch.load(init_argument_model_save_fpath)
        model.load_state_dict(init_argument_ckpt['model'])
        print_and_log('Initialized Argument model with checkpoint at {}'.format(init_argument_model_save_fpath), logger)
        del init_argument_ckpt
        torch.cuda.empty_cache()

    # print trainable weights
    n_trainable_params, n_nontrainable_params = 0, 0
    for p in model.parameters():
        n_params = torch.prod(torch.tensor(p.shape)).item()
        if p.requires_grad:
            n_trainable_params += n_params
        else:
            n_nontrainable_params += n_params
    logger.info('-' * 100)
    logger.info('> trainable params:')
    for name, param in model.named_parameters():
        if param.requires_grad:
            logger.info('>>> {0}: {1}'.format(name, param.shape))
    logger.info(
        'n_trainable_params: {0}, n_nontrainable_params: {1}'.format(n_trainable_params, n_nontrainable_params))
    logger.info('-' * 100)

    # optimizer
    param_groups = [
        {
            'params': [p for n, p in model.named_parameters() if n.startswith('xlmr')],
            'lr': config.xlmr_learning_rate, 'weight_decay': config.xlmr_weight_decay
        },
        {
            'params': [p for n, p in model.named_parameters() if not n.startswith('xlmr')
                       and 'crf' not in n],
            'lr': config.learning_rate, 'weight_decay': config.weight_decay
        },
        {
            'params': [p for n, p in model.named_parameters() if not n.startswith('xlmr')
                       and ('crf' in n)],
            'lr': config.learning_rate, 'weight_decay': 0
        }
    ]
    optimizer = AdamW(params=param_groups)
    schedule = get_linear_schedule_with_warmup(optimizer,
                                               num_warmup_steps=batch_num * 5,
                                               num_training_steps=batch_num * 100)

    # model state
    state = dict(model=model.state_dict(),
                 event_hidden_num=config.event_hidden_num,
                 role_hidden_num=config.role_hidden_num)

    global_step = 0

    best_english_dev_chinese = {
        'epoch': 0,
        'event': {'f1': 0},
        'argument': {'f1': 0},
        'combined': 0
    }
    best_english_test_chinese = {
        'epoch': 0,
        'event': {'f1': 0},
        'argument': {'f1': 0},
        'combined': 0
    }
    best_sample_chinese = {
        'epoch': 0,
        'event': {'f1': 0},
        'argument': {'f1': 0},
        'combined': 0
    }

    best_english_dev_korean = {
        'epoch': 0,
        'event': {'f1': 0},
        'argument': {'f1': 0},
        'combined': 0
    }
    best_english_test_korean = {
        'epoch': 0,
        'event': {'f1': 0},
        'argument': {'f1': 0},
        'combined': 0
    }
    best_sample_korean = {
        'epoch': 0,
        'event': {'f1': 0},
        'argument': {'f1': 0},
        'combined': 0
    }

    best_english_dev_russian = {
        'epoch': 0,
        'event': {'f1': 0},
        'argument': {'f1': 0},
        'combined': 0
    }
    best_english_test_russian = {
        'epoch': 0,
        'event': {'f1': 0},
        'argument': {'f1': 0},
        'combined': 0
    }
    best_sample_russian = {
        'epoch': 0,
        'event': {'f1': 0},
        'argument': {'f1': 0},
        'combined': 0
    }

    dev_infile_chinese = os.path.join('logs/trg/results.trigger.analysis.bp.json.lang-chinese.best')
    test_infile_chinese = os.path.join('logs/trg/results.trigger.devtest.bp.json.lang-chinese.best')

    dev_infile_korean = os.path.join('logs/trg/results.trigger.analysis.bp.json.lang-korean.best')
    test_infile_korean = os.path.join('logs/trg/results.trigger.devtest.bp.json.lang-korean.best')

    dev_infile_russian = os.path.join('logs/trg/results.trigger.analysis.bp.json.lang-russian.best')
    test_infile_russian = os.path.join('logs/trg/results.trigger.devtest.bp.json.lang-russian.best')

    chinese_infile = os.path.join('logs/trg/results.trigger.chinese-sample.bp.json.best')
    korean_infile = os.path.join('logs/trg/results.trigger.korean-sample.bp.json.best')
    russian_infile = os.path.join('logs/trg/results.trigger.russian-sample.bp.json.best')

    epoch2scores = []

    dev_sysfile_chinese = os.path.join(log_dir,
                                       'results.{}.analysis.bp.json'.format(config.model_name)) + '.lang-chinese'
    test_sysfile_chinese = os.path.join(log_dir,
                                        'results.{}.devtest.bp.json'.format(config.model_name)) + '.lang-chinese'

    dev_sysfile_korean = os.path.join(log_dir, 'results.{}.analysis.bp.json'.format(config.model_name)) + '.lang-korean'
    test_sysfile_korean = os.path.join(log_dir, 'results.{}.devtest.bp.json'.format(config.model_name)) + '.lang-korean'

    dev_sysfile_russian = os.path.join(log_dir,
                                       'results.{}.analysis.bp.json'.format(config.model_name)) + '.lang-russian'
    test_sysfile_russian = os.path.join(log_dir,
                                        'results.{}.devtest.bp.json'.format(config.model_name)) + '.lang-russian'

    chinese_sysfile = os.path.join(log_dir, 'results.{}.chinese-sample.bp.json'.format(config.model_name))
    korean_sysfile = os.path.join(log_dir, 'results.{}.korean-sample.bp.json'.format(config.model_name))
    russian_sysfile = os.path.join(log_dir, 'results.{}.russian-sample.bp.json'.format(config.model_name))

    dev_reffile = GRANULAR_ANALYSIS_REF_P3
    test_reffile = GRANULAR_DEVTEST_REF_P3
    chinese_reffile = SAMPLE_GRANULAR_CHINESE_P3
    korean_reffile = SAMPLE_GRANULAR_KOREAN_P3
    russian_reffile = SAMPLE_GRANULAR_RUSSIAN_P3

    for epoch in range(config.max_epoch):
        torch.cuda.empty_cache()
        print_and_log('*' * 30, logger)
        print_and_log('Epoch: {}'.format(epoch), logger)
        print_and_log('command: {}'.format(command), logger)
        # training set
        progress = tqdm.tqdm(total=batch_num, ncols=75,
                             desc='Train {}'.format(epoch))
        model.train()
        optimizer.zero_grad()
        losses = []
        for batch_idx, batch in enumerate(DataLoader(
                train_set, batch_size=config.batch_size // config.accumulate_step,
                shuffle=True, collate_fn=train_set.collate_fn)):

            loss = model(batch)
            loss = loss * (1 / config.accumulate_step)

            loss.backward()

            if (batch_idx + 1) % config.accumulate_step == 0:
                progress.update(1)
                global_step += 1
                torch.nn.utils.clip_grad_norm_(
                    model.parameters(), config.grad_clipping)
                optimizer.step()
                schedule.step()
                optimizer.zero_grad()
                print_and_log('{}: step: {}/{}, loss: {}'.format(datetime.now(), batch_idx + 1, batch_num, loss.item()),
                              logger,
                              printout=False)
                losses.append(loss.item())
        progress.close()

        model.eval()
        # dev set (Chinese)
        progress = tqdm.tqdm(total=dev_batch_num_chinese, ncols=75,
                             desc='Dev {}'.format(epoch))

        dev_predictions = []
        for batch in DataLoader(dev_set_chinese, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=dev_set_chinese.collate_fn):
            progress.update(1)
            graphs = model.predict(batch)
            dev_predictions.extend(graphs)
        progress.close()
        write_arg_predictions(
            dev_predictions,
            output_fpath=dev_sysfile_chinese + '.epoch-{}'.format(epoch),
            infile=dev_infile_chinese
        )
        remove_pronouns(dev_sysfile_chinese + '.epoch-{}'.format(epoch),
                        dev_sysfile_chinese + '.epoch-{}'.format(epoch))
        normalize_money_spans(dev_sysfile_chinese + '.epoch-{}'.format(epoch),
                              dev_sysfile_chinese + '.epoch-{}'.format(epoch))
        dev_score_chinese = compute_basic_score(
            reffile=dev_reffile,
            sysfile=dev_sysfile_chinese + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'analysis.arg.report.epoch-{}.csv'.format(epoch))
        )
        # test set
        progress = tqdm.tqdm(total=test_batch_num_chinese, ncols=75,
                             desc='Test {}'.format(epoch))

        test_predictions = []
        for batch in DataLoader(test_set_chinese, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=test_set_chinese.collate_fn):
            progress.update(1)
            graphs = model.predict(batch)
            test_predictions.extend(graphs)
        progress.close()

        write_arg_predictions(
            test_predictions,
            output_fpath=test_sysfile_chinese + '.epoch-{}'.format(epoch),
            infile=test_infile_chinese
        )
        remove_pronouns(test_sysfile_chinese + '.epoch-{}'.format(epoch),
                        test_sysfile_chinese + '.epoch-{}'.format(epoch))
        normalize_money_spans(test_sysfile_chinese + '.epoch-{}'.format(epoch),
                              test_sysfile_chinese + '.epoch-{}'.format(epoch))
        test_score_chinese = compute_basic_score(
            reffile=test_reffile,
            sysfile=test_sysfile_chinese + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'devtest.arg.report.epoch-{}.csv'.format(epoch))
        )

        # dev set (Korean)
        progress = tqdm.tqdm(total=dev_batch_num_korean, ncols=75,
                             desc='Dev {}'.format(epoch))

        dev_predictions = []
        for batch in DataLoader(dev_set_korean, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=dev_set_korean.collate_fn):
            progress.update(1)
            graphs = model.predict(batch)
            dev_predictions.extend(graphs)
        progress.close()
        write_arg_predictions(
            dev_predictions,
            output_fpath=dev_sysfile_korean + '.epoch-{}'.format(epoch),
            infile=dev_infile_korean
        )
        remove_pronouns(dev_sysfile_korean + '.epoch-{}'.format(epoch), dev_sysfile_korean + '.epoch-{}'.format(epoch))
        normalize_money_spans(dev_sysfile_korean + '.epoch-{}'.format(epoch),
                              dev_sysfile_korean + '.epoch-{}'.format(epoch))
        dev_score_korean = compute_basic_score(
            reffile=dev_reffile,
            sysfile=dev_sysfile_korean + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'analysis.arg.report.epoch-{}.csv'.format(epoch))
        )
        # test set
        progress = tqdm.tqdm(total=test_batch_num_korean, ncols=75,
                             desc='Test {}'.format(epoch))

        test_predictions = []
        for batch in DataLoader(test_set_korean, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=test_set_korean.collate_fn):
            progress.update(1)
            graphs = model.predict(batch)
            test_predictions.extend(graphs)
        progress.close()

        write_arg_predictions(
            test_predictions,
            output_fpath=test_sysfile_korean + '.epoch-{}'.format(epoch),
            infile=test_infile_korean
        )
        remove_pronouns(test_sysfile_korean + '.epoch-{}'.format(epoch),
                        test_sysfile_korean + '.epoch-{}'.format(epoch))
        normalize_money_spans(test_sysfile_korean + '.epoch-{}'.format(epoch),
                              test_sysfile_korean + '.epoch-{}'.format(epoch))
        test_score_korean = compute_basic_score(
            reffile=test_reffile,
            sysfile=test_sysfile_korean + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'devtest.arg.report.epoch-{}.csv'.format(epoch))
        )

        # dev set
        progress = tqdm.tqdm(total=dev_batch_num_russian, ncols=75,
                             desc='Dev {}'.format(epoch))

        dev_predictions = []
        for batch in DataLoader(dev_set_russian, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=dev_set_russian.collate_fn):
            progress.update(1)
            graphs = model.predict(batch)
            dev_predictions.extend(graphs)
        progress.close()
        write_arg_predictions(
            dev_predictions,
            output_fpath=dev_sysfile_russian + '.epoch-{}'.format(epoch),
            infile=dev_infile_russian
        )
        remove_pronouns(dev_sysfile_russian + '.epoch-{}'.format(epoch),
                        dev_sysfile_russian + '.epoch-{}'.format(epoch))
        normalize_money_spans(dev_sysfile_russian + '.epoch-{}'.format(epoch),
                              dev_sysfile_russian + '.epoch-{}'.format(epoch))
        dev_score_russian = compute_basic_score(
            reffile=dev_reffile,
            sysfile=dev_sysfile_russian + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'analysis.arg.report.epoch-{}.csv'.format(epoch))
        )
        # test set
        progress = tqdm.tqdm(total=test_batch_num_russian, ncols=75,
                             desc='Test {}'.format(epoch))

        test_predictions = []
        for batch in DataLoader(test_set_russian, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=test_set_russian.collate_fn):
            progress.update(1)
            graphs = model.predict(batch)
            test_predictions.extend(graphs)
        progress.close()

        write_arg_predictions(
            test_predictions,
            output_fpath=test_sysfile_russian + '.epoch-{}'.format(epoch),
            infile=test_infile_russian
        )
        remove_pronouns(test_sysfile_russian + '.epoch-{}'.format(epoch),
                        test_sysfile_russian + '.epoch-{}'.format(epoch))
        normalize_money_spans(test_sysfile_russian + '.epoch-{}'.format(epoch),
                              test_sysfile_russian + '.epoch-{}'.format(epoch))
        test_score_russian = compute_basic_score(
            reffile=test_reffile,
            sysfile=test_sysfile_russian + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'devtest.arg.report.epoch-{}.csv'.format(epoch))
        )

        # chinese test set
        progress = tqdm.tqdm(total=chinese_batch_num, ncols=75,
                             desc='Chinese {}'.format(epoch))

        chinese_argument_predictions = []
        for batch in DataLoader(chinese_set, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=chinese_set.collate_fn):
            progress.update(1)
            graphs = model.predict(batch)
            chinese_argument_predictions.extend(graphs)
        progress.close()

        write_arg_predictions(
            chinese_argument_predictions,
            output_fpath=chinese_sysfile + '.epoch-{}'.format(epoch),
            infile=chinese_infile
        )
        remove_pronouns(chinese_sysfile + '.epoch-{}'.format(epoch), chinese_sysfile + '.epoch-{}'.format(epoch))
        normalize_money_spans(chinese_sysfile + '.epoch-{}'.format(epoch), chinese_sysfile + '.epoch-{}'.format(epoch))
        chinese_score = compute_basic_score(
            reffile=chinese_reffile,
            sysfile=chinese_sysfile + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'chinese.arg.report.epoch-{}.csv'.format(epoch))
        )

        # korean test set
        progress = tqdm.tqdm(total=korean_batch_num, ncols=75,
                             desc='korean {}'.format(epoch))

        korean_argument_predictions = []
        for batch in DataLoader(korean_set, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=korean_set.collate_fn):
            progress.update(1)
            graphs = model.predict(batch)
            korean_argument_predictions.extend(graphs)
        progress.close()

        write_arg_predictions(
            korean_argument_predictions,
            output_fpath=korean_sysfile + '.epoch-{}'.format(epoch),
            infile=korean_infile
        )
        remove_pronouns(korean_sysfile + '.epoch-{}'.format(epoch), korean_sysfile + '.epoch-{}'.format(epoch))
        normalize_money_spans(korean_sysfile + '.epoch-{}'.format(epoch), korean_sysfile + '.epoch-{}'.format(epoch))
        korean_score = compute_basic_score(
            reffile=korean_reffile,
            sysfile=korean_sysfile + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'korean.arg.report.epoch-{}.csv'.format(epoch))
        )

        # russian test set
        progress = tqdm.tqdm(total=russian_batch_num, ncols=75,
                             desc='russian {}'.format(epoch))

        russian_argument_predictions = []
        for batch in DataLoader(russian_set, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=russian_set.collate_fn):
            progress.update(1)
            graphs = model.predict(batch)
            russian_argument_predictions.extend(graphs)
        progress.close()

        write_arg_predictions(
            russian_argument_predictions,
            output_fpath=russian_sysfile + '.epoch-{}'.format(epoch),
            infile=russian_infile
        )
        remove_pronouns(russian_sysfile + '.epoch-{}'.format(epoch), russian_sysfile + '.epoch-{}'.format(epoch))
        normalize_money_spans(russian_sysfile + '.epoch-{}'.format(epoch), russian_sysfile + '.epoch-{}'.format(epoch))
        russian_score = compute_basic_score(
            reffile=russian_reffile,
            sysfile=russian_sysfile + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'russian.arg.report.epoch-{}.csv'.format(epoch))
        )
        dev_score_chinese['epoch'] = epoch
        test_score_chinese['epoch'] = epoch

        dev_score_korean['epoch'] = epoch
        test_score_korean['epoch'] = epoch

        dev_score_russian['epoch'] = epoch
        test_score_russian['epoch'] = epoch

        chinese_score['epoch'] = epoch
        korean_score['epoch'] = epoch
        russian_score['epoch'] = epoch

        print_and_log('=' * 50, logger)
        print_and_log('Loss: {:.3f}'.format(sum(losses) / len(losses)), logger)
        print_and_log('-' * 20, logger)
        print_and_log('Dev Chinese: {}'.format(dev_score_chinese), logger, printout=not config.docker_run)
        print_and_log('Test Chinese: {}'.format(test_score_chinese), logger, printout=not config.docker_run)

        print_and_log('Dev Korean: {}'.format(dev_score_korean), logger, printout=not config.docker_run)
        print_and_log('Test Korean: {}'.format(test_score_korean), logger, printout=not config.docker_run)

        print_and_log('Dev Russian: {}'.format(dev_score_russian), logger, printout=not config.docker_run)
        print_and_log('Test Russian: {}'.format(test_score_russian), logger, printout=not config.docker_run)

        print_and_log('Chinese: {}'.format(chinese_score), logger, printout=not config.docker_run)
        print_and_log('Korean: {}'.format(korean_score), logger, printout=not config.docker_run)
        print_and_log('Russian: {}'.format(russian_score), logger, printout=not config.docker_run)

        if epoch == 0 or (dev_score_chinese['combined'] + chinese_score['combined']) > (
                best_english_dev_chinese['combined'] + best_sample_chinese['combined']):
            best_english_dev_chinese = dev_score_chinese
            best_english_test_chinese = test_score_chinese
            best_sample_chinese = chinese_score
            print_and_log('New best chinese model...', logger)

            torch.save(state, argument_model_save_fpath + '.lang-chinese')

            # save best prediction files

            copyfile(dev_sysfile_chinese + '.epoch-{}'.format(epoch), dev_sysfile_chinese + '.best')
            copyfile(test_sysfile_chinese + '.epoch-{}'.format(epoch), test_sysfile_chinese + '.best')
            copyfile(chinese_sysfile + '.epoch-{}'.format(epoch), chinese_sysfile + '.best')

        if epoch == 0 or (dev_score_korean['combined'] + korean_score['combined']) > (
                best_english_dev_korean['combined'] + best_sample_korean['combined']):
            best_english_dev_korean = dev_score_korean
            best_english_test_korean = test_score_korean
            best_sample_korean = korean_score
            print_and_log('New best korean model...', logger)

            torch.save(state, argument_model_save_fpath + '.lang-korean')

            # save best prediction files

            copyfile(dev_sysfile_korean + '.epoch-{}'.format(epoch), dev_sysfile_korean + '.best')
            copyfile(test_sysfile_korean + '.epoch-{}'.format(epoch), test_sysfile_korean + '.best')
            copyfile(korean_sysfile + '.epoch-{}'.format(epoch), korean_sysfile + '.best')

        if epoch == 0 or (dev_score_russian['combined'] + russian_score['combined']) > (
                best_english_dev_russian['combined'] + best_sample_russian['combined']):
            best_english_dev_russian = dev_score_russian
            best_english_test_russian = test_score_russian
            best_sample_russian = russian_score
            print_and_log('New best russian model...', logger)

            torch.save(state, argument_model_save_fpath + '.lang-russian')

            # save best prediction files

            copyfile(dev_sysfile_russian + '.epoch-{}'.format(epoch), dev_sysfile_russian + '.best')
            copyfile(test_sysfile_russian + '.epoch-{}'.format(epoch), test_sysfile_russian + '.best')
            copyfile(russian_sysfile + '.epoch-{}'.format(epoch), russian_sysfile + '.best')

        print_and_log('=' * 50, logger, printout=not config.docker_run)
        print_and_log('Current best result:', logger, printout=not config.docker_run)
        print_and_log('-' * 20, logger, printout=not config.docker_run)
        print_and_log('Dev for Chinese: {}'.format(best_english_dev_chinese), logger, printout=not config.docker_run)
        print_and_log('Test for Chinese: {}'.format(best_english_test_chinese), logger, printout=not config.docker_run)
        print_and_log('Chinese: {}'.format(best_sample_chinese), logger, printout=not config.docker_run)
        print_and_log('-' * 20, logger, printout=not config.docker_run)
        print_and_log('Dev for Korean: {}'.format(best_english_dev_korean), logger, printout=not config.docker_run)
        print_and_log('Test for Korean: {}'.format(best_english_test_korean), logger, printout=not config.docker_run)
        print_and_log('Korean: {}'.format(best_sample_korean), logger, printout=not config.docker_run)
        print_and_log('-' * 20, logger, printout=not config.docker_run)
        print_and_log('Dev for Russian: {}'.format(best_english_dev_russian), logger, printout=not config.docker_run)
        print_and_log('Test for Russian: {}'.format(best_english_test_russian), logger, printout=not config.docker_run)
        print_and_log('Russian: {}'.format(best_sample_russian), logger, printout=not config.docker_run)

elif config.model_name.startswith('template'):
    train_set = TemplateDataset(config.train_file, config)

    dev_set_chinese = TemplateDataset(config.dev_file + '.lang-chinese', config)
    test_set_chinese = TemplateDataset(config.test_file + '.lang-chinese', config)

    dev_set_korean = TemplateDataset(config.dev_file + '.lang-korean', config)
    test_set_korean = TemplateDataset(config.test_file + '.lang-korean', config)

    dev_set_russian = TemplateDataset(config.dev_file + '.lang-russian', config)
    test_set_russian = TemplateDataset(config.test_file + '.lang-russian', config)

    chinese_set = TemplateDataset(config.chinese_file, config)
    korean_set = TemplateDataset(config.korean_file, config)
    russian_set = TemplateDataset(config.russian_file, config)

    train_set.numberize(tokenizer, vocabs)

    dev_set_chinese.numberize(tokenizer, vocabs)
    test_set_chinese.numberize(tokenizer, vocabs)

    dev_set_korean.numberize(tokenizer, vocabs)
    test_set_korean.numberize(tokenizer, vocabs)

    dev_set_russian.numberize(tokenizer, vocabs)
    test_set_russian.numberize(tokenizer, vocabs)

    chinese_set.numberize(tokenizer, vocabs)
    korean_set.numberize(tokenizer, vocabs)
    russian_set.numberize(tokenizer, vocabs)

    batch_num = len(train_set) // config.batch_size

    dev_batch_num_chinese = len(dev_set_chinese) // config.eval_batch_size + \
                            (len(dev_set_chinese) % config.eval_batch_size != 0)
    test_batch_num_chinese = len(test_set_chinese) // config.eval_batch_size + \
                             (len(test_set_chinese) % config.eval_batch_size != 0)

    dev_batch_num_korean = len(dev_set_korean) // config.eval_batch_size + \
                           (len(dev_set_korean) % config.eval_batch_size != 0)
    test_batch_num_korean = len(test_set_korean) // config.eval_batch_size + \
                            (len(test_set_korean) % config.eval_batch_size != 0)

    dev_batch_num_russian = len(dev_set_russian) // config.eval_batch_size + \
                            (len(dev_set_russian) % config.eval_batch_size != 0)
    test_batch_num_russian = len(test_set_russian) // config.eval_batch_size + \
                             (len(test_set_russian) % config.eval_batch_size != 0)
    chinese_batch_num = len(chinese_set) // config.eval_batch_size + \
                        (len(chinese_set) % config.eval_batch_size != 0)
    korean_batch_num = len(korean_set) // config.eval_batch_size + \
                       (len(korean_set) % config.eval_batch_size != 0)
    russian_batch_num = len(russian_set) // config.eval_batch_size + \
                        (len(russian_set) % config.eval_batch_size != 0)

    # initialize the model
    model = TemplateModel(config, vocabs)
    model.cuda()

    init_template_model_save_fpath = os.path.join('logs/template/template-model.mdl') + '.lang-chinese'
    if os.path.exists(init_template_model_save_fpath):
        init_template_ckpt = torch.load(init_template_model_save_fpath)
        model.load_state_dict(init_template_ckpt['model'])
        print_and_log('Initialized Template model with checkpoint at {}'.format(init_template_model_save_fpath), logger)
        del init_template_ckpt
        torch.cuda.empty_cache()

    # print trainable weights
    n_trainable_params, n_nontrainable_params = 0, 0
    for p in model.parameters():
        n_params = torch.prod(torch.tensor(p.shape)).item()
        if p.requires_grad:
            n_trainable_params += n_params
        else:
            n_nontrainable_params += n_params
    logger.info('-' * 100)
    logger.info('> trainable params:')
    for name, param in model.named_parameters():
        if param.requires_grad:
            logger.info('>>> {0}: {1}'.format(name, param.shape))
    logger.info(
        'n_trainable_params: {0}, n_nontrainable_params: {1}'.format(n_trainable_params, n_nontrainable_params))
    logger.info('-' * 100)

    # optimizer
    param_groups = [
        {
            'params': [p for n, p in model.named_parameters() if n.startswith('xlmr')],
            'lr': config.xlmr_learning_rate, 'weight_decay': config.xlmr_weight_decay
        },
        {
            'params': [p for n, p in model.named_parameters() if not n.startswith('xlmr')
                       and 'crf' not in n],
            'lr': config.learning_rate, 'weight_decay': config.weight_decay
        },
        {
            'params': [p for n, p in model.named_parameters() if not n.startswith('xlmr')
                       and ('crf' in n)],
            'lr': config.learning_rate, 'weight_decay': 0
        }
    ]
    optimizer = AdamW(params=param_groups)
    schedule = get_linear_schedule_with_warmup(optimizer,
                                               num_warmup_steps=batch_num * 5,
                                               num_training_steps=batch_num * 100)

    # model state
    state = dict(model=model.state_dict(),
                 event_hidden_num=config.event_hidden_num,
                 role_hidden_num=config.role_hidden_num)

    global_step = 0

    best_english_dev_chinese = {
        'epoch': 0,
        'template': {'f1': 0},
        'slot': {'f1': 0},
        'combined': 0
    }
    best_english_test_chinese = {
        'epoch': 0,
        'template': {'f1': 0},
        'slot': {'f1': 0},
        'combined': 0
    }
    best_sample_chinese = {
        'epoch': 0,
        'template': {'f1': 0},
        'slot': {'f1': 0},
        'combined': 0
    }

    best_english_dev_korean = {
        'epoch': 0,
        'template': {'f1': 0},
        'slot': {'f1': 0},
        'combined': 0
    }
    best_english_test_korean = {
        'epoch': 0,
        'template': {'f1': 0},
        'slot': {'f1': 0},
        'combined': 0
    }
    best_sample_korean = {
        'epoch': 0,
        'template': {'f1': 0},
        'slot': {'f1': 0},
        'combined': 0
    }

    best_english_dev_russian = {
        'epoch': 0,
        'template': {'f1': 0},
        'slot': {'f1': 0},
        'combined': 0
    }
    best_english_test_russian = {
        'epoch': 0,
        'template': {'f1': 0},
        'slot': {'f1': 0},
        'combined': 0
    }
    best_sample_russian = {
        'epoch': 0,
        'template': {'f1': 0},
        'slot': {'f1': 0},
        'combined': 0
    }

    dev_infile_chinese = os.path.join('logs/arg/results.arg.analysis.bp.json.lang-chinese.best')
    test_infile_chinese = os.path.join('logs/arg/results.arg.devtest.bp.json.lang-chinese.best')

    dev_infile_korean = os.path.join('logs/arg/results.arg.analysis.bp.json.lang-korean.best')
    test_infile_korean = os.path.join('logs/arg/results.arg.devtest.bp.json.lang-korean.best')

    dev_infile_russian = os.path.join('logs/arg/results.arg.analysis.bp.json.lang-russian.best')
    test_infile_russian = os.path.join('logs/arg/results.arg.devtest.bp.json.lang-russian.best')

    chinese_infile = os.path.join('logs/arg/results.arg.chinese-sample.bp.json.best')
    korean_infile = os.path.join('logs/arg/results.arg.korean-sample.bp.json.best')
    russian_infile = os.path.join('logs/arg/results.arg.russian-sample.bp.json.best')

    dev_sysfile_chinese = os.path.join(log_dir,
                                       'results.{}.analysis.bp.json'.format(config.model_name)) + '.lang-chinese'
    test_sysfile_chinese = os.path.join(log_dir,
                                        'results.{}.devtest.bp.json'.format(config.model_name)) + '.lang-chinese'

    dev_sysfile_korean = os.path.join(log_dir,
                                      'results.{}.analysis.bp.json'.format(config.model_name)) + '.lang-korean'
    test_sysfile_korean = os.path.join(log_dir,
                                       'results.{}.devtest.bp.json'.format(config.model_name)) + '.lang-korean'

    dev_sysfile_russian = os.path.join(log_dir,
                                       'results.{}.analysis.bp.json'.format(config.model_name)) + '.lang-russian'
    test_sysfile_russian = os.path.join(log_dir,
                                        'results.{}.devtest.bp.json'.format(config.model_name)) + '.lang-russian'

    chinese_sysfile = os.path.join(log_dir, 'results.{}.chinese-sample.bp.json'.format(config.model_name))
    korean_sysfile = os.path.join(log_dir, 'results.{}.korean-sample.bp.json'.format(config.model_name))
    russian_sysfile = os.path.join(log_dir, 'results.{}.russian-sample.bp.json'.format(config.model_name))

    dev_reffile = GRANULAR_ANALYSIS_REF_P3
    test_reffile = GRANULAR_DEVTEST_REF_P3
    chinese_reffile = SAMPLE_GRANULAR_CHINESE_P3
    korean_reffile = SAMPLE_GRANULAR_KOREAN_P3
    russian_reffile = SAMPLE_GRANULAR_RUSSIAN_P3

    for epoch in range(config.max_epoch):
        torch.cuda.empty_cache()
        print_and_log('*' * 30, logger)
        print_and_log('Epoch: {}'.format(epoch), logger)
        print_and_log('command: {}'.format(command), logger)
        # training set
        progress = tqdm.tqdm(total=batch_num, ncols=75,
                             desc='Train {}'.format(epoch))
        model.train()
        optimizer.zero_grad()
        losses = []
        for batch_idx, batch in enumerate(DataLoader(
                train_set, batch_size=config.batch_size // config.accumulate_step,
                shuffle=True, collate_fn=train_set.collate_fn)):

            loss = model(batch)
            loss = loss * (1 / config.accumulate_step)

            loss.backward()

            if (batch_idx + 1) % config.accumulate_step == 0:
                progress.update(1)
                global_step += 1
                torch.nn.utils.clip_grad_norm_(
                    model.parameters(), config.grad_clipping)
                optimizer.step()
                schedule.step()
                optimizer.zero_grad()
                print_and_log('{}: step: {}/{}, loss: {}'.format(datetime.now(), batch_idx + 1, batch_num, loss.item()),
                              logger,
                              printout=False)
                losses.append(loss.item())
        progress.close()

        model.eval()
        # dev set (Chinese)
        progress = tqdm.tqdm(total=dev_batch_num_chinese, ncols=75,
                             desc='Dev {}'.format(epoch))

        dev_predictions_chinese = []
        for batch in DataLoader(dev_set_chinese, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=dev_set_chinese.collate_fn):
            progress.update(1)
            dev_predictions_chinese.extend(model.predict(batch))
        progress.close()

        # test set
        progress = tqdm.tqdm(total=test_batch_num_chinese, ncols=75,
                             desc='Test {}'.format(epoch))

        test_predictions_chinese = []
        for batch in DataLoader(test_set_chinese, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=test_set_chinese.collate_fn):
            progress.update(1)
            test_predictions_chinese.extend(model.predict(batch))
        progress.close()

        # dev set (Korean)
        progress = tqdm.tqdm(total=dev_batch_num_korean, ncols=75,
                             desc='Dev {}'.format(epoch))

        dev_predictions_korean = []
        for batch in DataLoader(dev_set_korean, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=dev_set_korean.collate_fn):
            progress.update(1)
            dev_predictions_korean.extend(model.predict(batch))
        progress.close()

        # test set
        progress = tqdm.tqdm(total=test_batch_num_korean, ncols=75,
                             desc='Test {}'.format(epoch))

        test_predictions_korean = []
        for batch in DataLoader(test_set_korean, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=test_set_korean.collate_fn):
            progress.update(1)
            test_predictions_korean.extend(model.predict(batch))
        progress.close()

        # dev set (Russian)
        progress = tqdm.tqdm(total=dev_batch_num_russian, ncols=75,
                             desc='Dev {}'.format(epoch))

        dev_predictions_russian = []
        for batch in DataLoader(dev_set_russian, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=dev_set_russian.collate_fn):
            progress.update(1)
            dev_predictions_russian.extend(model.predict(batch))
        progress.close()

        # test set
        progress = tqdm.tqdm(total=test_batch_num_russian, ncols=75,
                             desc='Test {}'.format(epoch))

        test_predictions_russian = []
        for batch in DataLoader(test_set_russian, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=test_set_russian.collate_fn):
            progress.update(1)
            test_predictions_russian.extend(model.predict(batch))
        progress.close()

        # chinese test set
        progress = tqdm.tqdm(total=chinese_batch_num, ncols=75,
                             desc='chinese {}'.format(epoch))

        chinese_template_predictions = []
        for batch in DataLoader(chinese_set, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=chinese_set.collate_fn):
            progress.update(1)
            chinese_template_predictions.extend(model.predict(batch))
        progress.close()

        # korean test set
        progress = tqdm.tqdm(total=korean_batch_num, ncols=75,
                             desc='korean {}'.format(epoch))

        korean_template_predictions = []
        for batch in DataLoader(korean_set, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=korean_set.collate_fn):
            progress.update(1)
            korean_template_predictions.extend(model.predict(batch))
        progress.close()

        # russian test set
        progress = tqdm.tqdm(total=russian_batch_num, ncols=75,
                             desc='russian {}'.format(epoch))

        russian_template_predictions = []
        for batch in DataLoader(russian_set, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=russian_set.collate_fn):
            progress.update(1)
            russian_template_predictions.extend(model.predict(batch))
        progress.close()

        write_template_predictions(
            dev_predictions_chinese,
            dev_infile_chinese,
            output_fpath=dev_sysfile_chinese + '.epoch-{}'.format(epoch)
        )
        dev_score_chinese = compute_granular_score(
            reffile=dev_reffile,
            sysfile=dev_sysfile_chinese + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'analysis.template.report.epoch-{}.csv'.format(epoch))
        )

        write_template_predictions(
            test_predictions_chinese,
            test_infile_chinese,
            output_fpath=test_sysfile_chinese + '.epoch-{}'.format(epoch)
        )

        test_score_chinese = compute_granular_score(
            reffile=test_reffile,
            sysfile=test_sysfile_chinese + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'devtest.template.report.epoch-{}.csv'.format(epoch))
        )

        # ===============

        write_template_predictions(
            dev_predictions_korean,
            dev_infile_korean,
            output_fpath=dev_sysfile_korean + '.epoch-{}'.format(epoch)
        )
        dev_score_korean = compute_granular_score(
            reffile=dev_reffile,
            sysfile=dev_sysfile_korean + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'analysis.template.report.epoch-{}.csv'.format(epoch))
        )

        write_template_predictions(
            test_predictions_korean,
            test_infile_korean,
            output_fpath=test_sysfile_korean + '.epoch-{}'.format(epoch)
        )

        test_score_korean = compute_granular_score(
            reffile=test_reffile,
            sysfile=test_sysfile_korean + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'devtest.template.report.epoch-{}.csv'.format(epoch))
        )

        # ===============

        write_template_predictions(
            dev_predictions_russian,
            dev_infile_russian,
            output_fpath=dev_sysfile_russian + '.epoch-{}'.format(epoch)
        )
        dev_score_russian = compute_granular_score(
            reffile=dev_reffile,
            sysfile=dev_sysfile_russian + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'analysis.template.report.epoch-{}.csv'.format(epoch))
        )

        write_template_predictions(
            test_predictions_russian,
            test_infile_russian,
            output_fpath=test_sysfile_russian + '.epoch-{}'.format(epoch)
        )

        test_score_russian = compute_granular_score(
            reffile=test_reffile,
            sysfile=test_sysfile_russian + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'devtest.template.report.epoch-{}.csv'.format(epoch))
        )

        write_template_predictions(
            chinese_template_predictions,
            chinese_infile,
            output_fpath=chinese_sysfile + '.epoch-{}'.format(epoch)
        )

        chinese_score = compute_granular_score(
            reffile=chinese_reffile,
            sysfile=chinese_sysfile + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'chinese.template.report.epoch-{}.csv'.format(epoch))
        )
        # ------------
        write_template_predictions(
            korean_template_predictions,
            korean_infile,
            output_fpath=korean_sysfile + '.epoch-{}'.format(epoch)
        )

        korean_score = compute_granular_score(
            reffile=korean_reffile,
            sysfile=korean_sysfile + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'korean.template.report.epoch-{}.csv'.format(epoch))
        )
        # ------------
        write_template_predictions(
            russian_template_predictions,
            russian_infile,
            output_fpath=russian_sysfile + '.epoch-{}'.format(epoch)
        )

        russian_score = compute_granular_score(
            reffile=russian_reffile,
            sysfile=russian_sysfile + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'russian.template.report.epoch-{}.csv'.format(epoch))
        )

        dev_score_chinese['epoch'] = epoch
        test_score_chinese['epoch'] = epoch

        dev_score_korean['epoch'] = epoch
        test_score_korean['epoch'] = epoch

        dev_score_russian['epoch'] = epoch
        test_score_russian['epoch'] = epoch

        chinese_score['epoch'] = epoch
        korean_score['epoch'] = epoch
        russian_score['epoch'] = epoch

        print_and_log('=' * 50, logger)
        print_and_log('Loss: {:.3f}'.format(sum(losses) / len(losses)), logger)
        print_and_log('-' * 20, logger)
        print_and_log('Dev Chinese: {}'.format(dev_score_chinese), logger, printout=not config.docker_run)
        print_and_log('Test Chinese: {}'.format(test_score_chinese), logger, printout=not config.docker_run)

        print_and_log('Dev Korean: {}'.format(dev_score_korean), logger, printout=not config.docker_run)
        print_and_log('Test Korean: {}'.format(test_score_korean), logger, printout=not config.docker_run)

        print_and_log('Dev Russian: {}'.format(dev_score_russian), logger, printout=not config.docker_run)
        print_and_log('Test Russian: {}'.format(test_score_russian), logger, printout=not config.docker_run)

        print_and_log('Sample Chinese: {}'.format(chinese_score), logger, printout=not config.docker_run)
        print_and_log('Sample Korean: {}'.format(korean_score), logger, printout=not config.docker_run)
        print_and_log('Sample Russian: {}'.format(russian_score), logger, printout=not config.docker_run)

        if epoch == 0 or (dev_score_chinese['template']['f1'] + chinese_score['template']['f1']) > (
                best_english_dev_chinese['template']['f1'] + best_sample_chinese['template']['f1']):
            best_english_dev_chinese = dev_score_chinese
            best_english_test_chinese = test_score_chinese
            best_sample_chinese = chinese_score
            print_and_log('New best chinese model...', logger)

            torch.save(state, template_model_save_fpath + '.lang-chinese')

            # save best prediction files
            copyfile(dev_sysfile_chinese + '.epoch-{}'.format(epoch), dev_sysfile_chinese + '.best')
            copyfile(test_sysfile_chinese + '.epoch-{}'.format(epoch), test_sysfile_chinese + '.best')
            copyfile(chinese_sysfile + '.epoch-{}'.format(epoch), chinese_sysfile + '.best')

        if epoch == 0 or (dev_score_korean['template']['f1'] + korean_score['template']['f1']) > (
                best_english_dev_korean['template']['f1'] + best_sample_korean['template']['f1']):
            best_english_dev_korean = dev_score_korean
            best_english_test_korean = test_score_korean
            best_sample_korean = korean_score
            print_and_log('New best korean model...', logger)

            torch.save(state, template_model_save_fpath + '.lang-korean')

            # save best prediction files
            copyfile(dev_sysfile_korean + '.epoch-{}'.format(epoch), dev_sysfile_korean + '.best')
            copyfile(test_sysfile_korean + '.epoch-{}'.format(epoch), test_sysfile_korean + '.best')
            copyfile(korean_sysfile + '.epoch-{}'.format(epoch), korean_sysfile + '.best')

        if epoch == 0 or (dev_score_russian['template']['f1'] + russian_score['template']['f1']) > (
                best_english_dev_russian['template']['f1'] + best_sample_russian['template']['f1']):
            best_english_dev_russian = dev_score_russian
            best_english_test_russian = test_score_russian
            best_sample_russian = russian_score
            print_and_log('New best russian model...', logger)

            torch.save(state, template_model_save_fpath + '.lang-russian')

            # save best prediction files
            copyfile(dev_sysfile_russian + '.epoch-{}'.format(epoch), dev_sysfile_russian + '.best')
            copyfile(test_sysfile_russian + '.epoch-{}'.format(epoch), test_sysfile_russian + '.best')
            copyfile(russian_sysfile + '.epoch-{}'.format(epoch), russian_sysfile + '.best')

        print_and_log('=' * 50, logger, printout=not config.docker_run)
        print_and_log('Current best result:', logger, printout=not config.docker_run)
        print_and_log('-' * 20, logger, printout=not config.docker_run)
        print_and_log('Dev for Chinese: {}'.format(best_english_dev_chinese), logger, printout=not config.docker_run)
        print_and_log('Test for Chinese: {}'.format(best_english_test_chinese), logger, printout=not config.docker_run)
        print_and_log('Chinese: {}'.format(best_sample_chinese), logger, printout=not config.docker_run)
        print_and_log('-' * 20, logger, printout=not config.docker_run)
        print_and_log('Dev for Korean: {}'.format(best_english_dev_korean), logger, printout=not config.docker_run)
        print_and_log('Test for Korean: {}'.format(best_english_test_korean), logger, printout=not config.docker_run)
        print_and_log('Korean: {}'.format(best_sample_korean), logger, printout=not config.docker_run)
        print_and_log('-' * 20, logger, printout=not config.docker_run)
        print_and_log('Dev for Russian: {}'.format(best_english_dev_russian), logger, printout=not config.docker_run)
        print_and_log('Test for Russian: {}'.format(best_english_test_russian), logger, printout=not config.docker_run)
        print_and_log('Russian: {}'.format(best_sample_russian), logger, printout=not config.docker_run)

elif config.model_name.startswith('slot'):
    train_set = SlotDataset(config.train_file, config)

    dev_set_chinese = SlotDataset(config.dev_file + '.lang-chinese', config)
    test_set_chinese = SlotDataset(config.test_file + '.lang-chinese', config)

    dev_set_korean = SlotDataset(config.dev_file + '.lang-korean', config)
    test_set_korean = SlotDataset(config.test_file + '.lang-korean', config)

    dev_set_russian = SlotDataset(config.dev_file + '.lang-russian', config)
    test_set_russian = SlotDataset(config.test_file + '.lang-russian', config)

    chinese_set = SlotDataset(config.chinese_file, config)
    korean_set = SlotDataset(config.korean_file, config)
    russian_set = SlotDataset(config.russian_file, config)

    train_set.numberize(tokenizer, vocabs)

    dev_set_chinese.numberize(tokenizer, vocabs)
    test_set_chinese.numberize(tokenizer, vocabs)

    dev_set_korean.numberize(tokenizer, vocabs)
    test_set_korean.numberize(tokenizer, vocabs)

    dev_set_russian.numberize(tokenizer, vocabs)
    test_set_russian.numberize(tokenizer, vocabs)

    chinese_set.numberize(tokenizer, vocabs)
    korean_set.numberize(tokenizer, vocabs)
    russian_set.numberize(tokenizer, vocabs)

    batch_num = len(train_set) // config.batch_size

    dev_batch_num_chinese = len(dev_set_chinese) // config.eval_batch_size + \
                            (len(dev_set_chinese) % config.eval_batch_size != 0)
    test_batch_num_chinese = len(test_set_chinese) // config.eval_batch_size + \
                             (len(test_set_chinese) % config.eval_batch_size != 0)

    dev_batch_num_korean = len(dev_set_korean) // config.eval_batch_size + \
                           (len(dev_set_korean) % config.eval_batch_size != 0)
    test_batch_num_korean = len(test_set_korean) // config.eval_batch_size + \
                            (len(test_set_korean) % config.eval_batch_size != 0)

    dev_batch_num_russian = len(dev_set_russian) // config.eval_batch_size + \
                            (len(dev_set_russian) % config.eval_batch_size != 0)
    test_batch_num_russian = len(test_set_russian) // config.eval_batch_size + \
                             (len(test_set_russian) % config.eval_batch_size != 0)
    chinese_batch_num = len(chinese_set) // config.eval_batch_size + \
                        (len(chinese_set) % config.eval_batch_size != 0)
    korean_batch_num = len(korean_set) // config.eval_batch_size + \
                       (len(korean_set) % config.eval_batch_size != 0)
    russian_batch_num = len(russian_set) // config.eval_batch_size + \
                        (len(russian_set) % config.eval_batch_size != 0)

    # initialize the model
    model = SlotModel(config, vocabs)
    model.cuda()

    init_slot_model_save_fpath = os.path.join('logs/slot/slot-model.mdl') + '.lang-chinese'
    if os.path.exists(init_slot_model_save_fpath):
        init_slot_ckpt = torch.load(init_slot_model_save_fpath)
        model.load_state_dict(init_slot_ckpt['model'])
        print_and_log('Initialized slot model with checkpoint at {}'.format(init_slot_model_save_fpath), logger)
        del init_slot_ckpt
        torch.cuda.empty_cache()

    # print trainable weights
    n_trainable_params, n_nontrainable_params = 0, 0
    for p in model.parameters():
        n_params = torch.prod(torch.tensor(p.shape)).item()
        if p.requires_grad:
            n_trainable_params += n_params
        else:
            n_nontrainable_params += n_params
    logger.info('-' * 100)
    logger.info('> trainable params:')
    for name, param in model.named_parameters():
        if param.requires_grad:
            logger.info('>>> {0}: {1}'.format(name, param.shape))
    logger.info(
        'n_trainable_params: {0}, n_nontrainable_params: {1}'.format(n_trainable_params, n_nontrainable_params))
    logger.info('-' * 100)

    # optimizer
    param_groups = [
        {
            'params': [p for n, p in model.named_parameters() if n.startswith('xlmr')],
            'lr': config.xlmr_learning_rate, 'weight_decay': config.xlmr_weight_decay
        },
        {
            'params': [p for n, p in model.named_parameters() if not n.startswith('xlmr')
                       and 'crf' not in n],
            'lr': config.learning_rate, 'weight_decay': config.weight_decay
        },
        {
            'params': [p for n, p in model.named_parameters() if not n.startswith('xlmr')
                       and ('crf' in n)],
            'lr': config.learning_rate, 'weight_decay': 0
        }
    ]
    optimizer = AdamW(params=param_groups)
    schedule = get_linear_schedule_with_warmup(optimizer,
                                               num_warmup_steps=batch_num * 5,
                                               num_training_steps=batch_num * 100)

    # model state
    state = dict(model=model.state_dict(),
                 event_hidden_num=config.event_hidden_num,
                 role_hidden_num=config.role_hidden_num)

    global_step = 0

    best_english_dev_chinese = {
        'epoch': 0,
        'template': {'f1': 0},
        'slot': {'f1': 0},
        'combined': 0
    }
    best_english_test_chinese = {
        'epoch': 0,
        'template': {'f1': 0},
        'slot': {'f1': 0},
        'combined': 0
    }
    best_sample_chinese = {
        'epoch': 0,
        'template': {'f1': 0},
        'slot': {'f1': 0},
        'combined': 0
    }

    best_english_dev_korean = {
        'epoch': 0,
        'template': {'f1': 0},
        'slot': {'f1': 0},
        'combined': 0
    }
    best_english_test_korean = {
        'epoch': 0,
        'template': {'f1': 0},
        'slot': {'f1': 0},
        'combined': 0
    }
    best_sample_korean = {
        'epoch': 0,
        'template': {'f1': 0},
        'slot': {'f1': 0},
        'combined': 0
    }

    best_english_dev_russian = {
        'epoch': 0,
        'template': {'f1': 0},
        'slot': {'f1': 0},
        'combined': 0
    }
    best_english_test_russian = {
        'epoch': 0,
        'template': {'f1': 0},
        'slot': {'f1': 0},
        'combined': 0
    }
    best_sample_russian = {
        'epoch': 0,
        'template': {'f1': 0},
        'slot': {'f1': 0},
        'combined': 0
    }

    dev_infile_chinese = os.path.join('logs/template/results.template.analysis.bp.json.lang-chinese.best')
    test_infile_chinese = os.path.join('logs/template/results.template.devtest.bp.json.lang-chinese.best')

    dev_infile_korean = os.path.join('logs/template/results.template.analysis.bp.json.lang-korean.best')
    test_infile_korean = os.path.join('logs/template/results.template.devtest.bp.json.lang-korean.best')

    dev_infile_russian = os.path.join('logs/template/results.template.analysis.bp.json.lang-russian.best')
    test_infile_russian = os.path.join('logs/template/results.template.devtest.bp.json.lang-russian.best')

    chinese_infile = os.path.join('logs/template/results.template.chinese-sample.bp.json.best')
    korean_infile = os.path.join('logs/template/results.template.korean-sample.bp.json.best')
    russian_infile = os.path.join('logs/template/results.template.russian-sample.bp.json.best')

    dev_sysfile_chinese = os.path.join(log_dir,
                                       'results.{}.analysis.bp.json'.format(config.model_name)) + '.lang-chinese'
    test_sysfile_chinese = os.path.join(log_dir,
                                        'results.{}.devtest.bp.json'.format(config.model_name)) + '.lang-chinese'

    dev_sysfile_korean = os.path.join(log_dir,
                                      'results.{}.analysis.bp.json'.format(config.model_name)) + '.lang-korean'
    test_sysfile_korean = os.path.join(log_dir,
                                       'results.{}.devtest.bp.json'.format(config.model_name)) + '.lang-korean'

    dev_sysfile_russian = os.path.join(log_dir,
                                       'results.{}.analysis.bp.json'.format(config.model_name)) + '.lang-russian'
    test_sysfile_russian = os.path.join(log_dir,
                                        'results.{}.devtest.bp.json'.format(config.model_name)) + '.lang-russian'

    chinese_sysfile = os.path.join(log_dir, 'results.{}.chinese-sample.bp.json'.format(config.model_name))
    korean_sysfile = os.path.join(log_dir, 'results.{}.korean-sample.bp.json'.format(config.model_name))
    russian_sysfile = os.path.join(log_dir, 'results.{}.russian-sample.bp.json'.format(config.model_name))

    dev_reffile = GRANULAR_ANALYSIS_REF_P3
    test_reffile = GRANULAR_DEVTEST_REF_P3
    chinese_reffile = SAMPLE_GRANULAR_CHINESE_P3
    korean_reffile = SAMPLE_GRANULAR_KOREAN_P3
    russian_reffile = SAMPLE_GRANULAR_RUSSIAN_P3

    for epoch in range(config.max_epoch):
        torch.cuda.empty_cache()
        print_and_log('*' * 30, logger)
        print_and_log('Epoch: {}'.format(epoch), logger)
        print_and_log('command: {}'.format(command), logger)
        # training set
        progress = tqdm.tqdm(total=batch_num, ncols=75,
                             desc='Train {}'.format(epoch))
        model.train()
        optimizer.zero_grad()
        losses = []
        for batch_idx, batch in enumerate(DataLoader(
                train_set, batch_size=config.batch_size // config.accumulate_step,
                shuffle=True, collate_fn=train_set.collate_fn)):

            loss = model(batch)
            loss = loss * (1 / config.accumulate_step)

            loss.backward()

            if (batch_idx + 1) % config.accumulate_step == 0:
                progress.update(1)
                global_step += 1
                torch.nn.utils.clip_grad_norm_(
                    model.parameters(), config.grad_clipping)
                optimizer.step()
                schedule.step()
                optimizer.zero_grad()
                print_and_log('{}: step: {}/{}, loss: {}'.format(datetime.now(), batch_idx + 1, batch_num, loss.item()),
                              logger,
                              printout=False)
                losses.append(loss.item())
        progress.close()

        model.eval()
        # dev set (Chinese)
        progress = tqdm.tqdm(total=dev_batch_num_chinese, ncols=75,
                             desc='Dev {}'.format(epoch))

        dev_predictions_chinese = []
        for batch in DataLoader(dev_set_chinese, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=dev_set_chinese.collate_fn):
            progress.update(1)
            dev_predictions_chinese.extend(model.predict(batch))
        progress.close()

        # test set
        progress = tqdm.tqdm(total=test_batch_num_chinese, ncols=75,
                             desc='Test {}'.format(epoch))

        test_predictions_chinese = []
        for batch in DataLoader(test_set_chinese, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=test_set_chinese.collate_fn):
            progress.update(1)
            test_predictions_chinese.extend(model.predict(batch))
        progress.close()

        # dev set (Korean)
        progress = tqdm.tqdm(total=dev_batch_num_korean, ncols=75,
                             desc='Dev {}'.format(epoch))

        dev_predictions_korean = []
        for batch in DataLoader(dev_set_korean, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=dev_set_korean.collate_fn):
            progress.update(1)
            dev_predictions_korean.extend(model.predict(batch))
        progress.close()

        # test set
        progress = tqdm.tqdm(total=test_batch_num_korean, ncols=75,
                             desc='Test {}'.format(epoch))

        test_predictions_korean = []
        for batch in DataLoader(test_set_korean, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=test_set_korean.collate_fn):
            progress.update(1)
            test_predictions_korean.extend(model.predict(batch))
        progress.close()

        # dev set (Russian)
        progress = tqdm.tqdm(total=dev_batch_num_russian, ncols=75,
                             desc='Dev {}'.format(epoch))

        dev_predictions_russian = []
        for batch in DataLoader(dev_set_russian, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=dev_set_russian.collate_fn):
            progress.update(1)
            dev_predictions_russian.extend(model.predict(batch))
        progress.close()

        # test set
        progress = tqdm.tqdm(total=test_batch_num_russian, ncols=75,
                             desc='Test {}'.format(epoch))

        test_predictions_russian = []
        for batch in DataLoader(test_set_russian, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=test_set_russian.collate_fn):
            progress.update(1)
            test_predictions_russian.extend(model.predict(batch))
        progress.close()

        # chinese test set
        progress = tqdm.tqdm(total=chinese_batch_num, ncols=75,
                             desc='chinese {}'.format(epoch))

        chinese_slot_predictions = []
        for batch in DataLoader(chinese_set, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=chinese_set.collate_fn):
            progress.update(1)
            chinese_slot_predictions.extend(model.predict(batch))
        progress.close()

        # korean test set
        progress = tqdm.tqdm(total=korean_batch_num, ncols=75,
                             desc='korean {}'.format(epoch))

        korean_slot_predictions = []
        for batch in DataLoader(korean_set, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=korean_set.collate_fn):
            progress.update(1)
            korean_slot_predictions.extend(model.predict(batch))
        progress.close()

        # russian test set
        progress = tqdm.tqdm(total=russian_batch_num, ncols=75,
                             desc='russian {}'.format(epoch))

        russian_slot_predictions = []
        for batch in DataLoader(russian_set, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=russian_set.collate_fn):
            progress.update(1)
            russian_slot_predictions.extend(model.predict(batch))
        progress.close()

        write_slot_predictions(
            dev_predictions_chinese,
            dev_infile_chinese,
            output_fpath=dev_sysfile_chinese + '.epoch-{}'.format(epoch)
        )
        dev_score_chinese = compute_granular_score(
            reffile=dev_reffile,
            sysfile=dev_sysfile_chinese + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'analysis.slot.report.epoch-{}.csv'.format(epoch))
        )

        write_slot_predictions(
            test_predictions_chinese,
            test_infile_chinese,
            output_fpath=test_sysfile_chinese + '.epoch-{}'.format(epoch)
        )

        test_score_chinese = compute_granular_score(
            reffile=test_reffile,
            sysfile=test_sysfile_chinese + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'devtest.slot.report.epoch-{}.csv'.format(epoch))
        )

        # ===============

        write_slot_predictions(
            dev_predictions_korean,
            dev_infile_korean,
            output_fpath=dev_sysfile_korean + '.epoch-{}'.format(epoch)
        )
        dev_score_korean = compute_granular_score(
            reffile=dev_reffile,
            sysfile=dev_sysfile_korean + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'analysis.slot.report.epoch-{}.csv'.format(epoch))
        )

        write_slot_predictions(
            test_predictions_korean,
            test_infile_korean,
            output_fpath=test_sysfile_korean + '.epoch-{}'.format(epoch)
        )

        test_score_korean = compute_granular_score(
            reffile=test_reffile,
            sysfile=test_sysfile_korean + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'devtest.slot.report.epoch-{}.csv'.format(epoch))
        )

        # ===============

        write_slot_predictions(
            dev_predictions_russian,
            dev_infile_russian,
            output_fpath=dev_sysfile_russian + '.epoch-{}'.format(epoch)
        )
        dev_score_russian = compute_granular_score(
            reffile=dev_reffile,
            sysfile=dev_sysfile_russian + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'analysis.slot.report.epoch-{}.csv'.format(epoch))
        )

        write_slot_predictions(
            test_predictions_russian,
            test_infile_russian,
            output_fpath=test_sysfile_russian + '.epoch-{}'.format(epoch)
        )

        test_score_russian = compute_granular_score(
            reffile=test_reffile,
            sysfile=test_sysfile_russian + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'devtest.slot.report.epoch-{}.csv'.format(epoch))
        )

        write_slot_predictions(
            chinese_slot_predictions,
            chinese_infile,
            output_fpath=chinese_sysfile + '.epoch-{}'.format(epoch)
        )

        chinese_score = compute_granular_score(
            reffile=chinese_reffile,
            sysfile=chinese_sysfile + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'chinese.slot.report.epoch-{}.csv'.format(epoch))
        )
        # ------------
        write_slot_predictions(
            korean_slot_predictions,
            korean_infile,
            output_fpath=korean_sysfile + '.epoch-{}'.format(epoch)
        )

        korean_score = compute_granular_score(
            reffile=korean_reffile,
            sysfile=korean_sysfile + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'korean.slot.report.epoch-{}.csv'.format(epoch))
        )
        # ------------
        write_slot_predictions(
            russian_slot_predictions,
            russian_infile,
            output_fpath=russian_sysfile + '.epoch-{}'.format(epoch)
        )

        russian_score = compute_granular_score(
            reffile=russian_reffile,
            sysfile=russian_sysfile + '.epoch-{}'.format(epoch),
            csv_fpath=os.path.join(log_dir, 'russian.slot.report.epoch-{}.csv'.format(epoch))
        )

        dev_score_chinese['epoch'] = epoch
        test_score_chinese['epoch'] = epoch

        dev_score_korean['epoch'] = epoch
        test_score_korean['epoch'] = epoch

        dev_score_russian['epoch'] = epoch
        test_score_russian['epoch'] = epoch

        chinese_score['epoch'] = epoch
        korean_score['epoch'] = epoch
        russian_score['epoch'] = epoch

        print_and_log('=' * 50, logger)
        print_and_log('Loss: {:.3f}'.format(sum(losses) / len(losses)), logger)
        print_and_log('-' * 20, logger)
        print_and_log('Dev Chinese: {}'.format(dev_score_chinese), logger, printout=not config.docker_run)
        print_and_log('Test Chinese: {}'.format(test_score_chinese), logger, printout=not config.docker_run)

        print_and_log('Dev Korean: {}'.format(dev_score_korean), logger, printout=not config.docker_run)
        print_and_log('Test Korean: {}'.format(test_score_korean), logger, printout=not config.docker_run)

        print_and_log('Dev Russian: {}'.format(dev_score_russian), logger, printout=not config.docker_run)
        print_and_log('Test Russian: {}'.format(test_score_russian), logger, printout=not config.docker_run)

        print_and_log('Sample Chinese: {}'.format(chinese_score), logger, printout=not config.docker_run)
        print_and_log('Sample Korean: {}'.format(korean_score), logger, printout=not config.docker_run)
        print_and_log('Sample Russian: {}'.format(russian_score), logger, printout=not config.docker_run)

        if epoch == 0 or (dev_score_chinese['slot']['f1'] + chinese_score['slot']['f1']) > (
                best_english_dev_chinese['slot']['f1'] + best_sample_chinese['slot']['f1']):
            best_english_dev_chinese = dev_score_chinese
            best_english_test_chinese = test_score_chinese
            best_sample_chinese = chinese_score
            print_and_log('New best chinese model...', logger)

            torch.save(state, slot_model_save_fpath + '.lang-chinese')

            # save best prediction files
            copyfile(dev_sysfile_chinese + '.epoch-{}'.format(epoch), dev_sysfile_chinese + '.best')
            copyfile(test_sysfile_chinese + '.epoch-{}'.format(epoch), test_sysfile_chinese + '.best')
            copyfile(chinese_sysfile + '.epoch-{}'.format(epoch), chinese_sysfile + '.best')

        if epoch == 0 or (dev_score_korean['slot']['f1'] + korean_score['slot']['f1']) > (
                best_english_dev_korean['slot']['f1'] + best_sample_korean['slot']['f1']):
            best_english_dev_korean = dev_score_korean
            best_english_test_korean = test_score_korean
            best_sample_korean = korean_score
            print_and_log('New best korean model...', logger)

            torch.save(state, slot_model_save_fpath + '.lang-korean')

            # save best prediction files
            copyfile(dev_sysfile_korean + '.epoch-{}'.format(epoch), dev_sysfile_korean + '.best')
            copyfile(test_sysfile_korean + '.epoch-{}'.format(epoch), test_sysfile_korean + '.best')
            copyfile(korean_sysfile + '.epoch-{}'.format(epoch), korean_sysfile + '.best')

        if epoch == 0 or (dev_score_russian['slot']['f1'] + russian_score['slot']['f1']) > (
                best_english_dev_russian['slot']['f1'] + best_sample_russian['slot']['f1']):
            best_english_dev_russian = dev_score_russian
            best_english_test_russian = test_score_russian
            best_sample_russian = russian_score
            print_and_log('New best russian model...', logger)

            torch.save(state, slot_model_save_fpath + '.lang-russian')

            # save best prediction files
            copyfile(dev_sysfile_russian + '.epoch-{}'.format(epoch), dev_sysfile_russian + '.best')
            copyfile(test_sysfile_russian + '.epoch-{}'.format(epoch), test_sysfile_russian + '.best')
            copyfile(russian_sysfile + '.epoch-{}'.format(epoch), russian_sysfile + '.best')

        print_and_log('=' * 50, logger, printout=not config.docker_run)
        print_and_log('Current best result:', logger, printout=not config.docker_run)
        print_and_log('-' * 20, logger, printout=not config.docker_run)
        print_and_log('Dev for Chinese: {}'.format(best_english_dev_chinese), logger, printout=not config.docker_run)
        print_and_log('Test for Chinese: {}'.format(best_english_test_chinese), logger, printout=not config.docker_run)
        print_and_log('Chinese: {}'.format(best_sample_chinese), logger, printout=not config.docker_run)
        print_and_log('-' * 20, logger, printout=not config.docker_run)
        print_and_log('Dev for Korean: {}'.format(best_english_dev_korean), logger, printout=not config.docker_run)
        print_and_log('Test for Korean: {}'.format(best_english_test_korean), logger, printout=not config.docker_run)
        print_and_log('Korean: {}'.format(best_sample_korean), logger, printout=not config.docker_run)
        print_and_log('-' * 20, logger, printout=not config.docker_run)
        print_and_log('Dev for Russian: {}'.format(best_english_dev_russian), logger, printout=not config.docker_run)
        print_and_log('Test for Russian: {}'.format(best_english_test_russian), logger, printout=not config.docker_run)
        print_and_log('Russian: {}'.format(best_sample_russian), logger, printout=not config.docker_run)

elif config.model_name.startswith('coref'):
    train_set = EventCorefDataset(config.train_file, config)

    dev_set_chinese = EventCorefDataset(config.dev_file + '.lang-chinese', config)
    test_set_chinese = EventCorefDataset(config.test_file + '.lang-chinese', config)

    dev_set_korean = EventCorefDataset(config.dev_file + '.lang-korean', config)
    test_set_korean = EventCorefDataset(config.test_file + '.lang-korean', config)

    dev_set_russian = EventCorefDataset(config.dev_file + '.lang-russian', config)
    test_set_russian = EventCorefDataset(config.test_file + '.lang-russian', config)

    chinese_set = EventCorefDataset(config.chinese_file, config)
    korean_set = EventCorefDataset(config.korean_file, config)
    russian_set = EventCorefDataset(config.russian_file, config)

    train_set.numberize(tokenizer, vocabs)

    dev_set_chinese.numberize(tokenizer, vocabs)
    test_set_chinese.numberize(tokenizer, vocabs)

    dev_set_korean.numberize(tokenizer, vocabs)
    test_set_korean.numberize(tokenizer, vocabs)

    dev_set_russian.numberize(tokenizer, vocabs)
    test_set_russian.numberize(tokenizer, vocabs)

    chinese_set.numberize(tokenizer, vocabs)
    korean_set.numberize(tokenizer, vocabs)
    russian_set.numberize(tokenizer, vocabs)

    batch_num = len(train_set) // config.batch_size

    dev_batch_num_chinese = len(dev_set_chinese) // config.eval_batch_size + \
                            (len(dev_set_chinese) % config.eval_batch_size != 0)
    test_batch_num_chinese = len(test_set_chinese) // config.eval_batch_size + \
                             (len(test_set_chinese) % config.eval_batch_size != 0)

    dev_batch_num_korean = len(dev_set_korean) // config.eval_batch_size + \
                           (len(dev_set_korean) % config.eval_batch_size != 0)
    test_batch_num_korean = len(test_set_korean) // config.eval_batch_size + \
                            (len(test_set_korean) % config.eval_batch_size != 0)

    dev_batch_num_russian = len(dev_set_russian) // config.eval_batch_size + \
                            (len(dev_set_russian) % config.eval_batch_size != 0)
    test_batch_num_russian = len(test_set_russian) // config.eval_batch_size + \
                             (len(test_set_russian) % config.eval_batch_size != 0)
    chinese_batch_num = len(chinese_set) // config.eval_batch_size + \
                        (len(chinese_set) % config.eval_batch_size != 0)
    korean_batch_num = len(korean_set) // config.eval_batch_size + \
                       (len(korean_set) % config.eval_batch_size != 0)
    russian_batch_num = len(russian_set) // config.eval_batch_size + \
                        (len(russian_set) % config.eval_batch_size != 0)

    # initialize the model
    model = EventCorefModel(config, vocabs)
    model.cuda()

    init_coref_model_save_fpath = os.path.join('logs/coref/coref-model.mdl') + '.lang-chinese'
    if os.path.exists(init_coref_model_save_fpath):
        init_coref_ckpt = torch.load(init_coref_model_save_fpath)
        model.load_state_dict(init_coref_ckpt['model'])
        print_and_log('Initialized Event Coref model with checkpoint at {}'.format(init_coref_model_save_fpath), logger)
        del init_coref_ckpt
        torch.cuda.empty_cache()

    # print trainable weights
    n_trainable_params, n_nontrainable_params = 0, 0
    for p in model.parameters():
        n_params = torch.prod(torch.tensor(p.shape)).item()
        if p.requires_grad:
            n_trainable_params += n_params
        else:
            n_nontrainable_params += n_params
    logger.info('-' * 100)
    logger.info('> trainable params:')
    for name, param in model.named_parameters():
        if param.requires_grad:
            logger.info('>>> {0}: {1}'.format(name, param.shape))
    logger.info(
        'n_trainable_params: {0}, n_nontrainable_params: {1}'.format(n_trainable_params, n_nontrainable_params))
    logger.info('-' * 100)

    # optimizer
    param_groups = [
        {
            'params': [p for n, p in model.named_parameters() if n.startswith('xlmr')],
            'lr': config.xlmr_learning_rate, 'weight_decay': config.xlmr_weight_decay
        },
        {
            'params': [p for n, p in model.named_parameters() if not n.startswith('xlmr')
                       and 'crf' not in n],
            'lr': config.learning_rate, 'weight_decay': config.weight_decay
        },
        {
            'params': [p for n, p in model.named_parameters() if not n.startswith('xlmr')
                       and ('crf' in n)],
            'lr': config.learning_rate, 'weight_decay': 0
        }
    ]
    optimizer = AdamW(params=param_groups)
    schedule = get_linear_schedule_with_warmup(optimizer,
                                               num_warmup_steps=batch_num * 5,
                                               num_training_steps=batch_num * 100)

    # model state
    state = dict(model=model.state_dict(),
                 event_hidden_num=config.event_hidden_num,
                 role_hidden_num=config.role_hidden_num)

    global_step = 0

    best_english_dev_chinese = {
        'epoch': 0,
        'template': {'f1': 0},
        'slot': {'f1': 0},
        'combined': 0
    }
    best_english_test_chinese = {
        'epoch': 0,
        'template': {'f1': 0},
        'slot': {'f1': 0},
        'combined': 0
    }
    best_sample_chinese = {
        'epoch': 0,
        'template': {'f1': 0},
        'slot': {'f1': 0},
        'combined': 0
    }

    best_english_dev_korean = {
        'epoch': 0,
        'template': {'f1': 0},
        'slot': {'f1': 0},
        'combined': 0
    }
    best_english_test_korean = {
        'epoch': 0,
        'template': {'f1': 0},
        'slot': {'f1': 0},
        'combined': 0
    }
    best_sample_korean = {
        'epoch': 0,
        'template': {'f1': 0},
        'slot': {'f1': 0},
        'combined': 0
    }

    best_english_dev_russian = {
        'epoch': 0,
        'template': {'f1': 0},
        'slot': {'f1': 0},
        'combined': 0
    }
    best_english_test_russian = {
        'epoch': 0,
        'template': {'f1': 0},
        'slot': {'f1': 0},
        'combined': 0
    }
    best_sample_russian = {
        'epoch': 0,
        'template': {'f1': 0},
        'slot': {'f1': 0},
        'combined': 0
    }

    dev_infile_chinese = os.path.join('logs/slot/results.slot.analysis.bp.json.lang-chinese.best')
    test_infile_chinese = os.path.join('logs/slot/results.slot.devtest.bp.json.lang-chinese.best')

    dev_infile_korean = os.path.join('logs/slot/results.slot.analysis.bp.json.lang-korean.best')
    test_infile_korean = os.path.join('logs/slot/results.slot.devtest.bp.json.lang-korean.best')

    dev_infile_russian = os.path.join('logs/slot/results.slot.analysis.bp.json.lang-russian.best')
    test_infile_russian = os.path.join('logs/slot/results.slot.devtest.bp.json.lang-russian.best')

    chinese_infile = os.path.join('logs/slot/results.slot.chinese-sample.bp.json.best')
    korean_infile = os.path.join('logs/slot/results.slot.korean-sample.bp.json.best')
    russian_infile = os.path.join('logs/slot/results.slot.russian-sample.bp.json.best')

    infile_eventid2templateid = {
        dev_infile_chinese: create_event2template_mapping(dev_infile_chinese),
        test_infile_chinese: create_event2template_mapping(test_infile_chinese),
        dev_infile_korean: create_event2template_mapping(dev_infile_korean),
        test_infile_korean: create_event2template_mapping(test_infile_korean),
        dev_infile_russian: create_event2template_mapping(dev_infile_russian),
        test_infile_russian: create_event2template_mapping(test_infile_russian),
        chinese_infile: create_event2template_mapping(chinese_infile),
        korean_infile: create_event2template_mapping(korean_infile),
        russian_infile: create_event2template_mapping(russian_infile),
    }

    dev_sysfile_chinese = os.path.join(log_dir,
                                       'results.{}.analysis.bp.json'.format(config.model_name)) + '.lang-chinese'
    test_sysfile_chinese = os.path.join(log_dir,
                                        'results.{}.devtest.bp.json'.format(config.model_name)) + '.lang-chinese'

    dev_sysfile_korean = os.path.join(log_dir,
                                      'results.{}.analysis.bp.json'.format(config.model_name)) + '.lang-korean'
    test_sysfile_korean = os.path.join(log_dir,
                                       'results.{}.devtest.bp.json'.format(config.model_name)) + '.lang-korean'

    dev_sysfile_russian = os.path.join(log_dir,
                                       'results.{}.analysis.bp.json'.format(config.model_name)) + '.lang-russian'
    test_sysfile_russian = os.path.join(log_dir,
                                        'results.{}.devtest.bp.json'.format(config.model_name)) + '.lang-russian'

    chinese_sysfile = os.path.join(log_dir, 'results.{}.chinese-sample.bp.json'.format(config.model_name))
    korean_sysfile = os.path.join(log_dir, 'results.{}.korean-sample.bp.json'.format(config.model_name))
    russian_sysfile = os.path.join(log_dir, 'results.{}.russian-sample.bp.json'.format(config.model_name))

    dev_reffile = GRANULAR_ANALYSIS_REF_P3
    test_reffile = GRANULAR_DEVTEST_REF_P3
    chinese_reffile = SAMPLE_GRANULAR_CHINESE_P3
    korean_reffile = SAMPLE_GRANULAR_KOREAN_P3
    russian_reffile = SAMPLE_GRANULAR_RUSSIAN_P3

    for epoch in range(config.max_epoch):
        torch.cuda.empty_cache()
        print_and_log('*' * 30, logger)
        print_and_log('Epoch: {}'.format(epoch), logger)
        print_and_log('command: {}'.format(command), logger)
        # training set
        progress = tqdm.tqdm(total=batch_num, ncols=75,
                             desc='Train {}'.format(epoch))
        model.train()
        optimizer.zero_grad()
        losses = []
        for batch_idx, batch in enumerate(DataLoader(
                train_set, batch_size=config.batch_size // config.accumulate_step,
                shuffle=True, collate_fn=train_set.collate_fn)):

            loss = model(batch)
            loss = loss * (1 / config.accumulate_step)

            loss.backward()

            if (batch_idx + 1) % config.accumulate_step == 0:
                progress.update(1)
                global_step += 1
                torch.nn.utils.clip_grad_norm_(
                    model.parameters(), config.grad_clipping)
                optimizer.step()
                schedule.step()
                optimizer.zero_grad()
                print_and_log('{}: step: {}/{}, loss: {}'.format(datetime.now(), batch_idx + 1, batch_num, loss.item()),
                              logger,
                              printout=False)
                losses.append(loss.item())
        progress.close()

        model.eval()
        # dev set (Chinese)
        progress = tqdm.tqdm(total=dev_batch_num_chinese, ncols=75,
                             desc='Dev {}'.format(epoch))

        dev_predictions_chinese = []
        for batch in DataLoader(dev_set_chinese, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=dev_set_chinese.collate_fn):
            progress.update(1)
            dev_predictions_chinese.extend(model.predict(batch))
        progress.close()

        # test set
        progress = tqdm.tqdm(total=test_batch_num_chinese, ncols=75,
                             desc='Test {}'.format(epoch))

        test_predictions_chinese = []
        for batch in DataLoader(test_set_chinese, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=test_set_chinese.collate_fn):
            progress.update(1)
            test_predictions_chinese.extend(model.predict(batch))
        progress.close()

        # dev set (Korean)
        progress = tqdm.tqdm(total=dev_batch_num_korean, ncols=75,
                             desc='Dev {}'.format(epoch))

        dev_predictions_korean = []
        for batch in DataLoader(dev_set_korean, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=dev_set_korean.collate_fn):
            progress.update(1)
            dev_predictions_korean.extend(model.predict(batch))
        progress.close()

        # test set
        progress = tqdm.tqdm(total=test_batch_num_korean, ncols=75,
                             desc='Test {}'.format(epoch))

        test_predictions_korean = []
        for batch in DataLoader(test_set_korean, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=test_set_korean.collate_fn):
            progress.update(1)
            test_predictions_korean.extend(model.predict(batch))
        progress.close()

        # dev set (Russian)
        progress = tqdm.tqdm(total=dev_batch_num_russian, ncols=75,
                             desc='Dev {}'.format(epoch))

        dev_predictions_russian = []
        for batch in DataLoader(dev_set_russian, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=dev_set_russian.collate_fn):
            progress.update(1)
            dev_predictions_russian.extend(model.predict(batch))
        progress.close()

        # test set
        progress = tqdm.tqdm(total=test_batch_num_russian, ncols=75,
                             desc='Test {}'.format(epoch))

        test_predictions_russian = []
        for batch in DataLoader(test_set_russian, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=test_set_russian.collate_fn):
            progress.update(1)
            test_predictions_russian.extend(model.predict(batch))
        progress.close()

        # chinese test set
        progress = tqdm.tqdm(total=chinese_batch_num, ncols=75,
                             desc='chinese {}'.format(epoch))

        chinese_coref_predictions = []
        for batch in DataLoader(chinese_set, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=chinese_set.collate_fn):
            progress.update(1)
            chinese_coref_predictions.extend(model.predict(batch))
        progress.close()

        # korean test set
        progress = tqdm.tqdm(total=korean_batch_num, ncols=75,
                             desc='korean {}'.format(epoch))

        korean_coref_predictions = []
        for batch in DataLoader(korean_set, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=korean_set.collate_fn):
            progress.update(1)
            korean_coref_predictions.extend(model.predict(batch))
        progress.close()

        # russian test set
        progress = tqdm.tqdm(total=russian_batch_num, ncols=75,
                             desc='russian {}'.format(epoch))

        russian_coref_predictions = []
        for batch in DataLoader(russian_set, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=russian_set.collate_fn):
            progress.update(1)
            russian_coref_predictions.extend(model.predict(batch))
        progress.close()

        for strategy in [1]:

            write_coref_predictions(
                dev_predictions_chinese,
                infile_eventid2templateid[dev_infile_chinese],
                infile=dev_infile_chinese,
                output_fpath=dev_sysfile_chinese + '.epoch-{}'.format(epoch) + '.strategy-{}'.format(strategy),
                strategy=strategy
            )
            dev_score_chinese = compute_granular_score(
                reffile=dev_reffile,
                sysfile=dev_sysfile_chinese + '.epoch-{}'.format(epoch) + '.strategy-{}'.format(strategy),
                csv_fpath=os.path.join(log_dir, 'analysis.coref.report.epoch-{}.csv'.format(epoch))
            )

            write_coref_predictions(
                test_predictions_chinese,
                infile_eventid2templateid[test_infile_chinese],
                infile=test_infile_chinese,
                output_fpath=test_sysfile_chinese + '.epoch-{}'.format(epoch) + '.strategy-{}'.format(strategy),
                strategy=strategy
            )

            test_score_chinese = compute_granular_score(
                reffile=test_reffile,
                sysfile=test_sysfile_chinese + '.epoch-{}'.format(epoch) + '.strategy-{}'.format(strategy),
                csv_fpath=os.path.join(log_dir, 'devtest.coref.report.epoch-{}.csv'.format(epoch))
            )

            # ===============

            write_coref_predictions(
                dev_predictions_korean,
                infile_eventid2templateid[dev_infile_korean],
                infile=dev_infile_korean,
                output_fpath=dev_sysfile_korean + '.epoch-{}'.format(epoch) + '.strategy-{}'.format(strategy),
                strategy=strategy
            )
            dev_score_korean = compute_granular_score(
                reffile=dev_reffile,
                sysfile=dev_sysfile_korean + '.epoch-{}'.format(epoch) + '.strategy-{}'.format(strategy),
                csv_fpath=os.path.join(log_dir, 'analysis.coref.report.epoch-{}.csv'.format(epoch))
            )

            write_coref_predictions(
                test_predictions_korean,
                infile_eventid2templateid[test_infile_korean],
                infile=test_infile_korean,
                output_fpath=test_sysfile_korean + '.epoch-{}'.format(epoch) + '.strategy-{}'.format(strategy),
                strategy=strategy
            )

            test_score_korean = compute_granular_score(
                reffile=test_reffile,
                sysfile=test_sysfile_korean + '.epoch-{}'.format(epoch) + '.strategy-{}'.format(strategy),
                csv_fpath=os.path.join(log_dir, 'devtest.coref.report.epoch-{}.csv'.format(epoch))
            )

            # ===============

            write_coref_predictions(
                dev_predictions_russian,
                infile_eventid2templateid[dev_infile_russian],
                infile=dev_infile_russian,
                output_fpath=dev_sysfile_russian + '.epoch-{}'.format(epoch) + '.strategy-{}'.format(strategy),
                strategy=strategy
            )
            dev_score_russian = compute_granular_score(
                reffile=dev_reffile,
                sysfile=dev_sysfile_russian + '.epoch-{}'.format(epoch) + '.strategy-{}'.format(strategy),
                csv_fpath=os.path.join(log_dir, 'analysis.coref.report.epoch-{}.csv'.format(epoch))
            )

            write_coref_predictions(
                test_predictions_russian,
                infile_eventid2templateid[test_infile_russian],
                infile=test_infile_russian,
                output_fpath=test_sysfile_russian + '.epoch-{}'.format(epoch) + '.strategy-{}'.format(strategy),
                strategy=strategy
            )

            test_score_russian = compute_granular_score(
                reffile=test_reffile,
                sysfile=test_sysfile_russian + '.epoch-{}'.format(epoch) + '.strategy-{}'.format(strategy),
                csv_fpath=os.path.join(log_dir, 'devtest.coref.report.epoch-{}.csv'.format(epoch))
            )

            write_coref_predictions(
                chinese_coref_predictions,
                infile_eventid2templateid[chinese_infile],
                infile=chinese_infile,
                output_fpath=chinese_sysfile + '.epoch-{}'.format(epoch) + '.strategy-{}'.format(strategy),
                strategy=strategy
            )

            chinese_score = compute_granular_score(
                reffile=chinese_reffile,
                sysfile=chinese_sysfile + '.epoch-{}'.format(epoch) + '.strategy-{}'.format(strategy),
                csv_fpath=os.path.join(log_dir, 'chinese.coref.report.epoch-{}.csv'.format(epoch))
            )
            # ------------
            write_coref_predictions(
                korean_coref_predictions,
                infile_eventid2templateid[korean_infile],
                infile=korean_infile,
                output_fpath=korean_sysfile + '.epoch-{}'.format(epoch) + '.strategy-{}'.format(strategy),
                strategy=strategy
            )

            korean_score = compute_granular_score(
                reffile=korean_reffile,
                sysfile=korean_sysfile + '.epoch-{}'.format(epoch) + '.strategy-{}'.format(strategy),
                csv_fpath=os.path.join(log_dir, 'korean.coref.report.epoch-{}.csv'.format(epoch))
            )
            # ------------
            write_coref_predictions(
                russian_coref_predictions,
                infile_eventid2templateid[russian_infile],
                infile=russian_infile,
                output_fpath=russian_sysfile + '.epoch-{}'.format(epoch) + '.strategy-{}'.format(strategy),
                strategy=strategy
            )

            russian_score = compute_granular_score(
                reffile=russian_reffile,
                sysfile=russian_sysfile + '.epoch-{}'.format(epoch) + '.strategy-{}'.format(strategy),
                csv_fpath=os.path.join(log_dir, 'russian.coref.report.epoch-{}.csv'.format(epoch))
            )

            dev_score_chinese['stategy'] = strategy
            test_score_chinese['stategy'] = strategy

            dev_score_korean['stategy'] = strategy
            test_score_korean['stategy'] = strategy

            dev_score_russian['stategy'] = strategy
            test_score_russian['stategy'] = strategy

            chinese_score['stategy'] = strategy
            korean_score['stategy'] = strategy
            russian_score['stategy'] = strategy

            dev_score_chinese['epoch'] = epoch
            test_score_chinese['epoch'] = epoch

            dev_score_korean['epoch'] = epoch
            test_score_korean['epoch'] = epoch

            dev_score_russian['epoch'] = epoch
            test_score_russian['epoch'] = epoch

            chinese_score['epoch'] = epoch
            korean_score['epoch'] = epoch
            russian_score['epoch'] = epoch

            print_and_log('=' * 50, logger)
            print_and_log('Loss: {:.3f}'.format(sum(losses) / len(losses)), logger)
            print_and_log('-' * 20, logger)
            print_and_log('Dev Chinese: {}'.format(dev_score_chinese), logger, printout=not config.docker_run)
            print_and_log('Test Chinese: {}'.format(test_score_chinese), logger, printout=not config.docker_run)

            print_and_log('Dev Korean: {}'.format(dev_score_korean), logger, printout=not config.docker_run)
            print_and_log('Test Korean: {}'.format(test_score_korean), logger, printout=not config.docker_run)

            print_and_log('Dev Russian: {}'.format(dev_score_russian), logger, printout=not config.docker_run)
            print_and_log('Test Russian: {}'.format(test_score_russian), logger, printout=not config.docker_run)

            print_and_log('Sample Chinese: {}'.format(chinese_score), logger, printout=not config.docker_run)
            print_and_log('Sample Korean: {}'.format(korean_score), logger, printout=not config.docker_run)
            print_and_log('Sample Russian: {}'.format(russian_score), logger, printout=not config.docker_run)

            if epoch == 0 or (dev_score_chinese['combined'] + chinese_score['combined']) > (
                    best_english_dev_chinese['combined'] + best_sample_chinese['combined']):
                best_english_dev_chinese = dev_score_chinese
                best_english_test_chinese = test_score_chinese
                best_sample_chinese = chinese_score
                print_and_log('New best chinese model...', logger)

                torch.save(state, coref_model_save_fpath + '.lang-chinese')

                # save best prediction files
                copyfile(dev_sysfile_chinese + '.epoch-{}'.format(epoch) + '.strategy-{}'.format(strategy),
                         dev_sysfile_chinese + '.best')
                copyfile(test_sysfile_chinese + '.epoch-{}'.format(epoch) + '.strategy-{}'.format(strategy),
                         test_sysfile_chinese + '.best')
                copyfile(chinese_sysfile + '.epoch-{}'.format(epoch) + '.strategy-{}'.format(strategy),
                         chinese_sysfile + '.best')

            if epoch == 0 or (dev_score_korean['combined'] + korean_score['combined']) > (
                    best_english_dev_korean['combined'] + best_sample_korean['combined']):
                best_english_dev_korean = dev_score_korean
                best_english_test_korean = test_score_korean
                best_sample_korean = korean_score
                print_and_log('New best korean model...', logger)

                torch.save(state, coref_model_save_fpath + '.lang-korean')

                # save best prediction files
                copyfile(dev_sysfile_korean + '.epoch-{}'.format(epoch) + '.strategy-{}'.format(strategy),
                         dev_sysfile_korean + '.best')
                copyfile(test_sysfile_korean + '.epoch-{}'.format(epoch) + '.strategy-{}'.format(strategy),
                         test_sysfile_korean + '.best')
                copyfile(korean_sysfile + '.epoch-{}'.format(epoch) + '.strategy-{}'.format(strategy),
                         korean_sysfile + '.best')

            if epoch == 0 or (dev_score_russian['combined'] + russian_score['combined']) > (
                    best_english_dev_russian['combined'] + best_sample_russian['combined']):
                best_english_dev_russian = dev_score_russian
                best_english_test_russian = test_score_russian
                best_sample_russian = russian_score
                print_and_log('New best russian model...', logger)

                torch.save(state, coref_model_save_fpath + '.lang-russian')

                # save best prediction files
                copyfile(dev_sysfile_russian + '.epoch-{}'.format(epoch) + '.strategy-{}'.format(strategy),
                         dev_sysfile_russian + '.best')
                copyfile(test_sysfile_russian + '.epoch-{}'.format(epoch) + '.strategy-{}'.format(strategy),
                         test_sysfile_russian + '.best')
                copyfile(russian_sysfile + '.epoch-{}'.format(epoch) + '.strategy-{}'.format(strategy),
                         russian_sysfile + '.best')

        print_and_log('=' * 50, logger, printout=not config.docker_run)
        print_and_log('Current best result:', logger, printout=not config.docker_run)
        print_and_log('-' * 20, logger, printout=not config.docker_run)
        print_and_log('Dev for Chinese: {}'.format(best_english_dev_chinese), logger, printout=not config.docker_run)
        print_and_log('Test for Chinese: {}'.format(best_english_test_chinese), logger, printout=not config.docker_run)
        print_and_log('Chinese: {}'.format(best_sample_chinese), logger, printout=not config.docker_run)
        print_and_log('-' * 20, logger, printout=not config.docker_run)
        print_and_log('Dev for Korean: {}'.format(best_english_dev_korean), logger, printout=not config.docker_run)
        print_and_log('Test for Korean: {}'.format(best_english_test_korean), logger, printout=not config.docker_run)
        print_and_log('Korean: {}'.format(best_sample_korean), logger, printout=not config.docker_run)
        print_and_log('-' * 20, logger, printout=not config.docker_run)
        print_and_log('Dev for Russian: {}'.format(best_english_dev_russian), logger, printout=not config.docker_run)
        print_and_log('Test for Russian: {}'.format(best_english_test_russian), logger, printout=not config.docker_run)
        print_and_log('Russian: {}'.format(best_sample_russian), logger, printout=not config.docker_run)

elif config.model_name.startswith('pipeline'):
    nlp = Pipeline('english', cache_dir='resource/trankit')
    nlp.add('chinese')
    nlp.add('traditional-chinese')
    nlp.add('korean')
    nlp.add('russian')

    # process test file
    # real test file without any annotation
    if config.docker_run:
        original_test_fpath = '/app/test_data.bp.json'
        config.output_file = '/app/results.json'
    else:
        original_test_fpath = os.path.join(WORKING_DIR, 'app/test_data.bp.json')
        config.output_file = 'app/results.json'

    # split test file into 3 separate files, each for 1 language
    lang2testfile = split_original_testfile(original_test_fpath)

    # run the following as usual
    # merge the predictions into 1 file in the end
    results_files = []
    for test_lang, test_raw_fpath in lang2testfile.items():
        trigger_model_save_fpath = os.path.join('logs/trg/trigger-model.mdl') + '.lang-{}'.format(test_lang)
        argument_model_save_fpath = os.path.join('logs/arg/argument-model.mdl') + '.lang-{}'.format(test_lang)
        template_model_save_fpath = os.path.join('logs/template/template-model.mdl') + '.lang-{}'.format(test_lang)
        slot_model_save_fpath = os.path.join('logs/slot/slot-model.mdl') + '.lang-{}'.format(test_lang)
        coref_model_save_fpath = os.path.join('logs/coref/coref-model.mdl') + '.lang-{}'.format(test_lang)

        ensure_dir('logs/pipeline')
        trigger_test_sysfile = 'logs/pipeline/results.trigger.test.bp.json' + '.lang-{}'.format(test_lang)
        argument_test_sysfile = 'logs/pipeline/results.argument.test.bp.json' + '.lang-{}'.format(test_lang)
        template_test_sysfile = 'logs/pipeline/results.template.test.bp.json' + '.lang-{}'.format(test_lang)
        slot_test_sysfile = 'logs/pipeline/results.slot.test.bp.json' + '.lang-{}'.format(test_lang)
        coref_test_sysfile = 'logs/pipeline/results.json' + '.lang-{}'.format(test_lang)

        trigger_test_processed_fpath = 'logs/pipeline/trigger.granular.test.{}.jsonl'.format(test_lang)
        read_and_build(nlp, test_raw_fpath,
                       trigger_test_processed_fpath,
                       lang=test_lang, model_name='trg')
        # -----------------
        trigger_test_set = TriggerDataset(trigger_test_processed_fpath, config)

        trigger_test_set.numberize(tokenizer, vocabs)

        trigger_test_batch_num = len(trigger_test_set) // config.eval_batch_size + \
                                 (len(trigger_test_set) % config.eval_batch_size != 0)

        # initialize the model
        trigger_ckpt = torch.load(trigger_model_save_fpath)
        config.event_hidden_num = trigger_ckpt['event_hidden_num']
        config.role_hidden_num = trigger_ckpt['role_hidden_num']
        trigger_model = TriggerModel(config, vocabs)
        trigger_model.cuda()
        trigger_model.load_state_dict(trigger_ckpt['model'])
        trigger_model.eval()
        torch.cuda.empty_cache()
        # test set
        progress = tqdm.tqdm(total=trigger_test_batch_num, ncols=75,
                             desc='Trigger')

        trigger_test_pred_graphs = []
        for batch in DataLoader(trigger_test_set, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=trigger_test_set.collate_fn):
            progress.update(1)
            graphs = trigger_model.predict(batch)
            trigger_test_pred_graphs.extend(graphs)
        progress.close()
        write_trigger_predictions(
            trigger_test_pred_graphs,
            output_fpath=trigger_test_sysfile,
            infile=test_raw_fpath
        )
        print('writing trigger predictions to {}... Done!'.format(trigger_test_sysfile))

        ############################################################################################
        del trigger_model
        del trigger_ckpt
        torch.cuda.empty_cache()

        argument_test_processed_fpath = 'logs/pipeline/argument.granular.test.{}.bp.jsonl'.format(test_lang)
        read_and_build(nlp, trigger_test_sysfile,
                       argument_test_processed_fpath,
                       lang=test_lang, model_name='arg')
        argument_test_set = ArgumentDataset(argument_test_processed_fpath, config, eval=True)

        argument_test_set.numberize(tokenizer, vocabs)

        argument_test_batch_num = len(argument_test_set) // config.eval_batch_size + \
                                  (len(argument_test_set) % config.eval_batch_size != 0)

        # initialize the model
        argument_ckpt = torch.load(argument_model_save_fpath)
        config.event_hidden_num = argument_ckpt['event_hidden_num']
        config.role_hidden_num = argument_ckpt['role_hidden_num']
        argument_model = ArgumentModel(config, vocabs)
        argument_model.cuda()
        argument_model.load_state_dict(argument_ckpt['model'])
        argument_model.eval()
        torch.cuda.empty_cache()
        # test set
        progress = tqdm.tqdm(total=argument_test_batch_num, ncols=75,
                             desc='Argument')

        argument_test_pred_graphs = []
        for batch in DataLoader(argument_test_set, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=argument_test_set.collate_fn):
            progress.update(1)
            graphs = argument_model.predict(batch)
            argument_test_pred_graphs.extend(graphs)
        progress.close()
        write_arg_predictions(
            argument_test_pred_graphs,
            output_fpath=argument_test_sysfile,
            infile=trigger_test_sysfile
        )
        remove_pronouns(argument_test_sysfile, argument_test_sysfile)
        normalize_money_spans(argument_test_sysfile, argument_test_sysfile)
        print('writing argument predictions to {}... Done!'.format(argument_test_sysfile))
        ############################################################################################
        del argument_model
        del argument_ckpt
        torch.cuda.empty_cache()

        template_test_processed_fpath = 'logs/pipeline/template.granular.test.{}.bp.jsonl'.format(test_lang)
        read_and_build(nlp, argument_test_sysfile,
                       template_test_processed_fpath,
                       lang=test_lang, model_name='template')

        template_test_set = TemplateDataset(template_test_processed_fpath, config)
        template_test_set.numberize(tokenizer, vocabs)
        template_test_batch_num = len(template_test_set) // config.eval_batch_size + \
                                  (len(template_test_set) % config.eval_batch_size != 0)

        template_ckpt = torch.load(template_model_save_fpath)
        config.event_hidden_num = template_ckpt['event_hidden_num']
        config.role_hidden_num = template_ckpt['role_hidden_num']
        template_model = TemplateModel(config, vocabs)
        template_model.cuda()
        template_model.load_state_dict(template_ckpt['model'])
        template_model.eval()
        torch.cuda.empty_cache()

        # test set
        progress = tqdm.tqdm(total=template_test_batch_num, ncols=75,
                             desc='Template')

        template_predictions = []
        for batch in DataLoader(template_test_set, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=template_test_set.collate_fn):
            progress.update(1)
            graphs = template_model.predict(batch)
            template_predictions.extend(graphs)
        progress.close()
        write_template_predictions(
            template_predictions,
            argument_test_sysfile,
            output_fpath=template_test_sysfile
        )
        print('writing template predictions to {}... Done!'.format(template_test_sysfile))
        ############################################################################################
        del template_model
        del template_ckpt
        torch.cuda.empty_cache()

        slot_test_processed_fpath = 'logs/pipeline/slot.granular.test.{}.bp.jsonl'.format(test_lang)
        read_and_build(nlp, template_test_sysfile,
                       slot_test_processed_fpath,
                       lang=test_lang, model_name='slot')

        slot_test_set = SlotDataset(slot_test_processed_fpath, config)
        slot_test_set.numberize(tokenizer, vocabs)
        slot_test_batch_num = len(slot_test_set) // config.eval_batch_size + \
                              (len(slot_test_set) % config.eval_batch_size != 0)

        slot_ckpt = torch.load(slot_model_save_fpath)
        config.event_hidden_num = slot_ckpt['event_hidden_num']
        config.role_hidden_num = slot_ckpt['role_hidden_num']
        slot_model = SlotModel(config, vocabs)
        slot_model.cuda()
        slot_model.load_state_dict(slot_ckpt['model'])
        slot_model.eval()
        torch.cuda.empty_cache()

        # test set
        progress = tqdm.tqdm(total=slot_test_batch_num, ncols=75,
                             desc='Slot')

        slot_predictions = []
        for batch in DataLoader(slot_test_set, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=slot_test_set.collate_fn):
            progress.update(1)
            graphs = slot_model.predict(batch)
            slot_predictions.extend(graphs)
        progress.close()
        write_slot_predictions(
            slot_predictions,
            template_test_sysfile,
            output_fpath=slot_test_sysfile
        )
        print('writing slot predictions to {}... Done!'.format(slot_test_sysfile))

        ############################################################################################
        del slot_model
        del slot_ckpt
        torch.cuda.empty_cache()

        infile_eventid2templateid = {
            slot_test_sysfile: create_event2template_mapping(fpath=slot_test_sysfile)
        }

        coref_test_processed_fpath = 'logs/pipeline/coref.granular.test.{}.bp.jsonl'.format(test_lang)
        read_and_build(nlp, slot_test_sysfile,
                       coref_test_processed_fpath,
                       lang=test_lang, model_name='coref')
        coref_test_set = EventCorefDataset(coref_test_processed_fpath, config)

        coref_test_set.numberize(tokenizer, vocabs)

        coref_test_batch_num = len(coref_test_set) // config.eval_batch_size + \
                               (len(coref_test_set) % config.eval_batch_size != 0)

        # initialize the model
        coref_ckpt = torch.load(coref_model_save_fpath)
        config.event_hidden_num = coref_ckpt['event_hidden_num']
        config.role_hidden_num = coref_ckpt['role_hidden_num']
        coref_model = EventCorefModel(config, vocabs)
        coref_model.cuda()
        coref_model.load_state_dict(coref_ckpt['model'])
        coref_model.eval()
        torch.cuda.empty_cache()
        # test set
        progress = tqdm.tqdm(total=coref_test_batch_num, ncols=75,
                             desc='Coref')

        coref_predictions = []
        for batch in DataLoader(coref_test_set, batch_size=config.eval_batch_size,
                                shuffle=False, collate_fn=coref_test_set.collate_fn):
            progress.update(1)
            preds = coref_model.predict(batch)
            coref_predictions.extend(preds)
        progress.close()
        write_coref_predictions(
            coref_predictions,
            infile_eventid2templateid[slot_test_sysfile],
            infile=slot_test_sysfile,
            output_fpath=coref_test_sysfile,
            strategy=1
        )
        print('writing coreference predictions to {}... Done!'.format(coref_test_sysfile))
        results_files.append(coref_test_sysfile)
        if not config.docker_run:
            test_score = compute_granular_score(
                reffile=test_raw_fpath,
                sysfile=coref_test_sysfile,
                csv_fpath='logs/pipeline/final_scores.csv' + '.lang-{}'.format(test_lang)
            )
            print(test_score)

    merge_predictions(results_files, output_file=config.output_file)
