import os
import sys
import json
import glob
from copy import deepcopy
import numpy as np
from collections import defaultdict
import random, re
from trankit import Pipeline
import torch
from shutil import copyfile
from constants import *
import hanzidentifier

WORKING_DIR = os.path.dirname(os.path.realpath(__file__))
PRECOMPUTED_DIR = os.path.join(WORKING_DIR, 'resource/precomputed')

ignore_sections = {'Byline', 'Dateline', 'Story-Lead', 'Section_Header'}

INFO_LEVEL = {
    'duration-mention': 0,
    'pronoun': 1,
    'nominal': 2,
    'name': 3
}

if os.path.exists('vocabs.json'):
    with open('vocabs.json') as f:
        vocabs = json.load(f)
else:
    vocabs = {}

vocabs['upos'] = {"[PAD]": 0, "[UNK]": 1, 'ADJ': 2, 'ADP': 3, 'ADV': 4, 'AUX': 5, 'CCONJ': 6, 'DET': 7, 'INTJ': 8,
                  'NOUN': 9, 'NUM': 10, 'PART': 11, 'PRON': 12, 'PROPN': 13, 'PUNCT': 14, 'SCONJ': 15, 'SYM': 16,
                  'VERB': 17, 'X': 18}

vocabs['deprel'] = {"[PAD]": 0, "[UNK]": 1, 'acl': 2, 'advcl': 3, 'advmod': 4, 'amod': 5, 'appos': 6, 'aux': 7,
                    'case': 8, 'cc': 9, 'ccomp': 10, 'compound': 11, 'conj': 12, 'cop': 13, 'csubj': 14, 'dep': 15,
                    'det': 16, 'discourse': 17, 'dislocated': 18, 'expl': 19, 'fixed': 20, 'flat': 21, 'goeswith': 22,
                    'iobj': 23, 'list': 24, 'mark': 25, 'nmod': 26, 'nsubj': 27, 'nummod': 28, 'obj': 29, 'obl': 30,
                    'orphan': 31, 'parataxis': 32, 'punct': 33, 'reparandum': 34, 'root': 35, 'vocative': 36,
                    'xcomp': 37}


def ensure_dir(dir_path):
    os.makedirs(dir_path, exist_ok=True)


ensure_dir(PRECOMPUTED_DIR)


def normalize_money_spans(in_file, out_file):
    with open(in_file) as f:
        data = json.load(f)
    entries = data['entries']
    removed = 0
    for _, entry in entries.items():
        event_set = entry['annotation-sets']['basic-events']['events']
        span_sets = entry['annotation-sets']['basic-events']['span-sets']
        for _, event in event_set.items():
            money_ssids = deepcopy(event['money'])
            selected = []
            for mssid in money_ssids:
                if '%' not in span_sets[mssid]['spans'][0]['string']:
                    selected.append(mssid)
                else:
                    removed += 1
            event['money'] = selected

    print('removing {} relative money arguments'.format(removed))

    with open(out_file, 'w') as f:
        json.dump(data, f, ensure_ascii=False)


def split_original_testfile(test_fpath):
    with open(test_fpath) as f:
        test_corpus = json.load(f)

    lang2corpus = {
        'korean': deepcopy(test_corpus),
        'russian': deepcopy(test_corpus),
        'chinese': deepcopy(test_corpus)
    }

    lang2corpus['korean']['entries'] = {}
    lang2corpus['russian']['entries'] = {}
    lang2corpus['chinese']['entries'] = {}

    for entry_id, entry in test_corpus['entries'].items():
        if entry['lang'] == 'kor':
            lang2corpus['korean']['entries'][entry_id] = entry
        elif entry['lang'] == 'rus':
            lang2corpus['russian']['entries'][entry_id] = entry
        else:
            lang2corpus['chinese']['entries'][entry_id] = entry

    lang2testfile = {
        'korean': os.path.join(WORKING_DIR, 'test_data.korean.bp.json'),
        'russian': os.path.join(WORKING_DIR, 'test_data.russian.bp.json'),
        'chinese': os.path.join(WORKING_DIR, 'test_data.chinese.bp.json')
    }
    for test_lang, test_raw_fpath in lang2testfile.items():
        with open(test_raw_fpath, 'w') as f:
            json.dump(lang2corpus[test_lang], f, ensure_ascii=False)

    return lang2testfile


def merge_predictions(results_files, output_file):
    assert len(results_files) > 0

    with open(results_files[0]) as f:
        merged_results = json.load(f)

    for fpath in results_files[1:]:
        with open(fpath) as f:
            results = json.load(f)

        for entry_id, entry in results['entries'].items():
            merged_results['entries'][entry_id] = entry

    with open(output_file, 'w') as f:
        json.dump(merged_results, f, ensure_ascii=False)
    print('-' * 20)
    print('Final results are written to {}!'.format(output_file))


class UnionFind:
    def __init__(self, size):
        self.parent = list(range(size))
        self.rank = [1] * size

    def find(self, x):
        if x == self.parent[x]:
            return x
        self.parent[x] = self.find(self.parent[x])
        return self.parent[x]

    def union(self, x, y):
        root_x = self.find(x)
        root_y = self.find(y)

        if root_x != root_y:
            if self.rank[root_x] > self.rank[root_y]:
                self.parent[root_y] = root_x
            elif self.rank[root_y] > self.rank[root_x]:
                self.parent[root_x] = root_y
            else:
                self.parent[root_x] = root_y
                self.rank[root_y] += 1


def create_event2template_mapping(fpath):
    with open(fpath) as f:
        data = json.load(f)

    mapping = defaultdict(dict)

    for entry_id, entry in data['entries'].items():
        if 'granular-templates' not in entry['annotation-sets']['basic-events']:
            mapping[entry_id] = {}
            continue

        ssid2start_end = {}
        span_sets = entry['annotation-sets']['basic-events']['span-sets']
        for span in span_sets.values():
            ssid = span['ssid']
            assert len(span['spans']) == 1, "reason: this code only works with system outputs, which have this feature"
            for m in span['spans']:
                ssid2start_end[ssid] = '{}-{}'.format(m['start'], m['end'])

        event_sets = entry['annotation-sets']['basic-events']['events']
        start_end2eventid = {}

        for event in event_sets.values():
            trigger_ssid = event['anchors']
            start_end2eventid[ssid2start_end[trigger_ssid]] = event['eventid']

        template_sets = entry['annotation-sets']['basic-events']['granular-templates']
        for template in template_sets.values():
            anchor_ssid = template['template-anchor']
            if anchor_ssid not in ssid2start_end:
                continue

            if ssid2start_end[anchor_ssid] in start_end2eventid:
                mapping[entry_id][start_end2eventid[ssid2start_end[anchor_ssid]]] = template['template-id']

    return mapping


def reorganize_events(data, sentence_predictions, event2template, strategy):
    '''
    sentence_predictions: predictions for all coref, ref events in the same sentence
    '''
    entry = data['entries'][sentence_predictions['entry-id']]
    event_set = entry['annotation-sets']['basic-events']['events']

    # populate ref-events
    ref2eventid = defaultdict(list)
    for pred in sentence_predictions['predictions']:
        head_eventid = pred['head-eventid']
        tail_eventid = pred['tail-eventid']
        score = pred['confidence']
        if pred['coref-type'] == 'RefEvent' and tail_eventid not in event_set[head_eventid]['ref-events']:
            event_set[head_eventid]['ref-events'].append(tail_eventid)
            event_set[head_eventid]['ref-events-confidence-scores'].append(score)

            ref2eventid[tail_eventid].append(head_eventid)

    # create event clusters
    eventid2idx = {}
    idx2eventid = {}
    for i in range(len(sentence_predictions['event_list'])):  # 'event_list' is already sorted by trigger spans
        eventid2idx[sentence_predictions['event_list'][i]] = i
        idx2eventid[i] = sentence_predictions['event_list'][i]

    node_idxs = list(range(len(sentence_predictions['event_list'])))
    num_nodes = len(node_idxs)

    uf = UnionFind(size=num_nodes)
    for pred in sentence_predictions['predictions']:
        if pred['coref-type'] == 'CorefEvent':
            head_eventid = pred['head-eventid']
            tail_eventid = pred['tail-eventid']
            i = eventid2idx[head_eventid]
            j = eventid2idx[tail_eventid]
            if i > j:
                uf.union(i, j)
            else:
                uf.union(j, i)

    head2tails = defaultdict(list)
    tail2head = {}
    for i in range(num_nodes):
        head = uf.find(i)

        head_eventid = idx2eventid[head]
        tail_eventid = idx2eventid[i]

        if tail_eventid != head_eventid:
            head2tails[head_eventid].append(tail_eventid)
            tail2head[tail_eventid] = head_eventid

    for eventid in sentence_predictions['event_list']:
        event = event_set[eventid]

        update_list = []
        score_update_list = []
        for i in range(len(event['ref-events'])):
            parent_eventid1 = idx2eventid[uf.find(eventid2idx[event['eventid']])]
            parent_eventid2 = idx2eventid[uf.find(eventid2idx[event['ref-events'][i]])]
            score_parent_eventid2 = event['ref-events-confidence-scores'][i]
            if parent_eventid1 != parent_eventid2:
                if parent_eventid2 not in update_list:
                    update_list.append(parent_eventid2)
                    score_update_list.append(score_parent_eventid2)

        event['ref-events'] = update_list
        event['ref-events-confidence-scores'] = score_update_list

    # delete duplicated events in 'basic-events' annotations
    for hevid in head2tails:
        head_event = event_set[hevid]
        for tevid in head2tails[hevid]:
            # delete
            del event_set[tevid]

    # delete duplicated slot fillers in 'granular-templates' annotations
    if 'granular-templates' in entry['annotation-sets']['basic-events']:
        template_set = entry['annotation-sets']['basic-events']['granular-templates']
        for template in template_set.values():
            new_template = {}
            for slot_name, fillers in template.items():
                if not isinstance(fillers, list):
                    new_template[slot_name] = template[slot_name]
                else:
                    new_template[slot_name] = []
                    filled_eventids = set()
                    for filler in fillers:
                        if 'event-id' not in filler:
                            new_template[slot_name].append(filler)
                        else:
                            if filler['event-id'] not in tail2head:
                                head_eventid = filler['event-id']
                            else:
                                head_eventid = tail2head[filler['event-id']]
                            filler['event-id'] = head_eventid

                            if filler['event-id'] not in filled_eventids:
                                filled_eventids.add(filler['event-id'])

                                new_template[slot_name].append(filler)
            entry['annotation-sets']['basic-events']['granular-templates'][template['template-id']] = new_template

        # ---------- merge duplicated templates -------------
        # do nothing


def write_coref_predictions(batch_predictions, event2template, infile, output_fpath, strategy):
    with open(infile) as f:
        write_out = json.load(f)

    # clean ref-events fields
    for entry in write_out['entries'].values():
        event_set = entry['annotation-sets']['basic-events']['events']
        for event in event_set.values():
            event['ref-events'] = []
            event['ref-events-confidence-scores'] = []

    # populate ref-events and collapse coref events
    for sent_predictions in batch_predictions:
        if len(sent_predictions['event_list']) == 1:
            continue
        reorganize_events(write_out, sent_predictions, event2template, strategy)

    # avoid missing entry error in the scorer
    with open(infile) as f:
        in_data = json.load(f)
    for entry_id, entry in in_data['entries'].items():
        if entry_id not in write_out['entries']:
            write_out['entries'][entry_id] = entry
    with open(output_fpath, 'w') as f:
        json.dump(write_out, f, ensure_ascii=False)


def write_trigger_predictions(pred_graphs, output_fpath, infile):
    write_out = {
        'corpus-id': 'trigger-output',
        'entries': {}
    }
    for graph in pred_graphs:
        write_entry = graph.write_entry
        if graph.entry_id not in write_out['entries']:
            write_out['entries'][graph.entry_id] = write_entry
            write_entry['segment-sections'] = []
            write_entry['segment-sections'].append(
                {
                    "start": graph.write_entry['start'],
                    "end": graph.write_entry['end'],
                    "structural-element": "Sentence"
                }
            )
        else:
            write_out['entries'][write_entry['entry-id']]['segment-sections'].append(
                {
                    "end": write_entry['end'],
                    "start": write_entry['start'],
                    "structural-element": "Sentence"
                }
            )
            offset_span_id = len(
                write_out['entries'][write_entry['entry-id']]['annotation-sets']['basic-events']['span-sets'])
            offset_event_id = len(
                write_out['entries'][write_entry['entry-id']]['annotation-sets']['basic-events']['events'])

            for event_id, event in write_entry['annotation-sets']['basic-events']['events'].items():
                new_ev_id = f"event-{offset_event_id + get_num_id(event['eventid'])}"
                new_event = {
                    "agents": [],
                    "agents-confidence-scores": [],
                    "money": [],
                    "money-confidence-scores": [],
                    "anchors": f"ss-{offset_span_id + get_num_id(event['anchors'])}",
                    "event-type": event['event-type'],
                    "event-confidence-score": event['event-confidence-score'],
                    "eventid": new_ev_id,
                    "patients": [],
                    "patients-confidence-scores": [],
                    "ref-events": [],
                    "ref-events-confidence-scores": []
                }
                aid = 0
                for agent_ssid in event['agents']:
                    new_ssid = f"ss-{offset_span_id + get_num_id(agent_ssid)}"
                    new_event['agents'].append(new_ssid)
                    new_event['agents-confidence-scores'].append(event['agents-confidence-scores'][aid])

                    aid += 1

                if 'money' not in event:
                    event['money'] = []

                mid = 0
                for money_ssid in event['money']:
                    new_ssid = f"ss-{offset_span_id + get_num_id(money_ssid)}"
                    new_event['money'].append(new_ssid)
                    new_event['money-confidence-scores'].append(event['money-confidence-scores'][mid])

                    mid += 1

                pid = 0
                for patient_ssid in event['patients']:
                    new_ssid = f"ss-{offset_span_id + get_num_id(patient_ssid)}"
                    new_event['patients'].append(new_ssid)
                    new_event['patients-confidence-scores'].append(event['patients-confidence-scores'][pid])

                    pid += 1

                rid = 0
                for tail_ev_id in event['ref-events']:
                    new_tail_ev_id = f"event-{offset_event_id + get_num_id(tail_ev_id)}"
                    new_event['ref-events'].append(new_tail_ev_id)
                    new_event['ref-events-confidence-scores'].append(event['ref-events-confidence-scores'][rid])

                    rid += 1

                write_out['entries'][write_entry['entry-id']]['annotation-sets']['basic-events']['events'][
                    new_ev_id] = new_event

            for span_id, spans in write_entry['annotation-sets']['basic-events']['span-sets'].items():
                new_span_id = f"ss-{offset_span_id + get_num_id(span_id)}"
                spans['ssid'] = new_span_id
                write_out['entries'][write_entry['doc-id']]['annotation-sets']['basic-events']['span-sets'][
                    new_span_id] = spans

    # avoid missing entry error in the scorer
    with open(infile) as f:
        in_data = json.load(f)
    for entry_id, entry in in_data['entries'].items():
        if entry_id not in write_out['entries']:
            write_out['entries'][entry_id] = entry

    with open(output_fpath, 'w') as f:
        json.dump(write_out, f, ensure_ascii=False)


def write_arg_predictions(pred_graphs, output_fpath, infile):
    global_output = {
        'corpus-id': 'argument-output',
        'entries': {}
    }
    entryid2segmentids = defaultdict(set)
    entryid_span2spans = defaultdict(dict)
    entryid_span2events = defaultdict(dict)

    for graph in pred_graphs:
        write_entry = graph.write_entry
        entry_id = write_entry['entry-id']
        segment_id = '{}-{}'.format(write_entry['start'], write_entry['end'])

        local_spansets = write_entry['annotation-sets']['basic-events']['span-sets']
        local_eventsets = write_entry['annotation-sets']['basic-events']['events']

        if entry_id not in global_output['entries']:
            global_output['entries'][entry_id] = write_entry
            write_entry['segment-sections'] = []
            write_entry['segment-sections'].append(
                {
                    "start": write_entry['start'],
                    "end": write_entry['end'],
                    "structural-element": "Sentence"
                }
            )
            entryid2segmentids[entry_id].add(segment_id)

            for span_id, span in local_spansets.items():
                if span['spans'][0]['synclass'] != 'event-anchor':
                    entryid_span2spans[entry_id][
                        '{}-{}'.format(span['spans'][0]['start'], span['spans'][0]['end'])] = span_id
            for event_id, event in local_eventsets.items():
                anchor_span_id = event['anchors']
                anchor_span = local_spansets[anchor_span_id]['spans'][0]
                entryid_span2events[entry_id]['{}-{}'.format(anchor_span['start'], anchor_span['end'])] = event_id

        else:
            if segment_id not in entryid2segmentids[entry_id]:
                entryid2segmentids[entry_id].add(segment_id)
                global_output['entries'][entry_id]['segment-sections'].append(
                    {
                        "start": write_entry['start'],
                        "end": write_entry['end'],
                        "structural-element": "Sentence"
                    }
                )
                offset_span_id = len(global_output['entries'][entry_id]['annotation-sets']['basic-events']['span-sets'])
                offset_event_id = len(global_output['entries'][entry_id]['annotation-sets']['basic-events']['events'])

                for event_id, event in local_eventsets.items():
                    new_ev_id = f"event-{offset_event_id + get_num_id(event['eventid'])}"
                    new_event = {
                        "agents": [],
                        "agents-confidence-scores": [],
                        "money": [],
                        "money-confidence-scores": [],
                        "anchors": f"ss-{offset_span_id + get_num_id(event['anchors'])}",
                        "event-type": event['event-type'],
                        'event-confidence-score': event['event-confidence-score'],
                        "eventid": new_ev_id,
                        "patients": [],
                        "patients-confidence-scores": [],
                        "ref-events": [],
                        "ref-events-confidence-scores": []
                    }
                    aid = 0
                    for agent_ssid in event['agents']:
                        new_ssid = f"ss-{offset_span_id + get_num_id(agent_ssid)}"
                        new_event['agents'].append(new_ssid)
                        new_event['agents-confidence-scores'].append(event['agents-confidence-scores'][aid])

                        aid += 1

                    if 'money' not in event:
                        event['money'] = []

                    mid = 0
                    for money_ssid in event['money']:
                        new_ssid = f"ss-{offset_span_id + get_num_id(money_ssid)}"
                        new_event['money'].append(new_ssid)
                        new_event['money-confidence-scores'].append(event['money-confidence-scores'][mid])

                        mid += 1

                    pid = 0
                    for patient_ssid in event['patients']:
                        new_ssid = f"ss-{offset_span_id + get_num_id(patient_ssid)}"
                        new_event['patients'].append(new_ssid)
                        new_event['patients-confidence-scores'].append(event['patients-confidence-scores'][pid])

                        pid += 1

                    rid = 0
                    for tail_ev_id in event['ref-events']:
                        new_tail_ev_id = f"event-{offset_event_id + get_num_id(tail_ev_id)}"
                        new_event['ref-events'].append(new_tail_ev_id)
                        new_event['ref-events-confidence-scores'].append(event['ref-events-confidence-scores'][rid])

                        rid += 1

                    if 'time-attachments' in event:
                        new_event['time-attachments'] = []
                        new_event['time-attachments-confidence-scores'] = []

                        tid = 0
                        for time_ssid in event['time-attachments']:
                            new_ssid = f"ss-{offset_span_id + get_num_id(time_ssid)}"
                            new_event['time-attachments'].append(new_ssid)
                            new_event['time-attachments-confidence-scores'].append(
                                event['time-attachments-confidence-scores'][tid])

                            tid += 1

                    global_output['entries'][entry_id]['annotation-sets']['basic-events']['events'][
                        new_ev_id] = new_event
                    event_span = local_spansets[event['anchors']]['spans'][0]
                    entryid_span2events[entry_id][
                        '{}-{}'.format(event_span['start'], event_span['end'])] = new_ev_id

                for span_id, spans in local_spansets.items():
                    new_span_id = f"ss-{offset_span_id + get_num_id(span_id)}"
                    spans['ssid'] = new_span_id
                    global_output['entries'][entry_id]['annotation-sets']['basic-events']['span-sets'][
                        new_span_id] = spans
                    if spans['spans'][0]['synclass'] != 'event-anchor':
                        entryid_span2spans[entry_id][
                            '{}-{}'.format(spans['spans'][0]['start'], spans['spans'][0]['end'])] = new_span_id
            else:
                offset_event_id = len(
                    global_output['entries'][write_entry['entry-id']]['annotation-sets']['basic-events']['events'])

                offset_span_id = len(
                    global_output['entries'][write_entry['entry-id']]['annotation-sets']['basic-events']['span-sets'])

                for event_id, event in local_eventsets.items():
                    event_span = local_spansets[event['anchors']]['spans'][0]

                    old_event_id = entryid_span2events[entry_id]['{}-{}'.format(
                        event_span['start'], event_span['end']
                    )]
                    old_event = \
                        global_output['entries'][write_entry['entry-id']]['annotation-sets']['basic-events']['events'][
                            old_event_id]

                    if 'time-attachments' in event:
                        if 'time-attachments' not in old_event:
                            old_event['time-attachments'] = []
                            old_event['time-attachments-confidence-scores'] = []

                        tid = -1
                        for time_ssid in event['time-attachments']:
                            tid += 1
                            time_span = '{}-{}'.format(local_spansets[time_ssid]['spans'][0]['start'],
                                                       local_spansets[time_ssid]['spans'][0]['end'])

                            if time_span in entryid_span2spans[entry_id]:
                                old_event['time-attachments'].append(entryid_span2spans[entry_id][time_span])
                                old_event['time-attachments-confidence-scores'].append(
                                    event['time-attachments-confidence-scores'][tid])
                            else:
                                new_ssid = f"ss-{offset_span_id + 1}"
                                offset_span_id += 1
                                old_event['time-attachments'].append(new_ssid)
                                old_event['time-attachments-confidence-scores'].append(
                                    event['time-attachments-confidence-scores'][tid])

                                global_output['entries'][write_entry['entry-id']]['annotation-sets']['basic-events'][
                                    'span-sets'][new_ssid] = {
                                    'ssid': new_ssid,
                                    'spans': [{
                                        'start': local_spansets[time_ssid]['spans'][0]['start'],
                                        'end': local_spansets[time_ssid]['spans'][0]['end'],
                                        'synclass': local_spansets[time_ssid]['spans'][0]['synclass'],
                                        'string': local_spansets[time_ssid]['spans'][0]['string']
                                    }]
                                }
                                entryid_span2spans[entry_id][time_span] = new_ssid

                    aid = -1
                    for agent_ssid in event['agents']:
                        aid += 1
                        agent_span = '{}-{}'.format(local_spansets[agent_ssid]['spans'][0]['start'],
                                                    local_spansets[agent_ssid]['spans'][0]['end'])
                        if agent_span in entryid_span2spans[entry_id]:
                            old_event['agents'].append(entryid_span2spans[entry_id][agent_span])
                            old_event['agents-confidence-scores'].append(event['agents-confidence-scores'][aid])
                        else:
                            new_ssid = f"ss-{offset_span_id + 1}"
                            offset_span_id += 1
                            old_event['agents'].append(new_ssid)
                            old_event['agents-confidence-scores'].append(event['agents-confidence-scores'][aid])

                            global_output['entries'][write_entry['entry-id']]['annotation-sets']['basic-events'][
                                'span-sets'][new_ssid] = {
                                'ssid': new_ssid,
                                'spans': [{
                                    'start': local_spansets[agent_ssid]['spans'][0]['start'],
                                    'end': local_spansets[agent_ssid]['spans'][0]['end'],
                                    'synclass': local_spansets[agent_ssid]['spans'][0]['synclass'],
                                    'string': local_spansets[agent_ssid]['spans'][0]['string']
                                }]
                            }
                            entryid_span2spans[entry_id][agent_span] = new_ssid

                    if 'money' not in event:
                        event['money'] = []
                        event['money-confidence-scores'] = []

                    mid = -1
                    for money_ssid in event['money']:
                        mid += 1
                        money_span = '{}-{}'.format(local_spansets[money_ssid]['spans'][0]['start'],
                                                    local_spansets[money_ssid]['spans'][0]['end'])
                        score = event['money-confidence-scores'][mid]

                        if money_span in entryid_span2spans[entry_id]:
                            old_event['money'].append(entryid_span2spans[entry_id][money_span])
                            old_event['money-confidence-scores'].append(score)
                        else:
                            new_ssid = f"ss-{offset_span_id + 1}"
                            offset_span_id += 1
                            old_event['money'].append(new_ssid)
                            old_event['money-confidence-scores'].append(score)

                            global_output['entries'][write_entry['entry-id']]['annotation-sets']['basic-events'][
                                'span-sets'][new_ssid] = {
                                'ssid': new_ssid,
                                'spans': [{
                                    'start': local_spansets[money_ssid]['spans'][0]['start'],
                                    'end': local_spansets[money_ssid]['spans'][0]['end'],
                                    'synclass': local_spansets[money_ssid]['spans'][0]['synclass'],
                                    'string': local_spansets[money_ssid]['spans'][0]['string']
                                }]
                            }
                            entryid_span2spans[entry_id][money_span] = new_ssid

                    pid = -1
                    for patient_ssid in event['patients']:
                        pid += 1
                        patient_span = '{}-{}'.format(local_spansets[patient_ssid]['spans'][0]['start'],
                                                      local_spansets[patient_ssid]['spans'][0]['end'])
                        score = event['patients-confidence-scores'][pid]
                        if patient_span in entryid_span2spans[entry_id]:
                            old_event['patients'].append(entryid_span2spans[entry_id][patient_span])
                            old_event['patients-confidence-scores'].append(score)
                        else:
                            new_ssid = f"ss-{offset_span_id + 1}"
                            offset_span_id += 1
                            old_event['patients'].append(new_ssid)
                            old_event['patients-confidence-scores'].append(score)

                            global_output['entries'][write_entry['entry-id']]['annotation-sets']['basic-events'][
                                'span-sets'][new_ssid] = {
                                'ssid': new_ssid,
                                'spans': [{
                                    'start': local_spansets[patient_ssid]['spans'][0]['start'],
                                    'end': local_spansets[patient_ssid]['spans'][0]['end'],
                                    'synclass': local_spansets[patient_ssid]['spans'][0]['synclass'],
                                    'string': local_spansets[patient_ssid]['spans'][0]['string']
                                }]
                            }
                            entryid_span2spans[entry_id][patient_span] = new_ssid

                    rid = -1
                    for tail_ev_id in event['ref-events']:
                        rid += 1
                        tail_event_span = '{}-{}'.format(
                            local_spansets[local_eventsets[tail_ev_id]['anchors']]['spans'][0]['start'],
                            local_spansets[local_eventsets[tail_ev_id]['anchors']]['spans'][0]['end'])
                        old_tail_event_id = entryid_span2events[entry_id][tail_event_span]
                        old_event['ref-events'].append(old_tail_event_id)
                        old_event['ref-events-confidence-scores'].append(event['ref-events-confidence-scores'][rid])

    # avoid missing entry error in the scorer
    with open(infile) as f:
        in_data = json.load(f)
    for entry_id, entry in in_data['entries'].items():
        if entry_id not in global_output['entries']:
            global_output['entries'][entry_id] = entry
    with open(output_fpath, 'w') as f:
        json.dump(global_output, f, ensure_ascii=False)


def write_template_predictions(pred_templates, infile, output_fpath):
    with open(infile) as f:
        write_out = json.load(f)

    for template in pred_templates:
        entry_id = template['entry-id']

        if 'granular-templates' not in write_out['entries'][entry_id]['annotation-sets']['basic-events']:
            write_out['entries'][entry_id]['annotation-sets']['basic-events']['granular-templates'] = {}

        offset_template_id = len(
            write_out['entries'][entry_id]['annotation-sets']['basic-events']['granular-templates']
        )

        new_tl_id = 'template-{}'.format(offset_template_id + 1)
        new_template = {
            "template-anchor": 'ss-1',
            "template-confidence-score": template['confidence'],
            "template-id": new_tl_id,
            "template-type": template['template-type']
        }
        if 'type' in template: new_template['type'] = template['type']
        if 'completion' in template: new_template['completion'] = template['completion']
        if 'over-time' in template: new_template['over-time'] = template['over-time']
        if 'coordinated' in template: new_template['coordinated'] = template['coordinated']

        write_out['entries'][entry_id]['annotation-sets']['basic-events']['granular-templates'][
            new_tl_id] = new_template

    # avoid missing entry error in the scorer
    with open(infile) as f:
        in_data = json.load(f)
    for entry_id, entry in in_data['entries'].items():
        if entry_id not in write_out['entries']:
            write_out['entries'][entry_id] = entry
    with open(output_fpath, 'w') as f:
        json.dump(write_out, f, ensure_ascii=False)


def is_pronoun(span_sets, span_id):
    spans = span_sets[span_id]
    for span in spans['spans']:
        if span['synclass'] == 'pronoun':
            return True
    return False


def write_slot_predictions(predictions, infile, output_fpath):
    with open(infile) as f:
        write_out = json.load(f)

    for pred in predictions:
        granular_template = \
            write_out['entries'][pred['entry-id']]['annotation-sets']['basic-events']['granular-templates'][
                pred['template-id']]
        if pred['slot'] not in granular_template:
            granular_template[pred['slot']] = []

        if pred['filler-id'].startswith('event'):
            granular_template[pred['slot']].append({
                'event-id': pred['filler-id'],
                'slot-confidence-score': pred['confidence']
            })
            basic_event = write_out['entries'][pred['entry-id']]['annotation-sets']['basic-events']['events'][
                pred['filler-id']]
            if 'time-attachments' in basic_event:
                granular_template[pred['slot']][-1]['time-attachments'] = basic_event['time-attachments']
        else:
            assert pred['filler-id'].startswith('ss'), 'pred: {}'.format(pred)
            granular_template[pred['slot']].append({
                'ssid': pred['filler-id'],
                'slot-confidence-score': pred['confidence']
            })
        if 'irrealis' in pred:
            granular_template[pred['slot']][-1]['irrealis'] = pred['irrealis']

    # remove pronoun slot fillers
    for entry_id, entry in write_out['entries'].items():
        if 'granular-templates' in entry['annotation-sets']['basic-events']:
            templates = entry['annotation-sets']['basic-events']['granular-templates']
            span_sets = entry['annotation-sets']['basic-events']['span-sets']

            for tmplt_id, tmplt in templates.items():
                for slot_name, slot_fillers in tmplt.items():
                    if isinstance(slot_fillers, list):
                        out_fillers = []
                        for filler in slot_fillers:
                            if 'ssid' in filler and is_pronoun(span_sets, filler['ssid']):
                                continue
                            out_fillers.append(filler)
                        tmplt[slot_name] = out_fillers

    # avoid missing entry error in the scorer
    with open(infile) as f:
        in_data = json.load(f)
    for entry_id, entry in in_data['entries'].items():
        if entry_id not in write_out['entries']:
            write_out['entries'][entry_id] = entry
    with open(output_fpath, 'w') as f:
        json.dump(write_out, f, ensure_ascii=False)


def compute_granular_score(reffile, sysfile, csv_fpath):
    os.system('python scorer/score.py -f -c {} Granular {} {}'.format(csv_fpath + '.granular', sysfile, reffile))

    try:
        with open(csv_fpath + '.granular') as f:
            scores = [line.strip() for line in f if line.strip()][-1].split(',')
            evt_prec = float(scores[-20])
            evt_rec = float(scores[-19])
            evt_fmeasure = float(scores[-18])

            arg_prec = float(scores[-12])
            arg_rec = float(scores[-11])
            arg_fmeasure = float(scores[-10])

            com_fmeasure = float(scores[-1])

        granular_score = {
            'template': {'p': evt_prec, 'r': evt_rec, 'f1': evt_fmeasure},
            'slot': {'p': arg_prec, 'r': arg_rec, 'f1': arg_fmeasure},
            'combined': com_fmeasure
        }
    except:
        granular_score = {
            'template': {'p': 0, 'r': 0, 'f1': 0},
            'slot': {'p': 0, 'r': 0, 'f1': 0},
            'combined': 0
        }
    return granular_score


def compute_basic_score(reffile, sysfile, csv_fpath):
    os.system('python scorer/score.py -f -c {} Basic {} {}'.format(csv_fpath + '.basic', sysfile, reffile))

    try:
        with open(csv_fpath + '.basic') as f:
            scores = [line.strip() for line in f if line.strip()][-1].split(',')
            evt_prec = float(scores[-20])
            evt_rec = float(scores[-19])
            evt_fmeasure = float(scores[-18])

            arg_prec = float(scores[-12])
            arg_rec = float(scores[-11])
            arg_fmeasure = float(scores[-10])

            com_fmeasure = float(scores[-1])

        basic_score = {
            'event': {'p': evt_prec, 'r': evt_rec, 'f1': evt_fmeasure},
            'argument': {'p': arg_prec, 'r': arg_rec, 'f1': arg_fmeasure},
            'combined': com_fmeasure
        }
    except:
        basic_score = {
            'event': {'p': 0, 'r': 0, 'f1': 0},
            'argument': {'p': 0, 'r': 0, 'f1': 0},
            'combined': 0
        }

    return basic_score


def remove_punctuations(string):
    pattern = r'''["’'\(\)\[\]\{\}<>:\,‒–—―…!\.«»\-‐\?‘’“”;/⁄␠·&@\*\\•\^¤¢\$€£¥₩₪†‡°¡¿¬\#№%‰‱¶′§~¨_\|¦⁂☞∴‽※"]+'''
    return re.sub(pattern, '', string)


def get_mid2token_span(tokens, span_set):
    mid_2_token_span = {}
    for spanid, span in span_set.items():
        spans = span['spans']
        for i in range(len(spans)):
            mid = '{}-{}'.format(spanid, i)
            span_words = spans[i]['string'].split()
            num_words = len(span_words)

            start, end = 1000000, -1
            if num_words == 1:
                for tid, token in enumerate(tokens):
                    if span_words[0] == token['text']:
                        start = tid
                        end = tid + 1
                        break
            elif num_words >= 2:
                for tid, token in enumerate(tokens):
                    if token['text'] == span_words[0]:
                        start = tid
                    if token['text'] == span_words[-1]:
                        end = tid
                if start == 1000000 or end != -1:
                    continue
                if start < end:
                    end += 1
                elif end < start:
                    tmp = start
                    start = end
                    end = tmp + 1
                else:
                    continue
            else:
                continue
            if start == 1000000 or end == -1:
                continue
            mid_2_token_span[mid] = (start, end, spans[i]['string'])
    return mid_2_token_span


def get_num_id(string_id):
    num_id = int(string_id.split('-')[-1])
    return num_id


def construct_graphs(eid2docgraph, subgraphs, vocabs):
    entryid2graph = {}
    for subgraph in subgraphs:
        entry_id = subgraph.entry_id
        docgraph = eid2docgraph[entry_id]
        if entry_id not in entryid2graph:
            entryid2graph[entry_id] = docgraph
        for mid1, mid2, rel_type in subgraph.relations:
            if vocabs['relation_type'][rel_type] > 0:
                m_index1 = docgraph.mention_id_2_mention_index[mid1]
                m_index2 = docgraph.mention_id_2_mention_index[mid2]
                docgraph.add_relation(m_index1, m_index2, rel_type)
        for tid, mid, role_type in subgraph.roles:
            if vocabs['role_type'][role_type] > 0:
                t_index = docgraph.trigger_id_2_trigger_index[tid]
                m_index = docgraph.mention_id_2_mention_index[mid]
                docgraph.add_role(t_index, m_index, role_type)
    graph_list = [(k, v) for k, v in entryid2graph.items()]
    graph_list.sort(key=lambda x: x[0])
    return [v for k, v in graph_list]


def aggregate_subgraphs(entryid2docgraph, gold_subgraphs, pred_subgraphs, vocabs):
    '''
    key fact: trigger_id and mention_id are the same -> use these to map
    '''
    if gold_subgraphs:
        gold_graphs = construct_graphs(
            deepcopy(entryid2docgraph), gold_subgraphs, vocabs)
    else:
        gold_graphs = None
    if pred_subgraphs:
        pred_graphs = construct_graphs(
            deepcopy(entryid2docgraph), pred_subgraphs, vocabs)
    else:
        pred_graphs = None
    return gold_graphs, pred_graphs


def parse_basic_scorer_report(csv_fpath):
    with open(csv_fpath) as f:
        scores = [line.strip() for line in f if line.strip()][-1].split(',')
        evt_prec = float(scores[-20])
        evt_rec = float(scores[-19])
        evt_fmeasure = float(scores[-18])

        arg_prec = float(scores[-12])
        arg_rec = float(scores[-11])
        arg_fmeasure = float(scores[-10])

        com_fmeasure = float(scores[-1])

    score = {
        'event': {'p': evt_prec, 'r': evt_rec, 'f': evt_fmeasure},
        'argument': {'p': arg_prec, 'r': arg_rec, 'f': arg_fmeasure},
        'combined': {'f': com_fmeasure}
    }
    return score


def build_vocabs(nlp, basic_fpath, lang):
    nlp.set_active(lang)
    empty_vocab = {
        'trigger_label': {'O': 0},
        'template-type': {'O': 0}, 'slot_type': {'O': 0}, 'irrealis_type': {'O': 0},
        'template-type2slots': defaultdict(list), 'slot2irrealis': defaultdict(list),
        'slot2filler_types': defaultdict(dict),
        'template_label': {'O': 0},
        'type_slot': {'O': 0}, 'completion_slot': {'O': 0},
        'overtime_slot': {'O': 0}, 'coordinated_slot': {'O': 0},
        'mention_label': {
            'O': 0,
            'B-anchor': 1,
            'I-anchor': 2,
            'B-name': 3,
            'I-name': 4,
            'B-nominal': 5,
            'I-nominal': 6,
            'B-pronoun': 7,
            'I-pronoun': 8,
            'B-time': 9,
            'I-time': 10
        },
        'event_type': {'O': 0},
        'mention_type': {'O': 0},
        'role_type': {'O': 0, 'agents': 1, 'patients': 2, 'ref-events': 3, 'money': 4, 'time-attachments': 5},
        'type': {'O': 0}
    }
    with open(basic_fpath) as f:
        data = json.load(f)
    c = 0
    for entry_id, entry in data['entries'].items():
        print('processed: ', entry_id)
        _ = parse_entry(
            nlp, entry, lang=lang, input_vocab=empty_vocab)
        c += 1
    print('Read {} entries'.format(c))
    for t in empty_vocab['template-type']:
        if t == 'O':
            continue

        for prefix in ['B', 'I']:
            empty_vocab['template_label']['{}-{}'.format(
                prefix, t
            )] = len(empty_vocab['template_label'])

    for t in empty_vocab['event_type']:
        if t == 'O':
            continue

        empty_vocab['type'][t] = len(empty_vocab['type'])

        for prefix in ['B', 'I']:
            empty_vocab['trigger_label']['{}-{}'.format(
                prefix, t)] = len(empty_vocab['trigger_label'])
    sorted_vocab = {}

    for key in empty_vocab:
        if key in ['template-type2slots', 'slot2irrealis', 'slot2filler_types']:
            sorted_vocab[key] = empty_vocab[key]
            continue

        tmp = [x for x in list(empty_vocab[key].keys()) if x != 'O']
        tmp.sort()
        sorted_vocab[key] = {}
        sorted_vocab[key]['O'] = 0
        for k in tmp:
            sorted_vocab[key][k] = len(sorted_vocab[key])
    with open('vocabs.json', 'w') as f:
        json.dump(sorted_vocab, f, ensure_ascii=False)


def collapse_events(ori_sysfile, collapsed_sysfile):
    with open(ori_sysfile) as f:
        data = json.load(f)

    for entry in data['entries'].values():
        event_set = entry['annotation-sets']['basic-events']['events']

        arg_id2_eventids = defaultdict(list)
        evid2headevids = defaultdict(list)

        for event in event_set.values():
            event['agents'].sort()
            if 'money' not in event:
                event['money'] = []
            event['money'].sort()
            event['patients'].sort()

            arg_id = '{}|{}|{}'.format(','.join(event['agents']), ','.join(event['money']), ','.join(event['patients']))
            arg_id2_eventids[arg_id].append(event['eventid'])

            for refevid in event['ref-events']:
                evid2headevids[refevid].append(event['eventid'])

        for arg_id in arg_id2_eventids:
            if len(arg_id2_eventids[arg_id]) > 1:
                kept_event = event_set[arg_id2_eventids[arg_id][0]]

                for del_eventid in arg_id2_eventids[arg_id][1:]:

                    for headid in evid2headevids[del_eventid]:
                        if headid in event_set:
                            idx = event_set[headid]['ref-events'].index(del_eventid)
                            del event_set[headid]['ref-events'][idx]

                            if kept_event['eventid'] not in event_set[headid]['ref-events']:
                                event_set[headid]['ref-events'].append(kept_event['eventid'])

                    del event_set[del_eventid]

    with open(collapsed_sysfile, 'w') as f:
        json.dump(data, f)


def remove_pronouns(ori_sysfile, fil_sysfile):
    with open(ori_sysfile) as f:
        data = json.load(f)
    entries = data['entries']
    num_remove = 0
    for _, entry in entries.items():
        event_set = entry['annotation-sets']['basic-events']['events']
        span_sets = entry['annotation-sets']['basic-events']['span-sets']
        for _, event in event_set.items():
            agent_ssids = deepcopy(event['agents'])
            patient_ssids = deepcopy(event['patients'])
            filtered_agent_ssids, filtered_patient_ssids = [], []
            for assid in agent_ssids:
                if span_sets[assid]['spans'][0]['synclass'] != 'pronoun':
                    filtered_agent_ssids.append(assid)
                else:
                    num_remove += 1
            for pssid in patient_ssids:
                if span_sets[pssid]['spans'][0]['synclass'] != 'pronoun':
                    filtered_patient_ssids.append(pssid)
                else:
                    num_remove += 1
            event['agents'] = filtered_agent_ssids
            event['patients'] = filtered_patient_ssids
    # print('removing {} pronoun arguments'.format(num_remove))

    with open(fil_sysfile, 'w') as f:
        json.dump(data, f, ensure_ascii=False)

    # collapse_events(fil_sysfile, fil_sysfile)


def read_and_build(nlp, basic_fpath, write_fpath, lang, model_name=''):
    nlp.set_active(lang)

    np.random.seed(1234)
    with open(basic_fpath) as f:
        data = json.load(f)

    pipeline_data = []

    num_skip_events, num_skip_mentions = 0, 0
    num_one_sent_relations = 0
    sentdist2numargs = defaultdict(int)
    for entry_id, entry in data['entries'].items():
        print('processed: ', entry_id)
        pipeline_insts = parse_entry(nlp, entry, lang=lang,
                                     sentdist2numargs=sentdist2numargs,
                                     model_name=model_name)
        if 'phase' in entry:
            for inst in pipeline_insts:
                inst['phase'] = int(entry['phase'])

        pipeline_data += pipeline_insts

    # ------------------------------------
    print('Read {} entries'.format(len(data['entries'])))
    print('Write {} examples to file: {}'.format(len(pipeline_data), write_fpath.format('pipeline')))

    with open(os.path.join(WORKING_DIR, write_fpath), 'w') as f:
        for d in pipeline_data:
            f.write(json.dumps(d, ensure_ascii=False) + '\n')


def tokenize_doc(nlp, sent_inst_list, lang):
    if lang == 'chinese' and len(sent_inst_list) > 0:
        if hanzidentifier.is_simplified(sent_inst_list[0]['text']):
            nlp.set_active('chinese')
        else:
            nlp.set_active('traditional-chinese')

    parsed = []
    token2sent = {}
    char2token = {}
    doc_token_id = 0

    if len(sent_inst_list) == 0:
        return parsed, char2token, token2sent
    else:
        for sent_id, sent_inst in enumerate(sent_inst_list):
            sent_feat_fpath = os.path.join(PRECOMPUTED_DIR,
                                           '{}-{}-{}.json'.format(sent_inst['entry_id'], sent_inst['start'],
                                                                  sent_inst['end']))
            if os.path.exists(sent_feat_fpath):
                with open(sent_feat_fpath) as f:
                    precomputed = json.load(f)
                    tokenized = precomputed['tokenized']
                    posdep_tokens = precomputed['posdep_tokens']
            else:
                tokenized = nlp.tokenize(sent_inst['text'], is_sent=True)['tokens']
                posdep_tokens = nlp.posdep([t['text'] for t in tokenized], is_sent=True)['tokens']

                with open(sent_feat_fpath, 'w') as f:
                    json.dump({
                        'tokenized': tokenized, 'posdep_tokens': posdep_tokens
                    }, f, ensure_ascii=False)

            sent_token_id = 0
            for k, pd_tok in enumerate(posdep_tokens):
                parsed.append({
                    'start': sent_inst['start'] + tokenized[k]['span'][0],  # document-level span
                    'end': sent_inst['start'] + tokenized[k]['span'][1],  # document-level span
                    'text': tokenized[k]['text']
                })
                sent_inst['tokens'].append(
                    {
                        'start': sent_inst['start'] + tokenized[k]['span'][0],  # document-level span
                        'end': sent_inst['start'] + tokenized[k]['span'][1],  # document-level span
                        'text': tokenized[k]['text'],
                        'upos': pd_tok['upos'],
                        'deprel': pd_tok['deprel'],
                        'head': pd_tok['head']
                    }
                )

                # char id 2 token id
                for char_id in range(parsed[-1]['start'], parsed[-1]['end']):
                    char2token[char_id] = doc_token_id
                # token id 2 sent id
                token2sent[doc_token_id] = (sent_id, sent_token_id)

                doc_token_id += 1
                sent_token_id += 1

    return parsed, char2token, token2sent


def parse_entry(nlp, basic_entry, lang, input_vocab=None, sentdist2numargs=None, model_name=''):
    original_doc_inst = {
        'doc_id': basic_entry['doc-id'],
        'entry_id': basic_entry['entry-id'],
        'text': '',  # done
        'tokens': [],  # done
        'mentions': {},  # done
        'relations': {},  # done
        'triggers': [],
        'roles': {},
        'token2sent': {}  # done
    }
    sent_inst_list = []
    merged_sent_pairs = set()
    doc_text = basic_entry['segment-text']
    original_doc_inst['text'] = doc_text

    for section in basic_entry['segment-sections']:
        start = section['start']
        end = section['end']
        if section['structural-element'] == 'Sentence':
            sent_inst = {
                'sent_id': basic_entry['entry-id'],
                'entry_id': basic_entry['entry-id'],
                'doc_id': basic_entry['doc-id'],
                'text': doc_text[start: end],  # done
                'doc_ori_text': doc_text,
                'start': start,
                'end': end,
                'tokens': [],  # done
                'templates': [],
                'events': [],  # done
                'mentions': [],  # done
            }
            if end - start > 1:
                sent_inst_list.append(sent_inst)
    sent_inst_list.sort(key=lambda x: (x['start'], x['end']))
    for sent_id in range(len(sent_inst_list)):
        sent_inst_list[sent_id]['sent_id'] += '|sent_id-{}'.format(sent_id)
    doc_parsed, char2doc_token, doc_token2sent_token = tokenize_doc(nlp, sent_inst_list, lang)

    original_doc_inst['tokens'] = doc_parsed
    # read annotations
    if 'annotation-sets' not in basic_entry:
        basic_entry['annotation-sets'] = {
            'basic-events': {'events': {}, 'span-sets': {}}}

    granular_templates = basic_entry['annotation-sets']['basic-events']['granular-templates'] if 'granular-templates' in \
                                                                                                 basic_entry[
                                                                                                     'annotation-sets'][
                                                                                                     'basic-events'] else {}
    basic_events = basic_entry['annotation-sets']['basic-events']['events'] if 'events' in \
                                                                               basic_entry['annotation-sets'][
                                                                                   'basic-events'] else {}
    span_sets = basic_entry['annotation-sets']['basic-events']['span-sets'] if 'span-sets' in \
                                                                               basic_entry['annotation-sets'][
                                                                                   'basic-events'] else {}
    # ------- read mentions
    mention_id_set = set()
    num_skip_mentions = 0
    num_skip_events = 0
    num_one_sent_relations = 0
    for span_id in span_sets:
        span_is_mention = False
        mention_id = -1
        for mention_id, mention in enumerate(span_sets[span_id]['spans']):
            if 'synclass' not in mention:
                mention['synclass'] = 'time'

            if mention['synclass'] not in {'event-anchor', 'template-anchor'}:
                start = 10000
                end = -1
                for charid in range(mention['start'], mention['end']):
                    if charid in char2doc_token:
                        doc_token_id = char2doc_token[charid]
                        start = min(start, doc_token_id)
                        end = max(end, doc_token_id)
                if start == 10000 and end == -1:
                    num_skip_mentions += 1
                    continue

                hstart = 10000
                hend = -1
                if 'hstart' not in mention or 'hend' not in mention:
                    mention['hstart'] = mention['start']
                    mention['hend'] = mention['end']
                if 'hstring' not in mention:
                    mention['hstring'] = mention['string']

                for hcharid in range(mention['hstart'], mention['hend']):
                    if hcharid in char2doc_token:
                        hdoc_token_id = char2doc_token[hcharid]
                        hstart = min(hstart, hdoc_token_id)
                        hend = max(hend, hdoc_token_id)
                if hstart == 10000 or hend == -1:
                    hstart = start
                    hend = end

                # skip this mention
                if doc_token2sent_token[end][0] - doc_token2sent_token[start][0] >= 2:
                    num_skip_mentions += 1
                    continue
                if doc_token2sent_token[end][0] - doc_token2sent_token[start][0] == 1:
                    merged_sent_pairs.add(
                        (doc_token2sent_token[start][0], doc_token2sent_token[end][0]))
                    offset = len(
                        sent_inst_list[doc_token2sent_token[start][0]]['tokens'])
                else:
                    offset = 0

                doc_mid = '{}-{}-{}'.format(
                    basic_entry['entry-id'], span_id, mention_id)
                sent_id, _ = doc_token2sent_token[start]
                sent_inst_list[sent_id]['mentions'].append({
                    'id': doc_mid,
                    'ssid': span_id,
                    'start': doc_token2sent_token[start][1],
                    'end': offset + doc_token2sent_token[end][1] + 1,
                    'ori_tok_id': start,
                    # pre-shift for later merging of parts of a broken sentence
                    'mention_type': mention['synclass'],
                    'text': mention['string'],
                    'head': {
                        'text': mention['hstring'],
                        'start': doc_token2sent_token[hstart][1],
                        'end': offset + doc_token2sent_token[hend][1] + 1
                    }
                })
                if input_vocab:
                    input_vocab['mention_type'][mention['synclass']] = input_vocab['mention_type'].get(
                        mention['synclass'], len(input_vocab['mention_type']))
                    input_vocab['type'][mention['synclass']] = input_vocab['type'].get(mention['synclass'],
                                                                                       len(input_vocab['type']))

                original_doc_inst['mentions'][doc_mid] = {
                    'id': doc_mid,
                    'start': start,
                    'end': end + 1,
                    'sent_id': sent_id,
                    'mention_type': mention['synclass'],
                    'text': mention['string']
                }

                mention_id_set.add(doc_mid)
                span_is_mention = True

        num_mentions = mention_id + 1
        if span_is_mention and num_mentions >= 2:
            # add Symmetric relations
            for mid1 in range(num_mentions):
                for mid2 in range(mid1):
                    arg1_id = '{}-{}-{}'.format(
                        basic_entry['entry-id'], span_id, mid1)
                    arg2_id = '{}-{}-{}'.format(
                        basic_entry['entry-id'], span_id, mid2)
                    if arg1_id in mention_id_set and arg2_id in mention_id_set:
                        rel_id = '{}-{}'.format(arg1_id, arg2_id)
                        original_doc_inst['relations'][rel_id] = {
                            'relation_type': 'Symmetric',
                            'id': rel_id,
                            'arguments': [
                                {
                                    'mention_id': arg1_id,
                                    'text': span_sets[span_id]['spans'][mid1]['string'],
                                    'role': 'Arg-1'
                                },
                                {
                                    'mention_id': arg2_id,
                                    'text': span_sets[span_id]['spans'][mid2]['string'],
                                    'role': 'Arg-2'
                                }
                            ]
                        }
                        if original_doc_inst['mentions'][arg1_id]['sent_id'] == original_doc_inst['mentions'][arg2_id][
                            'sent_id']:
                            num_one_sent_relations += 1

    # ------- read inclusive relations
    if 'includes-relations' in basic_entry['annotation-sets']['basic-events']:
        for whole_ssid, part_ssid_list in basic_entry['annotation-sets']['basic-events']['includes-relations'].items():
            whole_mentions = span_sets[whole_ssid]['spans']
            part_ssid_mid_list = []
            for part_ssid in part_ssid_list:
                for part_mid in range(len(span_sets[part_ssid]['spans'])):
                    part_ssid_mid_list.append((part_ssid, part_mid))
            for whole_m_id, whole_m in enumerate(whole_mentions):
                for _, part_ssid_mid in enumerate(part_ssid_mid_list):
                    part_ssid, part_mid = part_ssid_mid
                    arg1_id = '{}-{}-{}'.format(
                        basic_entry['entry-id'], whole_ssid, whole_m_id)
                    arg2_id = '{}-{}-{}'.format(
                        basic_entry['entry-id'], part_ssid, part_mid)
                    if arg1_id in mention_id_set and arg2_id in mention_id_set:
                        rel_id = '{}-{}'.format(arg1_id, arg2_id)
                        original_doc_inst['relations'][rel_id] = {
                            'relation_type': 'Includes',
                            'id': rel_id,
                            'arguments': [
                                {
                                    'mention_id': arg1_id,
                                    'text': span_sets[whole_ssid]['spans'][whole_m_id]['string'],
                                    'role': 'Arg-1'
                                },
                                {
                                    'mention_id': arg2_id,
                                    'text': span_sets[part_ssid]['spans'][part_mid]['string'],
                                    'role': 'Arg-2'
                                }
                            ]
                        }
                        if original_doc_inst['mentions'][arg1_id]['sent_id'] == original_doc_inst['mentions'][arg2_id][
                            'sent_id']:
                            num_one_sent_relations += 1

    eventid_2_timeattachments = defaultdict(list)
    for template_id, template in granular_templates.items():
        for slot_name, fillers in template.items():
            if isinstance(fillers, list):
                for filler in fillers:
                    if 'event-id' in filler and "time-attachments" in filler:
                        for timex in filler['time-attachments']:
                            eventid_2_timeattachments[filler['event-id']].append(timex)

    # ------- read events
    event_id_2_sent_id = {}
    event_id_2_event = {}
    se2evid = {}
    evid2new_events = defaultdict(list)
    for event in basic_events.values():
        ssid = event['anchors']

        for span in span_sets[ssid]['spans']:
            se2evid['{}-{}'.format(span['start'], span['end'])] = event['eventid']

    for event_id, event in basic_events.items():
        trigger_span_id = event['anchors']
        for trg_m_id, trigger_mention in enumerate(
                span_sets[trigger_span_id]['spans']):  # multiple events sharing the same set of arguments
            start = 10000
            end = -1
            for charid in range(trigger_mention['start'], trigger_mention['end']):
                if charid in char2doc_token:
                    doc_token_id = char2doc_token[charid]
                    start = min(start, doc_token_id)
                    end = max(end, doc_token_id)
            if start == 10000 and end == -1:
                num_skip_events += 1
                continue
            assert doc_token2sent_token[start][0] == doc_token2sent_token[end][0]
            sent_id_trg = doc_token2sent_token[start][0]
            trigger_id = '{}-{}-{}'.format(
                basic_entry['entry-id'], trigger_span_id, trg_m_id)
            new_event = {
                'event_type': event['event-type'],
                'event-confidence-score': event['event-confidence-score'] if 'event-confidence-score' in event else 1.0,
                'sent-id': sent_id_trg,
                'start-end': '{}-{}'.format(trigger_mention['start'], trigger_mention['end']),
                'id': '{}-{}-{}-{}'.format(basic_entry['entry-id'], event_id, trigger_span_id, trg_m_id),
                'eventid': event_id,
                'trigger': {
                    'id': trigger_id,
                    'start': doc_token2sent_token[start][1],
                    'end': doc_token2sent_token[end][1] + 1,
                    'ori_tok_id': start,
                    'text': trigger_mention['string'],
                    'event_type': event['event-type']
                },
                'agents': defaultdict(int), 'money': defaultdict(int), 'patients': defaultdict(int),
                'ref-events': defaultdict(int),
                'time-attachments': defaultdict(int)
            }
            evid2new_events[event_id].append(new_event)

            if input_vocab:
                input_vocab['event_type'][event['event-type']] = input_vocab['event_type'].get(
                    event['event-type'], len(input_vocab['event_type']))
            for agent_ssid in event['agents']:
                for amention_id, amention in enumerate(span_sets[agent_ssid]['spans']):
                    arg_id = '{}-{}-{}'.format(
                        basic_entry['entry-id'], agent_ssid, amention_id)
                    if arg_id in mention_id_set:
                        role_id = '{}-{}'.format(trigger_id, arg_id)
                        original_doc_inst['roles'][role_id] = {
                            'role_type': 'Agent',
                            'id': role_id,
                            'arguments': [
                                {
                                    'trigger_id': trigger_id
                                },
                                {
                                    'mention_id': arg_id
                                }
                            ]
                        }

                        sent_id_arg = original_doc_inst['mentions'][arg_id]['sent_id']

                        if sentdist2numargs is not None:
                            sentdist2numargs[np.abs(
                                sent_id_trg - sent_id_arg)] += 1
                        if sent_id_trg == sent_id_arg or (sent_id_trg, sent_id_arg) in merged_sent_pairs or (
                                sent_id_arg, sent_id_trg) in merged_sent_pairs:
                            new_event['agents'][agent_ssid] += 1

            if 'money' not in event:
                event['money'] = []
            for money_ssid in event['money']:
                for amention_id, amention in enumerate(span_sets[money_ssid]['spans']):
                    arg_id = '{}-{}-{}'.format(
                        basic_entry['entry-id'], money_ssid, amention_id)
                    if arg_id in mention_id_set:
                        role_id = '{}-{}'.format(trigger_id, arg_id)
                        original_doc_inst['roles'][role_id] = {
                            'role_type': 'Money',
                            'id': role_id,
                            'arguments': [
                                {
                                    'trigger_id': trigger_id
                                },
                                {
                                    'mention_id': arg_id
                                }
                            ]
                        }

                        sent_id_arg = original_doc_inst['mentions'][arg_id]['sent_id']

                        if sentdist2numargs is not None:
                            sentdist2numargs[np.abs(
                                sent_id_trg - sent_id_arg)] += 1
                        if sent_id_trg == sent_id_arg or (sent_id_trg, sent_id_arg) in merged_sent_pairs or (
                                sent_id_arg, sent_id_trg) in merged_sent_pairs:
                            new_event['money'][money_ssid] += 1

            for patient_ssid in event['patients']:
                for pmention_id, pmention in enumerate(span_sets[patient_ssid]['spans']):
                    arg_id = '{}-{}-{}'.format(
                        basic_entry['entry-id'], patient_ssid, pmention_id)
                    if arg_id in mention_id_set:
                        role_id = '{}-{}'.format(trigger_id, arg_id)
                        original_doc_inst['roles'][role_id] = {
                            'role_type': 'Patient',
                            'id': role_id,
                            'arguments': [
                                {
                                    'trigger_id': trigger_id
                                },
                                {
                                    'mention_id': arg_id
                                }
                            ]
                        }
                        sent_id_arg = original_doc_inst['mentions'][arg_id]['sent_id']
                        if sentdist2numargs is not None:
                            sentdist2numargs[np.abs(
                                sent_id_trg - sent_id_arg)] += 1
                        if sent_id_trg == sent_id_arg or (sent_id_trg, sent_id_arg) in merged_sent_pairs or (
                                sent_id_arg, sent_id_trg) in merged_sent_pairs:
                            new_event['patients'][patient_ssid] += 1

            for time_ssid in eventid_2_timeattachments[event['eventid']]:
                for tmention_id, tmention in enumerate(span_sets[time_ssid]['spans']):
                    arg_id = '{}-{}-{}'.format(
                        basic_entry['entry-id'], time_ssid, tmention_id)
                    if arg_id in mention_id_set:
                        role_id = '{}-{}'.format(trigger_id, arg_id)
                        original_doc_inst['roles'][role_id] = {
                            'role_type': 'TimeAttachments',
                            'id': role_id,
                            'arguments': [
                                {
                                    'trigger_id': trigger_id
                                },
                                {
                                    'mention_id': arg_id
                                }
                            ]
                        }

                        sent_id_arg = original_doc_inst['mentions'][arg_id]['sent_id']

                        if sentdist2numargs is not None:
                            sentdist2numargs[np.abs(
                                sent_id_trg - sent_id_arg)] += 1
                        if sent_id_trg == sent_id_arg or (sent_id_trg, sent_id_arg) in merged_sent_pairs or (
                                sent_id_arg, sent_id_trg) in merged_sent_pairs:
                            new_event['time-attachments'][time_ssid] += 1

            sent_inst_list[sent_id_trg]['events'].append(new_event)
            event_id_2_sent_id[new_event['id']] = sent_id_trg
            event_id_2_event[new_event['id']] = new_event

            doc_event = deepcopy(new_event)
            doc_event['trigger']['start'] = start
            doc_event['trigger']['end'] = end + 1
            original_doc_inst['triggers'].append(doc_event['trigger'])
    # ------- read referred events
    for event_id, event in basic_events.items():
        trigger_span_id = event['anchors']
        for trg_m_id, trigger_mention in enumerate(
                span_sets[trigger_span_id]['spans']):  # multiple events sharing the same set of arguments
            start = 10000
            end = -1
            for charid in range(trigger_mention['start'], trigger_mention['end']):
                if charid in char2doc_token:
                    doc_token_id = char2doc_token[charid]
                    start = min(start, doc_token_id)
                    end = max(end, doc_token_id)
            if start == 10000 and end == -1:
                continue
            assert doc_token2sent_token[start][0] == doc_token2sent_token[end][0]
            sent_id_trg = doc_token2sent_token[start][0]
            # event relations
            for ref_ev_id in event['ref-events']:
                ref_trg_ssid = basic_events[ref_ev_id]['anchors']
                for ref_trg_mid, ref_trg_m in enumerate(span_sets[ref_trg_ssid]['spans']):
                    arg1_id = '{}-{}-{}-{}'.format(basic_entry['entry-id'], event_id, trigger_span_id,
                                                   trg_m_id)
                    arg2_id = '{}-{}-{}-{}'.format(basic_entry['entry-id'], ref_ev_id, ref_trg_ssid,
                                                   ref_trg_mid)
                    if arg1_id in event_id_2_sent_id and arg2_id in event_id_2_sent_id and \
                            event_id_2_sent_id[arg1_id] == event_id_2_sent_id[arg2_id]:
                        event_id_2_event[arg1_id]['ref-events'][ref_ev_id] += 1

    # ------- read templates
    template_id_2_sent_id = {}
    template_id_2_template = {}

    document_templates = []

    for template_id, template in granular_templates.items():
        anchor_span_id = template['template-anchor']
        if anchor_span_id not in span_sets:
            continue

        for anc_m_id, anchor_mention in enumerate(span_sets[anchor_span_id]['spans']):
            start = 10000
            end = -1
            for charid in range(anchor_mention['start'], anchor_mention['end']):
                if charid in char2doc_token:
                    doc_token_id = char2doc_token[charid]
                    start = min(start, doc_token_id)
                    end = max(end, doc_token_id)
            if start == 10000 and end == -1:
                continue
            assert doc_token2sent_token[start][0] == doc_token2sent_token[end][0]
            sent_id_anc = doc_token2sent_token[start][0]

            new_template = {
                'template-id': template['template-id'],
                'template-type': template['template-type'],
                'sentence': [t['text'] for t in sent_inst_list[sent_id_anc]['tokens']],
                'type': str(template['type']) if 'type' in template else 'O',
                'completion': str(template['completion']) if 'completion' in template else 'O',
                'over-time': str(template['over-time']) if 'over-time' in template else 'O',
                'coordinated': str(template['coordinated']) if 'coordinated' in template else 'O',
                'id': '{}-{}-{}-{}'.format(basic_entry['entry-id'], template_id, anchor_span_id, anc_m_id),
                'anchor': {
                    'id': '{}-{}-{}-{}'.format(basic_entry['entry-id'], template_id, anchor_span_id, anc_m_id),
                    'start': doc_token2sent_token[start][1],
                    'end': doc_token2sent_token[end][1] + 1,
                    'ori_tok_id': start,
                    'text': anchor_mention['string'],
                    'template-type': template['template-type']
                },
                'slots': defaultdict(list), 'irrealis': {}
            }
            if input_vocab:
                input_vocab['template-type'][template['template-type']] = input_vocab['template-type'].get(
                    template['template-type'], len(input_vocab['template-type']))
                if 'type' in template:
                    input_vocab['type_slot'][str(template['type'])] = input_vocab['type_slot'].get(
                        str(template['type']), len(input_vocab['type_slot']))
                if 'completion' in template:
                    input_vocab['completion_slot'][str(template['completion'])] = input_vocab['completion_slot'].get(
                        str(template['completion']), len(input_vocab['completion_slot']))
                if 'over-time' in template:
                    input_vocab['overtime_slot'][str(template['over-time'])] = input_vocab['overtime_slot'].get(
                        str(template['over-time']), len(input_vocab['overtime_slot']))
                if 'coordinated' in template:
                    input_vocab['coordinated_slot'][str(template['coordinated'])] = input_vocab['coordinated_slot'].get(
                        str(template['coordinated']), len(input_vocab['coordinated_slot']))

            for slot_name, fillers in template.items():
                slot_name = slot_name.lower()

                ############ transmogrify_outcome_slots ###########
                # conversion is based on 'bp_lib-v1.4.7/lib/bp.py/BPDocumentGranularTemplate.transmogrify_outcome_slots'
                if template['template-type'] in ['Corruplate', 'Protestplate', 'Terrorplate']:
                    if slot_name == "outcome-averted":
                        slot_name = "outcome"
                        if isinstance(fillers, list):
                            for filler in fillers:
                                filler['irrealis'] = 'counterfactual'

                    elif slot_name == "outcome-hypothetical":
                        slot_name = "outcome"
                        if isinstance(fillers, list):
                            for filler in fillers:
                                filler['irrealis'] = 'hypothetical'

                    elif slot_name == "outcome-occurred":
                        slot_name = "outcome"
                        if isinstance(fillers, list):
                            for filler in fillers:
                                if 'irrealis' in filler:
                                    del filler['irrealis']

                if isinstance(fillers, list):
                    if input_vocab:
                        input_vocab['slot_type'][slot_name] = input_vocab['slot_type'].get(slot_name,
                                                                                           len(input_vocab[
                                                                                                   'slot_type']))
                        if slot_name not in input_vocab['template-type2slots'][template['template-type']]:
                            input_vocab['template-type2slots'][template['template-type']].append(slot_name)

                    for filler in fillers:
                        if input_vocab and 'irrealis' in filler:
                            input_vocab['irrealis_type'][filler['irrealis']] = input_vocab['irrealis_type'].get(
                                filler['irrealis'], len(input_vocab['irrealis_type']))
                            if filler['irrealis'] not in input_vocab['slot2irrealis'][slot_name]:
                                input_vocab['slot2irrealis'][slot_name].append(filler['irrealis'])

                        if 'event-id' in filler:
                            if input_vocab:
                                input_vocab['slot2filler_types'][slot_name]['event-id'] = \
                                    input_vocab['slot2filler_types'][slot_name].get(
                                        'event-id', len(input_vocab['slot2filler_types'][slot_name])
                                    )
                            new_template['slots'][slot_name].append(filler['event-id'])
                            if 'irrealis' in filler:
                                new_template['irrealis'][filler['event-id']] = filler['irrealis']
                        elif 'ssid' in filler:
                            if input_vocab:
                                input_vocab['slot2filler_types'][slot_name]['ssid'] = input_vocab['slot2filler_types'][
                                    slot_name].get(
                                    'ssid', len(input_vocab['slot2filler_types'][slot_name])
                                )

                            new_template['slots'][slot_name].append(filler['ssid'])
                            if 'irrealis' in filler:
                                new_template['irrealis'][filler['ssid']] = filler['irrealis']

            sent_inst_list[sent_id_anc]['templates'].append(new_template)
            document_templates.append(new_template)
            template_id_2_sent_id[new_template['id']] = sent_id_anc
            template_id_2_template[new_template['id']] = new_template

            anchor_start_end = '{}-{}'.format(anchor_mention['start'], anchor_mention['end'])

            if anchor_start_end in se2evid:
                added = 0
                for found_event in evid2new_events[se2evid[anchor_start_end]]:
                    if found_event['sent-id'] == sent_id_anc:
                        trigger_start_end = found_event['start-end']
                        if trigger_start_end != anchor_start_end:
                            coref_template = deepcopy(new_template)
                            coref_template['sentence'] = [t['text'] for t in
                                                          sent_inst_list[found_event['sent-id']]['tokens']]
                            coref_template['anchor']['id'] = coref_template['anchor']['id'] + '-' + found_event[
                                'eventid']
                            coref_template['anchor']['start'] = found_event['trigger']['start']
                            coref_template['anchor']['end'] = found_event['trigger']['end']
                            coref_template['anchor']['text'] = found_event['trigger']['text']
                            coref_template['anchor']['ori_tok_id'] = found_event['trigger']['ori_tok_id']
                            coref_template['id'] = coref_template['id'] + '-' + found_event['eventid']
                            coref_template['eventid'] = found_event['eventid']

                            sent_inst_list[found_event['sent-id']]['templates'].append(coref_template)
                            document_templates.append(coref_template)

                            added += 1

                if added > 0:
                    print('entry-id: {}, template-id: {}, event-id: {}'.format(basic_entry['entry-id'],
                                                                               template['template-id'],
                                                                               found_event['eventid']))
                    print('adding {} coref templates...'.format(added))

    final_sent_insts = []
    sent_id = 0

    triggerid2sentid = {}
    trigger_list = []

    mentionid2sentid = {}
    mention_list = []
    while sent_id < len(sent_inst_list):
        if (sent_id, sent_id + 1) in merged_sent_pairs:
            new_sent = deepcopy(sent_inst_list[sent_id])

            new_sent['sent_id'] += '[+]' + \
                                   sent_inst_list[sent_id + 1]['sent_id']
            a1, b1 = sent_inst_list[sent_id]['start'], sent_inst_list[sent_id]['end']
            a2, b2 = sent_inst_list[sent_id +
                                    1]['start'], sent_inst_list[sent_id + 1]['end']
            new_sent['text'] = doc_text[a1: b1] + \
                               ' ' * (a2 - b1) + doc_text[a2: b2]
            new_sent['end'] = sent_inst_list[sent_id + 1]['end']
            new_sent['tokens'] += sent_inst_list[sent_id + 1]['tokens']
            # as the second sentence is appended to the first sentence.
            # trigger and mention spans need to be shifted

            shifted_templates = []
            for template in sent_inst_list[sent_id + 1]['templates']:
                template['anchor']['start'] += len(sent_inst_list[sent_id]['tokens'])
                template['anchor']['end'] += len(sent_inst_list[sent_id]['tokens'])
                shifted_templates.append(template)

            new_sent['templates'] += shifted_templates
            for template in new_sent['templates']:
                template['sentence'] = [t['text'] for t in new_sent['tokens']]

            shifted_events = []
            for event in sent_inst_list[sent_id + 1]['events']:
                event['trigger']['start'] += len(
                    sent_inst_list[sent_id]['tokens'])
                event['trigger']['end'] += len(
                    sent_inst_list[sent_id]['tokens'])
                shifted_events.append(event)

            new_sent['events'] += shifted_events
            shifted_mentions = []
            for mention in sent_inst_list[sent_id + 1]['mentions']:
                mention['start'] += len(sent_inst_list[sent_id]['tokens'])
                mention['end'] += len(sent_inst_list[sent_id]['tokens'])
                mention['head']['start'] += len(sent_inst_list[sent_id]['tokens'])
                mention['head']['end'] += len(sent_inst_list[sent_id]['tokens'])
                shifted_mentions.append(mention)

            new_sent['mentions'] += shifted_mentions
            new_sent['triggers'] = []
            for ev in new_sent['events']:
                new_sent['triggers'].append(ev['trigger'])
            final_sent_insts.append(new_sent)
            # for document instance
            for mention in new_sent['mentions']:
                mention_list.append(mention)
                mentionid2sentid[mention['id']] = len(final_sent_insts) - 1
            # for argument instance
            for ev in new_sent['events']:
                trigger_list.append(ev['trigger'])
                triggerid2sentid[ev['trigger']['id']] = len(
                    final_sent_insts) - 1

            sent_id += 2

        else:
            final_sent_insts.append(sent_inst_list[sent_id])
            # for document instance
            for mention in sent_inst_list[sent_id]['mentions']:
                mention_list.append(mention)
                mentionid2sentid[mention['id']] = len(final_sent_insts) - 1
            # for argument instance
            sent_inst_list[sent_id]['triggers'] = []
            for ev in sent_inst_list[sent_id]['events']:
                trigger_list.append(ev['trigger'])
                triggerid2sentid[ev['trigger']['id']] = len(
                    final_sent_insts) - 1

                sent_inst_list[sent_id]['triggers'].append(ev['trigger'])

            sent_id += 1

    if sentdist2numargs is not None:
        sentdist2numargs['total'] += len(original_doc_inst['roles'])
    tmp = list(original_doc_inst['mentions'].values())
    tmp.sort(key=lambda x: x['start'])
    original_doc_inst['mentions'] = tmp
    original_doc_inst['triggers'].sort(key=lambda x: x['start'])
    assert len(original_doc_inst['tokens']) == sum(
        [len(s['tokens']) for s in final_sent_insts])

    if model_name in ['trg', 'coref', 'template']:
        # add left and right context to each sentence
        padded_insts = []
        for sid, sent in enumerate(final_sent_insts):
            lid = max(0, sid - 1)
            rid = min(len(final_sent_insts) - 1, sid + 1)
            ##############################################
            cur_sent = deepcopy(sent)

            if lid < sid:
                cur_sent['left_tokens'] = final_sent_insts[lid]['tokens']
            else:
                cur_sent['left_tokens'] = []

            cur_sent['start_token'] = len(cur_sent['left_tokens'])
            cur_sent['end_token'] = len(cur_sent['left_tokens']) + len(cur_sent['tokens'])

            if rid > sid:
                cur_sent['right_tokens'] = final_sent_insts[rid]['tokens']
            else:
                cur_sent['right_tokens'] = []

            padded_insts.append(cur_sent)

        final_sent_insts = padded_insts
    else:
        for inst in final_sent_insts:
            inst['start_token'] = 0
            inst['end_token'] = len(inst['tokens'])
    if model_name not in ['template', 'slot']:
        return final_sent_insts
    else:
        return [{'granular-templates': document_templates, 'sentences': final_sent_insts}]


def print_and_log(message, logger, printout=True):
    if printout:
        print(message)
    logger.info(message)


def save_result(output_file, gold_graphs, pred_graphs, sent_ids, tokens=None):
    with open(output_file, 'w', encoding='utf-8') as w:
        for i, (gold_graph, pred_graph, sent_id) in enumerate(
                zip(gold_graphs, pred_graphs, sent_ids)):
            output = {'sent_id': sent_id,
                      'gold': gold_graph.to_dict(),
                      'pred': pred_graph.to_dict()}
            if tokens:
                output['tokens'] = tokens[i]
            w.write(json.dumps(output, ensure_ascii=False) + '\n')


def mention_to_tab(start, end, entity_type, mention_type, mention_id, tokens, token_ids, score=1):
    tokens = tokens[start:end]
    token_ids = token_ids[start:end]
    span = '{}:{}-{}'.format(token_ids[0].split(':')[0],
                             token_ids[0].split(':')[1].split('-')[0],
                             token_ids[1].split(':')[1].split('-')[1])
    mention_text = tokens[0]
    previous_end = int(token_ids[0].split(':')[1].split('-')[1])
    for token, token_id in zip(tokens[1:], token_ids[1:]):
        start, end = token_id.split(':')[1].split('-')
        start, end = int(start), int(end)
        mention_text += ' ' * (start - previous_end) + token
        previous_end = end
    return '\t'.join([
        'json2tab',
        mention_id,
        mention_text,
        span,
        'NIL',
        entity_type,
        mention_type,
        str(score)
    ])


def json_to_mention_results(input_dir, output_dir, file_name,
                            bio_separator=' '):
    mention_type_list = ['nam', 'nom', 'pro', 'nam+nom+pro']
    file_type_list = ['bio', 'tab']
    writers = {}
    for mention_type in mention_type_list:
        for file_type in file_type_list:
            output_file = os.path.join(output_dir, '{}.{}.{}'.format(file_name,
                                                                     mention_type,
                                                                     file_type))
            writers['{}_{}'.format(mention_type, file_type)
            ] = open(output_file, 'w')

    json_files = glob.glob(os.path.join(input_dir, '*.json'))
    for f in json_files:
        with open(f, 'r', encoding='utf-8') as r:
            for line in r:
                result = json.loads(line)
                doc_id = result['doc_id']
                tokens = result['tokens']
                token_ids = result['token_ids']
                bio_tokens = [[t, tid, 'O']
                              for t, tid in zip(tokens, token_ids)]
                # separate bio output
                for mention_type in ['NAM', 'NOM', 'PRO']:
                    tokens_tmp = deepcopy(bio_tokens)
                    for start, end, enttype, mentype in result['graph']['entities']:
                        if mention_type == mentype:
                            tokens_tmp[start] = 'B-{}'.format(enttype)
                            for token_idx in range(start + 1, end):
                                tokens_tmp[token_idx] = 'I-{}'.format(
                                    enttype)
                    writer = writers['{}_bio'.format(mention_type.lower())]
                    for token in tokens_tmp:
                        writer.write(bio_separator.join(token) + '\n')
                    writer.write('\n')
                # combined bio output
                tokens_tmp = deepcopy(bio_tokens)
                for start, end, enttype, _ in result['graph']['entities']:
                    tokens_tmp[start] = 'B-{}'.format(enttype)
                    for token_idx in range(start + 1, end):
                        tokens_tmp[token_idx] = 'I-{}'.format(enttype)
                writer = writers['nam+nom+pro_bio']
                for token in tokens_tmp:
                    writer.write(bio_separator.join(token) + '\n')
                writer.write('\n')
                # separate tab output
                for mention_type in ['NAM', 'NOM', 'PRO']:
                    writer = writers['{}_tab'.format(mention_type.lower())]
                    mention_count = 0
                    for start, end, enttype, mentype in result['graph']['entities']:
                        if mention_type == mentype:
                            mention_id = '{}-{}'.format(doc_id, mention_count)
                            tab_line = mention_to_tab(
                                start, end, enttype, mentype, mention_id, tokens, token_ids)
                            writer.write(tab_line + '\n')
                # combined tab output
                writer = writers['nam+nom+pro_tab']
                mention_count = 0
                for start, end, enttype, mentype in result['graph']['entities']:
                    mention_id = '{}-{}'.format(doc_id, mention_count)
                    tab_line = mention_to_tab(
                        start, end, enttype, mentype, mention_id, tokens, token_ids)
                    writer.write(tab_line + '\n')
    for w in writers:
        w.close()


def normalize_score(scores):
    min_score, max_score = min(scores), max(scores)
    if min_score == max_score:
        return [0] * len(scores)
    return [(s - min_score) / (max_score - min_score) for s in scores]


def best_score_by_task(log_file, task, max_epoch=1000):
    with open(log_file, 'r', encoding='utf-8') as r:
        config = r.readline()

        best_scores = []
        best_dev_score = 0
        for line in r:
            record = json.loads(line)
            dev = record['dev']
            test = record['test']
            epoch = record['epoch']
            if epoch > max_epoch:
                break
            if dev[task]['f'] > best_dev_score:
                best_dev_score = dev[task]['f']
                best_scores = [dev, test, epoch]

        print('Epoch: {}'.format(best_scores[-1]))
        tasks = ['entity', 'mention', 'relation', 'trigger_id', 'trigger',
                 'role_id', 'role']
        for t in tasks:
            print('{}: dev: {:.2f}, test: {:.2f}'.format(t,
                                                         best_scores[0][t][
                                                             'f'] * 100.0,
                                                         best_scores[1][t][
                                                             'f'] * 100.0))


def read_json(fpath):
    with open(fpath) as f:
        data = json.load(f)
    return data


def write_json(data, out_fpath):
    with open(out_fpath, 'w') as f:
        json.dump(data, f, ensure_ascii=False)


def find_sdp_path_exclusive(head, obj1_tokens, obj2_tokens,
                            actual_len):  # "exclusive" means: the output path doesnt contain obj1 and obj2 tokens
    # find dependency path
    common_ancestors = None

    subj_ancestors = set(obj1_tokens)
    for s in obj1_tokens:
        h = head[s]
        tmp = [s]
        while h > 0:
            head_token = h - 1
            tmp += [head_token]
            subj_ancestors.add(head_token)
            h = head[head_token]  # get parent node

        if common_ancestors is None:
            common_ancestors = set(tmp)
        else:
            # apply join operation between two sets
            common_ancestors.intersection_update(tmp)

    obj_ancestors = set(obj2_tokens)
    for o in obj2_tokens:
        h = head[o]
        tmp = [o]
        while h > 0:
            head_token = h - 1
            tmp += [head_token]
            obj_ancestors.add(head_token)
            h = head[head_token]
        common_ancestors.intersection_update(tmp)

    # find lowest common ancestor
    if len(common_ancestors) == 1:
        lca = list(common_ancestors)[0]
    else:
        child_count = {k: 0 for k in common_ancestors}
        for ca in common_ancestors:
            head_of_ca = head[ca]
            head_token_of_ca = head_of_ca - 1
            if head_of_ca > 0 and head_token_of_ca in common_ancestors:
                child_count[head_token_of_ca] += 1

        # the LCA has no child in the CA set
        for ca in common_ancestors:
            if child_count[ca] == 0:
                lca = ca
                break

    inside_path_nodes = subj_ancestors.union(
        obj_ancestors).difference(common_ancestors)
    inside_path_nodes.add(lca)
    inside_path_nodes = list(inside_path_nodes)
    inside_path_nodes.sort()
    return inside_path_nodes


def create_trigger_dataset(nlp):
    read_and_build(nlp, GRANULAR_TRAIN_P1P2P3,
                   'datasets/basic-events/granular.train.trg.jsonl',
                   lang='english', model_name='trg')
    read_and_build(
        nlp,
        GRANULAR_ANALYSIS_EMPTY_P3,
        'datasets/basic-events/granular.dev.trg.jsonl',
        lang='english', model_name='trg'
    )
    # real test file without any annotation
    read_and_build(
        nlp,
        GRANULAR_DEVTEST_EMPTY_P3,
        'datasets/basic-events/granular.test.trg.jsonl',
        lang='english', model_name='trg')

    # test files for target languages
    read_and_build(nlp, SAMPLE_GRANULAR_CHINESE_P3,
                   'datasets/basic-events/chinese.granular.test.trg.jsonl',
                   lang='chinese', model_name='trg')
    read_and_build(nlp, SAMPLE_GRANULAR_KOREAN_P3,
                   'datasets/basic-events/korean.granular.test.trg.jsonl',
                   lang='korean', model_name='trg')
    read_and_build(nlp, SAMPLE_GRANULAR_RUSSIAN_P3,
                   'datasets/basic-events/russian.granular.test.trg.jsonl',
                   lang='russian', model_name='trg')


def create_argument_dataset(nlp):
    log_dir = 'logs/trg/'
    dev_sysfile_chinese = os.path.join(log_dir, 'results.trigger.analysis.bp.json.lang-chinese.best')
    test_sysfile_chinese = os.path.join(log_dir, 'results.trigger.devtest.bp.json.lang-chinese.best')

    dev_sysfile_korean = os.path.join(log_dir, 'results.trigger.analysis.bp.json.lang-korean.best')
    test_sysfile_korean = os.path.join(log_dir, 'results.trigger.devtest.bp.json.lang-korean.best')

    dev_sysfile_russian = os.path.join(log_dir, 'results.trigger.analysis.bp.json.lang-russian.best')
    test_sysfile_russian = os.path.join(log_dir, 'results.trigger.devtest.bp.json.lang-russian.best')

    chinese_sysfile = os.path.join(log_dir, 'results.trigger.chinese-sample.bp.json.best')
    korean_sysfile = os.path.join(log_dir, 'results.trigger.korean-sample.bp.json.best')
    russian_sysfile = os.path.join(log_dir, 'results.trigger.russian-sample.bp.json.best')

    read_and_build(nlp, GRANULAR_TRAIN_P1P2P3,
                   'datasets/basic-events/granular.train.arg.jsonl',
                   lang='english', model_name='arg')

    read_and_build(
        nlp,
        dev_sysfile_chinese,
        'datasets/basic-events/granular.dev.arg.jsonl' + '.lang-chinese',
        lang='english',
        model_name='arg'
    )
    # real test file without any annotation
    read_and_build(
        nlp,
        test_sysfile_chinese,
        'datasets/basic-events/granular.test.arg.jsonl' + '.lang-chinese',
        lang='english',
        model_name='arg')

    read_and_build(
        nlp,
        dev_sysfile_korean,
        'datasets/basic-events/granular.dev.arg.jsonl' + '.lang-korean',
        lang='english',
        model_name='arg'
    )
    # real test file without any annotation
    read_and_build(
        nlp,
        test_sysfile_korean,
        'datasets/basic-events/granular.test.arg.jsonl' + '.lang-korean',
        lang='english',
        model_name='arg')

    read_and_build(
        nlp,
        dev_sysfile_russian,
        'datasets/basic-events/granular.dev.arg.jsonl' + '.lang-russian',
        lang='english',
        model_name='arg'
    )
    # real test file without any annotation
    read_and_build(
        nlp,
        test_sysfile_russian,
        'datasets/basic-events/granular.test.arg.jsonl' + '.lang-russian',
        lang='english',
        model_name='arg')

    # test files for target languages
    read_and_build(nlp, chinese_sysfile,
                   'datasets/basic-events/chinese.granular.test.arg.jsonl',
                   lang='chinese', model_name='arg')
    read_and_build(nlp, korean_sysfile,
                   'datasets/basic-events/korean.granular.test.arg.jsonl',
                   lang='korean', model_name='arg')
    read_and_build(nlp, russian_sysfile,
                   'datasets/basic-events/russian.granular.test.arg.jsonl',
                   lang='russian', model_name='arg')


def create_template_dataset(nlp):
    log_dir = 'logs/arg/'
    dev_sysfile_chinese = os.path.join(log_dir, 'results.arg.analysis.bp.json.lang-chinese.best')
    test_sysfile_chinese = os.path.join(log_dir, 'results.arg.devtest.bp.json.lang-chinese.best')

    dev_sysfile_korean = os.path.join(log_dir, 'results.arg.analysis.bp.json.lang-korean.best')
    test_sysfile_korean = os.path.join(log_dir, 'results.arg.devtest.bp.json.lang-korean.best')

    dev_sysfile_russian = os.path.join(log_dir, 'results.arg.analysis.bp.json.lang-russian.best')
    test_sysfile_russian = os.path.join(log_dir, 'results.arg.devtest.bp.json.lang-russian.best')

    chinese_sysfile = os.path.join(log_dir, 'results.arg.chinese-sample.bp.json.best')
    korean_sysfile = os.path.join(log_dir, 'results.arg.korean-sample.bp.json.best')
    russian_sysfile = os.path.join(log_dir, 'results.arg.russian-sample.bp.json.best')

    read_and_build(nlp, GRANULAR_TRAIN_P1P2P3,
                   'datasets/basic-events/granular.train.template.jsonl',
                   lang='english', model_name='template')

    read_and_build(
        nlp,
        dev_sysfile_chinese,
        'datasets/basic-events/granular.dev.template.jsonl' + '.lang-chinese',
        lang='english',
        model_name='template'
    )
    # real test file without any annotation
    read_and_build(
        nlp,
        test_sysfile_chinese,
        'datasets/basic-events/granular.test.template.jsonl' + '.lang-chinese',
        lang='english',
        model_name='template')

    read_and_build(
        nlp,
        dev_sysfile_korean,
        'datasets/basic-events/granular.dev.template.jsonl' + '.lang-korean',
        lang='english',
        model_name='template'
    )
    # real test file without any annotation
    read_and_build(
        nlp,
        test_sysfile_korean,
        'datasets/basic-events/granular.test.template.jsonl' + '.lang-korean',
        lang='english',
        model_name='template')

    read_and_build(
        nlp,
        dev_sysfile_russian,
        'datasets/basic-events/granular.dev.template.jsonl' + '.lang-russian',
        lang='english',
        model_name='template'
    )
    # real test file without any annotation
    read_and_build(
        nlp,
        test_sysfile_russian,
        'datasets/basic-events/granular.test.template.jsonl' + '.lang-russian',
        lang='english',
        model_name='template')

    # test files for target languages
    read_and_build(nlp, chinese_sysfile,
                   'datasets/basic-events/chinese.granular.test.template.jsonl',
                   lang='chinese', model_name='template')
    read_and_build(nlp, korean_sysfile,
                   'datasets/basic-events/korean.granular.test.template.jsonl',
                   lang='korean', model_name='template')
    read_and_build(nlp, russian_sysfile,
                   'datasets/basic-events/russian.granular.test.template.jsonl',
                   lang='russian', model_name='template')


def create_slot_dataset(nlp):
    log_dir = 'logs/template/'
    dev_sysfile_chinese = os.path.join(log_dir, 'results.template.analysis.bp.json.lang-chinese.best')
    test_sysfile_chinese = os.path.join(log_dir, 'results.template.devtest.bp.json.lang-chinese.best')

    dev_sysfile_korean = os.path.join(log_dir, 'results.template.analysis.bp.json.lang-korean.best')
    test_sysfile_korean = os.path.join(log_dir, 'results.template.devtest.bp.json.lang-korean.best')

    dev_sysfile_russian = os.path.join(log_dir, 'results.template.analysis.bp.json.lang-russian.best')
    test_sysfile_russian = os.path.join(log_dir, 'results.template.devtest.bp.json.lang-russian.best')

    chinese_sysfile = os.path.join(log_dir, 'results.template.chinese-sample.bp.json.best')
    korean_sysfile = os.path.join(log_dir, 'results.template.korean-sample.bp.json.best')
    russian_sysfile = os.path.join(log_dir, 'results.template.russian-sample.bp.json.best')

    read_and_build(nlp, GRANULAR_TRAIN_P1P2P3,
                   'datasets/basic-events/granular.train.slot.jsonl',
                   lang='english', model_name='slot')

    read_and_build(
        nlp,
        dev_sysfile_chinese,
        'datasets/basic-events/granular.dev.slot.jsonl' + '.lang-chinese',
        lang='english',
        model_name='slot'
    )
    # real test file without any annotation
    read_and_build(
        nlp,
        test_sysfile_chinese,
        'datasets/basic-events/granular.test.slot.jsonl' + '.lang-chinese',
        lang='english',
        model_name='slot')

    read_and_build(
        nlp,
        dev_sysfile_korean,
        'datasets/basic-events/granular.dev.slot.jsonl' + '.lang-korean',
        lang='english',
        model_name='slot'
    )
    # real test file without any annotation
    read_and_build(
        nlp,
        test_sysfile_korean,
        'datasets/basic-events/granular.test.slot.jsonl' + '.lang-korean',
        lang='english',
        model_name='slot')

    read_and_build(
        nlp,
        dev_sysfile_russian,
        'datasets/basic-events/granular.dev.slot.jsonl' + '.lang-russian',
        lang='english',
        model_name='slot'
    )
    # real test file without any annotation
    read_and_build(
        nlp,
        test_sysfile_russian,
        'datasets/basic-events/granular.test.slot.jsonl' + '.lang-russian',
        lang='english',
        model_name='slot')

    # test files for target languages
    read_and_build(nlp, chinese_sysfile,
                   'datasets/basic-events/chinese.granular.test.slot.jsonl',
                   lang='chinese', model_name='slot')
    read_and_build(nlp, korean_sysfile,
                   'datasets/basic-events/korean.granular.test.slot.jsonl',
                   lang='korean', model_name='slot')
    read_and_build(nlp, russian_sysfile,
                   'datasets/basic-events/russian.granular.test.slot.jsonl',
                   lang='russian', model_name='slot')


def create_event_coref_dataset(nlp):
    log_dir = 'logs/slot/'
    dev_sysfile_chinese = os.path.join(log_dir, 'results.slot.analysis.bp.json.lang-chinese.best')
    test_sysfile_chinese = os.path.join(log_dir, 'results.slot.devtest.bp.json.lang-chinese.best')

    dev_sysfile_korean = os.path.join(log_dir, 'results.slot.analysis.bp.json.lang-korean.best')
    test_sysfile_korean = os.path.join(log_dir, 'results.slot.devtest.bp.json.lang-korean.best')

    dev_sysfile_russian = os.path.join(log_dir, 'results.slot.analysis.bp.json.lang-russian.best')
    test_sysfile_russian = os.path.join(log_dir, 'results.slot.devtest.bp.json.lang-russian.best')

    chinese_sysfile = os.path.join(log_dir, 'results.slot.chinese-sample.bp.json.best')
    korean_sysfile = os.path.join(log_dir, 'results.slot.korean-sample.bp.json.best')
    russian_sysfile = os.path.join(log_dir, 'results.slot.russian-sample.bp.json.best')

    read_and_build(nlp, GRANULAR_TRAIN_P1P2P3,
                   'datasets/basic-events/granular.train.coref.jsonl',
                   lang='english', model_name='coref')

    read_and_build(
        nlp,
        dev_sysfile_chinese,
        'datasets/basic-events/granular.dev.coref.jsonl' + '.lang-chinese',
        lang='english',
        model_name='coref'
    )

    read_and_build(
        nlp,
        test_sysfile_chinese,
        'datasets/basic-events/granular.test.coref.jsonl' + '.lang-chinese',
        lang='english',
        model_name='coref')

    # ==============

    read_and_build(
        nlp,
        dev_sysfile_korean,
        'datasets/basic-events/granular.dev.coref.jsonl' + '.lang-korean',
        lang='english',
        model_name='coref'
    )

    read_and_build(
        nlp,
        test_sysfile_korean,
        'datasets/basic-events/granular.test.coref.jsonl' + '.lang-korean',
        lang='english',
        model_name='coref')

    # ==============

    read_and_build(
        nlp,
        dev_sysfile_russian,
        'datasets/basic-events/granular.dev.coref.jsonl' + '.lang-russian',
        lang='english',
        model_name='coref'
    )

    read_and_build(
        nlp,
        test_sysfile_russian,
        'datasets/basic-events/granular.test.coref.jsonl' + '.lang-russian',
        lang='english',
        model_name='coref')

    # ==============

    # test files for target languages
    read_and_build(nlp, chinese_sysfile,
                   'datasets/basic-events/chinese.granular.test.coref.jsonl',
                   lang='chinese',
                   model_name='coref')
    read_and_build(nlp, korean_sysfile,
                   'datasets/basic-events/korean.granular.test.coref.jsonl',
                   lang='korean',
                   model_name='coref')
    read_and_build(nlp, russian_sysfile,
                   'datasets/basic-events/russian.granular.test.coref.jsonl',
                   lang='russian',
                   model_name='coref')


if __name__ == '__main__':
    nlp = Pipeline('english', cache_dir='resource/trankit')
    nlp.add('chinese')
    nlp.add('arabic')
    nlp.add('persian')
    nlp.add('traditional-chinese')
    nlp.add('korean')
    nlp.add('russian')

    if sys.argv[1] == '0':
        build_vocabs(nlp, GRANULAR_TRAIN_P1P2P3, lang='english')
    elif sys.argv[1] == '1':
        create_trigger_dataset(nlp)
    elif sys.argv[1] == '2':
        create_argument_dataset(nlp)
    elif sys.argv[1] == '3':
        create_template_dataset(nlp)
    elif sys.argv[1] == '4':
        create_slot_dataset(nlp)
    elif sys.argv[1] == '5':
        create_event_coref_dataset(nlp)
