# Copyright (C) 2008 - 2021 The MITRE Corporation. See the toplevel
# file LICENSE.txt for license terms.

from __future__ import print_function, absolute_import

# Here, we define the various aspects of a workspace.

# The workspace structure is pretty simple: it contains a number of
# predefined folders, each of which contains files. It also contains
# a properties file, and a model directory. The name of the file is
# consistent across folders, so, e.g., "foo" in the "core"
# folder would be the same as "foo" in the "reconciliation" folder.

# The core folders will be:

# core: this is where documents are imported to. They must be converted to mat-json
# on import.
# reconciliation: this folder exists for reconciliation operations between
# different versions in core. It is not used in the initial 2.0
# implementation - it's part of the TooCAAn extensions.
# export: where completed, frozen documents optionally go. Also not used yet.

# This set should be extensible.

# Each folder can have a predefined set of operations. The operations can be applied
# to the entire contents of the folder, or a subset.

# The workspace will impose a number of restrictions which aren't imposed
# in the bare MATEngine. For a given task, the workflows and step sequences
# for each of the batch operations is predefined, as well as the default
# arguments for the model construction. There's only
# one current model at a time, and whether or not previous models are removed
# is a global setting on the workspace (it'll be on by default, because these
# models are huge). If email notification is desired, the appropriate outgoing
# SMTP server will need to be set (can't require a password).

# The command-line tool will support the following actions:

# - configure a new workspace: creates the directory structure, sets the global properties.
# - load documents into a workspace
# - do a workspace operation: with options to do it on all eligible documents
# or a subset of names listed in a file; send email to the invoker or not

# Email invocation isn't implemented yet.

import os, shutil, sys, time, re, codecs

class WorkspaceError(Exception):
    pass

class WorkspaceUserLockError(WorkspaceError):
    pass

class WorkspaceFileLockedError(WorkspaceError):
    pass

# MAT imports

from .Operation import (Operation, OpArgument, OperationError, XMLOpArgumentAggregator,
                        Option, DynamicChoiceOption, OptionTemplate)
from . import Error
from .DocumentIO import getDocumentIO
from . import DocumentIO
from . import ExecutionContext
from .ExecutionContext import Tmpdir                
from . import PluginMgr
from .PluginMgr import FindPluginClass
from . import ToolChain

from .ComparisonDocument import (generateComparisonDocument, reconciliationDocumentFullyReconciled,
                                 exportReconciliationDocument)
from . import Bootstrap
from . import WorkspaceLogger
from . import WorkspaceDB
from .PyCompat import _strtypes

#
# Transactions
#

# Transactions are implemented as atomic operations across the DB
# and file system. They support commit and rollback.

class WorkspaceTransaction(object):

    def __init__(self, ws, op, filesToPreserve = None, filesToAdd = None, filesToRemove = None):
        self.workspace = ws
        self.op = op
        self.filesToPreserve = []
        if filesToPreserve is not None:
            self.addFilesToPreserve(filesToPreserve)
        self.filesToAdd = []
        if filesToAdd is not None:
            self.addFilesToAdd(filesToAdd)
        self.filesToRemove = []
        if filesToRemove is not None:
            self.addFilesToRemove(filesToRemove)
        self.db = self.workspace.getDB()
        self.db.beginTransaction()
        if self.workspace.logger:
            self.workspace.logger.reset()

    # These functions will raise an error if they can't
    # stash the right backup file. This is the appropriate behavior
    # even if the transaction has been started (it'll be rolled back).

    def addFilesToPreserve(self, paths, check = False):
        for curPath in paths:
            curPath = os.path.realpath(curPath)
            if check and curPath in self.filesToPreserve:
                continue
            if not os.path.exists(curPath):
                raise WorkspaceError("couldn't find existing pathname %s to preserve for transaction" % curPath)
            try:
                shutil.copyfile(curPath, curPath + ".bak")
            except Exception as e:
                raise WorkspaceError("couldn't cache copy of existing pathname %s for transaction (error was: %s)" % (curPath, e))
            self.filesToPreserve.append(curPath)

    def addFilesToAdd(self, paths, check = False):
        for curPath in paths:
            curPath = os.path.realpath(curPath)
            if check and curPath in self.filesToPreserve:
                continue
            self.filesToAdd.append(curPath)

    def addFilesToRemove(self, paths, check = False):
        for curPath in paths:
            curPath = os.path.realpath(curPath)
            if check and curPath in self.filesToPreserve:
                continue
            if not os.path.exists(curPath):
                raise WorkspaceError("couldn't find existing pathname %s to remove for transaction" % curPath)
            try:
                shutil.copyfile(curPath, curPath + ".bak")
            except Exception as e:
                raise WorkspaceError("couldn't cache copy of existing pathname %s for transaction (error was: %s)" % (curPath, e))
            self.filesToRemove.append(curPath)

    def commit(self):
        self.db.commitTransaction()
        for p in self.filesToPreserve:
            try:
                os.remove(p + ".bak")
            except:
                pass
        for p in self.filesToRemove:
            try:
                os.remove(p + ".bak")
            except:
                pass
        if self.workspace.logger:
            self.workspace.logger.commit(self.op)
        self.workspace.currentTransaction = None

    def rollback(self):
        try:
            self.db.rollbackTransaction()
        except Exception as e:
            print(("Warning: failed to rollback DB (error was: %s)" % e), file=sys.stderr)
        for p in self.filesToPreserve:
            try:
                shutil.copyfile(p + ".bak", p)
            except Exception as e:
                print(("Warning: failed to restore %s during rollback (error was: %s)" % (p, e)), file=sys.stderr)
            try:
                os.remove(p + ".bak")
            except:
                pass
        for p in self.filesToRemove:
            try:
                shutil.copyfile(p + ".bak", p)
            except Exception as e:
                print(("Warning: failed to restore %s during rollback (error was: %s)" % (p, e)), file=sys.stderr)
            try:
                os.remove(p + ".bak")
            except:
                pass
        for p in self.filesToAdd:
            if os.path.exists(p):
                try:
                    if os.path.isdir(p):
                        shutil.rmtree(p)
                    else:
                        os.remove(p)
                except Exception as e:
                    print(("Warning: failed to remove %s during rollback" % p), file=sys.stderr)
        self.workspace.currentTransaction = None

#
# WorkspaceFolder
#

# Each folder has a set of operations. We can add a folder to a workspace.
# Each folder corresponds to a directory which is a subdir of the
# workspace dir.

class WorkspaceFolder(object):

    def __init__(self, workspace, folderName, operations = None,
                 description = "", importTarget = True,
                 prettyName = None):
        self.workspace = workspace
        self.folderName = folderName
        self.prettyName = prettyName or self.folderName
        self.description = description
        # These are a holdover from when there could be raw files
        # in the workspace. Now, there's nothing but the rich file type.
        self.fileType = self.workspace.richFileType
        self.docIO = self.workspace.richFileIO
        self.importTarget = importTarget
        self.dir = os.path.join(self.workspace.folderDir, folderName)
        self.operations = {}
        if operations is not None:
            for op in operations:
                self.addOperation(op.name, op)

    def create(self):
        os.makedirs(self.dir)
    
    # So the logic here should be that we try to load the document
    # using the workspace's rich file type, as a test. We can't
    # realistically guess in all cases when it's a rich file but
    # we imported raw, so we'll have to settle for trying the
    # workspace's default rich format (which, so far, is mat-json).

    # By the time we get here, in the new workspaces, we will have
    # been guaranteed that the files are already in the rich
    # format of the workspace - and should be a document object.

    # In other words, all of the complexity of this is gone.

    def importFile(self, doc, basename):
        self.workspace.richFileIO.writeToTarget(doc, os.path.join(self.dir, basename))

    # This is an instance of a child of MAT.Operation.

    def addOperation(self, name, op):
        if op.name is None:
            raise WorkspaceError("operation class '%s' has no name attribute" % op.__name__)
        if op in self.workspace.toplevelOperations:
            raise WorkspaceError("can't define folder operation '%s', since it's already defined as a toplevel operation" % op.name)
        self.operations[name] = op

    def addOptions(self, aggregator):
        for op in self.getCmdlineOperations():
            op.addOptions(aggregator, self.workspace)

    def getCmdlineOperations(self, debug = False):
        if debug:
            avail = CMDLINE_DEBUG_AVAILABLE_MASK
        else:
            avail = CMDLINE_AVAILABLE_MASK
        return [op for op in self.operations.values() if (op.availability & avail)]

    def clear(self, basenames = None):
        # Remove all documents from the folder.
        files = os.listdir(self.dir)
        for f in files:
            if (basenames is None) or (f in basenames):
                self.removeFile(f)

    def removeFile(self, f):
        p = os.path.join(self.dir, f)
        if os.path.isfile(p):
            os.remove(p)
            if self.workspace.logger:
                self.workspace.logger.logRemoveFile(self, f)

    def move(self, targetFolderName, basenames = None):
        try:
            targetFolder = self.workspace.folders[targetFolderName]
        except KeyError:
            raise WorkspaceError("can't move files to unknown target folder '%s'" % targetFolderName)

        files = os.listdir(self.dir)
        moved = []
        for f in files:
            if (basenames is None) or (f in basenames):
                p = os.path.join(self.dir, f)
                if os.path.isfile(p):
                    shutil.move(p, os.path.join(targetFolder.dir, f))
                    moved.append(f)
        if self.workspace.logger:
            self.workspace.logger.logMove(self, targetFolderName, basenames)
        return moved    

    def getFiles(self, basenames = None):
        # Returns a list of all the full paths for all the files
        # in the folder, or just the files listed in the list
        # of basenames. Let's go completely bonkers with list
        # comprehensions. I check with the workspace so
        # any stray files in the directory don't get passed along.
        return [os.path.join(self.dir, b) for b in self.getBasenames(basenames)]

    def openFile(self, basename):
        basenameExists = self.getBasenames(basenames = [basename])
        if basenameExists:
            return self._openFileBasename(basename)
        else:
            return None

    # saveFile semantics: expects a fileBasename.

    def saveFile(self, doc, basename):
        self.docIO.writeToTarget(doc, os.path.join(self.dir, basename))
        if self.workspace.logger:
            self.workspace.logger.logSaveFile(doc, self, basename)
            
    def getBasenames(self, basenames = None):
        d = os.listdir(self.dir)
        return [b for b in self.workspace.getBasenames(basenames) if b in d]

    # Stub for core folder, below. This must filter
    # out bad basenames.

    def fileBasenameForUser(self, basename, user):
        if self.getBasenames([basename]):
            return basename
        else:
            return None
    
    def fileBasenameLocked(self, fileBasename):
        return None

    def prepareForEditing(self, doc, fileBasename, user, lockId):
        raise WorkspaceError("folder is not editable")

    # This is needed when I can't check the basenames yet, in import.
    # It's also needed in Workspace._openFile, where we compute
    # the basename and do all the actual checking before we load.
    
    def _openFileBasename(self, fileBasename):
        p = os.path.join(self.dir, fileBasename)
        return self.docIO.readFromSource(p, taskSeed = self.workspace.task)

    def basenameForDocname(self, docName):
        return docName

    def getOperation(self, operationName, basenames = None, transaction = None):
        try:
            cls = self.operations[operationName]
        except KeyError:
            raise OperationError("no operation named '%s' in folder '%s'" % (operationName, self.folderName))
        # Instantiate the class.
        return cls(self, basenames = basenames, transaction = transaction)

    def listContents(self, basenames, user = None, read_only = False):
        return [{"basename": basename, "openable": True} for basename in basenames]

    def userHasAppropriateRole(self, user, fileBasename = None):
        return True

    def _addFileBasenameStatus(self, webResult, fileBasename):
        return

# And now, the core workspace folder. It relies a LOT on the DB to
# find basenames, etc. 

class CoreWorkspaceFolder(WorkspaceFolder):
        
    # Every basename is in core - you can only import to core.
    # No need to look at the directory listing - you'd just
    # have to dismantle the filenames for the assigned users.
    
    def getBasenames(self, basenames = None):
        allBasenames = self.workspace.getDB().allBasenames()
        if basenames is None:
            return allBasenames
        else:
            return list(set(basenames) & set(allBasenames))

    # getFiles() semantics: return a fleshed-out list of
    # all the actual pathnames for the basenames listed.
    # This is NOT sensitive to users; it returns ALL of them.

    def getFiles(self, basenames = None):
        d = self.workspace.getDB().documentsForBasenames(basenames)
        r = []
        for v in d.values():
            r += v
        return [os.path.join(self.dir, p) for p in r]

    def fileBasenameForUser(self, basename, user):
        # So the idea here is to return the docname which
        # is appropriate for this user. If the user is None,
        # it should be the raw basename. But we also
        # have to ensure the file is present. Well, this is the
        # core folder - of COURSE it's present.
        return self.workspace.getDB().documentForBasenameAndUser(basename, user)

    def basenameForDocname(self, docName):
        return self.workspace.getDB().basenameForDocname(docName)

    def fileBasenameLocked(self, fileBasename):
        return self.workspace.getDB().coreDocumentLocked(fileBasename)

    def prepareForEditing(self, doc, fileBasename, user, lockId):
        db = self.workspace.getDB()
        db.lockCoreDocument(lockId, fileBasename, user)

    def listContents(self, basenames, user = None, read_only = False):
        db = self.workspace.getDB()
        bPairs = []
        # Let's say it's OK if there's no user.
        userHasAnnotatorRole = (user is None) or self.userHasAppropriateRole(user)
        for docName, basename, status, assignedUser, lockedBy, wfStatus, currentStep in db.basenameInfo(basenames):
            info = {"basename": basename, "status": status,
                    "workflow status": wfStatus,
                    "current step": self.workspace.workflow.getStepByName(currentStep).getVisibleName()}
            if assignedUser:
                info["assigned to"] = assignedUser
            if docName != basename:
                info["doc name"] = docName
            if lockedBy:
                info["locked by"] = lockedBy
            # Compute openability.            
            if read_only:
                # All files are openable read-only.
                info["openable"] = True
            elif not userHasAnnotatorRole:
                info["openable"] = False
            elif assignedUser and (user != assignedUser):
                # If there's an assigned user and it ain't you, can't open.
                # Doesn't matter if there's a user or not.
                info["openable"] = False
            elif lockedBy and (lockedBy != user):
                # If the file is locked, but not by you, can't open.
                # Doesn't matter if there's a user or not.
                info["openable"] = False
            else:
                info["openable"] = True
            bPairs.append(info)
        return bPairs

    def userHasAppropriateRole(self, user, fileBasename = None):
        return self.workspace.getDB().userHasRole(user, "annotator")

    # This should always add at least status, I think.
    def _addFileBasenameStatus(self, webResult, fileBasename):
        webResult["status"], webResult["current_step"], webResult["workflow_status"] = self.workspace.getDB().coreDocumentStatuses(fileBasename)

#
# Individual folders and their operations. Also, workspace-level operations.
#

# Each of the operations controls what its command-line and Web
# results look like. I need to do this here so we can specialize the
# operation structure for individual tasks.

NULL_RESULT, FN_RESULT, WEB_RESULT = 0, 1, 2
_jsonIO = getDocumentIO('mat-json')

# Originally, the toplevel workspace operations weren't actual operation
# objects, but then I fixed that, because I needed to be able to specialize
# them. Rather than rename WorkspaceOperation everywhere (even though I
# probably should), I've introduced a toplevel WorkspaceGeneralOperation.

# CMDLINE_AVAILABLE covers 2 bit positions - i.e., everything that
# is CMDLINE_AVAILABLE is CMDLINE_DEBUG_AVAILABLE.

NOT_AVAILABLE, UI_AVAILABLE, CMDLINE_DEBUG_AVAILABLE, CMDLINE_AVAILABLE = 0, 1, 2, 6
UI_AVAILABLE_MASK, CMDLINE_DEBUG_AVAILABLE_MASK, CMDLINE_AVAILABLE_MASK = 1, 2, 4

class WorkspaceGeneralOperation(Operation):

    # This needs to be an attribute, because I need to look at
    # it in WorkspaceLogger.py and if I try, I'll get a
    # circular reference because I can't import Workspace in WorkspaceLogger.
    isToplevel = False

    availability = UI_AVAILABLE | CMDLINE_AVAILABLE

    def __init__(self, *args, **kw):
        Operation.__init__(self, *args, **kw)
        self.parameters = {}
        # See _runOperation below.
        self.fromCmdline = False
        
    # I'm going to use the option where we create instances of
    # each operation when we run. 

    # I need this because some of the information is
    # required sometime when the operation is called
    # (i.e., from the UI) and other times not, and I don't
    # want to have to compute the whole thing each time, since
    # it can get expensive. So I'll return a result object from
    # which the necessary results can be calculated. One of the
    # important parts is to store the operation and the folder
    # AND the digested options.

    # So first, let's be sure to stash the parameters. This is
    # a special case - used to be in Operation.py, but I only need
    # it here, and I need to pass in the workspace.
    
    # Introduce a special case for when the aggregator is actually
    # already an extracted dictionary.

    # The params are treated as defaults; the aggregator contents
    # get whomped on top of them.

    def enhanceAndValidate(self, aggregator, ws, reportDefaults = False, **params):
        if aggregator:
            self.addOptions(aggregator, ws)
            params = aggregator.extract(reportDefaults = reportDefaults, defaults = params or None)
            # I used to call validate() on the OptionBearer, but the fact is that
            # the option bearer no longer has all the arguments; it may have an
            # DynamicChoiceOption which is a promise to add more arguments. So I can no longer
            # really tell which argument belongs to which item - and I mostly don't
            # need to, because everyone who makes use of this mechanism use **kw
            # and I've made sure that arguments can't clash. And anyway, validate()
            # method is only used here, and this method is currently only called in
            # _runOperation() in Workspace.py, which means it's fed only by the
            # operation. Short of getting rid of this entirely, the easiest thing
            # to do is to move this method to the aggregator, and ignore where
            # the options came from.
            # There's the additional problem that the aggregator may not exist.
            self.parameters = aggregator.validate(self.__class__, **params)
        else:
            self.parameters = params
        
    def doOperation(self):
        self.do(**self.parameters)

    @classmethod
    def customizeUsage(cls, w):
        return None

    def fnResult(self):
        return None

    def webResult(self):
        return None

    @classmethod
    def getUsage(cls, w):
        return "No usage available for %s." % cls.name

    @classmethod
    def addOptions(cls, aggregator, w, **kw):
        super(WorkspaceGeneralOperation, cls).addOptions(aggregator, **kw)

class WorkspaceToplevelOperation(WorkspaceGeneralOperation):

    # This needs to be an attribute, because I need to look at
    # it in WorkspaceLogger.py and if I try, I'll get a
    # circular reference because I can't import Workspace in WorkspaceLogger.
    
    isToplevel = True

    availability = CMDLINE_AVAILABLE

    def __init__(self, workspace, args, transaction = None):
        WorkspaceGeneralOperation.__init__(self)
        self.workspace = workspace
        self.args = args
        self.transaction = transaction

    def getWorkspace(self):
        return self.workspace
    
    def getOperationSettings(self):
        return self.workspace.getOperationSettings(self.name)

class WorkspaceOperation(WorkspaceGeneralOperation):

    def __init__(self, folder, basenames = None, transaction = None):
        WorkspaceGeneralOperation.__init__(self)
        self.transaction = transaction
        self.folder = folder
        self.state = None
        # We need to know EXACTLY what basenames were operated on
        # in order to report the document output. But sometimes,
        # we need to recalculate it. It MUST be fixed before
        # the operation, because we use _getTargetDocuments()
        # to report on the documents that made it into the target
        # folder, based on the documents that were in this folder
        # BEFORE the operation happened. But we only
        # want to do that if inputBasenames was None.
        self.inputBasenames = basenames
        self.affectedBasenames = self.folder.getBasenames(self.inputBasenames)

    def getWorkspace(self):
        return self.folder.workspace
    
    def getOperationSettings(self):
        return self.folder.workspace.getOperationSettings(self.name)

    def getAffectedFolders(self):
        return []

    def getTargetFolderAndDocuments(self):
        return None, []

    # By default, we get all of them.
    def getAffectedFileBasenames(self):
        # Invert the dictionary.
        docDict = self.folder.workspace.getDB().documentsForBasenames(self.affectedBasenames)
        baseDict = {}
        for k, vals in docDict.items():
            for v in vals:
                baseDict[v] = k
        return baseDict

    # In order to get the target documents, we have to know
    # which basenames were actually accessed, which is a complication
    # when None was passed in. See __init__.
    
    def _getTargetDocuments(self, folderName):
        target = self.folder.workspace.folders[folderName]
        return [(basename, fileBasename, target._openFileBasename(fileBasename))
                for (fileBasename, basename) in self.getAffectedFileBasenames().items()]

    @classmethod
    def getUsage(cls, w):
        return cls._getUsage(cls.name, w)

    @classmethod
    def _getUsage(cls, clsName, w):
        return """Usage: %prog [options] <dir> """ + clsName + """ [operation_options] <folder> [ <basename> ... ]

<folder>: The name of the folder to operate on.
<basename>: (optional) The basename or basenames to restrict the operation to."""

    def webResult(self):
        affectedFolders = self.getAffectedFolders()
        targetFolder, affectedDocuments = self.getTargetFolderAndDocuments()
        result = {"success": True,
                  "affected_folders": affectedFolders}
        # We also need the target folder and the document contents
        # from that folder. There may be no target (either no
        # documents were affected or the folder hasn't changed).
        if targetFolder:
            result["target"] = targetFolder
        if affectedDocuments:
            basename, fileBasename, doc = affectedDocuments[0]
            result["basename"] = basename
            result["file_basename"] = fileBasename
            if basename != fileBasename:
                result["assigned_to"] = fileBasename[len(basename) + 1:]
            result["doc"] = _jsonIO.renderJSONObj(doc)
            self.folder._addFileBasenameStatus(result, fileBasename)
        return result

# I need a registration mechanism, because without it, there's an
# import circularity with the experiment engine operation between
# CarafeTrain and Workspace.

_OPERATION_REPOSITORY = {"core": {},
                         "review": {},
                         "export": {},
                         "reconciliation": {},
                         "toplevel": {}}

def WorkspaceOp(*cats):
    def _decorator(cls):
        if issubclass(cls, WorkspaceToplevelOperation):
            _OPERATION_REPOSITORY["toplevel"][cls.name] = cls
        elif issubclass(cls, WorkspaceOperation):
            for cat in cats:
                _OPERATION_REPOSITORY[cat][cls.name] = cls
            return cls
    return _decorator

#
# toplevel
#

# First engine-based operation: import.

# Can we import to anything besides core? In the basic situation, no. And I can't think of a case
# where you should be able to import to anything else. But I'm not going to enforce
# that quite yet.

# WORKFLOW STATES:
# awaiting hand annotation
# in hand annotation
# done
# awaiting review: awaiting a hand human review
# in review
# awaiting reconciliation partner: when there are multiple basenames, waiting for others which match it
# awaiting crossvalidation: awaiting crossvalidation in response to a review request
# in reconciliation

@WorkspaceOp()
class ImportOperation(WorkspaceToplevelOperation):

    name = "import"

    # GRRRR. If this is called when the class is DEFINED, I have
    # no guarantee that all the document IO is available yet.

    args = [Option("--strip_suffix", action="store",
                   dest="strip_suffix",
                   help="remove this suffix from the file name when determining the basename for the file in the workspace. By default, the original file basename is used."),
            Option("--encoding", action = "store",
                   dest="encoding",
                   help="for raw documents, input encoding. Default is utf-8. All imported raw documents will be converted to utf-8."),
            # The file type is the default for the folder, which is JSON for rich folders,
            # raw for raw folders. Note that with DynamicChoiceOption, the value is
            # set to the actual class.
            
            DynamicChoiceOption("--file_type", DocumentIO.DocumentIO, 
                                help = "The file type of the document. Default is mat-json.",
                                storeLiteral = True),
            Option("--workflow", dest = "workflow",
                   type = "string", metavar = "workflow",
                   help = "The workflow defined for the workspace is applied to each imported document, up to the first hand annotation step. However, this is sometimes not adequate to bring documents to the appropriate status; e.g., you may want to tokenize and and align a document which has been previously annotated. If --workflow is specified, it will be applied to each imported document before the workspace workflow is (possibly vacuously) applied. If --steps is specified, those steps will be applied in this preprocessing phase; otherwise, all the steps in the workflow will be applied."),
            Option("--steps", dest = "steps",
                   type = "string", metavar = "step(,step...)",
                   help = "The workflow defined for the workspace is applied to each imported document, up to the first hand annotation step. However, this is sometimes not adequate to bring documents to the appropriate status; e.g., you may want to tokenize and and align a document which has been previously annotated. If --workflow is specified, it will be applied to each imported document before the workspace workflow is (possibly vacuously) applied. If --steps is specified, those steps will be applied in this preprocessing phase; otherwise, all the steps in the workflow will be applied."),
            Option("--status_user", dest = "status_user",
                   type="string", metavar = "user",
                   help = "By default, the current workflow step for the document and its status are inferred from the document itself. But when the current step has annotations, the workspace may not be able to tell what user the annotations should be attributed to. You can use this flag to fix the attribution. If the document has been tagged automatically but not corrected by a human, use MACHINE as your value; otherwise, use the name of the registered workspace user. This value will overwrite any attributions that are already in the document. The import process will usually signal an error if there are annotated segments which are not attributed."),
            Option("--step_status",
                   type = "choice", choices = ["reconciled", "gold"],
                   help = "By default, the current workflow step for the document and its status are inferred from the document itself. But when the current step has annotations, the workspace sometimes can't tell whether the step is partially annotated, gold, or reconciled. The default is to judge the step to be partially annotated; use this option to change the value to gold or reconciled. If the document is already explicitly marked as gold or reconciled for that step, this setting will be ignored (so you can't use it, for instance, to make a reconciled step gold). If this setting must be imposed, all steps previous to the current workflow step will also be marked in the same way."),
            Option("--suppress_advancement", type = "boolean", default = False,
                   help = "By default, the workspace automatically advances a document through its workflow as far as it can, stopping at the first non-completed human-reviewable step. If this flag is present, it suppresses this automatic advancement in two ways. First, it blocks the application of any workspace models, so no pre-tagging is applied. Second, it completely blocks advancement past gold and reconciled steps, even if the following step is a non-trainable, completely automated step. If you choose this option, the advance operation will allow you to apply automatic advancement later."),
            Option("--defer_reconciliation", type = "boolean", default = False,
                   help = "By default, if the current workflow step is gold, the engine will automatically check for scheduled reviews, then advance to the next hand-annotatable step, or complete the workflow if there are no such steps. If you choose --step_status gold, or if the current workflow step is already marked gold, and this step is scheduled for reconciliation review, this flag will hold the document in a state where it is waiting for reconciliation partners rather than assume that there are no other reconciliation partners available and resort to human review instead."),
            Option("--assign_to_users", dest = "assign_to_users",
                   type="string", metavar = "user(,user...)",
                   help = "If present, assign the imported document to the specified user or users. MACHINE is not an eligible target. NOTE: If multiple users are assigned, and the documents imported are gold or marked gold in the current step, any scheduled reviews will be skipped."),
            Option("--add_to_basename_set", dest = "add_to_basename_set",
                   type = "string", metavar="setname",
                   help = "Add the basenames to a given basename set. You might want to do this if, e.g., you were importing gold-standard documents which you wanted to set aside for evaluation.")]

    @classmethod
    def getUsage(cls, w):
        return """Usage: %prog [options] <dir> import [import_options] <folder> <file> ...

<folder>: The name of the folder to import documents into.
<file>: The file to import into the folder (can be repeated).

Available folders are:\n\n""" + w.describeFolders(importableOnly = True)

    def fnResult(self):
        return self.imported

    def webResult(self):
        import codecs
        import json
        folder = self.args[0]
        # Only want the first basename.
        fp = codecs.open(os.path.join(self.workspace.folders[folder].dir, self.basenames[0]), "r", "utf-8")
        s = fp.read()
        fp.close()
        return {"success": True,
                "affected_folders": [folder],
                "target": folder,
                "doc": json.loads(s)}

    # This is true whether or not we're doing a preprocess step or an import step.
    def getOperationSettings(self):
        operationSettings = WorkspaceToplevelOperation.getOperationSettings(self)
        if operationSettings:
            for key in ["output_file_type", "output_encoding"]:
                if key in operationSettings:
                    raise WorkspaceError("import operation settings don't permit %s option to MATEngine", key)
        return operationSettings

    # For the transaction, we want to make sure we plan to remove those files.
    
    # mark_human_gold = None, mark_gold_standard_reference = False,
    
    def do(self, assign_to_users = None,
           status_user = None, step_status = None,
           encoding = "utf-8", file_type = None, fileIO = None, strip_suffix = None, workflow = None,
           steps = None, add_to_basename_set = None, suppress_advancement = False,
           defer_reconciliation = False, **kw):

        #
        # First section: set up the context.
        #
        
        # We need to apply the import operation settings here. It's essentially
        # identical to the old tagprep. We'll use the mixin, and then the wrapup()
        # operation will deal with the actual save.
        
        files = self.args[1:]
        self.folderName = self.args[0]
        self.markGold = (step_status == "gold") 
        self.markReconciled = (step_status == "reconciled")
        db = self.workspace.getDB()
        knownUsers = set(db.listUsers())
        if assign_to_users:
            assign_to_users = assign_to_users.split(",")
            if assign_to_users == [""]:
                assign_to_users = []
            if assign_to_users:
                if "MACHINE" in assign_to_users:
                    raise WorkspaceError("can't assign documents to MACHINE")
                if not set(assign_to_users) <= knownUsers:
                    raise WorkspaceError("can't assign to unknown users %s" %
                                           ", ".join(set(assign_to_users) - knownUsers))
                # Same checks as we do in the assign operation.
                for user in assign_to_users:
                    # And it better be a kosher user.
                    if not db.userHasRole(user, "annotator"):
                        raise WorkspaceError("can't assign documents to a user (%s) without the annotator role" % user)
        if self.markGold:
            if status_user is None:
                if not assign_to_users:
                    raise WorkspaceError("can't mark gold without status user or user assignment")
            elif status_user == "MACHINE":
                raise WorkspaceError("can't attribute gold documents to MACHINE")
            elif status_user not in knownUsers:
                raise WorkspaceError("can't mark gold with unknown user %s" % status_user)
            elif assign_to_users:
                raise WorkspaceError(("can't specify a separate status user for gold marking if documents are assigned"))
        self.statusUser = status_user
        self.assignToUsers = assign_to_users
        self.suppressAdvancement = suppress_advancement
        self.deferReconciliation = defer_reconciliation
        fromCmdline = self.fromCmdline
        stripSuffix = strip_suffix
        if (fileIO is None) and (file_type is None):
            fileIO = self.workspace.richFileIO

        if (steps is not None) and (workflow is None):
            raise WorkspaceError("preprocess steps specified without preprocess workflow")
        self.basenameSet = add_to_basename_set

        #
        # Second section: check folders.
        #
        
        try:
            folder = self.workspace.folders[self.folderName]
        except KeyError:
            raise WorkspaceError("can't import files into unknown folder %s" % self.folderName)

        if not folder.importTarget:
            raise WorkspaceError("can't import into unimportable folder '%s'" % self.folderName)
            
        # There can't be more than one file with the same
        # name in a workspace. So I need to maintain
        # a hash.

        #
        # Third section: preprocessing the filenames.
        #
        
        # Make sure I get the most up-to-date.
        # If there are multiple imports going on simultaneously,
        # we're hosed. Locks will take care of this.

        # For ease of interaction, let's check all the basenames
        # first, so we can raise an error.

        # This is a hash from basenames to documents. If we're assigning
        # to users, we need to check for the existing assignment.
        
        bHash = db.documentsForBasenames()
        
        filePairs = {}
        alreadyExist = []
        duplicateBasenames = []
        compareSignals = {}

        for file in files:
            # No whitespace, dudes.
            basename = os.path.basename(file).strip()
            if (stripSuffix is not None) and \
               basename.endswith(stripSuffix):
                basename = basename[:-len(stripSuffix)]
            # Remove all spaces so we can space-concatenate
            # basenames on the command line and when
            # we list the contents.
            basename = basename.replace(" ", "_")
            if basename in filePairs:
                duplicateBasenames.append(file)
            elif basename in bHash:
                if not assign_to_users:
                    alreadyExist.append(file)
                else:
                    docCands = bHash[basename]
                    found = False
                    for u in assign_to_users:
                        if basename + "_" + u in docCands:
                            alreadyExist.append(file)
                            found = True
                            break
                    if not found:
                        # Must check to make sure it's the same signal. Ugh. And the
                        # basename also can't have anything in review or reconciliation.
                        signal = self.workspace.folders["core"]._openFileBasename(docCands[0]).signal
                        if "REVIEW" in [r[3] for r in db.fileBasenameInfo(docCands).values()]:
                            # It's already in review. Barf.
                            raise WorkspaceError("Basename %s already has documents in review or reconciliation and can't be used as an import target" % basename)
                        compareSignals[file] = signal
                        filePairs[file] = basename
            else:
                filePairs[file] = basename

        if duplicateBasenames:
            raise WorkspaceError("Basenames for files %s end up being duplicates of other computed basenames in the import; no files imported." % ", ".join(duplicateBasenames))

        if alreadyExist:
            raise WorkspaceError("Basenames for files %s already exist in workspace; no files imported." % ", ".join(alreadyExist))

        # At this point, we know that each basename can be imported, and
        # assigned to whatever users are declared.

        #
        # Fourth section: actual import and DB update
        #

        runParameters = {"input_encoding": encoding,
                         "output_file_type": self.workspace.richFileType,
                         "output_encoding": "utf-8",
                         "input_file_type": file_type,
                         "inputFileType": fileIO}

        # Now, run the MAT engine to do the prep and import. The
        # import itself will happen in wrapup(). This is the only stuff
        # that needs to be in the transaction.

        initialStepOperationSettings, ignore, finalEntry = self.workspace._getAutomatedStepOperationSettings(suppressModelApplication = self.suppressAdvancement)
        t = self.workspace.beginTransaction(self)
        try:            
            self.transaction = t
            if workflow:
                # Preprocess. We start with the runParameters shown, and then
                # update them once the import is done.
                if not steps:
                    try:
                        wf = self.workspace.task.getWorkflows()[workflow]
                    except KeyError:
                        raise WorkspaceError("unknown workflow '%s'" % workflow)
                    steps = [s.stepName for s in wf.stepList]
                dataPairs = self.workspace._applyToolChain("import", self.getOperationSettings(),
                                                           {"workflow": workflow, "steps": steps},
                                                           # Use list() to freeze dictionary keys in Python 3.
                                                           runParameters, list(filePairs.keys()),
                                                           fromCmdline = self.fromCmdline,
                                                           # Only on import.
                                                           synchronizePhasesDone = True)
                # OK, we've completed that. Now, update the run parameters, etc.
                runParameters["input_file_type"] = self.workspace.richFileType
                runParameters["input_encoding"] = "utf-8"
                # When we get to _storePreparedFiles(), the idea is that the filePairs
                # must map from the path in the dataPairs to the basename. The operation mixin
                # does NOT save the files. wrapup() does this instead. So once
                # the preprocess is run, we've got the document objects, but they're
                # NOT SAVED, and then we have to feed them to the main operation,
                # but the default do() method expects to read them, not
                # get them already as document objects. So we have to save them
                # to a tmpdir, then update filePairs and run the main do().
                with Tmpdir() as tdir:
                    newFilePairs = {}
                    for path, iData in dataPairs:
                        b = filePairs[path]
                        tmpPath = os.path.join(tdir, b)
                        self.workspace.richFileIO.writeToTarget(iData, tmpPath)
                        newFilePairs[tmpPath] = b
                    filePairs = newFilePairs
                    # Now, we need to run the main workflow operation.
                    dataPairs = self.workspace._applyToolChain("import", self.getOperationSettings(),
                                                               initialStepOperationSettings,
                                                               # Use list() to freeze dictionary keys in Python 3.
                                                               runParameters, list(filePairs.keys()),
                                                               fromCmdline = self.fromCmdline)
            else:
                # In the current workspace workflow, we might end up
                # autotagging. 
                dataPairs = self.workspace._applyToolChain("import", self.getOperationSettings(),
                                                           initialStepOperationSettings,
                                                           # Use list() to freeze dictionary keys in Python 3.
                                                           runParameters, list(filePairs.keys()),
                                                           fromCmdline = self.fromCmdline,
                                                           # Only on import.
                                                           synchronizePhasesDone = True)
            self._storePreparedFiles(filePairs, dataPairs, compareSignals, initialStepOperationSettings["steps"])
            # And finally, mark every document pristine.
            # Use list() to freeze dictionary keys in Python 3.
            self.workspace.getDB().markBasenamesPristine(list(filePairs.values()))
            self.transaction = None
            t.commit()
        except:
            self.transaction = None
            t.rollback()
            raise

        imported = len(filePairs.keys())

        if fromCmdline:
            if imported == 1:
                plural = ""
            else:
                plural = "s"
            print("Imported %d file%s into folder '%s' in workspace %s." % (imported, plural, self.folderName, self.workspace.dir))

        self.imported = imported
        # Use list() to freeze dictionary keys in Python 3.
        self.basenames = list(filePairs.values())

    # Obviously, this can only be called from within a transaction.
    # Remember, there may not be any initial steps.
    
    def _storePreparedFiles(self, filePairs, dataPairs, compareSignals, initialSteps):
        
        f = self.workspace.getFolder(self.folderName)

        # The idea is that we need to take a look at each step, starting
        # with the last step for the import, and see what the status is for the doc.
        # It should be an error for any steps to be skipped beyond that.
        # Every step BEFORE this which is at least partially done will be
        # regarded as entirely done. I should make sure that if those
        # sets are managed, the documents are marked gold. But who do
        # we attribute them to? Does it matter? I'm going to assume not, for
        # the moment.

        # The last step of the import might be not gold (if it was autotagged,
        # for instance).

        sList = self.workspace.workflow.stepList
        doneDict = {}
        for path, iData in dataPairs:
            b = filePairs[path]
            if path in compareSignals:
                if iData.signal != compareSignals[path]:
                    raise WorkspaceError("file %s, to be imported as basename %s and assigned to users, does not have the same signal as other assigned documents for that basename" % (path, b))
            # If there are no initial steps, we've passed the initial steps.
            passedInitial = not initialSteps
            lastDoneStep = None
            foundUndoneStep = False
            for step in sList:
                if (not passedInitial) and (step.stepName == initialSteps[-1]):
                    passedInitial = True
                # BEFORE we hit the final step, make sure all the
                # managed ones are marked gold.
                if not passedInitial:
                    if isinstance(step, PluginMgr.AnnotationStep):
                        step.markGold(iData)
                # Now, from this point onward, I see if the
                # document is done for the step. Keep going
                # until I find a step where it isn't; in that case,
                # mark that I've found a stopping point, and KEEP GOING.
                # If after that, I find another done step, barf.
                else:
                    if not step.believedAtLeastPartiallyDone(iData):
                        # We only update the undone step if the last done step
                        # was really, really completely done.
                        lastWasCompletelyDone = False
                        if lastDoneStep:
                            lastWasCompletelyDone = lastDoneStep.believedCompletelyDone(iData)
                            if lastWasCompletelyDone:
                                lastDoneStep.markGold(iData)
                            lastDoneStep = None
                        # If we haven't updated the done status yet for an
                        # undone case, do it. We ALWAYS want to be in an
                        # undone step.
                        if (b not in doneDict) or \
                           ((not foundUndoneStep) and lastWasCompletelyDone):
                            foundUndoneStep = True
                            doneDict[b] = (step, self.workspace._documentStatus(iData, step, "import", None))
                    else:
                        # Mark it, and keep marking as long as
                        # we keep finding done steps. If it's managed,
                        # ask the document; otherwise, it's partially corrected.
                        # See _documentStatus for the logic.
                        if isinstance(step, PluginMgr.AnnotationStep) and step.managedSets:
                            step._ensureSegments(iData, "non-gold", None)
                        doneDict[b] = (step, self.workspace._documentStatus(iData, step, "import", None))
                        if lastDoneStep:
                            lastDoneStep.markGold(iData)
                        elif initialSteps and (step.stepName != initialSteps[-1]):
                            # If we're not in the initial step list,
                            # (i.e., passedInitial is false and it's
                            # not the last one either), and there's
                            # no lastDoneStep, we've skipped a step
                            # and we need to barf.
                            raise WorkspaceError("Basename candidate %s has intermittent completed steps in the workspace workflow after the initial import steps" % filePairs[path])
                        lastDoneStep = step

        # At this point, each element in the dataPair dict has an
        # unbroken sequence of completed steps from the beginning of
        # the workflow. We should also know what step we're on.
        # If the step is done, we can compute the status; otherwise,
        # it's "unannotated". We're not yet going to worry about
        # submitting to reconciliation or advance; either of these
        # things could happen to an imported doc, probably. Right now,
        # we're looking at the current step for each doc, and its
        # status.

        # This hasn't broken, so let's import all the files, and then
        # log the imports.

        # And account for the assignments.

        for path, iData in dataPairs:
            b = filePairs[path]
            if not self.assignToUsers:
                f.importFile(iData, b)
            else:
                for u in self.assignToUsers:
                    f.importFile(iData, b + "_" + u)

        # For the logger, we only need to log the basenames,
        # since the logger is as it happens after the assignment.
        # Use list() to freeze dictionary keys in Python 3.
        basenames = list(filePairs.values())
        if self.workspace.logger:
            self.workspace.logger.logImport(f, basenames)

        # Now, we have to insert it into the DB.
        # We have to infer the workflow state from the current step
        # and the document state.
        
        # For markgold and markreconciled, you can only do that for
        # docs that don't have that status yet. So let's subdivide them.

        dontMarkGoldOrReconciled = []
        maybeMarkGoldOrReconciled = []

        basenameList = []
        db = self.workspace.getDB()
        for b in basenames:            
            # openFile is not available until insertDocument is called, because
            # it now uses the DB.
            currentStep, currentStatus = doneDict[b]
            wfStatus = self.workspace._docStatusToWorkflowStatus(currentStatus)

            # So this is complicated. We will always be in an undone step
            # if there's a step we can be in. So if the wfStatus is "done",
            # we're at the end of the workflow, and we don't need to
            # mark it gold. Similarly, if the currentStatus is "unannotated",
            # we're at the beginning of a step, and the human hasn't had a
            # chance to do anything. So we SHOULDN'T mark that gold. Anything
            # else is fair game.

            if (wfStatus == "done") or (currentStatus == "unannotated"):
                dontMarkGoldOrReconciled.append(b)
            else:
                maybeMarkGoldOrReconciled.append(b)

            # If the wfStatus isn't awaiting hand annotation, then we
            # should recall who last touched it.

            if not self.assignToUsers:
                db.insertDocument(b, b, currentStatus, currentStep.stepName, wfStatus,
                                  lastTouchedBy = ((wfStatus != "awaiting hand annotation") and self.statusUser) or None)
            else:
                for u in self.assignToUsers:
                    # Not exactly true that it was last touched by this person, but I think for
                    # the purposes to which this slot is put, it's good enough.
                    db.insertDocument(b + "_" + u, b, currentStatus, currentStep.stepName, wfStatus, assignedUser = u,
                                      lastTouchedBy = ((wfStatus != "awaiting hand annotation") and u) or None)

        # So there are two things that can happen when you mark gold:
        # you can enter review if a review is scheduled for that step,
        # or you can auto-advance if it hasn't been suppressed.

        # The ones which haven't already been marked done can be marked
        # gold or reconciled. The rest have to be processed normally.

        processableBasenames = dontMarkGoldOrReconciled

        if self.markGold and maybeMarkGoldOrReconciled:
            # Get all the imported basenames, and submit them to the markGold operation.
            o = f.getOperation("markgold", basenames = maybeMarkGoldOrReconciled, transaction = self.transaction)
            o.fromCmdline = self.fromCmdline
            skipReviewCompletely = self.assignToUsers and len(self.assignToUsers) > 1
            for u in self.assignToUsers or [self.statusUser]:
                o.do(user = u, suppress_advancement = self.suppressAdvancement,
                     deferReconciliation = self.deferReconciliation,
                     skipReviewCompletely = skipReviewCompletely)
        elif self.markReconciled and maybeMarkGoldOrReconciled:
            if self.assignToUsers:
                fileBasenames = []
                for b in maybeMarkGoldOrReconciled:
                    fileBasenames += [b + "_" + u for u in self.assignToUsers]
            else:
                fileBasenames = maybeMarkGoldOrReconciled
            self.workspace._updateDocumentStatuses(fileBasenames, f,
                                                   self.transaction,
                                                   # Which segment annots do we update?
                                                   ["non-gold", "human gold"],
                                                   # What do we update them to?
                                                   "reconciled",
                                                   # Who do we attribute it to?
                                                   "GOLD_STANDARD")
            if self.fromCmdline:
                print("Marked reconciled:", " ".join(fileBasenames))
            if not self.suppressAdvancement:
                if self.assignToUsers:
                    advancingBasenames = []
                    for u in self.assignToUsers:
                        advancingBasenames += [b + "_" + u for b in maybeMarkGoldOrReconciled]
                    self.workspace._autoAdvance(advancingBasenames, self.transaction, self.fromCmdline)
                else:
                    self.workspace._autoAdvance(maybeMarkGoldOrReconciled, self.transaction, self.fromCmdline)
        else:
            processableBasenames += maybeMarkGoldOrReconciled

        if processableBasenames:
            # We want to make sure that every annotated segment has an attribution,
            # if it's managed. That means that if the set contains only spanned
            # annotations, I should do it by ordering the segments and finding
            # the segments which contain annotations; otherwise, if it has any
            # spanless annotations, I just do the entire segment, since it'll cover
            # the whole document anyway (for now).
            # Actually, it's a bit more complicated than that - it's gotta
            # be that the label in question appears ONLY in sets which only
            # have spanned labels.
            repo = self.workspace.task.getAnnotationTypeRepository()
            spannedLabels = set([lab for (lab, atype) in repo.items() if atype.hasSpan])
            spanlessLabels = set([lab for (lab, atype) in repo.items() if (not atype.hasSpan)])
            managedSetDict = {}
            for basename in processableBasenames:
                currentStep, currentStatus = doneDict[basename]
                if self.assignToUsers:
                    fileBasenames = [(u, basename + "_" + u) for u in self.assignToUsers]
                else:
                    fileBasenames = [(self.statusUser, basename)]
                if currentStep.managedSets:
                    try:
                        spanOnlySets, spanOnlyLabels, otherSetMap = managedSetDict[currentStep]
                    except KeyError:                        
                        spanOnlySets = set()
                        spanOnlyLabels = set()
                        otherSetMap = {}
                        otherLabels = set()
                        # First, find all the sets which contain at least one spanless label.
                        # Then, "invalidate" all the sets which contain any of the labels in
                        # those sets.
                        spannedCandidates = []
                        for s in currentStep.managedSets:
                            labs = set(self.workspace.task.getAnnotationLabelsForSetNames([s], includeAttributeOnly = True))
                            if labs & spanlessLabels:
                                otherLabels |= labs
                                otherSetMap[s] = labs
                            else:
                                spannedCandidates.append((s, labs))
                        if not otherSetMap:
                            for s, labs in spannedCandidates:
                                spanOnlySets.add(s)
                                spanOnlyLabels |= labs
                        else:
                            # Ah, now it's tricky. We need to loop until we either
                            # don't invalidate a set, or until we don't have any sets
                            # left to invalidate.
                            prevLen = len(spannedCandidates)
                            while True:
                                for i in range(len(spannedCandidates)):
                                    s, labs = spannedCandidates[i]
                                    if labs & otherLabels:
                                        # Invalidate it, break, and try again.
                                        otherLabels |= labs
                                        otherSetMap[s] = labs
                                        spannedCandidates[i:i+1] = []
                                        break
                                if (len(spannedCandidates) == prevLen) or (not spannedCandidates):
                                    break
                                prevLen = len(spannedCandidates)
                        for s, labs in spannedCandidates:
                            spanOnlySets.add(s)
                            spanOnlyLabels |= labs
                        managedSetDict[currentStep] = spanOnlySets, spanOnlyLabels, otherSetMap
                    for user, b in fileBasenames:
                        doc = f._openFileBasename(b)
                        # Copy them, because I'm going to modify the list, probably.
                        # Only look at the current step. Originally, I was looking at
                        # just the segments 
                        segAnnots = [seg for seg in doc.orderAnnotations(["SEGMENT"])
                                     if seg.get("set") in spanOnlySets]
                        modified = False
                        segChecked = False
                        for annot in doc.orderAnnotations(spanOnlyLabels):
                            while segAnnots and (annot.start >= segAnnots[0].end):
                                segAnnots[0:1] = []
                                segChecked = False
                            # If we run out of segment annotations, punt.
                            if not segAnnots:
                                break
                            # So we know the annotation starts before the current segment
                            # ends, and it can't actually be before it, can it? Let's check,
                            # just to be paranoid.
                            if (annot.start >= segAnnots[0].start) and \
                               (annot.end <= segAnnots[0].end):
                                segChecked = True
                                # Now, the real work. If there's an annotation,
                                # either (a) the segment has to have an attribution,
                                # or (b) there has to be self.users. If neither,
                                # it's an error.
                                if user is not None:
                                    if segAnnots[0].get("annotator") != "MACHINE":
                                        segAnnots[0]["annotator"] = user
                                        modified = True
                                elif not segAnnots[0].get("annotator"):
                                    raise WorkspaceError("imported document has annotated segment without an annotator (use the --status_user option to correct this error)")
                        # And now, do the other sets, if they have any annotations.
                        for s, labs in otherSetMap.items():
                            if doc.hasAnnotations(labs):
                                for segAnnot in [seg for seg in doc.orderAnnotations(["SEGMENT"])
                                                 if seg.get("set") == s]:
                                    if user is not None:
                                        if segAnnot.get("annotator") != "MACHINE":
                                            segAnnot["annotator"] = user
                                            modified = True
                                    elif not segAnnot.get("annotator"):
                                        raise WorkspaceError("imported document has annotated segment without an annotator (use the --status_user option to correct this error)")
                        
                        if modified:
                            f.saveFile(doc, b)
                            newStatus = self.workspace._documentStatus(doc, currentStep, "import", None)
                            if newStatus != currentStatus:
                                db.updateDocumentStatus(b, newStatus)
            # By user.
            if self.assignToUsers:
                # If they're gold, they should NOT BE REVIEWED if there are multiple
                # assignees. The multiple human reviews
                # will be superfluous, and the reconciliation review will simply eject it
                # automatically, and mark it reconciled, which we haven't asked for.
                # So there's no point in submitting for review, period.
                theseBasenames = []
                for u in self.assignToUsers:
                    theseBasenames += [basename + "_" + u for basename in processableBasenames]
                if len(self.assignToUsers) == 1:
                    self.workspace._submitForReview(theseBasenames, self.assignToUsers[0], self.transaction,
                                                    deferReconciliation = self.deferReconciliation,
                                                    fromCmdline = self.fromCmdline)
                if not self.suppressAdvancement:
                    self.workspace._autoAdvance(theseBasenames, self.transaction, self.fromCmdline)
            else:
                self.workspace._submitForReview(processableBasenames,
                                                self.statusUser, self.transaction, deferReconciliation = self.deferReconciliation,
                                                fromCmdline = self.fromCmdline)
                if not self.suppressAdvancement:
                    self.workspace._autoAdvance(processableBasenames, self.transaction, self.fromCmdline)

        # And now, let's add them to a basename set.
        if self.basenameSet:
            # I don't see a reason not to do this directly.
            self.workspace._addToBasenameSet(self.basenameSet, *basenames)

#
# The more sophisticated list operation. Note that if I want
# the toplevel getFolderBasenames to do something comparable,
# I need to specialize that, too. Let's fix that.
#

@WorkspaceOp()
class ListOperation(WorkspaceToplevelOperation):

    availability = UI_AVAILABLE | CMDLINE_AVAILABLE

    name = "list"

    args = [Option("--user", 
                   help="The workspace user who's listing the folder (used to compute the user's ability to open the file)"),
            Option("--read_only", type="boolean", default=False,
                   help = "Whether the workspace has been opened read-only or not.")]

    @classmethod
    def getUsage(cls, w):
        
        return """Usage: %prog [options] <dir> list ( <folder> ...)

<folder>: (optional) the name of the folder to list the contents of. For certain folders,
  extended information will be shown. If no folders are named, all folders will be listed.

Available folders are:\n\n""" + w.describeFolders()

    # For each basename, get the status in the DB. Note that
    # the "IN" keyword has no support in parameter substitution,
    # apparently for good reasons having to do with query plans.
    # I can either write an adaptor (which isn't possible because
    # I don't have access to a method that converts the parameters
    # to internal string literals) or I can format the string myself.
    
    def do(self, user = None, read_only = False):
        # Use list() to freeze dictionary keys in Python 3.
        basenameDict = self.workspace._getFolderBasenames(*(self.args or list(self.workspace.folders.keys())))
        # I want to rewrite the basename dict so that it contains a pair of
        # the basename and the info.
        self.basenameDict = {}
        # Use list() to freeze dictionary items in Python 3.
        for key, basenames in list(basenameDict.items()):
            folder = self.workspace.getFolder(key)
            bInfo = []
            self.basenameDict[key] = bInfo
            if basenames:
                bInfo += folder.listContents(basenames, user = user, read_only = read_only)
                        
        if self.fromCmdline:
            for key, bInfo in self.basenameDict.items():
                print(key + ":")
                for info in bInfo:
                    basename = info["basename"]
                    del info["basename"]
                    if info:
                        print("  %s (%s)" % (basename, ", ".join([k + " " + str(v) for k, v in info.items()])))
                    else:
                        print("  %s" % basename)
                print()                    
        
    def fnResult(self):
        return self.basenameDict

    def webResult(self):
        folder = self.args[0]
        return {"success": True,
                "basename_info": self.basenameDict[folder]}

@WorkspaceOp()
class RemoveOperation(WorkspaceToplevelOperation):

    name = "remove"

    @classmethod
    def getUsage(cls, w):
        return """Usage: %prog [options] <dir> remove <basename> ...

<basename>...: the basename(s) to be removed from the workspace.

Available basenames are:\n\n""" + " ".join(w.getBasenames())

    def do(self):
        t = self.workspace.beginTransaction(self)
        try:
            # I was just calling the core remove, but unfortunately, I need
            # to pass in the transaction.
            self.removed = self.workspace._removeBasenames(t, basenameList = self.args)
            if self.fromCmdline:
                if self.removed:
                    print("Removed", " ".join(self.removed), ".")
                else:
                    print("No basenames removed.")
            t.commit()
        except:
            t.rollback()
            raise

@WorkspaceOp()
class OpenFileOperation(WorkspaceToplevelOperation):

    availability = CMDLINE_DEBUG_AVAILABLE | UI_AVAILABLE

    name = "open_file"

    args = [Option("--user", type="string",
                   help = "open the file with a particular user. Required unless --read_only is present."),
            Option("--read_only", type="boolean", default=False,
                   help = "if present, open the document for reading and don't lock it"),
            Option("--file_basename", type="string",
                   help = "open a particular file basename, rather than choosing the proper one based on the user.")]

    @classmethod
    def getUsage(cls, w):
        return """Usage: %prog [options] <dir> open_file [open_options] folder basename

<folder>: the folder to find the file in
<basename>: the basename to open

Available basenames are:\n\n""" + " ".join(w.getBasenames())

    def do(self, user = None, read_only = False, file_basename = None):
        # First, see if the document is locked.
        folder, basename = self.args
        t = self.workspace.beginTransaction(self)
        try:
            self.doc, fileBasename, lockId = self.workspace._openFile(folder, basename, user, read_only, file_basename)
            t.commit()
        except:
            t.rollback()
            raise
        if self.fromCmdline and lockId:
            print("Locking %s with lock ID %s" % (basename, lockId))

        # The UI should probably know what it's opening. The
        # basename and file_basename and assigned_to keys are
        # intended to mirror the webResult() method for the file operations.
        
        self.webResultSeed = {"read_only": read_only, "lock_id": lockId,
                              "basename": basename, "file_basename": fileBasename}
        if basename != fileBasename:
            self.webResultSeed["assigned_to"] = fileBasename[len(basename) + 1:]
        self.workspace.folders[folder]._addFileBasenameStatus(self.webResultSeed, fileBasename)

    def webResult(self):
        doc = self.doc
        folder, file = self.args
        if doc is None:
            d = {"success": False,
                 "error": "Couldn't find basename '%s' in folder '%s'" % (file, folder)}
        else:
            d = {"success": True,
                 "doc": _jsonIO.renderJSONObj(doc)}
        d.update(self.webResultSeed)
        return d

    # Believe it or not, this is ONLY called in test suites.
    def fnResult(self):
        return self.doc, self.webResultSeed["lock_id"]

#
# Here's the assignment operator.
#

# Can we assign a document if it's locked? Only if we're copying an
# assigned document and not renaming an unassigned document.

@WorkspaceOp()
class AssignOperation(WorkspaceToplevelOperation):

    name = "assign"
    
    args = [Option("--user", type="string",
                   help = "assign the basename to the named user or users (comma-separated). Required.")]

    @classmethod
    def getUsage(cls, w):
        
        return """Usage: %prog [options] <dir> assign [assign_options] <basename>... 

<basename>...: the basename(s) to be assigned to the specified users."""


    def do(self, user = None, fromImport = False):
        _in_transaction = self.transaction
        db = self.workspace.getDB()
        if user is None:
            raise WorkspaceError("Can't assign a document without a user")
        users = user.split(",")
        for user in users:
            # And it better be a kosher user.
            if not db.userIsRegistered(user):
                raise WorkspaceError("can't lock a document using unregistered user '%s'" % user)
            if not db.userHasRole(user, "annotator"):
                raise WorkspaceError("can't assign documents to a user without the annotator role")
        if _in_transaction:
            self._do(db, _in_transaction, users, fromImport)
        else:
            t = self.workspace.beginTransaction(self)
            try:
                self._do(db, t, users, fromImport)
                t.commit()
            except:
                t.rollback()
                raise

    def _do(self, db, t, users, fromImport):
        coreFolder = self.workspace.getFolder("core")
        basenames = self.args        
        curUserMap = db.usersForBasenames(basenames)
        bMap = db.documentsForBasenames(basenames)
        for basename in basenames:
            # First, we check the assignments and status for the basename.
            curUsers = curUserMap[basename]
            if not curUsers:
                raise WorkspaceError("unknown basename %s" % basename)
            elif curUsers == [None]:
                # If it's unassigned, it must not be locked in order to continue,
                # and if fromImport is False, it must be either unannotated or uncorrected.
                if db.coreDocumentLocked(basename):
                    raise WorkspaceError("can't assign unassigned basename %s because it's locked" % basename)
                if not db.documentIsPristine(basename):
                    raise WorkspaceError("can't assign unassigned basename %s because it's not pristine" % basename)
                status, curStep, wfStatus = db.coreDocumentStatuses(basename)
                # OK, we're clear. Rename the document, assign the basename.
                # Make a copy for each user.
                # The pattern we're going to use is basename_<username>.
                # We do NOT try to get
                # it into the autotagger for segmenting and active learning,
                # since only workspace-wide autotagging can do that.
                t.addFilesToPreserve([os.path.join(coreFolder.dir, basename)])
                for user in users:
                    userBasename = basename + "_" + user
                    shutil.copyfile(os.path.join(coreFolder.dir, basename),
                                    os.path.join(coreFolder.dir, userBasename))
                    if self.workspace.logger:
                        self.workspace.logger.logCopy(coreFolder, basename, coreFolder, userBasename)
                    t.addFilesToRemove([os.path.join(coreFolder.dir, userBasename)])
                    db.insertDocument(userBasename, basename, status, curStep, wfStatus, user)
                coreFolder.removeFile(basename)
                db.removeUnassignedBasenameEntry(basename)
            elif None in curUsers:
                raise WorkspaceError("basename % is both assigned and unassigned" % basename)            
            else:
                if fromImport:
                    raise WorkspaceError("basename %s can't be already assigned if it's just been imported" % basename)
                # If it's already assigned,  it can't be assigned to the user already. We need to find
                # a pristine copy.                

                userOverlap = set(users) & set(curUserMap[basename])
                if userOverlap:
                    raise WorkspaceError("users %s already assigned to basename %s" % \
                                           (",".join(userOverlap), basename))
                pristineDocname = None
                for docName in bMap[basename]:
                    if db.documentIsPristine(docName):
                        pristineDocname = docName
                        break
                if pristineDocname is None:
                    raise WorkspaceError("no pristine assigned document found for basename %s" % basename)
                status, curStep, wfStatus = db.coreDocumentStatuses(pristineDocname)
                # From here on, it's just like the other case, mostly.
                for user in users:
                    userBasename = basename + "_" + user
                    shutil.copyfile(os.path.join(coreFolder.dir, pristineDocname),
                                    os.path.join(coreFolder.dir, userBasename))
                    if self.workspace.logger:
                        self.workspace.logger.logCopy(coreFolder, pristineDocname, coreFolder, userBasename)
                    t.addFilesToRemove([os.path.join(coreFolder.dir, userBasename)])
                    db.insertDocument(userBasename, basename, status, curStep, wfStatus, user)

# We're going to make these all be classmethods.

def _computeRoles(roles, doingWhat):
    if roles is None:
        return ["annotator"]
    elif roles == "all":
        return ["annotator", "reviewer"]
    else:
        if type(roles) in _strtypes:
            inRoles = [r.strip() for r in roles.split(",")]
        else:
            inRoles = roles
        roles = set()
        for role in inRoles:
            if role == "all":
                roles.update(["annotator", "reviewer"])
            elif role not in ("annotator", "reviewer"):
                raise WorkspaceError("Unknown user role '%s', not %s." % (role, doingWhat))
            else:
                roles.add(role)
        return list(roles)

@WorkspaceOp()
class RegisterUserOperation(WorkspaceToplevelOperation):

    name = "register_users"

    args = [Option("--roles", type = "multiple_choice",
                   choices = ["all", "annotator", "reviewer"],
                   help = "Available roles are 'annotator', 'reviewer'. If omitted, the role will be 'annotator'. The string 'all' adds both roles. Otherwise, a comma-separated list of roles.")]

    @classmethod
    def getUsage(cls, w):
        return """Usage: %prog [options] <dir> register_users [register_user_options] <user>...

<user>: the name of a user to register for the workspace."""

    def do(self, roles = None):
        # Workspace is already locked. I'm not sure
        # this is going to be supported except on the command line.
        t = self.workspace.beginTransaction(self)
        try:            
            self.workspace._registerUsers(self.args, _computeRoles(roles, "registering users"))
            if self.fromCmdline:                
                o = self.workspace.getOperation("list_users", ())
                o.fromCmdline = True
                o.do()
            t.commit()
        except:
            t.rollback()
            raise

@WorkspaceOp()
class ListUserOperation(WorkspaceToplevelOperation):

    args = [OpArgument("no_roles",
                       help = "don't show the roles")]
    @classmethod
    def getUsage(cls, w):
        return """Usage: %prog [options] <dir> list_users [list_user_options]"""

    name = "list_users"

    @classmethod
    def getUsage(cls, w):
        return """Usage: %prog [options] <dir> list_users"""

    def do(self, no_roles = False):
        # Workspace is already locked. I'm not sure
        # this is going to be supported except on the command line.
        users = self.workspace._listUsers(no_roles = no_roles)
        if no_roles:
            if self.fromCmdline:
                for u in users: print(u)
            self.users = users
        else:
            if self.fromCmdline:
                for u, roles in users.items():
                    print(u, ":", " ".join(roles))
            self.users = users

    # No way yet to add new toplevel operations to the Web service, unfortunately.
    
    def fnResult(self):
        return self.users

    def webResult(self):
        # I never return roles.
        users = self.users
        if type(users) is dict:
            # Use list() to freeze dictionary keys in Python 3.
            users = list(users.keys())
        return {"success": True,
                "users": users}
                
#
# Here's a debugging object for you
#

@WorkspaceOp()
class DumpDatabaseOperation(WorkspaceToplevelOperation):

    name = "dump_database"

    @classmethod
    def getUsage(cls, w):
        db = w.getDB()
        return """Usage: %prog [options] <dir> dump_database [<table>...]

<table>: the name of a table to dump. If none are provided, the entire database will be dumped. Table names are: """ + " ".join(db.getTables())

    def do(self):
        db = self.workspace.getDB()
        data = db.dumpDatabase(tables = self.args or None)
        if self.fromCmdline:
            for d in data:
                print(d["table"])
                print("=" * len(d["table"]))
                strData = [[str(a) for a in r] for r in d["data"]]
                widths = [max([len(d["columns"][i])] + [len(r[i]) for r in strData]) for i in range(len(d["columns"]))]
                fmtString = "  ".join(["%-"+str(i)+"s" for i in widths])
                print(fmtString % tuple(d["columns"]))
                print(fmtString % tuple(["-"*i for i in widths]))
                for row in strData:
                    print(fmtString % tuple(row))
                print()

# For reference and completeness, I'm putting this here, even though
# it shows nothing interesting besides the list of users in MAT 2.0 initial.

@WorkspaceOp()
class WorkspaceConfigurationOperation(WorkspaceToplevelOperation):
    
    name = "workspace_configuration"

    @classmethod
    def getUsage(cls, w):
        return "Usage: %prog [options] <dir> workspace_configuration"

    def do(self):
        self.users = self.workspace.getDB().listUsers()
        self.prioritizationName = self.workspace.getPrioritizationClassName()
        if self.fromCmdline:
            print("Task:", self.workspace.task.name)
            print("Users:", ", ".join(self.users))
            print("Workflow: %s (workspace configuration '%s')" % (self.workspace.workflow.name, self.workspace.wsConfig))
            print("Language: %s" % (self.workspace.language))
            print("Logging:", (self.workspace.loggingEnabled and "yes") or "no")
            print("Prioritization:", (self.prioritizationName or "no"))
            if self.prioritizationName:
                print("Prioritization mode:", self.workspace.getPrioritizationMode())

    def webResult(self):
        return {"success": True,
                "task": self.workspace.task.name,
                "users": self.users,
                "workflow": self.workspace.workflow.name,
                "language": self.workspace.language,
                "operation_settings": self.workspace.operationSettings,
                "prioritization_class": self.prioritizationName,
                "prioritization_mode": self.workspace.getPrioritizationMode(),
                "logging_enabled": self.workspace.loggingEnabled}

# Stuff for experiments.

@WorkspaceOp()
class ListBasenameSetsOperation(WorkspaceToplevelOperation):

    availability = CMDLINE_AVAILABLE

    name = "list_basename_sets"

    @classmethod
    def getUsage(cls, w):
        return """Usage: %prog [options] <dir> list_basename_sets ( <set_name>... )

<set_name>: the name of a basename set"""

    def do(self):
        setMap = self.workspace.getDB().getBasenameSetMap()
        if self.fromCmdline:
            for setName, setContents in setMap.items():
                if (not self.args) or (setName in self.args):
                    print("Basename set:", setName)
                    print("Basenames:", ", ".join(setContents))
                    print()

@WorkspaceOp()
class AddToBasenameSetOperation(WorkspaceToplevelOperation):

    availability = CMDLINE_AVAILABLE

    name = "add_to_basename_set"

    @classmethod
    def getUsage(cls, w):
        return """Usage: %prog [options] <dir> add_to_basename_set <set_name> <basename>...

<set_name>: the name of a basename set
<basename>: a known workspace basename"""

    def do(self):
        t = self.workspace.beginTransaction(self)
        try:
            self.workspace._addToBasenameSet(self.args[0], *self.args[1:])
            t.commit()
        except:
            t.rollback()
            raise

@WorkspaceOp()
class RemoveFromBasenameSetOperation(WorkspaceToplevelOperation):

    availability = CMDLINE_AVAILABLE

    name = "remove_from_basename_set"

    @classmethod
    def getUsage(cls, w):
        return """Usage: %prog [options] <dir> remove_from_basename_set <set_name> <basename>...

<set_name>: the name of a basename set
<basename>: a workspace basename"""

    def do(self):
        t = self.workspace.beginTransaction(self)
        try:
            self.workspace._removeFromBasenameSet(self.args[0], *self.args[1:])
            t.commit()
        except:
            t.rollback()
            raise

#
# core folder
#

# Autotag. Operation on "core".
# Tags the documents, marks them as uncorrected.

# Uses the current model. Whether or not it's local depends on
# how it's invoked. By default, it uses the server, but it might
# be local. Right now, it's always local.

# The autotag operation applies to all documents, and tags those
# SEGMENTs which are either unassigned or assigned to MACHINE.

# What if there are basenames specified to autotag? We
# don't want to mix documents autotagged by previous models, but we
# don't want to affect too many documents.

# If specific files are requested, the file-level behavior
# (gathering all the documents) is bypassed.

@WorkspaceOp("core")
class AdvanceOperation(WorkspaceOperation):

    name = "advance"

    # Like markgold, I'm adding the option of providing parameters here. I
    # don't do it everywhere, unfortunately, because it would be all over the place
    # in completing review. Perhaps advance should be automatically disabled when
    # exiting review?
    
    @classmethod
    def addOptions(cls, aggregator, ws, **kw):
            
        super(AdvanceOperation, cls).addOptions(aggregator, ws, **kw)
        ws.advancerClass.addOptions(aggregator, ws, **kw)

    # I was hoping to be able to get away without specifying this, but
    # the Web service requires it in order to extract the arguments from the CGI request.
    
    args = [Option("--lock_id", type="string", help="lock ID (if document is locked)")]

    def getAffectedFolders(self):
        return self.advancer.getAffectedFolders()

    def getTargetFolderAndDocuments(self):
        return self.advancer.getTargetFolderAndDocuments()

    def getAffectedFileBasenames(self):
        return getattr(self, "affectedFileBasenames", {})

    # lock_id is only
    # used from the UI. If the requested basenames have a lock that doesn't
    # match the lock ID, you can't do anything.

    # Note, too, that updating the status in a particular step should NOT rely on
    # the document status if the sets aren't managed - in all cases, the
    # proper status is "uncorrected". See _documentStatus for the logic.
    
    def do(self, lock_id = None, **params):
        t = self.folder.workspace.beginTransaction(self)
        try:
            self._do(t, lock_id, params)
            t.commit()
        except:
            t.rollback()
            raise
            
    def _do(self, transaction, lock_id, wfParams):

        db = self.folder.workspace.getDB()
        # If there's a lock ID, prune this down to just those docs.
        # It's checked again in _autoAdvance, but that's just the way it goes.
        
        if lock_id is not None:
            fileBasename, basename, lockingUser = db.coreGetLockIDInfo(lock_id)
            for b in self.affectedBasenames:
                if b != basename:
                    raise WorkspaceError("wrong lock ID %s for basename %s" % (lock_id, b))
            docsToBasenames = {fileBasename: basename}
        else:
            # Convert this from basenames to doc names.
            bDict = db.documentsForBasenames(self.affectedBasenames)
            # Invert this, since I'm going to need the docnames back.
            docsToBasenames = {}
            for k, vList in bDict.items():
                for v in vList:
                    docsToBasenames[v] = k            

        # Use list() to freeze dictionary keys in Python 3.
        self.advancer = self.folder.workspace._autoAdvance(list(docsToBasenames.keys()), transaction, self.fromCmdline,
                                                           lockId = lock_id, workflowParams = wfParams)
        self.affectedFileBasenames = dict([(k, v) for (k, v) in docsToBasenames.items()
                                           if k in self.advancer.affectedFileBasenames])

# The save operation has to do save the document, but ALSO it has to
# check to see whether the document should pop the queue. And if all
# the segments are no longer marked MACHINE, and there's at least one SEGMENT,
# mark it completed.

# This save operation is ONLY for the core folder. There's a separate
# save operation for the reconciliation.

# From a transaction point of view, I need to ensure that when we roll
# back, we roll back to the original content of the file. So I ALWAYS
# have to stash the original file somewhere.

# For experimental purposes, we can get log fragments on every save.
# The first log fragments, at the moment, are XML, from Callisto (ugh). I'd
# really prefer to get CSV logs, but beggars can't be choosers. If 
# --log is present (and I'm declaring it in the args ONLY
# so I can get it via the Web service), I need to stash the file
# being saved, along with the log fragment, somewhere, checkpointed.

# When you save, and the document status changes, you have to return
# info that forces a folder update.

@WorkspaceOp("core")
class SaveOperation(WorkspaceOperation):

    name = "save"

    args = [OpArgument("doc", help = "document to save, as a JSON string", hasArg = True),
            Option("--lock_id", type="string", help="transaction ID for save"),
            Option("--release_lock", type="boolean", default=False, help="release the lock after save"),
            OpArgument("log", help = "log fragment (from UI) to trigger fine-grained progress monitoring", hasArg = True),
            OpArgument("log_format", help = "the format of the log fragment", hasArg = True),
            OpArgument("timestamp", help = "millisecond timestamp of log upload from the UI's point of view", hasArg = True),
            OpArgument("next_op", hasArg = True, help = "JSON string describing operation to perform after the save (for UI connectivity, mostly)")]

    availability = CMDLINE_DEBUG_AVAILABLE | UI_AVAILABLE

    def __init__(self, *args, **kw):
        WorkspaceOperation.__init__(self, *args, **kw)
        self.statusOp = None

    def getAffectedFolders(self):
        # Only markgold.
        if self.state == "statusop":
            return self.statusOp.getAffectedFolders()
        elif self.state == "statuschanged":
            return [self.folder.prettyName]
        else:
            return []

    def getTargetFolderAndDocuments(self):
        if self.state == "statusop":
            return self.statusOp.getTargetFolderAndDocuments()
        elif self.state == "statuschanged":
            return self.folder.prettyName, self._getTargetDocuments(self.folder.prettyName)
        else:
            return None, []

    def do(self, doc = None, next_op = None, lock_id = None, release_lock = False,
           log = None, log_format = None, timestamp = None):
        if lock_id is None:
            raise WorkspaceError("can't save without lock ID")
        # Now we get the basename. Must check to ensure that
        # the lock ID matches. Need to get the file basename
        # from the transaction.
        db = self.folder.workspace.getDB()
        fileBasename, basename, user = db.coreGetLockIDInfo(lock_id)
        if basename != self.affectedBasenames[0]:
            raise WorkspaceError("wrong lock ID %s for basename %s" % (lock_id, self.affectedBasenames[0]))
        self.affectedFileBasename = fileBasename
        t = self.folder.workspace.beginTransaction(self, filesToPreserve = [os.path.join(self.folder.dir, fileBasename)])
        try:
            curDocStatus, curStep, wfStatus = db.coreDocumentStatuses(fileBasename)
            # Do this as early as possible, so that the timings are as close to
            # the final log entry in the UI as we can make them.
            if log and self.folder.workspace.logger:
                self.folder.workspace.logger.logLog(log, log_format, timestamp)
            if doc is not None:
                # It can be none, if it's not dirty.
                # First, make it into a document. The document
                # string is almost certainly not Unicode yet.
                docObj = self.folder.docIO.readFromByteSequence(doc, 'utf-8')

                # There better only be one basename.
                self.folder.saveFile(docObj, fileBasename)
            else:
                docObj = self.folder._openFileBasename(fileBasename)

            # It used to  be that if the document had no segments
            # in the priority queue, it was done. But that's not really right.
            # The segments may not have been assigned a priority, and now
            # we have explicit markings on the document.
            # NOTE: Prioritizing is not supported in 2.0 initial. This code
            # is here because it'll be used eventually, and at the moment
            # it's too hard to excise it.
            segmenter = self.folder.workspace.getPrioritizer()
            if segmenter:
                segmenter.updateFromDoc(docObj, fileBasename)
            
            if next_op:
                import json
                nextOp = json.loads(next_op)
                opName = nextOp["operation"]
                del nextOp["operation"]
                o = self.folder.getOperation(opName, basenames = [basename], transaction = t)
                self.statusOp = o
                # Again, we hit the bug where in Python < 2.6.5,
                # the keys of **args can't be Unicode strings.
                # Well, I don't know whether that bug still lingers,
                # but in Python 3, these can't be byte sequences,
                # so we have to decode them again.
                nextOp = dict([(k.encode('ascii', 'replace').decode('ascii'), v) for k, v in nextOp.items()])
                o.do(**nextOp)
                self.state = "statusop"
            else:
                for s in self.folder.workspace.workflow.stepList:
                    if s.stepName == curStep:
                        thisStep = s
                        break
                newStatus = self.folder.workspace._documentStatus(docObj, thisStep, "save", curDocStatus)
                db.updateDocumentStatus(fileBasename, status = newStatus, wfStatus = self.folder.workspace._docStatusToWorkflowStatus(newStatus),
                                        lastTouchedBy = user)
                if newStatus != curDocStatus:
                    self.state = "statuschanged"
            if release_lock:
                if self.fromCmdline:
                    print("Releasing lock ID %s" % lock_id)
                o = self.folder.getOperation("release_lock",
                                             basenames = [basename],
                                             transaction = t)
                o.do(lock_id = lock_id)
            t.commit()
        except:
            t.rollback()
            raise

    # Only used opening the the affected files, I hope.
    def getAffectedFileBasenames(self):
        return {self.affectedFileBasename: self.affectedBasenames[0]}

# Build a model. Operation on gold only.
# Optionally auto-advance the result.

# This is noticeably trickier than it was in 2.0. In 2.0, we could rely on
# the Carafe tagger retagging any segment that's owned by "MACHINE"; but
# here, we might not have segments at all. In that case, it will NOT retag,
# because the step is done. So if the sets are managed, then we don't need
# to undo; but if the sets are NOT managed, we do. 

@WorkspaceOp("core")
class ModelBuildOperation(WorkspaceOperation):

    name = "modelbuild"

    args = [OpArgument("trainable_step", help = "a step in the task workflow which has a trainable engine. Required if there are multiple trainable steps in your workflow.", hasArg = True),
            OpArgument("config_name", help = "use a model settings configuration other than the default.", hasArg = True),
            OpArgument("do_autotag", help = "autotag eligible basenames with the model after the model is constructed."),
            OpArgument("autotag_basenames",
                       help = "(optional) if autotagging, a space-separated sequence of basenames to autotag. By default, all eligible files are retagged.",
                       hasArg = True),
            OpArgument("autotag_basename",
                       action = "append",
                       help = "(optional) if autotagging, a single basename to autotag. By default, all available files are retagged. Can be repeated.",
                       dest = "autotag_basename_list",
                       hasArg = True)]

    def __init__(self, *args, **kw):
        WorkspaceOperation.__init__(self, *args, **kw)
        self.autotagOps = []
    
    def getAffectedFolders(self):
        if self.state == "autotag":
            setA = set()
            for o in self.autotagOps:
                setA.add(o.getAffectedFolders())
            return list(setA)
        else:
            return []

    def allPaths(self):
        return self.folder.getFiles(self.affectedBasenames)

    # UPDATE FOR AUTOTAG

    def do(self, do_autotag = False, config_name = None, trainable_step = None,
           autotag_basenames = None, autotag_basename_list = None):

        if autotag_basename_list or autotag_basenames:
            autotag_basename_list = autotag_basename_list or []
            if type(autotag_basenames) is str:
                for b in autotag_basenames.split():
                    if b not in autotag_basename_list:
                        autotag_basename_list.append(b)

        operationSettings = self.getOperationSettings()
        
        if operationSettings is None:
            operationSettings = {}

        # Only train on gold segments.
        operationSettings["partial_tagging_on_gold_only"] = True

        # Figure out what the step is.

        trainableSteps = self.folder.workspace.workflow.getTrainableSteps()
        foundStep = None
        
        if trainable_step is not None:            
            for k in trainableSteps.values():
                if (trainable_step == k.stepName) or (trainable_step == k.prettyName):
                    foundStep = k
                    break
            if foundStep is None:
                raise WorkspaceError("no trainable step named '%s' was found in the workspace workflow" % trainable_step)

        elif not trainableSteps:
            raise WorkspaceError("no trainable steps found in the workspace")
        elif len(trainableSteps.values()) > 1:
            raise WorkspaceError("multiple trainable steps found in workspace; trainable_step must be specified")
        else:
            # In Python 3, values() isn't indexable because it's a view.
            foundStep = list(trainableSteps.values())[0]            
        
        # Copy all the files to the tempdir.

        allPaths = self.allPaths()
        if not allPaths:
            raise WorkspaceError("modelbuild: no paths affected")
        with Tmpdir(specifiedParent = self.folder.workspace.dir) as tmpDir:
            # Now, run the model engine.
            buildInfo = self.folder.workspace.task.getModelInfo(configName = config_name, step = foundStep.stepName)
            if buildInfo is None:
                raise WorkspaceError("unknown model build settings config")
            m = buildInfo.buildModelBuilder(**operationSettings)
            # Move old models aside, or delete them. 
            # Place the model in the "model" file. Put the basenames
            # in "model_basenames".
            modelPath = self.folder.workspace.getModelPathForStep(foundStep)
            m.run(modelPath, allPaths, tmpDir = tmpDir)
            fp = codecs.open(modelPath + "_basenames", "w", "utf-8")
            for p in allPaths:
                fp.write(os.path.basename(p) + "\n")
            fp.close()

        # Right now, if there are no paths affected, an error will be raised.
        # So I need to check that here and catch it.
        
        if do_autotag:
            self.state = "autotag"
            # First, we have to determine the step eligibility for the basenames
            # in question. autotag_basename_list may be None, in which case the right
            # things will happen. The file can't be locked.
            db = self.folder.workspace.getDB()
            bInfo = db.basenameInfo(autotag_basename_list, step = foundStep.stepName)
            docNames = []
            for docName, basename, status, assignedUser, lockedBy, wfStatus, currentStep in bInfo:
                # Looking only at uncorrected, instead of partially corrected, may
                # rule out partial annotation. But for now, that's fine. Also, if
                # later we enable partial annotation, we'll have to update the doc
                # status from the document, rather than asserting they're all uncorrected.
                # So right now, although it's logically possible to partially tag, it's
                # not supported, and we only retag those documents which are completely untouched
                # by humans for that step.
                if (lockedBy is None) and (status in ("unannotated", "uncorrected")):
                    docNames.append(docName)
            allPaths = [os.path.join(self.folder.dir, d) for d in docNames]
            # Start a transaction.
            t = self.folder.workspace.beginTransaction(self)            
            try:
                t.addFilesToPreserve(allPaths)
                opSettings = {"workflow": self.folder.workspace.workflow.name,
                              "steps": foundStep.stepName,
                              foundStep.stepName + "_local": "yes",
                              foundStep.stepName + "_model": modelPath}                
                if (not foundStep.managedSets) or (not foundStep.engine) or (not foundStep.engine.isRetaggable()):
                    opSettings["undo_through"] = foundStep.stepName
                runParameters = {"input_file_type": self.folder.workspace.richFileType,
                                 "input_encoding": "utf-8"}
                dataPairs = self.folder.workspace._applyToolChain("autotag", None, opSettings,
                                                                  runParameters, allPaths, checkPathsAffected = False,
                                                                  fromCmdline = self.fromCmdline)
                # Next, we'd better check to make sure that we can write each file.
                # If we can't, we want to raise an error. We should check each
                # individual file, because we don't want ANYthing to happen
                # if the writes can fail.
                if not os.access(self.folder.dir, os.W_OK | os.X_OK):
                    raise WorkspaceError("folder core not available for writing")
                status = "uncorrected"
                wfStatus = self.folder.workspace._docStatusToWorkflowStatus(status)
                for p, iData in dataPairs:
                    fileBasename = os.path.basename(p)
                    if not os.access(p, os.W_OK):
                        raise WorkspaceError("file %s in folder core not available for writing" % fileBasename)
                    self.folder.saveFile(iData, fileBasename)
                    db.updateDocumentStatus(fileBasename, status = status, step = foundStep.stepName,
                                            wfStatus = wfStatus)
                    db.clearLastTouchedBy(fileBasename)
                
                t.commit()
            except:
                t.rollback()
                raise

# And the markcompleted operation on autotagged must ensure that the
# affected basenames all have SEGMENTs none of which are MACHINE. And the segmenter
# has to be cleared.

@WorkspaceOp("core")
class MarkGoldOperation(WorkspaceOperation):

    availability = CMDLINE_DEBUG_AVAILABLE | UI_AVAILABLE

    name = "markgold"

    args = [OpArgument("user", hasArg = True,
                       help = "specify a user responsible for the gold marking"),
            Option("--suppress_advancement", type = "boolean", default = False,
                   help = "By default, if a document is gold, it will automatically check for scheduled reviews, then advance to the next hand-annotatable step, or complete the workflow if there are no such steps. This flag will suppress automatic advancement."),
            Option("--lock_id", type="string", help="lock ID for marking gold (obligatory if document is locked)"),
            Option("--release_lock", type="boolean", default=False, help="release the lock after marking gold")]

    # We need to be able to get all the parameters from the workflow in here.
    # So we need to call addOptions on the aggregator for the workflow, because
    # we're advancing. We have to do this in the advance operation too.

    @classmethod
    def addOptions(cls, aggregator, ws, **kw):
            
        super(MarkGoldOperation, cls).addOptions(aggregator, ws, **kw)
        ws.advancerClass.addOptions(aggregator, ws, **kw)

    def getAffectedFolders(self):
        folders = set(["core"])
        fileBasenamesSubmitted = getattr(self, "fileBasenamesSubmitted", None)
        if fileBasenamesSubmitted:
            folders.update([k for (k, v) in fileBasenamesSubmitted.items() if v])
        advancer = getattr(self, "advancer", None)
        if advancer:
            folders |= set(advancer.getAffectedFolders())
        return list(folders)

    def getTargetFolderAndDocuments(self):
        return "core", self._getTargetDocuments("core")

    def getAffectedFileBasenames(self):
        return self.affectedFileBasenames

    # skipReviewCompletely is set in import if we have multiple assignees.
    
    def do(self, user = None, suppress_advancement = False, lock_id = None,
           # These two are needed in import, but nowhere else.
           deferReconciliation = False, skipReviewCompletely = False, release_lock = False, **wfParams):
        _in_transaction = self.transaction
        if user is None:
            raise WorkspaceError("markgold requires a user")
        # lock_id can be None, but it's gotta match what's in the DB.
        # if it IS none, no need to lock the document.
        db = self.folder.workspace.getDB()
        if not db.userIsRegistered(user):
            raise WorkspaceError("unknown user %s" % user)
        # If we have a lock_id, use it to collect the fileBasenames.
        # Otherwise, collect them yourself.
        if lock_id is not None:
            fileBasenames = {}
            # There had better be only one of these, otherwise this will
            # break...
            fileBasename, basename, lockingUser = db.coreGetLockIDInfo(lock_id)
            for b in self.affectedBasenames:
                if (b != basename) or (lockingUser != user):
                    raise WorkspaceError("wrong lock ID %s for basename %s" % (lock_id, b))
            fileBasenames[fileBasename] = basename
        else:
            fileBasenames = {}
            for b in self.affectedBasenames:
                fileBasename = self.folder.fileBasenameForUser(b, user)
                lockingUser = self.folder.fileBasenameLocked(fileBasename)
                if lockingUser:
                    raise WorkspaceError("can't mark locked file %s gold without lock ID (locked by %s)" % (fileBasename, lockingUser))
                fileBasenames[fileBasename] = b
        self.affectedFileBasenames = fileBasenames
        if _in_transaction:
            self._do(db, fileBasenames, user, suppress_advancement, deferReconciliation, _in_transaction,
                     lock_id, skipReviewCompletely, release_lock, wfParams)
        else:
            t = self.folder.workspace.beginTransaction(self)
            try:
                self._do(db, fileBasenames, user, suppress_advancement, deferReconciliation, t,
                         lock_id, skipReviewCompletely, release_lock, wfParams)
                t.commit()
            except:
                t.rollback()
                raise

    # So here in the markgold operation, we need to do a few things. All the file basenames
    # are now gold.
    def _do(self, db, fileBasenameDict, user, suppress_advancement, defer_reconciliation, transaction,
            lock_id, skipReviewCompletely, release_lock, wfParams):
        # Use list() to freeze dictionary keys in Python 3.
        fileBasenames = list(fileBasenameDict.keys())
        self.folder.workspace._updateDocumentStatuses(fileBasenames, self.folder,
                                                      transaction,
                                                      # Which segment annots do we update?
                                                      ["non-gold"],
                                                      # What do we update them to?
                                                      "human gold",
                                                      # Who do we attribute it to?
                                                      user)
        if self.fromCmdline:
            print("Marked gold:", " ".join(fileBasenames))
        if not skipReviewCompletely:
            self.fileBasenamesSubmitted = self.folder.workspace._submitForReview(fileBasenames, user, transaction,
                                                                                 lockID = lock_id, deferReconciliation = defer_reconciliation,
                                                                                 fromCmdline = self.fromCmdline)
        if lock_id and release_lock:
            if self.fromCmdline:
                print("Releasing lock ID %s" % lock_id)
            o = self.folder.getOperation("release_lock",
                                         basenames = list(set(fileBasenameDict.values())),
                                         transaction = transaction)
            o.do(lock_id = lock_id)
            lock_id = None

        if not suppress_advancement:
            self.advancer = self.folder.workspace._autoAdvance(fileBasenames, transaction, self.fromCmdline,
                                                               lockId = lock_id, workflowParams = wfParams)

    def webResult(self):
        w = WorkspaceOperation.webResult(self)
        fileBasename = w.get("file_basename")
        if fileBasename:
            fileBasenamesSubmitted = getattr(self, "fileBasenamesSubmitted", None)
            if fileBasenamesSubmitted:
                # If the file_basename in w is in one of the lists of the basenames
                # submitted, it's submitted for review.
                for v in fileBasenamesSubmitted.values():
                    if fileBasename in v:
                        w["document_removed"] = True
                        break        
        return w

# I need this operation, really. Ultimately, this should happen
# segment by segment, but until I expose the segment boundaries and
# allow that, this will be a required option. Unlike markgold,
# I don't think we need to attribute a user here. But we need
# a user for validating the operation.

@WorkspaceOp("core")
class UnmarkGoldOperation(WorkspaceOperation):

    availability = CMDLINE_DEBUG_AVAILABLE | UI_AVAILABLE

    name = "unmarkgold"

    args = [OpArgument("user", hasArg = True,
                       help = "specify a user for the lock ID"),
            Option("--lock_id", type="string", help="lock ID for unmarking gold (obligatory if document is locked)")]

    def getAffectedFolders(self):
        return [self.folder.prettyName]

    def getTargetFolderAndDocuments(self):
        return "core", self._getTargetDocuments("core")

    def getAffectedFileBasenames(self):
        return self.affectedFileBasenames

    def do(self, user = None, lock_id = None):
        _in_transaction = self.transaction
        if user is None:
            raise WorkspaceError("unmarkgold requires a user")
        # lock_id can be None, but it's gotta match what's in the DB.
        # if it IS none, no need to lock the document.
        db = self.folder.workspace.getDB()
        if not db.userIsRegistered(user):
            raise WorkspaceError("unknown user %s" % user)
        # If we have a lock_id, use it to collect the fileBasenames.
        # Otherwise, collect them yourself.
        if lock_id is not None:
            fileBasenames = {}
            # There had better be only one of these, otherwise this will
            # break...
            fileBasename, basename, lockingUser = db.coreGetLockIDInfo(lock_id)
            for b in self.affectedBasenames:
                if (b != basename) or (lockingUser != user):
                    raise WorkspaceError("wrong lock ID %s for basename %s" % (lock_id, b))
            fileBasenames[fileBasename] = basename
        else:
            fileBasenames = {}
            for b in self.affectedBasenames:
                fileBasename = self.folder.fileBasenameForUser(b, user)
                lockingUser = self.folder.fileBasenameLocked(fileBasename)
                if lockingUser:
                    raise WorkspaceError("can't unmarkgold locked file %s without lock ID (locked by %s)" % (fileBasename, lockingUser))
                fileBasenames[fileBasename] = b
        self.affectedFileBasenames = fileBasenames
        # Check the current step for each one - if it's an automated step,
        # refuse.
        ungoldableSteps = set([step.stepName for step in self.folder.workspace.workflow.stepList
                               if isinstance(step, PluginMgr.AnnotationStep) and step.sType != "auto"])
        # Use list() to freeze dictionary keys in Python 3.
        bInfo = db.fileBasenameInfo(list(fileBasenames.keys()))
        for docName, (basename, status, assigned_user, locked_by, workflow_status, current_step) in bInfo.items():
            if current_step not in ungoldableSteps:
                raise WorkspaceError("can't unmarkgold file %s because its current step is fully automated" % docName)
        if _in_transaction:
            # Use list() to freeze dictionary keys in Python 3.
            self._do(db, list(fileBasenames.keys()), user, _in_transaction)
        else:
            t = self.folder.workspace.beginTransaction(self)
            try:
                # Use list() to freeze dictionary keys in Python 3.
                self._do(db, list(fileBasenames.keys()), user, t)
                t.commit()
            except:
                t.rollback()
                raise
            
    def _do(self, db, fileBasenames, user, transaction):
        self.folder.workspace._updateDocumentStatuses(fileBasenames, self.folder,
                                                      transaction,
                                                      # Which segment annots do we update?
                                                      ["human gold", "reconciled"],
                                                      # What do we update them to?
                                                      "non-gold",
                                                      # Who do we attribute it to?
                                                      None)
        for b in fileBasenames:
            db.updateDocumentStatus(b, lastTouchedBy = user)

@WorkspaceOp("core")            
class ReleaseLockOperation(WorkspaceOperation):

    name = "release_lock"

    availability = NOT_AVAILABLE

    args = [Option("--lock_id", type="string", help="lock ID")]

    def do(self, lock_id = None):
        _in_transaction = self.transaction
        if lock_id is None:
            raise WorkspaceError("Can't release a lock without an ID")
        db = self.folder.workspace.getDB()
        # I'm wrapping this because I don't know whether this
        # operation is going to remain atomic.
        if _in_transaction:
            self._do(db, lock_id)
        else:
            t = self.folder.workspace.beginTransaction(self)
            try:
                self._do(db, lock_id)
                t.commit()
            except:
                t.rollback()
                raise

    def _do(self, db, lock_id):
        db.unlockCoreLock(lock_id)

    def getAffectedFolders(self):
        return [self.folder.prettyName]

# Finally, the emergency force unlock.

# Originally, I coded this in such a way that it was more fragile
# than was appropriate. 

@WorkspaceOp("core", "reconciliation")
class ForceUnlockOperation(WorkspaceOperation):

    name = "force_unlock"

    availability = CMDLINE_AVAILABLE

    args = [OpArgument("user", hasArg = True,
                       help = "the user who's locked the basename")]

    def do(self, user = None):
        if user is None:
            raise WorkspaceError("can't force unlock a basename without a user")
        t = self.folder.workspace.beginTransaction(self)
        try:
            self._do(user)
            t.commit()
        except:
            t.rollback()
            raise

    def _do(self, user):
        if self.folder.folderName == "core":
            db = self.folder.workspace.getDB()
            unlocked = db.forceUnlockCoreBasenames(user, self.affectedBasenames)
            if self.fromCmdline:
                if unlocked:
                    print("Unlocked core documents:", " ".join(unlocked))
                else:
                    print("Unlocked no documents.")
        elif self.folder.folderName == "reconciliation":
            db = self.folder.workspace.getDB()
            unlocked = db.forceUnlockReconciliationBasenames(user, self.affectedBasenames)
            if self.fromCmdline:
                if unlocked:
                    print("Unlocked reconciliation basenames:", " ".join(unlocked))
                else:
                    print("Unlocked no basenames.")


#
# Review folder and operations
#

class ReviewWorkspaceFolder(WorkspaceFolder):

    def listContents(self, basenames, user = None, read_only = False):
        db = self.workspace.getDB()
        bPairs = []
        userHasReviewerRole = (user is None) or self.userHasAppropriateRole(user)
        for docName, basename, requestingUser, reviewingUser, repairRequest in db.reviewInfo(basenames):
            info = {"true basename": basename,
                    "basename": docName,
                    "requesting user": requestingUser,
                    "repair request": (repairRequest and True) or False}            
            if reviewingUser is not None:
                info["being reviewed by"] = reviewingUser
            # Compute openability. Read only: OK. No reviewer role: not OK.
            if read_only:
                info["openable"] = True
            elif (not repairRequest) and (not userHasReviewerRole):
                info["openable"] = False
            # Next, you can't be the requesting user, unless it's a
            # repair request, in which case you have to be.
            elif user and (((not repairRequest) and (user == requestingUser)) or (repairRequest and (user != requestingUser))):
                info["openable"] = False
            # Next, if it's locked and you're not the locking user, not OK.
            elif reviewingUser and (user != reviewingUser):
                info["openable"] = False
            else:
                info["openable"] = True
            bPairs.append(info)
        return bPairs

    def getFiles(self, basenames = None):
        db = self.workspace.getDB()
        return [os.path.join(self.dir, p[0]) for p in db.reviewInfo(basenames = basenames)]

    # the docnames are their own basenames here - the basenames are just decoration.
    
    def getBasenames(self, basenames = None):
        db = self.workspace.getDB()
        return list(set([p[0] for p in db.reviewInfo(docNames = basenames)]))

    def basenameForDocname(self, docName):
        return self.workspace.getDB().basenameForDocname(docName)

    def userHasAppropriateRole(self, user, fileBasename = None):
        db = self.workspace.getDB()
        if fileBasename is None:
            return db.userHasRole(user, "reviewer")
        else:
            info = db.reviewInfo(docNames = [fileBasename])
            if info and info[0][4]:
                return db.userHasRole(user, "annotator")
            else:
                # The appropriate role for repair requests is annotator, otherwise reviewer.
                return db.userHasRole(user, "reviewer")

    def prepareForEditing(self, doc, fileBasename, user, lockId):
        db = self.workspace.getDB()
        db.lockReviewDocument(lockId, fileBasename, user)

    # This should always add the current step, since I need it for annotation.
    def _addFileBasenameStatus(self, webResult, fileBasename):
        ignore, webResult["current_step"], ignore = self.workspace.getDB().coreDocumentStatuses(fileBasename)

@WorkspaceOp("review")
class ReviewSaveOperation(SaveOperation):

    args = [OpArgument("doc", help = "document to save, as a JSON string", hasArg = True),
            Option("--lock_id", type="string", help="transaction ID for save"),
            Option("--review_done", type="boolean", default=False, help="complete the review"),
            Option("--suppress_advancement", type = "boolean", default = False,
                   help = "By default, when the review is completed, the workspace advances the document to the next hand-annotatable step, or completes the workflow if there are no such steps. This flag will suppress automatic advancement."),
            Option("--release_lock", type="boolean", default=False, help="release the lock after save"),
            OpArgument("log", help = "log fragment (from UI) to trigger fine-grained progress monitoring", hasArg = True),
            OpArgument("log_format", help = "the format of the log fragment", hasArg = True),
            OpArgument("timestamp", help = "millisecond timestamp of log upload from the UI's point of view", hasArg = True)]

    # NOTE: Which folders are changed is governed by self.statusOp in SaveOperation.
    # At the moment, no folder is judged to be changed, except when we release the lock.
    # This doesn't happen in reconciliation save, but it does happen in review save.
    
    def do(self, doc = None, lock_id = None, review_done = False, suppress_advancement = False, release_lock = False,
           log = None, log_format = None, timestamp = None):
        self.status = None
        if lock_id is None:
            raise WorkspaceError("can't save without lock ID")        
        # Now we get the basename. Must check to ensure that
        # the transaction ID matches.
        affectedBasename = self.affectedBasenames[0]
        # See getAffectedFileBasenames in SaveOperation.
        db = self.folder.workspace.getDB()
        docName, basename, requestingUser = db.reviewGetLockIDInfo(lock_id)
        # Remember, the "basename" is the docname.
        if docName != affectedBasename:
            raise WorkspaceError("wrong lock ID %s for basename %s" % (lock_id, affectedBasename))
        
        self.affectedFileBasename = docName
        t = self.folder.workspace.beginTransaction(self, filesToPreserve = [os.path.join(self.folder.dir, docName)])
        try:
            # Do this as early as possible, to get the timings as close as possible.
            if log and self.folder.workspace.logger:
                self.folder.workspace.logger.logLog(log, log_format, timestamp)
            if doc is not None:
                # It can be none, if it's not dirty.
                # First, make it into a document. The document
                # string is almost certainly not Unicode yet.

                docObj = self.folder.docIO.readFromByteSequence(doc, 'utf-8')
                
                # There better only be one basename. If I'm going
                # to save the document, I had better update the IDs, in case
                # some yahoo has to forcibly unlock it.

                self.folder.saveFile(docObj, docName)

            if release_lock or review_done:
                if self.fromCmdline:
                    print("Releasing lock %s" % lock_id)
                o = self.folder.getOperation("release_lock",
                                             basenames = [docName],
                                             transaction = t)
                o.do(lock_id = lock_id)
                self.status = "statuschanged"

            if review_done:
                if self.fromCmdline:
                    print("Completing human review.")
                o = self.folder.getOperation("complete_human_review",
                                             basenames = [docName],
                                             transaction = t)
                self.status = "statusop"
                self.statusOp = o
                o.do(suppress_advancement = suppress_advancement)
                
            t.commit()
        except:
            t.rollback()
            raise
                
    @classmethod
    def _getUsage(cls, clsName, w):
        return """Usage: %prog [options] <dir> """ + clsName + """ [operation_options] <folder> [ <docname> ... ]

<folder>: The name of the folder to operate on.
<basename>: (optional) The docname or docnames to restrict the operation to."""

    def getAffectedFileBasenames(self):
        return dict([(b, b) for b in self.affectedBasenames])

    def getAffectedFolders(self):
        if self.affectedBasenames and (self.status == "statusop"):
            return [self.folder.prettyName]
        else:
            return []

    def webResult(self):
        w = WorkspaceOperation.webResult(self)
        fileBasename = w.get("file_basename")
        if fileBasename and (self.status == "statusop"):
            w["document_removed"] = True
        return w

@WorkspaceOp("review")
class ReviewReleaseLockOperation(WorkspaceOperation):

    name = "release_lock"

    availability = NOT_AVAILABLE

    args = [Option("--lock_id", type="string", help="lock ID")]

    def do(self, lock_id = None):
        _in_transaction = self.transaction
        if lock_id is None:
            raise WorkspaceError("Can't release a lock without an ID")
        db = self.folder.workspace.getDB()
        # We may need this later, and we're unlocking the transaction now.
        # I need the basename info to cache the current basename to
        # rollback the transaction if necessary.
        docName, basename, user = db.reviewGetLockIDInfo(lock_id)
        if basename is None:
            raise WorkspaceError("Couldn't find an active basename for lock ID '%s'" % lock_id)

        if _in_transaction:
            self._do(db, lock_id)
        else:            
            t = self.folder.workspace.beginTransaction(self, filesToPreserve = [os.path.join(self.folder.dir, docName)])
            try:
                self._do(db, lock_id)
                t.commit()
            except:
                t.rollback()
                raise

    def _do(self, db, lock_id):
        db.unlockReviewLock(lock_id)

    def getAffectedFolders(self):
        return [self.folder.prettyName]

    def getAffectedFileBasenames(self):
        return dict([(b, b) for b in self.affectedBasenames])
                
    @classmethod
    def _getUsage(cls, clsName, w):
        return """Usage: %prog [options] <dir> """ + clsName + """ [operation_options] <folder> [ <docname> ... ]

<folder>: The name of the folder to operate on.
<basename>: (optional) The docname or docnames to restrict the operation to."""

@WorkspaceOp("review")
class ReviewCompleteOperation(WorkspaceOperation):

    availability = CMDLINE_AVAILABLE

    args = [Option("--suppress_advancement", type = "boolean", default = False,
                   help = "By default, when the review is completed, the workspace advances the document to the next hand-annotatable step, or completes the workflow if there are no such steps. This flag will suppress automatic advancement."),
            Option("--lock_id", type="string", help="lock ID")]
    
    name = "complete_human_review"
                
    @classmethod
    def _getUsage(cls, clsName, w):
        return """Usage: %prog [options] <dir> """ + clsName + """ [operation_options] <folder> [ <docname> ... ]

<folder>: The name of the folder to operate on.
<basename>: (optional) The docname or docnames to restrict the operation to."""

    def do(self, suppress_advancement = False, lock_id = None):
        _in_transaction = self.transaction
        if _in_transaction:
            self._do(self.transaction, lock_id, suppress_advancement)
        else:            
            t = self.folder.workspace.beginTransaction(self, filesToPreserve = [os.path.join(self.folder.dir, docName) for docName in self.affectedBasenames])
            try:
                self._do(t, lock_id, suppress_advancement)
                t.commit()
            except:
                t.rollback()
                raise

    def _do(self, transaction, lockID, suppressAdvancement):
        # The affected basenames should be removed, as long as they're not locked.
        db = self.folder.workspace.getDB()
        reviewInfo = db.reviewInfo(docNames = self.affectedBasenames)
        if lockID:
            lockDoc, lockBasename, lockUser = db.reviewGetLockIDInfo(lockID)            
        for doc_name, basename, requesting_user, reviewing_user, repair_request in reviewInfo:
            if lockID:
                if lockDoc != doc_name:
                    raise WorkspaceFileLockedError("no documents removed from review: document '%s' doesn't match lock ID" % doc_name)
                if lockUser != reviewing_user:
                    raise WorkspaceFileLockedError("no documents removed from review: reviewing user doesn't match lock ID")
            elif reviewing_user:
                raise WorkspaceFileLockedError("no documents removed from review: document '%s' locked for human review" % doc_name)
            else:
                transaction.addFilesToPreserve([os.path.join(self.folder.dir, doc_name)], check = True)

        db.forceUnlockCoreDocnames("REVIEW", self.affectedBasenames)
        reviewDocs = {}
        for doc_name, basename, requesting_user, reviewing_user, repair_request in reviewInfo:
            db.removeFromReview(doc_name)
            # Load the documents before you remove them from the review
            # folder.
            recDoc = self.folder.openFile(doc_name)
            if recDoc is not None:
                reviewDocs[doc_name] = recDoc
        self.folder.clear(basenames = self.affectedBasenames)
        if self.fromCmdline:
            print("Removed documents from review:", " ".join(self.affectedBasenames))
        coreFolder = self.folder.workspace.getFolder("core")
        for doc_name, basename, requesting_user, reviewing_user, repair_request in reviewInfo:
            transaction.addFilesToPreserve([os.path.join(coreFolder.dir, doc_name)])
            try:
                reviewDoc = reviewDocs[doc_name]
            except KeyError:
                # There's no entry, because the document wasn't found.
                continue
            # All we need to do is save the review document and mark it reconciled.
            coreFolder.saveFile(reviewDoc, doc_name)
            # And finally, mark the appropriate segments reconciled.
            # Because they already have an assigned user, I don't need
            # to change it.
            db.clearLastTouchedBy(doc_name)
        docsToUpdate = [t[0] for t in reviewInfo if not t[4]]
        docsNotToUpdate = [t[0] for t in reviewInfo if t[4]]
        if docsToUpdate:
            self.folder.workspace._updateDocumentStatuses(docsToUpdate, coreFolder, transaction, ["human gold"],
                                                          "reconciled", None)
            # Don't advance any documents which are repair requests.
            if not suppressAdvancement:                
                self.folder.workspace._autoAdvance(docsToUpdate, transaction, self.fromCmdline)
        if docsNotToUpdate:
            # Restore their previous workflow status, by inferring it from the document, I guess.
            for docName, (basename, status, assignedUser, lockedBy, wfStatus, currentStep) in db.fileBasenameInfo(docsNotToUpdate).items():
                db.updateDocumentStatus(docName, wfStatus = self.folder.workspace._docStatusToWorkflowStatus(status))

    def getAffectedFileBasenames(self):
        return dict([(b, b) for b in self.affectedBasenames])

    # This isn't called from the UI, except indirectly, by virtue
    # of being called during save. SaveOperation delegates these next
    # two methods to this operation in that case.

    def getAffectedFolders(self):
        return [self.folder.prettyName, "core"]

    def getTargetFolderAndDocuments(self):
        return None, []

    def webResult(self):
        w = WorkspaceOperation.webResult(self)
        w["document_removed"] = True
        return w

@WorkspaceOp("review")
class ReviewForceUnlockOperation(WorkspaceOperation):

    name = "force_unlock"

    availability = CMDLINE_AVAILABLE

    args = [OpArgument("user", hasArg = True,
                       help = "the user who's locked the document")]

    def do(self, user = None):
        if user is None:
            raise WorkspaceError("can't force unlock a document without a user")
        t = self.folder.workspace.beginTransaction(self)
        try:
            self._do(user)
            t.commit()
        except:
            t.rollback()
            raise

    def _do(self, user):
        db = self.folder.workspace.getDB()
        unlocked = db.forceUnlockReviewBasenames(user, self.affectedBasenames)
        if self.fromCmdline:
            if unlocked:
                print("Unlocked review documents:", " ".join(unlocked))
            else:
                print("Unlocked no documents.")

    def getAffectedFileBasenames(self):
        return dict([(b, b) for b in self.affectedBasenames])

#
# Reconciliation folder and operations
#

class ReconciliationWorkspaceFolder(WorkspaceFolder):

    def fileBasenameLocked(self, fileBasename):
        return self.workspace.getDB().reconciliationBasenameLocked(fileBasename)

    # The semantics of "reviewed_by" are such that the frontend
    # can rely on the contents of that value to see whether
    # that segment has been reviewed. So in the ase where a review
    # is forced by the user, we have to make sure that the
    # user is not in the reviewed_by list.

    def prepareForEditing(self, doc, fileBasename, user, lockId):
        # lock it.
        db = self.workspace.getDB()
        db.lockReconciliationBasename(lockId, fileBasename, user)

    def listContents(self, basenames, user = None, read_only = False):
        db = self.workspace.getDB()
        bPairs = []
        userHasReviewerRole = (user is None) or self.userHasAppropriateRole(user)
        for basename, requestingUser, reviewingUser in db.reconciliationInfo(basenames):
            info = {"basename": basename, "requesting user": requestingUser}
            if reviewingUser:
                info["being reviewed by"] = reviewingUser
            # Compute openability. Read only: OK. No reviewer role: not OK.
            if read_only:
                info["openable"] = True
            elif not userHasReviewerRole:
                info["openable"] = False
            # Next, if it's locked and you're not the locking user, not OK.
            elif reviewingUser and (user != reviewingUser):
                info["openable"] = False
            else:
                info["openable"] = True        
            bPairs.append(info)
        return bPairs

    def userHasAppropriateRole(self, user, fileBasename = None):
        return self.workspace.getDB().userHasRole(user, "reviewer")

    def _addFileBasenameStatus(self, webResult, fileBasename):
        pass

@WorkspaceOp("reconciliation")
class ReconciliationSaveOperation(SaveOperation):

    args = [OpArgument("doc", help = "document to save, as a JSON string", hasArg = True),
            Option("--lock_id", type="string", help="lock ID for save"),
            Option("--release_lock", type="boolean", default=False, help="release the lock after save"),
            Option("--suppress_advancement", type = "boolean", default = False,
                   help = "By default, when the lock is released and reconciliation is done, the workspace advances the document to the next hand-annotatable step, or completes the workflow if there are no such steps. This flag will suppress automatic advancement."),
            OpArgument("log", help = "log fragment (from UI) to trigger fine-grained progress monitoring", hasArg = True),
            OpArgument("log_format", help = "the format of the log fragment", hasArg = True),
            OpArgument("timestamp", help = "millisecond timestamp of log upload from the UI's point of view", hasArg = True)]
    
    # NOTE: Which folders are changed is governed by self.statusOp in SaveOperation.
    # At the moment, no folder is judged to be changed, except when we release the lock.
    # But we never release the lock from the UI during a save operation. So we don't
    # need to worry.
    
    def do(self, doc = None, lock_id = None, release_lock = False, suppress_advancement = False,
           log = None, log_format = None, timestamp = None):
        if lock_id is None:
            raise WorkspaceError("can't save without lock ID")
        # Now we get the basename. Must check to ensure that
        # the transaction ID matches.
        basename = self.affectedBasenames[0]
        # See getAffectedFileBasenames in SaveOperation.
        self.affectedFileBasename = basename
        db = self.folder.workspace.getDB()
        if not db.reconciliationLockIDMatches(lock_id, basename):
            raise WorkspaceError("wrong lock ID %s for basename %s" % (lock_id, basename))

        t = self.folder.workspace.beginTransaction(self, filesToPreserve = [os.path.join(self.folder.dir, basename)])
        try:
            # Do this as early as possible, to get the timings as close as possible.
            if log and self.folder.workspace.logger:
                self.folder.workspace.logger.logLog(log, log_format, timestamp)
            if doc is not None:
                # It can be none, if it's not dirty.
                # First, make it into a document. The document
                # string is almost certainly not Unicode yet.

                docObj = self.folder.docIO.readFromByteSequence(doc, 'utf-8')

                basename, user = db.reconciliationGetLockInfo(lock_id)
                
                # There better only be one basename. If I'm going
                # to save the document, I had better update the IDs, in case
                # some yahoo has to forcibly unlock it.

                self.folder.saveFile(docObj, basename)

            if release_lock:
                if self.fromCmdline:
                    print("Releasing lock %s" % lock_id)
                o = self.folder.getOperation("release_lock",
                                             basenames = [basename],
                                             transaction = t)
                o.do(lock_id = lock_id, suppress_advancement = suppress_advancement)
                
            t.commit()
        except:
            t.rollback()
            raise

@WorkspaceOp("reconciliation")
class ReconciliationReleaseLockOperation(WorkspaceOperation):

    name = "release_lock"

    availability = NOT_AVAILABLE

    args = [Option("--lock_id", type="string", help="lock ID"),
            Option("--suppress_advancement", type = "boolean", default = False,
                   help = "By default, when reconciliation is done, the workspace advances the document to the next hand-annotatable step, or completes the workflow if there are no such steps. This flag will suppress automatic advancement.")]

    def do(self, lock_id = None, suppress_advancement = False):
        _in_transaction = self.transaction
        if lock_id is None:
            raise WorkspaceError("Can't release a lock without an ID")
        db = self.folder.workspace.getDB()
        # We may need this later, and we're unlocking the transaction now.
        # I need the basename info to cache the current basename to
        # rollback the transaction if necessary.
        basename, user = db.reconciliationGetLockInfo(lock_id)
        if basename is None:
            raise WorkspaceError("Couldn't find an active basename for lock ID '%s'" % lock_id)

        if _in_transaction:
            self._do(db, lock_id, basename, user, suppress_advancement, _in_transaction)
        else:            
            t = self.folder.workspace.beginTransaction(self, filesToPreserve = [os.path.join(self.folder.dir, basename)])
            try:
                self._do(db, lock_id, basename, user, suppress_advancement, t)
                t.commit()
            except:
                t.rollback()
                raise

    def _do(self, db, lock_id, basename, user, suppressAdvancement, transaction):

        db.unlockReconciliationLock(lock_id)

        # At this point, if the document is done with reconciliation,
        # we should remove it from reconciliation.

        # Load it, because we need to check to see if it can be removed.
        docObj = self.folder.openFile(basename)
        if reconciliationDocumentFullyReconciled(docObj):
            # First, export the document from reconciliation, and replace the
            # core docs with this one.
            # Finally, get it out of the reconciliation folder.
            self.folder.workspace._removeFromReconciliation([basename], self.folder, transaction,
                                                            suppressAdvancement,
                                                            verbose = self.fromCmdline)
            self.removed = True

    def getAffectedFolders(self):
        if getattr(self, "removed", False):
            return [self.folder.prettyName, "core"]
        else:
            return [self.folder.prettyName]

# Here's an emergency operation for you. I coded this much
# too fragilely in the first round. For instance, I can't
# pay attention to the affectedBasenames, since the file may have
# not made it into the folder due to an error.

@WorkspaceOp("reconciliation")
class RemoveFromReconciliationOperation(WorkspaceOperation):

    availability = CMDLINE_AVAILABLE

    name = "remove_from_reconciliation"

    args = [OpArgument("dont_reintegrate",
                       help = "By default, reconciliation updates are integrated back into the core documents. Use this flag to skip that step."),
            Option("--suppress_advancement", type = "boolean", default = False,
                   help = "By default, when reconciliation is done, the workspace advances the document to the next hand-annotatable step, or completes the workflow if there are no such steps. This flag will suppress automatic advancement.")]

    def do(self, dont_reintegrate = False, suppress_advancement = False):
        # Disabled a bunch of tests, too. See WorkspaceReconciliationTestCase in
        # test/mat_workspace_unittest.py. This must be re-enabled when we get
        # this working.

        # Let's first start by getting EITHER the inputBasenames
        # (those which were listed explicitly, whether or not they're in the
        # folder) OR the affectedBasenames. The effect of this is when
        # there was an explicit list, we ALWAYS use it.
        basenames = self.inputBasenames or self.affectedBasenames
        if not basenames:
            print("No basenames requested for removal.")
            return
        t = self.folder.workspace.beginTransaction(self)
        try:
            self.folder.workspace._removeFromReconciliation(basenames, self.folder, t,
                                                            suppress_advancement,
                                                            verbose = self.fromCmdline,
                                                            reintegrate = not dont_reintegrate)
            t.commit()
        except:
            t.rollback()
            raise

#
# User roles and reconciliation config
#

@WorkspaceOp()
class AddUserRoleOperation(WorkspaceToplevelOperation):

    name = "add_roles"

    args = [Option("--roles", type = "multiple_choice",
                   choices = ["all", "annotator", "reviewer"],
                   help = "Available roles are 'annotator', 'reviewer'. If omitted, the role will be 'annotator'. The string 'all' adds both roles. Otherwise, a comma-separated list of roles.")]

    @classmethod
    def getUsage(cls, w):
        return """Usage: %prog [options] <dir> add_roles [add_roles_options] <user>...

<user>: the name of a user to update the roles for."""

    def do(self, roles = None):
        # Workspace is already locked. I'm not sure
        # this is going to be supported except on the command line.
        t = self.workspace.beginTransaction(self)
        try:
            roles = _computeRoles(roles, "adding roles")
            if roles is None:
                raise WorkspaceError("Must specify roles to add")
            self.workspace._addRoles(self.args, roles)
            if self.fromCmdline:
                o = self.workspace.getOperation("list_users", ())
                o.fromCmdline = True
                o.do()
            t.commit()
        except:
            t.rollback()
            raise

@WorkspaceOp()
class RemoveUserRoleOperation(WorkspaceToplevelOperation):

    name = "remove_roles"

    args = [Option("--roles", type = "multiple_choice",
                   choices = ["all", "annotator", "reviewer"],
                   help = "Available roles are 'annotator', 'reviewer'. If omitted, the role will be 'annotator'. The string 'all' removes both roles. Otherwise, a comma-separated list of roles.")]

    @classmethod
    def getUsage(cls, w):
        return """Usage: %prog [options] <dir> remove_roles [remove_roles_options] <user>...

<user>: the name of a user to update the roles for."""

    def do(self, roles = None):
        # Workspace is already locked. I'm not sure
        # this is going to be supported except on the command line.
        t = self.workspace.beginTransaction(self)
        try:
            roles = _computeRoles(roles, "removing roles")
            if roles is None:
                raise WorkspaceError("Must specify roles to remove")
            self.workspace._removeRoles(self.args, roles)
            if self.fromCmdline:
                o = self.workspace.getOperation("list_users", ())
                o.fromCmdline = True
                o.do()
            t.commit()
        except:
            t.rollback()
            raise

#
# And another one: configuring review
#

@WorkspaceOp()
class ScheduleReviewOperation(WorkspaceToplevelOperation):

    name = "schedule_review"

    @classmethod
    def getUsage(cls, w):
        return """Usage: %prog [options] <dir> schedule_review <step> <review_type>

<step>: the name or pretty name of the step in the workflow for which the review is being scheduled.
        All basenames which complete this step will be submitted for this review,
        including the case where this step is the final step reached during the import process.
<review_type>: one of human, reconciliation, reconciliation_with_crossvalidation."""

    def do(self):
        if len(self.args) != 2:
            raise WorkspaceError("wrong number of arguments to schedule_review")
        t = self.workspace.beginTransaction(self)
        try:
            self.workspace._scheduleReview(self.args[0], self.args[1], fromCmdline = self.fromCmdline)
            t.commit()
        except:
            t.rollback()
            raise

@WorkspaceOp()
class UnscheduleReviewOperation(WorkspaceToplevelOperation):

    name = "unschedule_review"

    @classmethod
    def getUsage(cls, w):
        return """Usage: %prog [options] <dir> unschedule_review <step>

<step>: the name or pretty name of the step in the workflow for which the review schedule is being removed."""

    def do(self):
        if len(self.args) != 1:
            raise WorkspaceError("wrong number of arguments to unschedule_review")
        t = self.workspace.beginTransaction(self)
        try:
            self.workspace._unscheduleReview(self.args[0], fromCmdline = self.fromCmdline)
            t.commit()
        except:
            t.rollback()
            raise

@WorkspaceOp()
class ListReviewScheduleOperation(WorkspaceToplevelOperation):

    name = "list_review_schedule"

    @classmethod
    def getUsage(cls, w):
        return """Usage: %prog [options] <dir> list_review_schedule"""

    def do(self):
        if self.fromCmdline:
            db = self.workspace.getDB()
            scheds = db.getReviewSchedules()
            for step, [reviewType, simProfile] in scheds.items():
                if simProfile:
                    print(step, ":", reviewType, "(similarity profile %s)" % simProfile)
                else:
                    print(step, ":", reviewType)

#
# Apply crossvalidation.
#

# Everything marked awaiting crossvalidation will be run.

# Originally, I was importing TestRun and TrainingRun from Bootstrap,
# but I use the ones from MAT.CarafeExperiment in run_experiment, so...

@WorkspaceOp("core")
class ApplyCrossvalidationOperation(WorkspaceOperation):

    name = "apply_crossvalidation"

    DEFAULT_FOLDS = 8

    DEFAULT_CROSSVAL_COUNT = 10

    args = [Option("--folds", type="int", help="number of cross-validation folds. Default is 8."),
            Option("--crossvalidation_doc_count_threshold", type = "int",
                   help="number of documents awaiting crossvalidation required in a given step for crossvalidation to occur. Default is 10.")]
    
    availability = CMDLINE_AVAILABLE

    # We're going to apply crossvalidation, and then submit
    # to reconciliation review. We have to do this by step, because
    # we only want to use as training docs the docs which have
    # been touched by a human for that step, which means either
    # passing that step by or being in that step in a workflow
    # state other than "awaiting hand annotation". Perhaps I should
    # exclude the documents which are less than done? Maybe some
    # other time.
    
    def do(self, folds = None, crossvalidation_doc_count_threshold = None):
    
        db = self.folder.workspace.getDB()
        # First, let's see if there are any docs in the appropriate state,
        # and sort them into steps.
        stepsToCrossvalBasenames = {}
        otherDocNamesByStep = {}

        # Get the steps. Anything that's in this step counts as an other
        # basename for this or a previous step.

        steps = [s for s in self.folder.workspace.workflow.stepList if s.isTrainable()]

        coreFolder = self.folder

        lastTouchedByDict = {}

        for docName, basename, status, assignedUser, lockedBy, wfStatus, currentStep in db.basenameInfo(None):
            if wfStatus == "awaiting crossvalidation":
                # This is a candidate.
                try:
                    stepsToCrossvalBasenames[currentStep].append((docName, basename))
                except KeyError:
                    stepsToCrossvalBasenames[currentStep] = [(docName, basename)]
                lastTouchedByDict[docName] = assignedUser or db.getLastTouchedBy(docName)
                # For every PREVIOUS step, it should go into the otherDocNames
                for step in steps:
                    if step.stepName == currentStep:
                        break
                    else:
                        try:
                            otherDocNamesByStep[currentStep].add((docName, basename))
                        except KeyError:
                            otherDocNamesByStep[currentStep] = set([(docName, basename)])
            else:
                # For every previous step, or this step if the status isn't "awaiting hand annotation",
                # throw it in the otherDocNames bucket.
                for step in steps:
                    if (step.stepName != currentStep) or (wfStatus != "awaiting hand annotation"):
                        try:
                            otherDocNamesByStep[currentStep].add((docName, basename))
                        except KeyError:
                            otherDocNamesByStep[currentStep] = set([(docName, basename)])
                    # Now, we break.
                    if step.stepName == currentStep:
                        break

        # OK, at this point we have a mapping, for each step, of the documents that
        # are awaiting crossvalidation, and the documents that can be used to support them.
        # You can crossvalidate if you have something - anything - else to build a model against,
        # so either multiple crossvalidation basenames or something in the otherDocNames. If not,
        # I think I should skip it.

        params = self.getOperationSettings()
        if params is not None:
            if folds is None:
                # Not on the command line.
                folds = params.get("folds")
            if crossvalidation_doc_count_threshold is None:
                crossvalidation_doc_count_threshold = params.get("crossvalidation_doc_count_threshold")
        
        if folds is not None:
            folds = int(folds)
        else:
            # Use default folds.
            folds = self.DEFAULT_FOLDS
        
        if crossvalidation_doc_count_threshold is not None:
            crossvalidation_doc_count_threshold = int(crossvalidation_doc_count_threshold)
            if crossvalidation_doc_count_threshold < 1:
                # Can never be less than 1.
                crossvalidation_doc_count_threshold = 1
        else:
            crossvalidation_doc_count_threshold = self.DEFAULT_CROSSVAL_COUNT

        for step, crossValEntries in stepsToCrossvalBasenames.items():
            otherDocs = otherDocNamesByStep.get(step, [])
            if not (otherDocs or (len(set([p[1] for p in crossValEntries])) > 1)):
                # Skip it.
                if self.fromCmdline:
                    print("Waiting on crossvalidation basename %s in step %s because there aren't enough other docs to build a model against." % (crossValEntries[0][1], step))
                continue

            if len(crossValEntries) < crossvalidation_doc_count_threshold:
                print("Skipping crossvalidation in step %s because the number of documents awaiting crossvalidation (%d) is less than the threshold (%d). Change this behavior with the --crossvalidation_doc_count_threshold option." % (step, len(crossValEntries), crossvalidation_doc_count_threshold))
                continue

            basenamesForCrossval = set([b[1] for b in crossValEntries])

            # OK, so first, we prepare a document set for ALL the basenames which can be used here,
            # and partition it randomly, according to the number of folds.

            basenameHash = {}
            for docName, basename in crossValEntries + list(otherDocs):
                try:
                    basenameHash[basename].add(docName)
                except KeyError:
                    basenameHash[basename] = set([docName])

            # Split up all the BASENAMES into <folds> partitions. We're using the
            # basenames as an initial proxy, rather than try to sort all the docnames
            # into the same set.
            # Use list() to freeze dictionary keys in Python 3.
            dSet = Bootstrap.DocumentSet("cv", partitions = [("p%d" % i, 1) for i in range(folds)],
                                         prefix = self.folder.dir, fileList = list(basenameHash.keys()))

            # Now, we find all the partitions which have a basename among the basenames
            # awaiting crossvalidation, and construct a converted document set for that
            # partition which has just the relevant docs in it.

            partitionsToTest = []
            convertedSets = {}
            basenameToPartition = {}
            for p, files in dSet.partitionDict.items():
                # If the seed set was small enough, files may be an empty list.
                # Remember, we didn't actually check to see if these files
                # existed; they're just all the represented basenames.
                bList = [os.path.basename(f) for f in files]
                if set(bList) & basenamesForCrossval:
                    partitionsToTest.append(p)
                docs = []
                if bList:
                    for b in bList:
                        for docName in basenameHash[b]:
                            docs.append(docName)
                        if b in basenamesForCrossval:
                            # I only need the first one.
                            basenameToPartition[b] = (p, list(basenameHash[b])[0])
                    convertedSets[p] = Bootstrap.DocumentSet(p, prefix = self.folder.dir, fileList = docs)

            allPartitions = convertedSets.keys()

            # OK. Now I have a bunch of document sets, and they have the files I need.
            # Next, we want to configure a bootstrapper.
            bsDir = os.path.join(self.folder.workspace.dir, "crossvalidation_dir_" + step)
            if os.path.exists(bsDir):
                shutil.rmtree(bsDir)

            engineOptions = {"workflow": self.folder.workspace.workflow.name,
                             "steps": step}
            prepOptions = {"output_file_type": self.folder.fileType,
                           "workflow": self.folder.workspace.workflow.name,
                           "undo_through": step}
            modelOptions = {}

            e = Bootstrap.Bootstrapper(dir = bsDir, task = self.folder.workspace.task,
                                       # Use list() to freeze dictionary values in Python 3.
                                       corpora = list(convertedSets.values()),
                                       models = [Bootstrap.TrainingRun("model_excluding_" + p,
                                                                       trainingCorpora = [(otherP, None) for otherP in allPartitions
                                                                                          if otherP != p],
                                                                       trainableStep = step,
                                                                       engineSettings = modelOptions or None)
                                                 for p in partitionsToTest],
                                       runs = [Bootstrap.TestRun("test_" + p, model = "model_excluding_" + p,
                                                                 testCorpora = [(p, None)],
                                                                 engineOptions = engineOptions or None,
                                                                 enginePrepOptions = prepOptions or None)
                                               for p in partitionsToTest])
            e.run()

            # Now that we're done, we want to submit each basename to reconciliation.
            # Above, we set up basenameToPartition to capture a single docname
            # for each basename. Multiple docnames may have been tagged as part
            # of the experiment, but we only need one.

            # ALSO: note that we can submit individual documents, not just basenames,
            # to crossvalidation. Up to this point, it really doesn't matter, since
            # the crossvalidation happens the same way no matter what. But at this point,
            # we need to read the reconciliation jobs out of the reconciliation job table.            

            _jsonIO = DocumentIO.getDocumentIO('mat-json', task = self.folder.workspace.task)

            sourceDocNamesForCrossval = set([b[0] for b in crossValEntries])

            for recBasename, (coreBasename, sourceDocNames) in db.getReconciliationJobs().items():
                if set(sourceDocNames) <= sourceDocNamesForCrossval:
                    # If the source docs in this job are all mentioned in the list of
                    # things for crossvalidation, then let's process this.
                    # Let's have a transaction for each basename.
                    t = self.folder.workspace.beginTransaction(self)
                    try:
                        p, doc = basenameToPartition[coreBasename]                        
                        allDocs = [self.folder._openFileBasename(d) for d in sourceDocNames]

                        lastTouchedBy = None
                        for d in sourceDocNames:
                            db.updateDocumentStatus(d, wfStatus = "in reconciliation")
                            lastTouchedBy = lastTouchedBy or lastTouchedByDict.get(d)

                        d = _jsonIO.readFromSource(os.path.join(bsDir, "runs", "test_" + p, "model_excluding_" + p, "hyp", doc+".prepped.tag.json"))
                        allDocNames = ["crossvalidation"] + sourceDocNames
                        allDocs[0:0] = [d]
                        requestingUser = None
                        
                        self.folder.workspace._enterReconciliation(recBasename, lastTouchedBy, allDocNames, allDocs, t, self.fromCmdline)
                        t.commit()
                    except:
                        t.rollback()
                        raise

# And here we request an ad-hoc review. Only two types are available: human and reconciliation_with_crossvalidation.
# And you can't review a document which has a scheduled review. (You shouldn't be able to anyway, since
# you can block advancement, but not submitting to review, on markgold.)

@WorkspaceOp("core")
class RequestReviewOperation(WorkspaceOperation):

    availability = CMDLINE_AVAILABLE | UI_AVAILABLE

    name = "request_review"

    args = [Option("--review_type", help="The type of ad-hoc review. Available types are 'human', 'repair' and 'reconciliation_with_crossvalidation'."),            
            Option("--lock_id", type="string", help="lock ID (if document is locked)"),
            OpArgument("user", hasArg = True,
                       help = "specify a user requesting the review")]

    def getAffectedFolders(self):
        folders = set()
        fileBasenamesSubmitted = getattr(self, "fileBasenamesSubmitted", None)
        if fileBasenamesSubmitted:
            folders.update([k for (k, v) in fileBasenamesSubmitted.items() if v])
            folders.add("core")
        return list(folders)

    def getTargetFolderAndDocuments(self):
        fileBasenamesSubmitted = getattr(self, "fileBasenamesSubmitted", None)
        if fileBasenamesSubmitted:
            return "core", self._getTargetDocuments("core")
        else:
            return None, []

    def getAffectedFileBasenames(self):
        return self.affectedFileBasenames

    def do(self, review_type = None, lock_id = None, user = None):
        if review_type not in ('human', 'repair', 'reconciliation_with_crossvalidation'):
            raise WorkspaceError("Ad-hoc review type must be either human, repair or reconciliation_with_crossvalidation")
        if user is None:
            raise WorkspaceError("request_review requires a user")
        # lock_id can be None, but it's gotta match what's in the DB.
        # if it IS none, no need to lock the document.
        db = self.folder.workspace.getDB()
        if not db.userIsRegistered(user):
            raise WorkspaceError("unknown user %s" % user)
        # If we have a lock_id, use it to collect the fileBasenames.
        # Otherwise, collect them yourself.
        if lock_id is not None:
            fileBasenames = {}
            # There had better be only one of these, otherwise this will
            # break...
            fileBasename, basename, lockingUser = db.coreGetLockIDInfo(lock_id)
            for b in self.affectedBasenames:
                if (b != basename) or (lockingUser != user):
                    raise WorkspaceError("wrong lock ID %s for basename %s" % (lock_id, b))
            fileBasenames[fileBasename] = basename
        else:
            fileBasenames = {}
            for b in self.affectedBasenames:
                fileBasename = self.folder.fileBasenameForUser(b, user)
                lockingUser = self.folder.fileBasenameLocked(fileBasename)
                if lockingUser:
                    raise WorkspaceError("can't request review for file %s without lock ID (locked by %s)" % (fileBasename, lockingUser))
                fileBasenames[fileBasename] = b
        # Now, we get the file basename info, and we barf if there's a
        # review scheduled for the current step.
        # Use list() to freeze dictionary keys in Python 3.
        bInfo = db.fileBasenameInfo(list(fileBasenames.keys()))
        repairRequest = False
        if review_type == "repair":
            review_type = "human"
            repairRequest = True
        for docName, (basename, status, assignedUser, lockedBy, workflowStatus, currentStep) in bInfo.items():
            if repairRequest:
                if status == "reconciled":
                    raise WorkspaceError("can't request repair for reconciled file %s" % docName)
            elif status != "gold":
                raise WorkspaceError("can't request review for non-gold file %s" % docName)
            if (status == "gold") and db.getReviewSchedule(currentStep):
                raise WorkspaceError("can't request review for file %s, because its current step has a scheduled review" % docName)
            if (review_type == "reconciliation_with_crossvalidation") and \
               (not self.folder.workspace.task.getStep(self.folder.workspace.workflow.name, currentStep).isTrainable()):
                raise WorkspaceError("can't request crossvalidation review for non-trainable step '%s'" % currentStep)

        self.affectedFileBasenames = fileBasenames
        t = self.folder.workspace.beginTransaction(self)
        try:
            self._do(db, bInfo, user, review_type, t, repairRequest)
            t.commit()
        except:
            t.rollback()
            raise

    def _do(self, db, bInfo, user, review_type, t, repairRequest):
        coreFolder = self.folder
        reviewFolder = self.folder.workspace.getFolder("review")
        for docName, (basename, status, assignedUser, lockedBy, workflowStatus, currentStep) in bInfo.items():
            if review_type == "human":
                self.folder.workspace._submitForHumanReview(docName, coreFolder, reviewFolder, t, db,
                                                            basename, user, currentStep, repairRequest = repairRequest)
            elif review_type == "reconciliation_with_crossvalidation":
                # Here, we use the docName as the reconciliation basename.
                db.updateDocumentStatus(docName, wfStatus = "awaiting crossvalidation") 
                db.lockCoreDocument(None, docName, "REVIEW")
                db.ensureReconciliationJobEntry(docName, docName)

#
# Logging operations
#

@WorkspaceOp()
class EnableLoggingOperation(WorkspaceToplevelOperation):

    name = "enable_logging"

    @classmethod
    def getUsage(cls, w):
        return """Usage: %prog [options] <dir> enable_logging"""

    def do(self):
        self.workspace._enableLogging()

@WorkspaceOp()
class DisableLoggingOperation(WorkspaceToplevelOperation):

    args = [OpArgument("remove_log", help = "if provided, the log will be removed when logging is disabled. Otherwise, the log is stored separately, so that when logging is enabled again, a new log is created.")]

    name = "disable_logging"

    @classmethod
    def getUsage(cls, w):
        return """Usage: %prog [options] <dir> disable_logging"""
    
    def do(self, remove_log = False):
        self.workspace._disableLogging(removeLog = remove_log)

@WorkspaceOp()
class RerunLogOperation(WorkspaceToplevelOperation):
    
    args = [OpArgument("stop_at", help = "the log timestamp to stop immediately before", hasArg = True),
            OpArgument("restart", help = "go back to the beginning"),
            OpArgument("verbose", help = "describe the state of the workspace at each point")]

    name = "rerun_log"

    @classmethod
    def getUsage(cls, w):
        return """Usage: %prog [options] <dir> [rerun_options] rerun_log"""

    def do(self, stop_at = None, restart = False, verbose = False):
        self.workspace._rerunLog(stopAt = stop_at, restart = restart, verbose = verbose)

# Here's a separate operation to upload the log.

@WorkspaceOp()
class UploadUILogOperation(WorkspaceToplevelOperation):

    name = "upload_ui_log"

    availability = NOT_AVAILABLE

    args = [OpArgument("log", help = "log fragment (from UI) to trigger fine-grained progress monitoring", hasArg = True),
            OpArgument("log_format", help = "the format of the log fragment", hasArg = True),
            OpArgument("timestamp", help = "millisecond timestamp of log upload from the server's point of view", hasArg = True)]

    def do(self, log = None, log_format = None, timestamp = None):

        t = self.workspace.beginTransaction(self)
        try:
            if log and self.workspace.logger:
                self.workspace.logger.logLog(log, log_format, timestamp)
            t.commit()
        except:
            t.rollback()
            raise

    def webResult(self):
        return {"success": True}

#
# Lock object
#


# Try to obtain the lock. The open operation is cross-platform
# (looking forward) and atomic. Thanks to
# http://www.evanfosmark.com/2009/01/cross-platform-file-locking-support-in-python/

# We don't need to keep the file open. For debugging, we want it closed, in fact.

import errno, time

class WorkspaceLock(object):

    def __init__(self, operation, dir, timeout = None):
        self.lockfile = os.path.join(dir, "opLockfile")
        
        self.start_time = time.time()
        
        # If timeout is None, don't wait at all.
        # The idea is to raise an error if the file exists (os.O_EXCL).

        try:
            res = self._lock(operation, dir, timeout)
        except Exception as e:
            # If anything, anything at all, goes wrong in here, bail.
            self.unlock()
            raise WorkspaceError("encountered error trying to lock workspace: %s" % e)
        if res is False:
            # Read the lockfile.
            fp = open(self.lockfile, "rb")
            s = fp.read()
            fp.close()
            raise WorkspaceError("workspace is currently unavailable (processing other request: %s)" % s.decode("utf-8"))
        
    def _lock(self, operation, dir, timeout):
        while True:
            try:                
                fd = os.open(self.lockfile, os.O_CREAT|os.O_EXCL|os.O_RDWR)
                break
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise
                if (timeout is None) or \
                   ((time.time() - self.start_time) >= timeout):
                    return False
                time.sleep(.05)
        # For some reason, fdopen doesn't work here. Work directly on the fd.
        # os.getlogin() isn't portable. getpass.getuser() appears to be the only option.
        import getpass
        os.write(fd, ("%s by %s @ %s" % (operation, getpass.getuser(), time.ctime())).encode("utf-8"))
        os.close(fd)
        return True

    def unlock(self):
        # If this goes pear-shaped, we're in real trouble anyway.
        try:
            os.remove(self.lockfile)
        except:
            pass

#
# Auto advancement
#

# We need this in lots of places, and customizing its behavior is getting to
# be a pain. So we do it here.

class WorkspaceAutoAdvancer(Operation):

    # First, the options, if there are any.
                   
    @classmethod
    def addOptions(cls, aggregator, ws, **kw):

        super(WorkspaceAutoAdvancer, cls).addOptions(aggregator, **kw)
        
        for step in ws.workflow.stepList:
            if step.stepImplementation is None:
                continue
            stepName = step.stepName
            if step.prettyName:
                stepName += " (pretty name %s)" % step.prettyName
            stepcls = step.stepImplementation.__class__
            activator = (None, stepcls, None, step.stepName, None)
            stepcls.addOptions(aggregator, heading = "Options for auto-advancement for workflow step " + stepName, activator = activator)
            # And then, register the activator. I'd love this to actually be a
            # dynamic choice option, but I can't figure out how to make that happen
            # when the trigger is an instance rather than a class.
            aggregator.activatedSet.add(activator)

    def __init__(self, workspace, parentOperation, opName = None):
        self.workspace = workspace
        self.parentOperation = parentOperation
        self.affectedFileBasenames = set()
        self.affectedFolders = set()
        self.opName = opName or "advance"

    def advance(self, fileBasenames, transaction, fromCmdline, lockId, workflowParams):

        self.affectedFileBasenames = set()
        self.affectedFolders = set()
    
        db = self.workspace.getDB()
                
        # Get the basename info. Figure out what we might do with
        # the various documents. Then, take a look at the ones which
        # can be autotagged and see if any of them are locked.
        # If the document is unannotated, we start at the current step;
        # otherwise, we start after it. Because we're sorting by
        # step, we need to determine whether the step we
        # pass to _getAutomatedStepOperationSettings should be the one listed
        # or the one after it.

        # Conveniently, we need to do fileBasenameInfo twice anyway, once
        # for the things that are done and once for the things that are

        # unannotated. So...

        stepDict = {}

        def addToStepDict(docName, basename, status, assignedUser, lockedBy, currentStep):
            try:
                stepEntry = stepDict[currentStep]
            except KeyError:
                stepEntry = {"someLocked": False, "entries": []}
                # For each key in stepDict, find the next steps. Note that we should
                # advance to the next hand-annotatable step, even if there are no automated
                # steps to advance over. 
                stepEntry["advanceInfo"], stepEntry["stepObjs"], stepEntry["stepToEnter"] = self.workspace._getAutomatedStepOperationSettings(currentStep = currentStep)
                stepDict[currentStep] = stepEntry
            if lockedBy:
                stepEntry["someLocked"] = True
            stepEntry["entries"].append((docName, basename, assignedUser, lockedBy, status))

        basenameInfo = db.fileBasenameInfo(fileBasenames, status = "unannotated")

        for docName, (basename, status, assignedUser, lockedBy, wfStatus, currentStep) in basenameInfo.items():
            addToStepDict(docName, basename, status, assignedUser, lockedBy, currentStep)
        
        basenameInfo = db.fileBasenameInfo(fileBasenames, workflow_status = "done")
        
        nextSteps = {}
        lastStep = self.workspace.workflow.stepList[0]
        for s in self.workspace.workflow.stepList[1:]:
            nextSteps[lastStep.stepName] = s.stepName
            lastStep = s
        nextSteps[lastStep.stepName] = None

        for docName, (basename, status, assignedUser, lockedBy, wfStatus, currentStep) in basenameInfo.items():
            # Find the next step.
            currentStep = nextSteps[currentStep]
            if currentStep is None:
                # We're at the end of the workflow; can't do anything at all.
                continue
            addToStepDict(docName, basename, status, assignedUser, lockedBy, currentStep)

        someLocked = False
        trueAffectedFileBasenames = set()
        # Use list() to freeze dictionary keys in Python 3.
        for k, v in list(stepDict.items()):
            if (not v["stepObjs"]) and (not v["stepToEnter"]):
                # We don't need to worry about those entries where
                # we're at the end of the workflow.
                del stepDict[k]
                continue
            trueAffectedFileBasenames.update([p[0] for p in v["entries"]])
            if v["someLocked"]:
                someLocked = True

        idInfo = None
        if someLocked:
            # In this situation, the only way we can proceed is if
            # there's a lock_id passed in and there's only one
            # affected basename and it's locked with that lock_id.
            if (lockId is None) or (len(trueAffectedFileBasenames) > 1):
                raise WorkspaceError("can't advance while documents are locked")
            idInfo = db.coreGetLockIDInfo(lockId)
            if idInfo[0] != list(trueAffectedFileBasenames)[0]:
                raise WorkspaceError("can't advance while documents are locked")
        
        for currentStep, stepData in stepDict.items():

            if stepData["stepObjs"]:
                # We have steps to apply.
                self.affectedFileBasenames.update([p[0] for p in stepData["entries"]])
                lastStepObj = stepData["stepObjs"][-1]

                # In order to process the command lines really correctly, we
                # pass the operationSettings to an XMLOpArgumentAggregator.

                # Now that we have multiple tagging steps, we can't just look
                # for tagger_local or tagger_model.

                forbiddenKeys = set(["input_file", "input_file_type", "output_file",
                                     "output_dir", "input_file_re", "input_encoding",
                                     "input_dir", "output_file_type", "output_encoding",
                                     "output_fsuff", "workflow", "steps", "undo_through"])

                operationSettings = self.workspace.getOperationSettings("advance")
                if operationSettings:
                    for key in operationSettings.keys():
                        if (key in forbiddenKeys) or \
                           (key.endswith("_local")) or \
                           (key.endswith("_model")):
                            raise WorkspaceError("workspace operation settings don't permit %s option to MATEngine" % key)

                if workflowParams is not None:
                    if operationSettings is None:
                        operationSettings = {}
                    for k, v in workflowParams.items():
                        if (k in forbiddenKeys) or \
                           (k.endswith("_local")) or \
                           (k.endswith("_model")):
                            raise WorkspaceError("workspace operation settings don't permit %s option to MATEngine" % k)
                        # Overrides.
                        operationSettings[k] = v                           

                runParameters = {"input_file_type": self.workspace.folders["core"].fileType,
                                 "input_encoding": "utf-8"}

                pathDict = dict([(os.path.join(self.workspace.folders["core"].dir, v[0]), v) for v in stepData["entries"]])
                # Use list() to freeze dictionary keys in Python 3.
                allPaths = list(pathDict.keys())
                transaction.addFilesToPreserve(allPaths)

                self.applyToolChain(db, stepData, pathDict, operationSettings,
                                    runParameters, allPaths, fromCmdline)

            # Here's a special case. If we're executing a final auto step (in which case stepData
            # will not have a stepToEnter, and the final step will be an auto step),
            # we need to mark the document gold. We're not
            # updating the responsible user, so we can do them all at the same time.
            # I have to call them one at a time, because markgold wants the responsible user.

            if (not stepData["stepToEnter"]) and \
               isinstance(lastStepObj, PluginMgr.AnnotationStep) and \
               (lastStepObj.sType == "auto"):
                o = self.workspace.folders["core"].getOperation("markgold")
                self.affectedFolders.add("core")
                for  docName, basename, assignedUser, lockedBy, status in stepData["entries"]:
                    o._do(db, {docName: basename}, db.getLastTouchedBy(docName),
                          True, False, transaction, lockId, False, False, None)
                
            if stepData["stepToEnter"]:
                status = "unannotated"
                wfStatus = self.workspace._docStatusToWorkflowStatus(status)
                # Don't forget to ensure the segments.
                step = stepData["stepToEnter"]
                ensureSegments = False
                coreFolder = self.workspace.getFolder("core")
                self.affectedFolders.add("core")
                if isinstance(step, PluginMgr.AnnotationStep) and step.managedSets:
                    ensureSegments = True
                for docName, basename, assignedUser, lockedBy, prevStatus in stepData["entries"]:
                    db.updateDocumentStatus(docName, status = status, step = stepData["stepToEnter"].stepName,
                                            wfStatus = wfStatus)
                    # We've advanced a step.
                    db.clearLastTouchedBy(docName)
                    if ensureSegments:
                        iData = coreFolder._openFileBasename(docName)
                        step._ensureSegments(iData, "non-gold", None)
                        p = os.path.join(coreFolder.dir, docName)
                        if not os.access(p, os.W_OK):
                            raise WorkspaceError("file %s in folder core not available for writing" % docName)
                        coreFolder.saveFile(iData, docName)

    def applyToolChain(self, db, stepData, pathDict, operationSettings, runParameters, allPaths, fromCmdline):

        dataPairs = self.workspace._applyToolChain(self.opName, operationSettings, stepData["advanceInfo"],
                                                   runParameters, allPaths, checkPathsAffected = False,
                                                   fromCmdline = fromCmdline)

        lastStepObj = stepData["stepObjs"][-1]
        # Next, we'd better check to make sure that we can write each file.
        # If we can't, we want to raise an error. We should check each
        # individual file, because we don't want ANYthing to happen
        # if the writes can fail.
        if not os.access(self.workspace.folders["core"].dir, os.W_OK | os.X_OK):
            raise WorkspaceError("folder core not available for writing")
        for p, iData in dataPairs:
            fileBasename = os.path.basename(p)
            if not os.access(p, os.W_OK):
                raise WorkspaceError("file %s in folder core not available for writing" % fileBasename)
            self.workspace.folders["core"].saveFile(iData, fileBasename)

            # What's the document status? In all the cases where we have a finalEntry,
            # the status is unannotated. Otherwise, the status comes from the document.
            if not stepData["stepToEnter"]:
                status = self.workspace._documentStatus(iData, lastStepObj, self.opName, pathDict[p][4])
                db.updateDocumentStatus(fileBasename, status = status, step = lastStepObj.stepName,
                                        wfStatus = self.workspace._docStatusToWorkflowStatus(status))
        self.affectedFolders.add("core")

    def getTargetFolderAndDocuments(self):
        return "core", self.parentOperation._getTargetDocuments("core")
    
    def getAffectedFolders(self):
        return list(self.affectedFolders)
                
#
# Workspace
#

# Each workspace will have a set of folders, and also
# a properties file, and a model directory, and a list
# of files which are the file names which have been
# imported.

# We have to be able to add folders before we init.
# Belay that. Pass the create flag to addFolder, from
# the __init__ method.

# _fromRerunner will be True if we're creating a workspace from
# the workspace rerunner. In this case, we DO NOT CHECK for initial users.
# The logger will either be initialized in workspaceCustomize, in which
# case the registerUsers action will be captured in the log, or it
# will be initialized after the initial users were captured, and thus they'll
# already be in the DB.

class Workspace(object):

    def __init__(self, dir, create = False, taskName = None, workspaceConfig = None,
                 language = None, maxOldModels = 0,
                 pluginDir = None, initialUsers = None, similarityProfile = None,
                 _fromRerunner = False):
        
        self.dir = os.path.abspath(dir)
        self.task = None
        self.workflow = None
        self.operationSettings = None
        self.language = None
        self.similarityProfile = None
        pluginDir = pluginDir or PluginMgr.LoadPlugins()
        self.maxOldModels = maxOldModels
        self.folderDir = os.path.join(self.dir, "folders")
        self.modelDir = os.path.join(self.dir, "models")

        # Many of these settings here look forward to future versions
        # of MAT.
        self._db = None
        self.currentTransaction = None
        self.loggingEnabled = False
        self.logger = None
        self._prioritizer = None
        self.prioritizationClass = None
        self.toplevelOperations = {}
        # This can be updated in workspaceCustomize.
        self.advancerClass = WorkspaceAutoAdvancer

        createWorkspace = False
        if not os.path.isdir(self.dir):
            if create:
                self.task = self._findTask(taskName, pluginDir)
                self.language = self._findLanguage(language)
                self.workflow, self.operationSettings = self._findConfig(workspaceConfig)
                self.similarityProfile = self._findSimilarityProfile(similarityProfile, create = True)
                # You can only get a named config if you ask for it.
                self.wsConfig = workspaceConfig or self.workflow.name
                createWorkspace = True                
            else:
                raise WorkspaceError("no workspace at %s" % self.dir)
        elif create:
            raise WorkspaceError("can't create a workspace in existing directory %s" % self.dir)
        elif not os.access(self.dir, os.X_OK | os.R_OK):
            # If the user can't list or read the directory,
            # we've got trouble. Later, we'll test other things.
            raise WorkspaceError("insufficient permissions on workspace '%s'" % self.dir)
        elif not self._isWorkspace():
            raise WorkspaceError("directory %s is not a legal workspace" % self.dir)
        else:
            self._configureFromDir(pluginDir)

        # By this point, all the metadata is instantiated. Now we can
        # configure the IO.
        
        if not self.task:
            raise WorkspaceError("no task for workspace")
        self.richFileType = 'mat-json'
        self.richFileIO = getDocumentIO(self.richFileType, task = self.task)

        self.folders = {
            "core":
            CoreWorkspaceFolder(self, "core", description = "for all files during normal annotation",
                                operations = list(_OPERATION_REPOSITORY["core"].values())),
            "review":
            ReviewWorkspaceFolder(self, "review", description = "for human review files",
                                  importTarget = False,
                                  operations = list(_OPERATION_REPOSITORY["review"].values())),
            "export":
            WorkspaceFolder(self, "export", description = "for exported files",
                            importTarget = False,
                            operations = list(_OPERATION_REPOSITORY["export"].values())),
            "reconciliation":
            ReconciliationWorkspaceFolder(self, "reconciliation", 
                                          description = "for reconciliation files",
                                          importTarget = False,
                                          operations = list(_OPERATION_REPOSITORY["reconciliation"].values()))
        
            }
            
        # Not adding the OpenFile operation, yet. Don't want it available
        # on the command line, but it's available to be added and specialized
        # by tasks.
        
        self.addToplevelOperations(dict(_OPERATION_REPOSITORY["toplevel"]))

        if createWorkspace:
            if (initialUsers is None) and (not _fromRerunner):
                raise WorkspaceError("you must provide at least one user name to create a workspace (on the command line, use the --initial_users option)")
            self._createWorkspace()            
        
        self.task.workspaceCustomize(self, create = create)
        
        if createWorkspace and (not _fromRerunner):
            self.registerUsers(initialUsers)

    def getDB(self):
        if self._db is None:
            self._db = WorkspaceDB.WorkspaceDB(self)
            if self.prioritizationClass:
                self.prioritizationClass.enhanceDB(self._db)
        return self._db

    def closeDB(self):
        if self._db is not None:
            self._db.close()
            self._db = None

    # These next two are to support the "with" constructor,
    # to ensure that the DB gets shut down when the workspace context
    # is exited. This is important on Windows, where stray file
    # pointers appear not to be cleaned up properly.
    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.closeDB()        

    # These three return nothing in MAT 2. Someday, we may move over the prioritization
    # from TooCAAn, but not right now.
    
    def getPrioritizer(self):
        return None
    
    def getPrioritizationClassName(self):
        return None

    def getPrioritizationMode(self):
        return None
    
    def getModelPathForStep(self, step):
        return os.path.join(self.modelDir, "model_%s_%s" % \
                            (re.sub("\s+", "_", self.language),
                             re.sub("\s+", "_", step.stepName)))

    def describeFolders(self, importableOnly = False):
        if importableOnly:
            importValues = [True]
        else:
            importValues = [True, False]
        return "\n".join(["%s: %s" % (key, val.description)
                          for key, val in self.folders.items()
                          if val.importTarget in importValues])

    def _findTask(self, taskName, plugins):
        
        # Find the task, first and foremost.

        if len(plugins.values()) == 0:
            raise WorkspaceError("no tasks available")

        if taskName is None:
            if len(plugins.values()) == 1:
                # In Python 3, values() isn't indexable because it's a view.
                return list(plugins.values())[0]
            else:
                raise WorkspaceError("no task specified")
        
        t = plugins.getTask(taskName)
        if t is None:
            raise WorkspaceError("unknown task '%s'" % taskName)
        else:
            return t

    def _findConfig(self, wsConfig):
        try:
            return self.task.getWorkspaceDescription(wfOrConfigName = wsConfig, language = self.language)
        except PluginMgr.PluginError as e:
            raise WorkspaceError("error finding workspace configuration for task '%s': %s" % (self.task.name, str(e)))

    def _findLanguage(self, language):
        lgs = self.task.getLanguages()
        if language is None:
            if len(lgs.keys()) > 1:
                raise WorkspaceError("task %s defines multiple languages, and no language specified for workspace" % self.task.name)
            # In Python 3, keys() isn't indexable because it's a view.
            return list(lgs.keys())[0]
        elif language not in lgs:
            raise WorkspaceError("task %s does not define language %s" % (self.task.name, language))
        else:
            return language

    # If no similarity profile is specified, then any error will be that
    # there's no defined default, and that should be a warning. If a similarity
    # profile IS specified, the error should be reported.
    
    def _findSimilarityProfile(self, simProfile, create = False):
        if simProfile is None:
            try:
                p = self.task.getSimilarityProfile()
                return (p and p.get("name")) or None
            except PluginMgr.PluginError as e:
                if create:
                    print("Warning: no similarity profile was specified for the workspace, but the task has no default similarity profile. Consider recreating the workspace with the --similarity_profile option.", file=sys.stderr)
        else:
            try:
                p = self.task.getSimilarityProfile(name = simProfile)
            except PluginMgr.PluginError as e:
                raise WorkspaceError("error finding similarity profile for task '%s': %s" % (self.task.name, str(e)))
            if p is None:
                raise WorkspaceError("no similarity profile named '%s' in task '%s'" % (simProfile, self.task.name))
            return simProfile

    def _configureFromDir(self, pluginDir):
        db = self.getDB()
        self.task = self._findTask(db.getTaskName(), pluginDir)
        self.wsConfig = db.getWorkspaceConfig()
        self.language = self._findLanguage(db.getLanguage())
        self.workflow, self.operationSettings = self._findConfig(self.wsConfig)
        self.similarityProfile = self._findSimilarityProfile(db.getSimilarityProfile())
        self.loggingEnabled = db.loggingEnabled()
        if self.loggingEnabled:
            self.logger = WorkspaceLogger.WorkspaceLogger(self)
        self.maxOldModels = db.getMaxOldModels()
        pClass = db.getPrioritizationClass()
        if pClass:
            # Gotta reopen the DB.
            self.closeDB()
            self.prioritizationClass = FindPluginClass(pClass, self.task.name)
                
    def _createWorkspace(self):
        
        os.makedirs(self.dir)
        os.makedirs(self.folderDir)
        os.makedirs(self.modelDir)
        for folder in self.folders.values():
            folder.create()
        
        db = self.getDB()
        db.initializeWorkspaceState(self.task.name, self.loggingEnabled,
                                    self.wsConfig, self.language,
                                    self.prioritizationClass, self.maxOldModels,
                                    self.similarityProfile)

    def _isWorkspace(self):
        # It's gotta have a folders subdir and a models subdir.
        return os.path.isdir(self.folderDir) and \
               os.path.isdir(self.modelDir)

    # The possible statuses:

    # "reconciled" - all HASs are reconciled
    # "gold" - all HASs are gold
    # "partially gold" - at least one HAS is gold
    # "partially corrected" - at least one AI is not owned by MACHINE
    # "uncorrected" - no AI is not owned by MACHINE
    # "unannotated" - there are no content annotations

    # Note that there's a special case here where if all the 
    # modified sets are managed and gold, the document is gold.
    # Ditto partially gold. This actually looks like it falls right
    # out of the current logic.

    def _documentStatus(self, doc, step, currentOp, curStatus):
        
        def setsPresent(doc):
            return self.task.getAnnotationTypeRepository().filterSetsPresent(doc, step.setsAdded | step.setsModified)
        
        if isinstance(step, PluginMgr.AnnotationStep):
            if step.managedSets:
                segs = [seg for seg in doc.getAnnotations(["SEGMENT"])
                        if seg.get("set") in step.managedSets]
                segStatuses = [seg["status"] for seg in segs]
                s = set(segStatuses)
                if len(s) == 1:
                    if "reconciled" in s:
                        return "reconciled"
                    elif "human gold" in s:
                        return "gold"
                elif (len(s) == 2) and ("reconciled" in s) and ("human gold" in s):
                    return "gold"
                if ("reconciled" in s) or ("human gold" in s):
                    # At least one of them is gold.
                    return "partially gold"
                # It can only be unannotated if no annotator has touched it.
                owners = set([a.get("annotator") for a in segs])
                # What if there are None elements in there? There shouldn't be.
                # But there might be. It certainly doesn't count as machine.
                if (len(owners) == 1) and "MACHINE" in owners:
                    # It's the only thing.
                    return "uncorrected"
                elif (len(owners) > 1) or ((len(owners) == 1) and (list(owners) != [None])) or \
                         setsPresent(doc):
                    return "partially corrected"
                else:
                    return "unannotated"
            else:
                # If it's not managed, the logic is a bit different. First, we might
                # have to look at the current operation.
                if setsPresent(doc):
                    if currentOp == "save":
                        # If the current operation is 'save', it should not demote the status.
                        if curStatus == "uncorrected":
                            return "partially corrected"
                        else:
                            return curStatus                
                    else:
                        # If we find annotations, it's partially corrected.
                        return "partially corrected"
                else:
                    return "unannotated"
        elif step.believedAtLeastPartiallyDone(doc):
            return "gold"
        else:
            return "unannotated"

    def _registerUsers(self, userList, roles):
        if not userList:
            raise WorkspaceError("register_users requires at least one user")
        if not roles:
            raise WorkspaceError("register_users requires at least one role")
        if type(userList) in _strtypes:
            userList = [userList]
        for u in userList:
            if u.startswith("__"):
                raise WorkspaceError("can't register user %s: leading double underscore forbidden" % u)
            if "," in u:
                raise WorkspaceError("can't register user %s: comma forbidden" % u)
            # No whitespace, either.
            import re
            if re.search("\s", u):
                raise WorkspaceError("can't register user %s: whitespace forbidden" % u)
        db = self.getDB()
        existingUsers = db.listUsers()
        overlap = set(userList) & set(existingUsers)
        if overlap:
            raise WorkspaceError("can't register users %s: already registered" % ", ".join(overlap))
        db.registerUsers(userList, roles)

    def _listUsers(self, no_roles = False):
        db = self.getDB()
        if no_roles:
            return db.listUsers()
        else:
            return db.listUsersAndRoles()

    def _addRoles(self, userList, roles):
        if not userList:
            raise WorkspaceError("add_roles requires at least one user")
        if not roles:
            raise WorkspaceError("add_roles requires at least one role")
        if type(userList) in _strtypes:
            userList = [userList]
        db = self.getDB()
        existingUsers = db.listUsers()
        if set(userList) - set(existingUsers):
            raise WorkspaceError("can't update user roles: some users not registered")
        db.addUserRoles(userList, roles)
        
    def _removeRoles(self, userList, roles):
        if not userList:
            raise WorkspaceError("remove_roles requires at least one user")
        if not roles:
            raise WorkspaceError("remove_roles requires at least one role")
        if type(userList) in _strtypes:
            userList = [userList]
        db = self.getDB()
        existingUsers = db.listUsers()
        if set(userList) - set(existingUsers):
            raise WorkspaceError("can't update user roles: some users not registered")
        # Can't remove some roles when they're being used.
        docAssigned, docLocked, recLocked, revLocked = db.usersInUse()
        why = []
        for user in userList:            
            for role in roles:
                if role == "annotator":
                    if user in docAssigned:
                        why.append("%s assigned to core documents" % user)
                    elif user in docLocked:
                        why.append("%s editing core document" % user)
                elif user in recLocked:
                    why.append("%s editing a reconciliation document" % user)
                elif user in revLocked:
                    why.append("%s editing a review document" % user)
        if why:
            raise WorkspaceError("can't remove roles: %s" % (", ".join(why)))
        db.removeUserRoles(userList, roles)

    def _scheduleReview(self, step, reviewType, fromCmdline = False):
        if reviewType not in ("human", "reconciliation", "reconciliation_with_crossvalidation"):
            raise WorkspaceError("review type must be one of human, reconciliation, reconciliation_with_crossvalidation")
        thisStep = self.workflow.getStepByName(step)
        if thisStep is None:
            raise WorkspaceError("unknown step '%s' in workspace workflow" % step)
        if not isinstance(thisStep, PluginMgr.AnnotationStep):
            raise WorkspaceError("can't schedule review for non-annotation step '%s'" % step)
        if (thisStep.sType == "auto") and (thisStep is not self.workflow.stepList[-1]):
            raise WorkspaceError("can't schedule review for non-final fully automated step '%s'" % step)
        if (reviewType == "reconciliation_with_crossvalidation") and (not thisStep.isTrainable()):
            raise WorkspaceError("can't schedule crossvalidation review for non-trainable step '%s'" % step)
        # OK, we've got the step, and we've got the review type.
        # Record it. It may overwrite an existing schedule. But if there are
        # any documents awaiting review for the existing schedule, you can't overwrite it.
        db = self.getDB()
        curSched = db.getReviewSchedule(thisStep.stepName)
        if curSched is not None:
            # Check what's already there.
            if curSched == reviewType:
                if fromCmdline:
                    print("Review type %s is already scheduled for step %s." % (reviewType, thisStep.stepName))
                    # Nothing to do.
                return
            representedStatuses = db.getCoreDocumentWorkflowStatuses(thisStep.stepName)
            if representedStatuses and (set(representedStatuses) & set(["awaiting review", "in review", "awaiting reconciliation partner", "awaiting crossvalidation", "in reconciliation"])):
                raise WorkspaceError("can't change review type: documents currently awaiting or in review")
        db.setReviewSchedule(thisStep.stepName, reviewType)
        if fromCmdline:
            if curSched:
                print("Changed review schedule for step %s from %s to %s." % (thisStep.stepName, curSched, reviewType))
            else:
                print("Set review schedule for step %s to %s." % (thisStep.stepName, reviewType))

    def _unscheduleReview(self, step, fromCmdline = False):
        thisStep = self.workflow.getStepByName(step)
        if thisStep is None:
            raise WorkspaceError("unknown step '%s' in workspace workflow" % step)
        # OK, we've got the step, and we've got the review type.
        # Record it. It may overwrite an existing schedule. But if there are
        # any documents awaiting review for the existing schedule, you can't overwrite it.
        db = self.getDB()
        curSched = db.getReviewSchedule(thisStep.stepName)
        if curSched is None:
            if fromCmdline:
                print("No review currently scheduled for step %s." % (thisStep.stepName))
                # Nothing to do.
            return
        else:
            representedStatuses = db.getCoreDocumentWorkflowStatuses(thisStep.stepName)
            if representedStatuses and (set(representedStatuses) & set(["awaiting review", "in review", "awaiting reconciliation partner", "awaiting crossvalidation", "in reconciliation"])):
                raise WorkspaceError("can't remove review type: documents currently awaiting or in review")
            db.removeReviewSchedule(thisStep.stepName)
            if fromCmdline:
                print("Removed review schedule for step %s." % thisStep.stepName)
    
    # We need to take the document which has been removed from reconciliation
    # and reintegrate the votes into the core documents.

    # We're always going to be in a transaction. If the
    # transaction fails, we want to ensure that the reconciliation
    # document is restored (and it may already be in the list of documents
    # which we need to preserve), and if we reintegrate, we have
    # to ensure the core docs are preserved too.
    
    def _removeFromReconciliation(self, basenames, recFolder, transaction, suppressAdvancement, verbose = False, reintegrate = True):
        db = self.getDB()
        reintegrationDocs = {}

        # It may not actually be in the reconciliation folder yet (maybe it's just
        # awaiting crossvalidation, for instance). So we have to check.

        reintegrationBasenames = []
        for basename in basenames:
            if not db.beingReconciled(basename):
                print("Warning: the DB doesn't know that basename '%s' is being reconciled." % basename)
            elif db.reconciliationBasenameLocked(basename):
                raise WorkspaceFileLockedError("no basenames removed from reconciliation: basename '%s' locked for reconciliation review" % basename)
            elif os.path.isfile(os.path.join(recFolder.dir, basename)):
                reintegrationBasenames.append(basename)
                transaction.addFilesToPreserve([os.path.join(recFolder.dir, basename)], check = True)
                
        # Capture this info before we remove from reconciliation.
        recJobs = db.getReconciliationJobs()
        for basename in basenames:
            db.removeFromReconciliation(basename)
            coreBasename, coreDocnames = recJobs[basename]
            db.forceUnlockCoreDocnames("REVIEW", coreDocnames)
            for docName in coreDocnames:
                db.updateDocumentStatus(docName, wfStatus = "done")

        if reintegrate:
            # Load the documents before you remove them from the reconciliation
            # folder if you need to reintegrate.
            for basename in reintegrationBasenames:
                recDoc = recFolder.openFile(basename)
                if recDoc is not None:
                    reintegrationDocs[basename] = recDoc
        
        recFolder.clear(basenames = basenames)
        if verbose:
            print("Removed basenames from reconciliation:", " ".join(basenames))
        if reintegrate:
            coreFolder = self.getFolder("core")
            # Export the reconciliation document, and save it to each basename.            
            for basename in reintegrationBasenames:
                coreBasename, sourceDocNames = recJobs[basename]
                transaction.addFilesToPreserve([os.path.join(coreFolder.dir, n) for n in sourceDocNames])
                sourceDocs = [coreFolder._openFileBasename(n) for n in sourceDocNames]
                try:
                    recDoc = reintegrationDocs[basename]
                except KeyError:
                    # There's no entry, because the document wasn't found.
                    continue
                # This surgically alters the reconciliation document,
                # and turns it into a normal document. 
                # I think I ought to copy the segments for the current
                # step, if any, from the source docs into this document
                # and also the phasesDone information. 
                exportReconciliationDocument(self.task, recDoc)
                for n, doc in zip(sourceDocNames, sourceDocs):
                    recDoc.setSetsDone(doc.getSetsDone()[:])
                    recDoc.removeAnnotations(["SEGMENT"])
                    recDoc.importAnnotations(doc, atypes = ["SEGMENT"])
                    coreFolder.saveFile(recDoc, n)
                    db.clearLastTouchedBy(n)
                # And finally, mark the appropriate segments reconciled.
                # Because they already have an assigned user, I don't need
                # to change it.
                self._updateDocumentStatuses(sourceDocNames, coreFolder, transaction, ["human gold"],
                                             "reconciled", None)
        if not suppressAdvancement:
            fileBasenames = []
            for basename in basenames:
                coreBasename, docNameList = recJobs[basename]
                fileBasenames += docNameList
            self._autoAdvance(fileBasenames, transaction, verbose)

    # fileBasenames is the actual file name, not the workspace basename.

    # So in this case, we only want to update the SEGMENTs if they're the
    # last step that mentions the set.
    
    def _updateDocumentStatuses(self, fileBasenames, folder, transaction, segStatuses, newSegStatus,
                                responsibleUser):
        # This is used to update things to human gold and reconciled - so
        # when the doc is updated, everything in the doc can be removed from
        # the segmenter.
        # And, fixed it so that it can mark documents non-gold too - in this
        # case, they're already gold, so removing them from the segmenter is
        # superfluous. Also, if they're already gold, we don't want to change
        # the responsible user.
        if responsibleUser is None and ("non-gold" in segStatuses):
            raise WorkspaceError("updating non-gold statuses requires a user")
        segmenter = None
        db = self.getDB()
        bDict = db.fileBasenameInfo(fileBasenames)
        stepDict = dict([(step.stepName, step) for step in self.workflow.stepList])
        if newSegStatus == "reconciled":
            newDocStatus = "reconciled"
            newWfStatus = "done"
        elif newSegStatus == "human gold":
            newDocStatus = "gold"
            newWfStatus = "done"
        else:
            newDocStatus = "partially corrected"
            newWfStatus = "in hand annotation"
        for b in fileBasenames:
            doc = folder._openFileBasename(b)
            basename, status, assignedUser, lockedBy, wfStatus, currentStep = bDict[b]
            if status == newSegStatus:
                # Nothing to do.
                continue
            # Find the step. If the target status is gold or reconciled, we want
            # to mark all the previous SEGMENTs the same way, although we don't
            # want to update the annotator.
            step = stepDict[currentStep]
            modified = False
            segAnnots = [seg for seg in doc.getAnnotations(["SEGMENT"])]
            if newDocStatus in ("gold", "reconciled"):
                for prevStep in self.workflow.stepList:
                    if prevStep is step:
                        break
                    if isinstance(prevStep, PluginMgr.AnnotationStep) and prevStep.managedSets:
                        for annot in segAnnots:
                            if (annot.get("set") in prevStep.managedSets) and \
                               (annot.get("set") in prevStep.finalSetMentions) and \
                               (annot["status"] != newSegStatus):
                                annot["status"] = newSegStatus
                                modified = True
            if step.managedSets:
                # Gotta mark all the segments gold, and attribute all
                # the annotations which are attributed to MACHINE to the
                # relevant person. But if we're marking gold or reconciled,
                # only do the ones which are in the final step.
                for annot in segAnnots:
                    if (annot.get("set") in step.managedSets) and (annot["status"] in segStatuses):
                        if newSegStatus is not None:
                            if (newWfStatus == "done") and (annot.get("set") not in step.finalSetMentions):
                                pass
                            else:
                                annot["status"] = newSegStatus
                                modified = True
                        # FOR THE MOMENT, IN ORDER TO TICKLE THE TRAINER, WE NEED
                        # TO MARK THE SEGMENT.
                        if responsibleUser is not None:
                            annot["annotator"] = responsibleUser
                            modified = True
            if modified:
                transaction.addFilesToPreserve([os.path.join(folder.dir, b)], check = True)
                folder.saveFile(doc, b)
            if segmenter is None:
                segmenter = self.getPrioritizer()
            if segmenter is not None:
                segmenter.removeDoc(b)
            db.updateDocumentStatus(b, status = newDocStatus, wfStatus = newWfStatus, lastTouchedBy = responsibleUser)

    # This is used in multiple places, so they need to go on the workspace.
    # It happens in the context of an existing transaction.

    def _autoAdvance(self, fileBasenames, transaction, fromCmdline, lockId = None, workflowParams = None):

        # The transaction's operation is the original operation that created
        # the transaction, which should be the right one to ask the appropriate
        # questions of in the advancer.
        advancer = self.advancerClass(self, transaction.op)

        advancer.advance(fileBasenames, transaction, fromCmdline, lockId, workflowParams)

        return advancer

    def _docStatusToWorkflowStatus(self, currentStatus):
        if currentStatus in ("reconciled", "gold"):
            return "done"
        elif currentStatus in ("unannotated", "uncorrected"):
            return "awaiting hand annotation"
        elif currentStatus in ("partially gold", "partially corrected"):
            return "in hand annotation"                
        
    def _getAutomatedStepOperationSettings(self, currentStep = None, suppressModelApplication = False):
        # Which steps should we be applying? From the starting point, we should
        # apply all steps which are entirely automated, plus if we have
        # a step which is not auto and not hand, we should apply it if
        # we can (e.g., if it's trainable, make sure there's a model).
        # In addition, we want to indicate which step should finally
        # be entered after tagging.
        steps = []
        settings = {"workflow": self.workflow.name}
        finalEntry = None
        startChecking = (currentStep is None)
        
        for step in self.workflow.stepList:

            # See if we should start checking yet. If we have a
            # step name that matches, we start.
            if not startChecking:
                if step.stepName == currentStep:
                    startChecking = True
                else:
                    continue

            if isinstance(step, PluginMgr.TransformStep):
                steps.append(step)
            elif step.sType == "auto":
                steps.append(step)
            else:
                # We're going to stop here - the only
                # question is how. If it's not a hand step,
                # execute it if it's either not trainable or
                # there's a model for it.
                added = False
                if step.sType != "hand":
                    if step.isTrainable():
                        p = self.getModelPathForStep(step)
                        if os.path.exists(p) and (not suppressModelApplication):
                            steps.append(step)
                            added = True
                            settings[step.stepName + "_model"] = p
                            settings[step.stepName + "_local"] = "yes"
                    else:
                        steps.append(step)
                        added = True
                if not added:
                    # If we aren't executing anything in this step, we should at least enter it.
                    finalEntry = step
                break
        settings["steps"] = [s.stepName for s in steps]
        return settings, steps, finalEntry

    def _applyToolChain(self, opName, operationSettings, operationSettingsOverrides,
                        runParameters, allPaths, checkPathsAffected = True, fromCmdline = False,
                        synchronizePhasesDone = False):        

        # In workspaces the default is not to synchronize the phases done.
        
        if operationSettings is None:
            operationSettings = {}
        else:
            operationSettings = operationSettings.copy()
        if operationSettingsOverrides:
            operationSettings.update(operationSettingsOverrides)

        if not operationSettings:
            raise WorkspaceError("no operation settings provided for operation %s" % opName)

        # Now, we've got our settings. At least workflow and steps are defined.

        try:        
            workflow = operationSettings["workflow"]
        except KeyError:
            raise WorkspaceError("workflow undefined for operation %s" % opName)
        
        del operationSettings["workflow"]
        
        # Now, we run the engine, either on the specified
        # files (which should be just basenames), or on
        # everything in the folder.
        
        if (not allPaths) and checkPathsAffected:
            raise WorkspaceError("no paths affected for operation %s" % opName)

        runParameters = runParameters or {}
        
        # Now, run the operation. Note that aggregatorRun() will
        # use the runParameters which are passed both as defaults
        # for the original values and as defaults to be overridden,
        # so any values that need to be futzed with will be futzed with.

        aggr = XMLOpArgumentAggregator(operationSettings)
        try:
            e = ToolChain.MATEngine(workflow = workflow, taskObj = self.task,
                                    synchronizePhasesDone = synchronizePhasesDone)
            dataPairs = e.aggregatorRun(aggr, inputFileList = allPaths,
                                        **runParameters)
        except (Error.MATError, ToolChain.ConfigurationError) as exc:
            if ExecutionContext._DEBUG:
                raise
            else:
                if isinstance(exc, Error.MATError):
                    errstr = exc.errstr
                else:
                    ignore, errstr = exc.args
                raise WorkspaceError("toolchain application failed for operation %s: %s" % (opName, errstr))
        if fromCmdline:
            # Where were the steps? Well, there may not be any. Make sure you think about those cases.
            steps = operationSettings.get("steps") or runParameters.get("steps")
            if steps:
                if type(steps) not in _strtypes:
                    steps = ", ".join(steps)
                print("Automatically applied", steps, "to:", " ".join([os.path.basename(p) for p in allPaths]))

        return dataPairs

    # This is used in multiple places, so they need to go on the workspace.
    # It happens in the context of an existing transaction.

    # First, we group all the done fileBasenames by step. Then, if the
    # step is scheduled for a review, we process each basename. Actually,
    # we have to pay attention to the basenames which have at least one done
    # file, because we need to know whether they may be waiting for a
    # partner.

    # If it's locked, it can't be submitted for review. But if it's locked,
    # it shouldn't be done anyway. Well, if this is being called from markgold, for instance,
    # it definitely SHOULD be done if the basename in the lockInfo is the same
    # basename. 

    def _submitForReview(self, fileBasenames, requestingUser, transaction, lockID = None, deferReconciliation = False,
                         fromCmdline = False):
        fileBasenamesSubmitted = {"review": [], "reconciliation": [], "core": []}
        db = self.getDB()
        scheds = db.getReviewSchedules()
        # Obviously, if there are no schedules, don't bother going any farther.
        if not scheds:
            return
        humanSteps = [p[0] for p in scheds.items() if p[1][0] == "human"]
        if lockID:
            lockInfo = db.coreGetLockIDInfo(lockID)
        else:
            lockInfo = None, None, None            
        reviewFolder = self.folders["review"]
        coreFolder = self.folders["core"]
            
        for humanStep in humanSteps:            
            fbInfo = db.fileBasenameInfo(fileBasenames, step = humanStep, status = "gold")
            for docName, [basename, status, assigned_user, locked_by, workflow_status, current_step] in fbInfo.items():
                # Skip the ones which are locked by review.
                if locked_by == "REVIEW":
                    continue
                # Skip the ones which are locked, but not by you.
                if (lockID is None) and (locked_by is not None):
                    continue
                # If you've got a lock, but this isn't it, barf. It's a document you asked for,
                # and it's eligible to be submitted, but you can't do it. 
                if ((lockID is not None) and ((lockInfo[0] != docName) or (lockInfo[2] != locked_by))):
                    raise WorkspaceError("trying to submit a document %s for human review without the proper lock ID" % docName)

                self._submitForHumanReview(docName, coreFolder, reviewFolder, transaction, db,
                                           basename, requestingUser, humanStep)
                fileBasenamesSubmitted["review"].append(docName)
                if fromCmdline:
                    print("Submitted for human review:", docName, "at", humanStep)

            # Done with it. Remove.
            del scheds[humanStep]
        
        if not scheds:
            # All our schedules, if any, were human.
            return fileBasenamesSubmitted

        # The next bit is more complicated. In all these other cases, we need all the
        # basenames, and we need to see if the latest step of a doc is done.
        # Actually, we also need to distinguish between gold and reconciled status,
        # because reconciled must count as a "later" status than gold, so that
        # docs that subsequently enter "gold" don't get stranded in reconciliation. 

        # First, get the list of basenames, then get all the info for those basenames,
        # then find those which are done for the most recent step.

        bDict = {}

        # For each basename, we need to choose the latest current_step. That's
        # the step we check. If it's not "done", or if it doesn't have a schedule, we skip it.

        posDict = dict(zip([s.stepName for s in self.workflow.stepList], range(len(self.workflow.stepList))))
        schedsByIndex = dict([(posDict[k], v) for (k, v) in scheds.items()])

        # 1/13/21: bDict values are [stepIndex, hasGold, hasReconciled, hasSched, entryList].
        # We have to do something similar to what we do in the human pass, namely, keep
        # track of whether the current user has permission to do this. And we do check
        # the equivalence with the lockID passed in, later, in checkChangeability, but we have to
        # make sure that the value for hasGold doesn't bleed this check. The hasGold, hasReconciled,
        # and hasSched elements are used only once, in the loop which processes the bDict that's
        # constructed here. So there's probably more refactoring we can do, but I'm not
        # going to try. The minimal thing to do is to ensure that hasGold isn't blocked by
        # things that are locked by the current user.
        
        for docName, basename, status, assigned_user, locked_by, workflow_status, current_step in \
            db.basenameInfo(db.basenamesForDocnames(fileBasenames)):
            # Collect the documents for all the basenames.
            # We're doing some extra work here, since we don't know whether any
            # of the basenames are marked done, but what the hell.
            newIndex = posDict[current_step]
            lastTouchedBy = assigned_user or db.getLastTouchedBy(docName)
            # If the document is locked, and there's lock information but it isn't for this
            # document, you should be able to submit it.
            lockInfoOK = (locked_by is None) or ((lockInfo[0] == docName) and (locked_by == lockInfo[2]))
            try:
                e = bDict[basename]
                if newIndex > e[0]:
                    e[0] = curIndex = newIndex
                    e[1] = False
                    e[2] = False
                    e[3] = current_step in scheds
                e[1] |= (status == "gold") and lockInfoOK
                e[2] |= (status == "reconciled")
                e[4].append((docName, locked_by, workflow_status, newIndex, lastTouchedBy))
            except KeyError:
                bDict[basename] = [posDict[current_step], (status == "gold") and lockInfoOK,
                                   (status == "reconciled"),
                                   current_step in scheds,
                                   [(docName, locked_by, workflow_status, newIndex, lastTouchedBy)]]

        # OK, we've collected a ton of info about each eligible basename. 

        allDocs = set(fileBasenames)

        # So what might you be doing? You might be changing the status of a
        # document which is done, or a document which is already in some phase of review.
        # If you're changing the status of a document that's not in review yet, and
        # you have a lock ID, you'd better be holding the right document.

        def checkChangeability(docName, locked_by):
            if (locked_by != "REVIEW") and (lockID is not None) and \
               ((lockInfo[0] != docName) or (lockInfo[2] != locked_by)):
                raise WorkspaceError("trying to submit a document %s for review without the proper lock ID" % docName)
            # If you have no lock, and it's locked by somebody but not you,
            # skip it.
            return not ((lockID is None) and locked_by and (locked_by != "REVIEW"))

        def changeStatus(basename, currentStatus, docName, newStatus, folder):
            db.updateDocumentStatus(docName, wfStatus = newStatus)            
            db.lockCoreDocument(None, docName, "REVIEW")
            db.ensureReconciliationJobEntry(basename, docName)
            fileBasenamesSubmitted[folder].append(docName)
            
        def maybeChangeStatus(basename, currentStatus, docName, locked_by, newStatus, folder):
            if checkChangeability(docName, locked_by):
                changeStatus(basename, currentStatus, docName, newStatus, folder)

        # So what's the requesting user in the cases where we've delayed 
        # reconciliation by waiting for a partner or for crossvalidation?
        # It's one of the people to whom the document is assigned (not
        # sure it matters which one) or the last person who touched an
        # unassigned document.
        
        for basename, [stepIndex, hasGold, hasReconciled, hasSched, entryList] in bDict.items():
            # Even if
            # basename is represented by a step which is gold but not reconciled, unlocked and scheduled,
            # that might not be the latest basename. 
            if not (hasSched and hasGold and (not hasReconciled)):
                continue
            if stepIndex not in schedsByIndex:
                continue
            # OK, it's got a done document in a latest step which has a review scheduled.
            # Now, what status to assign any docName which we have in our list?
            # If they're not all done, it's gotta be awaiting reconciliation partner.
            # If they're all done and it's a crossvalidation review, they all have
            # to go into awaiting crossvalidation. Otherwise, we create a reconciliation
            # document and insert it into the reconciliation table.
            # But only change the status of things that are already locked by REVIEW, or
            # things that are locked by the lockID passed in.
            notDone = set()
            awaitingReconciliationPartner = set()
            notInThisStep = set()
            done = set()
            
            for entry in entryList:
                docName, locked_by, workflow_status, thisStepIndex, lastTouchedBy = entry
                if thisStepIndex != stepIndex:
                    notInThisStep.add(entry)
                elif workflow_status == "awaiting reconciliation partner":
                    awaitingReconciliationPartner.add(entry)
                elif workflow_status == "done":
                    done.add(entry)
                else:
                    notDone.add(entry)
            if notDone or notInThisStep:
                theseDone = [d for d in done if d[0] in allDocs]
                # Mark these all as awaiting a reconciliation partner.
                for docName, locked_by, workflow_status, thisStepIndex, lastTouchedBy in theseDone:
                    maybeChangeStatus(basename, workflow_status,
                                      docName, locked_by, "awaiting reconciliation partner", "core")
                    if fromCmdline:
                        print("Awaiting reconciliation partner:", docName)
            else:
                [reviewType, similarityProfile] = schedsByIndex[stepIndex]
                if [d for d in done if d[0] in allDocs]:
                    allDocsToReconcile = (done | awaitingReconciliationPartner)
                    if reviewType == "reconciliation_with_crossvalidation":
                        for docName, locked_by, workflow_status, thisStepIndex, lastTouchedBy in allDocsToReconcile:
                            maybeChangeStatus(basename, workflow_status,
                                              docName, locked_by, "awaiting crossvalidation", "core")
                            if fromCmdline:
                                print("Awaiting crossvalidation:", docName)
                    elif not deferReconciliation:
                        if len(allDocsToReconcile) > 1:
                            # Only proceed if they all check out.
                            allChangeable = True
                            for docName, locked_by, workflow_status, thisStepIndex, lastTouchedBy in allDocsToReconcile:
                                if not checkChangeability(docName, locked_by):
                                    allChangeable = False
                                    break
                            if allChangeable:
                                randomLastTouchedBy = None
                                for docName, locked_by, workflow_status, thisStepIndex, lastTouchedBy in allDocsToReconcile:
                                    changeStatus(basename, workflow_status, docName, "in reconciliation", "reconciliation")
                                    randomLastTouchedBy = randomLastTouchedBy or lastTouchedBy
                                self._enterReconciliation(basename, randomLastTouchedBy, [d[0] for d in allDocsToReconcile],
                                                          [coreFolder._openFileBasename(d[0]) for d in allDocsToReconcile],
                                                          transaction, fromCmdline)
                        else:
                            # We've only got one doc to reconcile. Punt to human review.
                            # If you want to hold documents on import until you accumulate
                            # more docs, then use --defer_reconciliation.
                            # If we import one by one and assign, we can't check to see if
                            # all docs are reconciled.
                            docName, locked_by, workflow_status, thisStepIndex, lastTouchedBy = allDocsToReconcile.pop()
                            step = self.workflow.stepList[thisStepIndex].stepName
                            self._submitForHumanReview(docName, coreFolder, reviewFolder, transaction, db,
                                                       basename, lastTouchedBy, step)
                            fileBasenamesSubmitted["review"].append(docName)
                            if fromCmdline:
                                print("Submitted for human review:", docName, "at", step)

        return fileBasenamesSubmitted

    def _submitForHumanReview(self, docName, coreFolder, reviewFolder, transaction, db,
                              basename, requestingUser, step, repairRequest = False):
        # Right now, make the state "in review", lock it for review,
        # copy it to the review folder.
        shutil.copyfile(os.path.join(coreFolder.dir, docName),
                        os.path.join(reviewFolder.dir, docName))
        if self.logger:
            self.logger.logCopy(coreFolder, docName, reviewFolder, docName)
        transaction.addFilesToAdd([os.path.join(reviewFolder.dir, basename)])
        # Note the lock is None. We're forcibly freeing the lock, which will
        # invalidate any future saves. The workspace UI should ensure that the
        # document is forced closed.
        db.lockCoreDocument(None, docName, "REVIEW")
        db.updateDocumentStatus(docName, wfStatus = "in review")
        db.insertReviewDocument(docName, basename, requestingUser, repairRequest)

    def _enterReconciliation(self, basename, requestingUser, docNames, docs, transaction, fromCmdline):
        db = self.getDB()
        # Enter reconciliation.
        recFolder = self.folders["reconciliation"]
        # We have to collect all the basenames and create a reconciliation
        # document.
        recDoc = generateComparisonDocument(self.task, docs[0], docs[1:],
                                            pivotLabel = docNames[0],
                                            otherLabels = docNames[1:],
                                            similarityProfile = db.getSimilarityProfile(),
                                            forReconciliation = True)
        # Check to see whether it's already fully reconciled.
        if reconciliationDocumentFullyReconciled(recDoc):
            db.forceUnlockCoreBasenames("REVIEW", [basename])
            docsForBasenames = db.documentsForBasenames(basenames = [basename])
            sourceDocNames = docsForBasenames[basename]
            
            coreFolder = self.getFolder("core")

            transaction.addFilesToPreserve([os.path.join(coreFolder.dir, n) for n in sourceDocNames])
            self._updateDocumentStatuses(sourceDocNames, coreFolder, transaction, ["human gold"],
                                         "reconciled", None)                
            if fromCmdline:
                print("Reconciliation not necessary for:", basename)
        else:
            transaction.addFilesToAdd([os.path.join(recFolder.dir, basename)])
            recFolder.saveFile(recDoc, basename)
            db.submitToReconciliation(basename, requestingUser)
            if fromCmdline:
                print("Entering reconciliation:", basename)
    
    # _openFile, by itself, needs no transaction. It only does DB updates
    # in any case, so the transaction stuff must be left to the callers.
    # We have to enforce that we require a user if it's not read only.

    def _openFile(self, folderName, basename, user, readOnly, fileBasename):
        folder = self.folders[folderName]
        if not fileBasename:
            # This is the usual case.
            fileBasename = folder.fileBasenameForUser(basename, user)
        elif (not readOnly) and (basename != fileBasename):
            # Coming from the UI, you might be asking to open a particular
            # file. If it's read only, it won't be locked, so just open it.
            # If it isn't, it better be the right file.
            properFileBasename = folder.fileBasenameForUser(basename, user)
            if properFileBasename != fileBasename:
                raise WorkspaceError("user %s can't open requested file for writing because it isn't assigned to that user" % user)
        # This can happen if we're called from the toplevel. It should
        # always be an error.
        if fileBasename is None:
            raise WorkspaceError("user %s can't open basename %s because no appropriate document was found" % (user, basename))
        return self._openFileBasename(folder, fileBasename, user, readOnly)

    # If the user is the same as the locking user, then free that lock
    # and relock. Don't reuse the lock - we want to make sure that previous
    # opens are cancelled.
    
    def _openFileBasename(self, folder, fileBasename, user, readOnly, doc = None):
        db = self.getDB()
        if not readOnly:
            lockingUser = folder.fileBasenameLocked(fileBasename)
            if lockingUser is not None:
                # If it's locked, mark it read only unless it has to be writeable,
                # in which case raise an error.
                if lockingUser != user:
                    # If the locking user is the same as the requesting user,
                    # ignore the lock - it'll get overwritten below.
                    raise WorkspaceFileLockedError("can't open a locked document for writing (locked by %s)" % lockingUser)
            # Make sure the specified user exists.
            if user is None:
                raise WorkspaceUserLockError("can't open document with unknown user")
            if not db.userIsRegistered(user):
                raise WorkspaceUserLockError("can't open a document using unregistered user '%s'" % user)
            # If the user is not assigned to this basename, then check                    
            # to make sure the user has the annotation role. If it's
            # already assigned, we assume the check was done there (later we'll
            # unwind this, when we do on-demand assignment). Nah, just
            # check it always.
            if not folder.userHasAppropriateRole(user, fileBasename):
                raise WorkspaceError("user %s can't open %s for writing because the user doesn't have the appropriate role" % (user, fileBasename))                       

        # This can't fail.
        doc = doc or folder._openFileBasename(fileBasename)
        lockId = None
        if not readOnly:            
            lockId = self._generateLockID()
            folder.prepareForEditing(doc, fileBasename, user, lockId)
        return doc, fileBasename, lockId

    def _generateLockID(self):
        import random, string
        return "".join([random.choice(string.ascii_letters + string.digits) for i in range(32)])

    def _addToBasenameSet(self, setName, *basenames):
        db = self.getDB()
        relevantBasenames = set(db.allBasenames(basenames = basenames))
        if set(basenames) != relevantBasenames:            
            raise WorkspaceError("unknown basename(s) %s" % ",".join(set(basenames) - relevantBasenames))        
        for b in basenames:
            db.addBasenameToBasenameSet(b, setName)

    def _removeFromBasenameSet(self, setName, *basenames):
        db = self.getDB()
        for b in basenames:
            db.removeBasenameFromBasenameSet(b, setName)

    # Three logging methods.
    def _enableLogging(self):
        if not self.loggingEnabled:
            self.loggingEnabled = True
            self.logger = WorkspaceLogger.WorkspaceLogger(self)
            self.getDB().enableLogging()

    def _disableLogging(self, removeLog = False):
        if self.loggingEnabled:
            self.logger = None
            # Remove or move aside the log. If we enable it again, it'll start
            # with a fresh seed. Otherwise, we might lose stuff
            # in the middle while it's temporarily disabled.
            loggerP = os.path.join(self.dir, WorkspaceLogger.WorkspaceLogger.loggerDir)
            if os.path.exists(loggerP):
                if removeLog:
                    shutil.rmtree(loggerP)
                else:
                    # Move it aside.
                    i = 0
                    while True:
                        p = loggerP + "_" + str(i)
                        if not os.path.exists(p):
                            shutil.move(loggerP, p)
                            break
                        i = i + 1
            self.loggingEnabled = False      
            self.getDB().disableLogging()

    def _rerunLog(self, stopAt = None, restart = False, verbose = False):
        r = WorkspaceLogger.WorkspaceRerunner(self, restart = restart)
        r.rollForward(stopAt = stopAt, verbose = verbose)
    
    # We need to keep track of the transactions because we need
    # to access the workspace logging, possibly.
    
    def beginTransaction(self, op, **kw):
        self.currentTransaction = WorkspaceTransaction(self, op, **kw)
        return self.currentTransaction

    #
    # Public methods
    #

    # Use this method to add a folder to the workspace.
    # This is meant to be called after super __init__
    # is called.

    def addFolder(self, name, dirName, folderClass = WorkspaceFolder, create = False, **kw):
        folder = folderClass(self, dirName, prettyName = name, **kw)
        self.folders[name] = folder
        if create:
            folder.create()
        elif not os.path.isdir(folder.dir):
            raise WorkspaceError("no folder named %s in workspace at %s" %
                                   (name, self.dir))

    # This method imports files.

    # Toplevel operation. MUST BE LOCKED. Originally, the core of the import
    # operation was also implemented here, but it now involves a MAT engine
    # operation, so we've moved it all to the toplevel operation itself.

    def importFiles(self, files, folderName, **kw):
        return self.runOperation("import", (folderName,) + tuple(files),
                                 fromCmdline = False, resultFormat = FN_RESULT,
                                 **kw)

    # This function returns all the legal basenames.
    # Overriding because the core behavior is now really
    # inefficient.

    def getBasenames(self, basenameList = None):
        return self.getDB().allBasenames(basenames = basenameList)

    # Toplevel operation. MUST BE LOCKED.

    def getFolderBasenames(self, *folderNames):

        return self.runOperation("list", folderNames, resultFormat = FN_RESULT)

    # Corresponding internal function.

    def _getFolderBasenames(self, *folderNames):
        return dict([(folderName, self.folders[folderName].getBasenames())
                     for folderName in folderNames if (folderName in self.folders)])

    # Toplevel operation. MUST BE LOCKED. This can't use
    # runOperation because it's not registered as an operation.

    def openWorkspaceFile(self, folder, basename, **kw):
        
        return self.runOperation("open_file", (folder, basename),
                                 resultFormat = FN_RESULT,
                                 **kw)

    # Toplevel operation. MUST BE LOCKED. Provided for backward compatibility.

    def removeBasenames(self, basenameList = None):

        return self.runOperation("remove", basenameList or [])

    def removeAllBasenames(self):

        return self.runOperation("remove", self.getBasenames())

    # COMPLETELY rewritten from the previous version, because I need
    # the transaction in here. And, I need to worry about getting files
    # from folders. And we no longer need to check the file list via
    # the previous method, AND we ought to migrate the call to discardBasenames
    # into here.

    def _removeBasenames(self, transaction, basenameList = None):

        # For ease of interaction, let's check all the basenames
        # first, so we can raise an error.

        logger = self.logger
        for f in self.folders.values():
            # This was f.clear(), but I have to copy it
            # to  get the transactions to work.
            for p in f.getFiles(basenameList):
                if os.path.isfile(p):
                    transaction.addFilesToPreserve([p], check = True)
                    os.remove(p)
                    if logger:
                        self.logger.logRemoveFile(f, os.path.basename(p))
        
        # Let's find the actual basenames we're removing.
        db = self.getDB()            
        basenamesRemoved = db.allBasenames()
        if basenameList is not None:
            basenamesRemoved = list(set(basenamesRemoved) & set(basenameList))

        db.discardBasenames(basenamesRemoved)

        return basenamesRemoved

    def getFolder(self, folderName):
        try:
            return self.folders[folderName]
        except KeyError:
            raise WorkspaceError("unknown folder %s" % folderName)

    def getFolderOperation(self, folderName, operationName, basenames = None, transaction = None):
        folder = self.getFolder(folderName)
        return folder.getOperation(operationName, basenames = basenames, transaction = transaction)

    # We're going to change the search method. Operations will no longer
    # require a folder, except when a dispatch is required. Furthermore,
    # you can define operations with the same name on different folders.
    
    def getOperation(self, operationName, args, transaction = None):
        try:
            cls = self.toplevelOperations[operationName]
            # Instantiate the class.
            return cls(self, args, transaction = transaction)
        except KeyError:
            if not args:
                raise WorkspaceError("no folder specified for folder operation")
            folderName = args[0]
            basenames = args[1:] or None
            return self.getFolderOperation(folderName, operationName, basenames = basenames, transaction = transaction)

    # We're going to use these classes to create separate command line option groups.
    # You shouldn't be able to add a folder operation which already exists
    # as a toplevel operation, or vice versa. I've now guaranteed that.
    
    def getOperationClasses(self, operationName):
        try:
            return [("<toplevel>", self.toplevelOperations[operationName])]
        except KeyError:
            return [(fName, f.operations[operationName]) for fName, f in self.folders.items()
                    if (operationName in f.operations)]

    def getCmdlineOperationNames(self, debug = False):
        s = set()
        for f in self.folders.values():
            s.update([o.name for o in f.getCmdlineOperations(debug = debug)])
        if debug:
            avail = CMDLINE_DEBUG_AVAILABLE_MASK
        else:
            avail = CMDLINE_AVAILABLE_MASK
        return [o.name for o in self.toplevelOperations.values() if (o.availability & avail)] + list(s)

    # Toplevel operation. MUST BE LOCKED.

    def runOperation(self, operationName, otherArgs,
                     aggregator = None, resultFormat = NULL_RESULT,
                     fromCmdline = False, **params):
        # This will raise an error if you can't get the lock.
        lock = WorkspaceLock("%s on %s" % (operationName, ", ".join(otherArgs)) , self.dir)

        try:
            o = self.getOperation(operationName, otherArgs)
            return self._runOperation(o, aggregator = aggregator, resultFormat = resultFormat,
                                      fromCmdline = fromCmdline,
                                      **params)
        finally:
            lock.unlock()

    # Toplevel convenience utility, for backward compatibility.

    def runFolderOperation(self, folderName, operationName,
                           aggregator = None, basenames = None, resultFormat = NULL_RESULT,
                           fromCmdline = False,
                           **params):

        lock = WorkspaceLock("%s on %s" % (operationName, folderName) , self.dir)

        try:
            o = self.getFolderOperation(folderName, operationName, basenames = basenames)
            return self._runOperation(o, aggregator = aggregator, resultFormat = resultFormat,
                                      fromCmdline = fromCmdline,
                                      **params)
        finally:
            lock.unlock()

    # Internal utilities. THESE MUST ALWAYS BE CALLED FROM A LOCKED FUNCTION.

    # OOOH, this is hideous. I have to
    # ensure that there's always a user. So for every operation, I have
    # to check this. What a waste. Of course, unless the operation is
    # registerUsers. Perhaps this can go away when create takes a user. And it has.

    def _runOperation(self, o, aggregator = None, resultFormat = NULL_RESULT, fromCmdline = False,
                      **params):
        try:
            o.enhanceAndValidate(aggregator, self, **params)
            o.fromCmdline = fromCmdline
            o.doOperation()
            if resultFormat == WEB_RESULT:
                return o.webResult()
            elif resultFormat == FN_RESULT:
                return o.fnResult()
            else:
                return None
        finally:
            self.closeDB()
            
    # This function is used to ensure that the current user can access the
    # workspace. It's used primarily in the CGI call. The other files
    # need to have been checked earlier.

    # On Windows, it doesn't appear that you can rely on os.access -
    # it will return True for writing even when the directory flags
    # have been disabled. Not sure what to do about this.
    
    def dirsAccessible(self, forWriting = True):
        if forWriting:
            flags = os.R_OK | os.W_OK
        else:
            flags = os.R_OK
        r = os.access(self.dir, os.X_OK | flags) and \
            os.access(self.folderDir, os.X_OK | flags) and \
            os.access(self.modelDir, os.X_OK | flags) and \
            len([folder for folder in self.folders.values()
                 if not os.access(folder.dir, os.X_OK | flags)]) == 0
        if r:
            if forWriting:
                flags = os.R_OK | os.W_OK
            else:
                flags = os.R_OK
            return os.access(self.getDB().wsDBFile, flags)
        else:
            return False

    def getOperationSettings(self, name):
        try:
            return self.operationSettings[name]
        except KeyError:
            return None

    def registerUsers(self, userList, roles = None):
        if type(userList) in _strtypes:
            userList = [userList]
        self.runOperation("register_users", userList, roles = roles)

    def listUsers(self):
        return self.runOperation("list_users", (), fromCmdline = False, resultFormat = FN_RESULT)

    def addToplevelOperations(self, opDict):
        for opName, op in opDict.items():
            # If it's already a folder operation, barf.
            for fName, f in self.folders.items():
                if opName in f.operations:
                    raise WorkspaceError("Can't add toplevel operation '%s', because it's already defined in folder '%s'" % (opName, fName))
            self.toplevelOperations[opName] = op
