####################################################################################################
### -----------------------------------------  NOTICE  ----------------------------------------- ###
###                                                                                              ###
### This software (or technical data) was produced for the U.S. Government under contract        ###
### 2015-14120200002-002, and is subject to the Rights in Data-General Clause 52.227-14 , ALT    ###
### IV (MAY 2014) or (DEC 2007).                                                                 ###
###                                                                                              ###
### © 2019 The MITRE Corporation. All Rights Reserved.                                           ###
###                                                                                              ###
### -------------------------------------------------------------------------------------------- ###
####################################################################################################

"""
BP document library for the bp_lib repository.

Rules:
    - Every class minimally needs to implement ``__init__``, ``__eq__``, ``__ne__``, ``__repr__``.
    - Every class should alphabetize its functions, grouping them first by built-ins and
      non-built-ins. The expection to this is the ``__init__`` function; that should always be
      first.
    - Every ``BPDocument<node>`` class minimally needs to implement the :class:`BPNode` members.
    - The BPProperty class only needs to implement the is_valid(self) function (in addition to the
      the specified built-ins).
    - Class definitions in this library are ordered by their axiomatic level within the BP JSON
      schema. Atomic objects (that don't refer to other objects) are implemented first, with their
      parent objects subsequently implemented until we arrive at the top-level BPDocument class.
"""

__author__  = "Leland E. Vakarian <lsperl@mitre.org>, Chelsea Deane <cdeane@mitre.org>"
__date__    = "22 Aug, 2022"


from abc import ABC, abstractmethod

from lib.utils import LOGGER


DIFF_INDENT = "  "


def _helper_listlike_diff(self_prop, other_prop, level, verbose):
    import click

    global DIFF_INDENT
    global LOGGER

    if verbose:
        warn_print = LOGGER.warning
    else:
        warn_print = click.echo

    if verbose:
        LOGGER.info(f"{level*DIFF_INDENT}Making sorted copies of incoming list-likes.")

    self_prop = sorted(list(self_prop))
    other_prop = sorted(list(other_prop))

    if self_prop != other_prop and not verbose:
        click.echo(f"{level*DIFF_INDENT}In list-like:")

    for i in range(max(len(self_prop), len(other_prop))):

        if verbose:
            LOGGER.info(f"{level*DIFF_INDENT}Comparing index {i}.")

        try:
            self_prop[i]
        except IndexError:
            warn_print(f"{DIFF_INDENT*(level+1)} self's list-like is shorter than other's.")
            break

        try:
            other_prop[i]
        except IndexError:
            warn_print(f"{DIFF_INDENT*(level+1)} other's list-like is shorter than self's.")
            break

        if verbose:
            LOGGER.info(f"{(level+1)*DIFF_INDENT}Self type: {type(self_prop[i])}.")
            LOGGER.info(f"{(level+1)*DIFF_INDENT}Other type: {type(other_prop[i])}.")

        if isinstance(self_prop[i], dict) and len(self_prop[i].keys()) == 0:
            # treat as a None
            if other_prop[i] == None or other_prop[i] == {}:
                if verbose:
                    LOGGER.info(f"{(level+1)*DIFF_INDENT}Values match.")
                    LOGGER.info(
                        f"{(level+2)*DIFF_INDENT}{self_prop[i]} | {other_prop[i]}"
                    )
            else:
                warn_print(f"{(level+1)*DIFF_INDENT}Values don't match:")
                warn_print(
                    f"{(level+2)*DIFF_INDENT}{self_prop[i]} | {other_prop[i]}"
                )
            continue

        elif isinstance(other_prop[i], dict) and \
            len(other_prop[i].keys()) == 0:

            # treat as a None
            if self_prop[i] == None or self_prop[i] == {}:
                if verbose:
                    LOGGER.info(f"{(level+1)*DIFF_INDENT}Values match.")
                    LOGGER.info(
                        f"{(level+2)*DIFF_INDENT}{self_prop[i]} | {other_prop[i]}"
                    )
            else:
                warn_print(f"{(level+1)*DIFF_INDENT}Values don't match:")
                warn_print(
                    f"{(level+2)*DIFF_INDENT}{self_prop[i]} | {other_prop[i]}"
                )
            continue

        if isinstance(self_prop[i], BPProperty):

            if not isinstance(other_prop[i], BPProperty):
                warn_print((
                    f"{(level+1)*DIFF_INDENT}Found BPProperty on self, but corresponding "
                    "element in other isn't a BPProperty."
                ))
                warn_print(f"{(level+2)*DIFF_INDENT}{type(other_prop[i])}")
                continue

            if verbose:
                LOGGER.info(
                    f"{(level+1)*DIFF_INDENT}Found {self_prop[i]}. Comparing to "
                    f"{other_prop[i]}."
                )

            self_prop[i].diff(
                other_prop[i],
                level=level+2,
                verbose=verbose
            )

        elif isinstance(self_prop[i], BPNode):

            if not isinstance(other_prop[i], BPNode):
                warn_print((
                    f"{(level+1)*DIFF_INDENT}Found BPNode on self, but corresponding element in "
                    "other isn't a BPNode."
                ))
                warn_print(f"{(level+2)*DIFF_INDENT}{type(other_prop[i])}")
                continue

            if verbose:
                LOGGER.info(
                    f"{(level+1)*DIFF_INDENT}Found BPNode {self_prop[i]}. Comparing to "
                    f"{other_prop[i]}."
                )

            self_prop[i].diff(
                other_prop[i],
                level=level+2,
                verbose=verbose
            )

        elif isinstance(self_prop[i], list) or isinstance(self_prop[i], set):

            if not isinstance(other_prop[i], list) and not isinstance(other_prop[i], set):
                warn_print((
                    f"{(level+1)*DIFF_INDENT}Found list-like on self, but corresponding element in "
                    "other isn't a list-like."
                ))
                warn_print(f"{(level+2)*DIFF_INDENT}{type(other_prop[i])}")
                continue

            if verbose:
                LOGGER.info(f"{(level+1)*DIFF_INDENT}Found list-like. Recursing.")

            _helper_listlike_diff(
                self_prop[i],
                other_prop[i],
                level+2,
                verbose
            )

        elif isinstance(self_prop[i], dict):

            if not isinstance(other_prop[i], dict):
                warn_print((
                    f"{(level+1)*DIFF_INDENT}Found dict-like on self, but corresponding "
                    "element in other isn't a dict-like."
                ))
                warn_print(
                    f"{(level+2)*DIFF_INDENT}{self_prop[i].keys()} | {type(other_prop[i])}"
                )
                continue

            if verbose:
                LOGGER.info(f"{(level+1)*DIFF_INDENT}Found dict-like. Recursing.")

            _helper_dictlike_diff(
                self_prop[i],
                other_prop[i],
                level+2,
                verbose
            )

        elif self_prop[i] != other_prop[i]:
            if verbose:
                LOGGER.info((
                    f"{(level+1)*DIFF_INDENT}Elements don't match, but they may be located "
                    "elsewhere in the list-like:"
                ))
                LOGGER.info(f"{(level+2)*DIFF_INDENT}{self_prop[i]} | {other_prop[i]}")

            if verbose:
                LOGGER.info(f"{(level+2)*DIFF_INDENT}Trying to find elements in each other's list.")

            if self_prop[i] not in other_prop:

                warn_print(
                    f"{(level+3)*DIFF_INDENT} Can't find {self_prop[i]} in other's list"
                )

            else:

                if verbose:
                    LOGGER.info(f"{(level+3)*DIFF_INDENT} Found {self_prop[i]} in other's list")

            if other_prop[i] not in self_prop:

                    warn_print(
                        f"{(level+3)*DIFF_INDENT} Can't find {other_prop[i]} in self's list"
                    )

            else:

                if verbose:
                    LOGGER.info(f"{(level+3)*DIFF_INDENT} Found {other_prop[i]} in other's list")

        else:

            if verbose:
                LOGGER.info(f"{(level+1)*DIFF_INDENT}Elements match.")
                LOGGER.info(
                    f"{(level+2)*DIFF_INDENT}{self_prop[i]} | {other_prop[i]}"
                )


def _helper_dictlike_diff(self_prop, other_prop, level, verbose):
    import click

    global DIFF_INDENT
    global LOGGER

    if verbose:
        warn_print = LOGGER.warning
    else:
        warn_print = click.echo

    keys = set(self_prop.keys())
    keys.union(set(other_prop.keys()))

    if self_prop != other_prop and not verbose:
        click.echo(f"{level*DIFF_INDENT}In dict-like:")

    for key in keys:

        if verbose:
            LOGGER.info(f"{level*DIFF_INDENT}Comparing key {key}.")

        try:
            self_prop[key]
        except KeyError:
            warn_print(
                f"{DIFF_INDENT*(level+1)} self's dict-like is missing key found in other's: {key}"
            )
            break

        try:
            other_prop[key]
        except KeyError:
            warn_print(
                f"{DIFF_INDENT*(level+1)} other's dict-like is missing key found in self's: {key}"
            )
            break
        if verbose:
            LOGGER.info(f"{(level+1)*DIFF_INDENT}Self type: {type(self_prop[key])}.")
            LOGGER.info(f"{(level+1)*DIFF_INDENT}Other type: {type(other_prop[key])}.")

        if isinstance(self_prop[key], dict) and len(self_prop[key].keys()) == 0:
            # treat as a None
            if other_prop[key] == None or other_prop[key] == {}:
                if verbose:
                    LOGGER.info(f"{(level+1)*DIFF_INDENT}Values match.")
                    LOGGER.info(
                        f"{(level+2)*DIFF_INDENT}{self_prop[key]} | {other_prop[key]}"
                    )
            else:
                warn_print(f"{(level+1)*DIFF_INDENT}Values don't match:")
                warn_print(
                    f"{(level+2)*DIFF_INDENT}{self_prop[key]} | {other_prop[key]}"
                )
            continue

        elif isinstance(other_prop[key], dict) and \
            len(other_prop[key].keys()) == 0:

            # treat as a None
            if self_prop[key] == None or self_prop[key] == {}:
                if verbose:
                    LOGGER.info(f"{(level+1)*DIFF_INDENT}Values match.")
                    LOGGER.info(
                        f"{(level+2)*DIFF_INDENT}{self_prop[key]} | {other_prop[key]}"
                    )
            else:
                warn_print(f"{(level+1)*DIFF_INDENT}Values don't match:")
                warn_print(
                    f"{(level+2)*DIFF_INDENT}{self_prop[key]} | {other_prop[key]}"
                )
            continue

        if isinstance(self_prop[key], BPProperty):

            if not isinstance(other_prop[key], BPProperty):
                warn_print((
                    f"{(level+1)*DIFF_INDENT}Found BPProperty on self, but corresponding "
                    "element in other isn't a BPProperty."
                ))
                warn_print(f"{(level+2)*DIFF_INDENT}{type(other_prop[key])}")
                continue

            if verbose:
                LOGGER.info(
                    f"{(level+1)*DIFF_INDENT}Found {self_prop[key]}. Comparing to "
                    f"{other_prop[key]}."
                )

            self_prop[key].diff(
                other_prop[key],
                level=level+2,
                verbose=verbose
            )

        elif isinstance(self_prop[key], BPNode):

            if not isinstance(other_prop[key], BPNode):

                warn_print((
                    f"{(level+1)*DIFF_INDENT}Found BPNode on self, but corresponding element in"
                    " other isn't a BPNode."
                ))
                warn_print(f"{(level+2)*DIFF_INDENT}{type(other_prop[key])}")
                continue

            if verbose:
                LOGGER.info(
                    f"{(level+1)*DIFF_INDENT}Found BPNode {self_prop[key]}. Comparing to "
                    f"{other_prop[key]}."
                )

            self_prop[key].diff(
                other_prop[key],
                level=level+2,
                verbose=verbose
            )

        elif isinstance(self_prop[key], list) or isinstance(self_prop[key], set):

            if not isinstance(other_prop[key], list) and not isinstance(other_prop[key], set):
                warn_print((
                    f"{(level+1)*DIFF_INDENT}Found list-like on self, but corresponding element"
                    " in other isn't a list-like."
                ))
                warn_print(f"{(level+2)*DIFF_INDENT}{type(other_prop[key])}")
                continue

            if verbose:
                LOGGER.info(f"{(level+1)*DIFF_INDENT}Found list-like. Recursing.")

            _helper_listlike_diff(
                self_prop[key],
                other_prop[key],
                level+2,
                verbose
            )

        elif isinstance(self_prop[key], dict):

            if not isinstance(other_prop[key], dict):
                warn_print((
                    f"{(level+1)*DIFF_INDENT}Found dict-like on self, but corresponding element"
                    " in other isn't a dict-like."
                ))
                warn_print(
                    f"{(level+2)*DIFF_INDENT}{self_prop[key].keys()} | {type(other_prop[key])}"
                )
                continue

            if verbose:
                LOGGER.info(f"{(level+1)*DIFF_INDENT}Found dict-like. Recursing.")

            _helper_dictlike_diff(
                self_prop[key],
                other_prop[key],
                level+2,
                verbose
            )

        elif self_prop[key] != other_prop[key]:

            warn_print(f"{(level+1)*DIFF_INDENT}Elements don't match:")
            warn_print(f"{(level+2)*DIFF_INDENT}{self_prop[key]} | {other_prop[key]}")

        else:
            if verbose:
                LOGGER.info(f"{(level+1)*DIFF_INDENT}Elements match.")
                LOGGER.info(
                    f"{(level+2)*DIFF_INDENT}{self_prop[key]} | {other_prop[key]}"
                )


class BPProperty(object):
    """Represents a property value of a BETTER Project document.

    Args:
        name (str): the name as it will appear in the JSON blob when serializing the document.
        required (bool): whether this property is a required value of its parent object, according
            to the schema. Defaults to False.
        value (*): the value of this property, the type of which should match the type specified in
            the ``type_`` argument.
        type_ (*): the Python type of the ``value``, ``default``, and elements of the
            ``choices`` arguments.
        default (*): the default value of this property, if it has one, the value of which should
            match the type specified in the ``type_`` argument.
        choices (list): if specified, a list of values constraining allowed values for the
            ``value`` and ``default`` arguments, the values of which should match the type specified
            in the ``type_`` argument.
        no_validation (bool): If True, skip validation when asked to. Defaults to False.
    """

    def __init__(
        self,
        name,
        required=False,
        value=None,
        type_=type(None),
        default=None,
        choices=None,
        no_validation=False
    ):

        assert type(name) == str, "Property names must be strings."
        assert type(choices) == list or type(choices) == type(None), \
            "choices must be a list or None."

        self.name = name
        self.required = required
        self.type_ = type_
        self.default = default
        self.choices = choices
        self.no_validation = no_validation

        # structural validation, doesn't belong in is_valid()
        if self.choices:
            for element in self.choices:
                assert type(element) == self.type_, (
                    f"Element {element} in choices is not of type {self.type_}!"
                )

        if value is not None:
            self.value = value
        else:
            self.value = self.default


    def __eq__(self, other):

        if not isinstance(other, self.__class__):
            return False

        __eq = lambda prop : getattr(self, prop) == getattr(other, prop)
        """'Normal' comparison."""

        def __leq(prop):
            """Compare lists in an unordered manner. Can't use sets with counters for some reason;
            two equivalent BPDocumentSpan objects (x == y is True) are distinct in a set, and I
            don't know why.
            """
            if not len(getattr(self, prop)) == len(getattr(other, prop)):
                return False
            found = []
            for element in getattr(self, prop):
                if element in getattr(other, prop):
                    found.append(element)
            return found == getattr(self, prop)

        def __deq(prop):
            """Sorts list-like elements of a dictionary property."""
            temp_dict_self = {}
            for key in getattr(self, prop):
                if type(getattr(self, prop)[key]) == list:
                    temp_dict_self[key] = sorted(getattr(self, prop)[key])
                else:
                    temp_dict_self[key] = getattr(self, prop)[key]
            temp_dict_other = {}
            for key in getattr(other, prop):
                if type(getattr(other, prop)[key]) == list:
                    temp_dict_other[key] = sorted(getattr(other, prop)[key])
                else:
                    temp_dict_other[key] = getattr(other, prop)[key]
            return temp_dict_self == temp_dict_other

        bools = []
        for prop in vars(self):

            if isinstance(getattr(self, prop), list) and isinstance(getattr(other, prop), list):
                bools.append(__leq(prop))

            elif isinstance(getattr(self, prop), dict) and len(getattr(self, prop).keys()) == 0 \
                and isinstance(getattr(other, prop), dict) and \
                len(getattr(other, prop).keys()) == 0:

                bools.append(True)

            elif isinstance(getattr(self, prop), dict) and len(getattr(self, prop).keys()) == 0:
                bools.append(None == getattr(other, prop))

            elif isinstance(getattr(other, prop), dict) and len(getattr(other, prop).keys()) == 0:
                bools.append(None == getattr(self, prop))

            elif isinstance(getattr(self, prop), dict) and isinstance(getattr(other, prop), dict):
                bools.append(__deq(prop))

            elif isinstance(getattr(self, prop), BPDocumentProvenance):
                continue
            else:
                bools.append(__eq(prop))

        return all(bools)


    def __lt__(self, other):
        return (self.name, self.value) < (other.name, other.value)


    def __ne__(self, other):
        return not self == other


    def __repr__(self):
        return (
            f"BPProperty({self.name}, required={self.required}, "
            f"value={self.value}, type_={self.type_}, default={self.default}, "
            f"choices={self.choices}, no_validation={self.no_validation})"
        )


    def __str__(self):
        return f"BPProperty<{self.name}>"


    def diff(self, other, level=0, verbose=False):
        """Compares this object with another of the same type.

        Comparisons are rendered as printable statements.

        Args:
            other (BPProperty): another object of the same type as this one.
            level (int): the indentation level to print for output messages.
            verbose (bool): show complete output, instead of just the differences.
        """
        import click

        global DIFF_INDENT
        global LOGGER

        if verbose:
            warn_print = LOGGER.warning
        else:
            warn_print = click.echo

        if not isinstance(other, self.__class__):
            raise ValueError(f"Can't compare {self.__class__} to {type(other)}.")

        if self != other and not verbose:
            click.echo(f"{level*DIFF_INDENT}In {self}:")

        for prop in vars(self):

            if verbose:
                LOGGER.info(f"{level*DIFF_INDENT}Comparing property {prop} in {self.name}.")

            if prop == "value":

                if isinstance(getattr(self, prop), dict) and len(getattr(self, prop).keys()) == 0:
                    # treat as a None
                    if getattr(other, prop) == None or getattr(other, prop) == {}:
                        if verbose:
                            LOGGER.info(f"{(level+1)*DIFF_INDENT}Values match.")
                            LOGGER.info(
                                f"{(level+2)*DIFF_INDENT}{getattr(self, prop)} | "
                                f"{getattr(other, prop)}"
                            )
                    else:
                        warn_print(f"{(level+1)*DIFF_INDENT}Values don't match:")
                        warn_print(
                            f"{(level+2)*DIFF_INDENT}{getattr(self, prop)} | "
                            f"{getattr(other, prop)}"
                        )
                    continue

                elif isinstance(getattr(other, prop), dict) and \
                    len(getattr(other, prop).keys()) == 0:

                    # treat as a None
                    if getattr(self, prop) == None or getattr(self, prop) == {}:
                        if verbose:
                            LOGGER.info(f"{(level+1)*DIFF_INDENT}Values match.")
                            LOGGER.info(
                                f"{(level+2)*DIFF_INDENT}{getattr(self, prop)} | "
                                f"{getattr(other, prop)}"
                            )
                    else:
                        warn_print(f"{(level+1)*DIFF_INDENT}Values don't match:")
                        warn_print(
                            f"{(level+2)*DIFF_INDENT}{getattr(self, prop)} | "
                            f"{getattr(other, prop)}"
                        )
                    continue

                if isinstance(getattr(self, prop), BPNode):

                    if not isinstance(getattr(other, prop), BPNode):

                        warn_print((
                            f"{(level+1)*DIFF_INDENT}Found BPNode on self, but corresponding "
                            "element in other isn't a BPNode."
                        ))
                        warn_print(f"{(level+2)*DIFF_INDENT}{type(getattr(other, prop))}")
                        continue

                    if verbose:
                        LOGGER.info(
                            f"{(level+1)*DIFF_INDENT}Found BPNode {getattr(self, prop)}. Comparing "
                            f"to {getattr(other, prop)}."
                        )
                    getattr(self, prop).diff(
                        getattr(other, prop),
                        level=level+2,
                        verbose=verbose
                    )

                elif isinstance(getattr(self, prop), BPProperty):

                    if not isinstance(getattr(other, prop), BPProperty):
                        warn_print((
                            f"{(level+1)*DIFF_INDENT}Found BPProperty on self, but "
                            "corresponding element in other isn't a BPProperty."
                        ))
                        warn_print(f"{(level+2)*DIFF_INDENT}{type(getattr(other, prop))}")
                        continue

                    if verbose:
                        LOGGER.info(
                            f"{(level+1)*DIFF_INDENT}Found {getattr(self, prop)}. Comparing "
                            f"to {getattr(other, prop)}."
                        )
                    getattr(self, prop).diff(
                        getattr(other, prop),
                        level=level+2,
                        verbose=verbose
                    )

                elif isinstance(getattr(self, prop), list) or isinstance(getattr(self, prop), set):

                    if not isinstance(getattr(other, prop), list) and \
                        not isinstance(getattr(other, prop), set):

                        warn_print((
                            f"{(level+1)*DIFF_INDENT}Found list-like on self, but corresponding"
                            " element in other isn't a list-like."
                        ))
                        warn_print(f"{(level+2)*DIFF_INDENT}{type(getattr(other, prop))}")
                        continue

                    if verbose:
                        LOGGER.info(f"{(level+1)*DIFF_INDENT}Found list-like. Comparing.")
                    _helper_listlike_diff(
                        getattr(self, prop),
                        getattr(other, prop),
                        level+2,
                        verbose
                    )

                elif isinstance(getattr(self, prop), dict):

                    if not isinstance(getattr(other, prop), dict):
                        warn_print((
                            f"{(level+1)*DIFF_INDENT}Found dict-like on self, but corresponding"
                            " element in other isn't a dict-like."
                        ))
                        warn_print((
                            f"{(level+2)*DIFF_INDENT}{getattr(self, prop).keys()} | "
                            f"{type(getattr(other, prop))}"
                        ))
                        continue

                    if verbose:
                        LOGGER.info(f"{(level+1)*DIFF_INDENT}Found dict-like. Comparing.")
                    _helper_dictlike_diff(
                        getattr(self, prop),
                        getattr(other, prop),
                        level+2,
                        verbose
                    )

                elif getattr(self, prop) != getattr(other, prop):

                    warn_print(f"{(level+1)*DIFF_INDENT}Values don't match:")
                    warn_print(
                        f"{(level+2)*DIFF_INDENT}{getattr(self, prop)} | {getattr(other, prop)}"
                    )

                else:
                    if verbose:
                        LOGGER.info(f"{(level+1)*DIFF_INDENT}Values match.")
                        LOGGER.info(
                            f"{(level+2)*DIFF_INDENT}{getattr(self, prop)} | {getattr(other, prop)}"
                        )

            elif getattr(self, prop) != getattr(other, prop):

                warn_print(f"{(level+1)*DIFF_INDENT}Values don't match:")
                warn_print(
                    f"{(level+2)*DIFF_INDENT}{getattr(self, prop)} | {getattr(other, prop)}"
                )

            else:
                if verbose:
                    LOGGER.info(f"{(level+1)*DIFF_INDENT}Values match.")
                    LOGGER.info(
                        f"{(level+2)*DIFF_INDENT}{getattr(self, prop)} | {getattr(other, prop)}"
                    )


    def is_valid(self):
        """Checks if this object is valid."""

        global DIFF_INDENT
        global LOGGER

        if self.no_validation:
            raise RuntimeError(f"Validation set to False: {self}")

        if not self.required and self.value is None:
            return True

        if self.required and self.value is None:
            LOGGER.error(f"BPProperty is required but no value is set: {self}")
            return False

        if self.type_ != type(self.value):
            LOGGER.error(f"BPProperty value does not match specified type_: {self.type_}")
            return False

        if self.choices and self.value not in self.choices:
            LOGGER.error((
                f"BPProperty value disallowed by choices: {self} -> "
                f"choices: {self.choices}"
            ))
            return False

        return True


class BPNode(ABC):
    """Abstract Base Class for elements of a BP document tree."""

    def __eq__(self, other):

        import traceback

        if not isinstance(other, self.__class__):
            return False

        __eq = lambda prop : getattr(self, prop) == getattr(other, prop)
        """'Normal' comparison."""
        def __leq(prop):
            """Compare lists in an unordered manner. Can't use sets with counters for some reason;
            two equivalent BPDocumentSpan objects (x == y is True) are distinct in a set, and I
            don't know why.
            """
            if not len(getattr(self, prop)) == len(getattr(other, prop)):
                return False
            found = []
            for element in getattr(self, prop):
                if element in getattr(other, prop):
                    found.append(element)
            return found == getattr(self, prop)

        def __deq(prop):
            temp_dict_self = {}
            for key in getattr(self, prop):
                if type(getattr(self, prop)[key]) == list:
                    temp_dict_self[key] = sorted(getattr(self, prop)[key])
                else:
                    temp_dict_self[key] = getattr(self, prop)[key]
            temp_dict_other = {}
            for key in getattr(other, prop):
                if type(getattr(other, prop)[key]) == list:
                    temp_dict_other[key] = sorted(getattr(other, prop)[key])
                else:
                    temp_dict_other[key] = getattr(other, prop)[key]
            return temp_dict_self == temp_dict_other

        bools = []
        for prop in vars(self):
            if isinstance(getattr(self, prop), list):
                bools.append(__leq(prop))

            elif isinstance(getattr(self, prop), dict) and len(getattr(self, prop).keys()) == 0 \
                and isinstance(getattr(other, prop), dict) and \
                len(getattr(other, prop).keys()) == 0:

                bools.append(True)

            elif isinstance(getattr(self, prop), dict) and len(getattr(self, prop).keys()) == 0:
                bools.append(None == getattr(other, prop))

            elif isinstance(getattr(other, prop), dict) and len(getattr(other, prop).keys()) == 0:
                bools.append(None == getattr(self, prop))

            elif isinstance(getattr(self, prop), dict) and isinstance(getattr(other, prop), dict):
                bools.append(__deq(prop))

            else:
                if isinstance(getattr(self, prop), BPProperty):
                    if isinstance(getattr(self, prop).value, BPDocumentProvenance):
                        continue
                    else:
                        bools.append(__eq(prop))

                else:
                    bools.append(__eq(prop))

        return all(bools)


    def __lt__(self, other):
        raise NotImplementedError


    def __ne__(self, other):
        return not self == other


    def __repr__(self):
        raise NotImplementedError


    def __str__(self):
        raise NotImplementedError


    def deidentify(self):
        return NotImplementedError


    def diff(self, other, level=0, verbose=False):
        """Compares this object with another of the same type.

        Comparisons are rendered as printable statements.

        Args:
            other (BPNode): another object of the same type as this one.
            level (int): the indentation level to print for output messages.
            verbose (bool): show complete output, instead of just the differences.
        """
        import click

        global DIFF_INDENT
        global LOGGER

        if verbose:
            warn_print = LOGGER.warning
        else:
            warn_print = click.echo

        if not isinstance(other, self.__class__):
            raise ValueError(f"Can't compare {self.__class__} to {type(other)}.")

        if self != other and not verbose:
            click.echo(f"{level*DIFF_INDENT}In {self}:")

        for prop in vars(self):

            if prop == "provenance":
                continue

            if isinstance(getattr(self, prop), dict) and len(getattr(self, prop).keys()) == 0:
                # treat as a None
                if getattr(other, prop) == None or getattr(other, prop) == {}:
                    if verbose:
                        LOGGER.info(f"{(level+1)*DIFF_INDENT}Values match.")
                        LOGGER.info(
                            f"{(level+2)*DIFF_INDENT}{getattr(self, prop)} | {getattr(other, prop)}"
                        )
                else:
                    warn_print(f"{(level+1)*DIFF_INDENT}Values don't match:")
                    warn_print(
                        f"{(level+2)*DIFF_INDENT}{getattr(self, prop)} | {getattr(other, prop)}"
                    )
                continue

            elif isinstance(getattr(other, prop), dict) and len(getattr(other, prop).keys()) == 0:

                # treat as a None
                if getattr(self, prop) == None or getattr(self, prop) == {}:
                    if verbose:
                        LOGGER.info(f"{(level+1)*DIFF_INDENT}Values match.")
                        LOGGER.info(
                            f"{(level+2)*DIFF_INDENT}{getattr(self, prop)} | {getattr(other, prop)}"
                        )
                else:
                    warn_print(f"{(level+1)*DIFF_INDENT}Values don't match:")
                    warn_print(
                        f"{(level+2)*DIFF_INDENT}{getattr(self, prop)} | {getattr(other, prop)}"
                    )
                continue

            if verbose:
                LOGGER.info(f"{level*DIFF_INDENT}Comparing property {prop}.")

            if isinstance(getattr(self, prop), BPNode):

                if not isinstance(getattr(other, prop), BPNode):

                    warn_print((
                        f"{(level+1)*DIFF_INDENT}Found BPNode on self, but corresponding "
                        "element in other isn't a BPNode."
                    ))
                    warn_print(f"{(level+2)*DIFF_INDENT}{type(getattr(other, prop))}")
                    continue

                if verbose:
                    LOGGER.info(
                        f"{(level+1)*DIFF_INDENT}Found BPNode {getattr(self, prop)}. Comparing to "
                        f"{getattr(other, prop)}."
                    )

                getattr(self, prop).diff(
                    getattr(other, prop),
                    level=level+2,
                    verbose=verbose
                )

            elif isinstance(getattr(self, prop), BPProperty):

                if not isinstance(getattr(self, prop), BPProperty):
                    warn_print((
                        f"{(level+1)*DIFF_INDENT}Found BPProperty on self, but corresponding "
                        "element in other isn't a BPProperty."
                    ))
                    warn_print(f"{(level+2)*DIFF_INDENT}{type(getattr(other, prop))}")
                    continue

                if verbose:
                    LOGGER.info(
                        f"{(level+1)*DIFF_INDENT}Found {getattr(self, prop)}. Comparing to "
                        f"{getattr(other, prop)}."
                    )
                getattr(self, prop).diff(
                    getattr(other, prop),
                    level=level+2,
                    verbose=verbose
                )

            elif isinstance(getattr(self, prop), list) or isinstance(getattr(self, prop), set):

                if not isinstance(getattr(other, prop), list) and \
                    not isinstance(getattr(other, prop), set):

                    warn_print((
                        f"{(level+1)*DIFF_INDENT}Found list-like on self, but corresponding "
                        "element in other isn't a list-like."
                    ))
                    warn_print(f"{(level+2)*DIFF_INDENT}{type(getattr(other, prop))}")
                    continue

                LOGGER.info(f"{(level+1)*DIFF_INDENT}Found list-like. Comparing.")
                _helper_listlike_diff(
                    getattr(self, prop),
                    getattr(other, prop),
                    level+2,
                    verbose
                )

            elif isinstance(getattr(self, prop), dict):

                if not isinstance(getattr(other, prop), dict):
                    warn_print((
                        f"{(level+1)*DIFF_INDENT}Found dict-like on self, but corresponding "
                        "element in other isn't a dict-like."
                    ))
                    warn_print((
                        f"{(level+2)*DIFF_INDENT}{getattr(self, prop).keys()} | "
                        f"{type(getattr(other, prop))}"
                    ))
                    continue

                if verbose:
                    LOGGER.info(f"{(level+1)*DIFF_INDENT}Found dict-like. Comparing.")
                _helper_dictlike_diff(
                    getattr(self, prop),
                    getattr(other, prop),
                    level+2,
                    verbose
                )

            elif getattr(self, prop) != getattr(other, prop):

                warn_print(f"{(level+1)*DIFF_INDENT}Values don't match:")
                warn_print(
                    f"{(level+2)*DIFF_INDENT}{getattr(self, prop)} | {getattr(other, prop)}"
                )

            else:
                if verbose:
                    LOGGER.info(f"{(level+1)*DIFF_INDENT}Values match.")
                    LOGGER.info(
                        f"{(level+2)*DIFF_INDENT}{getattr(self, prop)} | {getattr(other, prop)}"
                    )


    @classmethod
    def from_dict(klass):
        raise NotImplementedError


    def is_valid(self):
        raise NotImplementedError


    def set_validation(self, validation):
        raise NotImplementedError


    def to_dict(self):
        raise NotImplementedError


class BPDocumentSpan(BPNode):
    """Represents a Span object in a BETTER Project document.

    Args:
        string_ (str): the string literal contained within the bounds of the span.
        start (int): the starting index of the span within its parent segment.
        end (int): the ending index of the span within its parent segment.
        hstring (str): the head string literal contained within the bounds of the span, if it has
            one.
        hstart (int): the starting index of the head string within the span, if it has one.
        hend (int): the ending index of the head string within the span, if it has one.
        no_head (bool): set to True if the span is headless, and omit otherwise.
        synclass (str): the synclass this span, one of ``"Name"``, ``"Nominal"``, ``"Pronoun"``, or
            ``"Event-Anchor"``.
        is_anchor (bool): whether this Span represents an Event anchor.
        hinferred (bool): if True, asserts that this Span's Head information was the result of the
            assumed Head, rather than the Head annotation.
        no_validation (bool): If True, skip validation when asked to. Defaults to False.
    """

    def __init__(
        self,
        string_,
        start=None,
        end=None,
        hstring=None,
        hstart=None,
        hend=None,
        no_head=None,
        synclass=None,
        is_anchor=None,
        hinferred=None,
        no_validation=False
    ):

        from lib.utils import MAT_SPAN_SYNCLASS_TYPES

        # required values
        self.string_ = BPProperty(
            "string",
            True,
            value=string_,
            type_=str,
            no_validation=no_validation
        )

        # optionals
        self.start = BPProperty(
            "start",
            False,
            value=start,
            type_=int,
            no_validation=no_validation
        )
        self.end = BPProperty(
            "end",
            False,
            value=end,
            type_=int,
            no_validation=no_validation
        )
        self.hstring = BPProperty(
            "hstring",
            False,
            value=hstring,
            type_=str,
            no_validation=no_validation
        )
        self.hstart = BPProperty(
            "hstart",
            False,
            value=hstart,
            type_=int,
            no_validation=no_validation
        )
        self.hend = BPProperty(
            "hend",
            False,
            value=hend,
            type_=int,
            no_validation=no_validation
        )
        self.no_head = BPProperty(
            "no-head",
            False,
            value=no_head,
            type_=bool,
            no_validation=no_validation
        )
        self.synclass = BPProperty(
            "synclass",
            False,
            value=synclass,
            type_=str,
            choices=MAT_SPAN_SYNCLASS_TYPES,
            no_validation=no_validation
        )
        self.is_anchor = BPProperty(
            "anchor-string",
            False,
            value=is_anchor,
            type_=bool,
            no_validation=no_validation
        )
        self.hinferred = BPProperty(
            "hinferred",
            False,
            value=hinferred,
            type_=bool,
            no_validation=no_validation
        )
        self.no_validation = no_validation


    def __hash__(self):
        return hash(id(self))


    def __len__(self):
        return self.length()


    def __lt__(self, other):
        return (self.string_, self.start, self.end) < (other.string_, other.start, self.end)


    def __repr__(self):
        return (
            f"BPDocumentSpan({repr(self.string_.value)}, "
            f"start={repr(self.start.value)}, "
            f"end={repr(self.end.value)}, "
            f"hstring={repr(self.hstring.value)}, "
            f"hstart={repr(self.hstart.value)}, "
            f"hend={repr(self.hend.value)}, "
            f"no_head={repr(self.no_head.value)}, "
            f"synclass={repr(self.synclass.value)}, "
            f"is_anchor={repr(self.is_anchor.value)}, "
            f"hinferred={repr(self.hinferred.value)}, "
            f"no_validation={repr(self.no_validation)})"
        )


    def __str__(self):
        if self.start.value and self.end.value:
            return f"BPDocumentSpan<[{self.start.value}:{self.end.value}] '{self.string_.value}'>"
        else:
            return f"BPDocumentSpan<({id(self)}) '{self.string_.value}'>"


    @classmethod
    def from_dict(klass, dict_, no_validation=False):
        """Constructs an instance of :class:`BPDocumentSpan` from a Python dictionary following the
        BP schema for Span objects.

        Args:
            dict_ (dict): a Python dictionary following the BP schema for Span objects.
            no_validation (bool): If True, construct the object with validation turned off.
                Defaults to False.

        Returns:
            BPDocumentSpan: a new instance of this class built with the dictionary.

        .. Strategy:
           - primitives can be directly looked up and passed in from ``dict_``
           - collections (dicts, lists) and non-primitive types (like other ``BPDocument<type>``
             types defined in this library) must be constructed first if data is present in
             ``dict_``.
        """

        global LOGGER

        assert 'string' in dict_, "Span object dictionary is missing the 'string' key."

        # use the dummy to ensure that the name for the dict fields can be found in _one_ place
        dummy = klass("dummy")

        ## build new object

        obj = klass(
            dict_.get(dummy.string_.name),
            start=dict_.get(dummy.start.name),
            end=dict_.get(dummy.end.name),
            hstring=dict_.get(dummy.hstring.name),
            hstart=dict_.get(dummy.hstart.name),
            hend=dict_.get(dummy.hend.name),
            no_head=dict_.get(dummy.no_head.name),
            synclass=dict_.get(dummy.synclass.name),
            is_anchor=dict_.get(dummy.is_anchor.name),
            hinferred=dict_.get(dummy.hinferred.name),
            no_validation=no_validation
        )

        if not no_validation:
            assert obj.is_valid(), "Span object built from dictionary failed validation."

        return obj


    def hlength(self):
        """Gets the length of the head within this span."""
        if self.hstart.value and self.hend.value:
            return self.hend.value - self.hstart.value
        elif self.hstring.value:
            return len(self.hstring.value)
        else:
            return None


    def is_valid(self):
        """Checks if this object (and its constituents) are all valid."""

        global LOGGER

        if self.no_validation:
            raise RuntimeError(f"Validation set to False: {self}")

        for prop in [
            self.string_,
            self.start,
            self.end,
            self.hstring,
            self.hstart,
            self.hend,
            self.no_head,
            self.synclass,
            self.is_anchor,
            self.hinferred,
        ]:
            if not prop.is_valid():
                LOGGER.error(f"BPDocumentSpan property {prop} is invalid.")
                return False

        if self.no_head.value and (
            self.hstring.value or self.hstart.value is not None or self.hend.value is not None
        ):
            LOGGER.error((
                f"BPDocumentSpan no_head value is {self.no_head.value} and head values were"
                f"specified: {self.hstring} | {self.hstart} | {self.hend}"
            ))
            return False

        # # Disabling need-all-3 head info values for validation; `hstring` is sufficient
        # h_counter = 0
        # for val in [self.hstring.value, self.hstart.value, self.hend.value,]:
        #     if val is not None:
        #         h_counter += 1
        # if h_counter not in [0, 3,]:
        #     LOGGER.error((
        #         "BPDocumentSpan head values only partially provided, but either all 3 or none of "
        #         f"them should be: {self.hstring} | {self.hstart} | {self.hend}"
        #     ))
        #     return False

        if self.hstring.value and self.string_.value.find(self.hstring.value) == -1:
            LOGGER.error((
                f"BPDocumentSpan hstring '{self.hstring.value}' [{self.start.value}:"
                f"{self.end.value}] can't be found within string '{self.string_.value}' "
                f"[{self.start.value}:{self.end.value}]."
            ))
            return False

        return True


    def length(self):
        """Gets the length of this span."""
        if self.start.value and self.end.value:
            return self.end.value - self.start.value
        elif self.string_.value:
            return len(self.string_.value)
        else:
            return None


    def overlaps(self, other):
        """Checks if this Span overlaps another."""
        assert type(other) == BPDocumentSpan, \
            f"Other object isn't a Span: {type(other)}"
        return self.start.value <= other.end.value and other.start.value <= self.end.value


    def set_validation(self, validation):
        """Sets the validation mode for this node.

        Args:
            validation (bool): if True, sets ``no_validation`` for this node to ``False``.
                Otherwise, sets ``no_validation`` to ``True``.
        """

        assert type(validation) == bool

        self.no_validation = not validation

        self.string_.no_validation = self.no_validation
        self.start.no_validation = self.no_validation
        self.end.no_validation = self.no_validation
        self.hstring.no_validation = self.no_validation
        self.hstart.no_validation = self.no_validation
        self.hend.no_validation = self.no_validation
        self.no_head.no_validation = self.no_validation
        self.synclass.no_validation = self.no_validation
        self.is_anchor.no_validation = self.no_validation
        self.hinferred.no_validation = self.no_validation


    def to_dict(self):
        """Constructs a Python dictionary following the BP schema, suitable for JSON serialization.

        Returns:
            dict: a dictionary representing this BP Span.
        """

        if not self.no_validation:
            assert self.is_valid(), "Span object is invalid."

        dict_ = {}
        for prop in [
            self.string_,
            self.start,
            self.end,
            self.hstring,
            self.hstart,
            self.hend,
            self.no_head,
            self.synclass,
            self.is_anchor,
            self.hinferred,
        ]:
            if prop.value is not None:
                dict_[prop.name] = prop.value

        return dict_


    def trim(self):
        """Removes leading and trailing whitespace and punctuation symbols from ``string_`` and
        ``hstring``.

        Offsets are amended as needed.
        """
        from copy import deepcopy

        from lib.utils import PUNCTUATION
        from lib.utils import LTR_CURRENCY
        from lib.utils import RTL_CURRENCY

        not_strippable = LTR_CURRENCY + RTL_CURRENCY
        not_strippable_both = "%"
        strippable_both = "".join(
            [c for c in PUNCTUATION+" \t\n\r" if c not in not_strippable+not_strippable_both]
        )
        strippable_leading = "%"
        strippable_trailing = ""

        def _trim(var, start, end, h=False, mode="both"):

            if var.value is not None:

                s = deepcopy(var.value)
                if mode == "both":
                    s = s.strip(strippable_both)
                elif mode == "leading":
                    s = s.lstrip(strippable_leading)
                elif mode == "trailing":
                    s = s.rstrip(strippable_trailing)
                else:
                    raise ValueError("'mode' must be one of ['both', 'leading', 'trailing',].")

                assert var.value.find(s) > -1, (
                    f"[Both] Trimmed {'head for ' if h else ' '}span doesn't exist in original "
                    f"span!: '{var.value}' -> '{s}'"
                )

                if start.value is not None:
                    new_start = start.value + var.value.find(s)
                    start.value = new_start

                if end.value is not None:
                    new_end = new_start + len(s)
                    end.value = new_end

                var.value = s

        _trim(self.string_, self.start, self.end)
        _trim(self.hstring, self.hstart, self.hend, h=True)

        _trim(self.string_, self.start, self.end, mode="leading")
        _trim(self.hstring, self.hstart, self.hend, h=True, mode="leading")

        _trim(self.string_, self.start, self.end, mode="trailing")
        _trim(self.hstring, self.hstart, self.hend, h=True, mode="trailing")


class BPDocumentSpanSet(BPNode):
    """Represents a SpanSet object in a BETTER Project document.

    Args:
        ssid (str): a unique SpanSet ID for the collection of spans.
        spans (list): a list of :class:`BPDocumentSpan` objects to attribute to this SpanSet.
        no_validation (bool): If True, skip validation when asked to. Defaults to False.
    """

    def __init__(self, ssid, spans, irrealis=None, no_validation=False):

        # required
        self.ssid = BPProperty("ssid", True, value=ssid, type_=str, no_validation=no_validation)
        self.spans = BPProperty("spans", True, value=spans, type_=list, no_validation=no_validation)
        self.no_validation = no_validation


    def __len__(self):
        return len(self.spans)


    def __lt__(self, other):
        return (self.ssid, self.spans) < (other.ssid, other.spans)


    def __repr__(self):
        return f"BPDocumentSpanSet({self.ssid.value}, {repr(self.spans)}, {self.no_validation})"


    def __str__(self):
        return f"BPDocumentSpanSet<{self.ssid.value} ({id(self)})>"


    @classmethod
    def from_dict(klass, dict_, no_validation=False):
        """Constructs an instance of BPDocumentSpanSet from a Python dictionary following the BP
        schema for SpanSet objects.

        Args:
            dict_ (dict): a Python dictionary following the BP schema for SpanSet objects.
            no_validation (bool): If True, construct the object with validation turned off.
                Defaults to False.

        Returns:
            BPDocumentSpanSet: a new instance of this class built with the dictionary.

        .. Strategy:
           - primitives can be directly looked up and passed in from ``dict_``
           - collections (dicts, lists) and non-primitive types (like other ``BPDocument<type>``
             types defined in this library) must be constructed first if data is present in
             ``dict_``.
        """

        global LOGGER

        required = ['ssid', 'spans',]
        for key in required:
            assert key in dict_, f"SpanSet object dictionary is missing the '{key}' key."

        ## build non-primitive types and collections

        spans = []
        for span in dict_['spans']:
            spans.append(BPDocumentSpan.from_dict(span, no_validation))

        ## build new object

        obj = klass(
            dict_['ssid'],
            spans,
            no_validation=no_validation
        )

        if not no_validation:
            assert obj.is_valid(), "SpanSet object built from dictionary failed validation."

        return obj


    def is_valid(self):
        """Checks if this object (and its constituents) are all valid."""

        global LOGGER

        if self.no_validation:
            raise RuntimeError(f"Validation set to False: {self}")

        for prop in [self.ssid, self.spans,]:
            if not prop.is_valid():
                LOGGER.error(f"BPDocumentSpanSet property {prop} is invalid.")
                return False

        for element in self.spans.value:
            if type(element) is not BPDocumentSpan:
                i = self.spans.value.index(element)
                LOGGER.error((
                    f"BPDocumentSpanSet element at index {i} is not of type BPDocumentSpan: "
                    f"{element}"
                ))
                return False

        return True


    def set_validation(self, validation):
        """Sets the validation mode for this node and all children of this node.

        Args:
            validation (bool): if True, sets ``no_validation`` for this node and all of its children
                to ``False``. Otherwise, all nodes set ``no_validation`` to ``True``.
        """

        assert type(validation) == bool

        self.no_validation = not validation

        self.ssid.no_validation = self.no_validation
        self.spans.no_validation = self.no_validation

        ## recurse

        if self.spans.value:
            for span in self.spans.value:
                span.set_validation(validation)


    def to_dict(self):
        """Constructs a Python dictionary following the BP schema, suitable for JSON serialization.

        Returns:
            dict: a dictionary representing this BP SpanSet.
        """

        if not self.no_validation:
            assert self.is_valid(), "SpanSet object is invalid."

        dict_ = {
            self.ssid.name: self.ssid.value,
            self.spans.name: [span.to_dict() for span in self.spans.value],
        }

        return dict_


class BPDocumentSegmentSection(BPNode):
    """Represents a Segment Section object in a BETTER Project document.

    Args:
        start (int): the starting index of the segment section within its parent segment.
        end (int): the ending index of the segment section within its parent segment.
        structural_element (str): the structural element denoting this segment section, one of
            ``BASIC_MAT_STRUCTURAL_TYPES``.
        no_validation (bool): If True, skip validation when asked to. Defaults to False.
    """

    def __init__(self, start, end, structural_element, no_validation=False):

        from lib.utils import BASIC_MAT_STRUCTURAL_TYPES

        # required
        self.start = BPProperty("start", True, value=start, type_=int, no_validation=no_validation)
        self.end = BPProperty("end", True, value=end, type_=int, no_validation=no_validation)
        self.structural_element = BPProperty(
            "structural-element",
            True,
            value=structural_element,
            type_=str,
            choices=BASIC_MAT_STRUCTURAL_TYPES,
            no_validation=no_validation
        )
        self.no_validation = no_validation


    def __lt__(self, other):
        return (self.structural_element, self.start, self.end) < \
            (other.structural_element, other.start, other.end)


    def __repr__(self):
        return (
            f"BPDocumentSegmentSection({self.start.value}, {self.end.value}, "
            f"'{self.structural_element.value}', no_validation={self.no_validation})"
        )


    def __str__(self):
        return f"BPDocumentSegmentSection<{self.structural_element.value} ({id(self)})>"


    @classmethod
    def from_dict(klass, dict_, no_validation=False):
        """Constructs an instance of BPDocumentSegmentSection from a Python dictionary following the
        BP schema for Segment Section objects.

        Args:
            dict_ (dict): a Python dictionary following the BP schema for Segment Section objects.
            no_validation (bool): If True, construct the object with validation turned off.
                Defaults to False.

        Returns:
            BPDocumentSegmentSection: a new instance of this class built with the dictionary.

        .. Strategy:
           - primitives can be directly looked up and passed in from ``dict_``
           - collections (dicts, lists) and non-primitive types (like other ``BPDocument<type>``
             types defined in this library) must be constructed first if data is present in
             ``dict_``.
        """

        global LOGGER

        required = ["start", "end", "structural-element",]
        for key in required:
            assert key in dict_, f"Segment Section object dictionary is missing the '{key}' key."

        ## build new object

        obj = klass(
            dict_['start'],
            dict_['end'],
            dict_['structural-element'],
            no_validation=no_validation,
        )

        if not no_validation:
            assert obj.is_valid(), "Segment Section object built from dictionary failed validation."

        return obj


    def is_valid(self):
        """Checks if this object (and its constituents) are all valid."""

        global LOGGER

        if self.no_validation:
            raise RuntimeError(f"Validation set to False: {self}")

        for prop in [self.start, self.end, self.structural_element,]:
            if not prop.is_valid():
                LOGGER.error(f"BPDocumentSegmentSection property {prop} is invalid.")
                return False

        return True


    def overlaps(self, other):
        """Checks if this SegmentSection overlaps another."""
        assert type(other) == BPDocumentSegmentSection, \
            f"Other object isn't a SegmentSection: {type(other)}"
        return self.start.value <= other.end.value and other.start.value <= self.end.value


    def set_validation(self, validation):
        """Sets the validation mode for this node.

        Args:
            validation (bool): if True, sets ``no_validation`` for this node to ``False``.
                Otherwise, sets ``no_validation`` to ``True``.
        """

        assert type(validation) == bool

        self.no_validation = not validation

        self.start.no_validation = self.no_validation
        self.end.no_validation = self.no_validation
        self.structural_element.no_validation = self.no_validation


    def to_dict(self):
        """Constructs a Python dictionary following the BP schema, suitable for JSON serialization.

        Returns:
            dict: a dictionary representing this BP Segment Section.
        """

        if not self.no_validation:
            assert self.is_valid(), "Segment Section object is invalid."

        dict_ = {}
        for prop in [self.start, self.end, self.structural_element,]:
            dict_[prop.name] = prop.value

        return dict_


class BPDocumentProvenance(BPNode):
    """Represents a Provenance object in a BETTER Project document.

    Provenance objects keep track of book-keeping information, usually about how the data at the
    particular level the Provenance object is attached to was created, or where it came from.
    Multiple ``BPDocument<type>`` object types leverage it.

    Args:
        annotation_date (str): the date the data was annotated, in no particular format.
        annotation_procedure (str): one of ``"pretagger"``, ``"mturk"``, or ``"mat"``.
        annotator_class (str): one of ``lib.utils.BP_ANNOTATOR_CLASS_TYPES``.
        annotator_id (str): a unique ID of the annotator, such as a MITRE SUI or email address.
        annotations_role (str): either ``"reference"`` or ``"system"``.
        corpus_id (str): a unique ID/label for this corpus.
        corpus_creation_date (str): the date the corpus was created, in no particular format.
        source (str): the data source used to build the corpus, if any.
        no_validation (bool): If True, skip validation when asked to. Defaults to False.
    """

    def __init__(
        self,
        annotation_date=None,
        annotation_procedure=None,
        annotator_class=None,
        annotator_id=None,
        annotations_role=None,
        corpus_id=None,
        corpus_creation_date=None,
        source=None,
        no_validation=False
    ):

        from lib.utils import BP_ANNOTATOR_CLASS_TYPES

        self.annotation_date = BPProperty(
            "annotation-date",
            False,
            value=annotation_date,
            type_=str,
            no_validation=no_validation
        )
        self.annotation_procedure = BPProperty(
            "annotation-procedure",
            False,
            value=annotation_procedure,
            type_=str,
            choices=["pretagger", "mturk", "mat", "final-manual-reviewer",],
            no_validation=no_validation
        )
        self.annotator_class = BPProperty(
            "annotator-class",
            False,
            value=annotator_class,
            type_=str,
            choices=BP_ANNOTATOR_CLASS_TYPES,
            no_validation=no_validation
        )
        self.annotator_id = BPProperty(
            "annotator-id",
            False,
            value=annotator_id,
            type_=str,
            no_validation=no_validation
        )
        self.annotations_role = BPProperty(
            "annotations-role",
            False,
            value=annotations_role,
            type_=str,
            choices=["reference", "system",],
            no_validation=no_validation
        )
        self.corpus_id = BPProperty("corpus-id", False, value=corpus_id, type_=str)
        self.corpus_creation_date = BPProperty(
            "corpus-creation-date",
            False,
            value=corpus_creation_date,
            type_=str,
            no_validation=no_validation
        )
        self.source = BPProperty(
            "source",
            False,
            value=source,
            type_=str,
            no_validation=no_validation
        )
        self.no_validation = no_validation

        self.disallowed_key_names = [
            self.annotation_date.name,
            self.annotation_procedure.name,
            self.annotator_class.name,
            self.annotator_id.name,
            self.annotations_role.name,
            self.corpus_id.name,
            self.corpus_creation_date.name,
            self.source.name,
        ]
        self.misc_keys = {}


    def __repr__(self):
        return (
            f"BPDocumentProvenance(annotation_date={repr(self.annotation_date)}, "
            f"annotation_procedure={repr(self.annotation_procedure)}, "
            f"annotator_class={repr(self.annotator_class)}, annotator_id={repr(self.annotator_id)},"
            f" annotations_role={repr(self.annotations_role)}, corpus_id={repr(self.corpus_id)}, "
            f"corpus_creation_date={repr(self.corpus_creation_date)}, source={repr(self.source)}, "
            f"no_validation={self.no_validation})"
        )


    def __str__(self):
        return f"BPDocumentProvenance<{id(self)}>"


    def __update_disallowed_key_names(self):
        self.disallowed_key_names = [
            self.annotation_date.name,
            self.annotation_procedure.name,
            self.annotator_class.name,
            self.annotator_id.name,
            self.annotations_role.name,
            self.corpus_id.name,
            self.corpus_creation_date.name,
            self.source.name,
        ]


    def add_misc_key(self, key, value):
        """Adds a non-standard key-value pair to this Provenance object.

        The key must be a string, and the value must be ``json`` serializable.

        Args:
            key (str): the name of the new JSON object jey.
            value (*): the value to insert at the key. It must be ``json`` serializable, and it may
                not be the name of a standard key.
        """

        if key not in self.disallowed_key_names:
            self.misc_keys[key] = value


    @classmethod
    def from_dict(klass, dict_, no_validation=False):
        """Constructs an instance of BPDocumentProvenance from a Python dictionary following the BP
        schema for Provenance objects.

        Args:
            dict_ (dict): a Python dictionary following the BP schema for Provenance objects.
            no_validation (bool): If True, construct the object with validation turned off.
                Defaults to False.

        Returns:
            BPDocumentProvenance: a new instance of this class built with the dictionary.

        .. Strategy:
           - primitives can be directly looked up and passed in from ``dict_``
           - collections (dicts, lists) and non-primitive types (like other ``BPDocument<type>``
             types defined in this library) must be constructed first if data is present in
             ``dict_``.
        """

        global LOGGER

        ## build new object

        obj = klass(
            dict_.get("annotation-date", None),
            dict_.get("annotation-procedure", None),
            dict_.get("annotator-class", None),
            dict_.get("annotator-id", None),
            dict_.get("annotations-role", None),
            dict_.get("corpus-id", None),
            dict_.get("corpus-creation-date", None),
            dict_.get("source", None),
            no_validation=no_validation
        )

        for key in dict_:
            if key not in obj.disallowed_key_names:
                obj.add_misc_key(key, dict_.get(key, None))

        if not no_validation:
            assert obj.is_valid(), "Provenance object built from dictionary failed validation."

        return obj


    def is_valid(self):
        """Checks if this object (and its constituents) are all valid."""
        import json

        global LOGGER

        if self.no_validation:
            raise RuntimeError(f"Validation set to False: {self}")

        for key in self.misc_keys:

            if key in self.disallowed_key_names:
                LOGGER.error(f"Miscellaneous key {key} is not allowed!")
                return False

            try:
                json.dumps(self.misc_keys[key])
            except:
                LOGGER.error(f"Value at miscellaneous key {key} is not JSON serializable!")
                return False

        for prop in [
            self.annotation_date,
            self.annotation_procedure,
            self.annotator_class,
            self.annotator_id,
            self.annotations_role,
            self.corpus_id,
            self.corpus_creation_date,
            self.source,
        ]:
            if not prop.is_valid():
                LOGGER.error(f"BPDocumentProvenance property {prop} is invalid.")
                return False

        return True


    def set_validation(self, validation):
        """Sets the validation mode for this node.

        Args:
            validation (bool): if True, sets ``no_validation`` for this node to ``False``.
                Otherwise, sets ``no_validation`` to ``True``.
        """

        assert type(validation) == bool

        self.no_validation = not validation

        self.annotation_date.no_validation = self.no_validation
        self.annotation_procedure.no_validation = self.no_validation
        self.annotator_class.no_validation = self.no_validation
        self.annotator_id.no_validation = self.no_validation
        self.annotations_role.no_validation = self.no_validation
        self.corpus_id.no_validation = self.no_validation
        self.corpus_creation_date.no_validation = self.no_validation
        self.source.no_validation = self.no_validation


    def to_dict(self):
        """Constructs a Python dictionary following the BP schema, suitable for JSON serialization.

        Returns:
            dict: a dictionary representing this BP Provenance.
        """

        self.__update_disallowed_key_names()

        if not self.no_validation:
            assert self.is_valid(), "Provenance object is invalid."

        dict_ = {}
        for prop in [
            self.annotation_date,
            self.annotation_procedure,
            self.annotator_class,
            self.annotator_id,
            self.annotations_role,
            self.corpus_id,
            self.corpus_creation_date,
            self.source,
        ]:
            if prop.value is not None:
                dict_[prop.name] = prop.value

        for key in self.misc_keys:
            dict_[key] = self.misc_keys[key]

        return dict_


class BPDocumentGranularTemplateProperty(BPNode):
    """Represents an element of a Granular Template Property list.

    Args:
        event_id (str): a unique ID an event.
        ssid (str): the ID of the type of object referenced.
        irrealis (str): an Irrealis assertion, one of ``lib.utils.GRANULAR_MAT_IRREALIS_STATUSES``.
        time_attachment (list): a list of SSIDs, each referring to a ``Time-Mention`` or
            ``Duration-Mention``, that attributes timing information to this template property.
        no_validation (bool): If True, skip validation when asked to. Defaults to False.
    """

    def __init__(
        self,
        event_id=None,
        ssid=None,
        irrealis=None,
        time_attachments=None,
        no_validation=False
    ):

        from lib.utils import GRANULAR_MAT_IRREALIS_STATUSES

        if not no_validation:
            assert (event_id and not ssid) or (not event_id and ssid), \
                "GranularTemplateProperty takes either ssid or event-id, not both!"

        self.event_id = BPProperty(
            "event-id",
            False,
            value=event_id,
            type_=str,
            no_validation=no_validation
        )
        self.ssid = BPProperty(
            "ssid",
            False,
            value=ssid,
            type_=str,
            no_validation=no_validation
        )
        self.irrealis = BPProperty(
            "irrealis",
            False,
            value=irrealis,
            type_=str,
            choices=GRANULAR_MAT_IRREALIS_STATUSES,
            no_validation=no_validation
        )
        self.time_attachments = BPProperty(
            "time-attachments",
            False,
            value=time_attachments,
            type_=list,
            no_validation=no_validation
        )
        self.no_validation = no_validation


    def __lt__(self, other):
        return (self.event_id, self.ssid) < (other.event_id, other.ssid)


    def __repr__(self):
        return (
            f"BPDocumentGranularTemplateProperty({self.event_id.value}, "
            f"{self.ssid.value}, {self.irrealis.value}, {self.time_attachments.value}, "
            f"{self.no_validation})"
        )


    def __str__(self):
        return (
            f"GranTemplProp<type: {self.get_type()}: (evid: "
            f"{self.event_id.value} | ssid: {self.ssid.value})>"
        )

    @classmethod
    def from_dict(klass, dict_, no_validation=False):

        """Constructs an instance of BPDocumentGranularTemplateProperty from a Python dictionary
        following the BP schema for TemplateProperty objects.

        Args:
            dict_ (dict): a Python dictionary following the BP schema for TemplateProperty objects.
            no_validation (bool): If True, construct the object with validation turned off.
                Defaults to False.

        Returns:
            BPDocumentGranularTemplateProperty: a new instance of this class built with the
            dictionary.

        .. Strategy:
           - primitives can be directly looked up and passed in from ``dict_``
           - collections (dicts, lists) and non-primitive types (like other ``BPDocument<type>``
             types defined in this library) must be constructed first if data is present in
             ``dict_``.
        """

        global LOGGER

        ## build new object

        if not dict_:
            return

        if "ssid" in dict_ and "event-id" in dict_:
            raise KeyError(
                "Incoming dictionary for GranularTemplateProperty has both ssid and event-id!"
            )

        if "ssid" in dict_:
            obj = klass(
                ssid=dict_.get("ssid"),
                irrealis=dict_.get("irrealis"),
                time_attachments=dict_.get("time-attachments"),
                no_validation=no_validation
            )
        elif "event-id" in dict_:
            obj = klass(
                event_id=dict_.get("event-id"),
                irrealis=dict_.get("irrealis"),
                time_attachments=dict_.get("time-attachments"),
                no_validation=no_validation
            )
        else:
            raise KeyError(
                f"No 'ssid' or 'event-id' in dictionary: {dict_.keys()}"
            )

        if not no_validation:
            assert obj.is_valid(), \
                "GranularTemplateProperty object built from dictionary failed validation."

        return obj


    def get_type(self):
        """Gets the "type" of this TemplateProperty.

        Returns:
            str: one of ["spanset", "event", "both", None].
        """

        type_ = None
        if self.event_id.value and self.ssid.value:
            type_ = "both"
        elif self.event_id.value:
            type_ = "event"
        elif self.ssid.value:
            type_ = "spanset"
        return type_


    def is_valid(self):
        """Checks if this object (and its constituents) are all valid."""

        global LOGGER

        if self.no_validation:
            raise RuntimeError(f"Validation set to False: {self}")

        for prop in [self.event_id, self.ssid, self.irrealis, self.time_attachments,]:
            if not prop.is_valid():
                LOGGER.error(f"BPDocumentGranularTemplateProperty property {prop} is invalid.")
                return False

        assert (self.event_id.value and not self.ssid.value) or \
            (not self.event_id.value and self.ssid.value), \
            "GranularTemplateProperty takes either ssid or event-id, not both!"

        return True


    def set_validation(self, validation):
        """Sets the validation mode for this node.

        Args:
            validation (bool): if True, sets ``no_validation`` for this node to ``False``.
                Otherwise, sets ``no_validation`` to ``True``.
        """

        assert type(validation) == bool

        self.no_validation = not validation

        self.event_id.no_validation = self.no_validation
        self.ssid.no_validation = self.no_validation
        self.irrealis.no_validation = self.no_validation
        self.time_attachments.no_validation = self.no_validation


    def to_dict(self):
        """Constructs a Python dictionary following the BP schema, suitable for JSON serialization.

        Returns:
            dict: a dictionary representing this BP Span.
        """

        if not self.no_validation:
            assert self.is_valid(), "GranularTemplateProperty object is invalid."

        dict_ = {}

        if self.event_id.value:
            dict_[self.event_id.name] = self.event_id.value
        elif self.ssid.value:
            dict_[self.ssid.name] = self.ssid.value

        if self.irrealis.value:
            dict_[self.irrealis.name] = self.irrealis.value

        if self.time_attachments.value:
            dict_[self.time_attachments.name] = self.time_attachments.value

        return dict_


class BPDocumentGranularTemplate(BPNode):
    """Represents a Granular Template object in a BETTER Project document.

    Args:
        template_id (str): a unique ID for this event.
        type_ (str): one of the top-level keys in ``lib.utils.GRANULAR_MAT_TEMPLATE_TYPES``.
        anchor (str): an SSID referring to the MAT Annotation used to create this template.
            The SpanSet should only have one span in it.
        provenance (BPDocumentProvenance): a Provenance object pretaining to this event.
        no_validation (bool): If True, skip validation when asked to. Defaults to False.
    """

    def __init__(
        self,
        template_id,
        type_,
        anchor,
        provenance=None,
        no_validation=False
    ):
        from lib.utils import GRANULAR_MAT_TEMPLATE_TYPES

        self.template_id = BPProperty(
            "template-id",
            True,
            value=template_id,
            type_=str,
            no_validation=no_validation
        )
        self.type_ = BPProperty(
            "template-type",
            True,
            value=type_,
            type_=str,
            choices=list(GRANULAR_MAT_TEMPLATE_TYPES),
            no_validation=no_validation
        )
        self.anchor = BPProperty(
            "template-anchor",
            True,
            value=anchor,
            type_=str,
            no_validation=no_validation
        )
        self.provenance = BPProperty(
            "provenance",
            False,
            value=provenance,
            type_=BPDocumentProvenance,
            no_validation=no_validation
        )
        self.no_validation = no_validation

        self.properties = [
            self.template_id,
            self.type_,
            self.anchor,
            self.provenance,
        ]


    def __lt__(self, other):
        return (self.template_id, self.type_) < (other.template_id, other.type_)


    def __repr__(self):
        return (
            f"BPDocumentGranularTemplate({self.template_id.value}, {self.type_}, "
            f"{self.anchor.value}, {repr(self.provenance.value)}, {self.no_validation})"
        )


    def __str__(self):
        return f"GranTypelessTemplate<{self.template_id.value}>"


    def deidentify(self):
        """Obfuscates this BPNode and its children."""

        self.provenance.value = None


    @classmethod
    def from_dict(klass, dict_, no_validation=False):
        """Constructs an instance of :class:`BPDocumentGranularTemplate` from a Python dictionary
        following the BP schema for GranularTempalte objects.

        Args:
            dict_ (dict): a Python dictionary following the BP schema for GranularTemplate objects.
            no_validation (bool): If True, construct the object with validation turned off.
                Defaults to False.

        Returns:
            BPDocumentGranularTemplate: a new instance of this class built with the dictionary. More
            specifically, if the template is one of the recognized types in
            ``lib.utils.GRANULAR_MAT_TEMPLATE_TYPES``, this function will return an object of a type
            within the polymorphic descendants of :class:`BPDocumentGranularTemplate`.
        """

        from lib.utils import dict_get_icase
        from lib.utils import GRANULAR_MAT_TEMPLATE_TYPES

        global LOGGER

        required = ["template-id", "template-type", "template-anchor",]
        for key in required:
            assert key in dict_, f"GranularTemplate object dictionary is missing the '{key}' key."

        ## build non-primitive types and collections

        provenance = None
        if 'provenance' in dict_:
            provenance = BPDocumentProvenance.from_dict(dict_['provenance'], no_validation)

        template_id = dict_.get("template-id")
        anchor = dict_.get("template-anchor")

        type_ = dict_.get("template-type")
        assert type_ in GRANULAR_MAT_TEMPLATE_TYPES, (
            "Key 'template-type' in incoming dictionary is not one of "
            f"{list(GRANULAR_MAT_TEMPLATE_TYPES.keys())}: {type_}"
        )

        ## build new object

        if type_ == "Corruplate":
            obj = BPDocumentGranularCorruptionTemplate(
                template_id,
                anchor,
                charged_with=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "charged-with", default=[])],
                corrupt_event=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "corrupt-event", default=[])],
                judicial_actions=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "judicial-actions", default=[])],
                fine=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "fine", default=[])],
                over_time=dict_.get("over-time"),
                outcome=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "outcome", default=[])],
                outcome_averted=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "outcome-averted", default=[])],
                outcome_occurred=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "outcome-occurred", default=[])],
                outcome_hypothetical=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "outcome-hypothetical", default=[])],
                prison_term=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "prison-term", default=[])],
                where=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "where", default=[])],
                who=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "who", default=[])],
                provenance=provenance
            )
        elif type_ == "Epidemiplate":
            obj = BPDocumentGranularEpidemicTemplate(
                template_id,
                anchor,
                disease=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "disease", default=[])],
                exposed_count=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "exposed-count", default=[])],
                exposed_cumulative=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "exposed-cumulative", default=[])],
                exposed_individuals=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "exposed-individuals", default=[])],
                hospitalized_count=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "hospitalized-count", default=[])],
                hospitalized_cumulative=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "hospitalized-cumulative", default=[])],
                hospitalized_individuals=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "hospitalized-individuals", default=[])],
                infected_count=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "infected-count", default=[])],
                infected_cumulative=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "infected-cumulative", default=[])],
                infected_individuals=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "infected-individuals", default=[])],
                killed_count=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "killed-count", default=[])],
                killed_cumulative=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "killed-cumulative", default=[])],
                killed_individuals=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "killed-individuals", default=[])],
                NPI_Events=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "NPI-Events", default=[])],
                outbreak_event=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "outbreak-event", default=[])],
                tested_count=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "tested-count", default=[])],
                tested_cumulative=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "tested-cumulative", default=[])],
                tested_individuals=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "tested-individuals", default=[])],
                recovered_count=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "recovered-count", default=[])],
                recovered_cumulative=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "recovered-cumulative", default=[])],
                recovered_individuals=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "recovered-individuals", default=[])],
                vaccinated_individuals=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "vaccinated-individuals", default=[])],
                vaccinated_count=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "vaccinated-count", default=[])],
                vaccinated_cumulative=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "vaccinated-cumulative", default=[])],
                when=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "when", default=[])],
                where=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "where", default=[])],
                provenance=provenance
            )
        elif type_ == "Protestplate":
            obj = BPDocumentGranularProtestTemplate(
                template_id,
                anchor,
                arrested=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "arrested", default=[])],
                imprisoned=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "imprisoned", default=[])],
                killed=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "killed", default=[])],
                occupy=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "occupy", default=[])],
                over_time=dict_.get("over-time"),
                outcome=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "outcome", default=[])],
                organizer=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "organizer", default=[])],
                outcome_averted=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "outcome-averted", default=[])],
                outcome_occurred=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "outcome-occurred", default=[])],
                outcome_hypothetical=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "outcome-hypothetical", default=[])],
                protest_against=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "protest-against", default=[])],
                protest_event=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "protest-event", default=[])],
                protest_for=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "protest-for", default=[])],
                when=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "when", default=[])],
                where=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "where", default=[])],
                who=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "who", default=[])],
                wounded=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "wounded", default=[])],
                provenance=provenance
            )
        elif type_ == "Terrorplate":
            obj = BPDocumentGranularTerrorismTemplate(
                template_id,
                anchor,
                blamed_by=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "blamed-by", default=[])],
                claimed_by=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "claimed-by", default=[])],
                completion=dict_.get("completion"),
                coordinated=dict_.get("coordinated"),
                killed=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "killed", default=[])],
                kidnapped=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "kidnapped", default=[])],
                named_perp=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "named-perp", default=[])],
                named_perp_org=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "named-perp-org", default=[])],
                named_organizer=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "named-organizer", default=[])],
                over_time=dict_.get("over-time"),
                outcome=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "outcome", default=[])],
                outcome_averted=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "outcome-averted", default=[])],
                outcome_occurred=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "outcome-occurred", default=[])],
                outcome_hypothetical=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "outcome-hypothetical", default=[])],
                perp_captured=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "perp-captured", default=[])],
                perp_killed=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "perp-killed", default=[])],
                perp_objective=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "perp-objective", default=[])],
                perp_wounded=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "perp-wounded", default=[])],
                target_human=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "target-human", default=[])],
                target_physical=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "target-physical", default=[])],
                terror_event=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "terror-event", default=[])],
                terrorism_type=dict_.get("type"),
                weapon=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "weapon", default=[])],
                when=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "when", default=[])],
                where=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "where", default=[])],
                wounded=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "wounded", default=[])],
                provenance=provenance
            )
        elif type_ == "Disasterplate":
            obj = BPDocumentGranularDisasterTemplate(
                template_id,
                anchor,
                affected_cumulative_count=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "affected-cumulative-count", default=[])],
                announce_disaster_warnings=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "announce-disaster-warnings", default=[])],
                assistance_needed=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "assistance-needed", default=[])],
                assistance_provided=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "assistance-provided", default=[])],
                damage=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "damage", default=[])],
                declare_emergency=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "declare-emergency", default=[])],
                disease_outbreak_events=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "disease-outbreak-events", default=[])],
                human_displacement_events=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "human-displacement-events", default=[])],
                individuals_affected=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "individuals-affected", default=[])],
                injured_count=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "injured-count", default=[])],
                killed_count=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "killed-count", default=[])],
                major_disaster_event=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "major-disaster-event", default=[])],
                missing_count=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "missing-count", default=[])],
                outcome=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "outcome", default=[])],
                over_time=dict_.get("over-time"),
                rescue_events=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "rescue-events", default=[])],
                rescued_count=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "rescued-count", default=[])],
                related_natural_phenomena=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "related-natural-phenomena", default=[])],
                repair=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "repair", default=[])],
                responders=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "responders", default=[])],
                when=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "when", default=[])],
                where=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "where", default=[])],
                provenance=provenance
            )
        elif type_ == "Displacementplate":
            obj = BPDocumentGranularDisplacementTemplate(
                template_id,
                anchor,

                assistance_provided=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "assistance-provided", default=[])],
                assistance_needed=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "assistance-needed", default=[])],
                blocked_migration_count=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "blocked-migration-count", default=[])],
                current_location=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "current-location", default=[])],
                destination=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "destination", default=[])],
                detained_count=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "detained-count", default=[])],
                event_or_soa_at_origin=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "event-or-SoA-at-origin", default=[])],
                group_identity=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "group-identity", default=[])],
                human_displacement_event=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "human-displacement-event", default=[])],
                injured_count=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "injured-count", default=[])],
                killed_count=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "killed-count", default=[])],
                missing_count=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "missing-count", default=[])],
                origin=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "origin", default=[])],
                outcome=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "outcome", default=[])],
                over_time=dict_.get("over-time"),
                settlement_status_event_or_soa=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "settlement-status-event-or-SoA", default=[])],
                total_displaced_count=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "total-displaced-count", default=[])],
                transitory_events=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "Transitory-events", default=[])],
                transiting_location=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "transiting-location", default=[])],
                when=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "when", default=[])],
                provenance=provenance
            )
        elif type_ == "Cybercrimeplate":
            obj = BPDocumentGranularCyberCrimeTemplate(
                template_id,
                anchor,
                cybercrime_event=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "cybercrime-event", default=[])],
                information_stolen=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "information-stolen", default=[])],
                outcome=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "outcome", default=[])],
                over_time=dict_.get("over-time"),
                perpetrator=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "perpetrator", default=[])],
                related_crimes=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "related-crimes", default=[])],
                response=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "response", default=[])],
                victim=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "victim", default=[])],
                victim_impact=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "victim-impact", default=[])],
                when=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "when", default=[])],
                provenance=provenance
            )
        elif type_ == "Etiplate":
            obj = BPDocumentGranularETITemplate(
                template_id,
                anchor,

                agreement_length=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "agreement-length", default=[])],
                contract_amount=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "contract-amount", default=[])],
                contract_awardee=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "contract-awardee", default=[])],
                contract_awarder=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "contract-awarder", default=[])],
                etip_event=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "etip-event", default=[])],
                funding_amount=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "funding-amount", default=[])],
                funding_recipient=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "funding-recipient", default=[])],
                funding_source=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "funding-source", default=[])],
                outcome=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "outcome", default=[])],
                overall_project_value=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "overall-project-value", default=[])],
                over_time=dict_.get("over-time"),
                project_location=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "project-location", default=[])],
                project_name=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "project-name", default=[])],
                project_type=dict_.get("project-type"),
                signatories=[BPDocumentGranularTemplateProperty.from_dict(
                    item,
                    no_validation=no_validation
                ) for item in dict_get_icase(dict_, "signatories", default=[])],
                provenance=provenance
            )

        if not no_validation:
            assert obj.is_valid(), \
                "BPDocumentGranularTemplate object built from dictionary failed validation."

        ## handle Issue#52

        obj.transmogrify_outcome_slots()

        return obj


    @staticmethod
    def get_child_class(template_type):
        """Maps a polymorphic descendant class to a given ``template_type``.

        Args:
            template_type (str): the type of template to select, one of
                ``lib.utils.GRANULAR_MAT_TEMPLATE_TYPES``.

        returns:
            BPDocumentGranular<type>Template: a polymorphic descendant of
            :class:`BPDocumentGranularTemplate`.
        """

        from lib.utils import GRANULAR_MAT_TEMPLATE_TYPES

        assert template_type in GRANULAR_MAT_TEMPLATE_TYPES, "Not a valid template type!"

        if template_type == "Protestplate":
            return BPDocumentGranularProtestTemplate

        elif template_type == "Corruplate":
            return BPDocumentGranularCorruptionTemplate

        elif template_type == "Terrorplate":
            return BPDocumentGranularTerrorismTemplate

        elif template_type == "Epidemiplate":
            return BPDocumentGranularEpidemicTemplate

        elif template_type == "Disasterplate":
            return BPDocumentGranularDisasterTemplate

        elif template_type == "Displacementplate":
            return BPDocumentGranularDisplacementTemplate

        elif template_type == "Etiplate":
            return BPDocumentGranularETITemplate

        elif template_type == "Cybercrimeplate":
            return BPDocumentGranularCyberCrimeTemplate


    def is_valid(self):
        """Checks if this object (and its constituents) are all valid."""
        global LOGGER

        if self.no_validation:
            raise RuntimeError(f"Validation set to False: {self}")

        for prop in self.properties:
            if not prop.is_valid():
                LOGGER.error(f"BPDocumentGranularTemplate property {prop} is invalid.")
                return False

        return True


    def set_validation(self, validation):
        """Sets the validation mode for this node and all children of this node.

        Args:
            validation (bool): if True, sets ``no_validation`` for this node and all of its children
                to ``False``. Otherwise, all nodes set ``no_validation`` to ``True``.
        """

        assert type(validation) == bool

        self.no_validation = not validation

        for prop in self.properties:
            prop.no_validation = self.no_validation
            if prop.value:
                if prop.type_ == list:
                    for element in prop.value:
                        if hasattr(element, "set_validation"):
                            element.set_validation(self.no_validation)
                elif hasattr(prop.value, "set_validation"):
                    prop.value.set_validation(self.no_validation)


    def to_dict(self):
        """Constructs a Python dictionary following the BP schema, suitable for JSON serialization.

        Returns:
            dict: a dictionary representing this BP Span.
        """

        dict_ = {}
        for prop in [
            self.template_id,
            self.type_,
            self.anchor,
        ]:
            if prop.value is not None:
                dict_[prop.name] = prop.value

        if self.provenance.value is not None:
            dict_[self.provenance.name] = self.provenance.value.to_dict()

        return dict_


    def transmogrify_outcome_slots(self):
        """Converts any phase-1-era `outcome-X` slots to the phase-2-era `outcome` slots.

        Only affects Corruption, Protest, and Terrorism types.

        See [Issue#52](https://gitlab.mitre.org/better_program_eval/bp_lib/-/issues/52).
        """

        # Skip this function if we're working with an unaffected template type.
        affected_types = [
            BPDocumentGranularCorruptionTemplate, BPDocumentGranularProtestTemplate,
            BPDocumentGranularTerrorismTemplate,
        ]
        if type(self) not in affected_types:
            return

        # If an `outcome-averted` slot is populated, move all elements to the `outcome` slot, remove
        # any existing irrealis from each element, and add `counterfactual` irrealis to each
        # element.
        if self.outcome_averted.value:

            if self.outcome.value == None:
                self.outcome.value = []

            for element in self.outcome_averted.value:

                # check if this element is already in the outcome slot. if so, skip it.
                if element in self.outcome.value:
                    continue

                element.irrealis.value = 'counterfactual'
                if not self.outcome.value:
                    self.outcome.value = []
                self.outcome.value.append(element)

        # If an `outcome-hypothetical` slot is populated, move all elements to the `outcome` slot,
        # remove any existing irrealis from each element, and add `hypothetical` irrealis to
        # each element.
        if self.outcome_hypothetical.value:

            if self.outcome.value == None:
                self.outcome.value = []

            for element in self.outcome_hypothetical.value:

                if element in self.outcome.value:
                    continue

                element.irrealis.value = 'hypothetical'
                if not self.outcome.value:
                    self.outcome.value = []
                self.outcome.value.append(element)

        # If an `outcome-occurred` slot is populated, move all elements to the `outcome` slot and
        # remove any existing irrealis from each element.
        if self.outcome_occurred.value:

            if self.outcome.value == None:
                self.outcome.value = []

            for element in self.outcome_occurred.value:

                if element in self.outcome.value:
                    continue

                element.irrealis.value = None
                if not self.outcome.value:
                    self.outcome.value = []
                self.outcome.value.append(element)

        self.outcome_averted.value = None
        self.outcome_hypothetical.value = None
        self.outcome_occurred.value = None


class BPDocumentGranularCorruptionTemplate(BPDocumentGranularTemplate):
    """Represents a Granular Corruption Template object in a BETTER Project document.

    Args:
        template_id (str): a unique ID for this event.
        anchor (str): an SSID referring to the MAT Annotation used to create this template.
            The SpanSet should only have one span in it.
        charged_with (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        corrupt_event (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        judicial_actions (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        fine (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        over_time (bool): whether this instance of corruption ocurred over a length of time or not.
        outcome (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        outcome_averted (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        outcome_occurred (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        outcome_hypothetical (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        prison_term (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        where (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        who (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        provenance (BPDocumentProvenance): a Provenance object pretaining to this event.
        no_validation (bool): If True, skip validation when asked to. Defaults to False.
    """

    def __init__(
        self,
        template_id,
        anchor,
        charged_with=None,
        corrupt_event=None,
        judicial_actions=None,
        fine=None,
        over_time=None,
        outcome=None,
        outcome_averted=None,
        outcome_occurred=None,
        outcome_hypothetical=None,
        prison_term=None,
        where=None,
        who=None,
        provenance=None,
        no_validation=False
    ):

        super().__init__(template_id, "Corruplate", anchor, provenance, no_validation)

        self.charged_with = BPProperty(
            "charged-with",
            False,
            value=charged_with,
            type_=list,
            no_validation=no_validation
        )
        self.corrupt_event = BPProperty(
            "corrupt-event",
            False,
            value=corrupt_event,
            type_=list,
            no_validation=no_validation
        )
        self.judicial_actions = BPProperty(
            "judicial-actions",
            False,
            value=judicial_actions,
            type_=list,
            no_validation=no_validation
        )
        self.fine = BPProperty(
            "fine",
            False,
            value=fine,
            type_=list,
            no_validation=no_validation
        )
        self.over_time = BPProperty(
            "over-time",
            False,
            type_=bool,
            value=over_time,
            no_validation=no_validation
        )
        self.outcome = BPProperty(
            "outcome",
            False,
            value=outcome,
            type_=list,
            no_validation=no_validation
        )
        self.outcome_averted = BPProperty(
            "outcome-averted",
            False,
            value=outcome_averted,
            type_=list,
            no_validation=no_validation
        )
        self.outcome_occurred = BPProperty(
            "outcome-occurred",
            False,
            value=outcome_occurred,
            type_=list,
            no_validation=no_validation
        )
        self.outcome_hypothetical = BPProperty(
            "outcome-hypothetical",
            False,
            value=outcome_hypothetical,
            type_=list,
            no_validation=no_validation
        )
        self.prison_term = BPProperty(
            "prison-term",
            False,
            value=prison_term,
            type_=list,
            no_validation=no_validation
        )
        self.where = BPProperty(
            "where",
            False,
            value=where,
            type_=list,
            no_validation=no_validation
        )
        self.who = BPProperty(
            "who",
            False,
            value=who,
            type_=list,
            no_validation=no_validation
        )


    def __repr__(self):
        return (
            f"BPDocumentGranularCorruptionTemplate({self.template_id.value}, "
            f"{self.anchor.value}, "
            f"{self.charged_with.value}, "
            f"{self.corrupt_event.value}, "
            f"{self.judicial_actions.value}, "
            f"{self.fine.value}, "
            f"{self.over_time}, "
            f"{self.outcome.value}, "
            f"{self.outcome_averted.value}, "
            f"{self.outcome_occurred.value}, "
            f"{self.outcome_hypothetical.value}, "
            f"{self.prison_term.value}, "
            f"{self.where.value}, "
            f"{self.who.value}, "
            f"{repr(self.provenance.value)}, {self.no_validation})"
        )


    def __str__(self):
        return f"GranCorruplate<{self.template_id}>"


    def is_valid(self):
        """Checks if this object (and its constituents) are all valid."""
        global LOGGER

        if self.no_validation:
            raise RuntimeError(f"Validation set to False: {self}")

        base = super().is_valid()
        if not base:
            return False

        for prop in [
            self.charged_with,
            self.corrupt_event,
            self.judicial_actions,
            self.fine,
            self.over_time,
            self.outcome,
            self.outcome_averted,
            self.outcome_occurred,
            self.outcome_hypothetical,
            self.prison_term,
            self.where,
            self.who,
            self.provenance,
        ]:
            if not prop.is_valid():
                LOGGER.error(f"BPDocumentGranularCorruptionTemplate property {prop} is invalid.")
                return False

        return True


    def set_validation(self, validation):
        """Sets the validation mode for this node and all children of this node.

        Args:
            validation (bool): if True, sets ``no_validation`` for this node and all of its children
                to ``False``. Otherwise, all nodes set ``no_validation`` to ``True``.
        """

        assert type(validation) == bool

        super().set_validation(validation)

        self.charged_with.no_validation = self.no_validation
        self.corrupt_event.no_validation = self.no_validation
        self.judicial_actions.no_validation = self.no_validation
        self.fine.no_validation = self.no_validation
        self.over_time.no_validation = self.no_validation
        self.outcome.no_validation = self.no_validation
        self.outcome_averted.no_validation = self.no_validation
        self.outcome_occurred.no_validation = self.no_validation
        self.outcome_hypothetical.no_validation = self.no_validation
        self.prison_term.no_validation = self.no_validation
        self.where.no_validation = self.no_validation
        self.who.no_validation = self.no_validation

        ## recurse

        if self.charged_with.value:
            for prop in self.charged_with.value:
                prop.set_validation(validation)
        if self.corrupt_event.value:
            for prop in self.corrupt_event.value:
                prop.set_validation(validation)
        if self.judicial_actions.value:
            for prop in self.judicial_actions.value:
                prop.set_validation(validation)
        if self.fine.value:
            for prop in self.fine.value:
                prop.set_validation(validation)
        if self.outcome.value:
            for prop in self.outcome.value:
                prop.set_validation(validation)
        if self.outcome_averted.value:
            for prop in self.outcome_averted.value:
                prop.set_validation(validation)
        if self.outcome_occurred.value:
            for prop in self.outcome_occurred.value:
                prop.set_validation(validation)
        if self.outcome_hypothetical.value:
            for prop in self.outcome_hypothetical.value:
                prop.set_validation(validation)
        if self.prison_term.value:
            for prop in self.prison_term.value:
                prop.set_validation(validation)
        if self.where.value:
            for prop in self.where.value:
                prop.set_validation(validation)
        if self.who.value:
            for prop in self.who.value:
                prop.set_validation(validation)


    def to_dict(self):
        """Constructs a Python dictionary following the BP schema, suitable for JSON serialization.

        Returns:
            dict: a dictionary representing this BP Span.
        """

        if not self.no_validation:
            assert self.is_valid(), "Span object is invalid."

        dict_ = super().to_dict()

        for prop in [
            self.charged_with,
            self.corrupt_event,
            self.judicial_actions,
            self.fine,
            self.over_time,
            self.outcome,
            self.outcome_averted,
            self.outcome_occurred,
            self.outcome_hypothetical,
            self.prison_term,
            self.where,
            self.who,
        ]:
            if prop.value:
                if prop.type_ == list:
                    dict_[prop.name] = [temp_prop.to_dict() for temp_prop in prop.value]
                else:
                    dict_[prop.name] = prop.value

        return dict_


class BPDocumentGranularCyberCrimeTemplate(BPDocumentGranularTemplate):
    """Represents a Granular Cyber Crime Template object in a BETTER Project document.

    Args:
        template_id (str): a unique ID for this event.
        anchor (str): an SSID referring to the MAT Annotation used to create this template.
            The SpanSet should only have one span in it.
        cybercrime_event (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        information_stolen (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        outcome (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        over_time (bool): whether this act of terrorism ocurred over a length of time or not.
        perpetrator (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        related_crimes (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        response (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        victim (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        victim_impact (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        when (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        provenance (BPDocumentProvenance): a Provenance object pretaining to this Template.
        no_validation (bool): If True, skip validation when asked to. Defaults to False.
    """

    def __init__(
        self,
        template_id,
        anchor,
        cybercrime_event=None,
        information_stolen=None,
        outcome=None,
        over_time=None,
        perpetrator=None,
        related_crimes=None,
        response=None,
        victim=None,
        victim_impact=None,
        when=None,
        provenance=None,
        no_validation=False
    ):

        super().__init__(template_id, "Cybercrimeplate", anchor, provenance, no_validation)

        self.cybercrime_event = BPProperty(
            "cybercrime-event",
            False,
            value=cybercrime_event,
            type_=list,
            no_validation=no_validation
        )
        self.information_stolen = BPProperty(
            "information-stolen",
            False,
            value=information_stolen,
            type_=list,
            no_validation=no_validation
        )
        self.outcome = BPProperty(
            "outcome",
            False,
            value=outcome,
            type_=list,
            no_validation=no_validation
        )
        self.over_time = BPProperty(
            "over-time",
            False,
            value=over_time,
            type_=bool,
            no_validation=no_validation
        )
        self.perpetrator = BPProperty(
            "perpetrator",
            False,
            value=perpetrator,
            type_=list,
            no_validation=no_validation
        )
        self.related_crimes = BPProperty(
            "related-crimes",
            False,
            value=related_crimes,
            type_=list,
            no_validation=no_validation
        )
        self.response = BPProperty(
            "response",
            False,
            value=response,
            type_=list,
            no_validation=no_validation
        )
        self.victim = BPProperty(
            "victim",
            False,
            value=victim,
            type_=list,
            no_validation=no_validation
        )
        self.victim_impact = BPProperty(
            "victim-impact",
            False,
            value=victim_impact,
            type_=list,
            no_validation=no_validation
        )
        self.when = BPProperty(
            "when",
            False,
            value=when,
            type_=list,
            no_validation=no_validation
        )


    def __repr__(self):
        return (
            f"BPDocumentGranularCyberCrimeTemplate({self.template_id.value}, "
            f"{self.anchor.value}, "
            f"{self.cybercrime_event}, "
            f"{self.information_stolen}, "
            f"{self.outcome}, "
            f"{self.over_time}, "
            f"{self.perpetrator}, "
            f"{self.related_crimes}, "
            f"{self.response}, "
            f"{self.victim}, "
            f"{self.victim_impact}, "
            f"{self.when}, "
            f"{repr(self.provenance.value)}, {self.no_validation})"
        )


    def __str__(self):
        return f"GranCybercrimeplate<{self.template_id.value}>"


    def is_valid(self):
        """Checks if this object (and its constituents) are all valid."""
        global LOGGER

        if self.no_validation:
            raise RuntimeError(f"Validation set to False: {self}")

        base = super().is_valid()
        if not base:
            return False

        for prop in [
            self.cybercrime_event,
            self.information_stolen,
            self.outcome,
            self.over_time,
            self.perpetrator,
            self.related_crimes,
            self.response,
            self.victim,
            self.victim_impact,
            self.when,
        ]:
            if not prop.is_valid():
                LOGGER.error(f"BPDocumentGranularCyberCrimeTemplate property {prop} is invalid.")
                return False

        return True


    def set_validation(self, validation):
        """Sets the validation mode for this node and all children of this node.

        Args:
            validation (bool): if True, sets ``no_validation`` for this node and all of its children
                to ``False``. Otherwise, all nodes set ``no_validation`` to ``True``.
        """

        assert type(validation) == bool

        super().set_validation(validation)

        fields = [
            self.cybercrime_event,
            self.information_stolen,
            self.outcome,
            self.over_time,
            self.perpetrator,
            self.related_crimes,
            self.response,
            self.victim,
            self.victim_impact,
            self.when,
        ]

        for field in fields:
            field.no_validation = self.no_validation
            if field.value and field.type_ == list:
                for prop in field.value:
                    prop.set_validation(validation)


    def to_dict(self):
        """Constructs a Python dictionary following the BP schema, suitable for JSON serialization.

        Returns:
            dict: a dictionary representing this BP Span.
        """

        if not self.no_validation:
            assert self.is_valid(), "Span object is invalid."

        dict_ = super().to_dict()

        for field in [
            self.cybercrime_event,
            self.information_stolen,
            self.outcome,
            self.over_time,
            self.perpetrator,
            self.related_crimes,
            self.response,
            self.victim,
            self.victim_impact,
            self.when,
        ]:
            if field.value:
                if field.type_ == list:
                    dict_[field.name] = [temp_field.to_dict() for temp_field in field.value]
                else:
                    dict_[field.name] = field.value

        return dict_


class BPDocumentGranularDisasterTemplate(BPDocumentGranularTemplate):
    """Represents a Granular Disaster Template object in a BETTER Project document.

    Args:
        template_id (str): a unique ID for this template. Converted to a
            :class:`BPDocumentGranularTemplateProperty` on init.
        anchor (str): an SSID referring to the MAT Annotation used to create this template.
            The SpanSet should only have one span in it. Converted to a
            :class:`BPDocumentGranularTemplateProperty` on init.
        affected_cumulative_count  (list): a list of :class:`BPDocumentGranularTemplateProperty`
            objects.
        announce_disaster_warnings (list): a list of :class:`BPDocumentGranularTemplateProperty`
            objects.
        assistance_needed (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        assistance_provided (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        damage (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        declare_emergency (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        disease_outbreak_events (list): a list of :class:`BPDocumentGranularTemplateProperty`
            objects.
        human_displacement_events (list): a list of :class:`BPDocumentGranularTemplateProperty`
            objects.
        individuals_affected (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        injured_count (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        killed_count (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        major_disaster_event (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        missing_count (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        outcome (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        over_time (bool): whether this instance of corruption ocurred over a length of time or not.
        rescue_events (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        rescued_count (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        related_natural_phenomena (list): a list of :class:`BPDocumentGranularTemplateProperty`
            objects.
        repair (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        responders (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        where (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        when (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        provenance (BPDocumentProvenance): a Provenance object pretaining to this event.
        no_validation (bool): If True, skip validation when asked to. Defaults to False.
    """

    def __init__(
        self,
        template_id,
        anchor,
        affected_cumulative_count=None,
        announce_disaster_warnings=None,
        assistance_needed=None,
        assistance_provided=None,
        damage=None,
        declare_emergency=None,
        disease_outbreak_events=None,
        human_displacement_events=None,
        individuals_affected=None,
        injured_count=None,
        killed_count=None,
        major_disaster_event=None,
        missing_count=None,
        outcome=None,
        over_time=None,
        rescue_events=None,
        rescued_count=None,
        related_natural_phenomena=None,
        repair=None,
        responders=None,
        when=None,
        where=None,
        provenance=None,
        no_validation=False
    ):

        super().__init__(template_id, "Disasterplate", anchor, provenance, no_validation)

        self.affected_cumulative_count = BPProperty(
            "affected-cumulative-count",
            False,
            value=affected_cumulative_count,
            type_=list,
            no_validation=no_validation
        )
        self.announce_disaster_warnings = BPProperty(
            "announce-disaster-warnings",
            False,
            value=announce_disaster_warnings,
            type_=list,
            no_validation=no_validation
        )
        self.assistance_needed = BPProperty(
            "assistance-needed",
            False,
            value=assistance_needed,
            type_=list,
            no_validation=no_validation
        )
        self.assistance_provided = BPProperty(
            "assistance-provided",
            False,
            value=assistance_provided,
            type_=list,
            no_validation=no_validation
        )
        self.damage = BPProperty(
            "damage",
            False,
            value=damage,
            type_=list,
            no_validation=no_validation
        )
        self.declare_emergency = BPProperty(
            "declare-emergency",
            False,
            value=declare_emergency,
            type_=list,
            no_validation=no_validation
        )
        self.disease_outbreak_events = BPProperty(
            "disease-outbreak-events",
            False,
            value=disease_outbreak_events,
            type_=list,
            no_validation=no_validation
        )
        self.human_displacement_events = BPProperty(
            "human-displacement-events",
            False,
            value=human_displacement_events,
            type_=list,
            no_validation=no_validation
        )
        self.individuals_affected = BPProperty(
            "individuals-affected",
            False,
            value=individuals_affected,
            type_=list,
            no_validation=no_validation
        )
        self.injured_count = BPProperty(
            "injured-count",
            False,
            value=injured_count,
            type_=list,
            no_validation=no_validation
        )
        self.killed_count = BPProperty(
            "killed-count",
            False,
            value=killed_count,
            type_=list,
            no_validation=no_validation
        )
        self.major_disaster_event = BPProperty(
            "major-disaster-event",
            False,
            value=major_disaster_event,
            type_=list,
            no_validation=no_validation
        )
        self.missing_count = BPProperty(
            "missing-count",
            False,
            value=missing_count,
            type_=list,
            no_validation=no_validation
        )
        self.outcome = BPProperty(
            "outcome",
            False,
            value=outcome,
            type_=list,
            no_validation=no_validation
        )
        self.over_time = BPProperty(
            "over-time",
            False,
            type_=bool,
            value=over_time,
            no_validation=no_validation
        )
        self.rescue_events = BPProperty(
            "rescue-events",
            False,
            value=rescue_events,
            type_=list,
            no_validation=no_validation
        )
        self.rescued_count = BPProperty(
            "rescued-count",
            False,
            value=rescued_count,
            type_=list,
            no_validation=no_validation
        )
        self.related_natural_phenomena = BPProperty(
            "related-natural-phenomena",
            False,
            value=related_natural_phenomena,
            type_=list,
            no_validation=no_validation
        )
        self.repair = BPProperty(
            "repair",
            False,
            value=repair,
            type_=list,
            no_validation=no_validation
        )
        self.responders = BPProperty(
            "responders",
            False,
            value=responders,
            type_=list,
            no_validation=no_validation
        )
        self.where = BPProperty(
            "where",
            False,
            value=where,
            type_=list,
            no_validation=no_validation
        )
        self.when = BPProperty(
            "when",
            False,
            value=when,
            type_=list,
            no_validation=no_validation
        )


    def __repr__(self):
        return (
            f"BPDocumentGranularDisaterTemplate({self.template_id.value}, "
            f"{self.anchor.value}, "
            f"{self.affected_cumulative_count}, "
            f"{self.announce_disaster_warnings}, "
            f"{self.assistance_needed}, "
            f"{self.assistance_provided}, "
            f"{self.damage}, "
            f"{self.declare_emergency}, "
            f"{self.disease_outbreak_events}, "
            f"{self.human_displacement_events}, "
            f"{self.individuals_affected}, "
            f"{self.injured_count}, "
            f"{self.killed_count}, "
            f"{self.major_disaster_event}, "
            f"{self.missing_count}, "
            f"{self.outcome}, "
            f"{self.over_time}, "
            f"{self.rescue_events}, "
            f"{self.rescued_count}, "
            f"{self.related_natural_phenomena}, "
            f"{self.repair}, "
            f"{self.responders}, "
            f"{self.where}, "
            f"{self.when}, "
            f"{self.where.value}, "
            f"{self.when.value}, "
            f"{repr(self.provenance.value)}, {self.no_validation})"
        )


    def __str__(self):
        return f"GranDisasterplate<{self.template_id}>"


    def is_valid(self):
        """Checks if this object (and its constituents) are all valid."""
        global LOGGER

        if self.no_validation:
            raise RuntimeError(f"Validation set to False: {self}")

        base = super().is_valid()
        if not base:
            return False

        for prop in [
            self.affected_cumulative_count,
            self.announce_disaster_warnings,
            self.assistance_needed,
            self.assistance_provided,
            self.damage,
            self.declare_emergency,
            self.disease_outbreak_events,
            self.human_displacement_events,
            self.individuals_affected,
            self.injured_count,
            self.killed_count,
            self.major_disaster_event,
            self.missing_count,
            self.outcome,
            self.over_time,
            self.rescue_events,
            self.rescued_count,
            self.related_natural_phenomena,
            self.repair,
            self.responders,
            self.when,
            self.where,
            self.provenance,
        ]:
            if not prop.is_valid():
                LOGGER.error(f"BPDocumentGranularDisasterTemplate property {prop} is invalid.")
                return False

        return True


    def set_validation(self, validation):
        """Sets the validation mode for this node and all children of this node.

        Args:
            validation (bool): if True, sets ``no_validation`` for this node and all of its children
                to ``False``. Otherwise, all nodes set ``no_validation`` to ``True``.
        """

        assert type(validation) == bool

        super().set_validation(validation)

        self.affected_cumulative_count = self.no_validation
        self.announce_disaster_warnings = self.no_validation
        self.assistance_needed = self.no_validation
        self.assistance_provided = self.no_validation
        self.damage = self.no_validation
        self.declare_emergency = self.no_validation
        self.disease_outbreak_events = self.no_validation
        self.human_displacement_events = self.no_validation
        self.individuals_affected = self.no_validation
        self.injured_count = self.no_validation
        self.killed_count = self.no_validation
        self.major_disaster_event = self.no_validation
        self.missing_count = self.no_validation
        self.outcome = self.no_validation
        self.over_time = self.no_validation
        self.rescue_events = self.no_validation
        self.rescued_count = self.no_validation
        self.related_natural_phenomena = self.no_validation
        self.repair = self.no_validation
        self.responders = self.no_validation
        self.when = self.no_validation
        self.where = self.no_validation

        ## recurse for list types

        if self.affected_cumulative_count.value:
            for prop in self.affected_cumulative_count.value:
                prop.set_validation(validation)
        if self.announce_disaster_warnings.value:
            for prop in self.announce_disaster_warnings.value:
                prop.set_validation(validation)
        if self.assistance_needed.value:
            for prop in self.assistance_needed.value:
                prop.set_validation(validation)
        if self.assistance_provided.value:
            for prop in self.assistance_provided.value:
                prop.set_validation(validation)
        if self.damage.value:
            for prop in self.damage.value:
                prop.set_validation(validation)
        if self.declare_emergency.value:
            for prop in self.declare_emergency.value:
                prop.set_validation(validation)
        if self.disease_outbreak_events.value:
            for prop in self.disease_outbreak_events.value:
                prop.set_validation(validation)
        if self.human_displacement_events.value:
            for prop in self.human_displacement_events.value:
                prop.set_validation(validation)
        if self.individuals_affected.value:
            for prop in self.individuals_affected.value:
                prop.set_validation(validation)
        if self.injured_count.value:
            for prop in self.injured_count.value:
                prop.set_validation(validation)
        if self.killed_count.value:
            for prop in self.killed_count.value:
                prop.set_validation(validation)
        if self.major_disaster_event.value:
            for prop in self.major_disaster_event.value:
                prop.set_validation(validation)
        if self.missing_count.value:
            for prop in self.missing_count.value:
                prop.set_validation(validation)
        if self.outcome.value:
            for prop in self.outcome.value:
                prop.set_validation(validation)
        if self.rescue_events.value:
            for prop in self.rescue_events.value:
                prop.set_validation(validation)
        if self.rescued_count.value:
            for prop in self.rescued_count.value:
                prop.set_validation(validation)
        if self.related_natural_phenomena.value:
            for prop in self.related_natural_phenomena.value:
                prop.set_validation(validation)
        if self.repair.value:
            for prop in self.repair.value:
                prop.set_validation(validation)
        if self.responders.value:
            for prop in self.responders.value:
                prop.set_validation(validation)
        if self.when.value:
            for prop in self.when.value:
                prop.set_validation(validation)
        if self.where.value:
            for prop in self.where.value:
                prop.set_validation(validation)


    def to_dict(self):
        """Constructs a Python dictionary following the BP schema, suitable for JSON serialization.

        Returns:
            dict: a dictionary representing this BP Span.
        """

        if not self.no_validation:
            assert self.is_valid(), "Span object is invalid."

        dict_ = super().to_dict()

        for prop in [
            self.affected_cumulative_count,
            self.announce_disaster_warnings,
            self.assistance_needed,
            self.assistance_provided,
            self.damage,
            self.declare_emergency,
            self.disease_outbreak_events,
            self.human_displacement_events,
            self.individuals_affected,
            self.injured_count,
            self.killed_count,
            self.major_disaster_event,
            self.missing_count,
            self.outcome,
            self.over_time,
            self.rescue_events,
            self.rescued_count,
            self.related_natural_phenomena,
            self.repair,
            self.responders,
            self.when,
            self.where,
        ]:
            if prop.value:
                if prop.type_ == list:
                    dict_[prop.name] = [temp_prop.to_dict() for temp_prop in prop.value]
                else:
                    dict_[prop.name] = prop.value

        return dict_


class BPDocumentGranularDisplacementTemplate(BPDocumentGranularTemplate):
    """Represents a Granular Displacement Template object in a BETTER Project document.

    Args:
        template_id (str): a unique ID for this template. Converted to a
            :class:`BPDocumentGranularTemplateProperty` on init.
        anchor (str): an SSID referring to the MAT Annotation used to create this template.
            The SpanSet should only have one span in it. Converted to a
            :class:`BPDocumentGranularTemplateProperty` on init.
        assistance_provided (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        assistance_needed (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        blocked_migration_count (list): a list of :class:`BPDocumentGranularTemplateProperty`
            objects.
        current_location (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        destination (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        detained_count (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        event_or_soa_at_origin (list): a list of :class:`BPDocumentGranularTemplateProperty`
            objects.
        group_identity (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        human_displacement_event (list): a list of :class:`BPDocumentGranularTemplateProperty`
            objects.
        injured_count (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        killed_count (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        missing_count (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        origin (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        outcome (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        over_time (bool): whether this instance of corruption ocurred over a length of time or not.
        settlement_status_event_or_soa (list): a list of :class:`BPDocumentGranularTemplateProperty`
            objects.
        total_displaced_count (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        transitory_events (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        transiting_location (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        when (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        provenance (BPDocumentProvenance): a Provenance object pretaining to this event.
        no_validation (bool): If True, skip validation when asked to. Defaults to False.
    """

    def __init__(
        self,
        template_id,
        anchor,
        assistance_provided=None,
        assistance_needed=None,
        blocked_migration_count=None,
        current_location=None,
        destination=None,
        detained_count=None,
        event_or_soa_at_origin=None,
        group_identity=None,
        human_displacement_event=None,
        injured_count=None,
        killed_count=None,
        missing_count=None,
        origin=None,
        outcome=None,
        over_time=None,
        settlement_status_event_or_soa=None,
        total_displaced_count=None,
        transitory_events=None,
        transiting_location=None,
        when=None,
        provenance=None,
        no_validation=False
    ):

        super().__init__(template_id, "Displacementplate", anchor, provenance, no_validation)

        self.assistance_provided = BPProperty(
            "assistance-provided",
            False,
            type_=list,
            value=assistance_provided,
            no_validation=no_validation
        )
        self.assistance_needed = BPProperty(
            "assistance-needed",
            False,
            type_=list,
            value=assistance_needed,
            no_validation=no_validation
        )
        self.blocked_migration_count = BPProperty(
            "blocked-migration-count",
            False,
            type_=list,
            value=blocked_migration_count,
            no_validation=no_validation
        )
        self.current_location = BPProperty(
            "current-location",
            False,
            type_=list,
            value=current_location,
            no_validation=no_validation
        )
        self.destination = BPProperty(
            "destination",
            False,
            type_=list,
            value=destination,
            no_validation=no_validation
        )
        self.detained_count = BPProperty(
            "detained-count",
            False,
            type_=list,
            value=detained_count,
            no_validation=no_validation
        )
        self.event_or_soa_at_origin = BPProperty(
            "event-or-SoA-at-origin",
            False,
            type_=list,
            value=event_or_soa_at_origin,
            no_validation=no_validation
        )
        self.group_identity = BPProperty(
            "group-identity",
            False,
            type_=list,
            value=group_identity,
            no_validation=no_validation
        )
        self.human_displacement_event = BPProperty(
            "human-displacement-event",
            False,
            type_=list,
            value=human_displacement_event,
            no_validation=no_validation
        )
        self.injured_count = BPProperty(
            "injured-count",
            False,
            type_=list,
            value=injured_count,
            no_validation=no_validation
        )
        self.killed_count = BPProperty(
            "killed-count",
            False,
            type_=list,
            value=killed_count,
            no_validation=no_validation
        )
        self.missing_count = BPProperty(
            "missing-count",
            False,
            type_=list,
            value=missing_count,
            no_validation=no_validation
        )
        self.origin = BPProperty(
            "origin",
            False,
            type_=list,
            value=origin,
            no_validation=no_validation
        )
        self.outcome = BPProperty(
            "outcome",
            False,
            type_=list,
            value=outcome,
            no_validation=no_validation
        )
        self.over_time = BPProperty(
            "over-time",
            False,
            type_=bool,
            value=over_time,
            no_validation=no_validation
        )
        self.settlement_status_event_or_soa = BPProperty(
            "settlement-status-event-or-SoA",
            False,
            type_=list,
            value=settlement_status_event_or_soa,
            no_validation=no_validation
        )
        self.total_displaced_count = BPProperty(
            "total-displaced-count",
            False,
            type_=list,
            value=total_displaced_count,
            no_validation=no_validation
        )
        self.transitory_events = BPProperty(
            "Transitory-events",
            False,
            type_=list,
            value=transitory_events,
            no_validation=no_validation
        )
        self.transiting_location = BPProperty(
            "transiting-location",
            False,
            type_=list,
            value=transiting_location,
            no_validation=no_validation
        )
        self.when = BPProperty(
            "when",
            False,
            type_=list,
            value=when,
            no_validation=no_validation
        )


    def __repr__(self):
        return (
            f"BPDocumentGranularDisplacementTemplate({self.template_id.value}, "
            f"{self.anchor.value}, "
            f"{self.assistance_provided.value}, "
            f"{self.assistance_needed.value}, "
            f"{self.blocked_migration_count.value}, "
            f"{self.current_location.value}, "
            f"{self.destination.value}, "
            f"{self.detained_count.value}, "
            f"{self.event_or_soa_at_origin.value}, "
            f"{self.group_identity.value}, "
            f"{self.human_displacement_event.value}, "
            f"{self.injured_count.value}, "
            f"{self.killed_count.value}, "
            f"{self.missing_count.value}, "
            f"{self.origin.value}, "
            f"{self.outcome.value}, "
            f"{self.over_time.value}, "
            f"{self.settlement_status_event_or_soa.value}, "
            f"{self.total_displaced_count.value}, "
            f"{self.transitory_events.value}, "
            f"{self.transiting_location.value}, "
            f"{self.when.value}, "
            f"{repr(self.provenance.value)}, {self.no_validation})"
        )


    def __str__(self):
        return f"GranDisplacementplate<{self.template_id}>"


    def is_valid(self):
        """Checks if this object (and its constituents) are all valid."""
        global LOGGER

        if self.no_validation:
            raise RuntimeError(f"Validation set to False: {self}")

        base = super().is_valid()
        if not base:
            return False

        for prop in [
            self.assistance_provided,
            self.assistance_needed,
            self.blocked_migration_count,
            self.current_location,
            self.destination,
            self.detained_count,
            self.event_or_soa_at_origin,
            self.group_identity,
            self.human_displacement_event,
            self.injured_count,
            self.killed_count,
            self.missing_count,
            self.origin,
            self.outcome,
            self.over_time,
            self.settlement_status_event_or_soa,
            self.total_displaced_count,
            self.transitory_events,
            self.transiting_location,
            self.when,
            self.provenance,
        ]:
            if not prop.is_valid():
                LOGGER.error(f"BPDocumentGranularDisplacementTemplate property {prop} is invalid.")
                return False

        return True


    def set_validation(self, validation):
        """Sets the validation mode for this node and all children of this node.

        Args:
            validation (bool): if True, sets ``no_validation`` for this node and all of its children
                to ``False``. Otherwise, all nodes set ``no_validation`` to ``True``.
        """

        assert type(validation) == bool

        super().set_validation(validation)

        self.assistance_provided.no_validation = self.no_validation
        self.assistance_needed.no_validation = self.no_validation
        self.blocked_migration_count.no_validation = self.no_validation
        self.current_location.no_validation = self.no_validation
        self.destination.no_validation = self.no_validation
        self.detained_count.no_validation = self.no_validation
        self.event_or_soa_at_origin.no_validation = self.no_validation
        self.group_identity.no_validation = self.no_validation
        self.human_displacement_event.no_validation = self.no_validation
        self.injured_count.no_validation = self.no_validation
        self.killed_count.no_validation = self.no_validation
        self.missing_count.no_validation = self.no_validation
        self.origin.no_validation = self.no_validation
        self.outcome.no_validation = self.no_validation
        self.over_time.no_validation = self.no_validation
        self.settlement_status_event_or_soa.no_validation = self.no_validation
        self.total_displaced_count.no_validation = self.no_validation
        self.transitory_events.no_validation = self.no_validation
        self.transiting_location.no_validation = self.no_validation
        self.when.no_validation = self.no_validation

        ## recurse for list types

        if self.assistance_provided.value:
            for prop in self.assistance_provided.value:
                prop.set_validation(validation)
        if self.assistance_needed.value:
            for prop in self.assistance_needed.value:
                prop.set_validation(validation)
        if self.blocked_migration_count.value:
            for prop in self.blocked_migration_count.value:
                prop.set_validation(validation)
        if self.current_location.value:
            for prop in self.current_location.value:
                prop.set_validation(validation)
        if self.destination.value:
            for prop in self.destination.value:
                prop.set_validation(validation)
        if self.detained_count.value:
            for prop in self.detained_count.value:
                prop.set_validation(validation)
        if self.event_or_soa_at_origin.value:
            for prop in self.event_or_soa_at_origin.value:
                prop.set_validation(validation)
        if self.group_identity.value:
            for prop in self.group_identity.value:
                prop.set_validation(validation)
        if self.human_displacement_event.value:
            for prop in self.human_displacement_event.value:
                prop.set_validation(validation)
        if self.injured_count.value:
            for prop in self.injured_count.value:
                prop.set_validation(validation)
        if self.killed_count.value:
            for prop in self.killed_count.value:
                prop.set_validation(validation)
        if self.missing_count.value:
            for prop in self.missing_count.value:
                prop.set_validation(validation)
        if self.origin.value:
            for prop in self.origin.value:
                prop.set_validation(validation)
        if self.outcome.value:
            for prop in self.outcome.value:
                prop.set_validation(validation)
        if self.over_time.value:
            for prop in self.over_time.value:
                prop.set_validation(validation)
        if self.settlement_status_event_or_soa.value:
            for prop in self.settlement_status_event_or_soa.value:
                prop.set_validation(validation)
        if self.total_displaced_count.value:
            for prop in self.total_displaced_count.value:
                prop.set_validation(validation)
        if self.transitory_events.value:
            for prop in self.transitory_events.value:
                prop.set_validation(validation)
        if self.transiting_location.value:
            for prop in self.transiting_location.value:
                prop.set_validation(validation)
        if self.when.value:
            for prop in self.when.value:
                prop.set_validation(validation)


    def to_dict(self):
        """Constructs a Python dictionary following the BP schema, suitable for JSON serialization.

        Returns:
            dict: a dictionary representing this BP Span.
        """

        if not self.no_validation:
            assert self.is_valid(), "Span object is invalid."

        dict_ = super().to_dict()

        for prop in [
            self.assistance_provided,
            self.assistance_needed,
            self.blocked_migration_count,
            self.current_location,
            self.destination,
            self.detained_count,
            self.event_or_soa_at_origin,
            self.group_identity,
            self.human_displacement_event,
            self.injured_count,
            self.killed_count,
            self.missing_count,
            self.origin,
            self.outcome,
            self.over_time,
            self.settlement_status_event_or_soa,
            self.total_displaced_count,
            self.transitory_events,
            self.transiting_location,
            self.when,
        ]:
            if prop.value:
                if prop.type_ == list:
                    dict_[prop.name] = [temp_prop.to_dict() for temp_prop in prop.value]
                else:
                    dict_[prop.name] = prop.value

        return dict_


class BPDocumentGranularEpidemicTemplate(BPDocumentGranularTemplate):
    """Represents a Granular Epidemic Template object in a BETTER Project document.

    Args:
        template_id (str): a unique ID for this event.
        anchor (str): an SSID referring to the MAT Annotation used to create this template.
            The SpanSet should only have one span in it.
        disease (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        exposed_count (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        exposed_cumulative (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        exposed_individuals (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        hospitalized_count (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        hospitalized_cumulative (list): a list of :class:`BPDocumentGranularTemplateProperty`
            objects.
        hospitalized_individuals (list): a list of :class:`BPDocumentGranularTemplateProperty`
            objects.
        infected_count (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        infected_cumulative (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        infected_individuals (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        killed_count (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        killed_cumulative (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        killed_individuals (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        NPI_Events (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        outbreak_event (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        tested_count (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        tested_cumulative (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        tested_individuals (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        recovered_count (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        recovered_cumulative (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        recovered_individuals (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        vaccinated_individuals (list): a list of :class:`BPDocumentGranularTemplateProperty`
            objects.
        vaccinated_count (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        vaccinated_cumulative (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        when (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        where (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        provenance (BPDocumentProvenance): a Provenance object pretaining to this event.
        no_validation (bool): If True, skip validation when asked to. Defaults to False.
    """

    def __init__(
        self,
        template_id,
        anchor,
        disease=None,
        exposed_count=None,
        exposed_cumulative=None,
        exposed_individuals=None,
        hospitalized_count=None,
        hospitalized_cumulative=None,
        hospitalized_individuals=None,
        infected_count=None,
        infected_cumulative=None,
        infected_individuals=None,
        killed_count=None,
        killed_cumulative=None,
        killed_individuals=None,
        NPI_Events=None,
        outbreak_event=None,
        tested_count=None,
        tested_cumulative=None,
        tested_individuals=None,
        recovered_count=None,
        recovered_cumulative=None,
        recovered_individuals=None,
        vaccinated_individuals=None,
        vaccinated_count=None,
        vaccinated_cumulative=None,
        when=None,
        where=None,
        provenance=None,
        no_validation=False
    ):

        super().__init__(template_id, "Epidemiplate", anchor, provenance, no_validation)

        self.disease = BPProperty(
            "disease",
            False,
            value=disease,
            type_=list,
            no_validation=no_validation
        )
        self.exposed_count = BPProperty(
            "exposed-count",
            False,
            value=exposed_count,
            type_=list,
            no_validation=no_validation
        )
        self.exposed_cumulative = BPProperty(
            "exposed-cumulative",
            False,
            value=exposed_cumulative,
            type_=list,
            no_validation=no_validation
        )
        self.exposed_individuals = BPProperty(
            "exposed-individuals",
            False,
            value=exposed_individuals,
            type_=list,
            no_validation=no_validation
        )
        self.hospitalized_count = BPProperty(
            "hospitalized-count",
            False,
            value=hospitalized_count,
            type_=list,
            no_validation=no_validation
        )
        self.hospitalized_cumulative = BPProperty(
            "hospitalized-cumulative",
            False,
            value=hospitalized_cumulative,
            type_=list,
            no_validation=no_validation
        )
        self.hospitalized_individuals = BPProperty(
            "hospitalized-individuals",
            False,
            value=hospitalized_individuals,
            type_=list,
            no_validation=no_validation
        )
        self.infected_count = BPProperty(
            "infected-count",
            False,
            value=infected_count,
            type_=list,
            no_validation=no_validation
        )
        self.infected_cumulative = BPProperty(
            "infected-cumulative",
            False,
            value=infected_cumulative,
            type_=list,
            no_validation=no_validation
        )
        self.infected_individuals = BPProperty(
            "infected-individuals",
            False,
            value=infected_individuals,
            type_=list,
            no_validation=no_validation
        )
        self.killed_count = BPProperty(
            "killed-count",
            False,
            value=killed_count,
            type_=list,
            no_validation=no_validation
        )
        self.killed_cumulative = BPProperty(
            "killed-cumulative",
            False,
            value=killed_cumulative,
            type_=list,
            no_validation=no_validation
        )
        self.killed_individuals = BPProperty(
            "killed-individuals",
            False,
            value=killed_individuals,
            type_=list,
            no_validation=no_validation
        )
        self.NPI_Events = BPProperty(
            "NPI-Events",
            False,
            value=NPI_Events,
            type_=list,
            no_validation=no_validation
        )
        self.outbreak_event = BPProperty(
            "outbreak-event",
            False,
            value=outbreak_event,
            type_=list,
            no_validation=no_validation
        )
        self.tested_count = BPProperty(
            "tested-count",
            False,
            value=tested_count,
            type_=list,
            no_validation=no_validation
        )
        self.tested_cumulative = BPProperty(
            "tested-cumulative",
            False,
            value=tested_cumulative,
            type_=list,
            no_validation=no_validation
        )
        self.tested_individuals = BPProperty(
            "tested-individuals",
            False,
            value=tested_individuals,
            type_=list,
            no_validation=no_validation
        )
        self.recovered_count = BPProperty(
            "recovered-count",
            False,
            value=recovered_count,
            type_=list,
            no_validation=no_validation
        )
        self.recovered_cumulative = BPProperty(
            "recovered-cumulative",
            False,
            value=recovered_cumulative,
            type_=list,
            no_validation=no_validation
        )
        self.recovered_individuals = BPProperty(
            "recovered-individuals",
            False,
            value=recovered_individuals,
            type_=list,
            no_validation=no_validation
        )
        self.vaccinated_individuals = BPProperty(
            "vaccinated-individuals",
            False,
            value=vaccinated_individuals,
            type_=list,
            no_validation=no_validation
        )
        self.vaccinated_count = BPProperty(
            "vaccinated-count",
            False,
            value=vaccinated_count,
            type_=list,
            no_validation=no_validation
        )
        self.vaccinated_cumulative = BPProperty(
            "vaccinated-cumulative",
            False,
            value=vaccinated_cumulative,
            type_=list,
            no_validation=no_validation
        )
        self.when = BPProperty(
            "when",
            False,
            value=when,
            type_=list,
            no_validation=no_validation
        )
        self.where = BPProperty(
            "where",
            False,
            value=where,
            type_=list,
            no_validation=no_validation
        )

    def __repr__(self):
        return (
            f"BPDocumentGranularEpidemicTemplate({self.template_id.value}, "
            f"{self.anchor.value}, "
            f"{self.disease.value}, "
            f"{self.exposed_count.value}, "
            f"{self.exposed_cumulative.value}, "
            f"{self.exposed_individuals.value}, "
            f"{self.hospitalized_count.value}, "
            f"{self.hospitalized_cumulative.value}, "
            f"{self.hospitalized_individuals.value}, "
            f"{self.infected_count.value}, "
            f"{self.infected_cumulative.value}, "
            f"{self.infected_individuals.value}, "
            f"{self.killed_count.value}, "
            f"{self.killed_cumulative.value}, "
            f"{self.killed_individuals.value}, "
            f"{self.NPI_Events.value}, "
            f"{self.outbreak_event.value}, "
            f"{self.tested_count.value}, "
            f"{self.tested_cumulative.value}, "
            f"{self.tested_individuals.value}, "
            f"{self.recovered_count.value}, "
            f"{self.recovered_cumulative.value}, "
            f"{self.recovered_individuals.value}, "
            f"{self.vaccinated_individuals.value}, "
            f"{self.vaccinated_count.value}, "
            f"{self.vaccinated_cumulative.value}, "
            f"{self.when.value}, "
            f"{self.where.value}, "
            f"{repr(self.provenance.value)}, {self.no_validation})"
        )


    def __str__(self):
        return f"GranEpidemiplate<{self.template_id.value}>"


    def is_valid(self):
        """Checks if this object (and its constituents) are all valid."""
        global LOGGER

        if self.no_validation:
            raise RuntimeError(f"Validation set to False: {self}")

        base = super().is_valid()
        if not base:
            return False

        for prop in [
            self.disease,
            self.exposed_count,
            self.exposed_cumulative,
            self.exposed_individuals,
            self.hospitalized_count,
            self.hospitalized_cumulative,
            self.hospitalized_individuals,
            self.infected_count,
            self.infected_cumulative,
            self.infected_individuals,
            self.killed_count,
            self.killed_cumulative,
            self.killed_individuals,
            self.NPI_Events,
            self.outbreak_event,
            self.tested_count,
            self.tested_cumulative,
            self.tested_individuals,
            self.recovered_count,
            self.recovered_cumulative,
            self.recovered_individuals,
            self.vaccinated_individuals,
            self.vaccinated_count,
            self.vaccinated_cumulative,
            self.when,
            self.where,
            self.provenance,
        ]:
            if not prop.is_valid():
                LOGGER.error(f"BPDocumentGranularEpidemicTemplate property {prop} is invalid.")
                return False

        return True


    def set_validation(self, validation):
        """Sets the validation mode for this node and all children of this node.

        Args:
            validation (bool): if True, sets ``no_validation`` for this node and all of its children
                to ``False``. Otherwise, all nodes set ``no_validation`` to ``True``.
        """

        assert type(validation) == bool

        super().set_validation(validation)

        self.disease.no_validation = self.no_validation
        self.exposed_count.no_validation = self.no_validation
        self.exposed_cumulative.no_validation = self.no_validation
        self.exposed_individuals.no_validation = self.no_validation
        self.hospitalized_count.no_validation = self.no_validation
        self.hospitalized_cumulative.no_validation = self.no_validation
        self.hospitalized_individuals.no_validation = self.no_validation
        self.infected_count.no_validation = self.no_validation
        self.infected_cumulative.no_validation = self.no_validation
        self.infected_individuals.no_validation = self.no_validation
        self.killed_count.no_validation = self.no_validation
        self.killed_cumulative.no_validation = self.no_validation
        self.killed_individuals.no_validation = self.no_validation
        self.NPI_Events.no_validation = self.no_validation
        self.outbreak_event.no_validation = self.no_validation
        self.tested_count.no_validation = self.no_validation
        self.tested_cumulative.no_validation = self.no_validation
        self.tested_individuals.no_validation = self.no_validation
        self.recovered_count.no_validation = self.no_validation
        self.recovered_cumulative.no_validation = self.no_validation
        self.recovered_individuals.no_validation = self.no_validation
        self.vaccinated_individuals.no_validation = self.no_validation
        self.vaccinated_count.no_validation = self.no_validation
        self.vaccinated_cumulative.no_validation = self.no_validation
        self.when.no_validation = self.no_validation
        self.where.no_validation = self.no_validation

        ## recurse

        if self.disease.value:
            for prop in self.disease.value:
                prop.set_validation(validation)
        if self.exposed_count.value:
            for prop in self.exposed_count.value:
                prop.set_validation(validation)
        if self.exposed_cumulative.value:
            for prop in self.exposed_cumulative.value:
                prop.set_validation(validation)
        if self.exposed_individuals.value:
            for prop in self.exposed_individuals.value:
                prop.set_validation(validation)
        if self.hospitalized_count.value:
            for prop in self.hospitalized_count.value:
                prop.set_validation(validation)
        if self.hospitalized_cumulative.value:
            for prop in self.hospitalized_cumulative.value:
                prop.set_validation(validation)
        if self.hospitalized_individuals.value:
            for prop in self.hospitalized_individuals.value:
                prop.set_validation(validation)
        if self.infected_count.value:
            for prop in self.infected_count.value:
                prop.set_validation(validation)
        if self.infected_cumulative.value:
            for prop in self.infected_cumulative.value:
                prop.set_validation(validation)
        if self.infected_individuals.value:
            for prop in self.infected_individuals.value:
                prop.set_validation(validation)
        if self.killed_count.value:
            for prop in self.killed_count.value:
                prop.set_validation(validation)
        if self.killed_cumulative.value:
            for prop in self.killed_cumulative.value:
                prop.set_validation(validation)
        if self.killed_individuals.value:
            for prop in self.killed_individuals.value:
                prop.set_validation(validation)
        if self.NPI_Events.value:
            for prop in self.NPI_Events.value:
                prop.set_validation(validation)
        if self.outbreak_event.value:
            for prop in self.outbreak_event.value:
                prop.set_validation(validation)
        if self.tested_count.value:
            for prop in self.tested_count.value:
                prop.set_validation(validation)
        if self.tested_cumulative.value:
            for prop in self.tested_cumulative.value:
                prop.set_validation(validation)
        if self.tested_individuals.value:
            for prop in self.tested_individuals.value:
                prop.set_validation(validation)
        if self.recovered_count.value:
            for prop in self.recovered_count.value:
                prop.set_validation(validation)
        if self.recovered_cumulative.value:
            for prop in self.recovered_cumulative.value:
                prop.set_validation(validation)
        if self.recovered_individuals.value:
            for prop in self.recovered_individuals.value:
                prop.set_validation(validation)
        if self.vaccinated_individuals.value:
            for prop in self.vaccinated_individuals.value:
                prop.set_validation(validation)
        if self.vaccinated_count.value:
            for prop in self.vaccinated_count.value:
                prop.set_validation(validation)
        if self.vaccinated_cumulative.value:
            for prop in self.vaccinated_cumulative.value:
                prop.set_validation(validation)
        if self.when.value:
            for prop in self.when.value:
                prop.set_validation(validation)
        if self.where.value:
            for prop in self.where.value:
                prop.set_validation(validation)


    def to_dict(self):
        """Constructs a Python dictionary following the BP schema, suitable for JSON serialization.

        Returns:
            dict: a dictionary representing this BP Span.
        """

        if not self.no_validation:
            assert self.is_valid(), "Span object is invalid."

        dict_ = super().to_dict()

        for prop in [
            self.disease,
            self.exposed_count,
            self.exposed_cumulative,
            self.exposed_individuals,
            self.hospitalized_count,
            self.hospitalized_cumulative,
            self.hospitalized_individuals,
            self.infected_count,
            self.infected_cumulative,
            self.infected_individuals,
            self.killed_count,
            self.killed_cumulative,
            self.killed_individuals,
            self.NPI_Events,
            self.outbreak_event,
            self.tested_count,
            self.tested_cumulative,
            self.tested_individuals,
            self.recovered_count,
            self.recovered_cumulative,
            self.recovered_individuals,
            self.vaccinated_individuals,
            self.vaccinated_count,
            self.vaccinated_cumulative,
            self.when,
            self.where,
        ]:
            if prop.value:
                if prop.type_ == list:
                    dict_[prop.name] = [temp_prop.to_dict() for temp_prop in prop.value]
                else:
                    dict_[prop.name] = prop.value

        return dict_


class BPDocumentGranularETITemplate(BPDocumentGranularTemplate):
    """Represents a Granular ETI (Belt and Road Initiative) Template object in a BETTER Project
    document.

    Args:
        template_id (str): a unique ID for this event.
        anchor (str): an SSID referring to the MAT Annotation used to create this template.
            The SpanSet should only have one span in it.
        agreement_length (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        contract_amount (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        contract_awardee (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        contract_awarder (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        etip_event (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        funding_amount (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        funding_recipient (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        funding_source (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        outcome (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        over_time (bool): whether this act of terrorism ocurred over a length of time or not.
        overall_project_value (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        project_location (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        project_name (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        project_type (str): one of
            ``["road", "rail", "port", "energy", "airport", "multiple", "other", "unspecified"]``.
        signatories (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        when (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        provenance (BPDocumentProvenance): a Provenance object pretaining to this Template.
        no_validation (bool): If True, skip validation when asked to. Defaults to False.
    """

    def __init__(
        self,
        template_id,
        anchor,
        agreement_length=None,
        contract_amount=None,
        contract_awardee=None,
        contract_awarder=None,
        etip_event=None,
        funding_amount=None,
        funding_recipient=None,
        funding_source=None,
        outcome=None,
        over_time=None,
        overall_project_value=None,
        project_location=None,
        project_name=None,
        project_type=None,
        signatories=None,
        when=None,
        provenance=None,
        no_validation=False
    ):

        super().__init__(template_id, "Etiplate", anchor, provenance, no_validation)

        self.agreement_length = BPProperty(
            "agreement-length",
            False,
            value=agreement_length,
            type_=list,
            no_validation=no_validation
        )
        self.contract_amount = BPProperty(
            "contract-amount",
            False,
            value=contract_amount,
            type_=list,
            no_validation=no_validation
        )
        self.contract_awardee = BPProperty(
            "contract-awardee",
            False,
            value=contract_awardee,
            type_=list,
            no_validation=no_validation
        )
        self.contract_awarder = BPProperty(
            "contract-awarder",
            False,
            value=contract_awarder,
            type_=list,
            no_validation=no_validation
        )
        self.etip_event = BPProperty(
            "etip-event",
            False,
            value=etip_event,
            type_=list,
            no_validation=no_validation
        )
        self.funding_amount = BPProperty(
            "funding-amount",
            False,
            value=funding_amount,
            type_=list,
            no_validation=no_validation
        )
        self.funding_recipient = BPProperty(
            "funding-recipient",
            False,
            value=funding_recipient,
            type_=list,
            no_validation=no_validation
        )
        self.funding_source = BPProperty(
            "funding-source",
            False,
            value=funding_source,
            type_=list,
            no_validation=no_validation
        )
        self.outcome = BPProperty(
            "outcome",
            False,
            value=outcome,
            type_=list,
            no_validation=no_validation
        )
        self.over_time = BPProperty(
            "over-time",
            False,
            value=over_time,
            type_=bool,
            no_validation=no_validation
        )
        self.overall_project_value = BPProperty(
            "overall-project-value",
            False,
            value=overall_project_value,
            type_=list,
            no_validation=no_validation
        )
        self.project_location = BPProperty(
            "project-location",
            False,
            value=project_location,
            type_=list,
            no_validation=no_validation
        )
        self.project_name = BPProperty(
            "project-name",
            False,
            value=project_name,
            type_=list,
            no_validation=no_validation
        )
        self.project_type = BPProperty(
            "project-type",
            False,
            value=project_type,
            type_=str,
            default="unspecified",
            choices=[
                "road", "rail", "port", "energy", "airport", "multiple", "other", "unspecified",
            ],
            no_validation=no_validation
        )
        self.signatories = BPProperty(
            "signatories",
            False,
            value=signatories,
            type_=list,
            no_validation=no_validation
        )
        self.when = BPProperty(
            "when",
            False,
            value=when,
            type_=list,
            no_validation=no_validation
        )


    def __repr__(self):
        return (
            f"BPDocumentGranularETITemplate({self.template_id.value}, "
            f"{self.anchor.value}, "
            f"{self.agreement_length}, "
            f"{self.contract_amount}, "
            f"{self.contract_awardee}, "
            f"{self.contract_awarder}, "
            f"{self.etip_event}, "
            f"{self.funding_amount}, "
            f"{self.funding_recipient}, "
            f"{self.funding_source}, "
            f"{self.outcome}, "
            f"{self.over_time}, "
            f"{self.overall_project_value}, "
            f"{self.project_location}, "
            f"{self.project_name}, "
            f"{self.project_type}, "
            f"{self.signatories}, "
            f"{self.when}, "
            f"{repr(self.provenance.value)}, {self.no_validation})"
        )


    def __str__(self):
        return f"GranEtiplate<{self.template_id.value}>"


    def is_valid(self):
        """Checks if this object (and its constituents) are all valid."""
        global LOGGER

        if self.no_validation:
            raise RuntimeError(f"Validation set to False: {self}")

        base = super().is_valid()
        if not base:
            return False

        for prop in [
            self.agreement_length,
            self.contract_amount,
            self.contract_awardee,
            self.contract_awarder,
            self.etip_event,
            self.funding_amount,
            self.funding_recipient,
            self.funding_source,
            self.outcome,
            self.over_time,
            self.overall_project_value,
            self.project_location,
            self.project_name,
            self.project_type,
            self.signatories,
            self.when,
        ]:
            if not prop.is_valid():
                LOGGER.error(f"BPDocumentGranularETITemplate property {prop} is invalid.")
                return False

        return True


    def set_validation(self, validation):
        """Sets the validation mode for this node and all children of this node.

        Args:
            validation (bool): if True, sets ``no_validation`` for this node and all of its children
                to ``False``. Otherwise, all nodes set ``no_validation`` to ``True``.
        """

        assert type(validation) == bool

        super().set_validation(validation)

        fields = [
            self.agreement_length,
            self.contract_amount,
            self.contract_awardee,
            self.contract_awarder,
            self.etip_event,
            self.funding_amount,
            self.funding_recipient,
            self.funding_source,
            self.outcome,
            self.over_time,
            self.overall_project_value,
            self.project_location,
            self.project_name,
            self.project_type,
            self.signatories,
            self.when,
        ]

        for field in fields:
            field.no_validation = self.no_validation
            if field.value and field.type_ == list:
                for prop in field.value:
                    prop.set_validation(validation)


    def to_dict(self):
        """Constructs a Python dictionary following the BP schema, suitable for JSON serialization.

        Returns:
            dict: a dictionary representing this BP Span.
        """

        if not self.no_validation:
            assert self.is_valid(), "Span object is invalid."

        dict_ = super().to_dict()

        for field in [
            self.agreement_length,
            self.contract_amount,
            self.contract_awardee,
            self.contract_awarder,
            self.etip_event,
            self.funding_amount,
            self.funding_recipient,
            self.funding_source,
            self.outcome,
            self.over_time,
            self.overall_project_value,
            self.project_location,
            self.project_name,
            self.project_type,
            self.signatories,
            self.when,
        ]:
            if field.value:
                if field.type_ == list:
                    dict_[field.name] = [temp_field.to_dict() for temp_field in field.value]
                else:
                    dict_[field.name] = field.value

        return dict_


class BPDocumentGranularProtestTemplate(BPDocumentGranularTemplate):
    """Represents a Granular Protest Template object in a BETTER Project document.

    Args:
        template_id (str): a unique ID for this event.
        anchor (str): an SSID referring to the MAT Annotation used to create this template.
            The SpanSet should only have one span in it.
        arrested (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        imprisoned (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        killed (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        occupy (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        over_time (bool): whether this protest ocurred over a length of time or not.
        organizer (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        outcome (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        outcome_averted (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        outcome_occurred (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        outcome_hypothetical (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        protest_against (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        protest_event (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        protest_for (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        when (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        where (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        who (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        wounded (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        provenance (BPDocumentProvenance): a Provenance object pretaining to this event.
        no_validation (bool): If True, skip validation when asked to. Defaults to False.
    """

    def __init__(
        self,
        template_id,
        anchor,
        arrested=None,
        imprisoned=None,
        killed=None,
        occupy=None,
        over_time=None,
        organizer=None,
        outcome=None,
        outcome_averted=None,
        outcome_occurred=None,
        outcome_hypothetical=None,
        protest_against=None,
        protest_event=None,
        protest_for=None,
        when=None,
        where=None,
        who=None,
        wounded=None,
        provenance=None,
        no_validation=False
    ):

        super().__init__(template_id, "Protestplate", anchor, provenance, no_validation)

        self.arrested = BPProperty(
            "arrested",
            False,
            value=arrested,
            type_=list,
            no_validation=no_validation
        )
        self.imprisoned = BPProperty(
            "imprisoned",
            False,
            value=imprisoned,
            type_=list,
            no_validation=no_validation
        )
        self.killed = BPProperty(
            "killed",
            False,
            value=killed,
            type_=list,
            no_validation=no_validation
        )
        self.occupy = BPProperty(
            "occupy",
            False,
            value=occupy,
            type_=list,
            no_validation=no_validation
        )
        self.over_time = BPProperty(
            "over-time",
            False,
            type_=bool,
            value=over_time,
            no_validation=no_validation
        )
        self.organizer = BPProperty(
            "organizer",
            False,
            value=organizer,
            type_=list,
            no_validation=no_validation
        )
        self.outcome = BPProperty(
            "outcome",
            False,
            value=outcome,
            type_=list,
            no_validation=no_validation
        )
        self.outcome_averted = BPProperty(
            "outcome-averted",
            False,
            value=outcome_averted,
            type_=list,
            no_validation=no_validation
        )
        self.outcome_occurred = BPProperty(
            "outcome-occurred",
            False,
            value=outcome_occurred,
            type_=list,
            no_validation=no_validation
        )
        self.outcome_hypothetical = BPProperty(
            "outcome-hypothetical",
            False,
            value=outcome_hypothetical,
            type_=list,
            no_validation=no_validation
        )
        self.protest_against = BPProperty(
            "protest-against",
            False,
            value=protest_against,
            type_=list,
            no_validation=no_validation
        )
        self.protest_event = BPProperty(
            "protest-event",
            False,
            value=protest_event,
            type_=list,
            no_validation=no_validation
        )
        self.protest_for = BPProperty(
            "protest-for",
            False,
            value=protest_for,
            type_=list,
            no_validation=no_validation
        )
        self.when = BPProperty(
            "when",
            False,
            value=when,
            type_=list,
            no_validation=no_validation
        )
        self.where = BPProperty(
            "where",
            False,
            value=where,
            type_=list,
            no_validation=no_validation
        )
        self.who = BPProperty(
            "who",
            False,
            value=who,
            type_=list,
            no_validation=no_validation
        )
        self.wounded = BPProperty(
            "wounded",
            False,
            value=wounded,
            type_=list,
            no_validation=no_validation
        )

    def __repr__(self):
        return (
            f"BPDocumentGranularProtestTemplate({self.template_id.value}, "
            f"{self.anchor.value}, "
            f"{self.arrested.value}, "
            f"{self.imprisoned.value}, "
            f"{self.killed.value}, "
            f"{self.occupy.value}, "
            f"{self.over_time.value}, "
            f"{self.organizer.value}, "
            f"{self.outcome.value}, "
            f"{self.outcome_averted.value}, "
            f"{self.outcome_occurred.value}, "
            f"{self.outcome_hypothetical.value}, "
            f"{self.protest_against.value}, "
            f"{self.protest_event.value}, "
            f"{self.protest_for.value}, "
            f"{self.when.value}, "
            f"{self.where.value}, "
            f"{self.who.value}, "
            f"{self.wounded.value}, "
            f"{repr(self.provenance.value)}, {self.no_validation})"
        )


    def __str__(self):
        return f"GranProtestplate<{self.template_id.value}>"


    def is_valid(self):
        """Checks if this object (and its constituents) are all valid."""
        global LOGGER

        if self.no_validation:
            raise RuntimeError(f"Validation set to False: {self}")

        base = super().is_valid()
        if not base:
            return False

        for prop in [
            self.arrested,
            self.imprisoned,
            self.killed,
            self.occupy,
            self.over_time,
            self.organizer,
            self.outcome,
            self.outcome_averted,
            self.outcome_occurred,
            self.outcome_hypothetical,
            self.protest_against,
            self.protest_event,
            self.protest_for,
            self.when,
            self.where,
            self.who,
            self.wounded,
            self.provenance,
        ]:
            if not prop.is_valid():
                LOGGER.error(f"BPDocumentGranularProtestTemplate property {prop} is invalid.")
                return False

        return True


    def set_validation(self, validation):
        """Sets the validation mode for this node and all children of this node.

        Args:
            validation (bool): if True, sets ``no_validation`` for this node and all of its children
                to ``False``. Otherwise, all nodes set ``no_validation`` to ``True``.
        """

        assert type(validation) == bool

        super().set_validation(validation)

        self.arrested.no_validation = self.no_validation
        self.imprisoned.no_validation = self.no_validation
        self.killed.no_validation = self.no_validation
        self.occupy.no_validation = self.no_validation
        self.over_time.no_validation = self.no_validation
        self.organizer.no_validation = self.no_validation
        self.outcome.no_validation = self.no_validation
        self.outcome_averted.no_validation = self.no_validation
        self.outcome_occurred.no_validation = self.no_validation
        self.outcome_hypothetical.no_validation = self.no_validation
        self.protest_against.no_validation = self.no_validation
        self.protest_event.no_validation = self.no_validation
        self.protest_for.no_validation = self.no_validation
        self.when.no_validation = self.no_validation
        self.where.no_validation = self.no_validation
        self.who.no_validation = self.no_validation
        self.wounded.no_validation = self.no_validation

        ## recurse

        if self.arrested.value:
            for prop in self.arrested.value:
                prop.set_validation(validation)
        if self.imprisoned.value:
            for prop in self.imprisoned.value:
                prop.set_validation(validation)
        if self.killed.value:
            for prop in self.killed.value:
                prop.set_validation(validation)
        if self.occupy.value:
            for prop in self.occupy.value:
                prop.set_validation(validation)
        if self.organizer.value:
            for prop in self.organizer.value:
                prop.set_validation(validation)
        if self.outcome.value:
            for prop in self.outcome.value:
                prop.set_validation(validation)
        if self.outcome_averted.value:
            for prop in self.outcome_averted.value:
                prop.set_validation(validation)
        if self.outcome_occurred.value:
            for prop in self.outcome_occurred.value:
                prop.set_validation(validation)
        if self.outcome_hypothetical.value:
            for prop in self.outcome_hypothetical.value:
                prop.set_validation(validation)
        if self.protest_against.value:
            for prop in self.protest_against.value:
                prop.set_validation(validation)
        if self.protest_event.value:
            for prop in self.protest_event.value:
                prop.set_validation(validation)
        if self.protest_for.value:
            for prop in self.protest_for.value:
                prop.set_validation(validation)
        if self.when.value:
            for prop in self.when.value:
                prop.set_validation(validation)
        if self.where.value:
            for prop in self.where.value:
                prop.set_validation(validation)
        if self.who.value:
            for prop in self.who.value:
                prop.set_validation(validation)
        if self.wounded.value:
            for prop in self.wounded.value:
                prop.set_validation(validation)


    def to_dict(self):
        """Constructs a Python dictionary following the BP schema, suitable for JSON serialization.

        Returns:
            dict: a dictionary representing this BP Span.
        """

        if not self.no_validation:
            assert self.is_valid(), "Span object is invalid."

        dict_ = super().to_dict()

        for prop in [
            self.arrested,
            self.imprisoned,
            self.killed,
            self.occupy,
            self.over_time,
            self.organizer,
            self.outcome,
            self.outcome_averted,
            self.outcome_occurred,
            self.outcome_hypothetical,
            self.protest_against,
            self.protest_event,
            self.protest_for,
            self.when,
            self.where,
            self.who,
            self.wounded,
        ]:
            if prop.value:
                if prop.type_ == list:
                    dict_[prop.name] = [temp_prop.to_dict() for temp_prop in prop.value]
                else:
                    dict_[prop.name] = prop.value

        return dict_


class BPDocumentGranularTerrorismTemplate(BPDocumentGranularTemplate):
    """Represents a Granular Terrorism Template object in a BETTER Project document.

    Args:
        template_id (str): a unique ID for this event.
        anchor (str): an SSID referring to the MAT Annotation used to create this template.
            The SpanSet should only have one span in it.
        blamed_by (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        claimed_by (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        completion (str): one of ``["planned", "thwarted", "failed", "successful",]``.
        coordinated (bool): whether this act of terrorism was coordinated or spurrious.
        killed (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        kidnapped (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        named_perp (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        named_perp_org (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        named_organizer (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        over_time (bool): whether this act of terrorism ocurred over a length of time or not.
        outcome (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        outcome_averted (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        outcome_occurred (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        outcome_hypothetical (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        perp_captured (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        perp_killed (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        perp_objective (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        perp_wounded (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        target_human (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        target_physical (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        terror_event (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        terrorism_type (str): one of
            ``["arson", "assault", "bombing", "kidnapping", "murder", "unspecified"]``.
        weapon (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        when (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        where (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        wounded (list): a list of :class:`BPDocumentGranularTemplateProperty` objects.
        provenance (BPDocumentProvenance): a Provenance object pretaining to this event.
        no_validation (bool): If True, skip validation when asked to. Defaults to False.
    """

    def __init__(
        self,
        template_id,
        anchor,
        blamed_by=None,
        claimed_by=None,
        completion=None,
        coordinated=None,
        killed=None,
        kidnapped=None,
        named_perp=None,
        named_perp_org=None,
        named_organizer=None,
        over_time=None,
        outcome=None,
        outcome_averted=None,
        outcome_occurred=None,
        outcome_hypothetical=None,
        perp_captured=None,
        perp_killed=None,
        perp_objective=None,
        perp_wounded=None,
        target_human=None,
        target_physical=None,
        terror_event=None,
        terrorism_type=None,
        weapon=None,
        when=None,
        where=None,
        wounded=None,
        provenance=None,
        no_validation=False
    ):

        super().__init__(template_id, "Terrorplate", anchor, provenance, no_validation)

        self.blamed_by = BPProperty(
            "blamed-by",
            False,
            value=blamed_by,
            type_=list,
            no_validation=no_validation
        )
        self.claimed_by = BPProperty(
            "claimed-by",
            False,
            value=claimed_by,
            type_=list,
            no_validation=no_validation
        )
        self.completion = BPProperty(
            "completion",
            False,
            value=completion,
            type_=str,
            default="successful",
            choices=["planned", "thwarted", "failed", "successful",],
            no_validation=no_validation
        )
        self.coordinated = BPProperty(
            "coordinated",
            False,
            value=coordinated,
            type_=bool,
            no_validation=no_validation
        )
        self.killed = BPProperty(
            "killed",
            False,
            value=killed,
            type_=list,
            no_validation=no_validation
        )
        self.kidnapped = BPProperty(
            "kidnapped",
            False,
            value=kidnapped,
            type_=list,
            no_validation=no_validation
        )
        self.named_perp = BPProperty(
            "named-perp",
            False,
            value=named_perp,
            type_=list,
            no_validation=no_validation
        )
        self.named_perp_org = BPProperty(
            "named-perp-org",
            False,
            value=named_perp_org,
            type_=list,
            no_validation=no_validation
        )
        self.named_organizer = BPProperty(
            "named-organizer",
            False,
            value=named_organizer,
            type_=list,
            no_validation=no_validation
        )
        self.over_time = BPProperty(
            "over-time",
            False,
            value=over_time,
            type_=bool,
            no_validation=no_validation
        )
        self.outcome = BPProperty(
            "outcome",
            False,
            value=outcome,
            type_=list,
            no_validation=no_validation
        )
        self.outcome_averted = BPProperty(
            "outcome-averted",
            False,
            value=outcome_averted,
            type_=list,
            no_validation=no_validation
        )
        self.outcome_occurred = BPProperty(
            "outcome-occurred",
            False,
            value=outcome_occurred,
            type_=list,
            no_validation=no_validation
        )
        self.outcome_hypothetical = BPProperty(
            "outcome-hypothetical",
            False,
            value=outcome_hypothetical,
            type_=list,
            no_validation=no_validation
        )
        self.perp_captured = BPProperty(
            "perp-captured",
            False,
            value=perp_captured,
            type_=list,
            no_validation=no_validation
        )
        self.perp_killed = BPProperty(
            "perp-killed",
            False,
            value=perp_killed,
            type_=list,
            no_validation=no_validation
        )
        self.perp_objective = BPProperty(
            "perp-objective",
            False,
            value=perp_objective,
            type_=list,
            no_validation=no_validation
        )
        self.perp_wounded = BPProperty(
            "perp-wounded",
            False,
            value=perp_wounded,
            type_=list,
            no_validation=no_validation
        )
        self.target_human = BPProperty(
            "target-human",
            False,
            value=target_human,
            type_=list,
            no_validation=no_validation
        )
        self.target_physical = BPProperty(
            "target-physical",
            False,
            value=target_physical,
            type_=list,
            no_validation=no_validation
        )
        self.terror_event = BPProperty(
            "terror-event",
            False,
            value=terror_event,
            type_=list,
            no_validation=no_validation
        )
        self.terrorism_type = BPProperty(
            "type",
            False,
            value=terrorism_type,
            type_=str,
            default="unspecified",
            choices=["arson", "assault", "bombing", "kidnapping", "murder", "unspecified"],
            no_validation=no_validation
        )
        self.weapon = BPProperty(
            "weapon",
            False,
            value=weapon,
            type_=list,
            no_validation=no_validation
        )
        self.when = BPProperty(
            "when",
            False,
            value=when,
            type_=list,
            no_validation=no_validation
        )
        self.where = BPProperty(
            "where",
            False,
            value=where,
            type_=list,
            no_validation=no_validation
        )
        self.wounded = BPProperty(
            "wounded",
            False,
            value=wounded,
            type_=list,
            no_validation=no_validation
        )

    def __repr__(self):
        return (
            f"BPDocumentGranularTerrorismTemplate({self.template_id.value}, "
            f"{self.anchor.value}, "
            f"{self.blamed_by.value}, "
            f"{self.claimed_by.value}, "
            f"{self.completion.value}, "
            f"{self.coordinated.value}, "
            f"{self.killed.value}, "
            f"{self.kidnapped.value}, "
            f"{self.named_perp.value}, "
            f"{self.named_perp_org.value}, "
            f"{self.named_organizer.value}, "
            f"{self.over_time.value}, "
            f"{self.outcome.value}, "
            f"{self.outcome_averted.value}, "
            f"{self.outcome_occurred.value}, "
            f"{self.outcome_hypothetical.value}, "
            f"{self.perp_captured.value}, "
            f"{self.perp_killed.value}, "
            f"{self.perp_objective.value}, "
            f"{self.perp_wounded.value}, "
            f"{self.target_human.value}, "
            f"{self.target_physical.value}, "
            f"{self.terror_event.value}, "
            f"{self.terrorism_type.value}, "
            f"{self.weapon.value}, "
            f"{self.when.value}, "
            f"{self.where.value}, "
            f"{self.wounded.value}, "
            f"{repr(self.provenance.value)}, {self.no_validation})"
        )


    def __str__(self):
        return f"GranTerrorplate<{self.template_id.value}>"


    def is_valid(self):
        """Checks if this object (and its constituents) are all valid."""
        global LOGGER

        if self.no_validation:
            raise RuntimeError(f"Validation set to False: {self}")

        base = super().is_valid()
        if not base:
            return False

        for prop in [
            self.blamed_by,
            self.claimed_by,
            self.completion,
            self.coordinated,
            self.killed,
            self.kidnapped,
            self.named_perp,
            self.named_perp_org,
            self.named_organizer,
            self.over_time,
            self.outcome,
            self.outcome_averted,
            self.outcome_occurred,
            self.outcome_hypothetical,
            self.perp_captured,
            self.perp_killed,
            self.perp_objective,
            self.perp_wounded,
            self.target_human,
            self.target_physical,
            self.terror_event,
            self.terrorism_type,
            self.weapon,
            self.when,
            self.where,
            self.wounded,
            self.provenance,
        ]:
            if not prop.is_valid():
                LOGGER.error(f"BPDocumentGranularTerrorismTemplate property {prop} is invalid.")
                return False

        return True


    def set_validation(self, validation):
        """Sets the validation mode for this node and all children of this node.

        Args:
            validation (bool): if True, sets ``no_validation`` for this node and all of its children
                to ``False``. Otherwise, all nodes set ``no_validation`` to ``True``.
        """

        assert type(validation) == bool

        super().set_validation(validation)

        self.blamed_by.no_validation = self.no_validation
        self.claimed_by.no_validation = self.no_validation
        self.completion.no_validation = self.no_validation
        self.coordinated.no_validation = self.no_validation
        self.killed.no_validation = self.no_validation
        self.kidnapped.no_validation = self.no_validation
        self.named_perp.no_validation = self.no_validation
        self.named_perp_org.no_validation = self.no_validation
        self.named_organizer.no_validation = self.no_validation
        self.over_time.no_validation = self.no_validation
        self.outcome.no_validation = self.no_validation
        self.outcome_averted.no_validation = self.no_validation
        self.outcome_occurred.no_validation = self.no_validation
        self.outcome_hypothetical.no_validation = self.no_validation
        self.perp_captured.no_validation = self.no_validation
        self.perp_killed.no_validation = self.no_validation
        self.perp_objective.no_validation = self.no_validation
        self.perp_wounded.no_validation = self.no_validation
        self.target_human.no_validation = self.no_validation
        self.target_physical.no_validation = self.no_validation
        self.terror_event.no_validation = self.no_validation
        self.terrorism_type.no_validation = self.no_validation
        self.weapon.no_validation = self.no_validation
        self.when.no_validation = self.no_validation
        self.where.no_validation = self.no_validation
        self.wounded.no_validation = self.no_validation

        ## recurse

        if self.blamed_by.value:
            for prop in self.blamed_by.value:
                prop.set_validation(validation)
        if self.claimed_by.value:
            for prop in self.claimed_by.value:
                prop.set_validation(validation)
        if self.killed.value:
            for prop in self.killed.value:
                prop.set_validation(validation)
        if self.kidnapped.value:
            for prop in self.kidnapped.value:
                prop.set_validation(validation)
        if self.named_perp.value:
            for prop in self.named_perp.value:
                prop.set_validation(validation)
        if self.named_perp_org.value:
            for prop in self.named_perp_org.value:
                prop.set_validation(validation)
        if self.named_organizer.value:
            for prop in self.named_organizer.value:
                prop.set_validation(validation)
        if self.outcome.value:
            for prop in self.outcome.value:
                prop.set_validation(validation)
        if self.outcome_averted.value:
            for prop in self.outcome_averted.value:
                prop.set_validation(validation)
        if self.outcome_occurred.value:
            for prop in self.outcome_occurred.value:
                prop.set_validation(validation)
        if self.outcome_hypothetical.value:
            for prop in self.outcome_hypothetical.value:
                prop.set_validation(validation)
        if self.perp_captured.value:
            for prop in self.perp_captured.value:
                prop.set_validation(validation)
        if self.perp_killed.value:
            for prop in self.perp_killed.value:
                prop.set_validation(validation)
        if self.perp_objective.value:
            for prop in self.perp_objective.value:
                prop.set_validation(validation)
        if self.perp_wounded.value:
            for prop in self.perp_wounded.value:
                prop.set_validation(validation)
        if self.target_human.value:
            for prop in self.target_human.value:
                prop.set_validation(validation)
        if self.target_physical.value:
            for prop in self.target_physical.value:
                prop.set_validation(validation)
        if self.terror_event.value:
            for prop in self.terror_event.value:
                prop.set_validation(validation)
        if self.weapon.value:
            for prop in self.weapon.value:
                prop.set_validation(validation)
        if self.when.value:
            for prop in self.when.value:
                prop.set_validation(validation)
        if self.where.value:
            for prop in self.where.value:
                prop.set_validation(validation)
        if self.wounded.value:
            for prop in self.wounded.value:
                prop.set_validation(validation)


    def to_dict(self):
        """Constructs a Python dictionary following the BP schema, suitable for JSON serialization.

        Returns:
            dict: a dictionary representing this BP Span.
        """

        if not self.no_validation:
            assert self.is_valid(), "Span object is invalid."

        dict_ = super().to_dict()

        for prop in [
            self.blamed_by,
            self.claimed_by,
            self.completion,
            self.coordinated,
            self.killed,
            self.kidnapped,
            self.named_perp,
            self.named_perp_org,
            self.named_organizer,
            self.over_time,
            self.outcome,
            self.outcome_averted,
            self.outcome_occurred,
            self.outcome_hypothetical,
            self.perp_captured,
            self.perp_killed,
            self.perp_objective,
            self.perp_wounded,
            self.target_human,
            self.target_physical,
            self.terror_event,
            self.terrorism_type,
            self.weapon,
            self.when,
            self.where,
            self.wounded,
            self.wounded,
        ]:
            if prop.value:
                if prop.type_ == list:
                    dict_[prop.name] = [temp_prop.to_dict() for temp_prop in prop.value]
                else:
                    dict_[prop.name] = prop.value

        return dict_


class BPDocumentAbstractEvent(BPNode):
    """Represents an Abstract Event object in a BETTER Project document.

    Args:
        event_id (str): a unique ID for this event.
        anchors (str): a single SSID, referring to a SpanSet.
        agents (list): a list of SSIDs, each referring to a SpanSet. The list itself is required,
            but it may be empty.
        patients (list): a list of SSIDs, each referring to a SpanSet. The list itself is required,
            but it may be empty.
        helpful_harmful (str): part of this event's QuadClass, one of ``"helpful"``, ``"harmful"``,
            or ``"neutral"``.
        material_verbal (str): part of this event's QuadClass, one of ``"material"``, ``"verbal"``,
            ``"both"``, or ``"unk"``.
        noquad (bool): if True, this AbstractEvent has no QuadClass information.
        provenance (BPDocumentProvenance): a :class:`BPDocumentProvenance` object pretaining to this
            event.
        no_validation (bool): If True, skip validation when asked to. Defaults to False.
    """

    def __init__(
        self,
        event_id,
        anchors,
        agents,
        patients,
        helpful_harmful,
        material_verbal,
        noquad=False,
        provenance=None,
        no_validation=False
    ):
        self.event_id = BPProperty(
            "eventid",
            True,
            value=event_id,
            type_=str,
            no_validation=no_validation
        )
        self.anchors = BPProperty(
            "anchors",
            True,
            value=anchors,
            type_=str,
            no_validation=no_validation
        )
        self.agents = BPProperty(
            "agents",
            True,
            value=agents,
            type_=list,
            no_validation=no_validation
        )
        self.patients = BPProperty(
            "patients",
            True,
            value=patients,
            type_=list,
            no_validation=no_validation
        )
        self.hh = self.helpful_harmful = BPProperty(
            "helpful-harmful",
            not noquad,
            value=helpful_harmful,
            type_=str,
            choices=["helpful", "harmful", "neutral",],
            no_validation=no_validation
        )
        self.mv = self.material_verbal = BPProperty(
            "material-verbal",
            not noquad,
            value=material_verbal,
            type_=str,
            choices=["material", "verbal", "both", "unk",],
            no_validation=no_validation
        )
        self.noquad = BPProperty(
            "noquad",
            False,
            value=noquad,
            type_=bool,
            no_validation=no_validation
        )
        self.provenance = BPProperty(
            "provenance",
            False,
            value=provenance,
            type_=BPDocumentProvenance,
            no_validation=no_validation
        )
        self.no_validation = no_validation


    def __lt__(self, other):
        return (self.event_id, self.anchors, self.agents, self.patients, self.hh, self.mv) < \
            (other.event_id, other.anchors, other.agents, other.patients, other.hh, other.mv)


    def __repr__(self):
        return (
            f"BPDocumentAbstractEvent({self.event_id.value}, {self.anchors.value}, "
            f"{self.agents.value}, {self.patients.value}, {self.helpful_harmful},"
            f"{self.material_verbal}, noquad={self.noquad.value}, "
            f"provenance={repr(self.provenance.value)} no_validation={self.no_validation})"
        )


    def __str__(self):
        return f"BPDocumentAbstractEvent<{self.event_id.value} ({id(self)})>"


    def deidentify(self):
        """Obfuscates this BPNode and its children."""

        self.provenance.value = None


    @classmethod
    def from_dict(klass, dict_, no_validation=False):
        """Constructs an instance of BPDocumentAbstractEvent from a Python dictionary following the
        BP schema for Abstract Event objects.

        Args:
            dict_ (dict): a Python dictionary following the BP schema for Abstract Event objects.
            no_validation (bool): If True, construct the object with validation turned off.
                Defaults to False.

        Returns:
            BPDocumentAbstractEvent: a new instance of this class built with the dictionary.

        .. Strategy:
           - primitives can be directly looked up and passed in from ``dict_``
           - collections (dicts, lists) and non-primitive types (like other ``BPDocument<type>``
             types defined in this library) must be constructed first if data is present in
             ``dict_``.
        """

        global LOGGER

        required = [
            "eventid", "agents", "patients", "anchors",
        ]
        if "noquad" in dict_ and dict_["noquad"] == True:
            required.extend(["helpful-harmful", "material-verbal",])
        for key in required:
            assert key in dict_, f"Abstract Event object dictionary is missing the '{key}' key."

        ## build non-primitive types and collections

        agents = dict_['agents']
        patients = dict_['patients']
        provenance = None
        if 'provenance' in dict_:
            provenance = BPDocumentProvenance.from_dict(dict_['provenance'], no_validation)

        ## build new object

        obj = klass(
            dict_['eventid'],
            dict_['anchors'],
            agents,
            patients,
            dict_.get('helpful-harmful'),
            dict_.get('material-verbal'),
            dict_.get("noquad"),
            provenance=provenance,
            no_validation=no_validation
        )

        if not no_validation:
            assert obj.is_valid(), "Abstract Event object built from dictionary failed validation."

        return obj


    def has_duplicate_args(self, debug=False):
        """Checks if this event has duplicate values in its arguments."""

        from collections import Counter

        dupl = False

        counter = Counter(self.agents.value)
        for values in counter.values():
            if values > 1:
                if debug:
                    LOGGER.debug((
                        "BPDocumentAbstractEvent agents list has duplicate element: "
                        f"{self.agents.value}"
                    ))
                dupl = True

        counter = Counter(self.patients.value)
        for values in counter.values():
            if values > 1:
                if debug:
                    LOGGER.debug((
                        "BPDocumentAbstractEvent patients list has duplicate element: "
                        f"{self.patients.value}"
                    ))
                dupl = True

        return dupl


    def is_valid(self):
        """Checks if this object (and its constituents) are all valid."""

        global LOGGER

        if self.no_validation:
            raise RuntimeError(f"Validation set to False: {self}")

        for prop in [
            self.event_id,
            self.anchors,
            self.agents,
            self.patients,
            self.helpful_harmful,
            self.material_verbal,
            self.noquad,
            self.provenance,
        ]:
            if not prop.is_valid():
                LOGGER.error(f"BPDocumentAbstractEvent property {prop} is invalid.")
                return False

        for ssid in self.agents.value:
            if type(ssid) is not str:
                i = self.agents.value.index(ssid)
                LOGGER.error(
                    f"BPDocumentAbstractEvent agent ssid at index {i} is not of type string: {ssid}"
                )
                return False

        for ssid in self.patients.value:
            if type(ssid) is not str:
                i = self.patients.value.index(ssid)
                LOGGER.error((
                    f"BPDocumentAbstractEvent patient ssid at index {i} is not of type string: "
                    f"{ssid}"
                ))
                return False

        return True


    def set_validation(self, validation):
        """Sets the validation mode for this node and all children of this node.

        Args:
            validation (bool): if True, sets ``no_validation`` for this node and all of its children
                to ``False``. Otherwise, all nodes set ``no_validation`` to ``True``.
        """

        assert type(validation) == bool

        self.no_validation = not validation

        self.event_id.no_validation = self.no_validation
        self.anchors.no_validation = self.no_validation
        self.agents.no_validation = self.no_validation
        self.patients.no_validation = self.no_validation
        self.hh.no_validation = self.no_validation
        self.mv.no_validation = self.no_validation
        self.noquad.no_validation = self.no_validation
        self.provenance.no_validation = self.no_validation

        ## recurse

        if self.provenance.value:
            self.provenance.value.set_validation(validation)


    def to_dict(self):
        """Constructs a Python dictionary following the BP schema, suitable for JSON serialization.

        Returns:
            dict: a dictionary representing this BP Abstract Event.
        """

        if not self.no_validation:
            assert self.is_valid(), "Abstract Event object is invalid."

        dict_ = {}
        for prop in [
            self.event_id,
            self.anchors,
            self.agents,
            self.patients,
            self.helpful_harmful,
            self.material_verbal,
            self.noquad,
        ]:
            if prop.value is not None:
                dict_[prop.name] = prop.value

        if self.provenance.value is not None:
            dict_[self.provenance.name] = self.provenance.value.to_dict()

        return dict_


class BPDocumentBasicEvent(BPNode):
    """Represents a Basic Event object in a BETTER Project document.

    Args:
        event_id (str): a unique ID for this event.
        anchors (str): a single SSID, referencing a SpanSet.
        agents (list): a list of SSID strings, each referencing a SpanSet. The list itself is
            required, but it may be empty.
        patients (list): a list of SSID strings, each referencing a SpanSet. The list itself is
            required, but it may be empty.
        ref_events (list): a list of Event IDs, each referencing a Basic Event. The list itself is
            required, but it may be empty.
        event_type (str): one of lib.utils.BASIC_MAT_KEY_ANCHOR_TYPES.
        money (list): a list of SSID strings, each referencing a SpanSet. Optional.
        state_of_affairs (bool): whether or not this Event represents a state of affairs. Defaults
            to False.
        provenance (BPDocumentProvenance): a Provenance object pretaining to this event.
        event_template_filler_coref_events (list): the original list of equivalent_template_filler
            ssids. Each element is a string referring to an Event ID.
        no_validation (bool): ff True, skip validation when asked to. Defaults to False.
    """

    def __init__(
        self,
        event_id,
        anchors,
        agents,
        patients,
        ref_events,
        event_type,
        money=None,
        state_of_affairs=False,
        provenance=None,
        event_template_filler_coref_events=None,
        no_validation=False
    ):
        from lib.utils import BASIC_MAT_KEY_ANCHOR_TYPES
        from lib.utils import GRANULAR_MAT_IRREALIS_STATUSES

        self.event_id = BPProperty(
            "eventid",
            True,
            value=event_id,
            type_=str,
            no_validation=no_validation
        )
        self.anchors = BPProperty(
            "anchors",
            True,
            value=anchors,
            type_=str,
            no_validation=no_validation
        )
        self.agents = BPProperty(
            "agents",
            True,
            value=agents,
            type_=list,
            no_validation=no_validation
        )
        self.patients = BPProperty(
            "patients",
            True,
            value=patients,
            type_=list,
            no_validation=no_validation
        )
        self.ref_events = BPProperty(
            "ref-events",
            True,
            value=ref_events,
            type_=list,
            no_validation=no_validation
        )
        self.event_type = BPProperty(
            "event-type",
            True,
            value=event_type,
            type_=str,
            choices=BASIC_MAT_KEY_ANCHOR_TYPES,
            no_validation=no_validation
        )
        self.money = BPProperty(
            "money",
            False,
            value=money,
            type_=list,
            no_validation=no_validation
        )
        self.state_of_affairs = BPProperty(
            "state-of-affairs",
            False,
            value=state_of_affairs,
            type_=bool,
            no_validation=no_validation
        )
        self.provenance = BPProperty(
            "provenance",
            False,
            value=provenance,
            type_=BPDocumentProvenance,
            no_validation=no_validation
        )
        # we don't render these values to BP JSON, so they aren't wrapped in BPProperty objects
        self.event_template_filler_coref_events = event_template_filler_coref_events
        self.no_validation = no_validation


    def __lt__(self, other):
        return (
            self.event_id,
            self.anchors,
            self.agents,
            self.patients,
            self.ref_events,
            self.event_type,
            self.money
        ) < (
            other.event_id,
            other.anchors,
            other.agents,
            other.patients,
            other.ref_events,
            other.event_type,
            other.money
        )


    def __repr__(self):
        return (
            f"BPDocumentBasicEvent("
            f"'{self.event_id.value}', "
            f"{self.anchors.value}, "
            f"{self.agents.value}, "
            f"{self.patients.value}, "
            f"{self.ref_events.value}, "
            f"{self.event_type.value}, "
            f"{self.money.value}, "
            f"provenance={repr(self.provenance.value)}, "
            f"no_validation={self.no_validation})"
        )


    def __str__(self):
        return f"BPDocumentBasicEvent<{self.event_id.value} ({id(self)})>"


    def deidentify(self):
        """Obfuscates this BPNode and its children."""

        self.provenance.value = None


    @classmethod
    def from_dict(klass, dict_, no_validation=False):
        """Constructs an instance of BPDocumentBasicEvent from a Python dictionary following the BP
        schema for Basic Event objects.

        Args:
            dict_ (dict): a Python dictionary following the BP schema for Basic Event objects.
            no_validation (bool): If True, construct the object with validation turned off.
                Defaults to False.

        Returns:
            BPDocumentBasicEvent: a new instance of this class built with the dictionary.

        .. Strategy:
           - primitives can be directly looked up and passed in from ``dict_``
           - collections (dicts, lists) and non-primitive types (like other ``BPDocument<type>``
             types defined in this library) must be constructed first if data is present in
             ``dict_``.
        """

        global LOGGER

        required = ["eventid", "agents", "patients", "anchors", "ref-events", "event-type",]
        for key in required:
            assert key in dict_, f"Basic Event object dictionary is missing the '{key}' key."

        ## build non-primitive types and collections

        provenance = None
        if 'provenance' in dict_:
            provenance = BPDocumentProvenance.from_dict(dict_['provenance'], no_validation)

        ## build new object

        obj = klass(
            dict_['eventid'],
            dict_['anchors'],
            dict_['agents'],
            dict_['patients'],
            dict_['ref-events'],
            dict_['event-type'],
            money=dict_.get('money'),
            state_of_affairs=dict_.get('state-of-affairs'),
            provenance=provenance,
            no_validation=no_validation
        )

        if not no_validation:
            assert obj.is_valid(), "Basic Event object built from dictionary failed validation."

        return obj


    def has_duplicate_args(self, debug=False):
        """Checks if this event has duplicate values in its arguments."""

        from collections import Counter

        dupl = False

        counter = Counter(self.agents.value)
        for values in counter.values():
            if values > 1:
                if debug:
                    LOGGER.warning((
                        "BPDocumentBasicEvent agents list has duplicate element: "
                        f"{self.agents.value}"
                    ))
                dupl = True

        counter = Counter(self.patients.value)
        for values in counter.values():
            if values > 1:
                if debug:
                    LOGGER.warning((
                        "BPDocumentBasicEvent patients list has duplicate element: "
                        f"{self.patients.value}"
                    ))
                dupl = True

        counter = Counter(self.ref_events.value)
        for values in counter.values():
            if values > 1:
                if debug:
                    LOGGER.warning((
                        "BPDocumentBasicEvent ref_events list has duplicate element: "
                        f"{self.ref_events.value}"
                    ))
                dupl = True

        if self.money.value:
            counter = Counter(self.money.value)
            for values in counter.values():
                if values > 1:
                    if debug:
                        LOGGER.warning((
                            "BPDocumentBasicEvent money list has duplicate element: "
                            f"{self.money.value}"
                        ))
                    dupl = True

        return dupl


    def is_valid(self):
        """Checks if this object (and its constituents) are all valid."""

        global LOGGER

        if self.no_validation:
            raise RuntimeError(f"Validation set to False: {self}")

        for prop in [
            self.event_id,
            self.anchors,
            self.agents,
            self.patients,
            self.ref_events,
            self.event_type,
            self.money,
            self.state_of_affairs,
            self.provenance,
        ]:
            if not prop.is_valid():
                LOGGER.error(f"BPDocumentBasicEvent property {prop} is invalid.")
                return False

        for ssid in self.agents.value:
            if type(ssid) is not str:
                i = self.agents.value.index(ssid)
                LOGGER.error(
                    f"BPDocumentBasicEvent agent ssid at index {i} is not of type string: {ssid}"
                )
                return False

        for ssid in self.patients.value:
            if type(ssid) is not str:
                i = self.patients.value.index(ssid)
                LOGGER.error(
                    f"BPDocumentBasicEvent patient ssid at index {i} is not of type string: {ssid}"
                )
                return False

        for event in self.ref_events.value:
            if type(event) is not str:
                i = self.ref_events.value.index(event)
                LOGGER.error((
                    f"BPDocumentBasicEvent ref_event event at index {i} is not of type string: "
                    f"{event}"
                ))
                return False

        if self.money.value:
            for ssid in self.money.value:
                if type(ssid) is not str:
                    i = self.money.value.index(ssid)
                    LOGGER.error(
                        f"BPDocumentBasicEvent money ssid at index {i} is not of type string: "
                        f"{ssid}"
                    )
                    return False

        return True


    def set_validation(self, validation):
        """Sets the validation mode for this node and all children of this node.

        Args:
            validation (bool): if True, sets ``no_validation`` for this node and all of its children
                to ``False``. Otherwise, all nodes set ``no_validation`` to ``True``.
        """

        assert type(validation) == bool

        self.no_validation = not validation

        self.event_id.no_validation = self.no_validation
        self.anchors.no_validation = self.no_validation
        self.agents.no_validation = self.no_validation
        self.patients.no_validation = self.no_validation
        self.ref_events.no_validation = self.no_validation
        self.event_type.no_validation = self.no_validation
        self.money.no_validation = self.no_validation
        self.state_of_affairs.no_validation = self.no_validation
        self.provenance.no_validation = self.no_validation

        ## recurse

        if self.provenance.value:
            self.provenance.value.set_validation(validation)


    def to_dict(self):
        """Constructs a Python dictionary following the BP schema, suitable for JSON serialization.

        Returns:
            dict: a dictionary representing this BP Basic Event.
        """

        if not self.no_validation:
            assert self.is_valid(), "Basic Event object is invalid."

        dict_ = {}
        for prop in [
            self.event_id,
            self.anchors,
            self.agents,
            self.patients,
            self.ref_events,
            self.event_type,
            self.money,
            self.state_of_affairs,
        ]:
            if prop.value is not None:
                dict_[prop.name] = prop.value

        if self.provenance.value is not None:
            dict_[self.provenance.name] = self.provenance.value.to_dict()

        return dict_


class BPDocumentAbstractEventSet(BPNode):
    """Represents an Abstact Event Set object in a BETTER Project document.

    Args:
        events (dict): a dictionary keyed with Event IDs (strings) whose values are the
            corresponding :class:`BPDocumentAbstractEvent` objects.
        span_sets (dict): a dictionary keyed with SSIDs (strings) whose values are the corresponding
            class:`BPDocumentSpanSet` objects.
        includes_relations (dict): a dictionary keyed with SSIDs (strings) where each value is a
            list of SSIDs. This structure indicates that the SpanSets listed are "included" in (or
            coreferring to) the SpanSets specified by list elements.
        no_validation (bool): If True, skip validation when asked to. Defaults to False.
    """

    def __init__(
        self,
        events,
        span_sets,
        includes_relations=None,
        no_validation=False
    ):

        self.events = BPProperty(
            "events",
            True,
            value=events,
            type_=dict,
            no_validation=no_validation
        )
        self.span_sets = BPProperty(
            "span-sets",
            True,
            value=span_sets,
            type_=dict,
            no_validation=no_validation
        )
        self.includes_relations = BPProperty(
            "includes-relations",
            False,
            value=includes_relations,
            type_=dict,
            no_validation=no_validation
        )
        self.no_validation = no_validation


    def __lt__(self, other):
        return (self.events, self.span_sets) < (other.events, other.span_sets)


    def __repr__(self):
        return (
            f"BPDocumentAbstractEventSet({repr(self.events)}, {repr(self.span_sets)}), "
            f"includes_relations={repr(self.includes_relations)}, "
            f"no_validation={self.no_validation})"
        )


    def __str__(self):
        return f"BPDocumentAbstractEventSet<{id(self)}>"


    def deidentify(self):
        """Obfuscates this BPNode and its children."""

        if self.events.value:
            for evid in self.events.value:
                self.events.value[evid].deidentify()


    @classmethod
    def from_dict(klass, dict_, no_validation=False):
        """Constructs an instance of BPDocumentAbstractEventSet from a Python dictionary following
        the BP schema for Abstract Event Set objects.

        Args:
            dict_ (dict): a Python dictionary following the BP schema for Abstract Event Set
                objects.
            no_validation (bool): If True, construct the object with validation turned off.
                Defaults to False.

        Returns:
            BPDocumentAbstractEventSet: a new instance of this class built with the dictionary.

        .. Strategy:
           - primitives can be directly looked up and passed in from ``dict_``
           - collections (dicts, lists) and non-primitive types (like other ``BPDocument<type>``
             types defined in this library) must be constructed first if data is present in
             ``dict_``.
        """

        global LOGGER

        required = ["events", "span-sets",]
        for key in required:
            assert key in dict_, f"Abstract Event Set object dictionary is missing the '{key}' key."

        ## build non-primitive types and collections

        events = {}
        for event_id in dict_["events"]:
            events[event_id] = BPDocumentAbstractEvent.from_dict(
                dict_["events"][event_id],
                no_validation
            )

        span_sets = {}
        for ssid in dict_["span-sets"]:
            span_sets[ssid] = BPDocumentSpanSet.from_dict(dict_["span-sets"][ssid], no_validation)

        includes_relations = dict_.get("includes-relations", None)

        ## build new object

        obj = klass(
            events,
            span_sets,
            includes_relations=includes_relations,
            no_validation=no_validation
        )

        if not no_validation:
            assert obj.is_valid(), \
                "Abstract Event Set object built from dictionary failed validation."

        return obj


    def is_valid(self):
        """Checks if this object (and its constituents) are all valid."""

        global LOGGER

        if self.no_validation:
            raise RuntimeError(f"Validation set to False: {self}")

        for prop in [self.events, self.span_sets, self.includes_relations]:
            if not prop.is_valid():
                LOGGER.error(f"BPDocumentAbstractEventSet property {prop} is invalid.")
                return False

        for ssid in self.span_sets.value:

            if type(ssid) is not str:
                LOGGER.error(
                    f"BPDocumentAbstractEventSet ssid in span_sets is not of type string: {ssid}"
                )
                return False

            if type(self.span_sets.value[ssid]) is not BPDocumentSpanSet:
                LOGGER.error((
                    f"BPDocumentAbstractEventSet span_set with id {ssid} is not of type "
                    f"BPDocumentSpanSet: {type(self.span_sets.value[ssid])}"
                ))
                return False

        for event_id in self.events.value:

            if type(event_id) is not str:
                LOGGER.error(
                    "BPDocumentAbstractEventSet event_id in events is not of type string: "
                    f"{event_id}"
                )
                return False

            if type(self.events.value[event_id]) is not BPDocumentAbstractEvent:
                LOGGER.error((
                    f"BPDocumentAbstractEventSet event with id {event_id} is not of type "
                    f"BPDocumentAbstractEvent: {type(self.events.value[event_id])}"
                ))
                return False

            event = self.events.value[event_id]
            if event.anchors.value:
                if not self.span_sets.value:
                    LOGGER.error((
                        f"BPDocumentAbstractEventSet anchor populated in event with id {event_id} "
                        "when span sets is empty!"
                    ))
                    return False
                elif event.anchors.value not in self.span_sets.value:
                    LOGGER.error((
                        f"BPDocumentAbstractEventSet anchor {event.anchors.value} in event with id "
                        f"{event_id} is not found within its span sets!\nSSIDs: "
                        f"{self.span_sets.value.keys()}"
                    ))
                    return False

            if event.agents.value:
                if not self.span_sets.value:
                    LOGGER.error((
                        f"BPDocumentAbstractEventSet agents populated in event with id {event_id} "
                        "when span sets is empty!"
                    ))
                    return False

                for agent in event.agents.value:

                    if agent not in self.span_sets.value:
                        LOGGER.error((
                            f"BPDocumentAbstractEventSet agent {agent} in event with id {event_id} "
                            "is not found within its span sets!\n"
                            f"SSIDs: {self.span_sets.value.keys()}"
                        ))
                        return False

            if event.patients.value:
                if not self.span_sets.value:
                    LOGGER.error((
                        "BPDocumentAbstractEventSet patients populated in event with id "
                        f"{event_id} when span sets is empty!"
                    ))
                    return False

                for patient in event.patients.value:

                    if patient not in self.span_sets.value:
                        LOGGER.error((
                            f"BPDocumentAbstractEventSet patient {patient} in event with id "
                            f"{event_id} is not found within its span sets!\n"
                            f"SSIDs: {self.span_sets.value.keys()}"
                        ))
                        return False

        if self.includes_relations.value:
            for ssid in self.includes_relations.value:
                if type(ssid) is not str:
                    LOGGER.error(
                        "BPDocumentAbstractEventSet ssid in includes_relations is not of type "
                        f"string: {ssid}"
                    )
                    return False
                if type(self.includes_relations.value[ssid]) is not list:
                    LOGGER.error((
                        f"BPDocumentAbstractEventSet includes_relations list with id {ssid} is not "
                        f"of type list: {type(self.includes_relations.value[ssid])}"
                    ))
                    return False

                for inner_ssid in self.includes_relations.value[ssid]:
                    if type(inner_ssid) is not str:
                        LOGGER.error((
                            "BPDocumentAbstractEventSet includes_relations list element "
                            f"{inner_ssid} is not of type string: {type(inner_ssid)}"
                        ))
                        return False

        return True


    def set_validation(self, validation):
        """Sets the validation mode for this node and all children of this node.

        Args:
            validation (bool): if True, sets ``no_validation`` for this node and all of its children
                to ``False``. Otherwise, all nodes set ``no_validation`` to ``True``.
        """

        assert type(validation) == bool

        self.no_validation = not validation

        self.events.no_validation = self.no_validation
        self.span_sets.no_validation = self.no_validation
        self.includes_relations.no_validation = self.no_validation

        ## recurse

        if self.events.value:
            for evid in self.events.value:
                self.events.value[evid].set_validation(validation)

        if self.span_sets.value:
            for ssid in self.span_sets.value:
                self.span_sets.value[ssid].set_validation(validation)


    def to_dict(self):
        """Constructs a Python dictionary following the BP schema, suitable for JSON serialization.

        Returns:
            dict: a dictionary representing this BP Abstract Event Set.
        """

        if not self.no_validation:
            assert self.is_valid(), "Abstract Event Set object is invalid."

        dict_ = {
            self.events.name: {},
            self.span_sets.name: {},
        }

        for event_id in self.events.value:
            dict_[self.events.name][event_id] = self.events.value[event_id].to_dict()

        for ssid in self.span_sets.value:
            dict_[self.span_sets.name][ssid] = self.span_sets.value[ssid].to_dict()

        if self.includes_relations.value:
            dict_[self.includes_relations.name] = self.includes_relations.value

        return dict_


class BPDocumentBasicEventSet(BPNode):
    """Represents a Basic Event Set object in a BETTER Project document.

    Args:
        events (dict): a dictionary keyed with Event IDs (strings) whose values are
            corresponding :class:`BPDocumentBasicEvent` objects.
        span_sets (dict): a dictionary keyed with SSIDs (strings) whose values are corresponding
            :class:`BPDocumentSpanSet` objects.
        includes_relations (dict): a dictionary keyed with SSIDs (strings) where each value is a
            list of SSIDs. This structure indicates that the SpanSets listed are "included" in (or
            coreferring to) the SpanSets specified by list elements.
        granular_templates (dict): a dictionary keyed with Template IDs (strings) whose values are
            corresponding ``BPDocumentGranular<type>Template`` objects.
        template_filler_coref_events (list): a list of lists of strings. Each inner list represents
            a co-referential equivalence class. Each element of the inner list is an Event ID
            belonging to the equivalence class.
        no_validation (bool): If True, skip validation when asked to. Defaults to False.
    """

    def __init__(
        self,
        events,
        span_sets,
        includes_relations=None,
        granular_templates=None,
        template_filler_coref_events=None,
        no_validation=False
    ):

        self.events = BPProperty(
            "events",
            True,
            value=events,
            type_=dict,
            no_validation=no_validation
        )
        self.span_sets = BPProperty(
            "span-sets",
            True,
            value=span_sets,
            type_=dict,
            no_validation=no_validation
        )
        self.includes_relations = BPProperty(
            "includes-relations",
            False,
            value=includes_relations,
            type_=dict,
            no_validation=no_validation
        )
        self.granular_templates = BPProperty(
            "granular-templates",
            False,
            value=granular_templates,
            type_=dict,
            no_validation=no_validation
        )
        self.template_filler_coref_events=BPProperty(
            "template-filler-coref-events",
            False,
            value=template_filler_coref_events,
            type_=list,
            no_validation=no_validation
        )
        self.no_validation = no_validation


    def __lt__(self, other):
        return (self.events, self.span_sets, self.includes_relations, self.granular_templates) < \
            (other.events, other.span_sets, other.includes_relations, other.granular_templates)


    def __repr__(self):
        return (
            f"BPDocumentBasicEventSet({repr(self.events)}, {repr(self.span_sets)}), "
            f"includes_relations={repr(self.includes_relations)}, "
            f"granular_templates={repr(self.granular_templates)}, "
            f"template_filler_coref_events={repr(self.template_filler_coref_events)}, "
            f"no_validation={self.no_validation})"
        )


    def __str__(self):
        return f"BPDocumentBasicEventSet<{id(self)}>"


    def deidentify(self):
        """Obfuscates this BPNode and its children."""

        if self.events.value:
            for evid in self.events.value:
                self.events.value[evid].deidentify()

        if self.granular_templates.value:
            for gtid in self.granular_templates.value:
                self.granular_templates.value[gtid].deidentify()


    @classmethod
    def from_dict(klass, dict_, no_validation=False):
        """Constructs an instance of BPDocumentBasicEventSet from a Python dictionary following the
        BP schema for Basic Event Set objects.

        Args:
            dict_ (dict): a Python dictionary following the BP schema for Basic Event Set objects.
            no_validation (bool): If True, construct the object with validation turned off.
                Defaults to False.

        Returns:
            BPDocumentBasicEventSet: a new instance of this class built with the dictionary.

        .. Strategy:
           - primitives can be directly looked up and passed in from ``dict_``
           - collections (dicts, lists) and non-primitive types (like other ``BPDocument<type>``
             types defined in this library) must be constructed first if data is present in
             ``dict_``.
        """

        required = ["events", "span-sets",]
        for key in required:
            assert key in dict_, f"Basic Event Set object dictionary is missing the '{key}' key."

        ## build non-primitive types and collections

        events = {}
        for event_id in dict_["events"]:
            events[event_id] = BPDocumentBasicEvent.from_dict(
                dict_["events"][event_id],
                no_validation=no_validation
            )

        span_sets = {}
        for ssid in dict_["span-sets"]:
            span_sets[ssid] = BPDocumentSpanSet.from_dict(
                dict_["span-sets"][ssid],
                no_validation=no_validation
            )

        includes_relations = dict_.get("includes-relations", None)
        template_filler_coref_events = dict_.get("template-filler-coref-events", None)

        granular_templates = {}
        things = dict_.get("granular-templates")
        if things:
            for gtid in things:
                granular_templates[gtid] = BPDocumentGranularTemplate.from_dict(things[gtid])
        else:
            granular_templates = None

        ## build new object

        obj = klass(
            events,
            span_sets,
            includes_relations=includes_relations,
            granular_templates=granular_templates,
            template_filler_coref_events=template_filler_coref_events,
            no_validation=no_validation
        )

        if not no_validation:
            assert obj.is_valid(), "Basic Event Set object built from dictionary failed validation."

        return obj


    def is_valid(self):
        """Checks if this object (and its constituents) are all valid."""

        global LOGGER

        if self.no_validation:
            raise RuntimeError(f"Validation set to False: {self}")

        for prop in [
            self.events,
            self.span_sets,
            self.includes_relations,
            self.granular_templates,
        ]:

            if not prop.is_valid():
                LOGGER.error(f"BPDocumentBasicEventSet property {prop} is invalid.")
                return False

        ## validate span sets

        for ssid in self.span_sets.value:

            if type(ssid) is not str:
                LOGGER.error(
                    f"BPDocumentBasicEventSet ssid in span_sets is not of type string: {ssid}"
                )
                return False

            if type(self.span_sets.value[ssid]) is not BPDocumentSpanSet:
                LOGGER.error((
                    f"BPDocumentBasicEventSet span_set with id {ssid} is not of type "
                    f"BPDocumentSpanSet: {type(self.span_sets.value[ssid])}"
                ))
                return False

        ## validate event set

        for event_id in self.events.value:

            if type(event_id) is not str:
                LOGGER.error(
                    f"BPDocumentBasicEventSet event_id in events is not of type string: {event_id}"
                )
                return False

            if type(self.events.value[event_id]) is not BPDocumentBasicEvent:
                LOGGER.error((
                    f"BPDocumentBasicEventSet event with id {event_id} is not of type "
                    f"BPDocumentBasicEvent: {type(self.events.value[event_id])}"
                ))
                return False

            event = self.events.value[event_id]
            if event.anchors.value:
                if not self.span_sets.value:
                    LOGGER.error((
                        f"BPDocumentBasicEventSet anchor populated in event with id {event_id} "
                        "when span sets is empty!"
                    ))
                    return False
                elif event.anchors.value not in self.span_sets.value:
                    LOGGER.error((
                        f"BPDocumentBasicEventSet anchor {event.anchors.value} in event with id "
                        f"{event_id} is not found within its span sets!\nSSIDs: "
                        f"{self.span_sets.value.keys()}"
                    ))
                    return False

            if event.agents.value:
                if not self.span_sets.value:
                    LOGGER.error((
                        f"BPDocumentBasicEventSet agents populated in event with id {event_id} "
                        "when span sets is empty!"
                    ))
                    return False

                for ssid in event.agents.value:

                    if ssid not in self.span_sets.value:
                        LOGGER.error((
                            f"BPDocumentBasicEventSet agent {ssid} in event with id {event_id} "
                            "is not found within its span sets!\n"
                            f"SSIDs: {self.span_sets.value.keys()}"
                        ))
                        return False

            if event.patients.value:
                if not self.span_sets.value:
                    LOGGER.error((
                        "BPDocumentBasicEventSet patients populated in event with id "
                        f"{event_id} when span sets is empty!"
                    ))
                    return False

                for ssid in event.patients.value:

                    if ssid not in self.span_sets.value:
                        LOGGER.error((
                            f"BPDocumentBasicEventSet patient {ssid} in event with id "
                            f"{event_id} is not found within its span sets!\n"
                            f"SSIDs: {self.span_sets.value.keys()}"
                        ))
                        return False

            if event.money.value:
                if not self.span_sets.value:
                    LOGGER.error((
                        "BPDocumentBasicEventSet money populated in event with id "
                        f"{event_id} when span sets is empty!"
                    ))
                    return False

                for ssid in event.money.value:

                    if ssid not in self.span_sets.value:
                        LOGGER.error((
                            f"BPDocumentBasicEventSet patient {ssid} in event with id "
                            f"{event_id} is not found within its span sets!\n"
                            f"SSIDs: {self.span_sets.value.keys()}"
                        ))
                        return False

            if event.ref_events.value:

                for evid in event.ref_events.value:

                    if evid not in self.events.value:

                        LOGGER.error((
                            f"BPDocumentBasicEventSet ref_event {evid} in event with id "
                            f"{event_id} is not found within Entry's event set!\n"
                            f"Event IDs: {self.events.value.keys()}"
                        ))
                        return False

        ## validate includes relations

        if self.includes_relations.value:

            for ssid in self.includes_relations.value:

                if type(ssid) is not str:
                    LOGGER.error(
                        "BPDocumentBasicEventSet ssid in includes_relations is not of type string: "
                        f"{ssid}"
                    )
                    return False

                if type(self.includes_relations.value[ssid]) is not list:
                    LOGGER.error((
                        f"BPDocumentBasicEventSet includes_relations list with id {ssid} is not of "
                        f"type list: {type(self.includes_relations.value[ssid])}"
                    ))
                    return False

                for inner_ssid in self.includes_relations.value[ssid]:
                    if type(inner_ssid) is not str:
                        LOGGER.error((
                            f"BPDocumentBasicEventSet includes_relations list element {inner_ssid} "
                            f"is not of type string: {type(inner_ssid)}"
                        ))
                        return False

        ## validate granular template set

        if self.granular_templates.value:
            for gtid in self.granular_templates.value:

                if type(gtid) is not str:
                    LOGGER.error(
                        "BPDocumentBasicEventSet gtid in granular_templates is not of type string: "
                        f"{gtid}"
                    )
                    return False

                if BPDocumentGranularTemplate not in type(
                    self.granular_templates.value[gtid]
                ).__bases__:
                    LOGGER.error((
                        f"BPDocumentBasicEventSet granular_template with gtid {gtid} is not of type"
                        f" BPDocumentGranularTemplate: {type(self.granular_templates.value[gtid])}"
                    ))
                    return False

                if not self.granular_templates.value[gtid].is_valid():
                    return False

        ## validate equivalent template fillers

        if self.template_filler_coref_events.value:
            for class_ in self.template_filler_coref_events.value:

                if type(class_) is not list:
                    LOGGER.error(
                        "BPDocumentAbstractEventSet equivalence class in "
                        f"template_filler_coref_events is not of type list: {class_}"
                    )

                for evid in class_:
                    if type(evid) is not str:
                        LOGGER.error((
                            "BPDocumentAbstractEventSet equivalence class list element "
                            f"{evid} is not of type string: {type(evid)}"
                        ))
                        return False

        return True


    def set_validation(self, validation):
        """Sets the validation mode for this node and all children of this node.

        Args:
            validation (bool): if True, sets ``no_validation`` for this node and all of its children
                to ``False``. Otherwise, all nodes set ``no_validation`` to ``True``.
        """

        assert type(validation) == bool

        self.no_validation = not validation

        self.events.no_validation = self.no_validation
        self.span_sets.no_validation = self.no_validation
        self.includes_relations.no_validation = self.no_validation
        self.granular_templates.no_validation = self.no_validation
        self.template_filler_coref_events.no_validation = self.no_validation

        ## recurse

        if self.events.value:
            for evid in self.events.value:
                self.events.value[evid].set_validation(validation)

        if self.span_sets.value:
            for ssid in self.span_sets.value:
                self.span_sets.value[ssid].set_validation(validation)

        if self.granular_templates.value:
            for gtid in self.granular_templates.value:
                self.granular_templates.value[gtid].set_validation(validation)


    def to_dict(self):
        """Constructs a Python dictionary following the BP schema, suitable for JSON serialization.

        Returns:
            dict: a dictionary representing this BP Basic Event Set.
        """

        if not self.no_validation:
            assert self.is_valid(), "Basic Event Set object is invalid."

        dict_ = {
            self.events.name: {},
            self.span_sets.name: {},
        }

        if self.granular_templates.value:
            dict_[self.granular_templates.name] = {}

        for collection in [
            self.events,
            self.span_sets,
            self.granular_templates,
        ]:
            if collection.value:
                for id_ in collection.value:
                    dict_[collection.name][id_] = collection.value[id_].to_dict()

        if self.includes_relations.value:
            dict_[self.includes_relations.name] = self.includes_relations.value

        if self.template_filler_coref_events.value:
            dict_[self.template_filler_coref_events.name] = \
                self.template_filler_coref_events.value

        return dict_


class BPDocumentAnnotationSet(BPNode):
    """Represents an Annotation Set object in a BETTER Project document.

    Args:
        abstract_events (BPDocumentAbstractEventSet): a collection of Abstract Events, SpanSets,
            and coreference information.
        basic_events (BPDocumentBasicEventSet): a collection of Basic Events, Granular Templates,
            SpanSets, and coreference information.
        no_validation (bool): If True, skip validation when asked to. Defaults to False.
    """

    def __init__(self, abstract_events=None, basic_events=None, no_validation=False):

        self.abstract_events = BPProperty(
            "abstract-events",
            False,
            value=abstract_events,
            type_=BPDocumentAbstractEventSet,
            no_validation=no_validation
        )
        self.basic_events = BPProperty(
            "basic-events",
            False,
            value=basic_events,
            type_=BPDocumentBasicEventSet,
            no_validation=no_validation
        )
        self.no_validation = no_validation


    def __lt__(self, other):
        return (self.abstract_events, self.basic_events) < \
            (other.abstract_events, other.basic_events)


    def __repr__(self):
        return (
            f"BPDocumentAnnotationSet(abstract_events={repr(self.abstract_events)}, "
            f"basic_events={repr(self.basic_events)}, no_validation={self.no_validation})"
        )


    def __str__(self):
        return f"BPDocumentAnnotationSet<{id(self)}>"


    def deidentify(self):
        """Obfuscates this BPNode and its children."""

        if self.abstract_events.value:
            self.abstract_events.value.deidentify()

        if self.basic_events.value:
            self.basic_events.value.deidentify()


    @classmethod
    def from_dict(klass, dict_, no_validation=False):
        """Constructs an instance of BPDocumentAnnotationSet from a Python dictionary following the
        BP schema for Annotation Set objects.

        Args:
            dict_ (dict): a Python dictionary following the BP schema for Annotation Set objects.
            no_validation (bool): If True, construct the object with validation turned off.
                Defaults to False.

        Returns:
            BPDocumentAnnotationSet: a new instance of this class built with the dictionary.

        .. Strategy:
           - primitives can be directly looked up and passed in from ``dict_``
           - collections (dicts, lists) and non-primitive types (like other ``BPDocument<type>``
             types defined in this library) must be constructed first if data is present in
             ``dict_``.
        """

        ## build non-primitive types and collections

        if "abstract-events" in dict_:
            abstract_events = BPDocumentAbstractEventSet.from_dict(
                dict_["abstract-events"],
                no_validation
            )
        else:
            abstract_events = None

        if "basic-events" in dict_:
            basic_events = BPDocumentBasicEventSet.from_dict(dict_["basic-events"], no_validation)
        else:
            basic_events = None

        ## build new object

        obj = klass(
            abstract_events=abstract_events,
            basic_events=basic_events,
            no_validation=no_validation
        )

        if not no_validation:
            assert obj.is_valid(), "Annotation Set object built from dictionary failed validation."

        return obj


    def is_valid(self):
        """Checks if this object (and its constituents) are all valid."""

        global LOGGER

        if self.no_validation:
            raise RuntimeError(f"Validation set to False: {self}")

        for prop in [self.abstract_events, self.basic_events,]:
            if not prop.is_valid():
                LOGGER.error(f"BPDocumentAnnotationSet property {prop} is invalid.")
                return False

            if prop.value:
                if not prop.value.is_valid():
                    return False

        return True


    def set_validation(self, validation):
        """Sets the validation mode for this node and all children of this node.

        Args:
            validation (bool): if True, sets ``no_validation`` for this node and all of its children
                to ``False``. Otherwise, all nodes set ``no_validation`` to ``True``.
        """

        assert type(validation) == bool

        self.no_validation = not validation

        self.abstract_events.no_validation = self.no_validation
        self.basic_events.no_validation = self.no_validation

        ## recurse

        if self.abstract_events.value:
            self.abstract_events.value.set_validation(validation)

        if self.basic_events.value:
            self.basic_events.value.set_validation(validation)


    def to_dict(self):
        """Constructs a Python dictionary following the BP schema, suitable for JSON serialization.

        Returns:
            dict: a dictionary representing this BP Annotation Set.
        """

        if not self.no_validation:
            assert self.is_valid(), "Annotation Set object is invalid."

        dict_ = {}

        if self.abstract_events.value is not None:
            dict_[self.abstract_events.name] = self.abstract_events.value.to_dict()

        if self.basic_events.value is not None:
            dict_[self.basic_events.name] = self.basic_events.value.to_dict()

        return dict_


class BPDocumentEntry(BPNode):
    """Represents an Entry in a BETTER Project document.

    Args:
        entry_id (str): the ID for this Entry.
        doc_id (str): the ID of the document this Entry belongs to.
        lang (str): the ISO-639-3 language code for this Entry.
        sent_id (str): the ID of the sentence in the document this Entry belongs to.
        segment_type (str): either ``"document"`` or ``"sentence"``.
        segment_text (str): the string literal of for this Entry.
        segment_sections (list): a list of :class:`BPDocumentSegmentSection` objects, each detailing
            a Segment Section within this Entry.
        start_in_source (int): the start offset of the string literal for this Entry, relative to
            the beginning of the document this entry came from. Used if the ``segment_type`` is
            ``"sentence"``.
        end_in_source (int): the end offset of the string literal for this Entry, relative to
            the beginning of the document this entry came from. Used if the ``segment_type`` is
            ``"sentence"``.
        segment_start (int): the start offset of the string literal for this Entry, usually ``0``.
        segment_end (int): the end offset of the string literal for this Entry, usually
            ``len(segment_text)``.
        annotation_sets (BPDocumentAnnotationSet): the annotations found within this Entry.
        provenance (BPDocumentProvenance): a :class:`BPDocumentProvenance` object pretaining to this
            Entry.
        no_validation (bool): If True, skip validation when asked to. Defaults to False.
    """

    def __init__(
        self,
        entry_id,
        doc_id,
        lang=None,
        sent_id=None,
        segment_type=None,
        segment_text=None,
        segment_sections=None,
        start_in_source=None,
        end_in_source=None,
        segment_start=None,
        segment_end=None,
        annotation_sets=None,
        provenance=None,
        no_validation=False
    ):

        # required
        self.entry_id = BPProperty(
            "entry-id",
            True,
            value=entry_id,
            type_=str,
            no_validation=no_validation
        )
        self.doc_id = BPProperty(
            "doc-id",
            True,
            value=doc_id,
            type_=str,
            no_validation=no_validation
        )

        # optional
        self.lang = BPProperty(
            "lang",
            False,
            value=lang,
            type_=str,
            no_validation=no_validation
        )
        self.segment_type = BPProperty(
            "segment-type",
            False,
            value=segment_type,
            type_=str,
            choices=["sentence", "document",],
            no_validation=no_validation
        )
        self.sent_id = BPProperty(
            "sent-id",
            segment_type == "sentence",
            value=sent_id,
            type_=str,
            no_validation=no_validation
        )
        self.segment_text = BPProperty(
            "segment-text",
            False,
            value=segment_text,
            type_=str,
            no_validation=no_validation
        )
        self.segment_sections = BPProperty(
            "segment-sections",
            False,
            value=segment_sections,
            type_=list,
            no_validation=no_validation
        )
        self.start_in_source = BPProperty(
            "start-in-source",
            False,
            value=start_in_source,
            type_=int,
            no_validation=no_validation
        )
        self.end_in_source = BPProperty(
            "end-in-source",
            False,
            value=end_in_source,
            type_=int,
            no_validation=no_validation
        )
        self.segment_start = BPProperty(
            "segment-start",
            False,
            value=segment_start,
            type_=int,
            no_validation=no_validation
        )
        self.segment_end = BPProperty(
            "segment-end",
            False,
            value=segment_end,
            type_=int,
            no_validation=no_validation
        )
        self.annotation_sets = BPProperty(
            "annotation-sets",
            False,
            value=annotation_sets,
            type_=BPDocumentAnnotationSet,
            no_validation=no_validation
        )
        self.provenance = BPProperty(
            "provenance",
            False,
            value=provenance,
            type_=BPDocumentProvenance,
            no_validation=no_validation
        )
        self.no_validation = no_validation


    def __lt__(self, other):
        return (
            self.entry_id,
            self.doc_id,
            self.lang,
            self.sent_id,
            self.start_in_source,
            self.end_in_source,
            self.annotation_sets
        ) < (
            other.entry_id,
            other.doc_id,
            other.lang,
            other.sent_id,
            other.start_in_source,
            other.end_in_source,
            other.annotation_sets
        )


    def __repr__(self):
        return (
            f"BPDocumentEntry({self.entry_id.value}, {self.doc_id.value}, "
            f"lang='{self.lang.value}', "
            f"sent_id='{self.sent_id.value}', segment_type='{self.segment_type.value}', "
            f"segment_text='{self.segment_text.value}', "
            f"segment_sections={repr(self.segment_sections.value)}, "
            f"start_in_source={self.start_in_source.value}, "
            f"end_in_source={self.end_in_source.value}, "
            f"segment_start={self.segment_start.value}, segment_end={self.segment_end.value}, "
            f"annotation_sets={repr(self.annotation_sets.value)}, "
            f"provenance={repr(self.provenance.value)}, "
            f"no_validation={self.no_validation})"
        )


    def __str__(self):
        return f"BPDocumentEntry<{self.entry_id.value} ({id(self)})>"


    def deidentify(self):
        """Generates a UUID from this Entry's segment-text, to obfuscate where the document came
        from.

        Returns:
            str: a UUID keyed to this document's segment-text.
        """
        import uuid

        self.doc_id.value = str(uuid.uuid5(uuid.NAMESPACE_DNS, self.doc_id.value))
        self.entry_id.value = str(uuid.uuid5(uuid.NAMESPACE_DNS, self.entry_id.value))

        self.provenance.value = None

        if self.annotation_sets.value:
            self.annotation_sets.value.deidentify()

        return self.entry_id.value


    @classmethod
    def from_dict(klass, dict_, no_validation=False):
        """Constructs an instance of BPDocumentEntry from a Python dictionary following the BP
        schema for Entry objects.

        Args:
            dict_ (dict): a Python dictionary following the BP schema for Entry objects.
            no_validation (bool): If True, construct the object with validation turned off.
                Defaults to False.

        Returns:
            BPDocumentEntry: a new instance of this class built with the dictionary.

        .. Strategy:
           - primitives can be directly looked up and passed in from ``dict_``
           - collections (dicts, lists) and non-primitive types (like other ``BPDocument<type>``
             types defined in this library) must be constructed first if data is present in
             ``dict_``.
        """

        global LOGGER

        required = ["entry-id", "doc-id",]
        for key in required:
            assert key in dict_, f"Entry object dictionary is missing the '{key}' key."

        ## build non-primitive types and collections

        if "segment-sections" in dict_:
            segment_sections = []
            for element in dict_["segment-sections"]:
                segment_sections.append(BPDocumentSegmentSection.from_dict(
                    element,
                    no_validation=no_validation
                ))
        else:
            segment_sections = None

        if "annotation-sets" in dict_:
            annotation_sets = BPDocumentAnnotationSet.from_dict(
                dict_["annotation-sets"],
                no_validation=no_validation
            )
        else:
            annotation_sets = None

        if "provenance" in dict_:
            provenance = BPDocumentProvenance.from_dict(
                dict_["provenance"],
                no_validation=no_validation
            )
        else:
            provenance = None

        ## build new object

        obj = klass(
            dict_['entry-id'],
            dict_['doc-id'],
            lang=dict_.get('lang', None),
            sent_id=dict_.get('sent-id', None),
            segment_type=dict_.get('segment-type', None),
            segment_text=dict_.get('segment-text', None),
            segment_sections=segment_sections or None,
            start_in_source=dict_.get('start-in-source', None),
            end_in_source=dict_.get('end-in-source', None),
            segment_start=dict_.get('segment-start', None),
            segment_end=dict_.get('segment-end', None),
            annotation_sets=annotation_sets,
            provenance=provenance,
            no_validation=no_validation
        )

        if not no_validation:
            assert obj.is_valid(), "Entry object built from dictionary failed validation."

        return obj


    def get_event_set(self, dialect):
        """Gets the BPDocument<dialect>EventSet object from this document, or ``None`` if it doesn't
        exist.
        """

        from lib.utils import MAT_DIALECT_CHOICES

        assert dialect in MAT_DIALECT_CHOICES, f"{dialect} is not one of {MAT_DIALECT_CHOICES}!"

        if self.annotation_sets.value:
            if dialect == "Abstract":
                return self.annotation_sets.value.abstract_events.value
            elif dialect == "Basic" or dialect == "Granular":
                return self.annotation_sets.value.basic_events.value
        else:
            return None


    def is_valid(self):
        """Checks if this object (and its constituents) are all valid."""

        import pycountry

        global LOGGER

        if self.no_validation:
            raise RuntimeError(f"Validation set to False: {self}")

        def __set_helper(event_set):

            if event_set.span_sets.value and not self.segment_text.value:
                LOGGER.error((
                    f"BPDocumentEntry's {type(event_set).__name__} has defined spans but no"
                    "segment-text!"
                ))
                return False

            for ssid in event_set.span_sets.value:
                span_set = event_set.span_sets.value[ssid]

                if not span_set.is_valid():
                    return False

                for span in span_set.spans.value:

                    if span.start.value and span.end.value:

                        substring = self.segment_text.value[span.start.value:span.end.value]
                        if substring != span.string_.value:
                            LOGGER.error((
                                f"BPDocumentEntry string '{span.string_.value}' in span {ssid} in "
                                f"{type(event_set).__name__} does not match the substring in "
                                f"segment-text bounded by indices [{span.start.value},"
                                f"{span.end.value}]: {substring}"
                            ))
                            return False

                    else:

                        if self.segment_text.value.find(span.string_.value) == -1:

                            if len(self.segment_text.value) > 100:
                                LOGGER.error((
                                    f"BPDocumentEntry string '{span.string_.value}' in span {ssid} "
                                    f"in {type(event_set).__name__} can't be found within the "
                                    f"segment-text '{self.segment_text.value[:100]}...' "
                                    "(truncated)!"
                                ))
                            else:
                                LOGGER.error((
                                    f"BPDocumentEntry string '{span.string_.value}' in span {ssid} "
                                    f"in {type(event_set).__name__} can't be found within the "
                                    f"segment-text '{self.segment_text.value}'!"
                                ))

                            return False

                    if span.hstring.value:

                        if span.hstart.value and span.hend.value:

                            substring = self.segment_text.value[span.hstart.value:span.hend.value]
                            if substring != span.hstring.value:
                                LOGGER.error((
                                    f"BPDocumentEntry hstring '{span.hstring.value}' in span {ssid}"
                                    f" in {type(event_set).__name__} does not match the substring "
                                    f"in segment-text bounded by indices [{span.hstart.value},"
                                    f"{span.hend.value}]: {substring}"
                                ))
                                return False

                        else:

                            if self.segment_text.value.find(span.hstring.value) == -1:

                                if len(self.segment_text.value) > 100:
                                    LOGGER.error((
                                        f"BPDocumentEntry hstring '{span.hstring.value}' in span "
                                        f"{ssid} in {type(event_set).__name__} can't be found "
                                        f"within the segment-text '{self.segment_text.value[:100]}"
                                        "...' (truncated)!"
                                    ))
                                else:
                                    LOGGER.error((
                                        f"BPDocumentEntry hstring '{span.hstring.value}' in span "
                                        f"{ssid} in {type(event_set).__name__} can't be found "
                                        f"within the segment-text '{self.segment_text.value}'!"
                                    ))
                                return False

            return True

        for prop in [
            self.entry_id,
            self.doc_id,
            self.lang,
            self.sent_id,
            self.segment_type,
            self.segment_text,
            self.segment_sections,
            self.start_in_source,
            self.end_in_source,
            self.segment_start,
            self.segment_end,
            self.annotation_sets,
            self.provenance,
        ]:

            if not prop.is_valid():
                LOGGER.error(f"BPDocumentEntry property {prop} is invalid.")
                return False

        if self.lang.value and not pycountry.languages.get(alpha_3=self.lang.value):
            LOGGER.error(
                f"BPDocumentEntry language '{self.lang.value}' is not a value ISO-639-3 code!"
            )
            return False

        if self.segment_type.value and self.segment_type.value == "sentence":
            if not self.sent_id.value:
                LOGGER.error((
                    f"BPDocumentEntry segment_type is 'sentence' but sent_id is missing!"
                ))
                return False

        if self.annotation_sets.value:

            if not self.annotation_sets.value.is_valid():
                return False

            if self.annotation_sets.value.abstract_events.value:
                if not __set_helper(self.annotation_sets.value.abstract_events.value):
                    LOGGER.error(f"In: {str(self)}")
                    return False

            if self.annotation_sets.value.basic_events.value:
                if not __set_helper(self.annotation_sets.value.basic_events.value):
                    LOGGER.error(f"In: {str(self)}")
                    return False

        if self.segment_sections.value:
            for element in self.segment_sections.value:
                if type(element) is not BPDocumentSegmentSection:
                    i = self.segment_sections.value.index(element)
                    LOGGER.error((
                        f"BPDocumentEntry element at index {i} of segment_sections is not of type "
                        f"BPDocumentSegmentSection: {type(element)}"
                    ))
                    return False

        return True


    def set_validation(self, validation):
        """Sets the validation mode for this node and all children of this node.

        Args:
            validation (bool): if True, sets ``no_validation`` for this node and all of its children
                to ``False``. Otherwise, all nodes set ``no_validation`` to ``True``.
        """

        assert type(validation) == bool

        self.no_validation = not validation

        self.entry_id.no_validation = self.no_validation
        self.doc_id.no_validation = self.no_validation
        self.lang.no_validation=self.no_validation
        self.segment_type.no_validation = self.no_validation
        self.sent_id.no_validation = self.no_validation
        self.segment_text.no_validation = self.no_validation
        self.segment_sections.no_validation = self.no_validation
        self.start_in_source.no_validation = self.no_validation
        self.end_in_source.no_validation = self.no_validation
        self.segment_start.no_validation = self.no_validation
        self.segment_end.no_validation = self.no_validation
        self.annotation_sets.no_validation = self.no_validation
        self.provenance.no_validation = self.no_validation

        ## recurse

        if self.provenance.value:
            self.provenance.value.set_validation(validation)

        if self.segment_sections.value:
            for segsec in self.segment_sections.value:
                segsec.set_validation(validation)

        if self.annotation_sets.value:
            self.annotation_sets.value.set_validation(validation)


    def to_dict(self):
        """Constructs a Python dictionary following the BP schema, suitable for JSON serialization.

        Returns:
            dict: a dictionary representing this BP Entry.
        """

        if not self.no_validation:
            assert self.is_valid(), "Annotation Set object is invalid."

        dict_ = {
            self.entry_id.name: self.entry_id.value,
            self.doc_id.name: self.doc_id.value,
        }

        for prop in [
            self.lang, 
            self.sent_id,
            self.segment_type,
            self.segment_text,
            self.start_in_source,
            self.end_in_source,
            self.segment_start,
            self.segment_end,
        ]:
            if prop.value is not None:
                dict_[prop.name] = prop.value

        if self.segment_sections.value is not None:
            dict_[self.segment_sections.name] = [
                section.to_dict() for section in self.segment_sections.value
            ]

        if self.annotation_sets.value is not None:
            dict_[self.annotation_sets.name] = self.annotation_sets.value.to_dict()

        if self.provenance.value is not None:
            dict_[self.provenance.name] = self.provenance.value.to_dict()

        return dict_


class BPDocument(BPNode):
    """Represents a BETTER Project document.

    Args:
        corpus_id (str): passed when constructing new documents. The unique ID of the corpus being
            constructed. Required for new document construction.
        entries (dict): a dictionary keyed with Entry IDs (strings) whose values are the
            corresponding :class:`BPDocumentEntry` objects.
        provenance (BPDocumentProvenance): a :class:`BPDocumentProvenance` object pertaining to the
            whole document.
        no_validation (bool): If True, skip validation when asked to. Defaults to False.
    """

    def __init__(self, corpus_id, entries, provenance=None, no_validation=False):

        # passed-in values
        self.corpus_id = BPProperty(
            "corpus-id",
            True,
            value=corpus_id,
            type_=str,
            no_validation=no_validation
        )
        self.entries = BPProperty(
            "entries",
            True,
            value=entries,
            type_=dict,
            no_validation=no_validation
        )
        self.provenance = BPProperty(
            "provenance",
            False,
            value=provenance,
            type_=BPDocumentProvenance,
            no_validation=no_validation
        )
        self.no_validation = no_validation

        # static values (no longer passed in)
        self.format_type = BPProperty(
            "format-type",
            True,
            value="bp-corpus",
            type_=str,
            no_validation=no_validation
        )
        self.format_version = BPProperty(
            "format-version",
            True,
            value="v10",
            type_=str,
            default="v10",
            choices=["v10", "v9", "v9a",],
            no_validation=no_validation
        )


    def __lt__(self, other):
        return (self.corpus_id, self.entries) < (other.corpus_id, other.entries)


    def __repr__(self):
        return (
            f"BPDocument(corpus_id={repr(self.corpus_id)}, entries={repr(self.entries)}), "
            f"provenance={repr(self.provenance.value)}, no_validation={self.no_validation})"
        )


    def __str__(self):
        return f"BPDocument<{self.corpus_id.value} ({id(self)})>"


    def check_sentence_puncts(self):
        """Tries to ensure that all sentence segment sections include a trailing punctuation
        mark.

        See bp_lib::Issue#31.
        """

        from lib.utils import PUNCTUATION

        def __signal(entry, segment_section):
            start = segment_section.start.value
            end = segment_section.end.value
            return entry.segment_text.value[start:end]

        if self.entries.value:

            for enid in self.entries.value:

                entry = self.entries.value[enid]
                if entry and entry.segment_text.value:

                    segment_sections = entry.segment_sections.value

                    if segment_sections is not None:
                        for segment_section in segment_sections:
                            if segment_section.structural_element.value == "Sentence" and \
                                segment_section.start.value is not None and \
                                segment_section.end.value is not None:

                                # if the last char isn't a punct, extend sentence tag range by 1
                                if __signal(entry, segment_section)[-1] not in PUNCTUATION:
                                    segment_section.end.value += 1

                                # if the last char _still_ isn't a punct, revert
                                if __signal(entry, segment_section)[-1] not in PUNCTUATION:
                                    segment_section.end.value -= 1


    def dedupe_args(self):
        """Removes duplicate SSIDs and EventIDs from Event arguments."""

        def __dedupe(arg: list) -> list:
            """Preserve order when deduping."""
            seen = set()
            result = []
            for item in arg:
                if item not in seen:
                    seen.add(item)
                    result.append(item)
            return result

        if self.entries.value:

            for enid in self.entries.value:

                entry = self.entries.value[enid]
                if entry:

                    annot_sets = entry.annotation_sets.value
                    if annot_sets:

                        abstract_event_set = annot_sets.abstract_events.value
                        if abstract_event_set:

                            for event_id, event in abstract_event_set.events.value.items():

                                event.agents.value = __dedupe(event.agents.value)
                                event.patients.value = __dedupe(event.patients.value)

                        basic_event_set = annot_sets.basic_events.value
                        if basic_event_set:

                            for event_id, event in basic_event_set.events.value.items():

                                event.agents.value = __dedupe(event.agents.value)
                                event.patients.value = __dedupe(event.patients.value)
                                event.ref_events.value = __dedupe(event.ref_events.value)
                                event.money.value = __dedupe(event.money.value)


    def deidentify(self):
        """Obfuscates this Document's source information.

        Provenance objects are cleared and IDs are obfuscated.
        """

        new_entries = {}
        for enid in self.entries.value:
            entry = self.entries.value[enid]

            new_entries[entry.deidentify()] = entry

        self.provenance.value = None
        self.corpus_id.value += " (Obfuscated)"
        self.entries.value = new_entries


    @staticmethod
    def dict_is_schema_valid(doc, verbose=False):
        """Validates a Python dictionary representing BP JSON against the standard schema.

        Only the current version of the BP schema is considered.

        Args:
            doc (dict): a dictionary suitable for JSON serialization, presumably created using
                ``utils.open_json_blob()``.
            verbose (bool): print verbose output if True.

        Returns:
            bool: True if the document conforms to the JSON schema.
        """

        import os

        from jsonschema import validate
        from jsonschema.exceptions import ValidationError

        from lib.utils import BASEPATH
        from lib.utils import open_json_blob

        global LOGGER

        schema = open_json_blob(os.path.join(BASEPATH, "schema", "bp.schema.json"))

        try:

            validate(doc, schema)

        except ValidationError as err:

            if verbose:
                LOGGER.error(err)
            else:
                LOGGER.error(str(err)[:str(err).find("\n")])
            return False

        return True


    @classmethod
    def from_dict(klass, dict_, no_validation=False):
        """Constructs an instance of BPDocument from a Python dictionary following the BP schema for
        whole-Document objects.

        Args:
            dict_ (dict): a Python dictionary following the BP schema for Entry objects.
            no_validation (bool): If True, construct the object with validation turned off.
                Defaults to False.

        Returns:
            BPDocument: a new instance of this class built with the dictionary.

        .. Strategy:
           - primitives can be directly looked up and passed in from ``dict_``
           - collections (dicts, lists) and non-primitive types (like other ``BPDocument<type>``
             types defined in this library) must be constructed first if data is present in
             ``dict_``.
        """

        global LOGGER

        if not no_validation:
            assert klass.dict_is_schema_valid(dict_), f"Document object dictionary is invalid."

            required = ["format-type", "format-version", "corpus-id", "entries",]
            for key in required:
                assert key in dict_, f"Document object dictionary is missing the '{key}' key."

        ## build non-primitive types and collections

        entries = {}
        for entry_id in dict_['entries']:
            entries[entry_id] = BPDocumentEntry.from_dict(dict_["entries"][entry_id], no_validation)

        if "provenance" in dict_:
            provenance = BPDocumentProvenance.from_dict(dict_["provenance"], no_validation)
        else:
            provenance = None

        ## build new object

        obj = klass(
            dict_['corpus-id'],
            entries,
            provenance=provenance,
            no_validation=no_validation
        )

        if not no_validation:
            assert obj.is_valid(), "Document object built from dictionary failed validation."

        return obj


    @classmethod
    def from_json(klass, path=None, no_validation=False):
        """
        Constructs an instance of BPDocument from the contents of a JSON blob at the specified path
        following the BP schema for whole-document BP objects.

        Args:
            path (str): the path to a BP JSON document used to populate this object's values, or
                ``None`` if no document was read in. Not used when creating new documents.
            no_validation (bool): If True, construct the object with validation turned off.
                Defaults to False.

        Returns:
            BPDocument: a new instance of this class built with the contents of the JSON blob at the
            input path.
        """
        import codecs
        import json

        from lib.utils import open_json_blob

        return klass.from_dict(open_json_blob(path), no_validation)


    def is_valid(self):
        """Checks if this object (and its constituents) are all valid."""

        global LOGGER

        if self.no_validation:
            raise RuntimeError(f"Validation set to False: {self}")

        for prop in [self.format_type, self.format_version, self.corpus_id, self.provenance,]:

            if not prop.is_valid():
                LOGGER.error(f"BPDocument property {prop} is invalid.")
                return False

        for entry_id in self.entries.value:

            if type(entry_id) is not str:
                LOGGER.error(f"BPDocument entry_id is not of type string: {entry_id}")
                return False

            if type(self.entries.value[entry_id]) is not BPDocumentEntry:
                LOGGER.error((
                    "BPDocument entry is not of type BPDocumentEntry: "
                    f"{type(self.entries.value[entry_id])}"
                ))
                return False

            if not self.entries.value[entry_id].is_valid():
                return False

        return True


    def remove_annotation_sets(self):
        """Removes the annotation sets from the document."""

        if self.entries.value:
            for enid in self.entries.value:
                entry = self.entries.value[enid]
                if entry:
                    entry.annotation_sets.value = None


    def serialize(self, path, flat=False, verbose=False, force=True):
        """Serializes this BP Document object (and its constituents) to a JSON blob.

        Args:
            path (str): a path to write the JSON blob.
            flat (bool): if True, write the JSON blob to a single line, instead of indenting it.
            verbose (bool): print verbose output if True.
            force (bool): noninteractive if True, assume it's OK to overwrite existing files.
                Assumed default behavior for non-interactive procedures (i.e., if this wasn't called
                as the result of a CLI script).
        """

        from lib.utils import write_json_blob

        dict_ = self.to_dict()
        if not self.no_validation:
            assert self.dict_is_schema_valid(dict_), "Constructed BP document is invalid."

        if flat:
            write_json_blob(dict_, path, indent=0, verbose=verbose, force=force)
        else:
            write_json_blob(dict_, path, verbose=verbose, force=force)


    def set_validation(self, validation):
        """Sets the validation mode for all nodes of the document.

        Args:
            validation (bool): if True, sets ``no_validation`` for all nodes in the Document tree to
                ``False``. Otherwise, all nodes set ``no_validation`` to ``True``.
        """

        assert type(validation) == bool

        self.no_validation = not validation

        self.corpus_id.no_validation = self.no_validation
        self.entries.no_validation = self.no_validation
        self.provenance.no_validation = self.no_validation
        self.format_type.no_validation = self.no_validation
        self.format_version.no_validation = self.no_validation

        ## recurse

        if self.provenance.value:
            self.provenance.value.set_validation(validation)

        if self.entries.value:
            for enid in self.entries.value:
                self.entries.value[enid].set_validation(validation)


    def stats(
        self,
        pretty=False,
        abstract_events=False,
        basic_events=False,
        granular_templates=False,
        filter_by_annotator_class=None,
        filter_by_dataset_label=None
    ):
        """Gathers statistics about this BPDocument.

        Args:
            pretty (bool): if True, build pretty-printable statistics instead.
            abstract_events (bool): if True, count how many types of each Abstract Event type there
                are.
            basic_events (bool): if True, count how many types of each Basic Event type there
                are.
            granular_templates (bool): if True, count how many types of each Granular Template type
                there are.
            filter_by_annotator_class (str): if passed, only consider entries whose Provenance in
                the reference matches the given annotator class.
            filter_by_dataset_label (str): Only consider entries whose Provenance in the reference
                matches the given dataset label.

        Returns:
            dict: a collection of statistics. If ``pretty``, a string of statistics instead.
        """
        import re

        from tabulate import tabulate

        from lib.utils import BASIC_MAT_KEY_ANCHOR_TYPES
        from lib.utils import BP_ANNOTATOR_CLASS_TYPES
        from lib.utils import GRANULAR_MAT_TEMPLATE_TYPES

        if filter_by_dataset_label:
            assert type(filter_by_dataset_label) == str, "Filters must be strings."
        if filter_by_annotator_class:
            assert type(filter_by_annotator_class) == str, "Filters must be strings."
            assert filter_by_annotator_class in BP_ANNOTATOR_CLASS_TYPES, \
                f"Bad annotator class: '{filter_by_annotator_class}'"

        abstract_types = [
            ("helpful", "material",),
            ("helpful", "verbal",),
            ("helpful", "both",),
            ("helpful", "unk",),
            ("harmful", "material",),
            ("harmful", "verbal",),
            ("harmful", "both",),
            ("harmful", "unk",),
            ("neutral", "material",),
            ("neutral", "verbal",),
            ("neutral", "both",),
            ("neutral", "unk"),
        ]

        data = {
            "Entries": 0,
            "Filtered Entries": 0,
            "Tagged Sentences": 0,
            "Abstract Events": 0,
            "Abstract Span Sets": 0,
            "Abstract Span Sets w/Multiple Spans": 0,
            "Abstract Includes Relations": 0,
            "Basic Events": 0,
            "Basic Span Sets": 0,
            "Basic Span Sets w/Multiple Spans": 0,
            "Basic Includes Relations": 0,
            "Granular Templates": 0,
        }

        if abstract_events:
            for key in abstract_types:
                data[key] = 0

        if basic_events:
            for key in BASIC_MAT_KEY_ANCHOR_TYPES:
                data[key] = 0

        if granular_templates:
            for key in GRANULAR_MAT_TEMPLATE_TYPES:
                data[key] = 0

        if self.entries.value:

            for enid in self.entries.value:

                data["Entries"] += 1

                entry = self.entries.value[enid]

                if entry:

                    if filter_by_annotator_class:

                        provenance = entry.provenance.value
                        if not provenance:
                            continue

                        if provenance.annotator_class.value != filter_by_annotator_class:
                            continue

                    if filter_by_dataset_label:

                        provenance = entry.provenance.value
                        if not provenance:
                            continue

                        if "dataset-label" not in provenance.misc_keys:
                            continue

                        if provenance.misc_keys["dataset-label"] != filter_by_dataset_label:
                            continue

                    data["Filtered Entries"] += 1

                    segment_sections = entry.segment_sections.value

                    if segment_sections:
                        for segment_section in segment_sections:
                            if segment_section.structural_element.value == "Sentence":
                                data["Tagged Sentences"] += 1

                    annot_sets = entry.annotation_sets.value
                    if annot_sets:

                        abstract_event_set = annot_sets.abstract_events.value
                        if abstract_event_set:

                            events = abstract_event_set.events.value
                            if events:
                                data["Abstract Events"] += len(events.keys())

                                if abstract_events:
                                    for evid in events:
                                        event = events[evid]
                                        if event:
                                            hh = event.helpful_harmful.value
                                            mv = event.material_verbal.value
                                            if hh and mv:
                                                data[(hh, mv)] += 1

                            span_sets = abstract_event_set.span_sets.value
                            if span_sets:
                                data["Abstract Span Sets"] += len(span_sets.keys())
                                for ssid in span_sets:
                                    span_set = span_sets[ssid]
                                    if len(span_set.spans.value) > 1:
                                        data["Abstract Span Sets w/Multiple Spans"] += 1

                            includes_relations = abstract_event_set.includes_relations.value
                            if includes_relations:
                                data["Abstract Includes Relations"] += len(
                                    includes_relations.keys()
                                )

                        basic_event_set = annot_sets.basic_events.value
                        if basic_event_set:

                            events = basic_event_set.events.value
                            if events:
                                data["Basic Events"] += len(events.keys())

                                if basic_events:
                                    for evid in events:
                                        event = events[evid]
                                        if event:
                                            event_type = event.event_type.value
                                            if event_type:
                                                data[event_type] += 1

                            span_sets = basic_event_set.span_sets.value
                            if span_sets:
                                data["Basic Span Sets"] += len(span_sets.keys())
                                for ssid in span_sets:
                                    span_set = span_sets[ssid]
                                    if len(span_set.spans.value) > 1:
                                        data["Basic Span Sets w/Multiple Spans"] += 1

                            includes_relations = basic_event_set.includes_relations.value
                            if includes_relations:
                                data["Basic Includes Relations"] += len(includes_relations.keys())

                            gran_temps = basic_event_set.granular_templates.value
                            if gran_temps:
                                data["Granular Templates"] += len(gran_temps.keys())

                                if granular_templates:
                                    for gtid in gran_temps:
                                        gt = gran_temps[gtid]
                                        if gt:
                                            gt_type = gt.type_.value
                                            if gt_type:
                                                data[gt_type] += 1

        if pretty:

            data = tabulate(
                [[k, v,] for k, v in data.items()],
                headers=["Stat", "Count"],
                tablefmt="fancy_grid"
            )

        return data


    def to_dict(self):
        """Constructs a Python dictionary following the BP schema, suitable for JSON serialization.

        Returns:
            dict: a dictionary representing this BP document.
        """

        if not self.no_validation:
            assert self.is_valid(), "Document object is invalid."

        dict_ = {
            "entries": {},
        }

        for prop in [self.format_type, self.format_version, self.corpus_id,]:

            if prop.value is not None:
                dict_[prop.name] = prop.value

        if self.provenance.value is not None:
            dict_[self.provenance.name] = self.provenance.value.to_dict()

        for entry_id in self.entries.value:
            dict_[self.entries.name][entry_id] = self.entries.value[entry_id].to_dict()

        return dict_


    def trim_spans(self):
        """Trims all span objects in the document."""

        if self.entries.value:

            for enid in self.entries.value:

                entry = self.entries.value[enid]
                if entry:

                    annot_sets = entry.annotation_sets.value
                    if annot_sets:

                        abstract_event_set = annot_sets.abstract_events.value
                        if abstract_event_set:

                            span_sets = abstract_event_set.span_sets.value
                            if span_sets:
                                for ssid in span_sets:
                                    span_set = span_sets[ssid]
                                    for span in span_set.spans.value:
                                        span.trim()

                        basic_event_set = annot_sets.basic_events.value
                        if basic_event_set:

                            span_sets = basic_event_set.span_sets.value
                            if span_sets:
                                for ssid in span_sets:
                                    span_set = span_sets[ssid]
                                    for span in span_set.spans.value:
                                        span.trim()


